//Use this method to analyze the cap-trap data sets. This can be done
//with respect to different features, such as TSSs or enhancers

//Example:

//java -Xmx512m AnalyzeCapTrap -R mRNA/CapTrap/hr1_capTrap.tsv -I
//
//For enhancers
//
//java -Xmx512m AnalyzeCapTrap

//Use analyze_cap_trap.m for visualizing and post-processing

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class AnalyzeCapTrap {

    public static final String CAP_TRAP_ENDS_FILE_NAME = "cap_trap_ends";

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option rnaSeqFileNameFlag = argParser.addStringOption('R', "RNASeq");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('a', "saveFileName");
	    //use this flag when we only want to focus on features
	    //which are guaranteed to be isolated from one another.
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('I', "isolated");
	    //Use this flag to include a file with genes/peaks to filter against
	    CmdLineParser.Option selectFlag = argParser.addStringOption('s', "select");
	    //Use this flag when we have other species
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    //Use this flag if we want to center around enhancers instead of TSSs
	    CmdLineParser.Option enhancerFlag = argParser.addStringOption("enhancer");
	    //Different time points
	    CmdLineParser.Option timeFlag = argParser.addIntegerOption('t', "time");
	    CmdLineParser.Option upstreamFlag = argParser.addIntegerOption('u', "upstream");
	    CmdLineParser.Option downstreamFlag = argParser.addIntegerOption('d', "downstream");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(enhancerFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(timeFlag);
	    int time = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    tmp = argParser.getOptionValue(upstreamFlag);
	    int upstream = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(downstreamFlag);
	    int downstream = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(rnaSeqFileNameFlag);
	    String rnaSeqFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/hr" + time + ".tsv";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = null;
	    if (tmp!=null)
		saveFileName = (String)tmp;
	    else {
		if (enhancerFileName!=null) {
		    Pattern enhancerPattern = Pattern.compile("enhancers_([_|\\w]+)\\.txt");
		    Matcher m = enhancerPattern.matcher(enhancerFileName);
		    String stimulus = m.find() ? m.group(1) : "";
		    //saveFileName = (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/Enhancers/" + EALR_LENGTH_FILE_NAME + "_hr" + time + (shortRNA ? "_short" : "") + (mRNA ? "_mrna" : "") + (isolated ? "_isolated" : "");
		}
		else
		    saveFileName = (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/CapTrap/" + CAP_TRAP_ENDS_FILE_NAME + (isolated ? "_isolated" : "");
	    }
	    tmp = argParser.getOptionValue(selectFlag);
    	    String selectFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(speciesFlag);
    	    String species = tmp!=null ? (String)tmp : "Mm";
	    /////////////////////////////////
	    int nChr = 21;
	    if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
		nChr = 24;
	    	    int isolatedIDs[] = null;
	    if (isolated) {
		if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
		    isolatedIDs = FindReadsAcrossFeature.parseIsolatedFile((orchestra ? "/groups/neuroduo/martin/" : "/media/exthd/") + "Genomes/Hs/v36/isolated_genes.txt");
		else
		    isolatedIDs = FindReadsAcrossFeature.parseIsolatedFile(orchestra);
	    }
	    int geneIDsKeep[] = FindReadsAcrossFeature.parseSelectFile(selectFileName);
	    if (geneIDsKeep!=null) 
		Arrays.sort(geneIDsKeep);
	    Vector<Enhancer> enhancers = enhancerFileName==null ? null : FindReadsAcrossFeature.parseEnhancerFile(enhancerFileName);
	    int nReadsForward = 0, nReadsReverse = 0, nDiagonal = 0, nDiagonalRandom = 0;
	    TreeMap<Integer, Integer> senseAntisense5pEnds = new TreeMap<Integer, Integer>();
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    RefSeqGenes geneInfo = new RefSeqGenes(i, orchestra, species);
		    //Load all the RNA reads
		    ABIRead[] forwardReads = FindNongenicTranscripts.loadMRNAReadsChr(rnaSeqFileName, true, i);
		    nReadsForward += forwardReads.length;
		    ABIRead[] reverseReads = FindNongenicTranscripts.loadMRNAReadsChr(rnaSeqFileName, false, i);
		    nReadsReverse += reverseReads.length;
		    System.out.println("Loaded reads and annotation for chr " + i);
		    /////////////////////////////////
		    for (RefSeqGene g : geneInfo.getGenes()) {
			//Get all the reads near the TSS
			if (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) {
			    if (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0) {
				int start, end, tss = g.getStart();
				ABIRead[] senseReads = null, antisenseReads = null;
				if (g.getStrand()) {
				    senseReads = FindReadsAcrossFeature.getReads(i, tss, downstream, upstream, forwardReads);
				    antisenseReads = FindReadsAcrossFeature.getReads(i, tss, downstream, upstream, reverseReads);
				}
				else {
				    senseReads = FindReadsAcrossFeature.getReads(i, tss, downstream, upstream, reverseReads);
				    antisenseReads = FindReadsAcrossFeature.getReads(i, tss, downstream, upstream, forwardReads);
				}
				//Draw one read at random from each set
				if (senseReads.length>0 && antisenseReads.length>0) {
				    /*System.out.println("Found " + senseReads.length + " sense and " + antisenseReads.length + " antisense reads near " + g.toString());
				    if (g.getStrand())
					System.out.println("Sense read is at " + senseReads[0].getPosition() + ", rel " + (senseReads[0].getPosition() - g.getStart()) + " and AS " + antisenseReads[0].getPosition() + ", " + (antisenseReads[0].getPosition() - g.getStart()));
				    else
				    System.out.println("Sense read is at " + senseReads[0].getPosition() + ", rel " + (g.getStart() - senseReads[0].getPosition()) + " and AS " + antisenseReads[0].getPosition() + ", " + (g.getStart() - antisenseReads[0].getPosition()));*/
				    if (g.getStrand())
					senseAntisense5pEnds.put(senseReads[random.nextInt(senseReads.length)].getPosition() - g.getStart(), antisenseReads[random.nextInt(antisenseReads.length)].getPosition() - g.getStart());
				    else
					senseAntisense5pEnds.put(g.getStart() - senseReads[random.nextInt(senseReads.length)].getPosition(), g.getStart() - antisenseReads[random.nextInt(antisenseReads.length)].getPosition());
				    //Check if there are two reads which appear to be in the same position which would indicate a good diagonal
				    RefSeqGene gr = geneInfo.getRandomGene(random);
				    ABIRead randomSenseReads[] =  gr.getStrand() ? FindReadsAcrossFeature.getReads(i, tss, downstream, upstream, forwardReads) : FindReadsAcrossFeature.getReads(i, tss, downstream, upstream, reverseReads);
				    for (int j=0; j<antisenseReads.length; j++) {
					for (int k=0; k<senseReads.length; k++) {
					    if (Math.abs(antisenseReads[j].getPosition() - senseReads[k].getPosition())<120) {
						nDiagonal++;
						break; 
					    }
					}
					for (int k=0; k<randomSenseReads.length; k++) {
					    if (Math.abs(antisenseReads[j].getPosition() - randomSenseReads[k].getPosition())<120) {
						nDiagonalRandom++;
						break; 
					    }
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	    System.out.println("There were " + nDiagonal + " promoters with sense and antisense 5'-ends nearby and in the random control there were " + nDiagonalRandom);
	    //Save all of them in files
	    String saveName = "";
	    if (enhancerFileName!=null)
		saveName = saveFileName + "_forward" + (chr!=-1 ? "_" + chr : "") + ".txt";
	    else
		saveName = saveFileName + "_sense" + (chr!=-1 ? "_" + chr : "") + ".txt";
	    System.out.println("Saving to " + saveName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	    for (int s : senseAntisense5pEnds.keySet())
		bw.write(s + "\t" + senseAntisense5pEnds.get(s) + "\n");
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////


}
