//Use this class to analyze the controls from the Mef2
//experiment. This class is just a wrapper which calls on other
//classes to do the heavy lifting. Write output to files to be read
//and analyzed in Matlab. Due to the heavy memory-requirements, it
//cannot be easily launched from inside Matlab but must be started in
//a more pedestrian manner.
//
//Examples of how to use this class (should probably be split into
//separate ones later on):
//
//java AnalyzeControls -c 12 -n Mef2/KCl/mef2KCl
//
//Analyze the distribution of cluster-sizes with respect to a ChIP experiment
//
//java AnalyzeControls -c 12 -i Mef2/inputKCl/input -S -n Mef2/KCl/mef2KCl -u

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class AnalyzeControls{

    public static String GENOME_LOC = "/media/exthd/Genomes/Mm/v9/";
    public static int READ_LENGTH = 35;

    protected static int nMismatches, minDist, minReads, fragmentLength, promoterLength, nChr, windowSize, windowStep, minMappability;
    protected static boolean storeCode, bedReads, verbose, unique, orchestra, saveSequence, extend, gcCorrection;
    protected static String fileName, inputFileName, species;
    protected static double pValue, inputReadsRatio, qualityScore;
    
    protected static TreeSet regions = new TreeSet();

    public static DecimalFormat format = new DecimalFormat("###.######");

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    //c, C, d, D, e, g, G, i, l, m, M, n, o, p, P, q, R, s, u, U, v, w, W
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option falseDetectionRateFlag = argParser.addDoubleOption('f', "fdr");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('m', "mismatches");
	    CmdLineParser.Option codeFlag = argParser.addBooleanOption('C', "code");
	    CmdLineParser.Option nUpstreamFlag = argParser.addIntegerOption('U', "nUpstream");
	    CmdLineParser.Option nDownstreamFlag = argParser.addIntegerOption('d', "nDownstream");
	    CmdLineParser.Option minDistFlag = argParser.addIntegerOption('D', "minDist");
	    CmdLineParser.Option minReadsFlag = argParser.addIntegerOption('R', "minReads");
	    CmdLineParser.Option inputReadsRatioFlag = argParser.addDoubleOption('I', "inputReadsRatio");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    //Use this flag when determining the closest gene for an
	    //enriched region
	    CmdLineParser.Option promoterLengthFlag = argParser.addIntegerOption('p', "promoterLength");
	    //Use this flag for determining the windowsize when
	    //scanning the genome for various properties
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option windowStepFlag = argParser.addIntegerOption('W', "windowStep");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    //Use this flag to decide which species to
	    //investigate. Available species are: 
	    //Mouse (Mm) Default
	    //Yeast (Sc)
	    CmdLineParser.Option speciesFlag = argParser.addStringOption('s', "species");
	    CmdLineParser.Option uniqueFlag = argParser.addBooleanOption('u', "unique");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option genomeFlag = argParser.addBooleanOption('g', "genome");
	    //Use this flag when we want to apply the GC-correction
	    CmdLineParser.Option gcCorrectionFlag = argParser.addBooleanOption('G', "gcCorrection");
	    //Use this flag if we want to extend the reads to length fragmentLength
	    CmdLineParser.Option extendFlag = argParser.addBooleanOption('e', "extend");
	    //Exclude windows with very low mappability since they are
	    //considered outliers
	    CmdLineParser.Option minMappabilityFlag = argParser.addIntegerOption('M', "minMappability");
	    //Set the P-value when scanning for motifs
	    CmdLineParser.Option pValueFlag = argParser.addDoubleOption('P', "pValue");
	    //Minimum quality score required for a read to be included
	    CmdLineParser.Option qualityScoreFlag = argParser.addDoubleOption('q', "qualityScore");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    nMismatches = tmp!=null ? ((Integer)tmp).intValue() : 3;
	    storeCode = argParser.getOptionValue(codeFlag)!=null;
	    tmp = argParser.getOptionValue(minDistFlag);
	    minDist = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(minReadsFlag);
	    minReads = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(inputReadsRatioFlag);
	    inputReadsRatio = tmp!=null ? ((Double)tmp).intValue() : 4.0;
	    tmp = argParser.getOptionValue(promoterLengthFlag);
	    promoterLength = tmp!=null ? ((Integer)tmp).intValue() : 10000;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    windowSize = tmp!=null ? ((Integer)tmp).intValue() : 480;
	    tmp = argParser.getOptionValue(windowStepFlag);
	    windowStep = tmp!=null ? ((Integer)tmp).intValue() : 40;
	    verbose = argParser.getOptionValue(verboseFlag)!=null;
	    unique = argParser.getOptionValue(uniqueFlag)!=null;
	    orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    saveSequence = argParser.getOptionValue(genomeFlag)!=null;
	    gcCorrection = argParser.getOptionValue(gcCorrectionFlag)!=null;
	    extend = argParser.getOptionValue(extendFlag)!=null;
	    //Motif scanning flags
	    tmp = argParser.getOptionValue(minMappabilityFlag);
	    minMappability = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(pValueFlag);
	    pValue = tmp!=null ? ((Double)tmp).doubleValue() : .000000001;
	    tmp = argParser.getOptionValue(qualityScoreFlag);
	    qualityScore = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    tmp = argParser.getOptionValue(speciesFlag);
	    species = tmp!=null ? tmp.toString() : "Mm";
	    if (species.equals("Mm")) {
		GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
		nChr = 21;
	    }
	    else if (species.equals("Sc")) {
		GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Sc/sacCer1/" : "/media/exthd/Genomes/Sc/sacCer1/";
		nChr = 17;
	    }
	    //////////////////////////////////////////////////////////////////
	    if (chr==-1) { //Look at all chromosomes
		for (int i=1; i<=nChr; i++)
		    analyze(i);
	    }
	    else
		analyze(chr);
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    public static void analyze(int c) {
	if (inputFileName!=null) { //Analyze the inputs and normalize
	    //TreeSet regs = new TreeSet(PeakFinder.findSignificantClusters(c, fragmentLength, pValue, fileName, inputFileName, GENOME_LOC, extend, 0));
	    //TreeSet regs = PeakFinder2.findEnrichedRegionsFixedWidth(fileName, inputFileName, fragmentLength, extend, 0, pValue, c, orchestra);
	    ABIRead reads[] = PeakFinder2.parseFiles(c, fileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, getSaveFileName("read_stats", c, "txt"));
	    ABIRead inputReads[] = PeakFinder2.parseFiles(c, inputFileName, nMismatches, unique, -1, fragmentLength, extend, 0, qualityScore, null); //getSaveFileName("input_stats", c, "txt"));
	    TreeSet regs = PeakFinder2.findEnrichedRegionsScanningWindow(reads, inputReads, windowSize, windowStep, fileName, pValue, gcCorrection, c, 0, orchestra);
	    //TreeSet regs = PeakFinder2.findEnrichedRegionsScanningWindow(reads, inputReads, windowSize, pValue, gcCorrection, c, 0, orchestra);
	    if (regs!=null && !regs.isEmpty()) {
		System.out.println("Found " + regs.size() + " peaks for chromosome " + c + " and last peak is " + ((ChromosomeWindow)regs.last()).toString());
		//Read the file containing the mRNA info and store
		MRNAInfo mRNAInfo = readMRNAInfo(c);
		MiRNAInfo miRNAInfo = readMiRNAInfo(c);
		int i = 0;
		Iterator it = regs.iterator();
		while (it.hasNext()) {
		    ChromosomeWindow e = (ChromosomeWindow)it.next();
		    MRNAInfoEntry m = mRNAInfo.getNearestGene(e.getCenterOfMass());
		    //int geneDist = m.peakDistance(e.getCenter());
		    //e.setGeneDistance(geneDist);
		    e.setNearestGene(m);
		    //System.out.println(e.toString() + " is close to " + m.toString());
		    //Find the closest miRNA gene
		    e.setNearestMiRNA(miRNAInfo.getNearestMiRNA(e.getCenterOfMass()));
		}
		if (fileName.toLowerCase().indexOf("pol2")!=-1 || fileName.toLowerCase().indexOf("k36")!=-1) 
		    //Since the windowing method seems incapable of
		    //detecting regions which appear to be transcribed by
		    //eye, we will try to see if an algorithm using a
		    //window consisting of the entire coding region can do
		    //better.
		    comparePol2CodingRegions(PeakFinder2.parseFiles(c, fileName, fragmentLength, extend, 0, qualityScore, null), PeakFinder2.parseFiles(c, inputFileName, fragmentLength, extend, 0, qualityScore, null), mRNAInfo, pValue, getSaveFileName("transcribed", c, "txt"), getSaveFileName("transcribed", c, "bed"));
		if (fileName.toLowerCase().indexOf("pol2")!=-1) {
		    //For Pol2, we want to classify the genes by how much Pol2
		    //they have at the promoter, in the coding region and at
		    //the 3' end
		    analyzePol2Distribution(mRNAInfo, regs, getSaveFileName("tss_peaks", c, "txt"));	    
		    //findInternalPol2Enrichment(PeakFinder2.parseFiles(c, fileName, fragmentLength, extend, 0, qualityScore), PeakFinder2.parseFiles(c, inputFileName, fragmentLength, extend, 0, qualityScore), mRNAInfo, 2, 2, getSaveFileName("internal_pol2", c, "txt"));
		    int pol2ScanWin = 5000;
		    findInternalPol2EnrichmentWindow(PeakFinder2.parseFiles(c, fileName, fragmentLength, extend, 0, qualityScore, null), PeakFinder2.parseFiles(c, inputFileName, fragmentLength, extend, 0, qualityScore, null), mRNAInfo, pol2ScanWin, getSaveFileName("internal_pol2_" + pol2ScanWin, c, "txt"));
		}
		//For Npas4 we can compare with a list of targets
		//obtained elsewhere
		if (fileName.indexOf("Npas4")!=-1 && fileName.toLowerCase().indexOf("input")==-1) {
		    compareNpas4Targets(c, regs, mRNAInfo, getSaveFileName("npas4_RNAi_up", c, "txt"), true);
		    compareNpas4Targets(c, regs, mRNAInfo, getSaveFileName("npas4_RNAi_down", c, "txt"), false);
		}
		//compareGeneList(c, regs, mRNAInfo, fileName + "_autism_" + c + "_" + (extend ? "E" : "") + fragmentLength + ".txt", "GeneLists/autism_genes.txt");
		//compareGeneList(c, regs, mRNAInfo, fileName + "_npas4_" + c + "_" + (extend ? "E" : "") + fragmentLength + ".txt", "GeneLists/npas4_genes.txt");
		//compareGeneList(c, regs, mRNAInfo, fileName + "_mef2_" + c + "_" + (extend ? "E" : "") + fragmentLength + ".txt", "GeneLists/mef2_genes.txt");
		//compareGeneList(c, regs, mRNAInfo, fileName + "_KCl_" + c + "_" + (extend ? "E" : "") + fragmentLength + ".txt", "GeneLists/additional_KCl_genes.txt");
		//compareGeneList(c, regs, mRNAInfo, fileName + "_environment_" + c + "_" + (extend ? "E" : "") + fragmentLength + ".txt", "GeneLists/additional_environment_genes.txt");
		//compareGeneList(c, regs, mRNAInfo, fileName + "_npas4Andmef2_" + c + "_" + (extend ? "E" : "") + fragmentLength + ".txt", "GeneLists/npas4Andmef2_genes.txt");
		saveGeneList(c, regs, mRNAInfo);
		saveGeneList(c, regs, mRNAInfo, true);
		saveMiRNAList(c, miRNAInfo, regs, getSaveFileName("miRNA", c, "txt"));
		saveBedFile(getSaveFileName("peaks", c, "bed"), regs, (double)CreateBedFiles.READS_NORMALIZATION_FACTOR*(double)PeakFinder2.CHROMOSOME_LENGTH[c-1]/((double)reads.length*(double)PeakFinder2.TOTAL_CHROMOSOME_LENGTH));
		//savePeakReads(c, regs);
		//savePeakSequences(regs, getSaveFileName("seqs", c, "fa")); //Find the sequences
		//scanMotif(regs, fileName + "_motifs_" + c + "_" + (extend ? "E" : "") + fragmentLength + ".txt");
		regions.addAll(regs);
	    }	    
	}
    }

    public static String getChromosomeString(int c, String species) {
	return "chr" + getChromosomeNumber(c, species) + ".fa";
    }

    public static String getChromosomeString(int c) {
	return "chr" + getChromosomeNumber(c, "mm") + ".fa";
    }

    public static String getChromosomeNumber(int c, String species) {
	if (species.toLowerCase().equals("mm")) {
	    if (c==20)
		return "X";
	    else if (c==21)
		return "Y";
	    else if(c>0 && c<20)
		return String.valueOf(c);
	}
	if (species.toLowerCase().equals("hs")) {
	    if (c==23)
		return "X";
	    else if (c==24)
		return "Y";
	    else if(c>0 && c<23)
		return String.valueOf(c);
	}
	else if (species.toLowerCase().equals("sc")) {
	    if (c==17)
		return "M";
	    else if(c>0 && c<17)
		return String.valueOf(c);
	}
	System.err.println("Chr number " + c + " is not allowed for " + species);
	return new String("-1");
    }

    public static String getChromosomeNumber(int c) {
	return getChromosomeNumber(c, "mm");
    }
    public static String getSaveFileName(String prefix, int c, String suffix) {
	//Use to create the name for the files which we are saving
	if (c>0)
	    return fileName + "_" + prefix + "_" + c + "_" + (gcCorrection ? "" : "noGC_") + (extend ? "E" : "") + fragmentLength + "." + suffix;
	return fileName + "_" + prefix + "_" + (gcCorrection ? "" : "noGC_") + (extend ? "E" : "") + fragmentLength + "." + suffix;
    }

    public static String getSaveFileName(String fileName, String prefix, int c, int fragmentLength, String suffix) {
	//Use to create the name for the files which we are saving
	if (c>0)
	    return fileName + "_" + prefix + "_" + c + "_" + "E" + fragmentLength + "." + suffix;
	return fileName + "_" + prefix + "_" + "E" + fragmentLength + "." + suffix;
    }

    protected static void scanMotif(TreeSet regs, String fName) {
	//Use this method to scan for a motif with a known PFM. Save
	//information about the reads nearby with the strandedness
	//info so that we can reproduce fig 1a in the MACS paper and
	//try to determine the true fragmentation length.
	/*try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fName));
	    Vector motifDist = new Vector();
	    int nSeqsWithMotif = 0;
	    Iterator it = regs.iterator();
	    String tfmDir = null;
	    if (motifName.equals("mef2") || motifName.equals("MEF2"))
		tfmDir = orchestra ? "/groups/neuroduo/martin/Softwares/TFMscan1.0/Data/Mats/Jaspar/" : "/home/mhemberg/Softwares/TFMscan1.0/Data/Mats/Jaspar/";
	    if (tfmDir!=null) {
		while (it.hasNext()) {
		    ChromosomeWindow e = (ChromosomeWindow)it.next();
		    TFMScanner tfmScanner = null; //new TFMScanner(e.getSequence(), motifName.toUpperCase(), orchestra);
		    tfmScanner.findMotifs(.001/*motifPVal, tfmDir);
		    long tfbs[] = tfmScanner.getMotifLocations(motifName.toUpperCase());
		    //scan the reverse complement as well
		    tfmScanner = null; //new TFMScanner(MerTree.reverseComplement(e.getSequence()), motifName.toUpperCase(), orchestra);
		    tfmScanner.findMotifs(.001/*motifPVal, tfmDir);
		    long tfbsrc[] = tfmScanner.getMotifLocations(motifName.toUpperCase());
		    bw.write(e.toString());
		    if (tfbs.length + tfbsrc.length>0)	{ 
			nSeqsWithMotif++;
			for (int i=0; i<tfbs.length; i++) {
			    motifDist.addElement(new Integer((int)(e.getPeakLocation() - e.getStart() - tfbs[i])));
			    bw.write("\t" + (e.getPeakLocation() - e.getStart() - tfbs[i]));
			}
			for (int i=0; i<tfbsrc.length; i++) {
			    motifDist.addElement(new Integer((int)-(e.getPeakLocation() - e.getStart() - tfbsrc[i])));
			    bw.write("\t" + (-(e.getPeakLocation() - e.getStart() - tfbsrc[i])));
			}
		    }
		    bw.write("\n");
		    /*System.out.println("Found " + (tfbs.length + tfbsrc.length) + " motifs and peak is at " + (e.getPeakLocation() - e.getStart()));
		    for (int i=0; i<tfbs.length; i++)
			System.out.println("Motif is at " + tfbs[i] + ", dist " + (e.getPeakLocation() - e.getStart() - tfbs[i]));
		    for (int i=0; i<tfbsrc.length; i++)
		    System.out.println("Motif is at " + tfbsrc[i] + ", dist " + (e.getPeakLocation() - e.getStart() - tfbsrc[i]));
		}
	    }
	    else
		System.err.println("No PFM available for " + motifName);
	    bw.close();
	}
	catch (IOException e) { }
	*/
    }

    protected static void saveBedFile(String fName, TreeSet enrichedRegions, double normalizationFactor) {
	//Use this method to save the data about the enriched regions
	//into a .bed-file. This will allow us to have the info for
	//all chromosomes in one file and at the same time avoid
	//having all the background crap.
	try {
	    System.out.println("Saving peaks with factor " + normalizationFactor);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fName));
	    //Write some header info
	    String tName = fName.substring(fName.lastIndexOf("/") + 1, fName.lastIndexOf(".") - 1);
	    String trackName = tName, desc = tName;
	    Iterator it = enrichedRegions.iterator();
	    //Write the reads info
	    if (((ChromosomeWindow)enrichedRegions.first()).getChromosome()==1) {
		bw.write("browser position chr1:21315319-21327381\n");
		bw.write("track type=wiggle_0 name=\"" + trackName + "_data\" description=\"" + desc + "_data\" visibility=full autoScale=off viewLimits=0:2 color=\"255,0,0\" maxHeightPixels=70:40:10 priority=1\n");
	    }
	    int prevEnd = 0;
	    while (it.hasNext()) {
		ChromosomeWindow er = (ChromosomeWindow)it.next();
		String c = "chr" + getChromosomeNumber(er.getChromosome());
		int constRegions[][] = er.getBedFormat();
		for (int i=0; i<constRegions.length; i++) {
		    if (constRegions[i][2]>0 && constRegions[i][0]>=prevEnd && constRegions[i][1]>constRegions[i][0]) { //make sure that there are reads in the region
			bw.write(c + "\t" + constRegions[i][0] + "\t" + constRegions[i][1] + "\t" + format.format(normalizationFactor*(double)constRegions[i][2]) + "\n");
			prevEnd = Math.max(constRegions[i][1], prevEnd);
		    }
		}
	    }
	    bw.close();
	    //Write the input-info
	}
	catch (IOException e) { System.out.println("Error writing .bed file!"); }
    }

    public static Vector readMRNAInfo() {
	Vector info = new Vector(nChr); //Get the info for all genes
	for (int i=0; i<nChr; i++) 
	    info.addElement(readMRNAInfo(i+1));
	return info;
    }

    public static MRNAInfo readMRNAInfo(int c) { return readMRNAInfo(c, GENOME_LOC); }

    public static MRNAInfo readMRNAInfo(int c, String genomeLoc) {
	MRNAInfo mRNAInfo = new MRNAInfo(c, orchestra);
	try {
	    BufferedReader br = new BufferedReader(new FileReader(genomeLoc + "refSeqGenes/mrna_list.txt"));
	    String line = br.readLine();
	    //- gene geneID refseq - chr strand start end length n_coding n_intronic
	    Pattern mrnaPattern = Pattern.compile("\\d+\\s+([\\w|\\d]+)\\s+(\\d+)\\s+([X|N]M_\\d+\\.\\d)\\s+\\d+\\s+([X|Y|\\d]+)\\s+([+|-])\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = mrnaPattern.matcher(line);
		if (m.find()) {
		    int c2;
		    if (m.group(4).equals("X"))
			c2 = 20;
		    else if (m.group(4).equals("Y"))
			c2 = 21;
		    else
			c2 = Integer.parseInt(m.group(4));
		    if (c==c2) {
			boolean strand = m.group(5).equals("+");
			int start = strand ? Integer.parseInt(m.group(6)) : Integer.parseInt(m.group(7));
			int end = strand ? Integer.parseInt(m.group(7)) : Integer.parseInt(m.group(6));
			//check if a version of the gene already
			//exists. If so, we merge the two versions
			//into a super-gene which is contains the
			//union of the two versions
			MRNAInfoEntry g = mRNAInfo.getGene(Integer.parseInt(m.group(2)));
			if (g==null)
			    mRNAInfo.add(new MRNAInfoEntry(c, start, end, strand, m.group(1), Integer.parseInt(m.group(2)), m.group(3)));
			else { //found version already parsed
			    if (strand) {
				g.setStart(Math.min(start, g.getStart()));
				g.setEnd(Math.max(end, g.getEnd()));
			    }
			    else {
				g.setStart(Math.max(start, g.getStart()));
				g.setEnd(Math.min(end, g.getEnd()));
			    }
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { 
	    System.err.println("Error reading mRNA info from " + genomeLoc + "refSeqGenes/mrna_list.txt\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
	//System.out.println("Found " + mRNAInfo.size() + " genes for chromosome " + c);
	mRNAInfo.addExonInfo(GENOME_LOC);
	return mRNAInfo;
    }

    protected static void compareNpas4Targets(int c, TreeSet regs, MRNAInfo mRNAInfo, String saveFileName, boolean up) {
	//Find the peaks which are nearby genes which are affected by
	//Npas4-RNAi from Yingxi's list. There are two categories, up-
	//and downregulated.
	try {
	    System.out.println("Trying to open regulation file.");
	    BufferedReader br = new BufferedReader(new FileReader((orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/Npas4/Npas4_regulation.txt" : "/home/mhemberg/Projects/ChIP-Seq/TKData/Npas4/Npas4_regulation.txt")));
	    BufferedWriter bw = new BufferedWriter(new FileWriter((orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/Npas4/Npas4_regulation_names.txt" : "/home/mhemberg/Projects/ChIP-Seq/TKData/Npas4/Npas4_regulation_names.txt")));
	    TreeSet regulatedGenes = new TreeSet();
	    String line = br.readLine();
	    //groups are affy code and gene name
	    Pattern affyPattern = Pattern.compile("\\d+_[\\w_]*?at");
	    //Use this pattern to find the regulation activity which
	    //is last on the line. The second last column reflects
	    //activity regulated and can be [U|N|D] and the last is
	    //the Npas4-RNAi regulated which can be [U|D].
	    Pattern geneNamePattern = Pattern.compile("([\\w|\\d|-]+)\\s+.+\\s+([D|U|N])\\s+([D|U])\\s+");
	    //Pattern regulationPattern = Pattern.compile("([U|N|D])\s+([U|D])");
	    //Pattern geneNamePattern = Pattern.compile("([\\w|\\d]+)");
	    while (line!=null) {
		Matcher m = affyPattern.matcher(line);
		if (m.find()) {
		    //System.out.println("Found " + m.group(0) + " at " + line);
		    int end = m.end();
		    while (m.find(end)) { end = m.end(); }
		    //if (!line.contains("---")) {
			Matcher m2 = geneNamePattern.matcher(line);
			if (m2.find(end)) {
			    MRNAInfoEntry gene = mRNAInfo.getGene(m2.group(1));
			    if (gene!=null) {
				gene.setActivityRegulated(m2.group(2).equals("N") ? 0 : m2.group(2).equals("D") ? -1 : 1);
				gene.setNpas4RNAiRegulated(m2.group(3).equals("U") ? 1 : -1);
			    regulatedGenes.add(gene);
			    }
			    bw.write(m2.group(1) + "\t" + m2.group(2) + "\t" + m2.group(3) + "\n");
			}
			//}
		}
		line = br.readLine();
	    }
	    br.close();
	    bw.close();
	    System.out.println("Found " + regulatedGenes.size());
	    bw = new BufferedWriter(new FileWriter(saveFileName));
	    if (c==1) //Write a header
		bw.write("#" + ChromosomeWindow.getHeaderString() + "\t" + MRNAInfoEntry.getHeaderString() + "\tDistance\t5/3\tActivity\n");
	    Iterator it = regulatedGenes.iterator();
	    while (it.hasNext()) {
		MRNAInfoEntry gene = (MRNAInfoEntry)it.next();
		if ((gene.getNpas4RNAiRegulated()>0 && up) || (gene.getNpas4RNAiRegulated()<0 && !up)) {
		    TreeSet nearbyPeaks = gene.findProximalPeaks(regs, promoterLength);
		    nearbyPeaks.addAll(gene.findInternalPeaks(regs));
		    //Add those peaks which have this gene as its closest gene
		    nearbyPeaks.addAll(gene.findNearestPeaks(regs));
		    System.out.println("Found " + nearbyPeaks.size() + " peaks for " + gene.name);
		    //Save the peaks
		    Iterator it2 = nearbyPeaks.iterator();
		    while (it2.hasNext()) {
			ChromosomeWindow e = (ChromosomeWindow)it2.next();
			//Don't use the function since we have additional info.
			String str = e.toString() + "\t" + gene.geneInfo() + "\t" + gene.peakDistance(e.getCenter()) + "\t" + (gene.fivePrimeClosest(e.getCenterOfMass()) ? "5" : "3") + "\t" + gene.getActivityRegulated();
			bw.write(str + "\n");
		    }
		}
	    }
	    bw.close();
	}
	catch (IOException e) { }
    }

    protected static void saveHeader(BufferedWriter bw) {
	try {
	    bw.write("#CHR\tSTART\tSTOP\tWIDTH\tHEIGHT\tREADS\tINPUT_READS\tGENE\tREFSEQ\tGENE_ID\tDIST\tSTRAND\tI/P/D\n");
	}
	catch (IOException e) { }
    }

    protected static String saveGeneCentricPeakInfo(BufferedWriter bw, ChromosomeWindow e, MRNAInfoEntry gene) {
	String str = e.toString() + "\t" + gene.name + "\t" + gene.refSeq + "\t" + gene.id + "\t" + gene.peakDistance(e.getCenterOfMass());
	str += gene.strand ? "\t+" : "\t-";
	str += gene.peakInternal(e.getCenter()) ? "\tI" : gene.isProximal(e.getCenter(), promoterLength) ? "\tP" : "\tD";
	//Add info about which end of the gene is the closest one
	str += gene.fivePrimeClosest(e.getCenter()) ? "5" : "3";
	//str += "\t" + mRNAInfo.numberOfProximalOrInternal(e.getCenter(), promoterLength);
	try {
	    bw.write(str + "\n");
	}
	catch (IOException ex) { }
	return str;
    }

    protected static void compareGeneList(int c, TreeSet regs, MRNAInfo mRNAInfo, String saveFileName, String listFileName) {
	//A first try at writing a somewhat generic piece of code for
	//comparing the list of peaks with a list of genes which have
	//been shown to be activated from microarray studies.
	try {
	    BufferedReader br = new BufferedReader(new FileReader(listFileName));
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    if (c==1) { saveHeader(bw); }
	    String line = br.readLine();
	    TreeSet peaks = new TreeSet();
	    //Gene id and name
	    Pattern genePattern = Pattern.compile("(\\d+)\\s+(\\w+)");
	    while (line!=null) {
		Matcher m = genePattern.matcher(line);
		if (m.find()) {
		    int id = Integer.parseInt(m.group(1));
		    MRNAInfoEntry gene = mRNAInfo.getGene(m.group(2));
		    if (gene!=null) {
			//Find the peak which is closest to this gene as
			//well as all peaks which are proximal or
			//intronic.
			peaks = gene.findInternalPeaks(regs);
			peaks.addAll(gene.findProximalPeaks(regs, promoterLength));
			Iterator it = regs.iterator();
			while (it.hasNext()) {
			    ChromosomeWindow er = (ChromosomeWindow)it.next();
			    if (er.getNearestGene().name.toLowerCase().equals(gene.name.toLowerCase()))
				peaks.add(er);
			}
			//Save the list of peaks
			it = peaks.iterator();
			while (it.hasNext()) {
			    ChromosomeWindow er = (ChromosomeWindow)it.next();
			    saveGeneCentricPeakInfo(bw, er, gene);
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    bw.close();
	}
	catch (IOException e) { }
    }

    protected static void saveGeneList(int c, TreeSet regs, MRNAInfo mRNAInfo) { saveGeneList(c, regs, mRNAInfo, false); }

    protected static void saveGeneList(int c, TreeSet regs, MRNAInfo mRNAInfo, boolean intergenic) {
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(getSaveFileName((intergenic ? "intergenic" : "genes"), c, "txt")));
	    if (c==1) { //write the header info
		if (intergenic)
		    bw.write("#" + ChromosomeWindow.getHeaderString() + "\tUp/down\tGene_name\tGeneID\n");
		else
		    saveHeader(bw);
	    }
	    Iterator it = regs.iterator();
	    while (it.hasNext()) {
		ChromosomeWindow e = (ChromosomeWindow)it.next();
		if (intergenic) {
		    if (e.getNearestGeneDistance()>2*promoterLength && !e.getNearestGene().peakInternal(e.getCenterOfMass())) 
			bw.write(e.toString() + "\t" + /*e.getNearestGeneDistance() + "\t" +*/ (e.getNearestGene().peakUpstream(e.getCenterOfMass()) ? "U" : "D") + "\t" + e.getNearestGene().getName() + "\t" + e.getNearestGene().getID() + "\n");
		}
		else
		    bw.write(e.toString() + "\t" + e.getNearestGeneDistance() + "\t" + (e.getNearestGene().peakInternal(e.getCenterOfMass()) ? "I" : e.getNearestGene().peakUpstream(e.getCenterOfMass()) ? "U" : "D") + "\t" + e.getNearestGene().getName() + "\t" + e.getNearestGene().getID() + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) { }
    }
    
    protected static void savePeakSequences(TreeSet regs, String saveFileName) {
	try {
	    //Save the peak sequences
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    Iterator it = regs.iterator();
	    while (it.hasNext()) {
		ChromosomeWindow e = (ChromosomeWindow)it.next();
		bw.write(">chr" + e.getChromosome() + "_" + e.getSignificantStart() + "-" + e.getSignificantEnd() + "\n");
		int ss = e.getSignificantStart() - e.getStart(), se = e.getSignificantEnd() - e.getStart();
		bw.write(e.getSequence().substring(ss, se) + "\n");
	    }
	    bw.close();
	    //Also save a larger chunk to use as backgrounds for
	    //various motif-searching algorithms
	    /*bw = new BufferedWriter(new FileWriter(fileName + "_seqs_bg_" + c + ".fa"));
	      it = regs.iterator();
	      while (it.hasNext()) {
	      EnrichedRegion e = (EnrichedRegion)it.next();
	      bw.write(">chr" + e.getChromosome() + "_" + e.getStart() + "-" + e.getEnd() + "\n");
	      bw.write(e.getBackgroundSequence() + "\n");
	      }
	      bw.close();*/
	}
	catch (IOException e) { }
    }
	
	      
    /*protected static void savePeakSequences(int c, TreeSet regs) {
	Iterator it = regs.iterator();
	int peakStart[] = new int[regs.size()], peakLength[] = new int[regs.size()], i = 0;
	while (it.hasNext()) {
	    ChromosomeWindow e = (ChromosomeWindow)it.next();
	    peakStart[i] = e.getStart();
	    peakLength[i] = e.getLength();
	    //System.out.println(peakStart[i] + "\t" + peakLength[i]);
	    i++;
	}
	FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + getChromosomeString(c, species)); //Retrieve the sequences
	String seqs[] = fastaSeq.getSequences(peakStart, peakLength);
	i = 0;
	it = regs.iterator();
	while (it.hasNext()) {
	    ChromosomeWindow e = (ChromosomeWindow)it.next();
	    e.setSequence(seqs[i++]);
	    //e.setBackgroundSequence(orchestra, 2*fragmentLength);
	}
	//Save to a file
	try {
	    //Save the peak sequences
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName + "_seqs_" + c + "_" + (extend ? "E" : "") + fragmentLength + ".fa"));
	    it = regs.iterator();
	    while (it.hasNext()) {
		ChromosomeWindow e = (ChromosomeWindow)it.next();
		bw.write(">chr" + e.getChromosome() + "_" + (e.getStart()) + "-" + (e.getStart() + e.getLength()) + "\n");
		bw.write(e.getSequence() + "\n");
	    }
	    bw.close();
	    //Save only the most central part under the peak
	    bw = new BufferedWriter(new FileWriter(fileName + "_seqs_center_" + c + "_" + (extend ? "E" : "") + fragmentLength + ".fa"));
	    it = regs.iterator();
	    while (it.hasNext()) {
		ChromosomeWindow e = (ChromosomeWindow)it.next();
		//bw.write(">chr" + e.getChromosome() + "_" + (e.getStart() + e.getHalfWidthStart()) + "-" + (e.getStart() + e.getHalfWidthStart() + e.getHalfWidth()) + "\n");
		//bw.write(e.getSequence(true) + "\n");
	    }
	    bw.close();
	    //Also save a larger chunk to use as backgrounds for
	    //various motif-searching algorithms
	    /*bw = new BufferedWriter(new FileWriter(fileName + "_seqs_bg_" + c + ".fa"));
	    it = regs.iterator();
	    while (it.hasNext()) {
		EnrichedRegion e = (EnrichedRegion)it.next();
		bw.write(">chr" + e.getChromosome() + "_" + e.getStart() + "-" + e.getEnd() + "\n");
		bw.write(e.getBackgroundSequence() + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) { }
    }*/

    protected static void savePeakReads(int c, TreeSet regs){
	//Save the reads info for the peaks
	try {
	    //Save the peak sequences
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName + "_E_" + c));
	    Iterator it = regs.iterator();
	    while (it.hasNext()) {
		ChromosomeWindow e = (ChromosomeWindow)it.next();
		//e.saveReads(bw);
	    }
	    bw.close();
	}
	catch (IOException e) { }
    }

    public static void analyzePol2Distribution(MRNAInfo mRNAInfo, TreeSet regs, String saveFileName) {
	try {
	    int chr = mRNAInfo.getChromosome();
	    mRNAInfo.findNearbyPeaks(regs, promoterLength);
	    Iterator it = mRNAInfo.getGenes().iterator();
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    if (chr==1) { bw.write("#Name\tID\tChr\tStart\tEnd\tStrand\tTSS_reads\tTES_reads\n"); }
	    int nTSSonly = 0, nTSSandTES = 0, nTSSandInternal = 0;
	    while (it.hasNext()) { //Go through all genes and investigate the distribution of Pol2 reads
		MRNAInfoEntry gene = (MRNAInfoEntry)it.next();
		TreeSet nearbyPeaks = gene.getNearbyPeaks();
		//Find the peaks which overlap, are upstream of or is nearest
		//the annotated TSS.
		if (nearbyPeaks.size()>0) {
		    Iterator it2 = nearbyPeaks.iterator();
		    int totReads = 0, nTSSOverlapReads = 0, nUpstreamReads = 0, nTESOverlapReads = 0, nDownstreamReads = 0, nInternalReads = 0;
		    while (it2.hasNext()) {
			ChromosomeWindow peak = (ChromosomeWindow)it2.next();
			if (peak.overlaps(gene.getStart()))
			    nTSSOverlapReads = peak.numberOfReads();
			else if (gene.peakUpstream(peak.getCenterOfMass()) && peak.isNearestGene(gene))
			    nUpstreamReads += peak.numberOfReads();
			else if (peak.overlaps(gene.getEnd()))
			    nTESOverlapReads = peak.numberOfReads();
			else if (gene.peakDownstream(peak.getCenterOfMass()) && peak.isNearestGene(gene)) //Find the peaks which are beyond the TES
			    nDownstreamReads += peak.numberOfReads();
			else //Find everything which is in between
			    nInternalReads += peak.numberOfReads();
			totReads += peak.numberOfReads();
		    }
		    //Calculate what proportions of the reads are in the five categories
		    if (nTSSOverlapReads>0)
			bw.write(gene.getName() + "\t" + gene.getID() + "\t" + gene.getChromosome() + "\t" + gene.getStart() + "\t" + gene.getEnd() + "\t" + (gene.getStrand() ? "+" : "-") + "\t" + nTSSOverlapReads + "\t" + nTESOverlapReads + "\n");
		}
	    }
	    bw.close();
	}
	catch (IOException e) { }
    }

    protected static void comparePol2CodingRegions(ABIRead reads[], ABIRead inputReads[], MRNAInfo mRNAInfo, double pValue, String saveFileName, String bedSaveFileName) {
	try {
	    int chr = mRNAInfo.getChromosome();
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    BufferedWriter bedWriter = new BufferedWriter(new FileWriter(bedSaveFileName));
	    //Write a header
	    if (chr==1) { bw.write("#Name\tGeneID\tStart\tEnd\tStrand\tReads\tInput\n"); }
	    //Go through all the coding regions and check if there is
	    //significant enrichment compared with input
	    Mappability mappability = null; //new Mappability(GENOME_LOC, chr, windowSize, species);
	    FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(chr));
	    double readsPoiRatio = (double)reads.length/mappability.getNumberOfUniqueBases();
	    double inputPoiRatio = (double)inputReads.length/mappability.getNumberOfUniqueBases();
	    Iterator it = mRNAInfo.getGenes().iterator();
	    int readInd = 0, inputInd = 0;
	    while (it.hasNext()) {
		MRNAInfoEntry gene = (MRNAInfoEntry)it.next();
		int s = gene.getStrand() ? gene.getStart() : gene.getEnd();
		int e = gene.getStrand() ? gene.getEnd() : gene.getStart();
		TreeSet codingReads = new TreeSet(), codingInputReads = new TreeSet();
		while (readInd<reads.length && reads[readInd].getPosition()<e) {
		    if (reads[readInd].getPosition()>=s)
			codingReads.add(reads[readInd]);
		    readInd++;
		}
		while (inputInd<inputReads.length && inputReads[inputInd].getPosition()<e) {
		    if (inputReads[inputInd].getPosition()>=s)
			codingInputReads.add(inputReads[inputInd]);
		    inputInd++;
		}
		if (codingInputReads.size()>0 && codingReads.size()>0) {
		    //Calculate the potential number of reads for the region
		    int potReads = mappability.numberOfPotentialReadsRegion(s, e-windowSize);
		    //Calculate the p-value for the observed number of reads
		    double readsPVal = PeakFinder2.calculatePValue(codingReads.size(), potReads, readsPoiRatio);
		    double inputPVal = PeakFinder2.calculatePValue(codingInputReads.size(), potReads, inputPoiRatio);
		    if (readsPVal/inputPVal<pValue) {
			//System.out.println(gene.getName() + " at " + s + "-" + e + " has " + codingReads.size() + " reads and " + codingInputReads.size() + " input reads with p-values " + format.format(readsPVal) + " and " + format.format(inputPVal));
			bw.write(gene.getName() + "\t" + gene.getID() + "\t" + "\t" + gene.getStart() + "\t" + gene.getEnd() + "\t" + (gene.getStrand() ? "+" : "-") + "\t" + codingReads.size() + "\t" + codingInputReads.size() + "\n");
			ChromosomeWindow win = new ChromosomeWindow(chr, s, e, codingReads);
			bedWriter.write(win.printBedFormat());
		    }
		}
		//int randRegionsStart[] = new int[nRandRegions];
		//for (int i=0; i<nRandRegions; i++) { randRegionsStart[i] = random.nextInt(mappability.getNumberOfBases()); }
		//Arrays.sort(randRegionsStart);
		//while (false) { //count the number of random regions which appear significant to estimate the FDR
		    //Find a random region

		    //Maker sure that reads can be mapped to it

		    //Check if it comes out as significantly enriched when comparing with input
		//}
	    }
	    bw.close();
	    bedWriter.close();
	}
	catch (IOException ex) {

	}
    }

    protected static void findInternalPol2Enrichment(ABIRead reads[], ABIRead inputReads[], MRNAInfo mRNAInfo, int startExonOffset, int stopExonOffset, String saveFileName) {
	//The start and stop exon offset are used to determine after
	//which exon counting from start/end we should investigate the
	//region in between. Zero indicates the full gene and two
	//indicates the region between the second and second to last
	//exon.
	//
	//The output is a list containing the gene name, id, reads,
	//input reads, start, end of region and p-value for region
	//being actively transcribed. Use Matlab script
	//find_internal_pol2_enrichment.m to compare with the
	//unstimulated scenario.
	try {
	    int chr = mRNAInfo.getChromosome();
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    //Write a header
	    if (chr==1) 
		bw.write("#Name\tGeneID\tChr\tStart\tEnd\tStrand\tReads\tInput\tP-val\tInput P-val\n"); 
	    //Go through the region of interest
	    Mappability mappability = null; //new Mappability(GENOME_LOC, chr, windowSize, species);
	    FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(chr));
	    double readsPoiRatio = (double)reads.length/mappability.getNumberOfUniqueBases();
	    double inputPoiRatio = (double)inputReads.length/mappability.getNumberOfUniqueBases();
	    Iterator it = mRNAInfo.getGenes().iterator();
	    int readInd = 0, inputInd = 0;
	    while (it.hasNext()) {
		MRNAInfoEntry gene = (MRNAInfoEntry)it.next();
		if (gene.hasExonInfo()) { //Sometimes we fail to assign exon information
		    int nExons = gene.numberOfExons();
		    if (nExons>=startExonOffset + stopExonOffset) { 
			int s = gene.getStrand() ? gene.getExonStop(startExonOffset-1) : gene.getExonStart(nExons - stopExonOffset + 1);
			int e = gene.getStrand() ? gene.getExonStart(nExons - stopExonOffset + 1) : gene.getExonStop(startExonOffset-1);
			TreeSet codingReads = new TreeSet(), codingInputReads = new TreeSet();
			while (readInd<reads.length && reads[readInd].getPosition()<e) {
			    if (reads[readInd].getPosition()>=s)
				codingReads.add(reads[readInd]);
			    readInd++;
			}
			while (inputInd<inputReads.length && inputReads[inputInd].getPosition()<e) {
			    if (inputReads[inputInd].getPosition()>=s)
				codingInputReads.add(inputReads[inputInd]);
			    inputInd++;
			}
			if (codingInputReads.size()>0 && codingReads.size()>0) {
			    //Calculate the potential number of reads for the region
			    int potReads = mappability.numberOfPotentialReadsRegion(s, e-windowSize);
			    //Calculate the p-value for the observed number of reads
			    double readsPVal = PeakFinder2.calculateLogPValue(codingReads.size(), potReads, readsPoiRatio);
			    double inputPVal = PeakFinder2.calculateLogPValue(codingInputReads.size(), potReads, inputPoiRatio);
			    //#Name\tGeneID\tChr\tStart\tEnd\tStrand\tReads\tInput\tP-val\tInput P-val\n
			    bw.write(gene.getName() + "\t" + gene.getID() + "\t" + "\t" + gene.getChromosome() + "\t" + Math.min(s, e) + "\t" + Math.max(e, s) + "\t" + (gene.getStrand() ? "+" : "-") + "\t" + codingReads.size() + "\t" + codingInputReads.size() + "\t" + format.format(readsPVal) + "\t" + format.format(inputPVal) + "\n");
			}
		    }
		}
		else 
		    System.err.println("Failed to find exon info for " + gene.toString());
	    }
	    bw.close();
	}
	catch (IOException ex) {

	}
    }

    protected static void findInternalPol2EnrichmentWindow(ABIRead reads[], ABIRead inputReads[], MRNAInfo mRNAInfo, int windowLength, String saveFileName) {
	//Go through all the coding regions using a window of fixed
	//size. Count the number of Pol2 reads in each window and save
	//to file. This will later be compared across conditions.
	try {
	    int chr = mRNAInfo.getChromosome();
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    //Write a header
	    if (chr==1) 
		bw.write("#Name\tGeneID\tChr\tStart\tEnd\tStrand\tReads\tInput\tP-val\tInput P-val\n"); 
	    //Go through the region of interest
	    Mappability mappability = null; //new Mappability(GENOME_LOC, chr, windowSize, species);
	    FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(chr));
	    double readsPoiRatio = (double)reads.length/mappability.getNumberOfUniqueBases();
	    double inputPoiRatio = (double)inputReads.length/mappability.getNumberOfUniqueBases();
	    Iterator it = mRNAInfo.getGenes().iterator();
	    int readInd = 0, inputInd = 0;
	    while (it.hasNext()) {
		MRNAInfoEntry gene = (MRNAInfoEntry)it.next();
		int start = gene.getStrand() ? gene.getStart() : gene.getEnd();
		int end = gene.getStrand() ? gene.getEnd() : gene.getStart();
		int s = start, e = start + windowLength;
		while (e<end) { //We will be not be looking at the final truncated window, this should be a minor issue though, since we are expecting interesting stuff to happen considerably earlier.
		    TreeSet codingReads = new TreeSet(), codingInputReads = new TreeSet();
		    while (readInd<reads.length && reads[readInd].getPosition()<e) {
			if (reads[readInd].getPosition()>=s)
			    codingReads.add(reads[readInd]);
			readInd++;
		    }
		    while (inputInd<inputReads.length && inputReads[inputInd].getPosition()<e) {
			if (inputReads[inputInd].getPosition()>=s)
			    codingInputReads.add(inputReads[inputInd]);
			inputInd++;
		    }
		    if (codingInputReads.size()>0 && codingReads.size()>0) {
			//Calculate the potential number of reads for the region
			int potReads = mappability.numberOfPotentialReadsRegion(s, e-windowSize);
			//Calculate the p-value for the observed number of reads
			double readsPVal = PeakFinder2.calculateLogPValue(codingReads.size(), potReads, readsPoiRatio);
			double inputPVal = PeakFinder2.calculateLogPValue(codingInputReads.size(), potReads, inputPoiRatio);
			//#Name\tGeneID\tChr\tStart\tEnd\tStrand\tReads\tInput\tP-val\tInput P-val\n
			bw.write(gene.getName() + "\t" + gene.getID() + "\t" + "\t" + gene.getChromosome() + "\t" + Math.min(s, e) + "\t" + Math.max(e, s) + "\t" + (gene.getStrand() ? "+" : "-") + "\t" + codingReads.size() + "\t" + codingInputReads.size() + "\t" + format.format(readsPVal) + "\t" + format.format(inputPVal) + "\n");
		    }
		    s = e; //Move the window forward
		    e += windowLength;
		}
	    }
	    bw.close();
	}
	catch (IOException ex) {

	}
    }

    ///////////////////////////////////////////////////////

    public static MiRNAInfo readMiRNAInfo(int c) {
	MiRNAInfo miRNAInfo = new MiRNAInfo(c);
	try {
	    BufferedReader br = new BufferedReader(new FileReader(GENOME_LOC + "miRNA.txt"));
	    String line = br.readLine();
	    //- chr start end name  - strand
	    Pattern mirnaPattern = Pattern.compile("\\d+\\s+chr([X|Y|\\d]+)\\s+(\\d+)\\s+(\\d+)\\s+([-|\\w]+)\\s+\\d+\\s+([-|+])");
	    while (line!=null) {
		Matcher m = mirnaPattern.matcher(line);
		if (m.find()) {
		    int c2;
		    if (m.group(1).equals("X"))
			c2 = 20;
		    else if (m.group(1).equals("Y"))
			c2 = 21;
		    else
			c2 = Integer.parseInt(m.group(1));
		    if (c==c2) {
			boolean strand = m.group(5).equals("+");
			int start = strand ? Integer.parseInt(m.group(2)) : Integer.parseInt(m.group(3));
			int end = strand ? Integer.parseInt(m.group(3)) : Integer.parseInt(m.group(2));
			miRNAInfo.add(new MiRNAInfoEntry(c, start, end, strand, m.group(4)));
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { }
	System.out.println("Found " + miRNAInfo.size() + " miRNAs for chromosome " + c);
	return miRNAInfo;
    }

    public static void saveMiRNAList(int c, MiRNAInfo miRNAInfo, TreeSet regs, String saveFileName) {
	//use this method to find the peaks which are close to known
	//miRNAs. Parse the miRNA info using readMiRNAInfo()
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    if (c==1) //write the header info
		bw.write("#" + ChromosomeWindow.getHeaderString() + "\tDistance\tU/D\t" + MiRNAInfoEntry.getHeaderString() + "\n");
	    Iterator it = regs.iterator();
	    while (it.hasNext()) {
		ChromosomeWindow e = (ChromosomeWindow)it.next();
		if (e.getNearestMiRNADistance()<10*promoterLength) 
		    bw.write(e.toString() + "\t" + e.getNearestMiRNADistance() + "\t" + (e.getNearestMiRNA().peakUpstream(e.getCenterOfMass()) ? "U" : "D") + "\t" + e.getNearestMiRNA().toString() + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) { }
    }


}


