//Use this method to analyze the distribution of short RNAs. In
//particular, we want to test if there is a strong correlation to the
//PALRs. The first attempt to do this is to 

//Example:

//java -Xmx512m AnalyzePASRDistribution 
//
//Use analyze_pasr_distribution.m for visualizing and post-processing

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class AnalyzePASRDistribution {

    public static final String PASR_FILE_NAME = "pasr_densities";

    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option rnaSeqFileNameFlag = argParser.addStringOption('R', "RNASeq");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('a', "saveFileName");
	    //Determine how wide the bins that we study should be
	    CmdLineParser.Option binWidthFlag = argParser.addIntegerOption('w', "binWidth");
	    //use this flag when we only want to focus on features
	    //which are guaranteed to be isolated from one another.
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('I', "isolated");
	    //Use this flag to include a file with genes/peaks to filter against
	    CmdLineParser.Option selectFlag = argParser.addStringOption('s', "select");
	    //Use this flag for mRNA instead of total RNA
	    CmdLineParser.Option mRNAFlag = argParser.addBooleanOption('m', "mRNA");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    boolean mRNA = argParser.getOptionValue(mRNAFlag)!=null;
	    tmp = argParser.getOptionValue(rnaSeqFileNameFlag);
	    String rnaSeqFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/hr0s.tsv";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/" + PASR_FILE_NAME + (mRNA ? "_mrna" : "") + (isolated ? "_isolated" : "");
	    tmp = argParser.getOptionValue(binWidthFlag);
	    int binWidth = tmp!=null ? ((Integer)tmp).intValue() : 10;
	    tmp = argParser.getOptionValue(selectFlag);
    	    String selectFileName = tmp!=null ? (String)tmp : null;
	    /////////////////////////////////
	    int nChr = 21;
	    String transcriptStartFileName = "mRNA/hr0c_starts_n_stops_TSS_13kb" + (mRNA ? "_mRNA" : "") + ".tsv";
	    int isolatedIDs[] = isolated ? FindReadsAcrossFeature.parseIsolatedFile(orchestra) : null;
	    //Load the 5'-ends called by JMG
	    int transcriptStarts[][] = FindReadsAcrossFeature.parseInitiationFile(transcriptStartFileName, isolatedIDs, false, false);
	    //Load the annotation info  and the mRNA
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    Vector<ABIRead[]> forwardReads = new Vector<ABIRead[]>(nChr), reverseReads = new Vector<ABIRead[]>(nChr);
	    forwardReads.setSize(nChr);
	    reverseReads.setSize(nChr);
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra);
		    //Load all the RNA reads
		    forwardReads.setElementAt(FindNongenicTranscripts.loadMRNAReadsChr(rnaSeqFileName, true, i), i-1);
		    reverseReads.setElementAt(FindNongenicTranscripts.loadMRNAReadsChr(rnaSeqFileName, false, i), i-1);
		    System.out.println("Loaded reads and annotation for chr " + i);
		}
	    }
	    TreeMap<String, double[][][]> subsetSenseDensities = new TreeMap<String, double[][][]>();
	    TreeMap<String, double[][][]> subsetAntisenseDensities = new TreeMap<String, double[][][]>();
	    Vector<String> subsets = EstimatePALRLength.parseSelectFiles(selectFileName);
	    int[] sBinVec = getBinVec(FindReadsAcrossFeature.MAX_TSS_DISTANCE_TRANSCRIPT_START, binWidth), asBinVec = getBinVec(FindReadsAcrossFeature.MAX_TSS_DISTANCE_TRANSCRIPT_START, binWidth);
	    if (subsets.size()>0) {
		for (String s : subsets) 
		    subsetSenseDensities.put(s, getDensity(forwardReads, reverseReads, chr, sBinVec, binWidth, s, orchestra, geneInfo, transcriptStarts));
	    }
	    //Look at everything
	    subsetSenseDensities.put("all/all.tsv", getDensity(forwardReads, reverseReads, chr, sBinVec, binWidth, null, orchestra, geneInfo, transcriptStarts));
	    
	    //Save all of them in files
	    System.out.println("Saving to " + saveFileName + "_sense" + (chr!=-1 ? "_" + chr : "") + ".txt");
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName + "_sense" + (chr!=-1 ? "_" + chr : "") + ".txt"));
	    if (subsetSenseDensities.size()>1) {
		bw.write("#bin_center(width=" + binWidth + ")");
		for (String s : subsetSenseDensities.keySet()) {
		    if (s.lastIndexOf(".")>=0)
			bw.write("\t" + s.substring(s.lastIndexOf("/") + 1, s.lastIndexOf(".")));
		}
		bw.write("\n");
	    }
	    for (int i=0; i<sBinVec.length; i++) {
		bw.write(format.format(sBinVec[i]));
		for (String s : subsetSenseDensities.keySet()) {
		    for (int j=0; j<subsetSenseDensities.get(s)[i].length; j++) {
			for (int k=0; k<subsetSenseDensities.get(s)[i][j].length; k++)
			    bw.write("\t" + format.format(subsetSenseDensities.get(s)[i][j][k]));
		    }
		}
		bw.write("\n");
	    }
	    bw.close();
	    /////////////////////////////////
	    System.out.println("Saving to " + saveFileName + "_antisense" + (chr!=-1 ? "_" + chr : "") + ".txt");
	    bw = new BufferedWriter(new FileWriter(saveFileName + "_antisense" + (chr!=-1 ? "_" + chr : "") + ".txt"));
	    bw.write("#bin_center(width=" + binWidth + ")");
	    for (String s : subsetAntisenseDensities.keySet()) {
		if (s.lastIndexOf(".")>=0)
		    bw.write("\t" + s.substring(s.lastIndexOf("/") + 1, s.lastIndexOf(".")));
	    }
	    bw.write("\n");
	    for (int i=0; i<asBinVec.length; i++) {
		bw.write(format.format(asBinVec[i]));
		for (String s : subsetAntisenseDensities.keySet()) {
		    for (int j=0; j<subsetSenseDensities.get(s)[i].length; j++)
			bw.write("\t" + format.format(subsetAntisenseDensities.get(s)[i][j]));
		}
		bw.write("\n");
	    }
	    bw.close();
	} 
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static int[] getBinVec(int len, int step) {
	int vec[] = new int[2*len/step + 1];
	vec[0] = -len;
	for (int i=1; i<vec.length; i++)
	    vec[i] = vec[i-1] + step;
	return vec;
    }

    public static double[][][] getDensity(Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, int chr, int binVec[], int binWidth, String selectFileName, boolean orchestra, RefSeqGenes geneInfo[], int[][] transcriptStarts) {
	//returned array: {smallS_noT, smallS_senseT, smallS_antiT,
	//smallS_bi}, same but for smallAS.
	int nChr = 21;
	int geneIDsKeep[] = FindReadsAcrossFeature.parseSelectFile(selectFileName);
	/////////////////////////////////
	double density[][][][] = new double[nChr][binVec.length][4][3];
	int nGenes[] = new int[nChr], nGenesSum = 0;
	for (int i=1; i<=nChr; i++) {
	    if (chr==-1 || i==chr) {
		//System.out.println("Getting info for chr " + i);
		nGenes[i-1] = geneInfo[i-1].getStarts(geneIDsKeep).length;
		nGenesSum += nGenes[i - 1];
		density[i-1] = getDensity(forwardReads.elementAt(i-1), reverseReads.elementAt(i-1), geneInfo[i-1], geneIDsKeep, binVec, binWidth, transcriptStarts);
	    }
	}
	double averageDensity[][][] = new double[binVec.length][4][3];
	for (int i=0; i<nChr; i++) {
	    for (int k=0; k<binVec.length; k++) {
		for (int j=0; j<4; j++) {
		    if (chr==-1) { //Calculate the weighted average
			averageDensity[k][j][0] += nGenesSum>0 ? density[i][k][j][0]*nGenes[i]/nGenesSum : 0;
			averageDensity[k][j][1] += nGenesSum>0 ? density[i][k][j][1]*nGenes[i]/nGenesSum : 0;
			averageDensity[k][j][2] += nGenesSum>0 ? density[i][k][j][2]*nGenes[i]/nGenesSum : 0;
		    }
		    else if (i==chr-1)
			averageDensity[k] = density[chr - 1][k];
		}
	    }
	}
	return averageDensity;
    }

    public static double[][][] getDensity(ABIRead forwardReads[], ABIRead reverseReads[], RefSeqGenes geneInfo, int geneIDsKeep[], int binVec[], int binWidth, int transcriptStarts[][]) {
	//returned array: {smallS_noT, smallS_senseT, smallS_antiT,
	//smallS_bi}, same but for smallAS.
	double density[][][] = new double[binVec.length][4][3];
	int nGenes[][] = new int[binVec.length][4];
	for (int i=0; i<binVec.length; i++) {
	    for (RefSeqGene g : geneInfo.getGenes()) {
		if (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) {
		    //Find the part covered by S, AS, both and none
		    boolean senseTranscript = false, antisenseTranscript = false;
		    int ind = Arrays.binarySearch(transcriptStarts[0], g.getID());
		    int startSense = Integer.MIN_VALUE, startAntisense = Integer.MAX_VALUE;
		    if (ind>=0) {
			startSense = Math.max(transcriptStarts[1][ind], -FindReadsAcrossFeature.MAX_TSS_DISTANCE_TRANSCRIPT_START);
			startAntisense = Math.min(transcriptStarts[2][ind], FindReadsAcrossFeature.MAX_TSS_DISTANCE_TRANSCRIPT_START);
		    }
		    double densS = getDensity(g.getChromosome(), g.getStrand(), g.getStart(), binVec[i], binWidth, g.getStrand() ? forwardReads : reverseReads);
		    double densAS = getDensity(g.getChromosome(), g.getStrand(), g.getStart(), binVec[i], binWidth, g.getStrand() ? reverseReads : forwardReads);
		    senseTranscript = startSense>binVec[i];
		    antisenseTranscript = startAntisense<binVec[i];
		    if (!senseTranscript && !antisenseTranscript) { 
			density[i][0][0] += densS;
			density[i][0][1] += densAS; 
			density[i][0][2] += densS*densAS; //Math.min(densAS, densS); 
			nGenes[i][0]++; 
		    }
		    if (senseTranscript) { 
			density[i][1][0] += densS; 
			density[i][1][1] += densAS; 
			density[i][1][2] += densS*densAS; //Math.min(densAS, densS); 
			nGenes[i][1]++;
		    }
		    if (antisenseTranscript) { 
			density[i][2][0] += densS; 
			density[i][2][1] += densAS; 
			density[i][2][2] += densS*densAS; //Math.min(densAS, densS); 
			nGenes[i][2]++;
		    }
		    if (senseTranscript && antisenseTranscript) { 
			density[i][3][0] += densS; 
			density[i][3][1] += densAS; 
			density[i][3][2] += densS*densAS; //Math.min(densAS, densS); 
			nGenes[i][3]++;
		    //System.out.println("Found overlapping transcripts at " + binVec[i] + " for " + g.toString() + ", " + g.getID());
		    }
		}
	    }
	    for (int j=0; j<density[i].length; j++) {
		density[i][j][0] /= nGenes[i][j]>0 ? nGenes[i][j] : 1;
		density[i][j][1] /= nGenes[i][j]>0 ? nGenes[i][j] : 1;
		density[i][j][2] /= nGenes[i][j]>0 ? nGenes[i][j] : 1;
	    }
	    //System.out.println("Found " + Arrays.toString(nGenes[i]) + " genes with transcripts at " + binVec[i]);
	}
	//System.out.println("Found densities " + Arrays.toString(density));
	return density;
    }
    
    
    public static double getDensity(int chr, boolean strand, int tss, int bin, int binWidth, ABIRead[] reads) {
	//Find the absolute location of the bin
	int start, end;
	if (strand) {
	    start = tss + bin - binWidth/2;
	    end = tss + bin + binWidth/2;
	}
	else {
	    start = tss - bin - binWidth/2;
	    end = tss - bin + binWidth/2;
	}
	return Math.log(1 + (double)FindReadsAcrossFeature.numberOfReads(chr, start, end, reads)/binWidth);
    }

}
