//Use this class to calculate the critical cut-off for the difference
//between ChIP and input reads. Since there are difference between
//windows depending on their mappability and their GC-content, we
//split the parameter space into smaller parts and we compute the
//theoretical and empirical distributions leading to the cut-off
//differently for each grid point.
//
//This class runs one job for each chromosome and calculates the
//empirical CDF as well as the Poisson parameters for each grid-point
//for a given experiment. The results can then be compiled using the
//matlab file calculate_critical_window_difference.m and the cut-off
//values can be used by the PeakFinder2.java-class.
//
//Example:
//
//>java -Xmx512m CalculateCriticalWindowDifference -n Npas4/KCl/Npas4_KCl -i Mef2/inputKCl/inputKCl -w 240 -W 10 -m 4 -g 4 -q 10 -c 19 -f cutOffDist -u -e 
//
//In matlab we can subsequently run:
//
//>>z = calculate_critical_window_difference('../Npas4/KCl/Npas4_KCl', .001)
//
//For the Mikkelsen et al (2007) data from the Broad:
//
//java -Xmx512m CalculateCriticalWindowDifference -n /groups/neuroduo/martin/ChIP-Seq/MikkelsenData/ES/H3/ES_H3 -i /groups/neuroduo/martin/ChIP-Seq/MikkelsenData/ES/Input/ES_Input -w 1000 -W 100 -m 4 -g 4 -q 10 -c 19 -u -e -l 300 -o

import jargs.gnu.CmdLineParser;
import java.io.*;
import java.util.*;
import java.text.DecimalFormat;
import cern.jet.random.Poisson;
//import umontreal.iro.lecuyer.stat.Tally;
//import org.apache.commons.math.stat.descriptive.SummaryStatisticsImpl;


public class CalculateCriticalWindowDifference {

    public static DecimalFormat format = new DecimalFormat("###.########");

    public static void main(String args[]) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    //Use this flag to for the location of the .fa-files
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    //Use these flags when we are comparing two samples
	    CmdLineParser.Option fileName2Flag = argParser.addStringOption('N', "fileName2");
	    CmdLineParser.Option inputFileName2Flag = argParser.addStringOption('I', "inputFileName2");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('l', "fragmentLength");
	    CmdLineParser.Option uniqueFlag = argParser.addBooleanOption('u', "unique");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    //Use this flag if we want to extend the reads to length fragmentLength
	    CmdLineParser.Option extendFlag = argParser.addBooleanOption('e', "extend");
	    //Use this flag to decide which species to
	    //investigate. Available species are: 
	    //Mouse (Mm) Default
	    //Yeast (Sc)
	    CmdLineParser.Option speciesFlag = argParser.addStringOption('s', "species");
	    //Decide what file to write the results to
	    CmdLineParser.Option saveFileFlag = argParser.addStringOption('f', "saveFile");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option windowStepFlag = argParser.addIntegerOption('W', "windowStep");
	    //Remove the reads which don't meet the quality threshold
	    CmdLineParser.Option qualityFlag = argParser.addDoubleOption('q', "quality");
	    //Use this flag to determine the number of bins for the mappability index
	    CmdLineParser.Option mappabilityFlag = argParser.addIntegerOption('m', "mappabilityBins");
	    //Number of bins for the GC-bias
	    CmdLineParser.Option gcFlag = argParser.addIntegerOption('g', "gcBins");
	    //Use this flag when we want to calculate the z-scores for
	    //the coding regions
	    CmdLineParser.Option annotationFlag = argParser.addBooleanOption('a', "annotation");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    String GENOME_LOC = "./";
	    int nChr = 0;
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    boolean unique = argParser.getOptionValue(uniqueFlag)!=null;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean extend = argParser.getOptionValue(extendFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(fileName2Flag);
	    String fileName2 = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(inputFileName2Flag);
	    String inputFileName2 = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 480;
	    tmp = argParser.getOptionValue(windowStepFlag);
	    int windowStep = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(mappabilityFlag);
	    int nMappabilityBins = tmp!=null ? ((Integer)tmp).intValue() : 4;
	    int mappabilityBinSize = 2*(windowSize - 120)/nMappabilityBins;
	    tmp = argParser.getOptionValue(gcFlag);
	    int nGCBins = tmp!=null ? ((Integer)tmp).intValue() : 4;
	    int gcBinSize = windowSize/nGCBins;
	    tmp = argParser.getOptionValue(saveFileFlag);
	    boolean annotation = argParser.getOptionValue(annotationFlag)!=null;
	    String saveFileStart = tmp!=null && !tmp.toString().equals("") ? tmp.toString() : fileName;
	    //saveFileStart += "_";
	    tmp = argParser.getOptionValue(qualityFlag);
	    double qualityScore = tmp!=null ? ((Double)tmp).doubleValue() : 10;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? tmp.toString() : "Mm";
	    if (species.equals("Mm")) {
		GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
		nChr = 21;
	    }
	    else if (species.equals("Sc")) {
		GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Sc/sacCer1/" : "/media/exthd/Genomes/Sc/sacCer1/";
		nChr = 17;
	    }
	    else if (species.equals("Hs")) {
		GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Hs/v37/" : "/media/exthd/Genomes/Hs/v37/";
		nChr = 24;
	    }
	    boolean hs = species.equals("Hs");
	    //////////////////////////////////////////////////////////////////
	    int nWindows[][] = new int[nMappabilityBins][nGCBins], nReads[][] = new int[nMappabilityBins][nGCBins], nInputReads[][] = new int[nMappabilityBins][nGCBins], readsInd = 0, inputInd = 0;
	    //Keep track of the smallest and largest keys
	    int minZ = Integer.MAX_VALUE, maxZ = Integer.MIN_VALUE, minChip = Integer.MAX_VALUE, maxChip = Integer.MIN_VALUE, minInput = Integer.MAX_VALUE, maxInput = Integer.MIN_VALUE;
	    //Histogram for keeping track of the empirical CDF
	    TreeMap zHistogram[][] = new TreeMap[nMappabilityBins][nGCBins], chipHistogram[][] = new TreeMap[nMappabilityBins][nGCBins], inputHistogram[][] = new TreeMap[nMappabilityBins][nGCBins];
	    for (int i=0; i<nMappabilityBins; i++) {
		for (int j=0; j<nGCBins; j++) {
		    nWindows[i][j] = 0;
		    nReads[i][j] = 0;
		    nInputReads[i][j] = 0;
		    zHistogram[i][j] = new TreeMap();
		    chipHistogram[i][j] = new TreeMap();
		    inputHistogram[i][j] = new TreeMap();
		}
	    }
	    Mappability mappability = new Mappability(orchestra, chr, windowSize, fragmentLength, species.toLowerCase());
	    FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(chr, species));
	    ABIRead reads[] = null, inputReads[] = null;
	    if (fileName.contains("MikkelsenData")) {
		reads = PeakFinder2.parseBroadFiles(chr, fileName, fragmentLength, extend);
		inputReads = PeakFinder2.parseBroadFiles(chr, inputFileName, fragmentLength, extend);
	    }
	    else if (fileName.endsWith(".bed")) {
		reads = CreateWigFiles.parseBedFiles(chr, fileName, true, fragmentLength);
		inputReads = CreateWigFiles.parseBedFiles(chr, inputFileName, true, fragmentLength);
	    }
	    else {
		reads = PeakFinder2.parseFiles(chr, fileName, fragmentLength, extend, 0, qualityScore, AnalyzeControls.getSaveFileName(fileName, "read_stats", chr, fragmentLength, "txt"));
		inputReads = PeakFinder2.parseFiles(chr, inputFileName, fragmentLength, extend, 0, qualityScore, AnalyzeControls.getSaveFileName(inputFileName, "read_stats", chr, fragmentLength, "txt"));
	    }
	    //For saving the entire distribution and just not the summary
	    System.out.println("Saving " + reads.length + " reads to " + saveFileStart + "_win_dist_" + chr + ".txt");
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileStart + "_win_dist_" + chr + ".txt"));
	    ChromosomeWindow win = new ChromosomeWindow(chr, 0, windowSize), prevWin = null;
	    win.setSequence(fastaSeq.getNextSequence(1, windowSize));
	    while (win.getEnd()<(hs ? PeakFinder2.CHROMOSOME_LENGTH_HS[chr-1] : PeakFinder2.CHROMOSOME_LENGTH[chr-1])) {
		win.shiftRight(windowStep);
		while (readsInd<reads.length && reads[readsInd].getPosition()<=win.getStart() + windowSize - reads[readsInd].getLength()) //Found read going into the window
		    win.add(reads[readsInd++]);
		while (inputInd<inputReads.length && inputReads[inputInd].getPosition()<=win.getStart() + windowSize - inputReads[inputInd].getLength()) //Found input read going into the window
		    win.addInputRead(inputReads[inputInd++]);
		if (inputReads.length==0) {//No input available
		    int poi = Poisson.staticNextInt(1); //Get a random variable instead
		    //System.out.println(poi);
		    for (int i=0; i<poi; i++)
			win.addInputRead(new ABIRead(chr, true, win.getStart() + i));
		}
		win.addSequence(fastaSeq.getNextSequence(win.getEnd(), windowStep)); 
		//Calculate the potential number of reads
		win.setPotentialNumberOfReads(mappability.numberOfPotentialReadsNextWindow(windowStep));
		if (win.numberOfPotentialReads()>0) {
		    //Save the information about the window in a file
		    //System.out.println("Comparing " + win.toString() + "\n" + prevWin.toString());
		    bw.write(win.getStart() + "\t" + win.numberOfPotentialReads() + "\t" + win.gcContent() + "\t" + win.numberOfReads() + "\t" + win.numberOfNewReads(prevWin) + "\t" + win.numberOfInputReads() + "\t" + win.numberOfNewInputReads(prevWin) + "\n");
		    /*if (win.getStart()>=3242200 && win.getStart()<=3242500) {
			ABIRead rs[] = win.getReadsArray();
			System.out.println(win.getStart() + "-" + win.getEnd());
			for (int i=0; i<rs.length; i++)
			    System.out.println(rs[i].toString());
			    }*/
			
		    int mappabilityBinIndex = win.numberOfPotentialReads()/mappabilityBinSize;
		    if (mappabilityBinIndex==nMappabilityBins) { mappabilityBinIndex--; } //Adjust for the last bin
		    int gcBinIndex = win.gcContent()/gcBinSize;
		    //System.out.println(win.numberOfPotentialReads() + " " + mappabilityBinIndex + " " + win.gcContent() + " " + gcBinIndex);
		    if (gcBinIndex==nGCBins) { gcBinIndex--; } //Adjust for the last bin
		    nWindows[mappabilityBinIndex][gcBinIndex]++;
		    nReads[mappabilityBinIndex][gcBinIndex] += win.numberOfReads();
		    nInputReads[mappabilityBinIndex][gcBinIndex] += win.numberOfInputReads();
		    Integer z = new Integer(win.numberOfReads() - win.numberOfInputReads());
		    //System.out.println("Window " + win.getStart() + " with z = " + z + ", GC = " + win.gcContent() + " has mapInd = " + mappabilityBinIndex + " and gcInd = " + gcBinIndex);
		    int val = zHistogram[mappabilityBinIndex][gcBinIndex].containsKey(z) ? ((Integer)zHistogram[mappabilityBinIndex][gcBinIndex].get(z)).intValue() : 0;
		    zHistogram[mappabilityBinIndex][gcBinIndex].put(z, new Integer(val + 1));
		    minZ = minZ>z.intValue() ? z.intValue() : minZ;
		    maxZ = maxZ<z.intValue() ? z.intValue() : maxZ;
		    //Store the number of input and chip reads
		    Integer key = new Integer(win.numberOfReads());
		    val = chipHistogram[mappabilityBinIndex][gcBinIndex].containsKey(key) ? ((Integer)chipHistogram[mappabilityBinIndex][gcBinIndex].get(key)).intValue() : 0;
		    chipHistogram[mappabilityBinIndex][gcBinIndex].put(key, new Integer(val + 1));
		    minChip = minChip>win.numberOfReads() ? win.numberOfReads() : minChip;
		    maxChip = maxChip<win.numberOfReads() ? win.numberOfReads() : maxChip;
		    key = new Integer(win.numberOfInputReads());
		    val = inputHistogram[mappabilityBinIndex][gcBinIndex].containsKey(key) ? ((Integer)inputHistogram[mappabilityBinIndex][gcBinIndex].get(key)).intValue() : 0;
		    inputHistogram[mappabilityBinIndex][gcBinIndex].put(key, new Integer(val + 1));
		    minInput = minInput>win.numberOfInputReads() ? win.numberOfInputReads() : minInput;
		    maxInput = maxInput<win.numberOfInputReads() ? win.numberOfInputReads() : maxInput;
		}
		prevWin = new ChromosomeWindow(win);
	    }
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileStart + "_win_stat_diff_" + chr + ".txt"));
	    bw.write("#diff"); //write a header
	    for (int i=0; i<nMappabilityBins; i++) {
		for (int j=0; j<nGCBins; j++) 
		    bw.write("\tm=" + ((i+1)*mappabilityBinSize) + ",g=" + ((j+1)*gcBinSize));
	    }
	    bw.write("\n");
	    saveHistograms(bw, nMappabilityBins, nGCBins, minZ, maxZ, zHistogram);
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileStart + "_win_stat_chip_" + chr + ".txt"));
	    bw.write("#chip"); //write a header
	    for (int i=0; i<nMappabilityBins; i++) {
		for (int j=0; j<nGCBins; j++) 
		    bw.write("\tm=" + ((i+1)*mappabilityBinSize) + ",g=" + ((j+1)*gcBinSize));
	    }
	    bw.write("\n");
	    saveHistograms(bw, nMappabilityBins, nGCBins, minChip, maxChip, chipHistogram);
	    bw.close();
	    bw = new BufferedWriter(new FileWriter(saveFileStart + "_win_stat_input_" + chr + ".txt"));
	    bw.write("#input"); //write a header
	    for (int i=0; i<nMappabilityBins; i++) {
		for (int j=0; j<nGCBins; j++) 
		    bw.write("\tm=" + ((i+1)*mappabilityBinSize) + ",g=" + ((j+1)*gcBinSize));
	    }
	    bw.write("\n");
	    saveHistograms(bw, nMappabilityBins, nGCBins, minInput, maxInput, inputHistogram);
	    bw.close();
	    if (annotation) //Save the number of reads for each gene as well
		WiderPeakLocFDR.getZValues(chr, fileName, new RefSeqGenes(chr, orchestra), windowSize, windowStep, nMappabilityBins, nGCBins);
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    protected static void saveHistograms(BufferedWriter bw, int nMappabilityBins, int nGCBins, int min, int max, TreeMap histogram[][]) throws IOException {
	//Save the empircal PDF
	for (int k=min; k<=max; k++) {
	    bw.write((new Integer(k)).toString());
	    for (int i=0; i<nMappabilityBins; i++) {
		for (int j=0; j<nGCBins; j++) 
		    bw.write("\t" + getWindowCount(histogram[i][j], k)); 
	    }
	    bw.write("\n");
	}
    }


    protected static long getWindowCount(TreeMap zHistogram, int k) {
	//If the key k does not exist, get the one closest below
	if (zHistogram.containsKey(new Integer(k))) {
	    //System.out.println(k + "\t" + ((Integer)zHistogram.get(new Integer(k))).intValue());
	    try {
		return (long)((Integer)zHistogram.get(new Integer(k))).intValue();
	    }
	    catch (ClassCastException e) {
		return ((Long)zHistogram.get(new Integer(k))).longValue();
	    }
	}
	return 0;
    }

}
