//Use this method to extract 5' ends for CAGE data, EST data and our
//RNA Seq data (the called transcripts). Write to a file which can be
//used for plotting the densities relative to the TSS

//Example:

//java -Xmx512m CompareCageEstRNASeqStarts

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class CompareCageEstRNASeqStarts {

    public static final String CAGE_DATABASE_FILE_SUFFIX = "_mm9.gff";
    public static final String CAGE_DATABASE_FILE_PREFIX = "cage.rep_tag.chr";
    public static final String ALL_EST_POSITIONS_FILE_NAME = "all_est_positions_mm7.tsv";
    public static final String ALL_EST_POSITIONS_FILE_NAME_HS = "ESTs_all_hs36_2010_07_19.gtf";
    public static final String EST_VALID_5P_FILE_NAME = "valid_accession_numbers_est_5p_reads.tsv";
    public static final String EST_VALID_5P_FILE_NAME_HS = "ESTs_5p_hs36_2010_07_19_valid_accessions.gtf";
    public static final String EST_SPLICE_INFO_FILE_NAME = "est_splice_info.tsv";
    public static final String EST_SPLICE_INFO_FILE_NAME_HS = "ESTs_all_hs36_2010_07_19_splice_info.gtf";
    public static final String FIVE_PRIME_END_START_SUFFIX = "_five_prime_ends_";
    public static final String CAGE_EST_RNA_COMPARISON_FILE_NAME = "CAGE_EST_RNASeq_TSS_comparison";
    public static final String UPSTREAM_CONVERGENT_SENSE_SUFFIX = "_upstream_convergent_sense";

    //chr start end strand ID
    public static final String CAGE_PATTERN_STR = "^chr([\\d|X|Y]+)\\s+CAGE_tag\\s+expressed_sequence_match\\s+(\\d+)\\s+(\\d+)\\s+[\\d|\\.]+\\s+([\\+|-])\\s+\\.\\s+ID=(\\w+)";
    //chr start
    public static final String CAGE_CORTEX_PATTERN_STR = "^chr([\\d|X|Y]+)\\s\\d+\\s+(\\d+)";
    //id rna_lib_id tag_id
    public static final String CAGE_POLYA_PATTERN_STR = "^>(\\w+)\\s+rna_lib_id=(\\w+);rep_tag_id=(\\w+)";
    /////////////////////////////////
    //chr start end code strand
    public static final String ALL_EST_POSITIONS_PATTERN_MM_STR = "^chr([\\d|X|Y]+)[\\s|\\w|_]*\\s+(\\d+)\\s+(\\d+)\\s+(\\w+)\\s+\\d+\\s+([\\+|-])";
    public static final String EST_VALID_5P_PATTERN_HS_STR = "^chr([\\d|X|Y]+)\\s+[\\w|_]+\\s+[\\w|_]+\\s+(\\d+)\\s+(\\d+)\\s+[\\d|\\.]+\\s+([\\+|-]).+transcript_id\\s+\\\"(\\w+)\\\"";
    //#id     acc     version moddate type    direction       source  organism        library mrnaClone       sex     tissue  development     cell  cds      keyword description     geneName        productName     author  gi      mol
    public static final String EST_VALID_5P_PATTERN_MM_STR = "^\\d+\\s+(\\w+)";
    //n_splices code 
    public static final String EST_SPLICE_INFO_PATTERN_STR = "^\\d+\\s+\\d+\\s+\\d+\\s+\\d+\\s+\\d+\\s+\\d+\\s+\\d+\\s+(\\d+)\\s+\\d+\\s+[\\+|-]\\s+(\\w+)";
    public static final String EST_TISSUE_TYPE_FILE = "/groups/neuroduo/jesse/human_annot_for_mh/EST_gbCdnaInfo_hg18_2010_08_02.tsv";
    public static final String EST_TISSUE_TYPE_KEY_FILE = "/groups/neuroduo/jesse/UCSC/mysql_queries/tissue_key_for_ESTs.tsv";
    //id     acc     version moddate type    direction       source  organism        library mrnaClone       sex     tissue  development     cell    cds     keyword     description     geneName        productName     author  gi      mol
    public static final String EST_TISSUE_TYPE_PATTERN = "\\d+\\s+(\\w+)\\s+\\d+\\s+[\\d|-]+\\s+\\w+\\s+\\d+\\s+\\d+\\s+(\\d+)\\s+\\d+\\s+\\d+\\s+\\d+\\s+(\\d+)";
    

    public static final int MAX_TSS_DISTANCE = 10000;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option cageFileNameFlag = argParser.addStringOption('C', "CAGE");
	    CmdLineParser.Option estFileNameFlag = argParser.addStringOption('E', "EST");
	    CmdLineParser.Option rnaSeqFileNameFlag = argParser.addStringOption('R', "RNASeq");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option upstreamFlag = argParser.addIntegerOption('u', "upstream");
	    CmdLineParser.Option downstreamFlag = argParser.addIntegerOption('d', "downstream");
	    //use this flag when we only want to focus on features
	    //which are guaranteed to be isolated from one another.
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('I', "isolated");
	    //Use only the cortex CAGE tags
	    CmdLineParser.Option cortexFlag = argParser.addBooleanOption('x', "cortex");
	    //Skip the EST stuff (for faster calculations)
	    CmdLineParser.Option skipEstFlag = argParser.addBooleanOption('e', "skipEST");
	    //Use only the CAGE tags which have the polyA mark
	    CmdLineParser.Option polyAFlag = argParser.addBooleanOption('p', "polyA");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    boolean cortex = argParser.getOptionValue(cortexFlag)!=null;
	    boolean polyA = argParser.getOptionValue(polyAFlag)!=null;
	    boolean skipEst = argParser.getOptionValue(skipEstFlag)!=null;
	    tmp = argParser.getOptionValue(cageFileNameFlag);
	    String cageFileName = tmp!=null ? (String)tmp : polyA ? "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/CAGE/mm9/" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "deepCAGE/mm9/";
	    tmp = argParser.getOptionValue(estFileNameFlag);
	    String estFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/ESTs/" : null);
	    tmp = argParser.getOptionValue(rnaSeqFileNameFlag);
	    String rnaSeqFileName = tmp!=null ? (String)tmp : "mRNA/hr0c_starts_n_stops_TSS_13kb.tsv";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "deepCAGE/" + CAGE_EST_RNA_COMPARISON_FILE_NAME + (polyA ? "_polyA" : "") + (isolated ? "_isolated" : "") + (chr!=-1 ? "_" + chr : "");
	    tmp = argParser.getOptionValue(upstreamFlag);
	    int upstream = tmp!=null ? ((Integer)tmp).intValue() : 5000;
	    tmp = argParser.getOptionValue(downstreamFlag);
	    int downstream = tmp!=null ? ((Integer)tmp).intValue() : 5000;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    //////////////////////////////////////////////////////////////////
	    int nChr = 21;
	    int isolatedIDs[] = isolated ? FindReadsAcrossFeature.parseIsolatedFile(orchestra) : null;
	    //Key is position relative to TSS and value is number of tags
	    TreeMap<Integer, Integer> cageTagsS = new TreeMap<Integer, Integer>(), cageTagsAS = new TreeMap<Integer, Integer>();
	    TreeMap<Integer, Integer> estTagsS = new TreeMap<Integer, Integer>(), estTagsAS = new TreeMap<Integer, Integer>();
	    TreeMap<Integer, Integer> rnaSeqS = new TreeMap<Integer, Integer>(), rnaSeqAS = new TreeMap<Integer, Integer>();
	    //Key is tss and value is a vector of observed 5' ends
	    TreeMap<Integer, Vector<Integer>> fivePrimeCageEndsSenseNearTSS = new TreeMap<Integer, Vector<Integer>>(), fivePrimeCageEndsAntiSenseNearTSS = new TreeMap<Integer, Vector<Integer>>();
	    TreeMap<Integer, Vector<Integer>> fivePrimeEstEndsSenseNearTSS = new TreeMap<Integer, Vector<Integer>>(), fivePrimeEstEndsAntiSenseNearTSS = new TreeMap<Integer, Vector<Integer>>();
	    //Keep track of all ESTs that are upstream and convergent
	    TreeMap<String, Integer> upstreamConvergentEstSense = new TreeMap<String, Integer>(); //key is id and value is distance to tss
	    int nCageTagsS = 0, nCageTagsAS = 0;
	    HashSet<String> polyATagsKeepIDs = polyA ? filterCageTags("/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/CAGE/cage.all_tag.2004-11-16.fasta") : null;
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    System.out.println("Getting info for chr " + i);
		    //Load the annotation info
		    RefSeqGenes geneInfo = new RefSeqGenes(i, orchestra);
		    //Get the Cage tags
		    Vector<Integer> cageTagsForwardList = parseCageTags(cageFileName, i, true, cortex, polyATagsKeepIDs);
		    //Find the gene which it is closest to
		    for (int c : cageTagsForwardList) { 
			int tssS = geneInfo.getNearestTSS(c, true, true), tssAS = geneInfo.getNearestTSS(c, true, false);
			int nearestSenseTSS = c - tssS;
			int nearestAntiSenseTSS = tssAS - c;
			if (Math.abs(nearestSenseTSS)<Math.abs(nearestAntiSenseTSS)) {
			    cageTagsS.put(nearestSenseTSS, cageTagsS.containsKey(nearestSenseTSS) ? cageTagsS.get(nearestSenseTSS) + 1 : 1);
			    int id = geneInfo.getNearestGene(tssS).getID();
			    fivePrimeCageEndsSenseNearTSS = addFivePrimeEnd(nearestSenseTSS, id, fivePrimeCageEndsSenseNearTSS);
			}
			else {
			    cageTagsAS.put(nearestAntiSenseTSS, cageTagsAS.containsKey(nearestAntiSenseTSS) ? cageTagsAS.get(nearestAntiSenseTSS) + 1 : 1);
			    int id = geneInfo.getNearestGene(tssAS).getID();
			    fivePrimeCageEndsAntiSenseNearTSS = addFivePrimeEnd(nearestAntiSenseTSS, id, fivePrimeCageEndsAntiSenseNearTSS);
			}
		    }
		    Vector<Integer> cageTagsReverseList = parseCageTags(cageFileName, i, false, cortex, polyATagsKeepIDs);
		    //Find the gene which it is closest to
		    for (int c : cageTagsReverseList) { 
			int tssS = geneInfo.getNearestTSS(c, true, false), tssAS = geneInfo.getNearestTSS(c, true, true);
			int nearestSenseTSS = tssS - c;
			
			int nearestAntiSenseTSS = c - tssAS;
			if (Math.abs(nearestSenseTSS)<Math.abs(nearestAntiSenseTSS)) 
			    cageTagsS.put(nearestSenseTSS, cageTagsS.containsKey(nearestSenseTSS) ? cageTagsS.get(nearestSenseTSS) + 1 : 1);
			else
			    cageTagsAS.put(nearestAntiSenseTSS, cageTagsAS.containsKey(nearestAntiSenseTSS) ? cageTagsAS.get(nearestAntiSenseTSS) + 1 : 1);
			fivePrimeCageEndsSenseNearTSS = addFivePrimeEnd(nearestSenseTSS, tssS, fivePrimeCageEndsSenseNearTSS);
			fivePrimeCageEndsAntiSenseNearTSS = addFivePrimeEnd(nearestAntiSenseTSS, tssAS, fivePrimeCageEndsAntiSenseNearTSS);
		    }
		    /////////////////////////////////
		    //Save the histogram of the start sites
		    saveFivePrimeEnds(fivePrimeCageEndsSenseNearTSS, saveFileName + "_S_CAGE");
		    saveFivePrimeEnds(fivePrimeCageEndsAntiSenseNearTSS, saveFileName + "_AS_CAGE");
		    /////////////////////////////////
		    //Get the EST info
		    if (!skipEst) {
			LinkedHashMap<String, EST> estTagsForwardMap = parseEstTags(estFileName, i, true, species);
			//Collections.sort(estTagsForwardList);
			//Find the gene which it is closest to
			for (String k : estTagsForwardMap.keySet()) { 
			    int c = estTagsForwardMap.get(k).getStart();
			    int tssS = geneInfo.getNearestTSS(c, true, true), tssAS = geneInfo.getNearestTSS(c, true, false);
			    int nearestSenseTSS = c - tssS;
			    int nearestAntiSenseTSS = tssAS - c;
			    if (Math.abs(nearestSenseTSS)<Math.abs(nearestAntiSenseTSS)) {
				estTagsS.put(nearestSenseTSS, estTagsS.containsKey(nearestSenseTSS) ? estTagsS.get(nearestSenseTSS) + 1 : 1);
				if (nearestSenseTSS<0) //upstream convergent
				    upstreamConvergentEstSense.put(k, nearestSenseTSS);
			    }
			    else
				estTagsAS.put(nearestAntiSenseTSS, estTagsAS.containsKey(nearestAntiSenseTSS) ? estTagsAS.get(nearestAntiSenseTSS) + 1 : 1);
			    fivePrimeEstEndsSenseNearTSS = addFivePrimeEnd(nearestSenseTSS, tssS, fivePrimeEstEndsSenseNearTSS);
			    fivePrimeEstEndsAntiSenseNearTSS = addFivePrimeEnd(nearestAntiSenseTSS, tssAS, fivePrimeEstEndsAntiSenseNearTSS);
			}
			LinkedHashMap<String, EST> estTagsReverseMap = parseEstTags(estFileName, i, false, species);
			//Find the gene which it is closest to
			for (String k : estTagsReverseMap.keySet()) { 
			    int c = estTagsReverseMap.get(k).getStart();
			    int tssS = geneInfo.getNearestTSS(c, true, false), tssAS = geneInfo.getNearestTSS(c, true, true);
			    int nearestSenseTSS = tssS - c;
			    
			    int nearestAntiSenseTSS = c - tssAS;
			    if (Math.abs(nearestSenseTSS)<Math.abs(nearestAntiSenseTSS)) {
				estTagsS.put(nearestSenseTSS, estTagsS.containsKey(nearestSenseTSS) ? estTagsS.get(nearestSenseTSS) + 1 : 1);
				if (nearestSenseTSS<0) //upstream convergent
				    upstreamConvergentEstSense.put(k, nearestSenseTSS);
			    }
			    else
				estTagsAS.put(nearestAntiSenseTSS, estTagsAS.containsKey(nearestAntiSenseTSS) ? estTagsAS.get(nearestAntiSenseTSS) + 1 : 1);
			    fivePrimeEstEndsSenseNearTSS = addFivePrimeEnd(nearestSenseTSS, tssS, fivePrimeEstEndsSenseNearTSS);
			    fivePrimeEstEndsAntiSenseNearTSS = addFivePrimeEnd(nearestAntiSenseTSS, tssAS, fivePrimeEstEndsAntiSenseNearTSS);
			}
			/////////////////////////////////
			//Save the histogram of the start sites
			saveFivePrimeEnds(fivePrimeEstEndsSenseNearTSS, saveFileName + "_S_EST");
			saveFivePrimeEnds(fivePrimeEstEndsAntiSenseNearTSS, saveFileName + "_AS_EST");
		    }
		    /////////////////////////////////
		    //Compare the location of the CAGE tags with the one
		    //estimated from the WT-Seq data.
		    compareFivePrimeEnds(rnaSeqFileName, fivePrimeCageEndsSenseNearTSS, true, isolatedIDs, geneInfo);
		    compareFivePrimeEnds(rnaSeqFileName, fivePrimeCageEndsAntiSenseNearTSS, false, isolatedIDs, geneInfo);
		}
	    }
	    //Get the RNA Seq info
	    rnaSeqS = parseRNASeq(rnaSeqFileName, true, isolatedIDs);
	    rnaSeqAS = parseRNASeq(rnaSeqFileName, false, isolatedIDs);
	    //System.out.println("Found " + nCageTagsS + " CAGE sense tags and " + nCageTagsAS + " CAGE anti-sense tags.");
	    
	    System.out.println("Found sense CAGE tags for " + numberOfPromotersWithCageTags(fivePrimeCageEndsSenseNearTSS, true, false) + " genes and anti-sense CAGE tags for " + numberOfPromotersWithCageTags(fivePrimeCageEndsAntiSenseNearTSS, false, false));
	    System.out.println("Found sense CAGE tags in convergent region for " + numberOfPromotersWithCageTags(fivePrimeCageEndsSenseNearTSS, true, true) + " genes and anti-sense CAGE tags in convergent region for " + numberOfPromotersWithCageTags(fivePrimeCageEndsAntiSenseNearTSS, false, true));
	    System.out.println("There were " + numberOfCageTagsInPromoters(fivePrimeCageEndsSenseNearTSS, true, true) + " sense CAGE tags in convergent region and " + numberOfCageTagsInPromoters(fivePrimeCageEndsAntiSenseNearTSS, false, true) + " anti-sense CAGE tags in convergent region.");
	    /////////////////////////////////
	    //Save the histograms into one file
	    if (!skipEst) {
		System.out.println("Saving counts histogram to " + saveFileName + ".txt");
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName + ".txt"));
		bw.write("#Bin\tCAGE_S\tCAGE_AS\tEST_S\tEST_AS\tRNASeq_S\tRNASeq_AS\n");
		for (int i=-upstream; i<=downstream; i++) {
		    bw.write("" + i);
		    bw.write("\t" + (cageTagsS.containsKey(i) ? cageTagsS.get(i) : 0));
		    bw.write("\t" + (cageTagsAS.containsKey(i) ? cageTagsAS.get(i) : 0));
		    bw.write("\t" + (estTagsS.containsKey(i) ? estTagsS.get(i) : 0));
		    bw.write("\t" + (estTagsAS.containsKey(i) ? estTagsAS.get(i) : 0));
		    bw.write("\t" + (rnaSeqS.containsKey(i) ? rnaSeqS.get(i) : 0));
		    bw.write("\t" + (rnaSeqAS.containsKey(i) ? rnaSeqAS.get(i) : 0));
		    bw.write("\n");
		}
		bw.close();
		/////////////////////////////////
		System.out.println("Saving convergent upstream ESTs to " + saveFileName + UPSTREAM_CONVERGENT_SENSE_SUFFIX + ".txt");
		bw = new BufferedWriter(new FileWriter(saveFileName + UPSTREAM_CONVERGENT_SENSE_SUFFIX + ".txt"));
		for (String k : upstreamConvergentEstSense.keySet())
		    bw.write(k + "\t" + upstreamConvergentEstSense.get(k) + "\n");
		bw.close();
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static TreeMap<Integer, Vector<Integer>> addFivePrimeEnd(int dist, int id, TreeMap<Integer, Vector<Integer>> fivePrimeEndsNearTSS) {
	if (Math.abs(dist)<MAX_TSS_DISTANCE) {
	    if (!fivePrimeEndsNearTSS.containsKey(id))
		fivePrimeEndsNearTSS.put(id, new Vector<Integer>());
	    fivePrimeEndsNearTSS.get(id).addElement(dist);
	}
	return fivePrimeEndsNearTSS;
    }


    public static Vector<Integer> parseCageTags(String cageFileName, int chr, boolean forward, boolean cortex, HashSet<String> polyATagsKeepIDs) {
	//Return a list of CAGE tags found on each strand
	TreeMap<String, Integer> tagList = new TreeMap<String, Integer>(); //ID -> position
	try {
	    BufferedReader br = new BufferedReader(new FileReader(cageFileName + CAGE_DATABASE_FILE_PREFIX + AnalyzeControls.getChromosomeNumber(chr) + CAGE_DATABASE_FILE_SUFFIX));
	    String line = br.readLine();
	    Pattern cagePattern = Pattern.compile(CAGE_PATTERN_STR);
	    while (line!=null) {
		Matcher m = cagePattern.matcher(line);
		if (m.find()) {
		    boolean strand = m.group(4).equals("+");
		    if (strand==forward) 
			tagList.put(m.group(5), Integer.parseInt(m.group(forward ? 2 : 3)));
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading from Cage file : " + e.toString());
	}
	/*if (cortex) { //Now keep only those that are present in the cortex
	    Vector<Integer> somTagList = filterCageTags(tagList, cageFileName + "som_ctx" + (forward ? "+" : "-") + "_mm9.bed", chr);
	    Vector<Integer> visTagList = filterCageTags(tagList, cageFileName + "vis_ctx" + (forward ? "+" : "-") + "_mm9.bed", chr);
	    //Make sure that the tag was present in at least one of the sub-classes
	    Vector<Integer> removeSet = new Vector<Integer>();
	    for (Integer i : tagList) {
		if (!somTagList.contains(i) || visTagList.contains(i))
		    removeSet.addElement(i);
	    }
	    tagList.removeAll(removeSet);
	    }*/
	if (polyATagsKeepIDs!=null) { //Now keep only those that have the polyA mark
	    //Make sure that the tag was present in at least one of the sub-classes
	    Vector<String> removeSet = new Vector<String>();
	    for (String s : tagList.keySet()) {
		if (!polyATagsKeepIDs.contains(s))
		    removeSet.addElement(s);
	    }
	    for (String s : removeSet) {
		tagList.remove(s);
		polyATagsKeepIDs.remove(s);
	    }
	}
	System.out.println("Found " + tagList.size() + " " + (forward ? "forward" : " reverse") + " tags.");
	Vector<Integer> tags = new Vector<Integer>(tagList.values());
	Collections.sort(tags);
	return tags;
    }

    public static Vector<Integer> filterCageTags(Vector<Integer> tagList, String fileName, int chr) {
	Vector<Integer> keepList = new Vector<Integer>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern cagePattern = Pattern.compile(CAGE_CORTEX_PATTERN_STR);
	    while (line!=null) {
		Matcher m = cagePattern.matcher(line);
		if (m.find()) {
		    int c;
		    if (m.group(1).equals("X"))
			c = 20;
		    else if (m.group(1).equals("Y"))
			c = 21;
		    else
			c = Integer.parseInt(m.group(1));
		    if (chr==c) {
			//if (tagList.contains(Integer.parseInt(m.group(2))))
			keepList.addElement(Integer.parseInt(m.group(2)));
			//else
			//System.out.println("Could not find tag " + Integer.parseInt(m.group(2)) + " in orignal list");
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading from Cage file : " + e.toString());
	}
	return keepList;
    }
    
    public static HashSet<String> filterCageTags(String fileName) {
	HashSet<String> keepList = new HashSet<String>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern cagePattern = Pattern.compile(CAGE_POLYA_PATTERN_STR);
	    while (line!=null) {
		Matcher m = cagePattern.matcher(line);
		if (m.find()) {
		    if (m.group(2).equals("BC") || m.group(2).equals("BD") || m.group(2).equals("BO") || m.group(2).equals("CAW") || m.group(2).equals("CAX") || m.group(2).equals("CBA") || m.group(2).equals("CBD") || m.group(2).equals("CBR") || m.group(2).equals("CBI") || m.group(2).equals("CBM") || m.group(2).equals("CBU") || m.group(2).equals("CCI") || m.group(2).equals("CDW") || m.group(2).equals("CFU") || m.group(2).equals("CFV") || m.group(2).equals("CAA") || m.group(2).equals("CAB") || m.group(2).equals("CAC") || m.group(2).equals("CAD") || m.group(2).equals("CAE") || m.group(2).equals("CAF") || m.group(2).equals("CAG") || m.group(2).equals("CAH") || m.group(2).equals("CAI") || m.group(2).equals("CAJ") || m.group(2).equals("CCE") || m.group(2).equals("CFR") || m.group(2).equals("CDV") || m.group(2).equals("CAN"))
			//if (tagList.contains(Integer.parseInt(m.group(2))))
			keepList.add(m.group(3));
		    //else
		    //System.out.println("Could not find tag " + Integer.parseInt(m.group(2)) + " in orignal list");
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading from Cage file : " + e.toString());
	}
	System.out.println("Found " + keepList.size() + " tags to keep.");
	return keepList;
    }

    ///////////////////////////////////////////////////////

    public static LinkedHashMap<String, EST> parseEstTags(String estDirName, int chr, boolean forward) { return parseEstTags(estDirName, chr, forward, 0, "mm"); }

    public static LinkedHashMap<String, EST> parseEstTags(String estDirName, int chr, boolean forward, String species) { return parseEstTags(estDirName, chr, forward, 0, species); }

    public static LinkedHashMap<String, EST> parseEstTags(String estDirName, int chr, boolean forward, int spliced, String species) { return parseEstTags(estDirName, chr, forward, spliced, false, species); }

    public static LinkedHashMap<String, EST> parseEstTags(String estDirName, int chr, boolean forward, int spliced, boolean parseCellTypes, String species) {
	//Returns a list of valid ESTs
	//Value is ascession no and key is position
	//spliced = 0 -> use all, -1 unspliced only, 1 spliced 
	LinkedHashMap<String, EST> ests = new LinkedHashMap<String, EST>();
	//Find the position and strand info
	if (species.equals("mm")) {
	    try {
		String fileName = estDirName + ALL_EST_POSITIONS_FILE_NAME;
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		String line = br.readLine();
		Pattern allEstPositionsPattern = Pattern.compile(ALL_EST_POSITIONS_PATTERN_MM_STR);
		while (line!=null) {
		    Matcher m = allEstPositionsPattern.matcher(line);
		    if (m.find()) {
			int c;
			if (m.group(1).equals("X"))
			    c = 20;
			else if (m.group(1).equals("Y"))
			    c = 21;
			else
			    c = Integer.parseInt(m.group(1));
			boolean strand = m.group(5).equals("+");
			if (c==chr && strand==forward) 
			    ests.put(m.group(4), new EST(c, Integer.parseInt(m.group(forward ? 2 : 3)), Integer.parseInt(m.group(forward ? 3 : 2)), strand));
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error reading from EST position file : " + e.toString());
	    }
	    System.out.println("Found " + ests.size() + " ESTs for chr " + chr);
	    //keep the ones with valid 5' ends
	    HashSet<String> keepSet = new HashSet<String>();
	    HashSet<String> removeSet = new HashSet<String>();
	    try {
		String fileName = estDirName + EST_VALID_5P_FILE_NAME;
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		String line = br.readLine();
		Pattern estValid5pPattern = Pattern.compile(EST_VALID_5P_PATTERN_MM_STR);
		while (line!=null) {
		    Matcher m = estValid5pPattern.matcher(line);
		    if (m.find()) {
			if (ests.containsKey(m.group(1)))
			    keepSet.add(m.group(1));
		    }
		    line = br.readLine();
		}
		br.close();
		for (String k : ests.keySet()) {
		    if (!keepSet.contains(k))
			removeSet.add(k);
		}
		for (String k : removeSet) 
		    ests.remove(k);
	    }
	    catch (IOException e) {
		System.err.println("Error reading from EST 5' end file : " + e.toString());
	    }
	    System.out.println("Found " + ests.size() + " ESTs after filtering for 5'-ends for chr " + chr);
	    //if (spliced!=0) {
		try {
		    String fileName = estDirName + EST_SPLICE_INFO_FILE_NAME;
		    BufferedReader br = new BufferedReader(new FileReader(fileName));
		    String line = br.readLine();
		    Pattern estSpliceInfoPattern = Pattern.compile(EST_SPLICE_INFO_PATTERN_STR);
		    removeSet = new HashSet<String>();
		    keepSet = new HashSet<String>();
		    while (line!=null) {
			Matcher m = estSpliceInfoPattern.matcher(line);
			if (m.find()) {
			    boolean splice = Integer.parseInt(m.group(1))>1;
			    if (ests.containsKey(m.group(2))) {
				ests.get(m.group(2)).setSpliced(splice);
				if ((splice && spliced>0) || (!splice && spliced<0))
				    keepSet.add(m.group(2));
			    }
			}
			line = br.readLine();
		    }
		    br.close();
		    if (spliced!=0) {
			for (String k : ests.keySet()) {
			    if (!keepSet.contains(k))
				removeSet.add(k);
			}
			for (String k : removeSet) 
			    ests.remove(k);
			System.out.println("Removing " + removeSet.size() + " because of splicing, keeping " + ests.size());
		    }
		}
		catch (IOException e) {
		    System.err.println("Error reading from EST splice file : " + e.toString());
		}
		//}
	}
	else if (species.equals("hs")) {
	    try {
		String fileName = estDirName + EST_VALID_5P_FILE_NAME_HS;
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		String line = br.readLine();
		Pattern estValid5pPattern = Pattern.compile(EST_VALID_5P_PATTERN_HS_STR);
		while (line!=null) {
		    Matcher m = estValid5pPattern.matcher(line);
		    if (m.find()) {
			int c;
			if (m.group(1).equals("X"))
			    c = 23;
			else if (m.group(1).equals("Y"))
			    c = 24;
			else
			    c = Integer.parseInt(m.group(1));
			boolean strand = m.group(4).equals("+");
			if (c==chr && strand==forward) {
			    //Different file format than for mouse. Each exon of the EST is on its own line
			    EST est = new EST(c, Integer.parseInt(m.group(forward ? 2 : 3)), Integer.parseInt(m.group(forward ? 3 : 2)), strand, m.group(5));
			    if (!ests.containsKey(m.group(5))) 
				ests.put(m.group(5), est);
			    else //merge the two ESTs
				ests.get(m.group(5)).merge(est);
			    //System.out.println("Found EST " + m.group(5) + " -> " + est.toString());
			}
			/*else {
			    System.out.println("Could not find EST for " + line);
			    break;
			    }*/
		    }
		    line = br.readLine();
		} 
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error reading from EST 5' end file : " + e.toString());
	    }
	    System.out.println("Found " + ests.size() + " ESTs after filtering for 5'-ends for chr " + chr);
	    try {
		String fileName = estDirName + EST_SPLICE_INFO_FILE_NAME_HS;
		HashSet<String> keepSet = new HashSet<String>(), removeSet = new HashSet<String>();
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		String line = br.readLine();
		Pattern estSpliceInfoPattern = Pattern.compile(EST_SPLICE_INFO_PATTERN_STR);
		while (line!=null) {
		    Matcher m = estSpliceInfoPattern.matcher(line);
		    if (m.find()) {
			String k = m.group(2);
			if (ests.containsKey(k)) {
			    boolean splice = Integer.parseInt(m.group(1))>1;
			    ests.get(k).setSpliced(splice);
			    if (spliced!=0 && splice!=spliced>0)
				removeSet.add(k);
			}
		    }
		    line = br.readLine();
		}
		br.close();
		for (String k : removeSet) 
		    ests.remove(k);
	    }
	    catch (IOException e) {
		System.err.println("Error reading from EST splice file : " + e.toString());
	    }
	    System.out.println("Found " + ests.size() + " ESTs after filtering for splicing for chr " + chr);
	    if (parseCellTypes) {
		try {
		    BufferedReader br = new BufferedReader(new FileReader(EST_TISSUE_TYPE_FILE));
		    String line = br.readLine();
		    Pattern tissueTypePattern = Pattern.compile(EST_TISSUE_TYPE_PATTERN);
		    HashSet<String> removeSet = new HashSet<String>();
		    while (line!=null) {
			Matcher m = tissueTypePattern.matcher(line);
			if (m.find()) {
			    String k = m.group(1);
			    if (ests.containsKey(k)) {
				if (Integer.parseInt(m.group(3))==299)
				    removeSet.add(k);
				else
				    ests.get(k).setTissueType(Integer.parseInt(m.group(3)));
			    }
			}
			line = br.readLine();
		    }
		    br.close();
		    for (String k : removeSet) 
			ests.remove(k);
		}
		catch (IOException e) {
		    System.err.println("Error reading from EST tissue type file : " + e.toString());
		}
		System.out.println("Found " + ests.size() + " ESTs after filtering for organism for chr " + chr);
	    }
	}
	return ests;
    }

    public static TreeMap<Integer, Integer> getESTTissueTypeMergeMap() {
	//Use this method to parse the table containing the keys for the 

	return null;
    }

    public static TreeMap<Integer, Integer> parseRNASeq(String rnaSeqFileName, boolean sense, int isolatedIDs[]) {
	TreeMap<Integer, Integer> rnaSeq = new TreeMap<Integer, Integer>();
	int transcriptStarts[][] = FindReadsAcrossFeature.parseInitiationFile(rnaSeqFileName, isolatedIDs, false), colInd = sense ? 1 : 2;
	for (int i=0; i<transcriptStarts[colInd].length; i++) {
	    int t = transcriptStarts[colInd][i];
	    if (t>Integer.MIN_VALUE && t<Integer.MAX_VALUE)
		rnaSeq.put(t, rnaSeq.containsKey(t) ? rnaSeq.get(t) + 1 : 1);
	}
	return rnaSeq;
    }

    public static void saveFivePrimeEnds(TreeMap<Integer, Vector<Integer>> fivePrimeEndsNearTSS, String saveFileStart) {
	try {
	    String saveFileName = saveFileStart + FIVE_PRIME_END_START_SUFFIX  + ".txt";
	    System.out.println("Saving 5'-ends histogram to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    for (int id : fivePrimeEndsNearTSS.keySet()) {
		bw.write((new Integer(id)).toString());
		Vector<Integer> ends = fivePrimeEndsNearTSS.get(id);
		Collections.sort(ends);
		for (int p : ends)
		    bw.write("\t" + p);
		bw.write("\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving 5' ends : " + fivePrimeEndsNearTSS);
	}
    }

    public static void compareFivePrimeEnds(String rnaSeqFileName, TreeMap<Integer, Vector<Integer>> fivePrimeCageEndsNearTSS, boolean sense, int isolatedIDs[], RefSeqGenes geneInfo) {
	//Parse all the 5' ends from the RNA-seq
	int transcriptStarts[][] = FindReadsAcrossFeature.parseInitiationFile(rnaSeqFileName, isolatedIDs, false), colInd = sense ? 1 : 2;
	int furthestCageDistance = 0, modeDistance = 0, nFoundBothFurthest = 0, nFoundBothMode = 0, nModeNearby = 0;
	for (Integer i : fivePrimeCageEndsNearTSS.keySet()) { //Compare the location to the CAGE tags
	    int id = geneInfo.getNearestGene(i).getID();
	    int ind = Arrays.binarySearch(transcriptStarts[0], id);
	    if (ind>=0) {
		int t = transcriptStarts[colInd][ind];
		if (Math.abs(t)<FindReadsAcrossFeature.MAX_TSS_DISTANCE_TRANSCRIPT_START) {
		    //Calculate the distance to the most 5' CAGE tag (within 3.5 kb)
		    int furthest = fivePrimeCageEndsNearTSS.get(i).firstElement();
		    //if (Math.abs(furthest)<FindReadsAcrossFeature.MAX_TSS_DISTANCE_TRANSCRIPT_START) {
			nFoundBothFurthest++;
			furthestCageDistance += t - furthest;
			//}
		    //Calculate the distance to the mode of the CAGE distribution
		    int mode = findMode(fivePrimeCageEndsNearTSS.get(i));
		    //if (Math.abs(mode)<FindReadsAcrossFeature.MAX_TSS_DISTANCE_TRANSCRIPT_START) {
			nFoundBothMode++;
			modeDistance += t - mode;
			nModeNearby += Math.abs(t - mode)<120 ? 1 : 0;
			//}
		    //System.out.println("Found RNA-seq end " + t + ", furthest CAGE end " + fivePrimeCageEndsNearTSS.get(i).firstElement() + " and mode " + findMode(fivePrimeCageEndsNearTSS.get(i)) + " for " + id);
		}
	    }
	}
	System.out.println("The average distance to the furthest CAGE tag is " + format.format((double)furthestCageDistance/nFoundBothFurthest) + " and the average distance to the mode is " + format.format((double)modeDistance/nFoundBothMode) + " with " + format.format((double)nModeNearby/nFoundBothMode) + " within 120 bps.");
    }

    public static int numberOfPromotersWithCageTags(TreeMap<Integer, Vector<Integer>> fivePrimeCageEndsNearTSS, boolean sense, boolean convergentRegion) {
	if (!convergentRegion)
	    return fivePrimeCageEndsNearTSS.size();
	int nGenes = 0;
	for (Integer i : fivePrimeCageEndsNearTSS.keySet()) {
	    for (Integer p : fivePrimeCageEndsNearTSS.get(i)) {
		if (FindReadsAcrossFeature.inConvergentRegion(sense, p)) {
		    nGenes++;
		    break;
		}
	    }
	}
	return nGenes;
    }

    public static int numberOfCageTagsInPromoters(TreeMap<Integer, Vector<Integer>> fivePrimeCageEndsNearTSS, boolean sense, boolean convergentRegion) {
	int nTags = 0;
	for (Integer i : fivePrimeCageEndsNearTSS.keySet()) {
	    if (convergentRegion) {
		for (Integer p : fivePrimeCageEndsNearTSS.get(i)) {
		    if (FindReadsAcrossFeature.inConvergentRegion(sense, p)) 
			nTags++;
		}
	    }
	    else
		nTags += fivePrimeCageEndsNearTSS.get(i).size();
	}
	return nTags;
    }

    public static int findMode(Vector<Integer> ends) {
	int mostCommonCount = 0, prev = ends.elementAt(0), mostCommon = prev, count = 0;
	for (int i=1; i<ends.size(); i++) {
	    if (ends.elementAt(i)==prev) {
		count++;
		if (count>mostCommonCount) {
		    mostCommon = prev;
		    mostCommonCount = count;
		}
	    }
	    else {
		prev = ends.elementAt(i);
		count = 0;
	    }
	}
	return mostCommon;
    }

}
