//Use this method to create a set of random locations matching the
//TFBS, controlled for TSS distance and gene overlap
//
//First we need a list of all peak locations. This can for example be
//done with find_peaks_near_transcripts2.m and stored in
///home/mh168/Projects/ChIP-Seq/Transcripts/tf_kcl_peaks_all.txt
//
//bsub -q shared_12h java -Xmx4096m -server CreateRandomRegions -n /home/mh168/Projects/ChIP-Seq/TKData/Transcripts/tf_kcl_peaks_all.txt -o -s /home/mh168/Projects/ChIP-Seq/TKData/Transcripts/other_random_tfbs_locations_1.txt -c 1
//
//bsub -q shared_12h java -Xmx4096m -server CreateRandomRegions -n /home/mh168/Projects/ChIP-Seq/TKData/Transcripts/tf_kcl_peaks_all.txt -o -s /home/mh168/Projects/ChIP-Seq/TKData/Transcripts/intronic_random_tfbs_locations_1.txt -c 1 -I
//
//Create conserved random regions matching the enhancers
//
//bsub -q shared_12h java -Xmx4096m -server CreateRandomRegions -n /home/mh168/Projects/ChIP-Seq/TKData/Transcripts/PeakLocations/Extragenic_enhancer_tfbs_locations.txt -o -s /home/mh168/Projects/ChIP-Seq/TKData/Transcripts/Extragenic_enhancer_random_tfbs_locations_1.txt -c 1 -C

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;



public class CreateRandomRegions {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option intronicFlag = argParser.addBooleanOption('I', "intronic");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    //Make sure that the random regions are conserved
	    CmdLineParser.Option conservedFlag = argParser.addBooleanOption('C', "conserved");
	    //Make sure that the regions are not conserved
	    CmdLineParser.Option nonConservedFlag = argParser.addBooleanOption('N', "nonConserved");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option mappabilityWindowFlag = argParser.addIntegerOption('w', "mappabilityWindow");	    
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean intronic = argParser.getOptionValue(intronicFlag)!=null;
	    boolean conserved = argParser.getOptionValue(conservedFlag)!=null;
	    boolean nonConserved = argParser.getOptionValue(nonConservedFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "test";
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(mappabilityWindowFlag);
	    int mappabilityWindow = tmp!=null ? ((Integer)tmp).intValue() : 3000;
	    ///////////////////////////////////////////////////////
	    int nChr = FindTranscripts.numberOfChromosomes(species);
	    double minMappability = .8;
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    RefSeqUCSCEnsemblGenes geneInfo = new RefSeqUCSCEnsemblGenes(i, orchestra);
		    ConservationIslands conservationIslands = conserved || nonConserved ? new ConservationIslands(i, orchestra) : null;
		    int lociToMatch[] = FindReadsAcrossFeature2.parseFeatureFile(fileName, i);
		    lociToMatch = getIntronicSites(lociToMatch, intronic, geneInfo);
		    geneInfo.createRandomRegions(lociToMatch, saveFileName, mappabilityWindow, minMappability, intronic, conservationIslands, conserved, orchestra);
		}
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }

    public static int[] getIntronicSites(int[] lociToMatch, boolean intronic, RefSeqUCSCEnsemblGenes geneInfo) {
	Vector<Integer> lociToKeep = new Vector<Integer>();
	for (int i=0; i<lociToMatch.length; i++) {
	    boolean overlaps = geneInfo.overlapsGene(lociToMatch[i]);
	    if (overlaps==intronic)
		lociToKeep.addElement(lociToMatch[i]);
	}
	int keep[] = new int[lociToKeep.size()], i = 0;
	for (Integer k : lociToKeep)
	    keep[i++] = k;
	return keep;
    }

}