//Use this to find splice-junctions. In particular, we are interested
//in discovering splice junctions for the non-coding transcripts.
//
//The first version of this method relies on Tophat. The steps
//involved are:
//
//1. Find transcripts (from file)
//2. Extract sequence
//3. Call Tophat to see if any splice junctions were found
//
//Example:
//
//bsub -q shared_2h java DetectSpliceJunctions -b /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr0c -c 19 --sameBedFile -o
//


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;


public class DetectSpliceJunctions {

    public static final String TRANSCRIPT_SEQUENCE_SAVE_DIR = "/groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/ReadSequences";

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");

	    CmdLineParser.Option sameBedFileFlag = argParser.addBooleanOption("sameBedFile");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option bedFileNameFlag = argParser.addStringOption('b', "bedFileName");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option strandFlag = argParser.addBooleanOption('r', "reverseStrand");
	    ////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean reverseStrand = argParser.getOptionValue(strandFlag)!=null;
	    boolean sameBedFile = argParser.getOptionValue(sameBedFileFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "test";
	    tmp = argParser.getOptionValue(bedFileNameFlag);
	    String rnaFileName = tmp!=null ? (String)tmp : null;
	    String rnaFileStart = rnaFileName.substring(rnaFileName.lastIndexOf("/") + 1);
	    String bedFileName = rnaFileName!=null ? rnaFileName + FindTranscripts.BED_FILE_SUFFIX + "_" + (reverseStrand ? "r" : "f" ) + (chr==-1 || sameBedFile ? "" : "_" + chr) + ".bed" : null;
	    ///////////////////////////////////////////////////////
	    //Get the reads mapping to the chromosome

	    //Open file with transcript data
	    ReducedModel r = new ReducedModel(bedFileName, chr, !reverseStrand, species);
	    FastaSequence2 fastaReader = new FastaSequence2(chr, orchestra);
	    //Go through all the transcripts, look at all adjacent
	    //triplets of transcripts and search for possible splice
	    //junctions. Use a cut-off distance to ignore neighbors
	    //that are very far apart.
	    TranscribedRegion prevSegment = null, segment = null;
	    for (TranscribedRegion nextSegment : r.getSegments()) { 
		if (prevSegment!=null && segment!=null) {
		    //Extract the sequence
		    nextSegment.setSequence(reverseStrand ? FastaSequence2.complement(FastaSequence2.reverse(fastaReader.getSequence(nextSegment.getStart(), nextSegment.getLength()))) : fastaReader.getSequence(nextSegment.getStart(), nextSegment.getLength())); 
		    String sequenceSaveFileName = TRANSCRIPT_SEQUENCE_SAVE_DIR + "/" + rnaFileStart + "_" + (reverseStrand ? "r" : "f") + "_" + chr + ".fa";
		    BufferedWriter bw = new BufferedWriter(new FileWriter(sequenceSaveFileName));
		    //bw.write(prevSegment.getFastaStr() + "\n" + segment.getFastaStr() + "\n" + nextSegment.getFastaStr());
		    bw.close();
		    //Save the reads (including sequences mapped to these transcripts)
		}
		prevSegment = segment;
		segment = nextSegment;
	    }
	    //Save and create a bowtie index
	    String bowtieIndexFileName = TRANSCRIPT_SEQUENCE_SAVE_DIR + "/" + rnaFileStart + "_" + (reverseStrand ? "r" : "f") + "_" + chr;
	    /*String bowtieCmd = "/opt/bowtie/bowtie-build " + sequenceSaveFileName + " " + bowtieIndexFileName;
	    System.out.println("Running bowtie: " + bowtieCmd);
	    Runtime runtime = Runtime.getRuntime();
	    Process bowtieIndexBuildProcess = runtime.exec(bowtieCmd);
	    bowtieIndexBuildProcess.waitFor();
	    //Run Tophat
	    String codingSpliceReadsFileName = TRANSCRIPT_SEQUENCE_SAVE_DIR + "/" + rnaFileStart + "_spliceCoding_" + (reverseStrand ? "r" : "f") + ".fa";
	    String unmappedReadsFileName = TRANSCRIPT_SEQUENCE_SAVE_DIR + "/" + rnaFileStart + "_noMatches.fa";
	    String uniqueReadsFileName = TRANSCRIPT_SEQUENCE_SAVE_DIR + "/" + rnaFileStart + "_unique_" + (reverseStrand ? "r" : "f") + "_" + chr + ".fa";
	    String tophatCmd = "/opt/tophat/bin/tophat --num-threads 3 --segment-length 17 -o " + bowtieIndexFileName + "_tophat_out" + " " + bowtieIndexFileName + " " + codingSpliceReadsFileName + "," + unmappedReadsFileName + "," + uniqueReadsFileName;
	    //Wait for Tophat to finish
	    System.out.println("Running tophat: " + tophatCmd);
	    Process tophatProcess = runtime.exec(tophatCmd);
	    tophatProcess.waitFor();
	    System.out.println("Waiting for Tophat done!");*/
	    //Clean up some of the large files produced and left behind

	    //Move the .bed and .wig files to the www-directory

	    //Parse the output
	    //TreeSet<SpliceJunction> spliceJunctions = parseTophatOutput(tophatOutputFileName);
	    //if (spliceJunctions.size()>0) {
	    //System.out.println("Found splice junctions for transcript " + t.toString());
	    //for (SpliceJunction s : spliceJunctions)
	    //System.out.println(s.toString());
	    //}
	    
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    }


}