//Use this class to represent a region in the genome which has one or
//more TFBS.

import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;

public class DifferentialOverlapRegion implements Comparable {
    //Use this class to represent a region where there are TFBS for
    //more than one factor.

    public static DecimalFormat format = new DecimalFormat("###.##");

    protected int start, end, chr;
    protected double gcContent, conservationScoresVertebrates[] = null, conservationScoresEuarchontoglires[] = null;
    //Store the sequence under the peak
    protected String sequence = null;
    //Key is the factor name and value is the number of reads. Ignore
    //input for now.
    protected TreeMap tfPairs = new TreeMap();

    //////////////////////////////////////////////////////////////////

    public DifferentialOverlapRegion(DifferentialWindow win, String tfName, Vector ts) {
	chr = win.getChromosome();
	start = win.getStart();
	end = win.getEnd();
	gcContent = win.getGCContent();
	Iterator it = ts.iterator();
	while (it.hasNext())
	    tfPairs.put((String)it.next(), new double[]{0, 0});
	if (win.getDiffDiff()>0) //Figure out which peak is significant
	    tfPairs.put(tfName, new double[]{win.readsEnrichment(true), -win.readsEnrichment(false)});
	else
	    tfPairs.put(tfName, new double[]{-win.readsEnrichment(true), win.readsEnrichment(false)});
    }

    protected DifferentialOverlapRegion(DifferentialOverlapRegion r, DifferentialWindow win, String tfPairName) {
	//Copy-constructor
	chr = r.chr;
	start = r.start<win.getStart() ? r.start : win.getStart();
	end = r.end>win.getEnd() ? r.end : win.getEnd();
	gcContent = r.gcContent;
	conservationScoresEuarchontoglires = r.conservationScoresEuarchontoglires;
	conservationScoresVertebrates = r.conservationScoresVertebrates;
	sequence = r.sequence;
	Iterator it = r.tfPairs.keySet().iterator();
	while (it.hasNext()) {
	    String t = (String)it.next();
	    tfPairs.put(t, (double[])r.tfPairs.get(t));
	}
	if (win.getDiffDiff()>0) //Figure out which peak is significant
	    tfPairs.put(tfPairName, new double[]{win.readsEnrichment(true), -win.readsEnrichment(false)});
	else
	    tfPairs.put(tfPairName, new double[]{-win.readsEnrichment(true), win.readsEnrichment(false)});
    }

    //dummy constructor, use when we want to find head or tail sets
    public DifferentialOverlapRegion(int c, int s, int e, double g, Vector ts) {
	chr = c;
	start = s;
	end = e;
	gcContent = g;
	Iterator it = ts.iterator();
	while (it.hasNext())
	    tfPairs.put((String)it.next(), new double[]{0, 0});
    }
    
    //////////////////////////////////////////////////////////////////

    public int getStart() { return start; }

    public int getEnd() { return end; }
    
    public int getLength() { return Math.abs(getEnd() - getStart()); }

    public double getGCContent() { return gcContent; }

    public void addTFPair(DifferentialWindow win, String tfPairName) { addTFPair(win, tfPairName, true); }

    public void addTFPair(DifferentialWindow win, String tfName, boolean significant) {
	if (chr==win.getChromosome()) {
	    start = start<win.getStart() ? start : win.getStart();
	    end = end>win.getEnd() ? end : win.getEnd();
	    if (significant) {
		if (win.getDiffDiff()>0)
		    tfPairs.put(tfName, new double[]{win.readsEnrichment(true), -win.readsEnrichment(false)});
		else
		    tfPairs.put(tfName, new double[]{-win.readsEnrichment(true), win.readsEnrichment(false)});
	    }
	    else
		tfPairs.put(tfName, new double[]{-win.readsEnrichment(true), -win.readsEnrichment(false)});
	}
    }
    
    public boolean overlaps(DifferentialWindow win) { return win.getChromosome()==chr && win.getStart()<=end && win.getEnd()>=start; }

    public boolean overlaps(DifferentialOverlapRegion r) { return r.chr==chr && r.start<=end && r.end>=start; }

    public int numberOfTFs() { 
	int n = 0;
	Iterator it = tfPairs.keySet().iterator();
	while (it.hasNext()) {
	    String key = (String)it.next();
	    n += ((double[])tfPairs.get(key))[0]>0 || ((double[])tfPairs.get(key))[1]>0 ? 1 : 0;
	}
	return n;
    }

    //public int numberOfReads(String tfName) { return ((Integer)tfs.get(tfName)).intValue(); }

    public double[] readsEnrichment(String tfName) { return (double[])tfPairs.get(tfName); }

    public double readsEnrichment(String tfName, boolean one) { return one ? readsEnrichment(tfName)[0] : readsEnrichment(tfName)[1]; }

    public void setConservationScores(double s[], boolean vertebrate) { 
	if (vertebrate) 
	    conservationScoresVertebrates = s;
	else
	    conservationScoresEuarchontoglires = s;
    }

    public double meanConservationScore(boolean vertebrate) {
	double[] conservationScores = vertebrate ? conservationScoresVertebrates : conservationScoresEuarchontoglires;
	double m = 0;
	for (int i=0; i<conservationScores.length; i++) 
	    m += conservationScores[i];
	return m/conservationScores.length;
    }

    public double stdDevConservationScore(boolean vertebrate) {
	double[] conservationScores = vertebrate ? conservationScoresVertebrates : conservationScoresEuarchontoglires;
	double s = 0, m = meanConservationScore(vertebrate);
	for (int i=0; i<conservationScores.length; i++) 
	    s += (conservationScores[i] - m)*(conservationScores[i] - m);
	return s/(conservationScores.length - 1);
    }

    public double maxConservationScore(boolean vertebrate) {
	double[] conservationScores = vertebrate ? conservationScoresVertebrates : conservationScoresEuarchontoglires;
	double m = 0;
	for (int i=0; i<conservationScores.length; i++) 
	    m = Math.max(m, conservationScores[i]);
	return m;
    }

    //////////////////////////////////////////////////////////////////
    //Sequence stuff
    
    public void setSequence(String seq) { sequence = seq; }

    public String getSequence() { return sequence; }

    //////////////////////////////////////////////////////////////////

    public String toString() {
	String str = chr + "\t" + start + "\t" + end + "\t" + format.format(gcContent);
	if (conservationScoresVertebrates!=null) 
	    str += "\t" + format.format(meanConservationScore(true)) + "\t" + format.format(stdDevConservationScore(true)) + "\t" + format.format(maxConservationScore(true));
	if (conservationScoresEuarchontoglires!=null) 
	    str += "\t" + format.format(meanConservationScore(false)) + "\t" + format.format(stdDevConservationScore(false)) + "\t" + format.format(maxConservationScore(false));
	Iterator it = tfPairs.keySet().iterator();
	while (it.hasNext()) {
	    String key = (String)it.next();
	    str += "\t" + format.format(((double[])tfPairs.get(key))[0]) + "\t" + format.format(((double[])tfPairs.get(key))[1]);
	}
	return str;
    }
    
    public boolean equals(Object o) {
	if (!(o instanceof DifferentialOverlapRegion))
	    return false;
	DifferentialOverlapRegion r = (DifferentialOverlapRegion)o;
	//Check that they span the same region and contain the same
	//number of TFs
	return chr==r.chr && start==r.start && end==r.end; // && numberOfTFs()==r.numberOfTFs();
    }

    public int compareTo(Object o) {
	if (!(o instanceof DifferentialOverlapRegion))
	    return 0;
	DifferentialOverlapRegion r = (DifferentialOverlapRegion)o;
	if (chr<r.chr)
	    return -1;
	else if (chr>r.chr)
	    return 1;
	else {
	    if (end<r.start)
		return -1;
	    else if (start>r.end)
		return 1;
	    else
		return 0;
	}
    }
    
}
    
