//Use this class to represent a set of Encode peaks. In particular, we
//want to use the load-function from inside Matlab to speed up the
//file-reading process.

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.Tally;

public class EncodePeaks {

    protected TreeSet<Peak> peaks = new TreeSet<Peak>();
    protected int chr;
    protected String[] cellTypeNames = null, tfNames = null;
    protected String category = null;

    public EncodePeaks(String peakFileName, String tfName, int c) {
	chr = c;
	peaks = parseEncodePeakFileTF(peakFileName, tfName, chr);
	cellTypeNames = getEncodeCellTypes(peakFileName, tfName);
    }

    public EncodePeaks(String peakFileName, String peakFileName2, String tfName, int c) {
	//Use this constructor for biological replicates. We require
	//that each peak is present in both peakFileName and
	//peakFileName2 to be considered
	chr = c;
	peaks = parseEncodePeakFileTF(peakFileName, peakFileName2, tfName, chr);
	cellTypeNames = getEncodeCellTypes(peakFileName, tfName);
    }

    public EncodePeaks(String peakFileName, String tfName, int c, ConservationIslands conservationIslands) {
	chr = c;
	peaks = parseEncodePeakFileTF(peakFileName, tfName, chr, conservationIslands);
	cellTypeNames = getEncodeCellTypes(peakFileName, tfName);
    }

    public EncodePeaks(String peakFileName, String peakFileName2, String tfName, int c, ConservationIslands conservationIslands) {
	//Use this constructor for biological replicates. We require
	//that each peak is present in both peakFileName and
	//peakFileName2 to be considered
	chr = c;
	peaks = parseEncodePeakFileTF(peakFileName, peakFileName2, tfName, chr, conservationIslands);
	cellTypeNames = getEncodeCellTypes(peakFileName, tfName);
    }

    public EncodePeaks(String peakFileName, String cat, int c, boolean cons) {
	//Use this constructor when we only want to load the peaks from a given category as defined by PeaksNearTranscripts.java
	chr = c;
	category = cat;
	peaks = parseCategoryPeakFile(peakFileName, category, chr, cons);
    }

    //////////////////////////////////////////////////////////////////

    public static TreeSet<Peak> parseCategoryPeakFile(String peakFileName, String category, int chr, boolean cons) {
	TreeSet<Peak> peaks = new TreeSet<Peak>();
	try {
	    String fileName = "PeakLocations/" + category + "_" + peakFileName + "_" + (cons ? "cons_" : "") + "locations_" + chr + ".txt";
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    Pattern peakPattern = Pattern.compile("(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = peakPattern.matcher(line);
		if (m.find() && Integer.parseInt(m.group(1))==chr) 
		    peaks.add(new Peak(chr, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(2))));
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error parsing peak categories " + e.toString());
	}
	return peaks;
    }

    public TreeSet<Peak> addFactorTF(String pFileName, String tfName, int chr, ConservationIslands conservationIslands) {
	try {
	    System.out.println("Opening peak file " + pFileName);
	    BufferedReader br = new BufferedReader(new FileReader(pFileName));
	    String line = br.readLine();
	    Pattern peakPattern = Pattern.compile("chr(\\w+)\\s+(\\d+)\\s+(\\d+)\\s+\\.\\s+[\\d|\\.]+\\s+\\.\\s+[\\d|\\.]+\\s+([\\d|\\.]+)");
	    while (line!=null) {
		Matcher m = peakPattern.matcher(line);
		    if (m.find()) {
			int c;
			if (m.group(1).equals("X"))
			    c = 23;
			else if (m.group(1).equals("Y"))
			    c = 24;
			else
			    c = Integer.parseInt(m.group(1));
			if (c==chr || chr==-1) {
			    Peak p = new Peak(c, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)));
			    p.addFactor(tfName, Double.parseDouble(m.group(4)));
			    //Check if this peak overlaps with any other peak and if so merge the two
			    peaks = mergePeaks(p, peaks);
			}
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error opening " + e.toString());
	    }
	return peaks;
    }

    public TreeSet<Peak> removePromoterProximalPeaks(RefSeqUCSCEnsemblGenes geneInfo, boolean nonCoding, int maxDist) {
	//Remove the peaks that are within maxDist of a TSS
	TreeSet<Peak> removePeaks = new TreeSet<Peak>();
	for (Peak p : peaks) {
	    if (p.getChromosome()==geneInfo.getChromosome()) {
		RefSeqUCSCEnsemblGene nearestGene = geneInfo.getOverlappingTSS(p.getCenter() - maxDist, p.getCenter() + maxDist);
		if (nearestGene!=null && nearestGene.isNonCoding()==nonCoding)
		    removePeaks.add(p);
	    }
	}
	for (Peak p : removePeaks)
	    peaks.remove(p);
	return removePeaks;
    }

    public TreeSet<Peak> removeIntergenicPeaks(RefSeqUCSCEnsemblGenes geneInfo) {
	TreeSet<Peak> removePeaks = new TreeSet<Peak>();
	for (Peak p : peaks) {
	    if (p.getChromosome()==geneInfo.getChromosome() && geneInfo.overlapsGene(p.getCenter())) 
		removePeaks.add(p);
	}
	for (Peak p : removePeaks)
	    peaks.remove(p);
	return removePeaks;
    }

    public TreeSet<Peak> removeEnhancerOverlappingPeaks(TreeSet<Enhancer2> enhancers, int maxDist) {
	//Remove the peaks that are within maxDist of an enhancer
	TreeSet<Peak> removePeaks = new TreeSet<Peak>();
	for (Peak p : peaks) {
	    if (FindConservationIslands.overlapsEnhancer(enhancers, p, maxDist)) 
		removePeaks.add(p);
	}
	for (Peak p : removePeaks)
	    peaks.remove(p);
	return removePeaks;
    }

    public int[][] numberOfUniquePeaks(int nMC) {
	//use this method to calculate how many unique peaks that we
	//have as more cell types or TFs are added. Since this depends
	//on the order, we want to randomize the procedure and take
	//the average. The function returns the mean and the std error.
	String vec[] = tfNames==null ? cellTypeNames : tfNames;
	Tally peaksFoundTally[] = new Tally[vec.length];
	int meanStdNumberOfPeaks[][] = new int[vec.length][2];
	for (int i=0; i<peaksFoundTally.length; i++)
	    peaksFoundTally[i] = new Tally();
	for (int i=0; i<nMC; i++) {
	    int[] order = randperm(vec.length);
	    for (int j=0; j<order.length; j++) {
		String names[] = new String[j+1];
		for (int k=0; k<names.length; k++)
		    names[k] = vec[order[k]];
		int nPeaks = numberOfPeaks(names);
		peaksFoundTally[j].add(nPeaks);
	    }
	}
	if (nMC>0) {
	    for (int i=0; i<peaksFoundTally.length; i++) {
		meanStdNumberOfPeaks[i][0] = (int)peaksFoundTally[i].average();
		meanStdNumberOfPeaks[i][1] = (int)Math.sqrt(peaksFoundTally[i].variance());
	    }
	}
	return meanStdNumberOfPeaks;
    }

    public int[][] numberOfUniquePeaks(EncodePeaks overlapPeaks, int nMC) {
	//Use this function to find how many peaks overlap with the peak set of interest 
	String vec[] = tfNames==null ? cellTypeNames : tfNames; //Math.max(overlapPeaks.numberOfCellTypes(), overlapPeaks.numberOfTFs());
	Tally peaksFoundTally[] = new Tally[vec.length];
	int meanStdNumberOfPeaks[][] = new int[vec.length][2];
	for (int i=0; i<peaksFoundTally.length; i++)
	    peaksFoundTally[i] = new Tally();
	for (int i=0; i<nMC; i++) {
	    int[] order = randperm(vec.length);
	    for (int j=0; j<order.length; j++) {
		String names[] = new String[j+1];
		for (int k=0; k<names.length; k++)
		    names[k] = vec[order[k]];
		int nPeaks = numberOfPeaks(names, overlapPeaks);
		peaksFoundTally[j].add(nPeaks);
	    }
	}
	if (nMC>0) {
	    for (int i=0; i<peaksFoundTally.length; i++) {
		meanStdNumberOfPeaks[i][0] = (int)peaksFoundTally[i].average();
		meanStdNumberOfPeaks[i][1] = (int)Math.sqrt(peaksFoundTally[i].variance());
	    }
	}
	return meanStdNumberOfPeaks;
    }

    public int numberOfCellTypes() { return cellTypeNames!=null ? cellTypeNames.length : 1; }
    
    public int numberOfTFs() { return tfNames!=null ? tfNames.length : 1; }

    public int numberOfPeaks() { return peaks.size(); }

    public int numberOfPeaks(String names[]) { return numberOfPeaks(names, null); }

    public int numberOfPeaks(String names[], EncodePeaks overlapPeaks) {
	int nPeaksFound = 0;
	for (Peak p : peaks) {
	    Peak overlappingPeak = overlapPeaks==null ? null : p.getOverlappingPeak(overlapPeaks.peaks);
	    if ((names==null || p.hasFactor(names)) && (overlapPeaks==null || overlappingPeak!=null))
		nPeaksFound++;
	    //if (nPeaksFound==1 && overlappingPeak!=null)
	    //System.out.println("Found " + overlappingPeak.toString() + " overlapping " + p);
	}
	return nPeaksFound;
    }

    public static int[] randperm(int n) {
	//return a vector containing the numbers 0..n-1 in a random
	//order. Same as matlab function. Implementation taken from
	//Wikipedia http://en.wikipedia.org/wiki/Knuth_shuffle
	Random random = new Random();
	int array[] = new int[n];
	for (int i=0; i<n; i++) 
	    array[i] = i;
	for (int i = array.length; i > 1; i--) {
	    // Pick a random element to swap with the i-th element.
	    int j = random.nextInt(i);  // 0 <= j <= i-1 (0-based array)
	    // Swap array elements.
	    int tmp = array[j];
	    array[j] = array[i-1];
	    array[i-1] = tmp;
	}
	return array;
    }

    public TreeSet<Peak> findOverlappingEnrichment(EncodePeaks maskPeaks) { return findOverlappingEnrichment(maskPeaks, 0); }

    public TreeSet<Peak> findOverlappingEnrichment(EncodePeaks maskPeaks, int maxDist) {
	//Use this function to find how many peaks overlap with maskPeaks. 
	TreeSet<Peak> overlapPeaks = new TreeSet<Peak>();
	int prevStart = 0;
	for (Peak p : peaks) {
	    Peak overlappingPeak = p.getOverlappingPeak(maskPeaks.peaks);
	    Peak newPeak = new Peak(p.getChromosome(), p.getStart(), p.getEnd(), p.isConserved() ? 1.0 : 0.0);
	    for (String k : p.getFactorNames()) //There is a DHS, but no peaks, hence we should set the enrichment to zero
		newPeak.addFactor(k, 0);
	    //if (prevStart==p.getStart())
		//System.out.println("New peak is " + newPeak.toString() + " and DHS cell types are " + Arrays.toString(p.getFactorNames()));
	    if (overlappingPeak!=null) {
		for (String k : overlappingPeak.getFactorNames()) {
		    if (p.hasFactor(k))
			newPeak.addFactor(k, overlappingPeak.getFactor(k));
		    else //Peak, but no DHS use negative number to show
			newPeak.addFactor(k, -overlappingPeak.getFactor(k));
		}
		//if (prevStart==p.getStart())
		//System.out.println("Overlapping peak has " + overlappingPeak.toString() + " with " + Arrays.toString(overlappingPeak.getFactorNames()));
		//System.out.println(p.toString() + " has " + Arrays.toString(p.getFactorNames()) + " and " + overlappingPeak.toString() + " has " + Arrays.toString(overlappingPeak.getFactorNames()));
	    }
	    if (newPeak.getFactorNames(false).length==0)
		System.out.println("Found no factors for " + newPeak.toString());
	    overlapPeaks.add(newPeak);
	    prevStart = p.getStart();
	}
	return overlapPeaks;
    }

    public TreeSet<Peak> getPeaks() { return peaks; }

    ///////////////////////////////////////////////////////

    public static String parseTFName(String peakFileName, String cellType) {
	if (peakFileName.contains("Dnase")) //Stupid. Different format for the name...
	    return "Dnase";
	return peakFileName.substring(peakFileName.lastIndexOf(cellType) + cellType.length(), peakFileName.lastIndexOf("."));
    }

    public static String[] getEncodeTFNames(String peakFileName, String cellType) {
	Vector<String> peakFileNames = parsePeakFileNames(peakFileName);
	String tfNames[] = new String[peakFileNames.size()];
	int i = 0;
	for (String p : peakFileNames)
	    tfNames[i++] = parseTFName(p, cellType);
	return tfNames;
    }

    public static String parseCellType(String peakFileName, String tfName) {
	if (tfName.toLowerCase().contains("dnase")) { //Stupid. Different format for the name...
	    if (peakFileName.toLowerCase().contains("peaks")) {
		if (peakFileName.contains("Rep1"))
		    return peakFileName.substring(peakFileName.lastIndexOf("Rep1") + 4, peakFileName.lastIndexOf("."));
		else if (peakFileName.contains("Rep2"))
		    return peakFileName.substring(peakFileName.lastIndexOf("Rep2") + 4, peakFileName.lastIndexOf("."));
		else
		    return peakFileName.substring(peakFileName.lastIndexOf("Rep") + 3, peakFileName.lastIndexOf("."));
	    }
	    else {
		return peakFileName.substring(peakFileName.lastIndexOf("Dnase") + 5, peakFileName.lastIndexOf("Rep"));
	    }
	}
	else {
	    if (peakFileName.contains("Rep1"))
		return peakFileName.substring(peakFileName.lastIndexOf("Rep1") + 4, peakFileName.lastIndexOf(tfName));
	    else
		return peakFileName.substring(peakFileName.lastIndexOf("Peaks") + 5, peakFileName.lastIndexOf(tfName));
	}
    }

    public static String[] getEncodeCellTypes(String peakFileName, String tfName) {
	Vector<String> peakFileNames = parsePeakFileNames(peakFileName);
	String cellTypes[] = new String[peakFileNames.size()];
	int i = 0;
	for (String p : peakFileNames) {
	    //System.out.println("Found file " + p);
	    cellTypes[i++] = parseCellType(p, tfName);
	}
	return cellTypes;
    }

    public static Vector<String> parsePeakFileNames(String peakFileName) {
	Vector<String> peakFileNames = new Vector<String>();
	if (!peakFileName.endsWith("names.txt"))
	    peakFileNames.addElement(peakFileName);
	else {
	    try {
		BufferedReader br = new BufferedReader(new FileReader(peakFileName));
		String line = br.readLine();
		while (line!=null) {
		    if (line.length()>0 && !line.startsWith("#"))
			peakFileNames.addElement(line);
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error reading Encode peak file " + e.toString());
	    }
	}
	return peakFileNames;
    }

    public static TreeSet<Peak> parseEncodePeakFileCellType(String peakFileName, String cellType, int chr) {
	//Use this when we want to parse multiple factors from the same cell-type
	TreeSet<Peak> peaks = new TreeSet<Peak>();
	if (peakFileName!=null) {
	    TreeMap<String, String> peakFile2TF = new TreeMap<String, String>();
	    Vector<String> peakFileNames = new Vector<String>();
	    if (peakFileName.endsWith("narrowPeak"))
		peakFileNames.addElement(peakFileName);
	    else
		peakFileNames = parsePeakFileNames(peakFileName);
	    for (String pFileName : peakFileNames)
		peakFile2TF.put(pFileName, parseTFName(pFileName, cellType));
	    for (String pFileName : peakFileNames) {
		try {
		    System.out.println("Opening peak file " + pFileName);
		    BufferedReader br = new BufferedReader(new FileReader(pFileName));
		    String line = br.readLine();
		    Pattern peakPattern = Pattern.compile("chr(\\w+)\\s+(\\d+)\\s+(\\d+)\\s+\\.\\s+[\\d|\\.]+\\s+\\.\\s+[\\d|\\.]+\\s+([\\d|\\.]+)");
		    while (line!=null) {
			Matcher m = peakPattern.matcher(line);
			if (m.find()) {
			    int c;
			    if (m.group(1).equals("X"))
				c = 23;
			    else if (m.group(1).equals("Y"))
				c = 24;
			    else
				c = Integer.parseInt(m.group(1));
			    if (c==chr || chr==-1) {
				Peak p = new Peak(c, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)));
				p.addFactor(peakFile2TF.get(pFileName), Double.parseDouble(m.group(4)));
				//Check if this peak overlaps with any other peak and if so merge the two
				peaks = mergePeaks(p, peaks);
			    }
		    }
			line = br.readLine();
		}
		    br.close();
		}
		catch (IOException e) {
		    System.err.println("Error opening " + e.toString());
		}
	    }
	}
	return peaks;
    }

    public static TreeSet<Peak> mergePeaks(Peak p, TreeSet<Peak> peaks) { return mergePeaks(p, peaks, false); }

    public static TreeSet<Peak> mergePeaks(Peak p, TreeSet<Peak> peaks, boolean verbose) {
	TreeSet<Peak> mergedPeaks = new TreeSet<Peak>();
	TreeSet<Peak> overlapping = p.getAllOverlappingPeaks(peaks);
	//if (verbose && p.getCenter()>2755000 && p.getCenter()<2756000)
	//  System.out.println("Merging " + p.toString() + " with " + Arrays.toString(p.getFactorNames(true)) + " overlapping " + overlapping.size() + " out of " + peaks.size());
	if (overlapping.size()>0) {
	    for (Peak po : overlapping) {
		//Remove old entry
		peaks.remove(po);
		//merge the two peaks
		p.merge(po);
		//if (verbose && p.getCenter()>2755000 && p.getCenter()<2756000)
		//  System.out.println("Merging " + p.toString() + " with " + Arrays.toString(p.getFactorNames(true)) + " and " + po.toString() + " with " + Arrays.toString(po.getFactorNames(true)));
	    }
	}
	mergedPeaks.add(p);
	mergedPeaks.addAll(peaks);
	return mergedPeaks;
    }

    public static TreeSet<Peak> parseEncodePeakFileTF(String peakFileName, String tfName, int chr) { return parseEncodePeakFileTF(peakFileName, tfName, chr, null); }

    public static TreeSet<Peak> parseEncodePeakFileTF(String peakFileName, String tfName, int chr, ConservationIslands conservationIslands) {
	//Use this when we want to parse multiple cell-types from the same TF
	TreeSet<Peak> peaks = new TreeSet<Peak>();
	TreeMap<String, String> peakFile2CellType = new TreeMap<String, String>();
	Vector<String> peakFileNames = new Vector<String>();
	if (peakFileName.endsWith("narrowPeak"))
	    peakFileNames.addElement(peakFileName);
	else
	    peakFileNames = parsePeakFileNames(peakFileName);
	for (String pFileName : peakFileNames)
	    peakFile2CellType.put(pFileName, parseCellType(pFileName, tfName));
	for (String pFileName : peakFileNames) {
	    try {
		System.out.println("Opening peak file " + pFileName);
		BufferedReader br = new BufferedReader(new FileReader(pFileName));
		String line = br.readLine();
		Pattern peakPattern = Pattern.compile("chr(\\w+)\\s+(\\d+)\\s+(\\d+)\\s+\\.\\s+[\\d|\\.]+\\s+\\.\\s+[\\d|\\.]+\\s+([\\d|\\.]+)");
		while (line!=null) {
		    Matcher m = peakPattern.matcher(line);
		    if (m.find()) {
			int c;
			if (m.group(1).equals("X"))
			    c = 23;
			else if (m.group(1).equals("Y"))
			    c = 24;
			else
			    c = Integer.parseInt(m.group(1));
			if (c==chr || chr==-1) {
			    int start = Integer.parseInt(m.group(2)), end = Integer.parseInt(m.group(3));
			    if (conservationIslands==null || conservationIslands.overlaps(c, start, end)) {
				Peak p = new Peak(c, start, end);
				p.addFactor(peakFile2CellType.get(pFileName), Double.parseDouble(m.group(4)));
				//Check if this peak overlaps with any other peak and if so merge the two
				peaks = mergePeaks(p, peaks);
			    }
			}
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error opening " + e.toString());
	    }
	}
	return peaks;
    }

    public static TreeSet<Peak> parseEncodePeakFileTF(String peakFileName, String peakFileName2, String tfName, int chr) { return parseEncodePeakFileTF(peakFileName, peakFileName2, tfName, chr, null); }

    public static TreeSet<Peak> parseEncodePeakFileTF(String peakFileName, String peakFileName2, String tfName, int chr, ConservationIslands conservationIslands) {
	//Use this when we want to parse multiple cell-types from the same TF and require the peak to be present in two replicates.
	TreeSet<Peak> peaks = new TreeSet<Peak>();
	TreeMap<String, String> peakFile2CellType = new TreeMap<String, String>(), peakFile2CellType2 = new TreeMap<String, String>();
	Vector<String> peakFileNames = new Vector<String>(), peakFileNames2 = new Vector<String>();
	if (peakFileName.endsWith("narrowPeak")) {
	    peakFileNames.addElement(peakFileName);
	    peakFileNames2.addElement(peakFileName2);
	}
	else {
	    peakFileNames = parsePeakFileNames(peakFileName);
	    //Assume that files are in the same order
	    peakFileNames2 = parsePeakFileNames(peakFileName2);
	}
	for (String pFileName : peakFileNames)
	    peakFile2CellType.put(pFileName, parseCellType(pFileName, tfName));
	//System.out.println(peakFile2CellType.toString());
	for (String pFileName : peakFileNames2)
	    peakFile2CellType2.put(pFileName, parseCellType(pFileName, tfName) + "Rep2");
	//System.out.println(peakFile2CellType2.toString());
	for (int i=0; i<peakFileNames.size(); i++) {
	    //Keep track of all the peaks and first make sure that they can be replicated
	    TreeSet<Peak> peaks1 = new TreeSet<Peak>();
	    try {
		String pFileName = peakFileNames.elementAt(i);
		BufferedReader br = new BufferedReader(new FileReader(pFileName));
		String line = br.readLine();
		Pattern peakPattern = Pattern.compile("chr(\\w+)\\s+(\\d+)\\s+(\\d+)\\s+\\.\\s+[\\d|\\.]+\\s+\\.\\s+[\\d|\\.]+\\s+([\\d|\\.]+)");
		while (line!=null) {
		    Matcher m = peakPattern.matcher(line);
		    if (m.find()) {
			int c;
			if (m.group(1).equals("X"))
			    c = 23;
			else if (m.group(1).equals("Y"))
			    c = 24;
			else
			    c = Integer.parseInt(m.group(1));
			if (c==chr || chr==-1) {
			    int start = Integer.parseInt(m.group(2)), end = Integer.parseInt(m.group(3));
			    if (conservationIslands==null || conservationIslands.overlaps(c, start, end)) {
				Peak p = new Peak(c, start, end);
				p.addFactor(peakFile2CellType.get(pFileName), Double.parseDouble(m.group(4)));
				//Check if this peak overlaps with any other peak and if so merge the two
				peaks1 = mergePeaks(p, peaks1);
			    }
			}
		    }
		    line = br.readLine();
		}
		br.close();
		if (peaks1.size()>0) {
		    //System.out.println("Opened peak file " + pFileName + " with " + peaks1.size() + ", first peak is " + peaks1.first().toString());
		    //Now we can read the second replicate and keep only those peaks that are present in both
		    pFileName = peakFileNames2.elementAt(i);
		    //System.out.println("Opening peak file " + pFileName);
		    br = new BufferedReader(new FileReader(pFileName));
		    line = br.readLine();
		    while (line!=null) {
			Matcher m = peakPattern.matcher(line);
			if (m.find()) {
			    int c;
			    if (m.group(1).equals("X"))
				c = 23;
			    else if (m.group(1).equals("Y"))
				c = 24;
			    else
				c = Integer.parseInt(m.group(1));
			    if (c==chr || chr==-1) {
				int start = Integer.parseInt(m.group(2)), end = Integer.parseInt(m.group(3));
				if (conservationIslands==null || conservationIslands.overlaps(c, start, end)) {
				    Peak p = new Peak(c, start, end);
				    p.addFactor(peakFile2CellType2.get(pFileName) + "Rep2", Double.parseDouble(m.group(4)));
				    //Check if this peak overlaps with any other peak and if so merge the two
				    peaks1 = mergePeaks(p, peaks1);
				}
			    }
			}
			line = br.readLine();
		    }
		    br.close();
		    //System.out.println("Opened peak file " + pFileName + " with " + peaks1.size() + ", first peak is " + peaks1.first().toString());
		    //Go throught the list and throw away those peaks that were only found in the 1st replicate
		    for (Peak p : peaks1) {
			if (p.getFactors(true).size()==2) {
			    //Remove the second replicate
			    p.removeFactor(peakFile2CellType2.get(pFileName) + "Rep2");
			    peaks = mergePeaks(p, peaks, true);
			}
			//else
			//System.out.println(p + "\t" + p.getFactors(true));
		    }
		    System.out.println("Found " + peaks.size() + ", with first " + peaks.first().toString());
		}
		else
		    System.out.println("Opened peak file " + pFileName + " with no peaks.");
	    }
	    catch (IOException e) {
		System.err.println("Error opening " + e.toString());
	    }
	}
	return peaks;
    }
    
    
    
    public static int[][] loadPeaks(String fileName, String tfName, int chr) {
	//Use this method with matlab to read a file. The return array
	//has four column showing the chr, start, end and peak-height
	TreeSet<Peak> peaks = parseEncodePeakFileTF(fileName, tfName, chr);
	int peaksArray[][] = new int[peaks.size()][4], i = 0;
	for (Peak p : peaks) 
	    peaksArray[i++] = new int[]{p.getChromosome(), p.getStart(), p.getEnd(), 1};
	return peaksArray;
    }

		
    

}