//Use this method to calculate how the densities for intronic RNA
//changes as a function of the distance from the TSS.

//Example:

//java -Xmx512m EstimatePALRLength  -d 15000 -u 15000 -U 100 -D 100 -s mRNA/TSS_split_by_conv_div/ 
//
//For enhancers
//
//java -Xmx512m EstimatePALRLength  -d 5000 -u 5000 -U 50 -D 50 --enhancer Enhancers/enhancers_null_KCl.txt
//
//For enhancers in HeLa cells:
//
//bsub -q shared_2h java -Xmx512m EstimatePALRLength  -d 5000 -u 5000 -U 50 -D 50 --enhancer ../HeLa/Enhancers/heintzman_enhancer_list_hg18.bed --species Hs -R /groups/neuroduo/martin/ChIP-Seq/HeLa/mRNA/HeLac.tsv --noHists -o

//Use estimate_palr_length.m for visualizing and post-processing

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.probdist.NormalDist;

public class EstimatePALRLength {

    public static final String PALR_LENGTH_FILE_NAME = "palr_length";
    public static final String TALR_LENGTH_FILE_NAME = "talr_length";
    public static final String EALR_LENGTH_FILE_NAME = "ealr_length";

    public static final int BEFORE_DISTANCE = 10000;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option rnaSeqFileNameFlag = argParser.addStringOption('R', "RNASeq");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('a', "saveFileName");
	    CmdLineParser.Option antisenseFlag = argParser.addIntegerOption('u', "antisense");
	    CmdLineParser.Option senseFlag = argParser.addIntegerOption('d', "sense");
	    CmdLineParser.Option antisenseStepFlag = argParser.addIntegerOption('U', "antisenseStep");
	    CmdLineParser.Option senseStepFlag = argParser.addIntegerOption('D', "senseStep");
	    //Determine how wide the bins that we study should be
	    CmdLineParser.Option binWidthDownstreamFlag = argParser.addIntegerOption('w', "binWidthDownstream");
	    CmdLineParser.Option binWidthUpstreamFlag = argParser.addIntegerOption('W', "binWidthUpstream");
	    //use this flag when we only want to focus on features
	    //which are guaranteed to be isolated from one another.
	    CmdLineParser.Option isolatedFlag = argParser.addBooleanOption('I', "isolated");
	    //Use this flag to include a file with genes/peaks to filter against
	    CmdLineParser.Option selectFlag = argParser.addStringOption('s', "select");
	    //For looking at the short mRNAs
	    CmdLineParser.Option shortFlag = argParser.addBooleanOption('S', "short");
	    //For looking at the DNAse treated
	    CmdLineParser.Option dnaseFlag = argParser.addBooleanOption("DNAse");
	    //Use this flag if we want to track how the exonic densities change instead.
	    CmdLineParser.Option exonFlag = argParser.addBooleanOption('e', "exons");
	    //Use this flag for mRNA instead of total RNA
	    CmdLineParser.Option mRNAFlag = argParser.addBooleanOption('m', "mRNA");
	    //Use this flag to plot the location of the start of the nth intron
	    CmdLineParser.Option intronFlag = argParser.addIntegerOption('i', "intron");
	    //The overall enrichment of exons campared to introns due
	    //to the presence of mature transcripts
	    CmdLineParser.Option maturityFlag = argParser.addDoubleOption('M', "maturityFactor");
	    //Use this flag when we want to center around the TES
	    //instead of the TSS
	    CmdLineParser.Option endFlag = argParser.addBooleanOption('E', "end");
	    //Use this flag when we have other species
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    //Use this flag if we want to center around enhancers instead of TSSs
	    CmdLineParser.Option enhancerFlag = argParser.addStringOption("enhancer");
	    //Center around random regions
	    CmdLineParser.Option lociFlag = argParser.addStringOption("loci");
	    //Different time points
	    CmdLineParser.Option timeFlag = argParser.addIntegerOption('t', "time");
	    //Use the biological replicate
	    CmdLineParser.Option replicateFlag = argParser.addIntegerOption('b', "replicate");
	    //Align at the start of the first intron instead of the TSS
	    CmdLineParser.Option firstIntronFlag = argParser.addBooleanOption("firstIntron");
	    //Avoid flattening the reads
	    CmdLineParser.Option noFlatteningFlag = argParser.addBooleanOption("noFlattening");
	    //Avoid producing the files containing individual patterns for each loci
	    CmdLineParser.Option noHistsFlag = argParser.addBooleanOption("noHists");
	    //Consider all the internal starts instead of the outermost one
	    CmdLineParser.Option internalStartsFlag = argParser.addBooleanOption("internalStarts");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(enhancerFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(lociFlag);
	    String lociFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(timeFlag);
	    int time = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(replicateFlag);
	    int replicate = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    boolean isolated = argParser.getOptionValue(isolatedFlag)!=null;
	    boolean shortRNA = argParser.getOptionValue(shortFlag)!=null;
	    boolean dnase = argParser.getOptionValue(dnaseFlag)!=null;
	    boolean mRNA = argParser.getOptionValue(mRNAFlag)!=null;
	    boolean exons = argParser.getOptionValue(exonFlag)!=null;
	    boolean firstIntron = argParser.getOptionValue(firstIntronFlag)!=null;
	    boolean flattening = argParser.getOptionValue(noFlatteningFlag)==null;
	    boolean noHists = argParser.getOptionValue(noHistsFlag)!=null;
	    boolean ends = argParser.getOptionValue(endFlag)!=null;
	    boolean internalStarts = argParser.getOptionValue(internalStartsFlag)!=null;
	    tmp = argParser.getOptionValue(intronFlag);
	    int intron = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(rnaSeqFileNameFlag);
	    String rnaSeqFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/hr" + time + (shortRNA ? "s" /*"S_bp34"*/ : mRNA ? "m" : dnase ? "d" : "c") + (replicate>0 ? "B" + replicate : "") + ".tsv";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = null;
	    if (tmp!=null)
		saveFileName = (String)tmp;
	    else {
		if (enhancerFileName!=null) {
		    Pattern enhancerPattern = Pattern.compile("enhancers_([_|\\w]+)\\.txt");
		    Matcher m = enhancerPattern.matcher(enhancerFileName);
		    String stimulus = m.find() ? m.group(1) : "";
		    saveFileName = enhancerFileName.substring(0, enhancerFileName.lastIndexOf("/") + 1) + "mRNA/" + EALR_LENGTH_FILE_NAME + "_" + stimulus + "_hr" + time + (replicate>0 ? "_B" + replicate : "") + (shortRNA ? "_short" : "") + (mRNA ? "_mrna" : dnase ? "_dnase" : "") + (isolated ? "_isolated" : "");
		}
		else if (lociFileName!=null) {
		    saveFileName = lociFileName.substring(0, lociFileName.lastIndexOf(".")) + "_hr" + time + (replicate>0 ? "_B" + replicate : "") + (shortRNA ? "_short" : "") + (mRNA ? "_mrna" : "") + (dnase ? "_dnase" : "");
		}
		else
		    saveFileName = (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/" + (ends ? TALR_LENGTH_FILE_NAME : PALR_LENGTH_FILE_NAME) + "_hr" + time + (replicate>0 ? "_B" + replicate : "") + (shortRNA ? "_short" : "") + (mRNA ? "_mrna" : dnase ? "_dnase" : "") + (intron>0 ? "_intron_" + intron : "") + (exons ? "_exons" : "") + (isolated ? "_isolated" : "") +  (internalStarts ? "_internal" : "");
	    }
	    tmp = argParser.getOptionValue(antisenseFlag);
	    int antisense = tmp!=null ? ((Integer)tmp).intValue() : 1500;
	    tmp = argParser.getOptionValue(senseFlag);
	    int sense = tmp!=null ? ((Integer)tmp).intValue() : 5000;
	    tmp = argParser.getOptionValue(antisenseStepFlag);
	    int antisenseStep = tmp!=null ? ((Integer)tmp).intValue() : 100;
	    tmp = argParser.getOptionValue(senseStepFlag);
	    int senseStep = tmp!=null ? ((Integer)tmp).intValue() : 1000;
	    tmp = argParser.getOptionValue(binWidthDownstreamFlag);
	    int binWidthDownstream = tmp!=null ? ((Integer)tmp).intValue() : exons ? senseStep : senseStep;
	    tmp = argParser.getOptionValue(binWidthUpstreamFlag);
	    int binWidthUpstream = tmp!=null ? ((Integer)tmp).intValue() : exons ? antisenseStep : antisenseStep;
	    tmp = argParser.getOptionValue(selectFlag);
    	    String selectFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(maturityFlag);
    	    double maturityFactor = tmp!=null ? ((Double)tmp).doubleValue() : shortRNA ? 1 : 0;
	    tmp = argParser.getOptionValue(speciesFlag);
    	    String species = tmp!=null ? (String)tmp : "Mm";
	    /////////////////////////////////
	    int nChr = 21;
	    if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
		nChr = 24;
	    //Load the annotation info  and the mRNA
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    Vector<Enhancer> enhancers = enhancerFileName==null ? null : FindReadsAcrossFeature.parseEnhancerFile(enhancerFileName, species);
	    Vector<int[]> loci = lociFileName==null ? null : parseLociFile(lociFileName);
	    Vector<ABIRead[]> forwardReads = new Vector<ABIRead[]>(nChr), reverseReads = new Vector<ABIRead[]>(nChr);
	    int nReadsForward = 0, nReadsReverse = 0, nExons[] = new int[4];
	    double exonDensity[] = new double[4];
	    forwardReads.setSize(nChr);
	    reverseReads.setSize(nChr);
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || i==chr) {
		    forwardReads.setElementAt(FindNongenicTranscripts.loadMRNAReadsChr(rnaSeqFileName, true, i, flattening), i-1);
		    reverseReads.setElementAt(FindNongenicTranscripts.loadMRNAReadsChr(rnaSeqFileName, false, i, flattening), i-1);
		    nReadsForward += forwardReads.elementAt(i-1).length;
		    nReadsReverse += reverseReads.elementAt(i-1).length;
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra, species);
		    //Load all the RNA reads
		    //if (!intron) {
		    System.out.println("Loaded reads and annotation for chr " + i);
		    if (enhancerFileName==null && lociFileName==null) {
			//Calculate the average density of the second exon
			for (RefSeqGene g : geneInfo[i-1].getGenes()) {
			    for (int j=0; j<nExons.length; j++) {
				if (g.numberOfExons()>j) {
				    exonDensity[j] += calculateExonDensity(g, g.getStrand() ? forwardReads.elementAt(i-1) : reverseReads.elementAt(i-1), j);
				    nExons[j]++;
				}
			    }
			}
		    }
		    //}
		}
	    }
	    double normFactor = (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/((double)(nReadsForward + nReadsReverse));
	    if (enhancerFileName==null && lociFileName==null) {
		for (int j=0; j<nExons.length; j++) {
		    exonDensity[j] /= nExons[j];
		    System.out.println("Density of exon " + j + " is " + format.format(exonDensity[j]) + " from " + nExons[j] + " genes.");
		}
	    }
	    int isolatedIDs[] = null;
	    if (isolated) {
		if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
		    isolatedIDs = FindReadsAcrossFeature.parseIsolatedFile((orchestra ? "/groups/neuroduo/martin/" : "/media/exthd/") + "Genomes/Hs/v36/isolated_genes.txt");
		else
		    isolatedIDs = FindReadsAcrossFeature.parseIsolatedFile(orchestra);
	    }
	    TreeMap<String, double[]> subsetDownstreamDensities = new TreeMap<String, double[]>();
	    TreeMap<String, double[]> subsetUpstreamDensities = new TreeMap<String, double[]>();
	    Vector<String> subsets = parseSelectFiles(selectFileName);
	    int senseBinVec[] = getBinVec(sense, senseStep, true, exons/* || intron*/);
	    int senseBinCenterInd = Math.abs(senseBinVec.length - BEFORE_DISTANCE/senseStep - 1);
	    int antisenseBinVec[] = getBinVec(antisense, antisenseStep, false, exons/* || intron*/);
	    int antisenseBinCenterInd = Math.abs(antisenseBinVec.length - BEFORE_DISTANCE/antisenseStep - 1);
	    if (subsets.size()>0) {
		for (String s : subsets) {
		    subsetDownstreamDensities.put(s, getDensity(forwardReads, reverseReads, chr, true, s, isolatedIDs, senseBinVec, binWidthDownstream, senseBinCenterInd, orchestra, geneInfo, exons, shortRNA, intron, maturityFactor, ends, internalStarts, species, null));
		    if (!exons)
			subsetUpstreamDensities.put(s, getDensity(forwardReads, reverseReads, chr, false, s, isolatedIDs, antisenseBinVec, binWidthUpstream, antisenseBinCenterInd, orchestra, geneInfo, exons, shortRNA, intron, maturityFactor, ends, internalStarts, species, null));
		}
	    }
	    //Look at everything
	    //if (intron)
	    saveFirstIntronStartDensity(geneInfo, isolatedIDs, null, "mRNA/first_intron_start_density" + (chr!=-1 ? "_" + chr : "") + ".txt");
	    //else {
	    if (enhancerFileName!=null) {
		//sense intronic
		subsetDownstreamDensities.put("intronic.tsv", getEnhancerDensity(forwardReads, reverseReads, chr, enhancers, true, senseBinVec, antisenseBinVec, binWidthDownstream, orchestra, geneInfo, false, species, (noHists ? null : enhancerFileName.substring(0, enhancerFileName.lastIndexOf(".")) + FindReadsAcrossFeature.HISTOGRAM_SUFFIX + "intronic_" + rnaSeqFileName.substring(rnaSeqFileName.lastIndexOf("/") + 1, rnaSeqFileName.lastIndexOf(".")) + "_F_"), normFactor));
		//anti-sense intronic
		subsetUpstreamDensities.put("intronic.tsv", getEnhancerDensity(forwardReads, reverseReads, chr, enhancers, false, senseBinVec, antisenseBinVec, binWidthUpstream, orchestra, geneInfo, false, species, (noHists ? null : enhancerFileName.substring(0, enhancerFileName.lastIndexOf(".")) + FindReadsAcrossFeature.HISTOGRAM_SUFFIX + "intronic_" + rnaSeqFileName.substring(rnaSeqFileName.lastIndexOf("/") + 1, rnaSeqFileName.lastIndexOf(".")) + "_R_"), normFactor));
		//sense intergenic
		subsetDownstreamDensities.put("intergenic.tsv", getEnhancerDensity(forwardReads, reverseReads, chr, enhancers, true, senseBinVec, antisenseBinVec, binWidthDownstream, orchestra, geneInfo, true, species, (noHists ? null : enhancerFileName.substring(0, enhancerFileName.lastIndexOf(".")) + FindReadsAcrossFeature.HISTOGRAM_SUFFIX + "intergenic_" + rnaSeqFileName.substring(rnaSeqFileName.lastIndexOf("/") + 1, rnaSeqFileName.lastIndexOf(".")) + "_F_"), normFactor));
		//anti-sense intergenic
		subsetUpstreamDensities.put("intergenic.tsv", getEnhancerDensity(forwardReads, reverseReads, chr, enhancers, false, senseBinVec, antisenseBinVec, binWidthUpstream, orchestra, geneInfo, true, species, (noHists ? null : enhancerFileName.substring(0, enhancerFileName.lastIndexOf(".")) + FindReadsAcrossFeature.HISTOGRAM_SUFFIX + "intergenic_" + rnaSeqFileName.substring(rnaSeqFileName.lastIndexOf("/") + 1, rnaSeqFileName.lastIndexOf(".")) + "_R_"), normFactor));
	    }
	    else if (lociFileName!=null) {
		subsetDownstreamDensities.put("all.tsv", getLociDensity(forwardReads, chr, loci, true, senseBinVec, binWidthDownstream, orchestra, geneInfo, species, (noHists ? null : lociFileName.substring(0, lociFileName.lastIndexOf(".")) + FindReadsAcrossFeature.HISTOGRAM_SUFFIX + "hr" + time + "c_F_"), normFactor));
		subsetUpstreamDensities.put("all.tsv", getLociDensity(reverseReads, chr, loci, false, antisenseBinVec, binWidthUpstream, orchestra, geneInfo, species, (noHists ? null : lociFileName.substring(0, lociFileName.lastIndexOf(".")) + FindReadsAcrossFeature.HISTOGRAM_SUFFIX + "hr" + time + "c_R_"), normFactor));
	    }
	    else {
		subsetDownstreamDensities.put("all.tsv", getDensity(forwardReads, reverseReads, chr, true, null, isolatedIDs, senseBinVec, binWidthDownstream, senseBinCenterInd, orchestra, geneInfo, exons, shortRNA, intron, maturityFactor, ends, internalStarts, species, (noHists ? null : saveFileName + FindReadsAcrossFeature.HISTOGRAM_SUFFIX + "hr" + time + "c")));
		if (!exons)
		    subsetUpstreamDensities.put("all.tsv", getDensity(forwardReads, reverseReads, chr, false, null, isolatedIDs, antisenseBinVec, binWidthUpstream, antisenseBinCenterInd, orchestra, geneInfo, exons, shortRNA, intron, maturityFactor, ends, internalStarts, species, (noHists ? null : saveFileName + FindReadsAcrossFeature.HISTOGRAM_SUFFIX + "hr" + time + "c")));
	    }
		//Save all of them in files
		String saveName = "";
		if (lociFileName!=null)
		    saveName = saveFileName + "_forward" + (chr!=-1 ? "_" + chr : "") + ".txt";
		else
		    saveName = saveFileName + "_sense" + (chr!=-1 ? "_" + chr : "") + ".txt";
		System.out.println("Saving to " + saveName);
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
		if (subsetDownstreamDensities.size()>1) {
		    bw.write("#bin_center(width=" + binWidthUpstream + ")");
		    for (String s : subsetDownstreamDensities.keySet()) {
			if (s.lastIndexOf(".")>=0)
			    bw.write("\t" + s.substring(s.lastIndexOf("/") + 1, s.lastIndexOf(".")));
		    }
		    bw.write("\n");
		}
		for (int i=0; i<senseBinVec.length; i++) {
		    bw.write(format.format(senseBinVec[i]));
		    for (String s : subsetDownstreamDensities.keySet()) 
			bw.write("\t" + format.format(subsetDownstreamDensities.get(s)[i]*normFactor));
		    bw.write("\n");
		}
		bw.close();
		/////////////////////////////////
		if (!exons) { 
		    if (lociFileName!=null)
			saveName = saveFileName + "_reverse" + (chr!=-1 ? "_" + chr : "") + ".txt";
		    else
			saveName = saveFileName + "_antisense" + (chr!=-1 ? "_" + chr : "") + ".txt";
		    System.out.println("Saving to " + saveName);
		    bw = new BufferedWriter(new FileWriter(saveName));
		    if (subsetUpstreamDensities.size()>1) {
			bw.write("#bin_center(width=" + binWidthUpstream + ")");
			for (String s : subsetUpstreamDensities.keySet()) {
			    if (s.lastIndexOf(".")>=0)
				bw.write("\t" + s.substring(s.lastIndexOf("/") + 1, s.lastIndexOf(".")));
			}
			bw.write("\n");
		    }
		    for (int i=0; i<antisenseBinVec.length; i++) {
			bw.write(format.format(antisenseBinVec[i]));
			for (String s : subsetUpstreamDensities.keySet()) 
			    bw.write("\t" + format.format(subsetUpstreamDensities.get(s)[i]*normFactor));
			bw.write("\n");
		    }
		    bw.close();
		}
		//}
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static double getExonMaturity(int exonInd, Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, int chr, boolean down, String selectFileName, int isolatedIDs[], int binVec[], int binWidth, int binCenterInd, boolean orchestra, RefSeqGenes geneInfo[], String species) {
	int nChr = 21;
	if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
	    nChr = 24;
	int geneIDsKeep[] = FindReadsAcrossFeature.parseSelectFile(selectFileName);
	/////////////////////////////////
	double density[][] = new double[nChr][binVec.length];
	int nGenes[] = new int[nChr], nGenesSum = 0;
	for (int i=1; i<=nChr; i++) {
	    Arrays.fill(density[i-1], 0);
	    if (chr==-1 || i==chr) {
		//System.out.println("Getting info for chr " + i);
		nGenes[i-1] = geneInfo[i-1].getStarts(geneIDsKeep).length;
		nGenesSum += nGenes[i - 1];
		density[i-1] = getDensity(forwardReads.elementAt(i-1), reverseReads.elementAt(i-1), geneInfo[i-1], isolatedIDs, geneIDsKeep, binVec, binWidth, down, false, false, 0, 0, exonInd, 0, false, false, null);
	    }
	}
	double averageDensity[] = new double[binVec.length];
	if (chr==-1) { //Calculate the weighted average
	    for (int i=0; i<nChr; i++) {
		for (int j=0; j<binVec.length; j++) 
		    averageDensity[j] += nGenesSum>0 ? density[i][j]*nGenes[i]/nGenesSum : 0;
	    }
	}
	else 
	    averageDensity = density[chr - 1];
	//Find the bins before the first intron start
	int nSubtract = FindNongenicTranscripts.MRNA_READ_LENGTH/binWidth + 1, nExon = 50/binWidth, nIntron = 300/binWidth;
	double exonDensity = 0, intronDensity = 0;
	for (int i=0; i<nExon; i++) {
	    //System.out.println("Estimating first exon dens " + format.format(averageDensity[binCenterInd - nSubtract - i]));
	    exonDensity += averageDensity[binCenterInd - nSubtract - i]/nExon;
	}
	//Now we can calculate the average of the first intron
	for (int i=0; i<nIntron; i++) {
	    //System.out.println("Estimating first exon dens " + format.format(averageDensity[binCenterInd + 1 + i]));
	    intronDensity += averageDensity[binCenterInd + i + 1]/nIntron;
	}
	System.out.println("density for intron " + exonInd + " is " + format.format(intronDensity/binWidth) + " and preceeding exon density is " + format.format(exonDensity/binWidth) + " with ratio " + format.format(exonDensity/intronDensity));
	return exonDensity/intronDensity; //(firstExonDensity - firstIntronDensity)/binWidth;
    }

    public static double[] getDensity(Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, int chr, boolean down, String selectFileName, int isolatedIDs[], int binVec[], int binWidth, int binCenterInd, boolean orchestra, RefSeqGenes geneInfo[], boolean exons, boolean shortRNA, int intron, double maturityFactor, boolean ends, boolean internalStarts, String species, String saveHistName) {
	int nChr = 21;
	if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
	    nChr = 24;
	int geneIDsKeep[] = FindReadsAcrossFeature.parseSelectFile(selectFileName);
	/////////////////////////////////
	double density[][] = new double[nChr][binVec.length], exonMaturity[] = new double[4];
	int nGenes[] = new int[nChr], nGenesSum = 0;
	if (!exons) {
	    for (int i=0; i<exonMaturity.length; i++)
		exonMaturity[i] = down ? getExonMaturity(i+1, forwardReads, reverseReads, chr, down, selectFileName, isolatedIDs, binVec, binWidth, binCenterInd, orchestra, geneInfo, species) : 0;
	}
	//Now we subtract this from exonic bins
	for (int i=1; i<=nChr; i++) {
	    Arrays.fill(density[i-1], 0);
	    if (chr==-1 || i==chr) {
		nGenes[i-1] = geneInfo[i-1].getStarts(geneIDsKeep).length;
		nGenesSum += nGenes[i - 1];
		//density[i-1] = getDensity(forwardReads.elementAt(i-1), reverseReads.elementAt(i-1), geneInfo[i-1], isolatedIDs, geneIDsKeep, binVec, binWidth, down, exons, shortRNA, firstExonMaturity, (secondExonMaturity + thirdExonMaturity + fourthExonMaturity)/3, 0, 0, ends);
		density[i-1] = getDensity(forwardReads.elementAt(i-1), reverseReads.elementAt(i-1), geneInfo[i-1], isolatedIDs, geneIDsKeep, binVec, binWidth, down, exons, shortRNA, maturityFactor, maturityFactor, intron, 0, ends, internalStarts, saveHistName);
	    }
	}
	double averageDensity[] = new double[binVec.length];
	if (chr==-1) { //Calculate the weighted average
	    for (int i=0; i<nChr; i++) {
		for (int j=0; j<binVec.length; j++) 
		    averageDensity[j] += nGenesSum>0 ? density[i][j]*nGenes[i]/nGenesSum : 0;
	    }
	}
	else 
	    averageDensity = density[chr - 1];
	//System.out.println("Average density is " + Arrays.toString(averageDensity));
	return averageDensity;
    }

    public static int[] getBinVec(int len, int step, boolean sense, boolean exons) {
	len += exons ? 0 : BEFORE_DISTANCE;
	int vec[] = new int[len/step + 1];
	vec[0] = exons ? 0 : /*step/2 - 1 +*/ (sense ? -BEFORE_DISTANCE : BEFORE_DISTANCE);
	for (int i=1; i<vec.length; i++)
	    vec[i] = vec[i-1] + (sense ? step : -step);
	return vec;
    }

    public static void saveFirstIntronStartDensity(RefSeqGenes geneInfo[], int isolatedIDs[], int geneIDsKeep[], String saveFileName) {
	TreeMap<Integer, Integer> intronDistMap = new TreeMap<Integer, Integer>();
	for (int i=0; i<geneInfo.length; i++) {
	    if (geneInfo[i]!=null) {
		for (RefSeqGene g : geneInfo[i].getGenes()) {
		    if (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) {
			if (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0) {
			    if (g.hasIntronInfo() && g.numberOfIntrons()>0) { //Find out if the 1st intron of the gene starts in this bin
				int intronDist = Math.abs(g.getStart() - g.getIntronStart(0));
				intronDistMap.put(intronDist, intronDistMap.containsKey(intronDist) ? intronDistMap.get(intronDist) + 1 : 1);
			    }
			}
		    }
		}
	    }
	}
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    for (Integer d : intronDistMap.keySet())
		bw.write(d + "\t" + intronDistMap.get(d) + "\n");
	    bw.close();
	}
	catch (IOException e) {

	}
    }

    public static double[] getDensity(ABIRead forwardReads[], ABIRead reverseReads[], RefSeqGenes geneInfo, int isolatedIDs[], int geneIDsKeep[], int binVec[], int binWidth, boolean sense, boolean exons, boolean shortRNA, double maturityFactorFirst, double maturityFactorLater, int intronInd, double firstExonMaturity, boolean ends, boolean internalStarts, String saveHistName) {
	double density[] = new double[binVec.length], exonDensity[] = new double[binVec.length];
	int nGenesWithoutExon[] = new int[binVec.length], nGenesWithExon[] = new int[binVec.length];
	for (int i=0; i<binVec.length; i++) {
	    for (RefSeqGene g : geneInfo.getGenes()) {
		for (int j=0; j<g.numberOfTSS(); j++) {
		    if ((internalStarts && j>0) || (!internalStarts && j==0)) {
			if (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>=0) {
			    if (isolatedIDs==null || Arrays.binarySearch(isolatedIDs, g.getID())>=0) {
				if (intronInd==0 || (g.numberOfExons()>intronInd/* && Math.abs(g.getStart() - g.getIntronStart(0))<400 && Math.abs(g.getStart() - g.getIntronStart(0))>300*/)) {
				    int start, end, tss = ends ? g.getEnd() : intronInd>0 ? g.getIntronStart(intronInd-1) : g.getStart(j);
				    ABIRead[] reads;
				    if (g.getStrand()) {
					start = tss + binVec[i] - binWidth/2;
					end = tss + binVec[i] + binWidth/2;
					reads = sense ? forwardReads : reverseReads;
				    }
				    else {
					start = tss - binVec[i] - binWidth/2;
					end = tss - binVec[i] + binWidth/2;
					reads = sense ? reverseReads : forwardReads;
				    }
				    if (includeDensity(sense, exons, shortRNA, start, end, binVec[i]+binWidth/2, g, ends, intronInd)) {
					density[i] += FindReadsAcrossFeature.numberOfReads(geneInfo.getChromosome(), start, end, reads); //>0 ? 1 : 0;
					nGenesWithoutExon[i]++;
				    }
				    else {
					exonDensity[i] += FindReadsAcrossFeature.numberOfReads(geneInfo.getChromosome(), start, end, reads); //>0 ? 1 : 0;
					nGenesWithExon[i]++;
				    }
				}
			    }
			}
		    }
		}
	    }
	    //if (binVec[i]>=-10 && binVec[i]<=40 && sense)
	    //System.out.println("density is " + format.format(density[i]) + " for " + nGenesWithoutExon[i] + " and exon-density is " + format.format(exonDensity[i]/maturityFactorFirst + 0*binWidth*firstExonMaturity*nGenesWithExon[i]) + " for " + nGenesWithExon[i] + " with maturity " + format.format(firstExonMaturity) + " for " + binVec[i]);
	    if (maturityFactorFirst>0) {
		int n = nGenesWithoutExon[i] + nGenesWithExon[i];
		density[i] /= n>0 ? n : 1;
		exonDensity[i] /= n>0 ? n : 1;
		//if (binVec[i]>-4000 && binVec[i]<0 && sense) 
		//  density[i] = (density[i]*2*NormalDist.cdf(0, 2000, binVec[i]))/maturityFactor + density[i]*(1 - 2*NormalDist.cdf(0, 2000, binVec[i]));
		//density[i] += (exonDensity[i]/maturityFactor);
		if (binVec[i]>=0 && binVec[i]<1000)
		    density[i] += (exonDensity[i]/maturityFactorFirst);
		else if (binVec[i]>=1000)
		    density[i] += (exonDensity[i]/maturityFactorLater);
		    //  density[i] += exonDensity[i]/(maturityFactor*NormalDist.cdf(0, 2000, binVec[i])); 
		
	    }
	    else
		density[i] /= nGenesWithoutExon[i]>0 ? nGenesWithoutExon[i] : 1;
	}
	if (saveHistName!=null) { //Save histograms
	    try {
		String saveName = saveHistName + "_" + (sense ? "S" : "AS") + "_" + geneInfo.getChromosome() + ".txt";
		System.out.println("Saving histograms to " + saveName);
		BufferedWriter bw = saveHistName==null ? null : new BufferedWriter(new FileWriter(saveName));
		for (RefSeqGene g : geneInfo.getGenes()) {
		    int h[] = FindReadsAcrossFeature.getReadsHistogram(g.getChromosome(), g.getStart(), binWidth*binVec.length/2, binWidth*binVec.length/2, g.getStrand() ? (sense ? forwardReads : reverseReads) : sense ? reverseReads : forwardReads);
		    if (!g.getStrand())
			h = FindReadsAcrossFeature.flipArray(h);
		    bw.write(g.getStart() + "\t");
		    for (int i=0; i<h.length; i++)
			bw.write(format.format(h[i]) + "\t");
		    bw.write("\n");
		}
		bw.close();
	    }
	    catch (IOException e) {
		System.err.println("Error saving histogram to file " + e.toString());
	    }
	}
	//System.out.println("Found " + Arrays.toString(nGenesWithoutExon) + " genes without exons.");
	//System.out.println("Found densities " + Arrays.toString(density));
	return density;
    }

    public static int[] numberOfGenesWithFeature(RefSeqGenes geneInfo, int isolatedIDs[], int geneIDsKeep[], int binVec[], int binWidth, boolean sense, boolean exons) {
	//Count the number of genes that have either a fully exonic or
	//intronic bin at a given location
	int nGenes[] = new int[binVec.length];
	
	return nGenes;
    }

    public static boolean includeDensity(boolean sense, boolean exons, boolean shortRNA, int start, int end, int len, RefSeqGene g, boolean ends, int intronInd) {
	if (!sense || shortRNA)
	    return true;
	if (ends) {
	    if (-len>g.getLength())
		return false;
	    if (g.getStrand() && start>g.getEnd())
		return true;
	    if (!g.getStrand() && end<g.getEnd())
		return true;
	    return exons ? g.exonOnly(start, end) : g.intronOnly(start, end);
	}
	else {
	    if (len>g.getLength())
		return false;
	    if (g.getStrand() && (end<g.getStart() || (intronInd>0 && end<g.getIntronStart(intronInd-1))))
		return true;
	    if (!g.getStrand() && (start>g.getStart() || (intronInd>0 && start>g.getIntronStart(intronInd-1))))
		return true;
	    return exons ? g.exonOnly(start, end) : g.intronOnly(start, end);
	}
    }

    ////////////////////////////////////////////

    public static double[] getEnhancerDensity(Vector<ABIRead[]> readsForward, Vector<ABIRead[]> readsReverse, int chr, Vector<Enhancer> enhancers, boolean sense, int binVecForward[], int[] binVecReverse, int binWidth, boolean orchestra, RefSeqGenes geneInfo[], boolean nonGenic, String species, String saveHistName, double normFactor) {
	int nChr = 21;
	if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
	    nChr = 24;
	/////////////////////////////////
	double density[][] = new double[nChr][binVecReverse.length];
	int nEnhancers[] = new int[nChr], nEnhancersSum = 0;
	for (int i=1; i<=nChr; i++) {
	    Arrays.fill(density[i-1], 0);
	    if (chr==-1 || i==chr) {
		//Count the number of nonGenic enhancers
		for (Enhancer e : enhancers) {
		    if (e.getChromosome()==i) {
			RefSeqGene g = geneInfo[i-1].getOverlappingGene(e.getCenter());
			if ((nonGenic && g==null) || (!nonGenic && g!=null))
			    nEnhancers[i-1]++;
		    }
		}
		nEnhancersSum += nEnhancers[i - 1];
		density[i-1] = getEnhancerDensity(i, readsForward.elementAt(i-1), readsReverse.elementAt(i-1), enhancers, binVecForward, binVecReverse, binWidth, sense, nEnhancers[i-1], saveHistName==null ? null : saveHistName + i + ".txt", normFactor, geneInfo, nonGenic);
	    }
	}
	double averageDensity[] = new double[binVecReverse.length];
	if (chr==-1 && nEnhancersSum>0) { //Calculate the weighted average
	    for (int i=0; i<nChr; i++) {
		for (int j=0; j<averageDensity.length; j++) 
		    averageDensity[j] += density[i][j]*nEnhancers[i]/nEnhancersSum;
	    }
	}
	else 
	    averageDensity = density[chr - 1];
	//System.out.println("Average density is " + Arrays.toString(averageDensity));
	return averageDensity;
    }

    public static double[] getEnhancerDensity(int chr, ABIRead readsForward[], ABIRead readsReverse[], Vector<Enhancer> enhancers, int binVecForward[], int binVecReverse[], int binWidth, boolean sense, int nEnhancers, String saveHistName, double normFactor, RefSeqGenes geneInfo[], boolean nonGenic) {
	double density[] = new double[binVecReverse.length];
	try {
	    int nEns[] = new int[density.length];
	    BufferedWriter bw = saveHistName==null ? null : new BufferedWriter(new FileWriter(saveHistName));
	    for (Enhancer e : enhancers) {
		if (e.getChromosome()==chr) {
		    RefSeqGene g = geneInfo[chr-1].getOverlappingGene(e.getCenter());
		    if((nonGenic && g==null) || (!nonGenic && g!=null)) {
			//Use the strand of the nearest gene
			g = nonGenic ? geneInfo[chr-1].getNearestTSSGene(e.getCenter(), false) : g;
			boolean strand = g.getStrand();
			int center = e.getCenter(), start, end, binVec[] = null;
			ABIRead[] reads = null;
			if (strand) {
			    reads = sense ? readsForward : readsReverse;
			    binVec = sense ? binVecForward : binVecReverse;
			}
			else {
			    reads = sense ? readsReverse : readsForward;
			    binVec = sense ? binVecForward : binVecReverse;//binVec = sense ? binVecReverse : binVecForward;
			}
			if (reads!=null) {
			    for (int i=0; i<binVec.length; i++) {
				if (strand) {
				    start = center + binVec[i] - binWidth/2;
				    end = center + binVec[i] + binWidth/2;
				}
				else {
				    start = center - binVec[i] - binWidth/2;
				    end = center - binVec[i] + binWidth/2;
				}
				density[i] += FindReadsAcrossFeature.numberOfReads(chr, start, end, reads); //>0 ? 1 : 0;
				nEns[i]++;
			    }
			}
			if (bw!=null) { //Get the histogram
			    int h[] = FindReadsAcrossFeature.getReadsHistogram(chr, center, binWidth*binVec.length/2, binWidth*binVec.length/2, reads);
			    bw.write(center + "\t");
			    for (int i=0; i<h.length; i++)
				bw.write(format.format(normFactor*(double)h[i]) + "\t");
			    bw.write("\n");
			}
		    }
		}
	    }
	    for (int i=0; i<density.length; i++)
		density[i] /= nEns[i]>0 ? nEns[i] : 1;
	    if (bw!=null) {
		bw.close();
		System.out.println("Saved histogram to " + saveHistName);
	    }
	}
	catch (IOException e) {
	    System.err.println("Error saving histogram to file " + saveHistName);
	}
	return density;
    }

    /////////////////////////////////

    public static double[] getLociDensity(Vector<ABIRead[]> reads, int chr, Vector<int[]> loci, boolean forward, int binVec[], int binWidth, boolean orchestra, RefSeqGenes geneInfo[], String species, String saveHistName, double normFactor) {
	int nChr = 21, nLoci = 0;
	if (species.toLowerCase().equals("hs") || species.toLowerCase().equals("human"))
	    nChr = 24;
	/////////////////////////////////
	double density[][] = new double[nChr][binVec.length];
	for (int i=1; i<=nChr; i++) {
	    Arrays.fill(density[i-1], 0);
	    if (chr==-1 || i==chr) {
		density[i-1] = getLociDensity(i, reads.elementAt(i-1), loci.elementAt(i-1), binVec, binWidth, forward, saveHistName + i + ".txt", normFactor, geneInfo);
		nLoci += loci.elementAt(i-1).length;
	    }
	}
	double averageDensity[] = new double[binVec.length];
	if (chr==-1) { //Calculate the weighted average
	    for (int i=0; i<nChr; i++) {
		for (int j=0; j<binVec.length; j++) 
		    averageDensity[j] += density[i][j]*loci.elementAt(i).length/nLoci;
	    }
	}
	else 
	    averageDensity = density[chr - 1];
	//System.out.println("Average density is " + Arrays.toString(averageDensity));
	return averageDensity;
    }

    public static double[] getLociDensity(int chr, ABIRead reads[], int[] loci, int binVec[], int binWidth, boolean forward, String saveHistName, double normFactor, RefSeqGenes geneInfo[]) {
	double density[] = new double[binVec.length];
	try {
	    for (int j=0; j<loci.length; j++) {
		for (int i=0; i<binVec.length; i++) {
		    int start = loci[j] + binVec[i] - binWidth/2;
		    int end = loci[j] + binVec[i] + binWidth/2;
		    //if (includeDensity(sense, exons, shortRNA, start, end, binVec[i]+binWidth/2, g, ends)) {
		    density[i] += FindReadsAcrossFeature.numberOfReads(chr, start, end, reads)>0 ? 1 : 0;
		    //}
		}
	    }
	    for (int i=0; i<binVec.length; i++)
		density[i] /= loci.length;
	    //if (saveHistName!=null)
	    //FindReadsAcrossFeature.getHistograms(chr, loci, reads, null, binWidth*binVec.length/2, binWidth*binVec.length/2, FindNongenicTranscripts.MRNA_READ_LENGTH, saveHistName, normFactor, 0);
	    
	}
	catch (Exception e) {
	    System.err.println("Error saving histogram to file " + saveHistName);
	}
	return density;
    }

    ///////////////////////////////////////////////////////

    public static double calculateExonDensity(RefSeqGene g, ABIRead reads[], int exon) {
	double density = 0;
	if (g.numberOfExons()>exon) {
	    int start = g.getExonStart(exon), end = g.getExonStop(exon), len = Math.abs(start - end);
	    int nReads = FindReadsAcrossFeature.numberOfReads(g.getChromosome(), start, end, reads);
	    if (exon==0) { //Add the 5'UTR as well
		start = g.getUTRStart(true);
		end = g.getUTRStop(true);
		len += Math.abs(start - end);
		nReads += FindReadsAcrossFeature.numberOfReads(g.getChromosome(), start, end, reads);
	    }
	    if (exon==g.numberOfExons()-1) { //Add the 3'UTR as well
		start = g.getUTRStart(false);
		end = g.getUTRStop(false);
		len += Math.abs(start - end);
		nReads += FindReadsAcrossFeature.numberOfReads(g.getChromosome(), start, end, reads);
	    }
	    density = (double)nReads/len;
	}
	return density;
    }

    public static Vector<String> parseSelectFiles(String selectFileName) {
	//Find the files in the directory that contains the geneIDs that we want to investigate
	Vector<String> fileNames = new Vector<String>();
	if (selectFileName!=null) {
	    if (selectFileName.endsWith(".tsv") || selectFileName.endsWith(".txt"))
		fileNames.addElement(selectFileName);
	    else { //Parse the directory
		String dirFiles[] = (new File(selectFileName)).list();
		for (int i=0; i<dirFiles.length; i++) {
		    if (dirFiles[i].endsWith(".tsv") || dirFiles[i].endsWith(".txt")) {
			fileNames.add(selectFileName + dirFiles[i]);
			//System.out.println(dirFiles[i]);
		    }
		}
	    }
	}
	return fileNames;
    }

    public static Vector<int[]> parseLociFile(String lociFileName) {
	//use this method to parse files containing loci which we want
	//to investigate. Assume that each line starts with chr\tpos
	//and don't care about the rest.
	Vector<int[]> loci = new Vector<int[]>();
	try {
	    Vector<Integer> nLoci = new Vector<Integer>(); //Find the number of loci per chromosome
	    nLoci.addElement(0);
	    int chr = 0;
	    BufferedReader br = new BufferedReader(new FileReader(lociFileName));
	    Pattern lociPattern = Pattern.compile("^(\\d+)[\\s+|:](\\d+)");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = lociPattern.matcher(line);
		if (m.find()) {
		    int c = Integer.parseInt(m.group(1));
		    if (c!=chr && chr!=0) {
			loci.addElement(new int[nLoci.lastElement()]);
			nLoci.addElement(0);
		    }
		    chr = c;
		    nLoci.setElementAt(nLoci.lastElement() + 1, nLoci.size() - 1);
		}
		line = br.readLine();
	    }
	    loci.addElement(new int[nLoci.lastElement()]);
	    br.close();
	    br = new BufferedReader(new FileReader(lociFileName));
	    line = br.readLine();
	    int i[] = new int[loci.size()];
	    while (line!=null) {
		Matcher m = lociPattern.matcher(line);
		if (m.find()) {
		    int c = Integer.parseInt(m.group(1));
		    int p = Integer.parseInt(m.group(2));
		    loci.elementAt(c-1)[i[c-1]++] = p;
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading loci file " + lociFileName);
	}
	return loci;
    }

}
