//Use this method to investigate the correlation between enhancers and
//expression levels of nearby genes. This can be done by focusing on
//either the gene or the enhancers. As a first step, we are going to
//investigate the number of enhancers in the proximity of highly
//expressed genes.

//Example:

//java -Xmx512m ExpressionEnhancerCorrelation  -e Enhancers/enhancers_null_KCl.txt

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class ExpressionEnhancerCorrelation {

    public static final String TSS_2_ENHANCER_DISTANCE_SUFFIX = "_tss2enhancer_dists";

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption('f', "fdr");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    //Location of the enhancer list
	    CmdLineParser.Option enhancerFileNameFlag = argParser.addStringOption('e', "enhancerFileName");
	    CmdLineParser.Option maxEnhancer2TSSDistanceFlag = argParser.addIntegerOption('d', "maxEnhancer2TSSDistance");
	    //Use this flag to determine which expression profiles we are interested
	    CmdLineParser.Option expressionPatternFlag = argParser.addStringOption('E', "expressionPattern");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(maxEnhancer2TSSDistanceFlag);
	    int maxEnhancer2TSSDistance = tmp!=null ? ((Integer)tmp).intValue() : 2000000;
	    tmp = argParser.getOptionValue(enhancerFileNameFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "Enhancers/enhancers_null_KCl.txt";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : enhancerFileName.substring(0, enhancerFileName.lastIndexOf(".") - 1);
	    tmp = argParser.getOptionValue(expressionPatternFlag);
	    String expressionPattern = tmp!=null ? (String)tmp : "LHH|LHM|LHL|LMH";
	    String expressionPatternList[] = expressionPattern.split("\\|");
	    //////////////////////////////////////////////////////////////////
	    int nChr = 21;
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    TreeSet<Integer> tss2enhancerDistances = new TreeSet<Integer>();
	    //Load the enhancer sites
	    Vector<Enhancer> enhancers = FindReadsAcrossFeature.parseEnhancerFile(enhancerFileName);
	    //Load the expression info which has already been
	    //processed using the SymbolicDynamicsExpression.java
	    //class.
	    int ids2investigate[] = parseSymbolicExpressionFile(expressionPatternList, orchestra);
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || chr==i) {
		    //Load the annotation info
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra);
		    //Find the distance to the genes which we want to investigate
		    int ids[] = geneInfo[i-1].getIDs();
		    Arrays.sort(ids);
		    System.out.println("Loaded info for " + ids.length + " genes from chr " + i);
		    for (int j=0; j<ids2investigate.length; j++) {
			if (Arrays.binarySearch(ids, ids2investigate[j])>=0) {
			    RefSeqGene g = geneInfo[i-1].getGene(ids2investigate[j]);
			    System.out.println("Looking for enhancers near " + g.toString());
			    for (Enhancer en : enhancers) {
				if (en.getChromosome()==i && Math.abs(en.getCenter() - g.getStart())<maxEnhancer2TSSDistance) {
				    System.out.println("Found enhancer " + en.toString() + " at distance " + (g.getStart() - en.getCenter()) + " near " + g.toString());
				    tss2enhancerDistances.add(g.getStart() - en.getCenter());
				}
			    }
			}
		    }
		}
	    }
	    ///////////////////////////////////////////////////////
	    String saveName = saveFileName + TSS_2_ENHANCER_DISTANCE_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt";
	    System.out.println("Saving TSS distances to " + saveName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	    for (int d : tss2enhancerDistances) {
		bw.write(d + "\n");
	    }
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static int[] parseSymbolicExpressionFile(String expressionPatternList[], boolean orchestra) {
	//Get the geneIDs in those files that match expressionPattern
	Vector<Integer> ids = new Vector<Integer>();
	try {
	    File dir = new File("mRNA/SymbolicDynamics/");
	    File fs[] = dir.listFiles();
	    Vector<File> files = new Vector<File>();
	    for (int i=0; i<fs.length; i++) {
		for (int j=0; j<expressionPatternList.length; j++) {
		    if (fs[i].getName().contains(expressionPatternList[j])) {
			files.addElement(fs[i]);
			break;
		    }
		}
	    }
	    Pattern symbolicPattern = Pattern.compile("(\\d+)\\s+");
	    for (File f : files) {
		System.out.println("Opening file " + f.getName());
		BufferedReader br = new BufferedReader(new FileReader(f));
		String line = br.readLine();
		while (line!=null) {
		    Matcher m = symbolicPattern.matcher(line);
		    if (m.find())
			ids.addElement(Integer.parseInt(m.group(1)));
		    line = br.readLine();
		}
		br.close();
	    }
	}
	catch (IOException e) {
	    System.err.println("Error extracting ids in ExpressionEnhancerCorrelation.java from SymbolicDynamics files: " + e.toString());
	}
	int id[] = new int[ids.size()], j = 0;
	for (Integer i : ids) { id[j++] = i; }
	System.out.println("Found " + id.length + " genes to investigate.");
	return id;
    }


}

