//Use this class to store the bins used for classyfying windows and
//their different cutoffs for the z-statistic and the maximum #inputs.

import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import org.apache.commons.math.stat.regression.SimpleRegression;
import org.apache.commons.math.distribution.PoissonDistributionImpl;
import umontreal.iro.lecuyer.probdist.PoissonDist;

public class FDRCutOffs {

    public static final int FRAGMENT_LENGTH = 120;
    public static final String WIN_STAT_CHIP_SUFFIX = "_win_stat_chip_";
    public static final String WIN_STAT_INPUT_SUFFIX = "_win_stat_input_";
    public static final String WIN_STAT_DIFF_SUFFIX = "_win_stat_diff_";
    public static final String CUT_OFF_SUFFIX = "_win_cutoffs_";
    public static DecimalFormat format = new DecimalFormat("#####.#####");
    //Implicitly assume that all bins have the same size

    protected int cutOffs[][], nMappabilityBins, nGCBins, maxMappability, mappabilityBinSize, gcBinSize, maxGC;
    private int[][] maxInputReads;
    private TreeMap[][] readsHist, inputHist, diffHist;
    protected TreeMap[][] nullDist, localRatio;
    private TreeMap[][] nullDistInput, localRatioInput;
    protected double[][] p0;
    private double[][] p0Input, readPoissonRatio, inputPoissonRatio;
    protected double p0Average = 0;
    private double p0InputAverage = 0;
    protected long totWins = 0, nInputWindows[][] = null;

    //////////////////////////////////////////////////////////////////

    public FDRCutOffs(String fileStart, double fdr, boolean locFDR) {
	//Use this constructor when reading pre-computed results from Matlab.
	String fileName = fileStart + CUT_OFF_SUFFIX + (locFDR ? "loc" : "") + "fdr_" + format.format(fdr) + ".txt";
	try { //Try loading from file first
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine(); //First line is the header
	    line = br.readLine();
	    TreeSet mappabilities = new TreeSet(), gcs = new TreeSet();
	    Pattern cutOffPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = cutOffPattern.matcher(line);
		if (m.find()) {
		    Integer map = new Integer(m.group(1)), gc = new Integer(m.group(2));
		    mappabilities.add(map);
		    gcs.add(gc);
		}
		line = br.readLine();
	    }
	    br.close();
	    //Go through again now that we know how many bins there are
	    nMappabilityBins = mappabilities.size();
	    nGCBins = gcs.size();
	    maxMappability = ((Integer)mappabilities.last()).intValue();
	    maxGC = ((Integer)gcs.last()).intValue();
	    mappabilityBinSize = maxMappability/nMappabilityBins;
	    gcBinSize = maxGC/nGCBins;
	    cutOffs = new int[nMappabilityBins][nGCBins];
	    maxInputReads = new int[nMappabilityBins][nGCBins];
	    br = new BufferedReader(new FileReader(fileName));
	    line = br.readLine(); //First line is the header
	    line = br.readLine();
	    while (line!=null) {
		Matcher m = cutOffPattern.matcher(line);
		if (m.find()) {
		    Integer map = new Integer(m.group(1)), gc = new Integer(m.group(2)), cut = new Integer(m.group(3)), max = new Integer(m.group(4));
		    int mapInd = map.intValue()/mappabilityBinSize - 1, gcInd = gc.intValue()/gcBinSize - 1;
		    cutOffs[mapInd][gcInd] = cut.intValue();
		    maxInputReads[mapInd][gcInd] = max.intValue();
		    System.out.println("Found cut-off " + cutOffs[mapInd][gcInd] + " and max input " + maxInputReads[mapInd][gcInd]);
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Failed to load file with stratified pre-calculated cut-offs for " + fileName);
	}
    }

    public FDRCutOffs(String fileStart, int windowSize, int nMapB, int nGCB) {
	//Use this constructor when we want to calculate the whole
	//distribution. This will allow us to calculate the critical
	//FDR value for each window. By storing this, it will be
	//easier to perform the analysis of how many peaks are
	//detected for a given cut-off.
	nMappabilityBins = nMapB;
	nGCBins = nGCB;
	String str = "(-?\\d+)";
	for (int i=0; i<nMappabilityBins*nGCBins; i++) { str += "\\s+(\\d+)"; }
	Pattern pattern = Pattern.compile(str);
	maxMappability = (windowSize - FRAGMENT_LENGTH)*2;
	maxGC = windowSize;
	mappabilityBinSize = maxMappability/nMappabilityBins;
	gcBinSize = maxGC/nGCBins;
	//Normalize the window counts
	SimpleRegression simpleRegression[][] = new SimpleRegression[nMappabilityBins][nGCBins], simpleRegressionInput[][] = new SimpleRegression[nMappabilityBins][nGCBins];
	readsHist = new TreeMap[nMappabilityBins][nGCBins];
	inputHist = new TreeMap[nMappabilityBins][nGCBins];
	diffHist = new TreeMap[nMappabilityBins][nGCBins];
	nullDist = new TreeMap[nMappabilityBins][nGCBins];
	localRatio = new TreeMap[nMappabilityBins][nGCBins];
	nullDistInput = new TreeMap[nMappabilityBins][nGCBins];
	localRatioInput = new TreeMap[nMappabilityBins][nGCBins];
	p0 = new double[nMappabilityBins][nGCBins];
	p0Input = new double[nMappabilityBins][nGCBins];
	for (int i=0; i<nMappabilityBins; i++) { //Initialize
	    for (int j=0; j<nGCBins; j++) {
		readsHist[i][j] = new TreeMap();
		inputHist[i][j] = new TreeMap();
		diffHist[i][j] = new TreeMap();
		nullDist[i][j] = new TreeMap();
		localRatio[i][j] = new TreeMap();
		simpleRegression[i][j] = new SimpleRegression();
		nullDistInput[i][j] = new TreeMap();
		localRatioInput[i][j] = new TreeMap();
		simpleRegressionInput[i][j] = new SimpleRegression();
	    }
	}
	for (int i=1; i<=21; i++) {
	    //Open the file containing the histogram for the number of
	    //windows with a given read statistic.
	    readsHist = parseStatFile(fileStart + WIN_STAT_CHIP_SUFFIX + i + ".txt", pattern, readsHist);
	    inputHist = parseStatFile(fileStart + WIN_STAT_INPUT_SUFFIX + i + ".txt", pattern, inputHist);
	    diffHist = parseStatFile(fileStart + WIN_STAT_DIFF_SUFFIX + i + ".txt", pattern, diffHist);
	}
	//Compute the empirical PDF
	long[][] nReadWindows = new long[nMappabilityBins][nGCBins];
	nInputWindows = new long[nMappabilityBins][nGCBins];
	long[][] nReads = new long[nMappabilityBins][nGCBins], nInput = new long[nMappabilityBins][nGCBins];
	readPoissonRatio = new double[nMappabilityBins][nGCBins];
	inputPoissonRatio = new double[nMappabilityBins][nGCBins];
	for (int i=0; i<nMappabilityBins; i++) { //Find the total number of windows in each bin and the mean for ChIP and input
	    for (int j=0; j<nGCBins; j++) {
		Iterator it = readsHist[i][j].keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    nReadWindows[i][j] += ((Long)readsHist[i][j].get(key)).longValue();
		    nReads[i][j] += ((Long)readsHist[i][j].get(key)).longValue()*key.intValue();
		}
		it = inputHist[i][j].keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    nInputWindows[i][j] += ((Long)inputHist[i][j].get(key)).longValue();
		    nInput[i][j] += ((Long)inputHist[i][j].get(key)).longValue()*key.intValue();
		}
		readPoissonRatio[i][j] = (double)nReads[i][j]/nReadWindows[i][j];
		inputPoissonRatio[i][j] = (double)nInput[i][j]/nInputWindows[i][j];
		//System.out.println(readPoissonRatio[i][j] + "\t" + inputPoissonRatio[i][j] + "\t" + nReadWindows[i][j]);
	    }
	}
	for (int i=0; i<nMappabilityBins; i++) { 
	    for (int j=0; j<nGCBins; j++) { 
		Iterator it = (diffHist[i][j].keySet()).iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    double empiricalProb = (double)((Long)diffHist[i][j].get(key)).longValue()/nReadWindows[i][j];
		    double nullProb = inputPoissonRatio[i][j]>0 ? PeakFinder2.skellamPDF(key.intValue(), readPoissonRatio[i][j], inputPoissonRatio[i][j]) : PoissonDist.prob(readPoissonRatio[i][j], key.intValue());
		    diffHist[i][j].put(key, new Double(empiricalProb));
		    //Compute the null-distribution as a Skellam distribution.
		    nullDist[i][j].put(key, new Double(nullProb));
		    //System.out.println(i + "\t" + j + "\t" + key + "\t" + empiricalProb + "\t" + nullProb);
		    simpleRegression[i][j].addData(empiricalProb, nullProb);
		}
		//Now we normalize the input distribution to find the
		//maximal allowed count
		it = (inputHist[i][j].keySet()).iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    double empiricalProb = (double)((Long)inputHist[i][j].get(key)).longValue()/nInputWindows[i][j];
		    double nullProb = inputPoissonRatio[i][j]>0 ? (new PoissonDistributionImpl(inputPoissonRatio[i][j])).probability(key.intValue()) : 0; //Poisson
		    inputHist[i][j].put(key, new Double(empiricalProb));
		    nullDistInput[i][j].put(key, new Double(nullProb));
		    simpleRegressionInput[i][j].addData(empiricalProb, nullProb);
		}
	    }
	}
	//Find the FDRs corresponding to different #read-#input values
	for (int i=0; i<nMappabilityBins; i++) {
	    for (int j=0; j<nGCBins; j++) {
		//calculate p0, the proportion of reads which are in the
		//background. This is done via linear regression between the
		//null and empirical distribution.
		p0[i][j] = simpleRegression[i][j].getSlope();
		p0[i][j] = p0[i][j]>1 || Double.isNaN(p0[i][j]) || Double.isInfinite(p0[i][j]) ? 1 : p0[i][j]; //Make sure that it is a valid value
		p0Average += p0[i][j]*nReadWindows[i][j];
		totWins += nReadWindows[i][j];
		int lastValueGreaterThanOne = -1000;
		Iterator it = diffHist[i][j].keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    double ratio = p0[i][j]*((Double)nullDist[i][j].get(key)).doubleValue()/((Double)diffHist[i][j].get(key)).doubleValue();
		    ratio = Double.isNaN(ratio) || Double.isInfinite(ratio) ? 1 : ratio; //Make sure that it is a valid value
		    localRatio[i][j].put(key, new Double(ratio));
		    lastValueGreaterThanOne = ratio>1 ? key.intValue() : lastValueGreaterThanOne;
		}
		//Now we can change the ratios so that they become 1
		//in the whole regime where we only have a null
		//distribution and the non-null is not present.
		it = diffHist[i][j].keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    if (key.intValue()<lastValueGreaterThanOne)
			localRatio[i][j].put(key, new Double("1"));
		}
		//Do the same analysis for the input
		p0Input[i][j] = simpleRegressionInput[i][j].getSlope();
		p0Input[i][j] = p0Input[i][j]>1 || Double.isNaN(p0Input[i][j]) || Double.isInfinite(p0Input[i][j]) ? 1 : p0Input[i][j]; //Make sure that it is a valid value
		p0InputAverage += p0Input[i][j]*nInputWindows[i][j];
		it = inputHist[i][j].keySet().iterator();
		lastValueGreaterThanOne = -1000;
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    double ratio = p0Input[i][j]*((Double)nullDistInput[i][j].get(key)).doubleValue()/((Double)inputHist[i][j].get(key)).doubleValue();
		    localRatioInput[i][j].put(key, new Double(ratio));
		    lastValueGreaterThanOne = ratio>1 ? key.intValue() : lastValueGreaterThanOne;
		}
		it = inputHist[i][j].keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    if (key.intValue()<lastValueGreaterThanOne)
			localRatioInput[i][j].put(key, new Double("1"));
		}
	    }
	}
	p0Average /= totWins;
	p0InputAverage /= totWins;
    }


    //////////////////////////////////////////////////////////////////

    public static TreeMap[][] parseStatFile(String fileName, Pattern pattern, TreeMap hists[][]) {
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = pattern.matcher(line);
		if (m.find()) {
		    Integer key = new Integer(m.group(1));
		    //Assign the values for each histogram,
		    //corresponding to different bins in the
		    //mappability GC-content pairs.
		    for (int i=0; i<hists.length; i++) {
			for (int j=0; j<hists[i].length; j++) {
			    long val = hists[i][j].containsKey(key) ? ((Long)hists[i][j].get(key)).longValue() : 0;
			    hists[i][j].put(key, new Long(val + Long.parseLong(m.group(1 + i*hists[i].length + j + 1))));
			}
		    }
		}
		else if (!line.startsWith("#"))
		    System.err.println("Pattern: " + pattern.toString() + " does not match line \n" + line);
		line = br.readLine();
	    }  
	}
	catch (IOException e) {
	    System.err.println("Error reading from " + fileName);
	    return null;
	}
	return hists;
    }


    public int getDiffThreshold(int mappability, int gcContent) {
	int mapInd = mappability/mappabilityBinSize, gcInd = gcContent/gcBinSize;
	if (mapInd==nMappabilityBins) { mapInd--; }
	if (gcInd==nGCBins) { gcInd--; }
	return cutOffs[mapInd][gcInd];
    }

    public int getDiffThreshold(double fdr, int mappability, int gcContent) {
	//Use this method to calculate the threshold for what
	//difference is required to detect a window for the given GC
	//and mappability combination.
	int mapInd = getMappabilityIndex(mappability), gcInd = getGCContentIndex(gcContent);
	int key = ((Integer)localRatio[mapInd][gcInd].firstKey()).intValue() + 1;
	while (((Double)localRatio[mapInd][gcInd].get(new Integer(key))).doubleValue()>fdr) {
	    //System.out.println(((Double)localRatio[mapInd][gcInd].get(new Integer(key))).doubleValue() + "\t" + key);
	    key++;
	}
	//System.out.println("Returning " + key + " for fdr " + fdr);
	return key;
    }

    public int getMaxInput(int mappability, int gcContent) {
	int mapInd = mappability/mappabilityBinSize, gcInd = gcContent/gcBinSize;
	if (mapInd==nMappabilityBins) { mapInd--; }
	if (gcInd==nGCBins) { gcInd--; }
	return maxInputReads[mapInd][gcInd];
    }

    public double getFDRThreshold(int mappability, int gcContent, int nReads, int nInput) {
	//Use this method to calculate the FDR-value at which the
	//given window can be detected.
	int mapInd = mappability/mappabilityBinSize, gcInd = gcContent/gcBinSize, z = nReads - nInput;
	if (mapInd==nMappabilityBins) { mapInd--; }
	if (gcInd==nGCBins) { gcInd--; }
	if (mapInd<nMappabilityBins && gcInd<nGCBins && localRatio[mapInd][gcInd].containsKey(new Integer(z)))
	    return ((Double)localRatio[mapInd][gcInd].get(new Integer(z))).doubleValue();
	return 0;
    }

    public double getInputFDRThreshold(int mappability, int gcContent, int nInput) {
	//Use this method to calculate the FDR-value at which the
	//given window can be detected.
	int mapInd = mappability/mappabilityBinSize, gcInd = gcContent/gcBinSize;
	if (mapInd==nMappabilityBins) { mapInd--; }
	if (gcInd==nGCBins) { gcInd--; }
	if (mapInd<nMappabilityBins && gcInd<nGCBins && localRatioInput[mapInd][gcInd].containsKey(new Integer(nInput)))
	    return ((Double)localRatioInput[mapInd][gcInd].get(new Integer(nInput))).doubleValue();
	return 0;
    }

    public double getExpectedFDR(int mappability, int gcContent, double locFDR) {
	//Calculate the expected FDR as the area under the curve with
	//respect to the local FDR
	int mapInd = getMappabilityIndex(mappability), gcInd = getGCContentIndex(gcContent);
	double fdr = 0;
	Iterator it = localRatio[mapInd][gcInd].keySet().iterator();
	while (it.hasNext()) {
	    double r = ((Double)localRatio[mapInd][gcInd].get((Integer)it.next())).doubleValue();
	    if (r<locFDR)
		fdr += r;
	}
	return fdr;
    }

    public double[][] getReadsPoisson() { return readPoissonRatio; }
    
    public double getReadsPoisson(int mappability, int gcContent) { return readPoissonRatio[getMappabilityIndex(mappability)][getGCContentIndex(gcContent)]; }
    
    public double[][] getInputPoisson() { return inputPoissonRatio; }

    public double getInputPoisson(int mappability, int gcContent) { return inputPoissonRatio[getMappabilityIndex(mappability)][getGCContentIndex(gcContent)]; }

    public double[][] getNullInputWeights() { return p0Input; }

    public double[][] getNonNullInputWeights() { 
	double p1[][] = new double[p0Input.length][p0Input[0].length];
	for (int i=0; i<p0Input.length; i++) { 
	    for (int j=0; j<p0Input[0].length; j++) { p1[i][j] = 1 - p0Input[i][j]; }
	}
	return p1;
    }

    public double[][] getNullWeights() { return p0; }

    public double[][] getNonNullWeights() { 
	double p1[][] = new double[p0.length][p0[0].length];
	for (int i=0; i<p0.length; i++) { 
	    for (int j=0; j<p0[0].length; j++) { p1[i][j] = 1 - p0[i][j]; }
	}
	return p1;
    }

    public int getMappabilityIndex(int mappability) {
	int mapInd = mappability/mappabilityBinSize;
	return mapInd==nMappabilityBins ? mapInd-1 : mapInd;
    }

    public int getGCContentIndex(int gcContent) {
	int gcInd = gcContent/gcBinSize;
	return gcInd==nGCBins ? gcInd-1 : gcInd;
    }
    
    public double getNullInputAverageWeight() { return p0InputAverage; }

    public double getNonNullInputAverageWeight() { return 1 - p0InputAverage; }


    public double getNonNullInputDistribution(int mappability, int gcContent, int n) {
	//Caclulate the inferred non-null distribution
	int mapInd = getMappabilityIndex(mappability), gcInd = getGCContentIndex(gcContent);
	Integer key = new Integer(n);
	if (mapInd<nMappabilityBins && gcInd<nGCBins && localRatioInput[mapInd][gcInd].containsKey(key))
	    return (1 - (Math.min(1, ((Double)localRatioInput[mapInd][gcInd].get(key)).doubleValue())))*((Double)nullDistInput[mapInd][gcInd].get(key)).doubleValue()/(1 - p0Input[mapInd][gcInd]);
	return 0;
    }

    public double getNullInputDistribution(int mappability, int gcContent, int n) {
	//Caclulate the inferred non-null distribution
	int mapInd = getMappabilityIndex(mappability), gcInd = getGCContentIndex(gcContent);
	Integer key = new Integer(n);
	if (mapInd<nMappabilityBins && gcInd<nGCBins && localRatioInput[mapInd][gcInd].containsKey(key))
	    return p0Input[mapInd][gcInd]*((Double)nullDistInput[mapInd][gcInd].get(key)).doubleValue();
	return 0;
    }

    public double getNonNullDistribution(int mappability, int gcContent, int n) {
	//Caclulate the inferred non-null distribution, this will not
	//be properly normalized
	int mapInd = getMappabilityIndex(mappability), gcInd = getGCContentIndex(gcContent);
	Integer key = new Integer(n);
	if (mapInd<nMappabilityBins && gcInd<nGCBins && localRatio[mapInd][gcInd].containsKey(key))
	    return (1 - (Math.min(1, ((Double)localRatio[mapInd][gcInd].get(key)).doubleValue())))*getEmpiricalDistribution(mappability, gcContent, n)/(1 - p0[mapInd][gcInd]);
	return 0;
    }

    public double[] getNonNullDistribution(int mappability, int gcContent, int[] n) {
	double p[] = new double[n.length], s = 0;
	for (int i=0; i<n.length; i++) {
	    p[i] = getNonNullDistribution(mappability, gcContent, n[i]);
	    p[i] = (Double.isNaN(p[i]) || Double.isInfinite(p[i])) ? 0 : p[i];
	    s += p[i];
	}
	if (s>0)
	    for (int i=0; i<n.length; i++) { p[i] /= s; } //Normalize
	return p;
    }

    public double calculateExpectedPower(int mappability, int gcContent) {
	//Use this method to estimate the proporion of non-null cases detected
	int first = ((Integer)(new TreeSet(diffHist[getMappabilityIndex(mappability)][getGCContentIndex(gcContent)].keySet()).first())).intValue();
	int last = ((Integer)(new TreeSet(diffHist[getMappabilityIndex(mappability)][getGCContentIndex(gcContent)].keySet()).last())).intValue();
	int nBins = last - first + 1;
	int n[] = new int[nBins];
	//double f1[] = new double[nBins];
	for (int i=first; i<=last; i++)
	    n[i-first] = i; 
	double f1[] = getNonNullDistribution(mappability, gcContent, n);
	double efdr = 0;
	//System.out.println("f1 = " + Arrays.toString(f1));
	for (int i=0; i<nBins; i++) { 
	    if (!Double.isNaN(f1[i]) && !Double.isInfinite(f1[i]) && !Double.isNaN(getFDRThreshold(mappability, gcContent, i+first, 0)) && !Double.isInfinite(getFDRThreshold(mappability, gcContent, i+first, 0)))
		efdr += f1[i]*getFDRThreshold(mappability, gcContent, i+first, 0); 
	}
	//System.out.println(mappability + "\t" + gcContent + "\t" + first + "\t" + last + "\t" + efdr);
	return efdr;
    }

    public double calculateExpectedPower() {
	//Use this method to estimate the proporion of non-null cases
	//detected, averaging over all bins
	double efdr = 0;
	for (int i=0; i<nMappabilityBins; i++) {
	    for (int j=0; j<nGCBins; j++) {
		//System.out.println(nInputWindows[i][j]*calculateExpectedPower(i*mappabilityBinSize + 1, j*gcBinSize + 1));
		efdr += nInputWindows[i][j]*calculateExpectedPower(i*mappabilityBinSize + 1, j*gcBinSize + 1);
	    }
	}
	return efdr/totWins;
    }

    public double calculateStdDevPower(int mappability, int gcContent) {
	//Use this method to estimate the proporion of non-null cases detected
	int first = ((Integer)(new TreeSet(diffHist[getMappabilityIndex(mappability)][getGCContentIndex(gcContent)].keySet()).first())).intValue(), last = ((Integer)(new TreeSet(diffHist[getMappabilityIndex(mappability)][getGCContentIndex(gcContent)].keySet()).last())).intValue();
	int nBins = last - first + 1;
	int n[] = new int[nBins];
	for (int i=first; i<=last; i++) { n[i-first] = i; }
	double f1[] = getNonNullDistribution(mappability, gcContent, n), sd1 = 0;
	for (int i=0; i<nBins; i++) { sd1 += f1[i]*getFDRThreshold(mappability, gcContent, i+first, 0)*getFDRThreshold(mappability, gcContent, i+first, 0); }
	return Math.sqrt(sd1 - calculateExpectedPower(mappability, gcContent)*calculateExpectedPower(mappability, gcContent));
    }

    public double calculateStdDevPower() {
	//Use this method to estimate the proporion of non-null cases
	//detected, averaging over all bins
	double sd1 = 0;
	for (int i=0; i<nMappabilityBins; i++) {
	    for (int j=0; j<nGCBins; j++) {
		sd1 += nInputWindows[i][j]*calculateStdDevPower(i*mappabilityBinSize + 1, j*gcBinSize + 1);
	    }
	}
	return sd1/totWins;
    }

    public double getNullDistribution(int mappability, int gcContent, int n) {
	//Caclulate the inferred non-null distribution
	int mapInd = getMappabilityIndex(mappability), gcInd = getGCContentIndex(gcContent);
	Integer key = new Integer(n);
	if (mapInd<nMappabilityBins && gcInd<nGCBins && localRatio[mapInd][gcInd].containsKey(key))
	    return p0[mapInd][gcInd]*((Double)nullDist[mapInd][gcInd].get(key)).doubleValue();
	return 0;
    }

    public double getEmpiricalDistribution(int mappability, int gcContent, int n) {
	int mapInd = getMappabilityIndex(mappability), gcInd = getGCContentIndex(gcContent);
	Integer key = new Integer(n);
	if (mapInd<nMappabilityBins && gcInd<nGCBins && diffHist[mapInd][gcInd].containsKey(key))
	    return ((Double)diffHist[mapInd][gcInd].get(key)).doubleValue();
	return 0;
    }

    public double calculateNonNullDetectionProportion(int mappability, int gcContent, double fdr) {
	//Calculate the proportion of non-null cases which we are able
	//to detect for a given threshold. For specific gc and
	//mappability values.
	double g = 0;
	int first = ((Integer)(new TreeSet(diffHist[getMappabilityIndex(mappability)][getGCContentIndex(gcContent)].keySet()).first())).intValue(), last = ((Integer)(new TreeSet(diffHist[getMappabilityIndex(mappability)][getGCContentIndex(gcContent)].keySet()).last())).intValue();
	int nBins = last - first + 1;
	int n[] = new int[nBins];
	for (int i=first; i<=last; i++) { n[i-first] = i; }
	double f1[] = getNonNullDistribution(mappability, gcContent, n);
	for (int i=0; i<nBins; i++) { 
	    if (getFDRThreshold(mappability, gcContent, i+first, 0)<fdr) {
		//System.out.println((i+first) + "\t" + getFDRThreshold(mappability, gcContent, i+first, 0) + "\t" + f1[i]);
		g += f1[i];
	    }
	    
	}
	return g;
    }
	
    public double calculateNonNullDetectionProportion(double fdr) {
	double g = 0;
	for (int i=0; i<nMappabilityBins; i++) {
	    for (int j=0; j<nGCBins; j++) {
		g += (double)nInputWindows[i][j]*calculateNonNullDetectionProportion(i*mappabilityBinSize+1, j*gcBinSize+1, fdr);
	    }
	}
	return g/totWins;
    }

    public double[] calculateNonNullDetectionProportion(double fdrs[]) {
	double g[] = new double[fdrs.length];
	for (int i=0; i<fdrs.length; i++) { g[i] = calculateNonNullDetectionProportion(fdrs[i]); }
	return g;
    }

}
