//Use this method to read a Fasta file and extract the sequences
//corresponding to probe locations.


import java.util.*;
import java.io.*;

public class FastaSequence {

    protected String fileName;
    protected int chr = -1;
    protected Vector seqs = null, starts = null, inds = null;

    public FastaSequence(String f){
	fileName = f;
    }

    public long getChromosomeLength(){
	long len = 0;
	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    while (line!=null) {
		len += (long)line.length();
		line = br.readLine();
	    }
	    br.close();
	}
	catch (Exception e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	}
	return len;
    }

    public String[] getContiguousSequences(long pos[], int probeLength){
	//Use this method to extract the non-overlapping sequences
	//corresponding to the given tiling array.
	seqs = new Vector();
	starts = new Vector();
	inds = new Vector();
	int posIndex = 0;
	long genomeIndex = 0, rem = 0, genomePartLength = 0;	
	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //Do not use the fancy BioJava stuff. Instead, we just
	    //read the data since it is more or less just a flat
	    //file. This version assumes that there is only one
	    //sequence and header in the entire file. This will be a
	    //little bit tricky since the pos argument are based on an
	    //index starting at 1 while the strings used in Java are
	    //indexed from 0.
	    String line = br.readLine(), genomePart = "";
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    int segStartIndex = 0;
	    int segStopIndex = findContiguousRegion(pos, segStartIndex, probeLength);
	    while (segStopIndex<pos.length) {
		line = br.readLine();
		int lineLength = line.length();
		while (line!=null && genomeIndex + lineLength<=pos[segStartIndex]-1) {
		    genomePart = line;
		    genomeIndex += genomePart.length();
		    line = br.readLine();
		}
		while (line!=null && genomeIndex + genomePartLength<pos[segStopIndex] + probeLength - 1 && genomeIndex<=pos[segStartIndex]-1) {
		    lineLength = line.length();
		    genomePart += line;
		    genomePartLength = genomePart.length();
		    line = br.readLine();
		}
		//Read enough to be able to extract the sequence
		//for probe starting at pos[posIndex]
		seqs.addElement(genomePart.substring((int)(pos[segStartIndex] - genomeIndex - 1), (int)(pos[segStopIndex] - genomeIndex + probeLength - 1)));
		starts.addElement(new Long(pos[segStartIndex]));
		inds.addElement(new Integer(segStartIndex));
		segStartIndex = segStopIndex + 1;
		segStopIndex = findContiguousRegion(pos, segStartIndex, probeLength);
		//Remove the parts which are no longer needed from
		//the String genomePart
		if (segStartIndex<pos.length && pos[segStartIndex] - 1<genomeIndex + genomePartLength) {
		    genomePart = genomePart.substring((int)(pos[segStartIndex] - 1 - genomeIndex));
		    genomeIndex += pos[segStopIndex] - 1 - genomeIndex;
		    genomePartLength = genomePart.length();
		}
		else {
		    genomeIndex += genomePartLength;
		    genomePart = new String("");
		    genomePartLength = 0;
		}
	    }
	    br.close();
	}
	catch (IOException e) {

	}
	String[] s = new String[seqs.size()];
	for (int i=0; i<seqs.size(); i++)
	    //make sure that all letters are capital which is required by TFMScan
	    s[i] = ((String)seqs.elementAt(i)).toUpperCase();
	return s;
    }

    public long[] getContiguousSequencesStart(){
	//Use this method to extract the starting positions of the
	//contiguous sequences.
	if (starts==null) {
	    System.err.println("Error, must use getContiguousSequences(long[] pos, int probeLength) first.");
	    return null;
	}
	else {
	    long[] s = new long[starts.size()];
	    for (int i=0; i<starts.size(); i++)
		s[i] = ((Long)starts.elementAt(i)).longValue();
	    return s;
	}
    }

    public int[] getContiguousSequencesStartIndexes(){
	//Use this method to extract the starting positions of the
	//contiguous sequences.
	if (inds==null) {
	    System.err.println("Error, must use getContiguousSequences(long[] pos, int probeLength) first.");
	    return null;
	}
	else {
	    int[] s = new int[inds.size()];
	    for (int i=0; i<inds.size(); i++)
		s[i] = ((Integer)inds.elementAt(i)).intValue();
	    return s;
	}
    }

    private int findContiguousRegion(long[] pos, int startInd, int probeLength){
	//Returns the index of the last element of pos which forms a
	//contiguous region with all indices starting from startInd.
	while (startInd+1<pos.length && pos[startInd] + (long)probeLength>pos[startInd+1])
	    startInd++;
	//if ((startInd+1)==pos.length) //Handle the last probe
	//return pos[startInd] + probeLength>pos[pos.length-1] ? pos.length - 1 : pos.length - 2;
	return startInd;
    }


    public void saveSequences(String fileName, long pos[], int probeLength){
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    String seqs[] = getSequences(pos, probeLength);
	    for (int i=0; i<pos.length; i++) {
		bw.write(">" + pos[i] + "-" + (pos[i] + probeLength - 1) + "\n");
		bw.write(seqs[i] + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {

	}
    }

    public void saveSequence(String fileName, long start, int length) {
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    bw.write(">" + start + "-" + (start + length - 1) + "\n");
	    long pos[] = new long[1];
	    pos[0] = start;
	    bw.write(getSequences(pos, length)[0] + "\n");
	    bw.close();
	}
	catch (IOException e) {

	}
    }

    public static void saveSequence(String fileName, String header, String sequence) {
	//Use this method for saving a sequence in Fasta-format
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    bw.write(">" + header + "\n");
	    bw.write(sequence + "\n");
	    bw.close();
	}
	catch (IOException e) {

	}
    }

    public String[] getSequences(long pos[], long probeLength){
	//Use this method to extract all the sequences starting at pos
	//and extending probeLength basepairs. Assume that pos is
	//sorted in ascending order.
	String seqs[] = new String[pos.length];
	int posIndex = 0;
	long genomeIndex = 0, rem = 0, genomePartLength = 0;
	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //Do not use the fancy BioJava stuff. Instead, we just
	    //read the data since it is more or less just a flat
	    //file. This version assumes that there is only one
	    //sequence and header in the entire file. This will be a
	    //little bit tricky since the pos argument are based on an
	    //index starting at 1 while the strings used in Java are
	    //indexed from 0.
	    String line = br.readLine(), genomePart = "";
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    while (line!=null && posIndex<pos.length) {
		int lineLength = line.length();
		genomePart += line;
		genomePartLength = genomePart.length();
		while (posIndex<pos.length && genomeIndex + genomePartLength>pos[posIndex] + probeLength - 1 && genomeIndex<=pos[posIndex]-1) {
		    //Read enough to be able to extract the sequence
		    //for probe starting at pos[posIndex]
		    seqs[posIndex] = genomePart.substring((int)(pos[posIndex] - genomeIndex - 1), (int)(pos[posIndex] - genomeIndex + probeLength - 1));
		    //Remove the parts which are no longer needed from
		    //the String genomePart
		    posIndex++;
		}
		if (posIndex<pos.length && pos[posIndex] - 1<genomeIndex + genomePartLength) {
		    genomePart = genomePart.substring((int)(pos[posIndex] - 1 - genomeIndex));
		    genomeIndex += pos[posIndex] - 1 - genomeIndex;
		}
		else {
		    genomeIndex += genomePartLength;
		    genomePart = new String("");
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (Exception e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	    System.err.println(posIndex + " = posIndex, " + genomeIndex + " = genomeIndex, " + genomePartLength + " = genomePartLength");
	    System.err.println(pos[posIndex] - 1 - genomeIndex);
	    //return null;
	}
	return seqs;
    }

    public String getSequence(int pos, int probeLength){
	//Use this method to extract all the sequences starting at pos
	//and extending probeLength basepairs. Assume that pos is
	//sorted in ascending order.
	int posIndex = 0;
	int genomeIndex = 0, rem = 0, genomePartLength = 0;
	String seq = "";
	try {
	    //Open the file
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    //Do not use the fancy BioJava stuff. Instead, we just
	    //read the data since it is more or less just a flat
	    //file. This version assumes that there is only one
	    //sequence and header in the entire file. This will be a
	    //little bit tricky since the pos argument are based on an
	    //index starting at 1 while the strings used in Java are
	    //indexed from 0.
	    String line = br.readLine(), genomePart = "";
	    //Discard the header
	    while (line!=null && line.substring(0, 1).equals(">"))
		line = br.readLine();
	    while (line!=null) {
		int lineLength = line.length();
		genomePart += line;
		genomePartLength = genomePart.length();
		while (genomeIndex + genomePartLength>pos + probeLength - 1 && genomeIndex<=pos-1) {
		    //Read enough to be able to extract the sequence
		    //for probe starting at pos[posIndex]
		    seq = genomePart.substring((int)(pos - genomeIndex - 1), (int)(pos - genomeIndex + probeLength - 1));
		}
		if (seq.length()>0)
		    break;
		if (pos - 1<genomeIndex + genomePartLength) {
		    genomePart = genomePart.substring((int)(pos - 1 - genomeIndex));
		    genomeIndex += pos - 1 - genomeIndex;
		}
		else {
		    genomeIndex += genomePartLength;
		    genomePart = new String("");
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (Exception e) {
	    System.err.println("Error reading file " + fileName + ": " + e.toString());
	}
	return seq;
    }

    public int[][] getMATCoefficients(long pos[], int probeLength){
	String seqs[] = getSequences(pos, probeLength);
	int[][] coeffs = new int[pos.length][probeLength*3 + 5];
	for (int i=0; i<pos.length; i++){
	    int nA = 0, nC = 0, nG = 0, nT = 0;
	    for (int j=0; j<probeLength; j++) {
		if (seqs[i].substring(j, j+1).toUpperCase().equals("A")) {
		    coeffs[i][j] = 1;
		    nA++;
		}
		else if (seqs[i].substring(j, j+1).toUpperCase().equals("C")) {
		    coeffs[i][j+probeLength] = 1;
		    nC++;
		}
		else if (seqs[i].substring(j, j+1).toUpperCase().equals("G")) {
		    coeffs[i][j+probeLength*2] = 1;
		    nG++;
		}
		else if (seqs[i].substring(j, j+1).toUpperCase().equals("T")) 
		    nT++;
	    }
	    coeffs[i][probeLength*3] = nT;
	    coeffs[i][probeLength*3+1] = nA*nA;
	    coeffs[i][probeLength*3+2] = nC*nC;
	    coeffs[i][probeLength*3+3] = nG*nG;
	    coeffs[i][probeLength*3+4] = nT*nT;
	}
	return coeffs;
    }

    public int[] scoreMotif(int[][] pfm, long start, int length) {
	return scoreMotif(pfm, start, length, .7, true);
    }

    public int[] scoreMotif(int[][] pfm, long start, int length, double threshold) {
	return scoreMotif(pfm, start, length, threshold, true);
    }

    public int[] scoreMotif(int[][] pfm, long start, int length, double threshold, boolean single) {
	//Use this method to scan for a motif between start and start
	//+ length given a representation pfm. Return an array with
	//the indices for the motif of interest. These indices are
	//with respect to the true coordinates.
	return null;
    }

}
