//Use this method to decide which regions of the genome are considered
//transcribed based on the mRNA data. The first method that we use is
//very simple, relying only on a distance cut-off. This distance is
//based on a FDR calculation from the overall statistics of all mRNA
//reads outside the exonic and intronic regions.

//Example:

//java -Xmx512m FindEnhancerSites -d 2000 -p MmDataPeakFiles/enhancer_marker_names -P MmDataPeakFiles/h3k4me1_b2_names -s MmDataEnhancers/enhancers_b2 -r 3 -c 1 -s /groups/neuroduo/martin/ChIP-Seq/MmData/Enhancers/enhancers_b2 --EST


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class FindEnhancerSites {

    public static final String ENHANCER_SUFFIX = "enhancers";
    public static final String ENHANCER_HIGH_QUALITY_SUFFIX = "_hq";
    public static final String ENHANCER_MEDIUM_QUALITY_SUFFIX = "_mq";
    public static final String ENHANCER_NULL_SUFFIX = "_null";
    public static final String P300_SUFFIX = "_p300";
    public static final String ENHANCER_NUMBERS_SUFFIX = "_nEnhancers";
    public static final String ENHANCER_NO_CBP_SUFFIX = "_noCBP";
    public static final String TRANSCRIBED_REGION_LENGTH_SUFFIX = "_transcribedRegionLengths";
    public static final String TRANSCRIBED_REGION_NUMBER_SUFFIX = "_transcribedRegionNumbers";
    public static final String ENHANCER_EST_FILE_SUFFIX = "ESTs";

    public static final String CBP_KCL_FILTER_STR = "CBP_Millipore_KCl";
    public static final String CBP_KCL_B2_FILTER_STR = "CBP_Abcam_KCl";
    public static final String REPRODUCIBLE_CBP_KCL_FILTER_STR = "CBP_KCl_rep";
    public static final String CBP_UN_FILTER_STR = "CBP_Abcam_un";
    public static final String CBP_MP_UN_FILTER_STR = "CBP_Millipore_un";
    public static final String P300_UN_B1_FILTER_STR = "p300_un_B1";
    public static final String P300_UN_B2_FILTER_STR = "p300_un_B2";
    public static final String P300_KCL_B1_FILTER_STR = "p300_KCl_B1";
    public static final String P300_KCL_B2_FILTER_STR = "p300_KCl_B2";
    public static final String CBP_FILTER_STR = "CBP";
    public static final String CONSERVATION_FILTER_STR = "Cons";
    public static final String POL2_KCL_FILTER_STR = "Pol2_KCl";
    public static final String POL2_UN_FILTER_STR = "Pol2_un";
    public static final String POL2_FILTER_STR = "Pol2";
    public static final String RNA_SS_FILTER_STR = "RNA_Ss";
    public static final String RNA_BI_FILTER_STR = "RNA_Bi";
    public static final String DISTAL_FILTER_STR = "Distal";
    public static final String INTRONIC_FILTER_STR = "Intronic";
    public static final String PROXIMAL_FILTER_STR = "Proximal";
    public static final String H3K4ME1_FILTER_STR = "H3K4Me1";
    public static final String H3K4ME3_FILTER_STR = "H3K4Me3";
    public static final String H3K4ME1_KCL_FILTER_STR = "H3K4Me1_KCl";
    public static final String H3K4ME3_KCL_FILTER_STR = "H3K4Me3_KCl";
    public static final String H3K4ME1_UN_FILTER_STR = "H3K4Me1_un";
    public static final String H3K4ME3_UN_FILTER_STR = "H3K4Me3_un";

    public static final int LONG_FACTOR = 4;
    public static final int PROXIMAL_PROMOTER_CUTOFF = 10000;
    public static final int ENHANCER_SAVE_WIDTH = 500;
    public static final int MIN_TSS_DISTANCE = 2000;
    public static final int EST_ENHANCER_DISTANCE = 2000;

    public static final String H3K4ME1_STR = "H3K4Me1";
    public static final String H3K4ME3_STR = "H3K4Me3";

    public static final String H3K4ME1_KCL_STR = "H3K4Me1_KCl";
    public static final String H3K4ME3_KCL_STR = "H3K4Me3_KCl";
    public static final String H3K4ME1_UN_STR = "H3K4Me1_un";
    public static final String H3K4ME3_UN_STR = "H3K4Me3_un";

    public static Vector<String> filterStrs = new Vector<String>();

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //String[] tmpStr = new String[]{CBP_FILTER_STR, CBP_UN_FILTER_STR, POL2_FILTER_STR, POL2_UN_FILTER_STR, POL2_KCL_FILTER_STR, H3K4ME1_FILTER_STR, H3K4ME3_FILTER_STR, H3K4ME1_KCL_FILTER_STR, H3K4ME3_KCL_FILTER_STR, H3K4ME1_UN_FILTER_STR, H3K4ME3_UN_FILTER_STR, CONSERVATION_FILTER_STR, RNA_SS_FILTER_STR, RNA_BI_FILTER_STR, PROXIMAL_FILTER_STR, DISTAL_FILTER_STR, INTRONIC_FILTER_STR};
	    String[] tmpStr = new String[]{POL2_KCL_FILTER_STR, H3K4ME1_KCL_FILTER_STR, CONSERVATION_FILTER_STR, RNA_SS_FILTER_STR, RNA_BI_FILTER_STR, PROXIMAL_FILTER_STR, DISTAL_FILTER_STR, INTRONIC_FILTER_STR};
	    filterStrs = new Vector(Arrays.asList(tmpStr));
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    //File containing a list of all the mRNA data sets that we
	    //want to investigate. This needs to correspond to the
	    //file which was used for generating the null peaks
	    CmdLineParser.Option mRNAFileNameFlag = argParser.addStringOption('m', "mRNAFileName");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option replicateFlag = argParser.addBooleanOption('b', "replicate");
	    CmdLineParser.Option reproducibleFlag = argParser.addBooleanOption("reproducible");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption('f', "fdr");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    //Decide what type of method we are going to use
	    CmdLineParser.Option typeFlag = argParser.addStringOption('t', "type");
	    //Restrict the length of the transcripts which will be
	    //candidates for the overlap regions
	    CmdLineParser.Option maxEnhancerTranscriptLengthFlag = argParser.addIntegerOption('l', "maxEnhancerTranscriptLength");
	    //The location of the peak file containing Pol2 and CBP peaks
	    CmdLineParser.Option peakFileFlag = argParser.addStringOption('p', "peakFileName");
	    //The location of the H3K4Me1 marks
	    CmdLineParser.Option h3K4Me1PeakFileFlag = argParser.addStringOption('P', "H3K4Me1PeakFileName");
	    //The max distance to the next read in the 3' direction
	    CmdLineParser.Option maxReadDistanceFlag = argParser.addIntegerOption('d', "maxReadDistance");
	    //The minimum number of reads in the transcript
	    CmdLineParser.Option minReadsFlag = argParser.addIntegerOption('r', "minReads");
	    //Compare with EST-tags
	    CmdLineParser.Option estFileNameFlag = argParser.addBooleanOption("EST");
	    //Compare with CAGE tags
	    CmdLineParser.Option cageFileNameFlag = argParser.addBooleanOption("CAGE");
	    //Use only the cortex CAGE tags
	    CmdLineParser.Option cortexFlag = argParser.addBooleanOption('x', "cortex");
	    //Use only the CAGE tags which have the polyA mark
	    CmdLineParser.Option polyAFlag = argParser.addBooleanOption('p', "polyA");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean replicate = argParser.getOptionValue(replicateFlag)!=null;
	    boolean reproducible = argParser.getOptionValue(reproducibleFlag)!=null;
	    if (replicate)
		filterStrs.setElementAt(CBP_KCL_B2_FILTER_STR, 1);
	    else if (reproducible)
		filterStrs.setElementAt(REPRODUCIBLE_CBP_KCL_FILTER_STR, 1);
	    else
		filterStrs.setElementAt(CBP_KCL_FILTER_STR, 1);
	    Collections.sort(filterStrs);
	    tmp = argParser.getOptionValue(mRNAFileNameFlag);
	    String mRNAFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/hr";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "Enhancers/" + ENHANCER_SUFFIX + (replicate ? "_b2" : "") + (reproducible ? "_rep" : "");
	    tmp = argParser.getOptionValue(peakFileFlag);
	    String peakFileName = tmp!=null ? (String)tmp : "PeakFiles/enhancer_markers_" + (replicate ? "b2_" : "") + (reproducible ? "rep_" : "") + "names";
	    tmp = argParser.getOptionValue(h3K4Me1PeakFileFlag);
	    String h3K4Me1PeakFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(typeFlag);
	    String type = tmp!=null ? (String)tmp : "all";
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(maxEnhancerTranscriptLengthFlag);
	    int maxEnhancerTranscriptLength = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(minReadsFlag);
	    int minReads = tmp!=null ? ((Integer)tmp).intValue() : 2;
	    tmp = argParser.getOptionValue(maxReadDistanceFlag);
	    int maxReadDistance = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    //EST and CAGE stuff
	    boolean cortex = argParser.getOptionValue(cortexFlag)!=null;
	    boolean polyA = argParser.getOptionValue(polyAFlag)!=null;
	    String cageFileName = argParser.getOptionValue(cageFileNameFlag)!=null ? (polyA ? "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/CAGE/mm9/" : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "deepCAGE/mm9/") : null;
	    String estFileName = argParser.getOptionValue(estFileNameFlag)!=null ? (orchestra ? "/groups/neuroduo/jesse/SOLiD/deepCAGE_and_EST/ESTs/" : null) : null;
	    //////////////////////////////////////////////////////////////////
	    //Get the peak info
	    TreeMap<Integer, Peak> peaks = parsePeakFile(peakFileName, chr, false);
	    TreeSet<String> peakNames = new TreeSet<String>(parsePeakFileHeader(peakFileName, chr).values());
	    System.out.println("Found " + peaks.size() + " combo peaks with " + peakNames.size() + " factors from " + peakFileName);
	    //Get the K4Me1 marks, these are typically obtained using
	    //different parameters for the peakfinder.
	    TreeMap<Integer, Peak> h3K4Me1Peaks = parsePeakFile(h3K4Me1PeakFileName, chr, false);
	    TreeSet<String> h3K4PeakNames = new TreeSet<String>(parsePeakFileHeader(h3K4Me1PeakFileName, chr).values());
	    System.out.println("Found " + h3K4Me1Peaks.size() + " H3K4Me1 peaks with " + h3K4PeakNames.size() + " factors for chromosome.");
	    /////////////////////////////////
	    //Get the annotation info
	    int nChr = 21;
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || chr==i) 
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra);
	    }
	    /////////////////////////////////
	    //Keep track of the number of enhancers that will be
	    //extracted depending on what combination of criteria are
	    //used for selection. The candidates are:
	    //
	    //1. CPB peak
	    //2. Conservation
	    //3. Pol2 peak
	    //4. Bidirectional transcription
	    //5. More than 10kb away from gene
	    int nEnhancers = 0, nEnhancerWithRNA = 0, nEnhancerWithPol2 = 0, nEnhancerWithPol2AndRNA = 0, meanEnhancerDistance = 0, nConsCBP = 0;
	    //Keep track of the distance to the nearest TSS for each
	    //putative enhancer site
	    TreeMap<Integer, Integer> enhancerTSSDistance = new TreeMap<Integer, Integer>();
	    //Keep track of the number of enhancers per gene
	    TreeMap<RefSeqGene, Integer> geneEnhancers = new TreeMap<RefSeqGene, Integer>();
	    TreeMap<Peak, Enhancer> enhancers = new TreeMap<Peak, Enhancer>();
	    LinkedHashMap<Filter, TreeSet<Enhancer>> filter2enhancersMap = new LinkedHashMap<Filter, TreeSet<Enhancer>>();
	    //The enhancers that did not have CBP mark
	    TreeSet<Enhancer> enhancersNoCBP = new TreeSet<Enhancer>();
	    Vector<Integer> shortestLength = new Vector<Integer>();
	    int timePoints[] = new int[]{0, 1, 6};
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || chr==i) {
		    Vector<ABIRead[]> forwardReads = new Vector<ABIRead[]>(3);
		    Vector<ABIRead[]> reverseReads = new Vector<ABIRead[]>(3);
		    for (int j=0; j<timePoints.length; j++) {
			//Get the the mRNA info
			forwardReads.addElement(null);
			reverseReads.addElement(null);
			forwardReads.setElementAt(FindNongenicTranscripts.loadMRNAReadsChr(mRNAFileName + timePoints[j] + "c.tsv", true, i), j);
			reverseReads.setElementAt(FindNongenicTranscripts.loadMRNAReadsChr(mRNAFileName + timePoints[j] + "c.tsv", false, i), j);
			System.out.println("Loaded reads from t = " + timePoints[j] + " for chr " + i);
			///////////////////////////////////////////////////////
			//Find the length distribution of the
			//transcribed regions at each putative
			//enhancer location
			shortestLength.addAll(calculateRNALengthDistribution(i, peaks, h3K4Me1Peaks, geneInfo, forwardReads.lastElement(), reverseReads.lastElement(), minReads, maxReadDistance));
		    }
		    filter2enhancersMap = generateFilterCombos(i, peaks, h3K4Me1Peaks, geneInfo, forwardReads, reverseReads, minReads, maxReadDistance, estFileName, saveFileName, filter2enhancersMap);
		    //enhancersNoCBP = getEnhancersNoCBP(i, peaks, h3K4Me1Peaks, geneInfo, forwardReads, reverseReads, minReads, maxReadDistance, enhancersNoCBP);
		}
	    }
	    //////////////////////////////////////////////////////////////////
	    //Save the numbers to a file
	    String saveName = saveFileName + ENHANCER_NUMBERS_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt";
	    System.out.println("Saving enhancer sites to " + saveName); 
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	    bw.write("#");
	    for (String f : filterStrs)
		bw.write(f + "\t");
	    bw.write("nEnhancers\n");
	    for (Filter filters : filter2enhancersMap.keySet()) {
		bw.write(filters.saveString(filterStrs) + filter2enhancersMap.get(filters).size() + "\n");
	    }
	    bw.close();
	    ///////////////////////////////////////////////////////
	    //Pool all the enhancers into one long list
	    TreeMap<Enhancer, Filter> nullEnhancers = new TreeMap<Enhancer, Filter>(), ncEnhancers = new TreeMap<Enhancer, Filter>(), remainingEnhancers = new TreeMap<Enhancer, Filter>(), rnaEnhancers = new TreeMap<Enhancer, Filter>();
	    TreeMap<Enhancer, Filter> p300B1Enhancers = new TreeMap<Enhancer, Filter>(), p300B2Enhancers = new TreeMap<Enhancer, Filter>();
	    TreeMap<Enhancer, Filter> p300Enhancers = new TreeMap<Enhancer, Filter>();
	    //Store enahncer as key and strictest filter as value
	    boolean kclMe1 = h3K4Me1Peaks.get(h3K4Me1Peaks.firstKey()).hasFactor("KCl");
	    if (peaks.get(peaks.firstKey()).hasFactor("CBP")) {
		for (Filter filters : filter2enhancersMap.keySet()) {
		    for (Enhancer en : filter2enhancersMap.get(filters)) {
			if (filters.nullFilter(kclMe1) && !nullEnhancers.containsKey(en))
			    nullEnhancers.put(en, filters);
		    }
		}
		for (Filter filters : filter2enhancersMap.keySet()) {
		    for (Enhancer en : filter2enhancersMap.get(filters)) {
			if (filters.nullFilter(kclMe1, false) && !nullEnhancers.containsKey(en) && !ncEnhancers.containsKey(en))
			    ncEnhancers.put(en, filters);
		    }
		}
		for (Filter filters : filter2enhancersMap.keySet()) {
		    for (Enhancer en : filter2enhancersMap.get(filters)) {
			if (filters.hasRequired(kclMe1) && !nullEnhancers.containsKey(en) && !ncEnhancers.containsKey(en) && !remainingEnhancers.containsKey(en))
			    remainingEnhancers.put(en, filters);
		    }
		}
		for (Filter filters : filter2enhancersMap.keySet()) {
		    for (Enhancer en : filter2enhancersMap.get(filters)) {
			if (filters.hasRequired(kclMe1) && (filters.contains(RNA_BI_FILTER_STR) || filters.contains(RNA_SS_FILTER_STR)) && !rnaEnhancers.containsKey(en))
			    rnaEnhancers.put(en, filters);
		    }
		}
		/////////////////////////////////
		saveEnhancers(saveFileName + ENHANCER_NULL_SUFFIX + (kclMe1 ? "": "_un") + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, nullEnhancers);
		saveEnhancers(saveFileName + ENHANCER_NULL_SUFFIX + (kclMe1 ? "": "_un") + "_nc" + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, ncEnhancers);
		saveEnhancers(saveFileName + ENHANCER_NULL_SUFFIX + (kclMe1 ? "": "_un") + "_noMe1" + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, remainingEnhancers);
		saveEnhancers(saveFileName + ENHANCER_NULL_SUFFIX + (kclMe1 ? "": "_un") + "_rna" + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, rnaEnhancers);
	    }
	    else { //p300 enhancers
		for (Filter filters : filter2enhancersMap.keySet()) {
		    for (Enhancer en : filter2enhancersMap.get(filters)) {
			if (filters.p300Filter(kclMe1) && !p300Enhancers.containsKey(en))
			    p300Enhancers.put(en, filters);
		    }
		}
		/*for (Filter filters : filter2enhancersMap.keySet()) {
		    for (Enhancer en : filter2enhancersMap.get(filters)) {
			if (filters.p300Filter(kclMe1, false) && !p300Enhancers.containsKey(en) && !p300ncEnhancers.containsKey(en))
			    p300ncEnhancers.put(en, filters);
		    }
		    }*/
		/////////////////////////////////
		saveEnhancers(saveFileName + P300_SUFFIX + (kclMe1 ? "": "_un") + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, p300Enhancers);
		//saveEnhancers(saveFileName + P300_SUFFIX + (kclMe1 ? "": "_un") + "_nc" + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, p300ncEnhancers);
	    }
	    //////////////////////////////////////////////////////////////////
	    //Find the number of enhancers within a given distance of each TSS
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || chr==i) {

		}
	    }
	    ///////////////////////////////////////////////////////
	    //Save the distribution of shortest transcribed regions 
	    saveName = saveFileName + TRANSCRIBED_REGION_LENGTH_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt";
	    bw = new BufferedWriter(new FileWriter(saveName));
	    System.out.println("Saving lengths to " + saveName);
	    for (Integer i : shortestLength)
		bw.write(i + "\n");
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static void saveEnhancerTranscriptLengths(String saveName, TreeMap<Enhancer, Filter> enhancers) throws IOException {
	System.out.println("Saving to " + saveName);

    }

    public static void saveEnhancers(String saveName, TreeSet<String> peakNames, TreeMap<Enhancer, Filter> enhancers) throws IOException {
	System.out.println("Saving " + enhancers.size() + " enhancer sites to " + saveName);
	BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	bw.write("#Loc\tnForward\tnReverse\tH3K4Me1_KCl");
	for (String s : peakNames)
	    bw.write("\t" + s);
	for (String f : filterStrs)
	    bw.write("\t" + f);
	bw.write("\n");
	for (Enhancer en : enhancers.keySet()) {
	    Peak p = en.getPeak();
	    Filter f = enhancers.get(en);
	    bw.write("chr" + AnalyzeControls.getChromosomeNumber(p.getChromosome()) + ":" + (p.getCenter() - ENHANCER_SAVE_WIDTH) + "-" + (p.getCenter() + ENHANCER_SAVE_WIDTH) + "\t" + en.numberOfReadsForward() + "\t" + en.numberOfReadsReverse() + "\t" + format.format(en.getH3K4Me1(false)) + "\t" + format.format(en.getH3K4Me1(true)));
	    //save information about the peaks that we found here
	    for (String s : peakNames)
		bw.write("\t" + format.format(p.getFactor(s)));
	    //Save information about which filters it satisfied
	    bw.write("\t" + f.saveString(filterStrs));
	    bw.write("\n");
	}
	bw.close();
	//Create a .bed file
	bw = new BufferedWriter(new FileWriter(saveName.replace("txt", "bed")));
	bw.write("browser position chr1:21315319-29327381\n");
	bw.write("track type=wiggle_0 name=\"" + saveName.substring(saveName.lastIndexOf("/") + 1, saveName.lastIndexOf(".")) + "\" description=\"" + saveName.substring(saveName.lastIndexOf("/") + 1, saveName.lastIndexOf(".")) + "\" visibility=dense autoScale=off viewLimits=0:1 color=\"255,0,0\" maxHeightPixels=70:40:10 priority=1\n");
	int prevEnd = Integer.MIN_VALUE;
	for (Enhancer en : enhancers.keySet()) {
	    Peak p = en.getPeak();
	    bw.write("chr" + AnalyzeControls.getChromosomeNumber(p.getChromosome()) + "\t" + (Math.max(prevEnd, p.getCenter() - ENHANCER_SAVE_WIDTH)) + "\t" + (p.getCenter() + ENHANCER_SAVE_WIDTH) + "\t1\n");
	    prevEnd = p.getCenter() + ENHANCER_SAVE_WIDTH;
	}
	bw.close();
    }

    public static void saveEnhancers(String saveName, TreeSet<String> peakNames, TreeSet<Enhancer> enhancers) throws IOException {
	//For the ones without CBP
	System.out.println("Saving " + enhancers.size() + " enhancer sites to " + saveName);
	BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	bw.write("#Loc\tnForward\tnReverse\tH3K4Me1_KCl");
	for (String s : peakNames)
	    bw.write("\t" + s);
	bw.write("\n");
	for (Enhancer en : enhancers) {
	    Peak p = en.getPeak();
	    bw.write("chr" + AnalyzeControls.getChromosomeNumber(p.getChromosome()) + ":" + (p.getCenter() - ENHANCER_SAVE_WIDTH) + "-" + (p.getCenter() + ENHANCER_SAVE_WIDTH) + "\t" + en.numberOfReadsForward() + "\t" + en.numberOfReadsReverse() + "\t" + format.format(en.getH3K4Me1(true)) + "\t" + format.format(en.getH3K4Me1(false)));
	    //save information about the peaks that we found here
	    for (String s : peakNames)
		bw.write("\t" + format.format(p.getFactor(s)));
	    bw.write("\n");
	}
	bw.close();

    }

    //////////////////////////////////////////////////////////////////

    public static LinkedHashMap<Filter, TreeSet<Enhancer>> generateFilterCombos(int chr, TreeMap<Integer, Peak> peaks, TreeMap<Integer, Peak> h3K4Me1Peaks, RefSeqGenes[] geneInfo, Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, int minReads, int maxReadDistance, String estFileName, String saveFileName, LinkedHashMap<Filter, TreeSet<Enhancer>> map) {
	LinkedHashSet<Filter> filters = map.size()>0 ? new LinkedHashSet<Filter>(map.keySet()) : getAllFilters(peaks.get(peaks.firstKey()).hasFactor("CBP"));
	//for (Filter filter : filters) {
	return getEnhancers(chr, peaks, h3K4Me1Peaks, geneInfo, forwardReads, reverseReads, filters, minReads, maxReadDistance, estFileName, saveFileName, map);
	    /*if (map.containsKey(filter))
		map.get(filter).addAll(e);
	    else
		map.put(filter, e);
	}
	return map;*/
    }

    public static LinkedHashSet<Filter> getAllFilters(boolean cbp) {
	LinkedHashSet<Filter> filters = new LinkedHashSet<Filter>();
	if (cbp) {
	    filters.add(new Filter(new String[]{CBP_KCL_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_KCL_STR}));
	    filters.add(new Filter(new String[]{CBP_KCL_FILTER_STR, H3K4ME1_KCL_STR}));
	    filters.add(new Filter(new String[]{CBP_KCL_FILTER_STR}));
	    filters.add(new Filter(new String[]{CBP_KCL_FILTER_STR, H3K4ME1_KCL_STR, RNA_BI_FILTER_STR}));
	    filters.add(new Filter(new String[]{CBP_KCL_FILTER_STR, H3K4ME1_KCL_STR}));
	    filters.add(new Filter(new String[]{CBP_KCL_FILTER_STR}));
	    //filters.add(new Filter(new String[]{CBP_MP_UN_FILTER_STR, H3K4ME1_UN_STR}));
	    //filters.add(new Filter(new String[]{CBP_UN_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_UN_STR}));
	    //filters.add(new Filter(new String[]{CBP_UN_FILTER_STR, H3K4ME1_UN_STR}));
	}
	else { //p300
	    filters.add(new Filter(new String[]{P300_KCL_B1_FILTER_STR, P300_KCL_B2_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_KCL_STR}));
	    filters.add(new Filter(new String[]{P300_UN_B1_FILTER_STR, P300_UN_B2_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_UN_STR}));
	    filters.add(new Filter(new String[]{P300_KCL_B1_FILTER_STR, P300_KCL_B2_FILTER_STR, H3K4ME1_KCL_STR}));
	    filters.add(new Filter(new String[]{P300_UN_B1_FILTER_STR, P300_UN_B2_FILTER_STR, H3K4ME1_UN_STR}));
	    filters.add(new Filter(new String[]{P300_KCL_B1_FILTER_STR, P300_KCL_B2_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_KCL_STR}));
	    filters.add(new Filter(new String[]{P300_UN_B1_FILTER_STR, P300_UN_B2_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_UN_STR}));
	    filters.add(new Filter(new String[]{P300_KCL_B1_FILTER_STR, P300_KCL_B2_FILTER_STR, H3K4ME1_KCL_STR}));
	    filters.add(new Filter(new String[]{P300_UN_B1_FILTER_STR, P300_UN_B2_FILTER_STR, H3K4ME1_UN_STR}));
	    //filters.add(new Filter(new String[]{P300_KCL_B1_FILTER_STR, P300_UN_B1_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_KCL_STR}));
	    //filters.add(new Filter(new String[]{P300_UN_B2_FILTER_STR, P300_KCL_B2_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_KCL_STR}));
	    //filters.add(new Filter(new String[]{P300_KCL_B1_FILTER_STR, P300_UN_B1_FILTER_STR, H3K4ME1_KCL_STR}));
	    //filters.add(new Filter(new String[]{P300_UN_B2_FILTER_STR, P300_KCL_B2_FILTER_STR, H3K4ME1_KCL_STR}));
	}
	//filters.add(new Filter(new String[]{CBP_KCL_B2_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_KCL_STR}));
	//filters.add(new Filter(new String[]{CBP_KCL_B2_FILTER_STR, H3K4ME1_KCL_STR}));
	//filters.add(new Filter(new String[]{CBP_KCL_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_KCL_STR, PROXIMAL_FILTER_STR}));
	//filters.add(new Filter(new String[]{CBP_KCL_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_KCL_STR, INTRONIC_FILTER_STR}));
	//filters.add(new Filter(new String[]{CBP_KCL_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_KCL_STR, DISTAL_FILTER_STR}));
	/*for (String f : filterStrs) {
	    Filter filter = new Filter(f);
	    if (filter.hasRequired()) {
		filters.add(filter);
		//Recursively create the rest of the filters
		filters.addAll(getAllFilters(filter, filters));
	    }
	    }*/
	System.out.println("Found " + filters.size() + " filters.");
	return filters;
    }

    public static LinkedHashSet<Filter> getAllFilters(Filter filter, LinkedHashSet<Filter> filters) {
	for (String f : filterStrs) {
	    if (filter.filterOK(f)) {
		Filter newFilter = new Filter(filter, f);
		if (!filters.contains(newFilter) && newFilter.hasRequired()) {
		    filters.add(newFilter);
		    //System.out.println("Added filter " + FILTER_STRS[i] + " to " + newFilter.toString());
		    filters.addAll(getAllFilters(newFilter, filters));
		}
	    }
	}
	return filters;
    }
    
    /////////////////////////////////

    public static Vector<Integer> calculateRNALengthDistribution(int chr, TreeMap<Integer, Peak> peaks, TreeMap<Integer, Peak> h3K4Me1Peaks, RefSeqGenes[] geneInfo, ABIRead[] forwardReads, ABIRead[] reverseReads, int minReads, int maxReadDistance) {
	//Use this method to calculate the distribution of the
	//transcribed regions for all the putative enhancer sites.
	Vector<Integer> shortestLengths= new Vector<Integer>();
	for (Integer center : peaks.keySet()) {
	    Peak p = peaks.get(center);
	    int start = center - maxReadDistance*LONG_FACTOR, end = center + maxReadDistance*LONG_FACTOR, c = p.getChromosome();
	    if (c==chr || chr==-1) {
		//Check for conservation and nearby K4Me1
		Peak h3K4Me1Peak = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME1_STR, maxReadDistance, c);
		if (p.isConserved() && h3K4Me1Peak!=null && h3K4Me1Peak.getFactor(H3K4ME1_STR)>0) {
		    RefSeqGene overlappingGene = geneInfo[c - 1].getOverlappingGene(center);
		    if (overlappingGene==null && !geneInfo[c - 1].overlapsTSS(p.getCenter() - MIN_TSS_DISTANCE, p.getCenter() + MIN_TSS_DISTANCE)) {
			//Find the lengths of the transcribed regions
			int nReadsF = 0, nReadsR = 0;
			ABIRead readsForward[] = FindReadsAcrossFeature.getReads(c, start, end, forwardReads);
			int transcribedRegionForward[] = null;//IsolateBidirectionalTranscription.findTranscribedRegionBoundaries(readsForward, true, maxReadDistance);
			ABIRead readsReverse[] = FindReadsAcrossFeature.getReads(c, start, end, reverseReads);
			int transcribedRegionReverse[] = null;//IsolateBidirectionalTranscription.findTranscribedRegionBoundaries(readsReverse, false, maxReadDistance);
			shortestLengths.addElement(Math.max(Math.abs(transcribedRegionForward[0] - transcribedRegionForward[1]), Math.abs(transcribedRegionReverse[0] - transcribedRegionReverse[1])));
		    }
		}
	    }
	}
	return shortestLengths;
    }

    public static LinkedHashMap<Filter, TreeSet<Enhancer>> getEnhancers(int chr, TreeMap<Integer, Peak> peaks, TreeMap<Integer, Peak> h3K4Me1Peaks, RefSeqGenes[] geneInfo, Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, LinkedHashSet<Filter> filters, int minReads, int maxReadDistance, String estFileName, String saveFileName, LinkedHashMap<Filter, TreeSet<Enhancer>> enhancers) {
	int nNearTSS = 0, nMe1 = 0, nLong = 0;
	LinkedHashMap<String, EST> estTagsMapSplicedForward = estFileName==null ? null : CompareCageEstRNASeqStarts.parseEstTags(estFileName, chr, true);
	LinkedHashMap<String, EST> estTagsMapSplicedReverse = estFileName==null ? null : CompareCageEstRNASeqStarts.parseEstTags(estFileName, chr, false);
	LinkedHashMap<String, EST> estTagsMapUnsplicedForward = estFileName==null ? null : CompareCageEstRNASeqStarts.parseEstTags(estFileName, chr, true);
	LinkedHashMap<String, EST> estTagsMapUnsplicedReverse = estFileName==null ? null : CompareCageEstRNASeqStarts.parseEstTags(estFileName, chr, false);
	try {
	    String saveName = saveFileName + "_" + ENHANCER_EST_FILE_SUFFIX + "_w_" + EST_ENHANCER_DISTANCE + "_" + chr + ".txt";
	    System.out.println("Saving enhancer sites to " + saveName);
	    BufferedWriter estWriter = estFileName==null ? null : new BufferedWriter(new FileWriter(saveName));
	    if (estWriter!=null && chr==1)
		estWriter.write("#chr\tCBP\tGeneID\tEST_5p\tEST_3p\tStrand\tSpliced\t\tOverlapsTSS\tName\n");
	for (Integer center : peaks.keySet()) {
	    Peak p = peaks.get(center);
	    boolean peakEnhancer = true;
	    int start = center - maxReadDistance, end = center + maxReadDistance, c = p.getChromosome();
	    if (c==chr && p.hasFactor(CBP_KCL_FILTER_STR)) {
		//Make sure that we are not in the middle of a
		//long transcript. This rerquirment not required
		//if we are looking at intronic sites.
		//if (p.hasFactor(CBP_KCL_FILTER_STR) && p.hasFactor(CBP_KCL_B2_FILTER_STR))
		//  System.out.println(p.toString() + " has reproducible CBP peak.");
		RefSeqGene overlappingGene = geneInfo[c - 1].getOverlappingGene(center);
		if (estFileName!=null) { //discard if there are too many ESTs starting near the enhancer
		    findNearbyESTs(estTagsMapSplicedForward, chr, true, center, EST_ENHANCER_DISTANCE, overlappingGene==null ? -1 : overlappingGene.getID(), geneInfo[c - 1], estWriter);
		    findNearbyESTs(estTagsMapSplicedReverse, chr, false, center, EST_ENHANCER_DISTANCE, overlappingGene==null ? -1 : overlappingGene.getID(), geneInfo[c - 1], estWriter);
		    findNearbyESTs(estTagsMapUnsplicedForward, chr, true, center, EST_ENHANCER_DISTANCE, overlappingGene==null ? -1 : overlappingGene.getID(), geneInfo[c - 1], estWriter);
		    findNearbyESTs(estTagsMapUnsplicedReverse, chr, false, center, EST_ENHANCER_DISTANCE, overlappingGene==null ? -1 : overlappingGene.getID(), geneInfo[c - 1], estWriter);
		}
		int startLong = center - maxReadDistance*LONG_FACTOR, endLong = center + maxReadDistance*LONG_FACTOR;
		if (!geneInfo[c - 1].overlapsTSS(p.getCenter() - MIN_TSS_DISTANCE, p.getCenter() + MIN_TSS_DISTANCE)) {
		    int[] nReadsForward = new int[forwardReads.size()], nReadsReverse = new int[reverseReads.size()], nReadsForwardLong = new int[forwardReads.size()], nReadsReverseLong = new int[reverseReads.size()];
		    int nReadsF = 0, nReadsR = 0;
		    boolean tooLong = false;
		    for (int i=0; i<forwardReads.size(); i++) {
			nReadsForward[i] = FindReadsAcrossFeature.numberOfReads(c, start, end, forwardReads.elementAt(i));
			nReadsReverse[i] = FindReadsAcrossFeature.numberOfReads(c, start, end, reverseReads.elementAt(i));
			nReadsForwardLong[i] = FindReadsAcrossFeature.numberOfReads(c, startLong, endLong, forwardReads.elementAt(i));
			nReadsReverseLong[i] = FindReadsAcrossFeature.numberOfReads(c, startLong, endLong, reverseReads.elementAt(i));
			//Make sure that we are not in the middle of a
			//longer transcript, unless we are in the
			//middle of an annotated gene.
			if (overlappingGene==null)
			    tooLong = !(nReadsForwardLong[i]<minReads*LONG_FACTOR*6 && nReadsReverseLong[i]<minReads*LONG_FACTOR*6);
			else
			    tooLong = !(overlappingGene.getStrand() ? nReadsReverseLong[i]<minReads*LONG_FACTOR*6 : nReadsForwardLong[i]<minReads*LONG_FACTOR*6);
			if (tooLong) { break; }
			nReadsF = Math.max(nReadsForward[i], nReadsF);
			nReadsR = Math.max(nReadsReverse[i], nReadsR);
		    }
		    //Find the closest K4Me3/1 peak
		    Peak h3K4Me1PeakUn = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME1_UN_STR, maxReadDistance, c);
		    Peak h3K4Me1PeakKCl = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME1_KCL_STR, maxReadDistance, c);
		    if (peakEnhancer) {
			RefSeqGene nearestGene = geneInfo[p.getChromosome()-1].getNearestTSSGene(center, false);
			int nearestDistance = nearestGene.getStrand() ? center - nearestGene.getStart() : nearestGene.getStart() - center;
			RefSeqGene nearestTESGene = geneInfo[p.getChromosome()-1].getNearestTESGene(center, false);
			int nearestTESDistance = nearestTESGene.getStrand() ? center - nearestGene.getEnd() : nearestGene.getEnd() - center;
			Enhancer en = new Enhancer(p, start, end, nReadsF, nReadsR, h3K4Me1PeakKCl!=null ? h3K4Me1PeakKCl.getFactor(H3K4ME1_KCL_STR) : 0, h3K4Me1PeakUn!=null ? h3K4Me1PeakUn.getFactor(H3K4ME1_UN_STR) : 0/*, h3K4Me3PeakUp!=null ? h3K4Me3PeakUp.getFactor(H3K4ME3_STR) : 0, h3K4Me3PeakDown!=null ? h3K4Me3PeakDown.getFactor(H3K4ME3_STR) : 0*/);
			//System.out.println("Found " + en.toString() + " and " + en.getPeak().isConserved());
			for (Filter filter : filters) {
			    peakEnhancer = true;
			    if (peakEnhancer && filter.contains(CONSERVATION_FILTER_STR)) {
				peakEnhancer = p.isConserved();
				//if (!peakEnhancer) { break; }
			    }
			    if (peakEnhancer && (filter.contains(RNA_SS_FILTER_STR) || filter.contains(RNA_BI_FILTER_STR))) {
				for (int i=0; i<forwardReads.size(); i++) {
				    if (filter.contains(RNA_BI_FILTER_STR))
					peakEnhancer = peakEnhancer && (nReadsForward[i]>minReads && nReadsReverse[i]>minReads);
				    else {
					peakEnhancer = peakEnhancer && (nReadsForward[i]>minReads || nReadsReverse[i]>minReads);
				    }
				    if (!peakEnhancer) { break; }
				}
			    }
			    /////////////////////////////////
			    /*if (peakEnhancer && filter.contains(POL2_FILTER_STR))
				peakEnhancer = p.hasFactor("8WG16") || p.hasFactor("4H8");
			    if (peakEnhancer && filter.contains(POL2_KCL_FILTER_STR))
				peakEnhancer = p.hasFactor("8WG16_KCl") || p.hasFactor("4H8_KCl");
			    if (peakEnhancer && filter.contains(POL2_UN_FILTER_STR))
			    peakEnhancer = p.hasFactor("8WG16_un") || p.hasFactor("4H8_un");*/
			    /////////////////////////////////
			    //if (peakEnhancer && filter.contains(CBP_FILTER_STR))
			    //peakEnhancer = p.hasFactor(CBP_FILTER_STR);
			    if (peakEnhancer && filter.contains(CBP_KCL_FILTER_STR)) {
				peakEnhancer = p.hasFactor(CBP_KCL_FILTER_STR);
			    }
			    if (peakEnhancer && filter.contains(CBP_UN_FILTER_STR))
				peakEnhancer = p.hasFactor(CBP_UN_FILTER_STR);
			    if (peakEnhancer && filter.contains(CBP_MP_UN_FILTER_STR))
				peakEnhancer = p.hasFactor(CBP_MP_UN_FILTER_STR);
			    /////////////////////////////////
			    //if (peakEnhancer && filter.contains(H3K4ME1_FILTER_STR))
			    //peakEnhancer = p.overlaps(h3K4Me1Peaks, H3K4ME1_FILTER_STR, maxReadDistance);
			    if (peakEnhancer && filter.contains(H3K4ME1_KCL_FILTER_STR))
				peakEnhancer = p.overlaps(h3K4Me1Peaks, H3K4ME1_KCL_FILTER_STR, maxReadDistance);
			    if (peakEnhancer && filter.contains(H3K4ME1_UN_FILTER_STR))
				peakEnhancer = p.overlaps(h3K4Me1Peaks, H3K4ME1_UN_FILTER_STR, maxReadDistance);
			    if (peakEnhancer && filter.contains(P300_UN_B1_FILTER_STR))
				peakEnhancer = p.hasFactor(P300_UN_B1_FILTER_STR);
			    if (peakEnhancer && filter.contains(P300_UN_B2_FILTER_STR))
				peakEnhancer = p.hasFactor(P300_UN_B2_FILTER_STR);
			    if (peakEnhancer && filter.contains(P300_KCL_B1_FILTER_STR))
				peakEnhancer = p.hasFactor(P300_KCL_B1_FILTER_STR);
			    if (peakEnhancer && filter.contains(P300_KCL_B2_FILTER_STR))
				peakEnhancer = p.hasFactor(P300_KCL_B2_FILTER_STR);
			    if (peakEnhancer && filter.contains(INTRONIC_FILTER_STR)) 
				peakEnhancer = overlappingGene!=null;
			    if (peakEnhancer && (filter.contains(DISTAL_FILTER_STR) || filter.contains(PROXIMAL_FILTER_STR))) { 
				if (overlappingGene==null) {
				    if (filter.contains(DISTAL_FILTER_STR)) 
					peakEnhancer = Math.abs(nearestDistance)>PROXIMAL_PROMOTER_CUTOFF && Math.abs(nearestTESDistance)>PROXIMAL_PROMOTER_CUTOFF;
				    else 
					peakEnhancer = (Math.abs(nearestDistance)<PROXIMAL_PROMOTER_CUTOFF || Math.abs(nearestTESDistance)<PROXIMAL_PROMOTER_CUTOFF);
				}
				else
				    peakEnhancer = false;
			    }
			    if (peakEnhancer) {
				if (!enhancers.containsKey(filter)) { enhancers.put(filter, new TreeSet<Enhancer>()); }
				//if (!tooLong)
				    enhancers.get(filter).add(en);
				    /*else*/ if (tooLong && filter.contains(H3K4ME1_KCL_FILTER_STR) && filter.contains(CBP_KCL_FILTER_STR)) 
				    nLong++;
			    }
			    else if (filter.contains(H3K4ME1_KCL_FILTER_STR) && filter.contains(CBP_KCL_FILTER_STR) && filter.size()==2) {
				nMe1++;
			    }
			}
		    }
		}
		else
		    nNearTSS++;
	    }
	}
	    if (estWriter!=null)
		estWriter.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving EST info " + e.toString());
	}
	for (Filter f : enhancers.keySet())
	    System.out.println("Found " + enhancers.get(f).size() + " enhancers for " + f.toString());
	try { //Save the failures to file, this is a truly ugly last-minute hack
	    BufferedWriter bw = new BufferedWriter(new FileWriter("removed_enhancers_" + chr + ".txt"));
	    if (chr==1)
		bw.write("#Near_TSS\tNo_Me1\tToo_long\n");
	    bw.write(nNearTSS + "\t" + nMe1 + "\t" + nLong + "\n");
	    bw.close();
	}
	catch (IOException e) {

	}
	return enhancers;
    }

    public static int findNearbyESTs(LinkedHashMap<String, EST> estTagsMap, int chr, boolean forward, int center, int maxDist, int geneID, RefSeqGenes geneInfo) { return findNearbyESTs(estTagsMap, chr, forward, center, maxDist, geneID, geneInfo, null); }

    public static int findNearbyESTs(LinkedHashMap<String, EST> estTagsMap, int chr, boolean forward, int center, int maxDist, int geneID, RefSeqGenes geneInfo, BufferedWriter bw) { 
		int nNear = 0;
	//System.out.println("Checking for ESTs near " + chr + ":" + center);
	//System.out.println(estTagsMap.get(estTagsMap.keySet().iterator().next()).toString());
	try {
	    for (String k : estTagsMap.keySet()) { //go through all ESTs
		EST est = estTagsMap.get(k);
		if (Math.abs(center - est.getStart())<maxDist) {
		    nNear++;
		    //System.out.println("Found " + est.toString() + " near chr" + chr + ":" + center);
		    if (bw!=null) {
			//check if intergenic EST spans any TSSs
			int s = forward ? est.getStart() : est.getEnd();
			int e = forward ? est.getEnd() : est.getStart();
			RefSeqGene overlappingTSS = geneInfo.getOverlappingTSS(s, e, true);
			int spansTSS = geneID==-1 && overlappingTSS!=null ? Math.abs(center - overlappingTSS.getStart()) : -1;
			bw.write(chr + "\t" + center + "\t" + geneID + "\t" + est.getStart() + "\t" + est.getEnd() + "\t" + (forward ? 1 : 0) + "\t" + (est.isSpliced() ? 1 : 0) + "\t" + spansTSS + "\t" + k + "\n");
		    }
		}
	    }
	}
	catch (IOException e) {
	    System.err.println("Error saving EST info " + e.toString());
	}
	return nNear;
    }

    public static HashSet<EST> getNearbyESTs(LinkedHashMap<String, EST> estTagsMap, int chr, boolean forward, int center, int maxDist) {
	HashSet<EST> ests = new HashSet<EST>();
	for (String k : estTagsMap.keySet()) { //go through all ESTs
	    EST est = estTagsMap.get(k);
	    if (Math.abs(center - est.getStart())<maxDist) {
		ests.add(est);
	    }
	}
	return ests;
    }

    public static HashSet<EST> getNearbyESTs(LinkedHashMap<String, EST> estTagsMap, int chr, boolean forward, int center, boolean spliced, int maxDist) {
	HashSet<EST> ests = new HashSet<EST>();
	for (String k : estTagsMap.keySet()) { //go through all ESTs
	    EST est = estTagsMap.get(k);
	    if (est.isSpliced()==spliced && Math.abs(center - est.getStart())<maxDist) {
		ests.add(est);
	    }
	}
	return ests;
    }

    public static TreeSet<Enhancer> getEnhancersNoCBP(int chr, TreeMap<Integer, Peak> peaks, TreeMap<Integer, Peak> h3K4Me1Peaks, RefSeqGenes[] geneInfo, Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, int minReads, int maxReadDistance, TreeSet<Enhancer> enhancers) {
	//Use this method to try to isolate enhancers that lack the CBP peak
	for (Integer center : peaks.keySet()) {
	    //First look for the ones with Pol2 peaks
	    Peak p = peaks.get(center);
	    boolean peakEnhancer = true;
	    int start = center - maxReadDistance, end = center + maxReadDistance, c = p.getChromosome();
	    if (c==chr) {
		//Make sure that we have pol2, but no CBP
		if (!p.hasFactor(CBP_FILTER_STR) && (p.hasFactor("8WG16") || p.hasFactor("4H8")) && p.isConserved()) {
		    RefSeqGene overlappingGene = geneInfo[c - 1].getOverlappingGene(center);
		    if (overlappingGene==null) { //non-intronic
			int startLong = center - maxReadDistance*LONG_FACTOR, endLong = center + maxReadDistance*LONG_FACTOR;
			if (!geneInfo[c - 1].overlapsTSS(p.getCenter() - MIN_TSS_DISTANCE, p.getCenter() + MIN_TSS_DISTANCE)) { //Distal
			    int[] nReadsForward = new int[forwardReads.size()], nReadsReverse = new int[reverseReads.size()], nReadsForwardLong = new int[forwardReads.size()], nReadsReverseLong = new int[reverseReads.size()];
			    int nReadsF = 0, nReadsR = 0;
			    for (int i=0; i<forwardReads.size(); i++) {
				nReadsForward[i] = FindReadsAcrossFeature.numberOfReads(c, start, end, forwardReads.elementAt(i));
				nReadsReverse[i] = FindReadsAcrossFeature.numberOfReads(c, start, end, reverseReads.elementAt(i));
				nReadsForwardLong[i] = FindReadsAcrossFeature.numberOfReads(c, startLong, endLong, forwardReads.elementAt(i));
				nReadsReverseLong[i] = FindReadsAcrossFeature.numberOfReads(c, startLong, endLong, reverseReads.elementAt(i));
				//Make sure that we are not in the middle of a
				//longer transcript, unless we are in the
				//middle of an annotated gene.
				peakEnhancer = nReadsForwardLong[i]<minReads*LONG_FACTOR*6 && nReadsReverseLong[i]<minReads*LONG_FACTOR*6;
				if (!peakEnhancer) { break; }
				nReadsF = Math.max(nReadsForward[i], nReadsF);
				nReadsR = Math.max(nReadsReverse[i], nReadsR);
			    }
			    if (peakEnhancer) {
				RefSeqGene nearestGene = geneInfo[p.getChromosome()-1].getNearestTSSGene(center, false);
				int nearestDistance = nearestGene.getStrand() ? center - nearestGene.getStart() : nearestGene.getStart() - center;
				RefSeqGene nearestTESGene = geneInfo[p.getChromosome()-1].getNearestTESGene(center, false);
				int nearestTESDistance = nearestTESGene.getStrand() ? center - nearestGene.getEnd() : nearestGene.getEnd() - center;
				if (Math.abs(nearestDistance)>PROXIMAL_PROMOTER_CUTOFF && Math.abs(nearestTESDistance)>PROXIMAL_PROMOTER_CUTOFF) {
				    //Find the closest K4Me3/1 peak
				    //Peak h3K4Me3PeakKCl = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME3_STR, maxReadDistance, true);
				    Peak h3K4Me1PeakKCl = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME1_STR, maxReadDistance, true);
				    //Peak h3K4Me3PeakUn = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME3_STR, maxReadDistance, false);
				    Peak h3K4Me1PeakUn = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME1_STR, maxReadDistance, false);
				    //if ((h3K4Me3PeakKCl==null || h3K4Me3PeakKCl.getFactor(H3K4ME3_KCL_STR)==0) && (h3K4Me3PeakUn==null || h3K4Me3PeakUn.getFactor(H3K4ME3_UN_STR)==0)) { //Avoid Me3
					if ((h3K4Me1PeakKCl!=null && h3K4Me1PeakKCl.getFactor(H3K4ME1_STR)>0) || (h3K4Me1PeakUn!=null && h3K4Me1PeakUn.getFactor(H3K4ME1_STR)>0)) { //Require Me1
					    Enhancer en = new Enhancer(p, start, end, nReadsF, nReadsR, h3K4Me1PeakKCl!=null ? h3K4Me1PeakKCl.getFactor(H3K4ME1_STR) : 0, h3K4Me1PeakUn!=null ? h3K4Me1PeakUn.getFactor(H3K4ME1_STR) : 0/*, h3K4Me3PeakKCl!=null ? h3K4Me3PeakKCl.getFactor(H3K4ME3_STR) : 0, h3K4Me3PeakUn!=null ? h3K4Me3PeakUn.getFactor(H3K4ME3_STR) : 0*/);
					    enhancers.add(en);
					}
					//} 
				}
			    }
			}
		    }
		}
	    }
	}
	System.out.println("Found " + enhancers.size() + " enhancers without CBP for chr " + chr);
	return enhancers;
    }
    
    //////////////////////////////////////////////////////////////////

    public static TreeMap<Integer, Peak> parsePeakFile(String peakFileName, int chr) { return parsePeakFile(peakFileName, chr, false); }

    public static TreeMap<Integer, Peak> parsePeakFile(String peakFileName, int chr, boolean conservedOnly) {
	TreeMap<Integer, String> factorNames = parsePeakFileHeader(peakFileName, chr);
	int nConsColumns = numberOfColumnsWith("cons", peakFileName, chr);
	int nGCColumns = numberOfColumnsWith("GC", peakFileName, chr);
	//System.out.println("Found " + nGCColumns + " GC columns and " + nConsColumns + " conservation columns.");
	TreeMap<Integer, Peak> peaks = new TreeMap<Integer, Peak>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(peakFileName + "_combos" + /*(chr==-1 ? "" : "_" + chr) + */".txt"));
	    //chr start end [gc mean_cons11 std_cons11 max_cons30 mean_cons30 std_cons11 max_cons30]
	    Pattern peakPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+([\\d|\\.]*)\\s*([\\d|\\.]*)\\s*([\\d|\\.]*)\\s*([\\d|\\.]*)\\s*([\\d|\\.]*)\\s+([\\d|\\.]*)\\s*([\\d|\\.]*)");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = peakPattern.matcher(line);
		if (m.find()) {
		    if (chr==-1 || Integer.parseInt(m.group(1))==chr) {
			//if (!conservedOnly || Double.parseDouble(m.group(7))>Peak.PEAK_CONSERVED_THRESHOLD) {
			Peak p = new Peak(Integer.parseInt(m.group(1)), Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)), (nConsColumns>0 ? Double.parseDouble(m.group(7)) : 0.0));
			    //Parse the peak sizes
			    String tokens[] = line.substring(m.end(3 + nConsColumns + nGCColumns)).trim().split("\\s+");
			    //System.out.println(Arrays.toString(tokens));
			    for (int i=0; i<factorNames.size(); i++) 
				p.addFactor(factorNames.get(i), Double.parseDouble(tokens[i]));
			    peaks.put(p.getCenter(), p);
			    //}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error, failed parsing " + peakFileName + " with error " + e.toString());
	}
	return peaks;
    }

    public static int numberOfColumnsWith(String text, String peakFileName, int chr) {
	int nCols = 0;
	String name = peakFileName + "_combos" + /*(chr==-1 ? "" : "_" + chr) +*/ ".txt";
	try {
	    BufferedReader br = new BufferedReader(new FileReader(name));
	    String line = br.readLine();
	    if (line.startsWith("#")) {
		String heads[] = line.substring(1).split("\t");
		int j = 0;
		for (int i=0; i<heads.length; i++) {
		    if (heads[i].contains(text))
			nCols++;
		}
	    }
	    else 
		System.err.println("Error, could not find header for " + name);
	    br.close();
	    }
	catch (IOException e) {
	    System.err.println("Error, failed parsing " + name);
	}
	return nCols;
    }

    public static TreeMap<Integer, String> parsePeakFileHeader(String peakFileName, int chr) {
	TreeMap<Integer, String> factorNames = new TreeMap<Integer, String>();
	String name = peakFileName + "_combos" + /*(chr==-1 ? "" : "_" + chr) +*/ ".txt";
	try {
	    BufferedReader br = new BufferedReader(new FileReader(name));
	    String line = br.readLine();
	    if (line.startsWith("#")) {
		String heads[] = line.substring(1).split("\t");
		int j = 0;
		for (int i=0; i<heads.length; i++) {
		    if (!heads[i].contains("cons") && !heads[i].contains("chr") && !heads[i].contains("Start") && !heads[i].contains("End") && !heads[i].contains("start") && !heads[i].contains("end") && !heads[i].contains("GC") && !heads[i].contains("Gene") && !heads[i].contains("TSS") && !heads[i].contains("TSE") && !heads[i].contains("hr")) {
			factorNames.put(j++, heads[i]);
		    }
		}
	    }
	    else 
		System.err.println("Error, could not find header for " + name);
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error, failed parsing " + name);
	}
	return factorNames;
    }

}

