//Use this method to decide which regions of the genome are considered
//transcribed based on the mRNA data. The first method that we use is
//very simple, relying only on a distance cut-off. This distance is
//based on a FDR calculation from the overall statistics of all mRNA
//reads outside the exonic and intronic regions.

//This is a slimmed down version of the previous
//FindEnhancerSites.java which only pulls out the enhancers which were
//previously in the null category.

//Example:

//java -Xmx512m FindEnhancerSites2 -d 2000 -p PeakFiles/enhancer_markers_names.txt

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class FindEnhancerSites2 {

    public static final String ENHANCER_SUFFIX = "enhancers";
    public static final String TRANSCRIBED_REGION_LENGTH_SUFFIX = "_transcribedRegionLengths";

    public static final String CBP_KCL_FILTER_STR = "CBP_KCl";
    public static final String CBP_KCL_B2_FILTER_STR = "CBP_KCl_B2";
    public static final String REPRODUCIBLE_CBP_KCL_FILTER_STR = "CBP_KCl_rep";
    public static final String CBP_UN_FILTER_STR = "CBP_un";
    public static final String CBP_FILTER_STR = "CBP";
    public static final String CONSERVATION_FILTER_STR = "Cons";
    public static final String POL2_KCL_FILTER_STR = "Pol2_KCl";
    public static final String POL2_UN_FILTER_STR = "Pol2_un";
    public static final String POL2_FILTER_STR = "Pol2";
    public static final String RNA_SS_FILTER_STR = "RNA_Ss";
    public static final String RNA_BI_FILTER_STR = "RNA_Bi";
    public static final String DISTAL_FILTER_STR = "Distal";
    public static final String INTRONIC_FILTER_STR = "Intronic";
    public static final String PROXIMAL_FILTER_STR = "Proximal";
    public static final String H3K4ME1_FILTER_STR = "H3K4Me1";
    public static final String H3K4ME3_FILTER_STR = "H3K4Me3";
    public static final String H3K4ME1_KCL_FILTER_STR = "H3K4Me1_KCl";
    public static final String H3K4ME3_KCL_FILTER_STR = "H3K4Me3_KCl";
    public static final String H3K4ME1_UN_FILTER_STR = "H3K4Me1_un";
    public static final String H3K4ME3_UN_FILTER_STR = "H3K4Me3_un";

    public static final int LONG_FACTOR = 4;
    public static final int PROXIMAL_PROMOTER_CUTOFF = 10000;
    public static final int ENHANCER_SAVE_WIDTH = 10000;
    public static final int MIN_TSS_DISTANCE = 1000;

    public static final String H3K4ME1_STR = "H3K4Me1";
    public static final String H3K4ME3_STR = "H3K4Me3";

    public static final String H3K4ME1_KCL_STR = "H3K4Me1_KCl";
    public static final String H3K4ME3_KCL_STR = "H3K4Me3_KCl";
    public static final String H3K4ME1_UN_STR = "H3K4Me1_un";
    public static final String H3K4ME3_UN_STR = "H3K4Me3_un";

    public static Vector<String> filterStrs = new Vector<String>();

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    String[] tmpStr = new String[]{CBP_FILTER_STR, CBP_UN_FILTER_STR, POL2_FILTER_STR, POL2_UN_FILTER_STR, POL2_KCL_FILTER_STR, H3K4ME1_FILTER_STR, H3K4ME3_FILTER_STR, H3K4ME1_KCL_FILTER_STR, H3K4ME3_KCL_FILTER_STR, H3K4ME1_UN_FILTER_STR, H3K4ME3_UN_FILTER_STR, CONSERVATION_FILTER_STR, RNA_SS_FILTER_STR, RNA_BI_FILTER_STR, PROXIMAL_FILTER_STR, DISTAL_FILTER_STR, INTRONIC_FILTER_STR};
	    filterStrs = new Vector(Arrays.asList(tmpStr));
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    //File containing a list of all the mRNA data sets that we
	    //want to investigate. This needs to correspond to the
	    //file which was used for generating the null peaks
	    CmdLineParser.Option mRNAFileNameFlag = argParser.addStringOption('m', "mRNAFileName");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option replicateFlag = argParser.addBooleanOption('b', "replicate");
	    CmdLineParser.Option reproducibleFlag = argParser.addBooleanOption("reproducible");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption('f', "fdr");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    //Restrict the length of the transcripts which will be
	    //candidates for the overlap regions
	    CmdLineParser.Option maxEnhancerTranscriptLengthFlag = argParser.addIntegerOption('l', "maxEnhancerTranscriptLength");
	    //The location of the peak file containing Pol2 and CBP peaks
	    CmdLineParser.Option peakFileFlag = argParser.addStringOption('p', "peakFileName");
	    //The location of the H3K4Me1 marks
	    CmdLineParser.Option h3K4Me1PeakFileFlag = argParser.addStringOption('P', "H3K4Me1PeakFileName");
	    //The max distance to the next read in the 3' direction
	    CmdLineParser.Option maxReadDistanceFlag = argParser.addIntegerOption('d', "maxReadDistance");
	    //The minimum number of reads in the transcript
	    CmdLineParser.Option minReadsFlag = argParser.addIntegerOption('r', "minReads");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean replicate = argParser.getOptionValue(replicateFlag)!=null;
	    boolean reproducible = argParser.getOptionValue(reproducibleFlag)!=null;
	    if (replicate)
		filterStrs.setElementAt(CBP_KCL_B2_FILTER_STR, 1);
	    else if (reproducible)
		filterStrs.setElementAt(REPRODUCIBLE_CBP_KCL_FILTER_STR, 1);
	    else
		filterStrs.setElementAt(CBP_KCL_FILTER_STR, 1);
	    tmp = argParser.getOptionValue(mRNAFileNameFlag);
	    String mRNAFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/hr";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "Enhancers/" + ENHANCER_SUFFIX + (replicate ? "_b2" : "") + (reproducible ? "_rep" : "");
	    tmp = argParser.getOptionValue(peakFileFlag);
	    String peakFileName = tmp!=null ? (String)tmp : "PeakFiles/enhancer_markers_" + (replicate ? "b2_" : "") + (reproducible ? "rep_" : "") + "names";
	    tmp = argParser.getOptionValue(h3K4Me1PeakFileFlag);
	    String h3K4Me1PeakFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(typeFlag);
	    String type = tmp!=null ? (String)tmp : "all";
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(maxEnhancerTranscriptLengthFlag);
	    int maxEnhancerTranscriptLength = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    tmp = argParser.getOptionValue(minReadsFlag);
	    int minReads = tmp!=null ? ((Integer)tmp).intValue() : 2;
	    tmp = argParser.getOptionValue(maxReadDistanceFlag);
	    int maxReadDistance = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    //////////////////////////////////////////////////////////////////
	    //Get the peak info
	    TreeMap<Integer, Peak> peaks = parsePeakFile(peakFileName, chr);
	    TreeSet<String> peakNames = new TreeSet<String>(parsePeakFileHeader(peakFileName, chr).values());
	    System.out.println("Found " + peaks.size() + " combo peaks with " + peakNames.size() + " factors from " + peakFileName);
	    //Get the K4Me1 marks, these are typically obtained using
	    //different parameters for the peakfinder.
	    TreeMap<Integer, Peak> h3K4Me1Peaks = parsePeakFile(h3K4Me1PeakFileName, chr, true);
	    TreeSet<String> h3K4PeakNames = new TreeSet<String>(parsePeakFileHeader(h3K4Me1PeakFileName, chr).values());
	    System.out.println("Found " + h3K4Me1Peaks.size() + " H3K4Me1 peaks with " + h3K4PeakNames.size() + " factors for chromosome.");
	    /////////////////////////////////
	    //Get the annotation info
	    int nChr = 21;
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || chr==i) 
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra);
	    }
	    /////////////////////////////////
	    //Keep track of the number of enhancers that will be
	    //extracted depending on what combination of criteria are
	    //used for selection. The candidates are:
	    //
	    //1. CPB peak
	    //2. Conservation
	    //3. Pol2 peak
	    //4. Bidirectional transcription
	    //5. More than 10kb away from gene
	    int nEnhancers = 0, nEnhancerWithRNA = 0, nEnhancerWithPol2 = 0, nEnhancerWithPol2AndRNA = 0, meanEnhancerDistance = 0, nConsCBP = 0;
	    //Keep track of the distance to the nearest TSS for each
	    //putative enhancer site
	    TreeMap<Integer, Integer> enhancerTSSDistance = new TreeMap<Integer, Integer>();
	    //Keep track of the number of enhancers per gene
	    TreeMap<RefSeqGene, Integer> geneEnhancers = new TreeMap<RefSeqGene, Integer>();
	    TreeMap<Peak, Enhancer> enhancers = new TreeMap<Peak, Enhancer>();
	    LinkedHashMap<Filter, TreeSet<Enhancer>> filter2enhancersMap = new LinkedHashMap<Filter, TreeSet<Enhancer>>();
	    //The enhancers that did not have CBP mark
	    TreeSet<Enhancer> enhancersNoCBP = new TreeSet<Enhancer>();
	    Vector<Integer> shortestLength = new Vector<Integer>();
	    int timePoints[] = new int[]{0, 1, 6};
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || chr==i) {
		    Vector<ABIRead[]> forwardReads = new Vector<ABIRead[]>(3);
		    Vector<ABIRead[]> reverseReads = new Vector<ABIRead[]>(3);
		    for (int j=0; j<timePoints.length; j++) {
			//Get the the mRNA info
			forwardReads.addElement(null);
			reverseReads.addElement(null);
			forwardReads.setElementAt(FindNongenicTranscripts.loadMRNAReadsChr(mRNAFileName + timePoints[j] + "c.tsv", true, i), j);
			reverseReads.setElementAt(FindNongenicTranscripts.loadMRNAReadsChr(mRNAFileName + timePoints[j] + "c.tsv", false, i), j);
			System.out.println("Loaded reads from t = " + timePoints[j] + " for chr " + i);
			///////////////////////////////////////////////////////
			//Find the length distribution of the
			//transcribed regions at each putative
			//enhancer location
			shortestLength.addAll(calculateRNALengthDistribution(i, peaks, h3K4Me1Peaks, geneInfo, forwardReads.lastElement(), reverseReads.lastElement(), minReads, maxReadDistance));
		    }
		    filter2enhancersMap = generateFilterCombos(i, peaks, h3K4Me1Peaks, geneInfo, forwardReads, reverseReads, minReads, maxReadDistance, filter2enhancersMap);
		    enhancersNoCBP = getEnhancersNoCBP(i, peaks, h3K4Me1Peaks, geneInfo, forwardReads, reverseReads, minReads, maxReadDistance, enhancersNoCBP);
		}
	    }
	    //////////////////////////////////////////////////////////////////
	    //Save the numbers to a file
	    String saveName = saveFileName + ENHANCER_NUMBERS_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt";
	    System.out.println("Saving enhancer sites to " + saveName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	    bw.write("#");
	    for (String f : filterStrs)
		bw.write(f + "\t");
	    bw.write("nEnhancers\n");
	    for (Filter filters : filter2enhancersMap.keySet()) {
		bw.write(filters.saveString(filterStrs) + filter2enhancersMap.get(filters).size() + "\n");
	    }
	    bw.close();
	    ///////////////////////////////////////////////////////
	    //Pool all the enhancers into one long list
	    TreeMap<Enhancer, Filter> allEnhancers = new TreeMap<Enhancer, Filter>(), highQualityEnhancers = new TreeMap<Enhancer, Filter>(), highQualityKClEnhancers = new TreeMap<Enhancer, Filter>(), highQualityUnEnhancers = new TreeMap<Enhancer, Filter>(), highQualityBothEnhancers = new TreeMap<Enhancer, Filter>(), mediumQualityEnhancers = new TreeMap<Enhancer, Filter>();
	    TreeMap<Enhancer, Filter> nullEnhancers = new TreeMap<Enhancer, Filter>(), nullEnhancersKCl = new TreeMap<Enhancer, Filter>(), nullEnhancersKClRep = new TreeMap<Enhancer, Filter>(), nullEnhancersUn = new TreeMap<Enhancer, Filter>(), nullEnhancersBoth = new TreeMap<Enhancer, Filter>(), nullEnhancersK4Me3 = new TreeMap<Enhancer, Filter>();
	    for (Filter filters : filter2enhancersMap.keySet()) {
		//Store enahncer as key and strictest filter as value
		if (!filters.distanceOK()) {
		    for (Enhancer en : filter2enhancersMap.get(filters)) {
			if (!allEnhancers.containsKey(en))
			    allEnhancers.put(en, filters);
			else if (allEnhancers.get(en).numerOfFilters()<filters.numerOfFilters())
			    allEnhancers.put(en, filters);
			if (!highQualityEnhancers.containsKey(en) && filters.highQuality())
			    highQualityEnhancers.put(en, filters);
			if (!highQualityKClEnhancers.containsKey(en) && filters.highQuality(true))
			    highQualityKClEnhancers.put(en, filters);
			if (!highQualityUnEnhancers.containsKey(en) && filters.highQuality(false))
			    highQualityUnEnhancers.put(en, filters);
			if (highQualityUnEnhancers.containsKey(en) && highQualityKClEnhancers.containsKey(en) && !highQualityBothEnhancers.containsKey(en))
			    highQualityBothEnhancers.put(en, filters);
			//////////////////////
			if (!mediumQualityEnhancers.containsKey(en) && filters.mediumQuality())
			    mediumQualityEnhancers.put(en, filters);
			if (!nullEnhancers.containsKey(en) && filters.nullFilter()/* && en.getH3K4Me3()==0*/)
			    nullEnhancers.put(en, filters);
			if (!nullEnhancersKCl.containsKey(en) && filters.nullFilter(true)/* && en.getH3K4Me3(true)==0*/) 
			    nullEnhancersKCl.put(en, filters);
			if (reproducible && !nullEnhancersKClRep.containsKey(en) && filters.nullFilterRep(true)/* && en.getH3K4Me3(true)==0*/) {
			    nullEnhancersKClRep.put(en, filters);
			    System.out.println(en.toString() + " has filters " + filters.toString() + " and CBP_KCl " + format.format(en.getPeak().getFactor(CBP_KCL_FILTER_STR)) + " and CBP_KCl_B2 " + format.format(en.getPeak().getFactor(CBP_KCL_B2_FILTER_STR)));
			}
			if (!nullEnhancersUn.containsKey(en) && filters.nullFilter(false)/* && en.getH3K4Me3(false)==0*/)
			    nullEnhancersUn.put(en, filters);
			if (nullEnhancersUn.containsKey(en) && nullEnhancersKCl.containsKey(en) && !nullEnhancersBoth.containsKey(en)/* && en.getH3K4Me3()==0*/)
			    nullEnhancersBoth.put(en, filters);
			//////////////////////
			if (!nullEnhancersK4Me3.containsKey(en) && filters.nullFilter(true) && en.getH3K4Me3(true)==0) 
			    nullEnhancersK4Me3.put(en, filters);
		    }
		}
	    }
	    saveName = saveFileName + (chr==-1 ? "" : "_" + chr) + ".txt";
	    saveEnhancers(saveName, peakNames, allEnhancers);
	    saveName = saveFileName + ENHANCER_HIGH_QUALITY_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt";
	    saveEnhancers(saveName, peakNames, highQualityEnhancers);
	    saveEnhancers(saveFileName + ENHANCER_HIGH_QUALITY_SUFFIX + "_KCl" + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, highQualityKClEnhancers);
	    saveEnhancers(saveFileName + ENHANCER_HIGH_QUALITY_SUFFIX + "_un" + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, highQualityUnEnhancers);
	    saveEnhancers(saveFileName + ENHANCER_HIGH_QUALITY_SUFFIX + "_both" + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, highQualityBothEnhancers);
	    //////////////////////
	    saveName = saveFileName + ENHANCER_MEDIUM_QUALITY_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt";
	    saveEnhancers(saveName, peakNames, mediumQualityEnhancers);
	    saveName = saveFileName + ENHANCER_NULL_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt";
	    saveEnhancers(saveName, peakNames, nullEnhancers);
	    saveName = saveFileName + ENHANCER_NULL_SUFFIX + "_KCl" + (chr==-1 ? "" : "_" + chr) + ".txt";
	    saveEnhancers(saveName, peakNames, nullEnhancersKCl);
	    //Save the distribtution of the transcribed regions at these enhancers
	    if (reproducible) { 
		saveName = saveFileName + ENHANCER_NULL_SUFFIX + "_KCl_rep" + (chr==-1 ? "" : "_" + chr) + ".txt";
		saveEnhancers(saveName, peakNames, nullEnhancersKClRep);
	    }
	    saveName = saveFileName + ENHANCER_NULL_SUFFIX + "_un" + (chr==-1 ? "" : "_" + chr) + ".txt";
	    saveEnhancers(saveName, peakNames, nullEnhancersUn);
	    saveEnhancers(saveFileName + ENHANCER_NULL_SUFFIX + "_both" + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, nullEnhancersBoth);
	    //////////////////////
	    //Null-enhancers with K4Me3
	    saveName = saveFileName + ENHANCER_NULL_SUFFIX + "_H3K4Me3" + (chr==-1 ? "" : "_" + chr) + ".txt";
	    saveEnhancers(saveName, peakNames, nullEnhancersK4Me3);
	    /////////////////////////////////
	    //Enhancers without CBP
	    saveEnhancers(saveFileName + ENHANCER_NO_CBP_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt", peakNames, enhancersNoCBP);
	    //////////////////////////////////////////////////////////////////
	    //Find the number of enhancers within a given distance of each TSS
	    for (int i=1; i<=nChr; i++) {
		if (chr==-1 || chr==i) {

		}
	    }
	    ///////////////////////////////////////////////////////
	    //Save the distribution of shortest transcribed regions 
	    saveName = saveFileName + TRANSCRIBED_REGION_LENGTH_SUFFIX + (chr==-1 ? "" : "_" + chr) + ".txt";
	    bw = new BufferedWriter(new FileWriter(saveName));
	    System.out.println("Saving lengths to " + saveName);
	    for (Integer i : shortestLength)
		bw.write(i + "\n");
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static void saveEnhancers(String saveName, TreeSet<String> peakNames, TreeMap<Enhancer, Filter> enhancers) throws IOException {
	System.out.println("Saving " + enhancers.size() + " enhancer sites to " + saveName);
	BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	bw.write("#Loc\tnForward\tnReverse\tH3K4Me1_Up\tH3K4Me3_Up\tH3K4Me1_Down\tH3K4Me3_Down");
	for (String s : peakNames)
	    bw.write("\t" + s);
	for (String f : filterStrs)
	    bw.write("\t" + f);
	bw.write("\n");
	for (Enhancer en : enhancers.keySet()) {
	    Peak p = en.getPeak();
	    Filter f = enhancers.get(en);
	    bw.write("chr" + AnalyzeControls.getChromosomeNumber(p.getChromosome()) + ":" + (p.getCenter() - ENHANCER_SAVE_WIDTH) + "-" + (p.getCenter() + ENHANCER_SAVE_WIDTH) + "\t" + en.numberOfReadsForward() + "\t" + en.numberOfReadsReverse() + "\t" + format.format(en.getH3K4Me1(true)) + "\t" + format.format(en.getH3K4Me3(true)) + "\t" + format.format(en.getH3K4Me1(false)) + "\t" + format.format(en.getH3K4Me3(false)));
	    //save information about the peaks that we found here
	    for (String s : peakNames)
		bw.write("\t" + format.format(p.getFactor(s)));
	    //Save information about which filters it satisfied
	    bw.write("\t" + f.saveString(filterStrs));
	    bw.write("\n");
	}
	bw.close();

    }

    public static void saveEnhancers(String saveName, TreeSet<String> peakNames, TreeSet<Enhancer> enhancers) throws IOException {
	//For the ones without CBP
	System.out.println("Saving " + enhancers.size() + " enhancer sites to " + saveName);
	BufferedWriter bw = new BufferedWriter(new FileWriter(saveName));
	bw.write("#Loc\tnForward\tnReverse\tH3K4Me1_Up\tH3K4Me1_Down");
	for (String s : peakNames)
	    bw.write("\t" + s);
	bw.write("\n");
	for (Enhancer en : enhancers) {
	    Peak p = en.getPeak();
	    bw.write("chr" + AnalyzeControls.getChromosomeNumber(p.getChromosome()) + ":" + (p.getCenter() - ENHANCER_SAVE_WIDTH) + "-" + (p.getCenter() + ENHANCER_SAVE_WIDTH) + "\t" + en.numberOfReadsForward() + "\t" + en.numberOfReadsReverse() + "\t" + format.format(en.getH3K4Me1(true)) + "\t" + format.format(en.getH3K4Me1(false)));
	    //save information about the peaks that we found here
	    for (String s : peakNames)
		bw.write("\t" + format.format(p.getFactor(s)));
	    bw.write("\n");
	}
	bw.close();

    }

    //////////////////////////////////////////////////////////////////

    public static LinkedHashMap<Filter, TreeSet<Enhancer>> generateFilterCombos(int chr, TreeMap<Integer, Peak> peaks, TreeMap<Integer, Peak> h3K4Me1Peaks, RefSeqGenes[] geneInfo, Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, int minReads, int maxReadDistance, LinkedHashMap<Filter, TreeSet<Enhancer>> map) {
	LinkedHashSet<Filter> filters = map.size()>0 ? new LinkedHashSet<Filter>(map.keySet()) : getAllFilters();
	//for (Filter filter : filters) {
	    return getEnhancers(chr, peaks, h3K4Me1Peaks, geneInfo, forwardReads, reverseReads, filters, minReads, maxReadDistance, map);
	    /*if (map.containsKey(filter))
		map.get(filter).addAll(e);
	    else
		map.put(filter, e);
	}
	return map;*/
    }

    public static LinkedHashSet<Filter> getAllFilters() {
	LinkedHashSet<Filter> filters = new LinkedHashSet<Filter>();
	//filters.add(new Filter(new String[]{CBP_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_STR}));
	//filters.add(new Filter(new String[]{CBP_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_STR, PROXIMAL_FILTER_STR}));
	//filters.add(new Filter(new String[]{CBP_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_STR, INTRONIC_FILTER_STR}));
	//filters.add(new Filter(new String[]{CBP_FILTER_STR, CONSERVATION_FILTER_STR, H3K4ME1_STR, DISTAL_FILTER_STR}));
	for (String f : filterStrs) {
	    Filter filter = new Filter(f);
	    if (filter.hasRequired()) {
		filters.add(filter);
		//Recursively create the rest of the filters
		filters.addAll(getAllFilters(filter, filters));
	    }
	}
	System.out.println("Found " + filters.size() + " filters.");
	return filters;
    }

    public static LinkedHashSet<Filter> getAllFilters(Filter filter, LinkedHashSet<Filter> filters) {
	for (String f : filterStrs) {
	    if (filter.filterOK(f)) {
		Filter newFilter = new Filter(filter, f);
		if (!filters.contains(newFilter) && newFilter.hasRequired()) {
		    filters.add(newFilter);
		    //System.out.println("Added filter " + FILTER_STRS[i] + " to " + newFilter.toString());
		    filters.addAll(getAllFilters(newFilter, filters));
		}
	    }
	}
	return filters;
    }
    
    /////////////////////////////////

    public static Vector<Integer> calculateRNALengthDistribution(int chr, TreeMap<Integer, Peak> peaks, TreeMap<Integer, Peak> h3K4Me1Peaks, RefSeqGenes[] geneInfo, ABIRead[] forwardReads, ABIRead[] reverseReads, int minReads, int maxReadDistance) {
	//Use this method to calculate the distribution of the
	//transcribed regions for all the putative enhancer sites.
	Vector<Integer> shortestLengths= new Vector<Integer>();
	for (Integer center : peaks.keySet()) {
	    Peak p = peaks.get(center);
	    int start = center - maxReadDistance*LONG_FACTOR, end = center + maxReadDistance*LONG_FACTOR, c = p.getChromosome();
	    if (c==chr || chr==-1) {
		//Check for conservation and nearby K4Me1
		Peak h3K4Me1Peak = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME1_STR, maxReadDistance);
		if (p.isConserved() && h3K4Me1Peak!=null && h3K4Me1Peak.getFactor(H3K4ME1_STR)>0) {
		    RefSeqGene overlappingGene = geneInfo[c - 1].getOverlappingGene(center);
		    if (overlappingGene==null && !geneInfo[c - 1].overlapsTSS(p.getCenter() - MIN_TSS_DISTANCE, p.getCenter() + MIN_TSS_DISTANCE)) {
			//Find the lengths of the transcribed regions
			int nReadsF = 0, nReadsR = 0;
			ABIRead readsForward[] = FindReadsAcrossFeature.getReads(c, start, end, forwardReads);
			int transcribedRegionForward[] = IsolateBidirectionalTranscription.findTranscribedRegionBoundaries(readsForward, true, maxReadDistance);
			ABIRead readsReverse[] = FindReadsAcrossFeature.getReads(c, start, end, reverseReads);
			int transcribedRegionReverse[] = IsolateBidirectionalTranscription.findTranscribedRegionBoundaries(readsReverse, false, maxReadDistance);
			shortestLengths.addElement(Math.max(Math.abs(transcribedRegionForward[0] - transcribedRegionForward[1]), Math.abs(transcribedRegionReverse[0] - transcribedRegionReverse[1])));
		    }
		}
	    }
	}
	return shortestLengths;
    }

    public static LinkedHashMap<Filter, TreeSet<Enhancer>> getEnhancers(int chr, TreeMap<Integer, Peak> peaks, TreeMap<Integer, Peak> h3K4Me1Peaks, RefSeqGenes[] geneInfo, Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, LinkedHashSet<Filter> filters, int minReads, int maxReadDistance, LinkedHashMap<Filter, TreeSet<Enhancer>> enhancers) {
	for (Integer center : peaks.keySet()) {
	    Peak p = peaks.get(center);
	    boolean peakEnhancer = true;
	    int start = center - maxReadDistance, end = center + maxReadDistance, c = p.getChromosome();
	    if (c==chr) {
		//Make sure that we are not in the middle of a
		//long transcript. This rerquirment not required
		//if we are looking at intronic sites.
		//if (p.hasFactor(CBP_KCL_FILTER_STR) && p.hasFactor(CBP_KCL_B2_FILTER_STR))
		//  System.out.println(p.toString() + " has reproducible CBP peak.");
		RefSeqGene overlappingGene = geneInfo[c - 1].getOverlappingGene(center);
		int startLong = center - maxReadDistance*LONG_FACTOR, endLong = center + maxReadDistance*LONG_FACTOR;
		if (!geneInfo[c - 1].overlapsTSS(p.getCenter() - MIN_TSS_DISTANCE, p.getCenter() + MIN_TSS_DISTANCE)) {
		    int[] nReadsForward = new int[forwardReads.size()], nReadsReverse = new int[reverseReads.size()], nReadsForwardLong = new int[forwardReads.size()], nReadsReverseLong = new int[reverseReads.size()];
		    int nReadsF = 0, nReadsR = 0;
		    for (int i=0; i<forwardReads.size(); i++) {
			nReadsForward[i] = FindReadsAcrossFeature.numberOfReads(c, start, end, forwardReads.elementAt(i));
			nReadsReverse[i] = FindReadsAcrossFeature.numberOfReads(c, start, end, reverseReads.elementAt(i));
			nReadsForwardLong[i] = FindReadsAcrossFeature.numberOfReads(c, startLong, endLong, forwardReads.elementAt(i));
			nReadsReverseLong[i] = FindReadsAcrossFeature.numberOfReads(c, startLong, endLong, reverseReads.elementAt(i));
			//Make sure that we are not in the middle of a
			//longer transcript, unless we are in the
			//middle of an annotated gene.
			if (overlappingGene==null)
			    peakEnhancer = nReadsForwardLong[i]<minReads*LONG_FACTOR*6 && nReadsReverseLong[i]<minReads*LONG_FACTOR*6;
			else
			    peakEnhancer = overlappingGene.getStrand() ? nReadsReverseLong[i]<minReads*LONG_FACTOR*6 : nReadsForwardLong[i]<minReads*LONG_FACTOR*6;
			if (!peakEnhancer) { break; }
			nReadsF = Math.max(nReadsForward[i], nReadsF);
			nReadsR = Math.max(nReadsReverse[i], nReadsR);
		    }
		    if (peakEnhancer) {
			RefSeqGene nearestGene = geneInfo[p.getChromosome()-1].getNearestTSSGene(center, false);
			int nearestDistance = nearestGene.getStrand() ? center - nearestGene.getStart() : nearestGene.getStart() - center;
			RefSeqGene nearestTESGene = geneInfo[p.getChromosome()-1].getNearestTESGene(center, false);
			int nearestTESDistance = nearestTESGene.getStrand() ? center - nearestGene.getEnd() : nearestGene.getEnd() - center;
			//Find the closest K4Me3/1 peak
			Peak h3K4Me1Peak = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME1_STR, maxReadDistance);
			Enhancer en = new Enhancer(p, start, end, nReadsF, nReadsR, h3K4Me1Peak);
			for (Filter filter : filters) {
			    peakEnhancer = true;
			    if (peakEnhancer && filter.contains(CONSERVATION_FILTER_STR))
				peakEnhancer = p.isConserved();
			    if (peakEnhancer && (filter.contains(RNA_SS_FILTER_STR) || filter.contains(RNA_BI_FILTER_STR))) {
				for (int i=0; i<forwardReads.size(); i++) {
				    if (filter.contains(RNA_BI_FILTER_STR))
					peakEnhancer = peakEnhancer && (nReadsForward[i]>minReads && nReadsReverse[i]>minReads);
				    else {
					peakEnhancer = peakEnhancer && (nReadsForward[i]>minReads || nReadsReverse[i]>minReads);
				    }
				    if (!peakEnhancer) { break; }
				}
			    }
			    /////////////////////////////////
			    if (peakEnhancer && filter.contains(POL2_FILTER_STR))
				peakEnhancer = p.hasFactor("8WG16") || p.hasFactor("4H8");
			    if (peakEnhancer && filter.contains(POL2_KCL_FILTER_STR))
				peakEnhancer = p.hasFactor("8WG16_KCl") || p.hasFactor("4H8_KCl");
			    if (peakEnhancer && filter.contains(POL2_UN_FILTER_STR))
				peakEnhancer = p.hasFactor("8WG16_un") || p.hasFactor("4H8_un");
			    /////////////////////////////////
			    if (peakEnhancer && filter.contains(CBP_FILTER_STR))
				peakEnhancer = p.hasFactor(CBP_FILTER_STR);
			    if (peakEnhancer && filter.contains(CBP_KCL_FILTER_STR))
				peakEnhancer = p.hasFactor(CBP_KCL_FILTER_STR);
			    if (peakEnhancer && filter.contains(CBP_KCL_B2_FILTER_STR))
				peakEnhancer = p.hasFactor(CBP_KCL_B2_FILTER_STR);
			    if (peakEnhancer && filter.contains(REPRODUCIBLE_CBP_KCL_FILTER_STR))
				peakEnhancer = p.hasFactor(CBP_KCL_FILTER_STR, true) && p.hasFactor(CBP_KCL_B2_FILTER_STR, true);
			    if (peakEnhancer && filter.contains(CBP_UN_FILTER_STR))
				peakEnhancer = p.hasFactor(CBP_UN_FILTER_STR);
			    /////////////////////////////////
			    if (peakEnhancer && filter.contains(H3K4ME1_FILTER_STR))
				peakEnhancer = p.overlaps(h3K4Me1Peaks, H3K4ME1_FILTER_STR, maxReadDistance);
			    if (peakEnhancer) {
				if (!enhancers.containsKey(filter)) { enhancers.put(filter, new TreeSet<Enhancer>()); }
				enhancers.get(filter).add(en);
			    }
			}
		    }
		}
	    }
	}
	//System.out.println("Found " + enhancers.size() + " enhancers for " + filter.toString());
	return enhancers;
    }

    public static TreeSet<Enhancer> getEnhancersNoCBP(int chr, TreeMap<Integer, Peak> peaks, TreeMap<Integer, Peak> h3K4Me1Peaks, RefSeqGenes[] geneInfo, Vector<ABIRead[]> forwardReads, Vector<ABIRead[]> reverseReads, int minReads, int maxReadDistance, TreeSet<Enhancer> enhancers) {
	//Use this method to try to isolate enhancers that lack the CBP peak
	for (Integer center : peaks.keySet()) {
	    //First look for the ones with Pol2 peaks
	    Peak p = peaks.get(center);
	    boolean peakEnhancer = true;
	    int start = center - maxReadDistance, end = center + maxReadDistance, c = p.getChromosome();
	    if (c==chr) {
		//Make sure that we have pol2, but no CBP
		if (!p.hasFactor(CBP_FILTER_STR) && (p.hasFactor("8WG16") || p.hasFactor("4H8")) && p.isConserved()) {
		    RefSeqGene overlappingGene = geneInfo[c - 1].getOverlappingGene(center);
		    if (overlappingGene==null) { //non-intronic
			int startLong = center - maxReadDistance*LONG_FACTOR, endLong = center + maxReadDistance*LONG_FACTOR;
			if (!geneInfo[c - 1].overlapsTSS(p.getCenter() - MIN_TSS_DISTANCE, p.getCenter() + MIN_TSS_DISTANCE)) { //Distal
			    int[] nReadsForward = new int[forwardReads.size()], nReadsReverse = new int[reverseReads.size()], nReadsForwardLong = new int[forwardReads.size()], nReadsReverseLong = new int[reverseReads.size()];
			    int nReadsF = 0, nReadsR = 0;
			    for (int i=0; i<forwardReads.size(); i++) {
				nReadsForward[i] = FindReadsAcrossFeature.numberOfReads(c, start, end, forwardReads.elementAt(i));
				nReadsReverse[i] = FindReadsAcrossFeature.numberOfReads(c, start, end, reverseReads.elementAt(i));
				nReadsForwardLong[i] = FindReadsAcrossFeature.numberOfReads(c, startLong, endLong, forwardReads.elementAt(i));
				nReadsReverseLong[i] = FindReadsAcrossFeature.numberOfReads(c, startLong, endLong, reverseReads.elementAt(i));
				//Make sure that we are not in the middle of a
				//longer transcript, unless we are in the
				//middle of an annotated gene.
				peakEnhancer = nReadsForwardLong[i]<minReads*LONG_FACTOR*6 && nReadsReverseLong[i]<minReads*LONG_FACTOR*6;
				if (!peakEnhancer) { break; }
				nReadsF = Math.max(nReadsForward[i], nReadsF);
				nReadsR = Math.max(nReadsReverse[i], nReadsR);
			    }
			    if (peakEnhancer) {
				RefSeqGene nearestGene = geneInfo[p.getChromosome()-1].getNearestTSSGene(center, false);
				int nearestDistance = nearestGene.getStrand() ? center - nearestGene.getStart() : nearestGene.getStart() - center;
				RefSeqGene nearestTESGene = geneInfo[p.getChromosome()-1].getNearestTESGene(center, false);
				int nearestTESDistance = nearestTESGene.getStrand() ? center - nearestGene.getEnd() : nearestGene.getEnd() - center;
				if (Math.abs(nearestDistance)>PROXIMAL_PROMOTER_CUTOFF && Math.abs(nearestTESDistance)>PROXIMAL_PROMOTER_CUTOFF) {
				    //Find the closest K4Me3/1 peak
				    Peak h3K4Me3PeakKCl = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME3_STR, maxReadDistance, true);
				    Peak h3K4Me1PeakKCl = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME1_STR, maxReadDistance, true);
				    Peak h3K4Me3PeakUn = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME3_STR, maxReadDistance, false);
				    Peak h3K4Me1PeakUn = p.getOverlappingPeak(h3K4Me1Peaks, H3K4ME1_STR, maxReadDistance, false);
				    if ((h3K4Me3PeakKCl==null || h3K4Me3PeakKCl.getFactor(H3K4ME3_KCL_STR)==0) && (h3K4Me3PeakUn==null || h3K4Me3PeakUn.getFactor(H3K4ME3_UN_STR)==0)) { //Avoid Me3
					if ((h3K4Me1PeakKCl!=null && h3K4Me1PeakKCl.getFactor(H3K4ME1_STR)>0) || (h3K4Me1PeakUn!=null && h3K4Me1PeakUn.getFactor(H3K4ME1_STR)>0)) { //Require Me1
					    Enhancer en = new Enhancer(p, start, end, nReadsF, nReadsR, h3K4Me1PeakKCl!=null ? h3K4Me1PeakKCl.getFactor(H3K4ME1_STR) : 0, h3K4Me1PeakUn!=null ? h3K4Me1PeakUn.getFactor(H3K4ME1_STR) : 0, h3K4Me3PeakKCl!=null ? h3K4Me3PeakKCl.getFactor(H3K4ME3_STR) : 0, h3K4Me3PeakUn!=null ? h3K4Me3PeakUn.getFactor(H3K4ME3_STR) : 0);
					    enhancers.add(en);
					}
				    } 
				}
			    }
			}
		    }
		}
	    }
	}
	System.out.println("Found " + enhancers.size() + " enhancers without CBP for chr " + chr);
	return enhancers;
    }
    
    //////////////////////////////////////////////////////////////////

    public static TreeMap<Integer, Peak> parsePeakFile(String peakFileName, int chr) { return parsePeakFile(peakFileName, chr, false); }

    public static TreeMap<Integer, Peak> parsePeakFile(String peakFileName, int chr, boolean conservedOnly) {
	TreeMap<Integer, String> factorNames = parsePeakFileHeader(peakFileName, chr);
	TreeMap<Integer, Peak> peaks = new TreeMap<Integer, Peak>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(peakFileName + "_combos" + /*(chr==-1 ? "" : "_" + chr) + */".txt"));
	    Pattern peakPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+([\\d|\\.]+)\\s+([\\d|\\.]+)\\s+([\\d|\\.]+)\\s+([\\d|\\.]+)\\s+([\\d|\\.]+)\\s+([\\d|\\.]+)\\s+([\\d|\\.]+)");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = peakPattern.matcher(line);
		if (m.find()) {
		    if (chr==-1 || Integer.parseInt(m.group(1))==chr) {
			if (!conservedOnly || Double.parseDouble(m.group(7))>Peak.PEAK_CONSERVED_THRESHOLD) {
			    Peak p = new Peak(Integer.parseInt(m.group(1)), Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)), Double.parseDouble(m.group(7)));
			    //Parse the peak sizes
			    String tokens[] = line.substring(m.end()).trim().split("\\s+");
			    for (int i=0; i<factorNames.size(); i++) 
				p.addFactor(factorNames.get(i), Double.parseDouble(tokens[i]));
			    peaks.put(p.getCenter(), p);
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error, failed parsing " + peakFileName);
	}
	return peaks;
    }

    public static TreeMap<Integer, String> parsePeakFileHeader(String peakFileName, int chr) {
	TreeMap<Integer, String> factorNames = new TreeMap<Integer, String>();
	String name = peakFileName + "_combos" + /*(chr==-1 ? "" : "_" + chr) +*/ ".txt";
	try {
	    BufferedReader br = new BufferedReader(new FileReader(name));
	    String line = br.readLine();
	    if (line.startsWith("#")) {
		String heads[] = line.substring(1).split("\t");
		int j = 0;
		for (int i=0; i<heads.length; i++) {
		    if (!heads[i].contains("cons") && !heads[i].contains("chr") && !heads[i].contains("Start") && !heads[i].contains("End") && !heads[i].contains("GC") && !heads[i].contains("Gene") && !heads[i].contains("TSS") && !heads[i].contains("TSE") && !heads[i].contains("hr")) {
			factorNames.put(j++, heads[i]);
		    }
		}
	    }
	    else 
		System.err.println("Error, could not find header for " + name);
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error, failed parsing " + name);
	}
	return factorNames;
    }

}

