//Use this method to decide which regions of the genome are considered
//transcribed based on the mRNA data. The first method that we use is
//very simple, relying only on a distance cut-off. This distance is
//based on a FDR calculation from the overall statistics of all mRNA
//reads outside the exonic and intronic regions.

//Example:

//java -Xmx512m FindNongenicTranscripts -m mRNA/hr0c.tsv -f .01 -s mRNA/NullDistributions/hr0c

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class FindNongenicTranscripts{

    public static final String OVERLAPPING_SUFFIX = "_overlapping_transcripts";
    public static final String TRANSCRIPT_SUFFIX = "_transcripts";
    public static final String NONGENIC_STR = "nongenic";

    public static final int MIN_READS_TRANSCRIPT = 4;
    public static final int MRNA_READ_LENGTH = 35;
    public static final int MAX_READS_PER_LOCUS = 20000;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    //File containing a list of all the mRNA data sets that we
	    //want to investigate. This needs to correspond to the
	    //file which was used for generating the null peaks
	    CmdLineParser.Option mRNAFileNameFlag = argParser.addStringOption('m', "mRNAFileName");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption('f', "fdr");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    //Decide what type of method we are going to use
	    CmdLineParser.Option typeFlag = argParser.addStringOption('t', "type");
	    //Restrict the length of the transcripts which will be
	    //candidates for the overlap regions
	    CmdLineParser.Option maxEnhancerTranscriptLengthFlag = argParser.addIntegerOption('l', "maxEnhancerTranscriptLength");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    tmp = argParser.getOptionValue(mRNAFileNameFlag);
	    String mRNAFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(typeFlag);
	    String type = tmp!=null ? (String)tmp : NONGENIC_STR;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(maxEnhancerTranscriptLengthFlag);
	    int maxEnhancerTranscriptLength = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    //////////////////////////////////////////////////////////////////
	    int nChr = 21;
	    TreeSet<NonGenicTranscript> transcriptsForward = findTranscripts(mRNAFileName, true, nChr, orchestra, type, fdr, MIN_READS_TRANSCRIPT, saveFileName);
	    TreeSet<NonGenicTranscript> transcriptsReverse = findTranscripts(mRNAFileName, false, nChr, orchestra, type, fdr, MIN_READS_TRANSCRIPT, saveFileName);
	    //Find the overlapping transcripts
	    TreeMap<NonGenicTranscript, NonGenicTranscript> overlappingTranscripts = new TreeMap<NonGenicTranscript, NonGenicTranscript>(); //key is forward and value is reverse transcript
	    for (NonGenicTranscript tf : transcriptsForward) {
		if (tf.getLength()<maxEnhancerTranscriptLength) { //Make sure that the transcript is not too long
		    for (NonGenicTranscript tr : transcriptsReverse) {
			if (tr.getLength()<maxEnhancerTranscriptLength) { //Make sure that the transcript is not too long
			    //The indexes are always wrt the start of the forward strand
			    if (tf.getStart()<tr.getEnd() && tf.getEnd()>tr.getEnd())  //tf_s tr_e tf_e
				overlappingTranscripts.put(tf, tr);
			    else if (tf.getStart()<tr.getStart() && tf.getEnd()>tr.getStart())  //tf_s tr_s tf_e
				overlappingTranscripts.put(tf, tr);
			    else if (tf.getStart()>tr.getStart() && tf.getEnd()<tr.getEnd())  //tr_s tf_s tf_e tr_e
				overlappingTranscripts.put(tf, tr);
			    else if (tf.getStart()<tr.getStart() && tf.getEnd()>tr.getEnd())  //tf_s tr_s tr_e tf_e
				overlappingTranscripts.put(tf, tr);
			}
		    }
		}
	    }
	    String fileName = saveFileName + OVERLAPPING_SUFFIX + ".txt"; //Save stuff
	    BufferedWriter bw = new BufferedWriter(new FileWriter(fileName));
	    for (NonGenicTranscript tf : overlappingTranscripts.keySet()) {
		NonGenicTranscript tr = overlappingTranscripts.get(tf);
		int start = Math.min(tf.getStart(), tr.getStart());
		int end = Math.max(tf.getEnd(), tr.getEnd());
		bw.write(tf.getChromosome() + ":" + start + "-" + end + "\t" + tf.numberOfReads() + "\t" + tr.numberOfReads() + "\n");
	    }
	    bw.close();
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static TreeSet<NonGenicTranscript> findTranscripts(String mRNAFileName, boolean forward, int nChr, boolean orchestra, String type, double fdr, int minReads, String saveFileName) {
	RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	for (int i=1; i<=nChr; i++)
	    geneInfo[i-1] = new RefSeqGenes(i, orchestra);
	Vector<ABIRead[]> reads = loadMRNAReads(mRNAFileName, forward, nChr);
	TreeMap<Integer, Integer> distanceDistribution = mRNADistanceDistribution(type, mRNAFileName, forward, orchestra, geneInfo, reads);
	//Determine the max distance between two transcripts based on the desired FDR
	int maxReadDistance = findMaxReadDistance(distanceDistribution, fdr);
	return findTranscripts(-1, reads, geneInfo, forward, type, maxReadDistance, minReads, saveFileName);
    }

    public static TreeSet<NonGenicTranscript> findTranscripts(int chr, String mRNAFileName, boolean forward, RefSeqGenes geneInfo[], String type, int maxReadDistance, int minReads, String saveFileName) {
	int nChr = geneInfo.length;
	Vector<ABIRead[]> reads = new Vector<ABIRead[]>(nChr);
	for (int i=1; i<=nChr; i++) {
	    reads.addElement(null);
	    if (chr==-1 || chr==i) 
		reads.setElementAt(loadMRNAReadsChr(mRNAFileName, forward, i), i-1);
	}
	return findTranscripts(chr, reads, geneInfo, forward, type, maxReadDistance, minReads, saveFileName);
    }

    protected static TreeSet<NonGenicTranscript> findTranscripts(int chr, Vector<ABIRead[]> reads, RefSeqGenes geneInfo[], boolean forward, String type, int maxReadDistance, int minReads, String saveFileName) {
	int nIsolated = 0, sumReadsInTranscript = 0, nChr = reads.size();
	TreeMap<Integer, Integer> transcriptLengthDist = new TreeMap<Integer, Integer>();
	//Go through all transcripts and find which ones are close
	//enough to be considered belonging to the same transcript.
	TreeSet<NonGenicTranscript> nonGenicTranscripts = new TreeSet<NonGenicTranscript>();
	for (int j=1; j<=nChr; j++) {
	    if (chr==-1 || j==chr) {
		ABIRead[] r = reads.elementAt(j-1);
		if (forward) {
		    int is = 0, transcriptStart = r[is].getPosition(), nReadsInTranscript = 0;
		    //while (is<r.length-1 && geneInfo[j-1].overlapsGene(transcriptStart, transcriptStart + MRNA_READ_LENGTH))
		    transcriptStart = r[0].getPosition();
		    for (int i=is+1; i<r.length; i++) {
			if (r[i].getPosition() - r[i-1].getPosition()>maxReadDistance || (type.equals(NONGENIC_STR) && geneInfo[j-1].overlapsGene(r[i-1].getPosition(), r[i].getPosition(), true))) { //transcript ended
			    if (nReadsInTranscript>minReads) {
				nonGenicTranscripts.add(new NonGenicTranscript(j, forward, transcriptStart, r[i-1].getPosition() + MRNA_READ_LENGTH, nReadsInTranscript));
				sumReadsInTranscript += nReadsInTranscript;
			    }
			    else
				nIsolated += nReadsInTranscript;
			    transcriptStart = r[i].getPosition();
			    nReadsInTranscript = 1;
			}
			else
			nReadsInTranscript++;
		    }
		}
		else {
		    int transcriptStart = r[r.length-1].getPosition() + MRNA_READ_LENGTH, nReadsInTranscript = 0;
		    for (int i=r.length-2; i>=0; i--) {
			if (r[i+1].getPosition() - r[i].getPosition()>maxReadDistance || (type.equals(NONGENIC_STR) && geneInfo[j-1].overlapsGene(r[i].getPosition() + MRNA_READ_LENGTH, r[i+1].getPosition() + MRNA_READ_LENGTH, false))) { //transcript ended
			    if (nReadsInTranscript>minReads) {
				nonGenicTranscripts.add(new NonGenicTranscript(j, forward, transcriptStart, r[i+1].getPosition(), nReadsInTranscript));
				sumReadsInTranscript += nReadsInTranscript;
			    }
			    else
				nIsolated += nReadsInTranscript;
			    transcriptStart = r[i].getPosition() + MRNA_READ_LENGTH;
			    nReadsInTranscript = 1;
			}
			else
			    nReadsInTranscript++;
		    }
		}
	    }
	}
    	if (saveFileName!=null) {
	    try {
		System.out.println("Saving transcripts to " + saveFileName);
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName + TRANSCRIPT_SUFFIX + (forward ? "_f" : "_r") + ".txt"));
		for (NonGenicTranscript n : nonGenicTranscripts)
		    bw.write("chr" + n.toString() + "\n");
		bw.close();
	    }
	    catch (IOException e) {
		System.err.println("Error saving transcript data " + e.toString());
	    }
	}
	System.out.println("Found " + nIsolated + " isolated reads which were not part of a transcript and " + nonGenicTranscripts.size() + " transcripts with an average of " + format.format((double)sumReadsInTranscript/nonGenicTranscripts.size()));
	return nonGenicTranscripts;
    }

    public static int findMaxReadDistance(TreeMap<Integer, Integer> distanceDistribution, double fdr) {
	//Use this method to figure out what proportion of reads that
	//are further than a given distance from their
	//neighbors. First we need to normalize the distribution.
	int n = 0;
	for (int i : distanceDistribution.keySet()) 
	    n += distanceDistribution.get(i);
	//Now we can find out when the proportion of reads goes beyond the threshold
	int s = 0;
	Iterator it = distanceDistribution.keySet().iterator();
	while (it.hasNext() && (double)s/n<1-fdr)
	    s += distanceDistribution.get((Integer)it.next());
	return it.hasNext() ? (Integer)it.next() : 1 + (int)distanceDistribution.lastKey();
    }

    public static Vector<ABIRead[]> loadMRNAReads(String mRNAFileName, boolean forward, int nChr) { return loadMRNAReads(mRNAFileName, forward, nChr, true); }

    public static Vector<ABIRead[]> loadMRNAReads(String mRNAFileName, boolean forward, int nChr, boolean flatten) {
	Vector<ABIRead[]> reads = new Vector<ABIRead[]>();
	try {
	    int nReads[] = new int[nChr];
	    Arrays.fill(nReads, 0);
	    BufferedReader br = new BufferedReader(new FileReader(mRNAFileName));
	    String line = br.readLine();
	    Pattern mRNAPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([F|R])\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    if (chr<=nChr) { //22 is the splice chromosome
			if ((m.group(3).equals("F") && forward) || (m.group(3).equals("R") && !forward))
			    nReads[chr - 1] += flatten ? 1 : Integer.parseInt(m.group(4));
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    for (int i=0; i<nChr; i++) 
		reads.addElement(new ABIRead[nReads[i]]);
	    br = new BufferedReader(new FileReader(mRNAFileName));
	    line = br.readLine();
	    int i[] = new int[nChr];
	    Arrays.fill(i, 0);
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    if (chr<=nChr) { //22 is the splice chromosome
			if ((m.group(3).equals("F") && forward)) {
			    for (int j=0; j<(flatten ? 1 : Integer.parseInt(m.group(4))); j++)
				reads.elementAt(chr-1)[i[chr-1]++] = new ABIRead(chr, true, Integer.parseInt(m.group(2)), MRNA_READ_LENGTH);
			}
			else if (m.group(3).equals("R") && !forward) {
			    for (int j=0; j<(flatten ? 1 : Integer.parseInt(m.group(4))); j++)
				reads.elementAt(chr-1)[i[chr-1]++] = new ABIRead(chr, false, Integer.parseInt(m.group(2)) - MRNA_READ_LENGTH, MRNA_READ_LENGTH);
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading file " + mRNAFileName + " : " + e.toString());
	}
	return reads;
    }

    public static ABIRead[] loadMRNAReadsChr(String mRNAFileName, int c) { 
	ABIRead f[] = loadMRNAReadsChr(mRNAFileName, true, c, true); 
	ABIRead r[] = loadMRNAReadsChr(mRNAFileName, false, c, true); 
	ABIRead t[] = new ABIRead[f.length + r.length];
	for (int i=0; i<f.length; i++)
	    t[i] = f[i];
	for (int i=0; i<r.length; i++)
	    t[i + f.length] = r[i];
	Arrays.sort(t);
	return t;
    }
    
    public static Vector<ABIRead[]> groupRNAReads(ABIRead reads[], int sepDist) {
	//Use this method to split reads into groups when they are
	//separated by at least sepDist.
	Vector<ABIRead[]> readGroups = new Vector<ABIRead[]>();
	int startInd = 0, endInd, ind = 0;
	while (ind<reads.length-1) {
	    while (reads[ind+1].getPosition() - reads[ind].getPosition()<sepDist) 
		ind++;
	    endInd = ind;
	    ABIRead r[] = new ABIRead[endInd - startInd];
	    for (int i=startInd; i<endInd; i++)
		r[i-startInd] = reads[i];
	    readGroups.addElement(r);
	    startInd = endInd;
	}
	System.out.println("Found " + readGroups.size() + " separate groups.");
	return readGroups;
    }



    public static int[] loadMRNAReadsChrBin(String mRNAFileName, int binSize, boolean forward, int c) { return loadMRNAReadsChrBin(mRNAFileName, binSize, forward, c, true); }

    public static int[] loadMRNAReadsChrBin(String mRNAFileName, int binSize, boolean forward, int c, boolean flatten) {
	ABIRead reads[] = loadMRNAReadsChr(mRNAFileName, forward, c, flatten);
	int lastBinInd = (1 + reads[reads.length-1].getPosition()/binSize);
	int bins[] = new int[lastBinInd];
	for (int i=0; i<reads.length; i++)
	    bins[reads[i].getPosition()/binSize]++;
	return bins;
    }

    public static ABIRead[] loadMRNAReadsChr(String mRNAFileName, boolean forward, int c) { return loadMRNAReadsChr(mRNAFileName, forward, c, true); }
    
    public static ABIRead[] loadMRNAReadsChr(String mRNAFileName, boolean forward, int c, boolean flatten) {
	ABIRead[] reads = null;
	try {
	    int nReads = 0;
	    BufferedReader br = null;
	    try {
		br = new BufferedReader(new FileReader(mRNAFileName.replace(".tsv", "_" + c + ".tsv")));
	    }
	    catch (IOException ex) {
		br = new BufferedReader(new FileReader(mRNAFileName));
	    }
	    String line = br.readLine();
	    Pattern mRNAPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([F|R])\\s+(\\d+)\\s*(\\d*)");
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    if (chr==c) { //22 is the splice chromosome
			if ((m.group(3).equals("F") && forward) || (m.group(3).equals("R") && !forward))
			    nReads += flatten ? 1 : Math.min(Integer.parseInt(m.group(4)), MAX_READS_PER_LOCUS);
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    reads = new ABIRead[nReads];
	    try {
		br = new BufferedReader(new FileReader(mRNAFileName.replace(".tsv", "_" + c + ".tsv")));
	    }
	    catch (IOException ex) {
		br = new BufferedReader(new FileReader(mRNAFileName));
	    }
	    line = br.readLine();
	    int i = 0;
	    while (line!=null) {
		Matcher m = mRNAPattern.matcher(line);
		if (m.find()) {
		    int chr = Integer.parseInt(m.group(1));
		    if (chr==c) { //22 is the splice chromosome
			nReads = Math.min(Integer.parseInt(m.group(4)), MAX_READS_PER_LOCUS);
			if ((m.group(3).equals("F") && forward)) {
			    for (int j=0; j<(flatten ? 1 : nReads); j++) {
				if (m.group(5)!=null && !m.group(5).equals("") && Integer.parseInt(m.group(5))>0)
				    reads[i++] = new ABIRead(chr, true, Integer.parseInt(m.group(2)), Integer.parseInt(m.group(5))/nReads);
				else
				    reads[i++] = new ABIRead(chr, true, Integer.parseInt(m.group(2)), MRNA_READ_LENGTH);
			    }
			}
			else if (m.group(3).equals("R") && !forward) {
			    //System.out.println(line + "\t" + nReads);
			    for (int j=0; j<(flatten ? 1 : nReads); j++) {
				if (m.group(5)!=null && !m.group(5).equals("") && Integer.parseInt(m.group(5))>0)
				    reads[i++] = new ABIRead(chr, false, Integer.parseInt(m.group(2)) - Integer.parseInt(m.group(5))/nReads, Integer.parseInt(m.group(5))/nReads);
				else
				    reads[i++] = new ABIRead(chr, false, Integer.parseInt(m.group(2)) - MRNA_READ_LENGTH, MRNA_READ_LENGTH);
			    }
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading file " + mRNAFileName + " : " + e.toString());
	}
	return reads;
    }


    public static boolean geneInRegion(int start, int end, boolean forward, RefSeqGenes geneInfo) {
	//Returns true if there is an annotated gene between start and
	//end. Start and end are always given with respect to the
	//forward strand. This is different from the annotation data
	//where the start is alwasy the 5' end.
	return geneInfo.overlapsGene(start, end, forward);
	/*Iterator it = geneInfo.getGenes().iterator();
	while (it.hasNext()) {
	    RefSeqGene g = (RefSeqGene)it.next();
	    if (forward && g.getStrand()==forward) {
		if (g.getStart()>start && g.getStart()<end) //gene starts in region
		    return true;
		else if (g.getStart()<=start && g.getEnd()>start) //start internal
		    return true;
	    }
	    else if (!forward && g.getStrand()==forward) {
		if (g.getEnd()>start && g.getEnd()<end) //gene starts in region
		    return true;
		else if (g.getEnd()<=start && g.getStart()>start) //start internal
		    return true;
		    }
	    
	}
	return false;*/
    }

    public static TreeMap<Integer, Integer> mRNADistanceDistribution(String type, String mRNAKey, boolean forward, boolean orchestra, RefSeqGenes geneInfo[], Vector<ABIRead[]> mRNAs) {
	//Use this method to calculate the
	String fileStart = mRNAKey.substring(mRNAKey.lastIndexOf("/") + 1, mRNAKey.lastIndexOf(".") - 1);
	String saveFileName = (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/TKData/" : "") + "mRNA/NullDistributions/" + fileStart + "_" + type + "_distances_" + (forward ? "f" : "r") + ".txt"; //Try to open file with results
	TreeMap<Integer, Integer> distHist = new TreeMap<Integer, Integer>();
	try {
	    BufferedReader br = new BufferedReader(new FileReader(saveFileName));
	    String line = br.readLine();
	    Pattern pattern = Pattern.compile("(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = pattern.matcher(line);
		if (m.find()) 
		    distHist.put(new Integer(m.group(1)), new Integer(m.group(2)));
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Could not find file with results for distances, recomputing....");
	    for (int j=1; j<=21; j++) {
		//Map has pos as key and number of reads as key
		ABIRead mRNA[] = mRNAs.elementAt(j-1); //GeneratePeakmRNANullDist.readMRNAInfo(mRNAKey, forward, j);
		//System.out.println("Found results for chr " + j);
		int nReads = 0, nBins = 0;
		if (type.equals("nongenic")) {
		    //Find all reads in non-genic regions. We find the
		    //regions where there are no genes and study the
		    //subset of of reads which are in these regions.
		    int nonGenicStart = 0, nonGenicEnd = 0;
		    Iterator it = geneInfo[j-1].getGenes().iterator();
		    RefSeqGene g = null;
		    while (it.hasNext()) {
			g = (RefSeqGene)it.next();
			if (g.getStrand()==forward) {
			    nonGenicEnd = forward ? g.getStart() : g.getEnd();
			    break;
			}
		    }
		    //Get all the reads in the non-genic region
		    //and find their distances
		    int prevMRNAStart = mRNA[0].getPosition();
		    for (int i=1; i<mRNA.length; i++) {
			int mRNAStart = mRNA[i].getPosition();
			if (prevMRNAStart>nonGenicStart && mRNAStart<nonGenicEnd) {
			    int dist = mRNAStart - prevMRNAStart;
			    distHist.put(dist, 1 + (distHist.containsKey(dist) ? distHist.get(dist) : 0));
			}
			else if (g!=null) {
			    nonGenicStart = forward ? g.getEnd() : g.getStart();
			    while (it.hasNext()) {
				g = (RefSeqGene)it.next();
				if (g.getStrand()==forward)
				    nonGenicEnd = g.getStrand() ? g.getStart() : g.getEnd();
				else
				    g = null;
			    }
			    if (g==null) //reached last gene
				nonGenicEnd = PeakFinder2.CHROMOSOME_LENGTH[j - 1];
			}
			prevMRNAStart = mRNAStart;
		    }
		}
	    }	    
	    System.err.println("Saving results to " + saveFileName);
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		Iterator it = distHist.keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    bw.write(key.toString() + "\t" + ((Integer)distHist.get(key)).toString() + "\n");
		}
		bw.close();
	    }
	    catch (IOException ex) {
		System.err.println("Error saving results to " + saveFileName);
	    }
	}
	return distHist;
    }

    public static int[][] findNonGenicRegions(int chr, boolean strands[], int tsss[], int tses[], boolean forward) {
	int nGenes = 0;
	for (int i=0; i<strands.length; i++) {
	    if (strands[i]==forward) { nGenes++; }
	}
	int genic[][] = new int[nGenes][2], j = 0;
	for (int i=0; i<strands.length; i++) {
	    if (strands[i]==forward) { 
		genic[j][0] = tsss[i];
		genic[j++][1] = tses[i];
	    }
	}
	int nonGenic[][] = new int[nGenes + 2][2];
	nonGenic[0][0] = 0;
	nonGenic[0][1] = genic[0][0];
	for (int i=0; i<genic.length-1; i++) {
	    nonGenic[i+1][0] = genic[i][1] + 1;
	    nonGenic[i+1][1] = genic[i+1][0];
	}
	nonGenic[nonGenic.length-1][0] = genic[genic.length-1][1] + 1;
	nonGenic[nonGenic.length-1][1] = PeakFinder2.CHROMOSOME_LENGTH[chr-1];
	return nonGenic;
    }

}
