//use to detect transcripts using the wavelet based method. The
//current version only contains three parameters: two length scales
//and a cut-off for the number of break points and what densities to
//include. The last parameter is similar to an FDR type cut-off and to
//sweep for different values, we may use:

//java -Xmx2548m FindTranscripts -c 1 -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr6c  -l 10 -L 18 -R 35 --merge --minDens .01 --fdr .001 --fdrMax .2 --nfdr 10 -o --peakFileName ../MmDataPeakFiles/Pol2_8WG16_KCl_B1_names --gmm --log

//It seems as if the best value when maximimizing read inclusion and
//minimizing genome coverage is fdr = .025.

//To compare the obtained transcripts against the annotation

//java -Xmx2048m FindTranscripts -c 1 -n /groups/neuroduo/martin/ChIP-Seq/TKData/mRNA/hr6m  -l 9 -L 16 -R 35 --merge --fdr .025 --compare .1 --compareMax .9 --ncompare 9

//To look at human HeLa cells, we might want to do:

//java -Xmx2048m FindTranscripts -c 18 -n /groups/neuroduo/martin/ChIP-Seq/HeLa/mRNA/HeLac  -l 9 -L 16 -R 35 --merge --fdr .025 -o --species hs

//For the yeast:
//
//java -Xmx1024m FindTranscripts -c 1 -n ../../SacCer/mRNA/c1_shot --species sc -L 12 -l 6 --merge --fdr .001 --fdrMax .1 --nfdr 10 --minDens .01 --gmm -R 50

//For the Bayesian method
//
//java -Xmx1024m FindTranscripts -c 19 -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr6c -b 9 -K 50 -R 35 -l 14 --bayes
//
//bsub -q shared_2h java -Xmx1024m FindTranscripts -c 19 -n /groups/neuroduo/martin/ChIP-Seq/MmData/MeCP2/MeCP2/30KI_Joe/B1/MeCP2_MeCP2_30KI_Joe_B1 -i /groups/neuroduo/martin/ChIP-Seq/MmData/Input/Input/30KI_Joe/B1/Input_Input_30KI_Joe_B1 -b 8 -K 50 -R 35 -l 10 --bayes --chip --log 2 --gmm 2
//
//For the Plasmodium stuff:
//
//bsub -q shared_2h java -Xmx1024m FindTranscripts -c 14 -n /groups/neuroduo/martin/ChIP-Seq/Plasmodium/mRNA/pf5-RNA-Seq-ET.bed -b 2 -R 35 -l 4 -L 12 --log 1 --gmm 2 -o --species pf --cost coverage


import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.TallyStore;

public class FindTranscripts {

    public static final String BED_FILE_SUFFIX = "_transcripts";
    public static final String BED_FILE_SUFFIX_CHIP = "_regions";
    public static final String LATEX_TABLE_FILE_NAME_SUFFIX = "_summary_stats.tex";
    public static final String TRANSCRIPT_LENGTHS_SUFFIX = "_lengths";
    public static final String TRANSCRIPT_DENSITIES_SUFFIX = "_dens";
    public static final String TRANSCRIPT_WITHIN_DENISTY_STD_DEV_SUFFIX = "_densStdDev";
    public static final String TRANSCRIPT_PEAKS_SUFFIX = "_peaks";
    public static final String TRANSCRIPT_CONSERVATION_SUFFIX = "_cons";
    public static final String TRANSCRIPT_WITHIN_CONSERVATION_STD_DEV_SUFFIX = "_consStdDev";

    
    public static final String WWW_DIR_NAME = "neuroduo.hms.harvard.edu/docroot/bigWig/mh/Transcripts/";
    public static final String URL_NAME = "http://neuroduo.hms.harvard.edu/bigWig/mh/Transcripts/";
    public static final String TRACK_INFO_FILE_PATH = "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/Transcripts/";
    public static final String TRACK_INFO_FILE_NAME = "track_info.txt";

    //Location of the files which have parsed DH's huge files. 
    public static final String ID_EXPRESSION_FILE_LOCATION = "mRNA/SymbolicDynamics/symbolic_dynamics_xpt.txt";
    //Pattern for parsing, columns are ID name rank0 rank1 rank6 exp0 exp1 exp6
    public static final Pattern ID_EXPRESSION_FILE_PATTERN = Pattern.compile("(\\d+)\\s+(\\w+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+([\\d|\\.]+)\\s+([\\d|\\.]+)\\s+([\\d|\\.]+)");

    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option readLengthFlag = argParser.addIntegerOption('R', "readLength");	    
	    CmdLineParser.Option strandFlag = argParser.addBooleanOption('r', "reverseStrand");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option logFlag = argParser.addIntegerOption("log");
	    CmdLineParser.Option gmmFlag = argParser.addIntegerOption("gmm");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption("conservation");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option sequenceFlag = argParser.addBooleanOption("sequence");
	    CmdLineParser.Option cpcFlag = argParser.addBooleanOption("CPC");
	    //Parameters for the transcript finding
	    CmdLineParser.Option lMinFlag = argParser.addIntegerOption('l', "lMin");
	    CmdLineParser.Option lMaxFlag = argParser.addIntegerOption('L', "lMax");
	    //For coarse-graining
	    CmdLineParser.Option binSizeFlag = argParser.addIntegerOption('b', "binSize");
	    CmdLineParser.Option deltaLFlag = argParser.addIntegerOption("deltaL");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption("fdr");
	    CmdLineParser.Option fdrMaxFlag = argParser.addDoubleOption("fdrMax");
	    CmdLineParser.Option nfdrFlag = argParser.addIntegerOption("nfdr");
	    CmdLineParser.Option nIterationsFlag = argParser.addIntegerOption("nIterations");
	    CmdLineParser.Option bicFlag = argParser.addBooleanOption("bic");
	    CmdLineParser.Option mergeFlag = argParser.addBooleanOption("merge");
	    CmdLineParser.Option flattenFlag = argParser.addBooleanOption("flatten");
	    //include the compare highest expressed genes when testing
	    //how well the transcribed regions match the annotation.
	    CmdLineParser.Option compareFlag = argParser.addDoubleOption("compare");
	    CmdLineParser.Option compareMaxFlag = argParser.addDoubleOption("compareMax");
	    CmdLineParser.Option minDensFlag = argParser.addDoubleOption("minDens");
	    CmdLineParser.Option maxDensFlag = argParser.addDoubleOption("maxDens");
	    CmdLineParser.Option nDensFlag = argParser.addIntegerOption("ndens");
	    CmdLineParser.Option nCompareFlag = argParser.addIntegerOption("ncompare");
	    CmdLineParser.Option costFunctionFlag = argParser.addStringOption("cost");
	    //Use these flags when we want to down-sample the total number of reads to investigate how the optimal parameters change

	    //Use this to indicate a peak file, eg Pol2 or H3K4Me3
	    //that we wish to compare with for the optimization
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption("peakFileName");
	    //Use this flag to provide a set of peaks that are to be
	    //used for spiking the set of breakpoints, eg Pol2 peak
	    //locations
	    CmdLineParser.Option breakPointsFileNameFlag = argParser.addStringOption("breakPointsFileName");
	    //Use this flag to save information about intermediate steps of the algorithm
	    CmdLineParser.Option saveInfoFileNameFlag = argParser.addBooleanOption("saveInfo");
	    //Use these flags for ChIP data. We need to have a
	    //fragment length and possibly a negative control data set
	    //(if available)
	    CmdLineParser.Option chipFlag = argParser.addBooleanOption("chip");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('f', "fragmentLength");
	    //Use this flag to save the .bed files for the raw data as well
	    CmdLineParser.Option dataBedFlag = argParser.addBooleanOption("bed");
	    //Use this flag to indicate that we want to use the Bayesian method
	    CmdLineParser.Option bayesFlag = argParser.addBooleanOption("bayes");
	    CmdLineParser.Option nSegmentsMaxFlag = argParser.addIntegerOption('K', "nSegmentsMax");
	    //Decide what error model to use
	    CmdLineParser.Option errorModelFlag = argParser.addStringOption("errorModel");
	    //For comparing against known elements
	    CmdLineParser.Option enhancerFileNameFlag = argParser.addStringOption("enhancers");
	    CmdLineParser.Option lncRNAFileNameFlag = argParser.addStringOption("lncRNA");
	    CmdLineParser.Option macroRNAFileNameFlag = argParser.addStringOption("macroRNA");
	    CmdLineParser.Option miRNAFileNameFlag = argParser.addStringOption("miRNA");
	    CmdLineParser.Option miRNAFileNamePrecursorsFlag = argParser.addStringOption("miRNAPrecursors");
	    CmdLineParser.Option maxDivergentTSSDistanceFlag = argParser.addIntegerOption("maxDivergent");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean chip = argParser.getOptionValue(chipFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    boolean sequence = argParser.getOptionValue(sequenceFlag)!=null; 
	    boolean cpc = argParser.getOptionValue(cpcFlag)!=null; 
	    tmp = argParser.getOptionValue(gmmFlag);
	    int nComponents = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    boolean gmm = nComponents>1;
	    boolean reverseStrand = argParser.getOptionValue(strandFlag)!=null;
	    boolean bayes = argParser.getOptionValue(bayesFlag)!=null;
	    tmp = argParser.getOptionValue(errorModelFlag);
	    String errorModel = tmp!=null ? (String)tmp : TranscriptFinder.GAUSSIAN_ERROR;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp + (chip || ((String)tmp).contains(".") ? "" : ".tsv") : "";
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(breakPointsFileNameFlag);
	    String breakPointsFileName = tmp!=null ? (String)tmp : null;
	    boolean saveInfo = argParser.getOptionValue(saveInfoFileNameFlag)!=null;
	    String saveInfoFileName = saveInfo ? (chip ? fileName : fileName.substring(0, fileName.lastIndexOf(".")) + (reverseStrand ? "_r" : "_f") ) : null;
	    String latexTableFileName = saveInfoFileName + LATEX_TABLE_FILE_NAME_SUFFIX;
	    tmp = argParser.getOptionValue(costFunctionFlag);
	    String costFunction = tmp!=null ? ((String)tmp).toLowerCase() : "sens";
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : (chip && !fileName.contains(".") ? fileName.substring(0, fileName.lastIndexOf("_")) : fileName.substring(0, fileName.lastIndexOf(".")));
	    tmp = argParser.getOptionValue(logFlag);
	    int nLogs = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    boolean log = nLogs>0;
	    tmp = argParser.getOptionValue(readLengthFlag);
	    int readLength = tmp!=null ? ((Integer)tmp).intValue() : 35;
	    tmp = argParser.getOptionValue(lMinFlag);
	    int lMin = tmp!=null ? ((Integer)tmp).intValue() : 5;
	    tmp = argParser.getOptionValue(lMaxFlag);
	    int lMax = tmp!=null ? ((Integer)tmp).intValue() : 15;
	    tmp = argParser.getOptionValue(binSizeFlag);
	    int binSize = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(deltaLFlag);
	    int deltaL = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .01;
	    tmp = argParser.getOptionValue(fdrMaxFlag);
	    double fdrMax = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(nfdrFlag);
	    int nfdr = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(nIterationsFlag);
	    int nIterations = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(nSegmentsMaxFlag);
	    int nSegmentsMax = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    //String penalty = TranscriptFinder.AIC_STR;
	    //penalty = argParser.getOptionValue(bicFlag)!=null ? TranscriptFinder.BIC_STR : penalty;
	    tmp = argParser.getOptionValue(enhancerFileNameFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_lMe3_real.tsv";
	    tmp = argParser.getOptionValue(lncRNAFileNameFlag);
	    String lncRNAFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/landerRegions.bed";
	    tmp = argParser.getOptionValue(macroRNAFileNameFlag);
	    String macroRNAFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/macroRNA.txt";
	    tmp = argParser.getOptionValue(miRNAFileNameFlag);
	    String miRNAFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/miRNA.txt";
	    tmp = argParser.getOptionValue(miRNAFileNamePrecursorsFlag);
	    String miRNAFileNamePrecursors = null; //tmp!=null ? (String)tmp : (orchestra ? "/home/mh168" : "/home/mhemberg" ) + "/Projects/ChIP-Seq/TKData/mRNA/miRNA/landgrafTuschl_2007_C_mm_miRNA_precursors_names.txt";
	    tmp = argParser.getOptionValue(maxDivergentTSSDistanceFlag);
	    int maxDivergentTSSDistance = tmp!=null ? ((Integer)tmp).intValue() : 2000;
	    boolean mergeSegments = argParser.getOptionValue(mergeFlag)!=null;
	    boolean flatten = argParser.getOptionValue(flattenFlag)!=null;
	    boolean bed = argParser.getOptionValue(dataBedFlag)!=null;
	    tmp = argParser.getOptionValue(compareFlag);
	    double compare = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    tmp = argParser.getOptionValue(compareMaxFlag);
	    double compareMax = tmp!=null ? ((Double)tmp).doubleValue() : 1;
	    tmp = argParser.getOptionValue(minDensFlag);
	    double minDens = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    tmp = argParser.getOptionValue(maxDensFlag);
	    double maxDens = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    tmp = argParser.getOptionValue(nDensFlag);
	    int nDens = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(nCompareFlag);
	    int nCompare = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    //////////////////////////////////////////////////////////////////
	    int nChr = numberOfChromosomes(species) - 1;
	    int lVec[] = getLengthVec(lMin, lMax, deltaL), lVecCoeffs[] = getLengthVec(binSize, lMax, deltaL);
	    double fdrVec[] = getFDRVec(fdr, fdrMax, nfdr);
	    double densVec[] = getFDRVec(minDens, maxDens, nDens);
	    double compareVec[] = species.equals("mm") ? getCompareVec(compare, compareMax, nCompare) : null;
	    Pattern rnaFileNamePattern = Pattern.compile("hr(\\d)\\w");
	    Matcher m = rnaFileNamePattern.matcher(fileName);
	    int time = m.find() ? Integer.parseInt(m.group(1)) : 6;
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    TranscriptFinder tf[] = new TranscriptFinder[nChr];
	    ReducedModel reducedModel[] = new ReducedModel[nChr];
	    //Load the enhancers
	    TreeSet<Enhancer2> enhancers = parseEnhancerFile(enhancerFileName, chr);
	    //Load the lncRNAs
	    TreeSet<LncRNA> lncRNAs = AnalyzeTranscriptReadProfiles.loadLncRNA(lncRNAFileName, species);
	    TreeSet<MacroRNA> macroRNAs = AnalyzeTranscriptReadProfiles.loadMacroRNA(macroRNAFileName, species);
	    //Load the miRNAs
	    TreeSet<MiRNA> miRNAs = AnalyzeTranscriptReadProfiles.loadMiRNA(miRNAFileName, miRNAFileNamePrecursors, species);
	    //Load the peaks
	    Vector<Peak> peaks = peakFileName==null ? null : new Vector<Peak>(FindEnhancerSites.parsePeakFile(peakFileName, chr).values());
	    Vector<Peak> breakPointPeaks = breakPointsFileName==null ? peaks : new Vector<Peak>(FindEnhancerSites.parsePeakFile(breakPointsFileName, chr).values());
	    System.out.println("FDR\tn\ttranscribed\treads\tPPV\tSensitivity\t#unique\t#missed_g\t#g_spanned_by_multiple_t\t#non-coding_t\t#t_spanning_mutliple_g\tPeak_overlap");
	    for (int i=1; i<=nChr; i++) {
		if (i==chr || chr==-1) {
		    if (chip) {
			//ABIRead[] rnaReads = ABIReads.loadMRNAReadsChr(fileName, true, i, readLength, false);
			geneInfo[i-1] = new RefSeqGenes(i, orchestra, species);
			//geneInfo[i-1].setFactorReads(fileName, rnaReads, 1);
			//AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, fileName, rnaReads, i, 0);
			//AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, fileName, rnaReads, i);
			/*if (fileName.contains("Plasmodium"))
			    tf[i-1] = new TranscriptFinder(rnaReads, i, !reverseStrand, readLength, species, log);
			else
			tf[i-1] = new ChIPRegionFinder(rnaReads, i, fragmentLength, species, log);*/
		    }
		    else {
			ABIRead[] rnaReadsForward = ABIReads.loadMRNAReadsChr(fileName, true, i, readLength, false);
			ABIRead[] rnaReadsReverse = ABIReads.loadMRNAReadsChr(fileName, false, i, readLength, false);
			geneInfo[i-1] = new RefSeqGenes(i, orchestra, species);
			geneInfo[i-1].setFactorReads(fileName, rnaReadsForward, 1);
			AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, fileName, rnaReadsForward, i, 1);
			AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, rnaReadsForward, i, 1);
			AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, fileName, rnaReadsForward, i);
			geneInfo[i-1].setFactorReads(fileName, rnaReadsReverse, -1);
			AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, fileName, rnaReadsReverse, i, -1);
			AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, rnaReadsReverse, i, -1);
			AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, fileName, rnaReadsReverse, i);
			tf[i-1] = new TranscriptFinder(reverseStrand ? rnaReadsReverse : rnaReadsForward, i, !reverseStrand, readLength, species, log);
		    }
		    if (bed)
			ABIReads.saveBedFile(fileName.substring(0, fileName.lastIndexOf(".")) + (reverseStrand ? "_r" : "_f") + "_" + chr + ".bed", tf[i-1].getReads(), reverseStrand, true, false, fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")) + (reverseStrand ? "_r" : "_f"), species);
		    if (bayes)
			tf[i-1].calculateBayesianPiecewiseConstantRegression((int)Math.pow(2, lMin), binSize, nSegmentsMax, nLogs, errorModel, nComponents, geneInfo[i-1]);
		    else
			tf[i-1].calculateHaarCoefficients(lVecCoeffs, binSize, saveInfoFileName);
		    String saveDensInfo = saveFileName + "_dens" + (reverseStrand ? "_r" : "_f") + (chr==-1 ? "" : "_" + i) + ".txt";
		    if (densVec.length>1)
			System.out.println("Saving density info to " + saveDensInfo);
		    BufferedWriter bw = densVec.length>1 ? new BufferedWriter(new FileWriter(saveDensInfo)) : null;
		    if (bw!=null)
			bw.write("dens\tFDR\tn\ttranscribed\treads\tPPV\t#unique\t#missed_g\t#g_spanned_by_multiple_t\t#non-coding_t\t#t_spanning_mutliple_g\tPeak_overlap\n");
		    for (int k=0; k<densVec.length; k++) {
			geneInfo[i-1].setGeneIDsKeep(true, fileName, minDens);
			geneInfo[i-1].setGeneIDsKeep(false, fileName, minDens);
			if (!bayes) {
			    for (int j=0; j<fdrVec.length; j++) {
				int bestIterations = tf[i-1].findRefinedReducedModel(lMin, lMax, binSize, fdrVec[j], mergeSegments, gmm, saveInfoFileName, geneInfo[i-1], breakPointPeaks, peaks, maxDivergentTSSDistance, costFunction, nIterations);
				reducedModel[i-1] = tf[i-1].getReducedModel();
				//reducedModel[i-1].calculateLogEMALikelihood(log);
				//reducedModel[i-1].compare2Annotation(maxDivergentTSSDistance, geneInfo[i-1], lncRNAs, miRNAs, macroRNAs, enhancers);
				//reducedModel[i-1].categorizeTranscribedRegions(false);
				//Find out how many of the transcripts have a peak near their start
				/*TreeMap<TranscribedRegion, Peak> transcriptsWithPeaks = peaks==null ? new TreeMap<TranscribedRegion, Peak>() : reducedModel[i-1].compare2Peaks(peaks, maxDivergentTSSDistance);
				if (bw!=null)
				    bw.write(format.format(densVec[k]) + "\t" + printStats(fdrVec[j], reducedModel[i-1], transcriptsWithPeaks, peaks) + "\n");
				else
				    printStats(fdrVec[j], reducedModel[i-1], transcriptsWithPeaks, peaks);
				if (verbose) {
				    reducedModel[i-1].findGenesPerTranscript();
				    reducedModel[i-1].findTranscriptsPerGene();
				}
				reducedModel[i-1].removeGenes();*/
			    }
			}
			else {
			    reducedModel[i-1] = tf[i-1].findReducedModel(mergeSegments);
			    System.out.println("Found " + reducedModel[i-1].numberOfTranscribedRegions()); //reducedModel.toString());
			    /*double minDetectedDens = reducedModel[i-1].minTranscribedRegionDensity(log); 
			    reducedModel[i-1].compare2Annotation(geneInfo[i-1]);
			    reducedModel[i-1].categorizeTranscribedRegions(false);
			    //printTypeStats(reducedModel[i-1], tf[i-1].getReads().length, null, lncRNAs);
			    //Find out how many of the transcripts have a peak near their start
			    TreeMap<TranscribedRegion, Peak> transcriptsWithPeaks = peaks==null ? new TreeMap<TranscribedRegion, Peak>() : reducedModel[i-1].compare2Peaks(peaks, maxDivergentTSSDistance);*/
			    //printStats(fdrVec[j], reducedModel[i-1], geneIDsKeep, transcriptsWithPeaks);
			}
		    }
		    if (saveFileName!=null && fdrVec.length<=1) {
			String saveFile = saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + (chr!=-1 ? "_" + chr : "") + ".bed";
			//reducedModel[i-1].saveReducedModel(saveFile, (chr==1 || chr==-1) && fileName.contains("Plasmodium") ? saveFile.substring(saveFile.lastIndexOf("/") + 1, saveFile.lastIndexOf(".")) : null, chr==-1 && i>1);
			if (orchestra && chr!=-1 && !fileName.contains("Plasmodium")) //Put it in a directory on the www-site for easy upload to the genome browser
			    createWigFile(saveFileName, chip, reverseStrand, chr, bayes);
			//Save the lengths of the different transcripts
			saveTranscriptProperties(reducedModel[i-1], binSize, peaks!=null ? peaks.firstElement().getFactorNames(false) : null, orchestra, saveFileName, reverseStrand, chr, conservation, bayes, chip);			
			if (sequence) { //Save the sequences for each transcript
			    Runtime runtime = cpc ? Runtime.getRuntime() : null;
			    for (int j=0; j<TranscribedRegion.TYPES_ARRAY_STR.length; j++) { 
				//get one file for each category. This will allow us to split up the jobs for CPC speeding the whole thing up significantly
				String seqSaveFileName = saveFileName + "_transcripts_" + TranscribedRegion.TYPES_ARRAY_STR[j] + "_" + (reverseStrand ? "r" : "f") + (chr==-1 ? "" : "_" + chr) + ".fa";
				//reducedModel[i-1].saveSequences(orchestra, seqSaveFileName, TranscribedRegion.TYPES_ARRAY_STR[j], false);
				if (cpc) { //Run CPC to analyze if the sequences are
				    String cpcOutDir = seqSaveFileName.substring(0, seqSaveFileName.lastIndexOf("/") + 1) + "CPC/" + (reverseStrand ? "r" : "f") + "_" + chr + "/" + TranscribedRegion.TYPES_ARRAY_STR[j] + "/";
				    (new File(cpcOutDir)).mkdirs();
				    String cpcCmd = "bsub -o /dev/null -q all_7d ./run_cpc.sh " + seqSaveFileName + " " + cpcOutDir + seqSaveFileName.substring(seqSaveFileName.lastIndexOf("/") + 1, seqSaveFileName.lastIndexOf(".")) + " " + cpcOutDir;
				    System.out.println("Running CPC:\n" + cpcCmd);
				    Process cpcProcess = runtime.exec(cpcCmd);
				    //cpcProcess.waitFor();
				}
			    }
			}
			//Save statistics for the genes. Since the
			//number of classes of genes is fixed (unique,
			//split, multi and missed) we use the opposite
			//scheme than for the transcripts when saving
			//and use one file for each category with the
			//columns corresponding to properties instead
			//of the other way around.
			if (geneInfo[i-1].size()>0) {
			    //reducedModel[i - 1].saveGeneInfo(saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + "_" + RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
			    //reducedModel[i - 1].saveGeneInfo(saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + "_" + RefSeqGene.MERGED_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.MERGED_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
			    //reducedModel[i - 1].saveGeneInfo(saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + "_" + RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
			    //reducedModel[i - 1].saveGeneInfo(saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + "_" + RefSeqGene.MISSED_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.MISSED_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
			}
		    }
		    if (bw!=null)
			bw.close();
		    tf[i-1] = null;
		    reducedModel[i-1] = null;
		    geneInfo[i-1] = null;
		}
	    }
	    ///////////////////////////////////////////////////////
	    //Save stuff, mosly used for yeast and other small genomes
	    //where we don't have to split jobs by chromosome
	    if (orchestra && chr==-1) //Put it in a directory on the www-site for easy upload to the genome browser
		createWigFile(saveFileName, chip, reverseStrand, chr, bayes);
	    /*if (saveFileName!=null && fdrVec.length<=1 && chr==-1) {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + (chr!=-1 ? "_" + chr : "") + ".bed"));
		for (int i=1; i<nChr; i++) {
		    bw.write(reducedModel[i-1].getBedFileString(reverseStrand, i==1 ? saveFileName.substring(saveFileName.lastIndexOf("/") + 1) + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) : null, species));
		    System.out.println("Saving " + reducedModel[i-1].numberOfTranscribedRegions() + " for chr " + i);
		}
		bw.close();
		}*/
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static void printTypeStats(ReducedModel r, int totReads) { printTypeStats(r, totReads, null, null); }

    public static void printTypeStats(ReducedModel r, int totReads, String latexTableFileName) { printTypeStats(r, totReads, latexTableFileName, null); }

    public static void printTypeStats(ReducedModel r, int totReads, String latexTableFileName, TreeSet<LncRNA> lncRNAs) {
	TreeMap<String, Integer> typeCount = new TreeMap<String, Integer>();
	int nGenesMerged = 0;
	TreeSet<RefSeqGene> genesSplit = new TreeSet<RefSeqGene>();
	/*for (TranscribedRegion t : r.getSegments()) {
	    typeCount.put(t.getType(), typeCount.containsKey(t.getType()) ? 1 + typeCount.get(t.getType()) : 1);
	    if (t.getType().equals(TranscribedRegion.CODING_FRAGMENT_STR)) 
		genesSplit.add((RefSeqGene)t.getNeighbors().iterator().next());
	    else if (t.getType().equals(TranscribedRegion.MULTI_GENE_STR)) 
		nGenesMerged += t.inDegree();
		}*/
	/*TreeMap<String, TallyStore> segmentTypeLengths = r.getTranscribedRegionTypesLengths();
	TreeMap<String, Integer> segmentTypeReads = r.getTranscribedRegionTypesReads();
	System.out.println("Found " + r.numberOfTranscribedRegions() + " with " + typeCount.size() + " different types.");
	try {
	    BufferedWriter bw = latexTableFileName!=null ? new BufferedWriter(new FileWriter(latexTableFileName)) : null;
	    if (bw!=null) //write a header
		bw.write("\\begin{tabular}{ccccc}\n\t\\textbf{Category} & \\textbf{#} & \\textbf{$<\\log_2$L$>$} & \\textbf{Genome} & \\textbf{Reads} \\\\\n\t\\hline\n");
	    for (String t : typeCount.keySet()) {
		System.out.print(t + "\t" + typeCount.get(t) + "\t" + format.format(segmentTypeLengths.get(t).average()) + "\t" + format.format((double)segmentTypeReads.get(t)/totReads));
		if (bw!=null) 
		bw.write(t + "\t" + typeCount.get(t) + "\t" + format.format(segmentTypeLengths.get(t).average()) + "\t" + format.format((double)segmentTypeReads.get(t)/totReads) + "\n");
		if (t.equals(TranscribedRegion.MULTI_GENE_STR))
		    System.out.print("\t" + nGenesMerged);
		if (t.equals(TranscribedRegion.CODING_FRAGMENT_STR))
		    System.out.print("\t" + genesSplit.size());
		System.out.print("\n");
	    }
	    if (bw!=null) {
		bw.write("\\end{tabular}\n"); //footer
		bw.close();
	    }
	}
	catch (IOException e) {
	    System.err.println("Error saving latex table " + e.toString());
	}
	//Give more info about the eRNAs and the lncRNAs
	for (TranscribedRegion t : r.getSegments()) {
	    if (t.getType().equals(TranscribedRegion.LNC_RNA_STR))
		System.out.println(t.getType() + ":\t" + t.toString() + "\t" + GeneTranscriptsGraph.overlappingLncRNA(r.getChromosome(), lncRNAs, t).toString());
	    if (t.getType().equals(TranscribedRegion.ERNA_STR))
		System.out.println(t.getType() + ":\t" + t.toString());
	    else if (t.getType().equals(TranscribedRegion.NON_CODING_AS_STR)) {
		RefSeqGene g = (RefSeqGene)t.getNeighbors().iterator().next();
		System.out.println(t.getType() + ":\t" + t.toString() + ", " + g.toString() + ", overlap = " + format.format((double)GeneTranscriptsGraph.overlap(g, t)/g.getLength()));
	    }
	    else if (t.getType().equals(TranscribedRegion.CODING_STR)) {
		RefSeqGene g = (RefSeqGene)t.getNeighbors().iterator().next();
		System.out.println(t.getType() + ":\t" + t.toString() + ", " + g.toString() + ", overlap = " + format.format((double)GeneTranscriptsGraph.overlap(g, t)/g.getLength()));
	    }
	    else if (t.getType().equals(TranscribedRegion.CODING_FRAGMENT_STR)) {
		RefSeqGene g = (RefSeqGene)t.getNeighbors().iterator().next();
		System.out.println(t.getType() + ":\t" + t.toString() + ", " + g.toString() + ", overlap = " + format.format((double)GeneTranscriptsGraph.overlap(g, t)/g.getLength()));
	    }
	    else if (t.getType().equals(TranscribedRegion.DIVERGENT_STR) || t.getType().equals(TranscribedRegion.DIVERGENT_END_STR)) {
		RefSeqGene g = (RefSeqGene)t.getNeighbors().iterator().next();
		System.out.println(t.getType() + ":\t" + t.toString() + ", " + g.toString() + ", overlap = " + format.format((double)GeneTranscriptsGraph.overlap(g, t)/g.getLength()));
	    }
	    else if (t.getType().equals(TranscribedRegion.MULTI_GENE_STR))
		System.out.println(t.getType() + ":\t" + t.toString() + "\t" + t.inDegree());
	    else if (t.getType().equals(TranscribedRegion.NON_CODING_OTHER_RNA_STR))
		System.out.println(t.getType() + ":\t" + t.toString());
		}*/
    }

    public static void saveLatexTable(String latexTableFileName, int totReads, TreeMap<String, Integer> typeCount, TreeMap<String, TallyStore> segmentTypeLengths, TreeMap<String, Integer> segmentTypePol2Peaks, TreeMap<String, Integer> segmentTypeReads, int nSplit, int nMerged) {
	try {
	    System.out.println("Saving summary statistics for all chromosomes to " + latexTableFileName);
	    BufferedWriter bw = latexTableFileName!=null ? new BufferedWriter(new FileWriter(latexTableFileName)) : null;
	    if (bw!=null) //write a header
		bw.write("\\begin{tabular}{lrcrcrr}\n\t\\textbf{Category} & \\multicolumn{2}{c}{\\textbf{\\# transcripts}} & \\multicolumn{2}{c}{\\textbf{RNAPII}} & \\textbf{Genome} & \\textbf{Reads} \\\\\n\t\\hline\\\\\n");
	    double lenSum = 0, readsSum = 0, nPol2 = 0;
	    int nTranscripts = 0;
	    for (String t : typeCount.keySet()) 
		nTranscripts += typeCount.get(t);
	    for (String t : typeCount.keySet()) {
		int nPeaks = segmentTypePol2Peaks.containsKey(t) ? segmentTypePol2Peaks.get(t) : 0;
		System.out.print(t + "\t" + typeCount.get(t) + "\t" + nPeaks + "\t" + format.format((double)segmentTypeReads.get(t)/totReads));
		if (bw!=null) {
		    bw.write(t.replace("_", " "));
		    if (t.equals(TranscribedRegion.MULTI_GENE_STR))
			bw.write(" (" + nMerged + ") ");
		    else if (t.equals(TranscribedRegion.CODING_FRAGMENT_STR))
			bw.write(" (" + nSplit + ") ");
		    bw.write(" & " + typeCount.get(t) + " & (" + format.format((double)typeCount.get(t)/nTranscripts) + ") & " + nPeaks + " & (" + format.format((double)nPeaks/typeCount.get(t)) + ") & " + format.format(((double)segmentTypeLengths.get(t).average()*segmentTypeLengths.get(t).numberObs())/RefSeqGenes.TOTAL_CHROMOSOME_LENGTH) + " & " + format.format((double)segmentTypeReads.get(t)/totReads) + "\\\\\n");
		}
		lenSum += (double)segmentTypeLengths.get(t).average()*segmentTypeLengths.get(t).numberObs();
		readsSum += (double)segmentTypeReads.get(t)/totReads;
		nPol2 += nPeaks;
		/*if (t.equals(TranscribedRegion.MULTI_GENE_STR))
		    System.out.print("\t" + nGenesMerged);
		if (t.equals(TranscribedRegion.CODING_FRAGMENT_STR))
		System.out.print("\t" + genesSplit.size());*/
		System.out.print("\n");
	    }
	    if (bw!=null) {
		bw.write("\\hline\n");
		bw.write("Total & " + nTranscripts + " & & " + nPol2 + " & & " + format.format(lenSum/RefSeqGenes.TOTAL_CHROMOSOME_LENGTH) + " & " + format.format(readsSum) + "\n");
		bw.write("\\end{tabular}\n"); //footer
		bw.close();
	    }
	}
	catch (IOException e) {
	    System.err.println("Error saving latex table " + e.toString());
	}
    }

    public static void saveLatexTable(String latexTableFileName, int totReads, int[] typeCount, int[] segmentTypeTotLengths, int[] segmentTypePol2Peaks, int[] segmentTypeTotReads) { saveLatexTable(latexTableFileName, totReads, typeCount, segmentTypeTotLengths, segmentTypePol2Peaks, segmentTypeTotReads, null); }

    public static void saveLatexTable(String latexTableFileName, int totReads, int[] typeCount, int[] segmentTypeTotLengths, int[] segmentTypePol2Peaks, int[] segmentTypeTotReads, int[] segmentTypePolyA) {
	//Assume that everything has the same order as TYPES_ARRAY_STR
	try {
	    System.out.println("Saving summary statistics for all chromosomes to " + latexTableFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(latexTableFileName));
	    bw.write("\\begin{tabular}{lrcrcrrr}\n\t\\textbf{Category} & \\multicolumn{2}{c}{\\textbf{\\# transcripts}} & \\multicolumn{2}{c}{\\textbf{RNAPII}} & \\textbf{Genome} & \\textbf{Reads} & \\textbf{polyA$^+$} \\\\\n\t\\hline\\\\\n");
	    double lenSum = 0, readsSum = 0, nPol2 = 0, nPolyA = 0;
	    int nTranscripts = 0;
	    for (int i=0; i<typeCount.length; i++) {
		nTranscripts += typeCount[i];
		lenSum += segmentTypeTotLengths[i];
		readsSum += segmentTypeTotReads[i];
		nPol2 += segmentTypePol2Peaks[i];
		if (segmentTypePolyA!=null)
		    nPolyA += segmentTypePolyA[i];
	    }
	    for (int i=0; i<TranscribedRegion.TYPES_ARRAY_STR.length; i++) {
		if (typeCount[i]>0) {
		    bw.write(TranscribedRegion.TYPES_ARRAY_STR[i].replace("_", " "));
		    bw.write(" & " + typeCount[i] + " & (" + format.format((double)typeCount[i]/nTranscripts) + ") & " + segmentTypePol2Peaks[i] + " & (" + format.format((double)segmentTypePol2Peaks[i]/typeCount[i]) + ") & " + format.format((double)segmentTypeTotLengths[i]/RefSeqGenes.TOTAL_CHROMOSOME_LENGTH) + " & " + format.format((double)segmentTypeTotReads[i]/totReads) + " & " + format.format((double)segmentTypePolyA[i]/typeCount[i]) + "\\\\\n");
		    System.out.print("\n");
		}
	    }
	    bw.write("\\hline\n");
	    bw.write("Total & " + nTranscripts + " & & " + nPol2 + " & & " + format.format(lenSum/RefSeqGenes.TOTAL_CHROMOSOME_LENGTH) + " & " + format.format(readsSum/totReads) + " & " + format.format((double)nTranscripts/nPolyA) + "\n");
	    bw.write("\\end{tabular}\n"); //footer
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving latex table " + e.toString());
	}
    }

    public static void saveLatexTable(String latexTableFileName, String[] types, int[] typesCount, LinkedHashMap header2DataMap, boolean includeFractions[], boolean weightedAverage[]) {
	//Assume that everything has the same order as TYPES_ARRAY_STR
	DecimalFormat format = new DecimalFormat("###.###");
	try {
	    System.out.println("Saving summary statistics for all chromosomes to " + latexTableFileName);
	    //Find out how many columns we need
	    int nHeaders = 1 + header2DataMap.size(), nFractions = 0;
	    for (int i=0; i<includeFractions.length; i++)
		nFractions += includeFractions[i] ? 1 : 0;
	    //Keep track of the totals
	    LinkedHashMap<String, Double> header2SumMap = new LinkedHashMap<String, Double>();
	    int totTranscripts = 0;
	    BufferedWriter bw = new BufferedWriter(new FileWriter(latexTableFileName));
	    //write the header
	    bw.write("\\begin{tabular}{l");
	    for (int i=0; i<nHeaders + nFractions; i++)
		bw.write("r");
	    bw.write("}\n\\textbf{Category} & \\multicolumn{2}{c}{\\textbf{\\# Transcripts}}");
	    for (int i=0; i<typesCount.length; i++)
		totTranscripts += typesCount[i];
	    int i = 0;
	    Iterator it = header2DataMap.keySet().iterator();
	    while (it.hasNext()) {
		String h = (String)it.next();
		double d = 0;
		bw.write(" & ");
		if (includeFractions[i])
		    bw.write("\\multicolumn{2}{c}{"); 
		bw.write("\\textbf{" + h + "}");
		if (includeFractions[i])
		    bw.write("}");
		i++;
		//Find the sum for the category
		for (int j=0; j<((double[])header2DataMap.get(h)).length; j++) {
		    if (weightedAverage[i])
			d += ((double[])header2DataMap.get(h))[j]*((double)typesCount[j])/totTranscripts;
		    else
			d += ((double[])header2DataMap.get(h))[j];
		}
		header2SumMap.put(h, d);
	    }
	    bw.write("\\\\\n\\hline\\\\\n");
	    //write the data for each category
	    for (int j=0; j<types.length; j++) {
		bw.write(types[j] + " & " + typesCount[j] + " & (" + format.format((double)typesCount[j]/totTranscripts) + ")");
		i = 0;
		it = header2DataMap.keySet().iterator();
		while (it.hasNext()) {
		    String h = (String)it.next();
		    bw.write(" & " + format.format(((double[])header2DataMap.get(h))[j]));
		    if (includeFractions[i++])
			bw.write(" & (" + format.format(((double[])header2DataMap.get(h))[j]/header2SumMap.get(h)) + ")");
		}
		bw.write("\\\\\n");
	    }
	    //Write the sum
	    bw.write("\\hline\n");
	    bw.write("Total & " + totTranscripts + " & ");
	    i = 0;
	    for (String h : header2SumMap.keySet()) {
		bw.write(" & " + format.format(header2SumMap.get(h))); 
		if (includeFractions[i++])
		    bw.write(" & ");
	    }
	    bw.write("\n");
	    bw.write("\\end{tabular}\n"); //footer
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving latex table " + e.toString());
	}
    }

    public static void saveGeneLatexTable(String geneLatexTableFileName, int totReads, int nUniqueGenes, int meanUniqueLengths, int totUniqueReads, int nSplitGenes, int meanSplitLengths, int totSplitReads, int nMergedGenes, int meanMergedLengths, int totMergedReads, int nMissedGenes, int meanMissedLength, int totMissedReads) { 
	DecimalFormat format = new DecimalFormat("###.###");
	try {
	    System.out.println("Saving summary statistics for genes to " + geneLatexTableFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(geneLatexTableFileName));
	    bw.write("\\begin{tabular}{lrcrr}\n\t\\textbf{Category} & \\multicolumn{2}{c}{\\textbf{\\# genes}} & \\textbf{$<$Length$>$} & \\textbf{Reads} \\\\\n\t\\hline\\\\\n");
	    double lenSum = 0;
	    int nReads = totUniqueReads + totMergedReads + totSplitReads + totMissedReads;
	    int nGenes = nUniqueGenes + nMergedGenes + nSplitGenes + nMissedGenes;
	    bw.write(RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nUniqueGenes + " & (" + format.format((double)nUniqueGenes/nGenes) + ") & " + meanUniqueLengths + " & " + format.format((double)totUniqueReads/totReads) + "\\\\\n");
	    bw.write(RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nSplitGenes + " & (" + format.format((double)nSplitGenes/nGenes) + ") & " + meanSplitLengths + " & " + format.format((double)totSplitReads/totReads) + "\\\\\n");
	    bw.write(RefSeqGene.MERGED_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nMergedGenes + " & (" + format.format((double)nMergedGenes/nGenes) + ") & " + meanMergedLengths + " & " + format.format((double)totMergedReads/totReads) + "\\\\\n");
	    bw.write(RefSeqGene.MISSED_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nMissedGenes + " & (" + format.format((double)nMissedGenes/nGenes) + ") & " + meanMissedLength + " & " + format.format((double)totMissedReads/totReads) + "\\\\\n");
	    bw.write("\\hline\n");
	    bw.write("Total & " + nGenes + " & & & " + format.format((double)nReads/totReads) + "\n");
	    bw.write("\\end{tabular}\n"); //footer
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving latex table " + e.toString());
	}
    }

    public static void saveTranscriptProperties(ReducedModel reducedModel, int binSize, String peakFileNames[], boolean orchestra, String saveFileName, boolean reverseStrand, int chr, boolean conservation) { saveTranscriptProperties(reducedModel, binSize, peakFileNames, orchestra, saveFileName, reverseStrand, chr, conservation, false, false); }

    public static void saveTranscriptProperties(ReducedModel reducedModel, int binSize, String peakFileNames[], boolean orchestra, String saveFileName, boolean reverseStrand, int chr, boolean conservation, boolean bayes, boolean chip) {
	//saveTranscriptProperty(reducedModel.getTranscribedRegionTypesLengths(), saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + TRANSCRIPT_LENGTHS_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	//saveTranscriptProperty(reducedModel.getTranscribedRegionTypesDensities(), saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + TRANSCRIPT_DENSITIES_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	//saveTranscriptProperty(reducedModel.getTranscribedRegionTypesWithinDensityStdDevs((int)Math.pow(2, binSize)), saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + TRANSCRIPT_WITHIN_DENISTY_STD_DEV_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	if (peakFileNames!=null) {
	    //for (int i=0; i<peakFileNames.length; i++)
	    //	saveTranscriptProperty(reducedModel.getTranscribedRegionTypesPeaks(peakFileNames[i]), saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + "_" + peakFileNames[i] + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	}
	if (conservation) {
	    //reducedModel.calculateConservation(orchestra);
	    //saveTranscriptProperty(reducedModel.getTranscribedRegionTypesConservation(), saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + TRANSCRIPT_CONSERVATION_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	    //saveTranscriptProperty(reducedModel.getTranscribedRegionTypesWithinConservationStdDevs(binSize), saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + TRANSCRIPT_WITHIN_CONSERVATION_STD_DEV_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	}
    }

    public static void saveTranscriptProperty(TreeMap<String, TallyStore> type2data, String saveFileName, boolean header) {
	//Use this method to save data of a specific transcript
	//property to a file. Each column represents a certain type of
	//transcript and they are padded with NaN to all have the same
	//number of rows which will make it easier to merge.
	try {
	    System.out.println("Saving to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    if (header) {
		bw.write("#");
		for (int j=0; j<TranscribedRegion.TYPES_ARRAY_STR.length; j++) 
		    bw.write(TranscribedRegion.TYPES_ARRAY_STR[j] + "\t");
		bw.write("\n");
	    }
	    int maxCount = 0; //Find the category with the largest no elements
	    for (TallyStore t : type2data.values())
		maxCount = maxCount>t.numberObs() ? maxCount : t.numberObs();
	    for (int i=0; i<maxCount; i++) {
		for (int j=0; j<TranscribedRegion.TYPES_ARRAY_STR.length; j++) {
		    if (type2data.containsKey(TranscribedRegion.TYPES_ARRAY_STR[j]) && i<type2data.get(TranscribedRegion.TYPES_ARRAY_STR[j]).numberObs() && !Double.isNaN(type2data.get(TranscribedRegion.TYPES_ARRAY_STR[j]).getArray()[i]) && !Double.isInfinite(type2data.get(TranscribedRegion.TYPES_ARRAY_STR[j]).getArray()[i]))
			bw.write(format.format(type2data.get(TranscribedRegion.TYPES_ARRAY_STR[j]).getArray()[i]) + "\t");
		    else
			bw.write("NaN\t");
		}
		bw.write("\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving " + e.toString());
	}
    }

    public static String printStats(double fdr, ReducedModel reducedModel, TreeMap<TranscribedRegion, Peak> transcriptsWithPeaks) { return printStats(fdr, reducedModel, transcriptsWithPeaks, null); }

    public static String printStats(double fdr, ReducedModel reducedModel, TreeMap<TranscribedRegion, Peak> transcriptsWithPeaks, Vector<Peak> peaks) {
	//n_regions, proportion transcribed, proportion of reads
	String str = format.format(fdr) + "\t" + reducedModel.numberOfTranscribedRegions() + "\t" + format.format(reducedModel.proportionTranscribed()) + "\t" + format.format(reducedModel.proportionOfReads()) + "\t";
	//PPV sensitivity
	//str += format.format(reducedModel.getPositivePredictiveValue()) + "\t";
	//n_unique, n_missed_genes, n_genes_w_multiple
	str += format.format(reducedModel.numberOfGenesWithOneTranscript()) + "\t" + format.format(reducedModel.numberOfGenesWithNTranscripts(0, false, false)) + "\t" + format.format(reducedModel.numberOfGenesWithNTranscripts(2, true, false)) + "\t";
	//n_nc_transcripts, n_transcripts_covering_multiple
	str += format.format(reducedModel.numberOfTranscriptsWithNGenes(0, false, false)) + "\t" + format.format(reducedModel.numberOfTranscriptsWithNGenes(2, true, false)) + "\t";
	//proportion with peaks
	if (peaks!=null)
	    str += format.format((double)transcriptsWithPeaks.size()/(peaks.size() - transcriptsWithPeaks.size() + reducedModel.numberOfTranscribedRegions()));
	else
	    str += format.format((double)transcriptsWithPeaks.size()/reducedModel.numberOfTranscribedRegions());
	System.out.println(str);
	return str;
    }

    public static void createWigFile(String saveFileName, boolean chip, boolean reverseStrand, int chr, boolean bayes) {
	String saveFile = saveFileName + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f")) + (chr!=-1 ? "_" + chr : "") + ".bed";
	String name = saveFileName.substring(saveFileName.lastIndexOf("mRNA/") + 5) + (bayes ? "_bayes" : "") + (chip ? BED_FILE_SUFFIX_CHIP : BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f"));
	String colorStr = CreateWigFiles.getColorString(saveFileName);
	int priority = CreateWigFiles.getPriority(saveFileName);
	String trackInfoStr = "track type=bigWig name=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" description=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" priority=" + priority + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=" + URL_NAME + name + (chr==-1 ? "" : "_" + chr) + ".bw";
	CreateWigFiles.addTrackInfo(trackInfoStr, TRACK_INFO_FILE_PATH + (saveFileName.contains("MeCP2") ? "MeCP2/" : saveFileName.substring(saveFileName.lastIndexOf("mRNA/") + 4, saveFileName.lastIndexOf("/") + 1)) + TRACK_INFO_FILE_NAME);
	//CreateWigFiles.createWWWLink(saveFile, "/www/" + WWW_DIR_NAME + (saveFileName.contains("MeCP2") ? "MeCP2/" : saveFileName.substring(saveFileName.lastIndexOf("mRNA/") + 5, saveFileName.lastIndexOf("/") + 1)));
	//createWigFile(saveFileName, name, chr);
    }

    /*public static void createWigFile(String saveFileName, int chr) { createWigFile(saveFileName, saveFileName.replace(".bed", ""), chr); }

    public static void createWigFile(String saveFileName, String name, int chr) {
	String colorStr = CreateWigFiles.getColorString(saveFileName);
	int priority = CreateWigFiles.getPriority(saveFileName);
	String trackInfoStr = "track type=bigWig name=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" description=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" priority=" + priority + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=" + URL_NAME + name + (chr==-1 ? "" : "_" + chr) + ".bw";
	CreateWigFiles.addTrackInfo(trackInfoStr, TRACK_INFO_FILE_PATH + (saveFileName.contains("MeCP2") ? "MeCP2/" : "") + TRACK_INFO_FILE_NAME);
	CreateWigFiles.createWWWLink(saveFile, "/www/" + WWW_DIR_NAME + (saveFileName.contains("MeCP2") ? "MeCP2/" : ""));
	}*/

    public static int numberOfChromosomes(String species) {
	if (species.equals("hs") || species.equals("human"))
	    return 24;
	else if (species.equals("sc") || species.equals("cerevisiae"))
	    return 17;
	else if (species.equals("pf") || species.equals("plasmodium"))
	    return 15;
	return 21; //default is mouse
    }

    public static int[] getLengthVec(int lMin, int lMax, int deltaL) {
	int lVec[] = new int[(lMax-lMin)/deltaL];
	for (int i=0; i<(lMax-lMin)/deltaL; i++)
	    lVec[i] = lMin + i*deltaL;
	return lVec;
    }

    public static double[] getFDRVec(double fdrMin, double fdrMax, int nFDR) {
	//create nFDR points equally spaced on a log-scale btwn fdrMin and fdrMax
	if (nFDR<=1)
	    return new double[]{fdrMin};
	double fdrVec[] = new double[nFDR], deltaFDR = (Math.log10(fdrMax) - Math.log10(fdrMin))/(nFDR-1);
	for (int i=0; i<nFDR; i++)
	    fdrVec[i] = Math.pow(10, Math.log10(fdrMin) + deltaFDR*i);
	return fdrVec;
    }

    public static double[] getCompareVec(double compareMin, double compareMax, int nCompare) {
	if (nCompare<=1)
	    return new double[]{compareMin};
	double compareVec[] = new double[nCompare], delta = (compareMax - compareMin)/(nCompare - 1);
	for (int i=0; i<nCompare; i++)
	    compareVec[i] = compareMin + delta*i;
	return compareVec;
    }

    ///////////////////////////////////////////////////////

    public static TreeMap<Double, Integer> parseRankedGenes(int time) {
	//Find the compare proportion of genes with highest
	//expression. Returns a map where the key is the read density
	//and the value is the geneID.
	TreeMap<Double, Integer> map = new TreeMap<Double, Integer>();
	try {
	    int nCollisions = 0, expGroup = time==0 ? 6 : (time==1 ? 7 : 8);
	    BufferedReader br = new BufferedReader(new FileReader(ID_EXPRESSION_FILE_LOCATION));
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = ID_EXPRESSION_FILE_PATTERN.matcher(line);
		if (m.find()) {
		    int id = Integer.parseInt(m.group(1));
		    double exp = Double.parseDouble(m.group(expGroup));
		    while (map.containsKey(exp)) {
			exp += random.nextGaussian()*.00000001;
			nCollisions++;
		    }
		    map.put(exp, id);
		}
		line = br.readLine();
	    }
	    //System.out.println("There were " + nCollisions + " collisions when sorting the list of expressed genes.");
	}
	catch (IOException e) {
	    System.err.println("Error parsing symbolic expression file: " + e.toString());
	}		    
	return map;
    }

    public static TreeMap<Double, Integer> findGeneDensities(RefSeqGenes geneInfo, boolean reverseStrand, String fileName) { return findGeneDensities(geneInfo, reverseStrand ? -1 : 1, fileName); }

    public static TreeMap<Double, Integer> findGeneDensities(RefSeqGenes geneInfo, String fileName) { return findGeneDensities(geneInfo, 0, fileName); }

    protected static TreeMap<Double, Integer> findGeneDensities(RefSeqGenes geneInfo, int strand, String fileName) {
	TreeMap<Double, Integer> exp2IdMap = new TreeMap<Double, Integer>();
	for (RefSeqGene g : geneInfo.getGenes()) {
	    if (strand==0 || (g.getStrand() && strand>0) || (!g.getStrand() && strand<0)) {
		double k = g.getReadDensity(fileName);
		//if (k>.0001)
		//  System.out.println(format.format(k) + "\t" + g.toString());
		while (exp2IdMap.containsKey(k))
		    k += Math.abs(random.nextGaussian()*.0000001);
		exp2IdMap.put(k, g.getID());
	    }
	}
	return exp2IdMap;
    }

    public static int[] getHighestExpressedGenes(TreeMap<Double, Integer> exp2IdMap, double compare, RefSeqGenes geneInfo) {
	//Find how many expressed genes there are on this chr
	int n = 0;
	for (Integer v : exp2IdMap.values())
	    n += geneInfo.getGene(v)!=null ? 1 : 0;
	int s = (int)(compare*(double)n), i = 0, j = 0;
	int geneIDsKeep[] = new int[s];
	Vector<Double> keys = new Vector<Double>(exp2IdMap.keySet());
	while (i<s && j<exp2IdMap.size()) {
	    int k = exp2IdMap.get(keys.elementAt(keys.size() - j - 1));
	    if (geneInfo.getGene(k)!=null)
		geneIDsKeep[i++] = k;
	    j++;
	}
	Arrays.sort(geneIDsKeep);
	return geneIDsKeep;
    }

    public static int[] getGenesWithDensityGreaterThan(double minDetectedDens, TreeMap<Double, Integer> exp2IdMap, RefSeqGenes geneInfo, boolean forward) {
	int n = 0;
	TreeMap<Double, Integer> highMap = new TreeMap<Double, Integer>(exp2IdMap.tailMap(minDetectedDens));
	for (Integer v : highMap.values()) 
	    n += geneInfo.getGene(v)!=null && geneInfo.getGene(v).getStrand()==forward ? 1 : 0;
	int i = 0, j = 0;
	int geneIDsKeep[] = new int[n];
	Vector<Integer> values = new Vector<Integer>(highMap.values());
	while (i<n && j<values.size()) {
	    if (geneInfo.getGene(values.elementAt(j))!=null && geneInfo.getGene(values.elementAt(j)).getStrand()==forward) {
		geneIDsKeep[i++] = values.elementAt(j);
		//System.out.println(format.format(values.elementAt(j)) + "\t" + geneInfo.getGene(values.elementAt(j)).toString());
	    }
	    j++;
	}
	Arrays.sort(geneIDsKeep);
	return geneIDsKeep;
    }

    public static int[] getGenesWithDensityGreaterThan(double minDetectedDens, TreeMap<Double, Integer> exp2IdMap, RefSeqGenes geneInfo) {
	int n = 0;
	TreeMap<Double, Integer> highMap = new TreeMap<Double, Integer>(exp2IdMap.tailMap(minDetectedDens));
	for (Integer v : highMap.values()) 
	    n += geneInfo.getGene(v)!=null ? 1 : 0;
	int i = 0, j = 0;
	int geneIDsKeep[] = new int[n];
	Vector<Integer> values = new Vector<Integer>(highMap.values());
	while (i<n && j<values.size()) {
	    if (geneInfo.getGene(values.elementAt(j))!=null)
		geneIDsKeep[i++] = values.elementAt(j);
	    j++;
	}
	Arrays.sort(geneIDsKeep);
	return geneIDsKeep;
    }

    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName) { return parseEnhancerFile(fileName, "Mm"); }
		  
    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName, String species) { return parseEnhancerFile(fileName, species, -1); }
		  
    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName, int chr) { return parseEnhancerFile(fileName, "Mm", chr); }

    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName, String species, int chr) {
	//Use the subset files by JMG that contain chr, pos and TRUE|FALSE
	TreeSet<Enhancer2> enhancers = new TreeSet<Enhancer2>();
	if (fileName!=null) {
	    try {
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		String line = br.readLine();
		//chr pos TRUE|FALSE
		Pattern enhancerPattern = Pattern.compile("^(\\d+)\\s+(\\d+)\\s+([TRUE|FALSE])");
		while (line!=null) {
		    Matcher m = enhancerPattern.matcher(line);
		    if (m.find()) {
			int c = Integer.parseInt(m.group(1));
			if ((c==chr || chr==-1) && m.group(3).equals("TRUE")) 
			    enhancers.add(new Enhancer2(c, Integer.parseInt(m.group(2))));
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error reading enhancers from " + fileName + ": " + e.toString());
	    }
	}
	System.out.println("Found " + enhancers.size() + " enhancers.");
	return enhancers;
    }

}

