//use to detect transcripts using the wavelet based method. The
//current version only contains three parameters: two length scales
//and a cut-off for the number of break points and what densities to
//include. The last parameter is similar to an FDR type cut-off and to
//sweep for different values, we may use:
//
//Version 2 is a slimmed down version of FindTranscripts.java and some
//stuff that is no longer used has been removed, including the
//Bayesian method and everything that was used for ChIP experiments.

//java -Xmx2548m FindTranscripts2 -c 1 -n /groups/neuroduo/martin/ChIP-Seq/MmData/mRNA/hr6c  -l 10 -L 18 -R 35 --merge --minDens .01 --fdr .001 --fdrMax .2 --nfdr 10 -o --peakFileName ../MmDataPeakFiles/Pol2_8WG16_KCl_B1_names --gmm --log

//To compare the obtained transcripts against the annotation

//java -Xmx2048m FindTranscripts2 -c 1 -n /groups/neuroduo/martin/ChIP-Seq/TKData/mRNA/hr6m  -l 9 -L 16 -R 35 --merge --fdr .025 --compare .1 --compareMax .9 --ncompare 9

//To look at human HeLa cells, we might want to do:

//java -Xmx2048m FindTranscripts2 -c 18 -n /groups/neuroduo/martin/ChIP-Seq/HeLa/mRNA/HeLac  -l 9 -L 16 -R 35 --merge --fdr .025 -o --species hs

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;
import umontreal.iro.lecuyer.stat.TallyStore;

public class FindTranscripts2 {

    public static final String BED_FILE_SUFFIX = "_transcripts";
    public static final String RANDOM_READS_FILE_SUFFIX = "_random_reads_transcripts";
    public static final String DOWNSAMPLING_FILE_SUFFIX = "_downsampling_transcripts";
    public static final String LATEX_TABLE_FILE_NAME_SUFFIX = "_summary_stats.tex";
    public static final String TRANSCRIPT_LENGTHS_SUFFIX = "_lengths";
    public static final String TRANSCRIPT_DENSITIES_SUFFIX = "_dens";
    public static final String TRANSCRIPT_WITHIN_DENISTY_STD_DEV_SUFFIX = "_densStdDev";
    public static final String TRANSCRIPT_PEAKS_SUFFIX = "_peaks";
    public static final String TRANSCRIPT_CONSERVATION_SUFFIX = "_cons";
    public static final String TRANSCRIPT_WITHIN_CONSERVATION_STD_DEV_SUFFIX = "_consStdDev";
    public static final String NON_CODING_AS_INFO_SUFFIX = "Noncoding_AS_info";
    
    public static final String WWW_DIR_NAME = "neuroduo.hms.harvard.edu/docroot/bigWig/mh/Transcripts/";
    public static final String URL_NAME = "http://neuroduo.hms.harvard.edu/bigWig/mh/Transcripts/";
    public static final String TRACK_INFO_FILE_PATH = "/www/neuroduo.hms.harvard.edu/docroot/bigWig/mh/Transcripts/";
    public static final String TRACK_INFO_FILE_NAME = "track_info.txt";
    public static final String RANDOM_DIR_NAME = "Random/";
    public static final String SEQUENCE_DIR_NAME = "Sequences/";
    public static final String AA_DIR_NAME = "AA/";

    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option readLengthFlag = argParser.addIntegerOption('R', "readLength");	    
	    CmdLineParser.Option strandFlag = argParser.addBooleanOption('r', "reverseStrand");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option logFlag = argParser.addIntegerOption("log");
	    CmdLineParser.Option gmmFlag = argParser.addIntegerOption("gmm");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption("conservation");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option sequenceFlag = argParser.addBooleanOption("sequence");
	    CmdLineParser.Option cpcFlag = argParser.addBooleanOption("CPC");
	    CmdLineParser.Option aaFlag = argParser.addIntegerOption("AA");
	    //Parameters for the transcript finding
	    CmdLineParser.Option lMinFlag = argParser.addIntegerOption('l', "lMin");
	    CmdLineParser.Option lMaxFlag = argParser.addIntegerOption('L', "lMax");
	    //For coarse-graining
	    CmdLineParser.Option binSizeFlag = argParser.addIntegerOption('b', "binSize");
	    CmdLineParser.Option deltaLFlag = argParser.addIntegerOption("deltaL");
	    CmdLineParser.Option fdrFlag = argParser.addDoubleOption("fdr");
	    CmdLineParser.Option fdrMaxFlag = argParser.addDoubleOption("fdrMax");
	    CmdLineParser.Option nfdrFlag = argParser.addIntegerOption("nfdr");
	    CmdLineParser.Option mergeFlag = argParser.addBooleanOption("merge");
	    CmdLineParser.Option flattenFlag = argParser.addBooleanOption("flatten");
	    CmdLineParser.Option minDensFlag = argParser.addDoubleOption("minDens");
	    CmdLineParser.Option costFunctionFlag = argParser.addStringOption("cost");
	    //Use these flags when we want to down-sample the total number of reads to investigate how the optimal parameters change
	    CmdLineParser.Option downSampleFlag = argParser.addIntegerOption("downSample");
	    //Use this to indicate a peak file, eg Pol2 or H3K4Me3
	    //that we wish to compare with for the optimization
	    CmdLineParser.Option peakFileNameFlag = argParser.addStringOption("peakFileName");
	    //Use this flag to provide a set of peaks that are to be
	    //used for spiking the set of breakpoints, eg Pol2 peak
	    //locations
	    CmdLineParser.Option breakPointsFileNameFlag = argParser.addStringOption("breakPointsFileName");
	    //Use this flag to save information about intermediate steps of the algorithm
	    CmdLineParser.Option saveInfoFileNameFlag = argParser.addBooleanOption("saveInfo");
	    //Use this flag to create a set of randomized regions with
	    //the same lengths as the ones detected by the algorithm
	    CmdLineParser.Option randomFlag = argParser.addIntegerOption("random");
	    //Use this flag when we want to randomize the location of
	    //the reads and the breakpoints instead to determine the
	    //background detection level.
	    CmdLineParser.Option randomReadsFlag = argParser.addIntegerOption("randomReads");
	    //For comparing against known elements
	    CmdLineParser.Option enhancerFileNameFlag = argParser.addStringOption("enhancers");
	    CmdLineParser.Option lncRNAFileNameFlag = argParser.addStringOption("lncRNA");
	    CmdLineParser.Option macroRNAFileNameFlag = argParser.addStringOption("macroRNA");
	    CmdLineParser.Option miRNAFileNameFlag = argParser.addStringOption("miRNA");
	    CmdLineParser.Option miRNAFileNamePrecursorsFlag = argParser.addStringOption("miRNAPrecursors");
	    CmdLineParser.Option maxDivergentTSSDistanceFlag = argParser.addIntegerOption("maxDivergent");
	    //If we want to use the alternative categorization based
	    //only on how close the unannotated start is to the
	    //annotated start.
	    CmdLineParser.Option promoterLocationCategoriesFlag = argParser.addBooleanOption("promoterLocation");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    boolean sequence = argParser.getOptionValue(sequenceFlag)!=null; 
	    boolean cpc = argParser.getOptionValue(cpcFlag)!=null; 
	    tmp = argParser.getOptionValue(aaFlag);
	    int minAAs = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(randomFlag);
	    int nMC = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(randomReadsFlag);
	    int nReadsMC = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    boolean rand = nMC>0; 
	    tmp = argParser.getOptionValue(gmmFlag);
	    int nComponents = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    boolean gmm = nComponents>1;
	    boolean reverseStrand = argParser.getOptionValue(strandFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp + (((String)tmp).contains(".") ? "" : ".tsv") : "";
	    tmp = argParser.getOptionValue(peakFileNameFlag);
	    String peakFileName = tmp!=null ? (String)tmp : null;
	    tmp = argParser.getOptionValue(breakPointsFileNameFlag);
	    String breakPointsFileName = tmp!=null ? (String)tmp : null;
	    boolean saveInfo = argParser.getOptionValue(saveInfoFileNameFlag)!=null;
	    String saveInfoFileName = saveInfo ? fileName.substring(0, fileName.lastIndexOf(".")) + (reverseStrand ? "_r" : "_f") : null;
	    String latexTableFileName = saveInfoFileName + LATEX_TABLE_FILE_NAME_SUFFIX;
	    tmp = argParser.getOptionValue(costFunctionFlag);
	    String costFunction = tmp!=null ? ((String)tmp).toLowerCase() : "sens";
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : fileName.substring(0, fileName.lastIndexOf("."));
	    String saveFileNameRandom = saveFileName.substring(0, saveFileName.lastIndexOf("/") + 1) + RANDOM_DIR_NAME + saveFileName.substring(saveFileName.lastIndexOf("/") + 1);
	    tmp = argParser.getOptionValue(logFlag);
	    int nLogs = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    boolean log = nLogs>0;
	    tmp = argParser.getOptionValue(readLengthFlag);
	    int readLength = tmp!=null ? ((Integer)tmp).intValue() : 35;
	    tmp = argParser.getOptionValue(lMinFlag);
	    int lMin = tmp!=null ? ((Integer)tmp).intValue() : 5;
	    tmp = argParser.getOptionValue(lMaxFlag);
	    int lMax = tmp!=null ? ((Integer)tmp).intValue() : 15;
	    tmp = argParser.getOptionValue(binSizeFlag);
	    int binSize = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(deltaLFlag);
	    int deltaL = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(fdrFlag);
	    double fdr = tmp!=null ? ((Double)tmp).doubleValue() : .01;
	    tmp = argParser.getOptionValue(fdrMaxFlag);
	    double fdrMax = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(nfdrFlag);
	    int nfdr = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(downSampleFlag);
	    int nDownSample = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(enhancerFileNameFlag);
	    String enhancerFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/ChIP-Seq/MmData/" : "/home/mhemberg/Projects/ChIP-Seq/TKData/MmData" ) + "Enhancers/Subsets/CBP_Millipore_KCl_B1_subset_enh_ex_lMe3_real.tsv";
	    tmp = argParser.getOptionValue(lncRNAFileNameFlag);
	    String lncRNAFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/home/mh168/" : "/home/mhemberg/" ) + "Projects/ChIP-Seq/TKData/mRNA/lncRNA/landerRegions.bed";
	    tmp = argParser.getOptionValue(macroRNAFileNameFlag);
	    String macroRNAFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/macroRNA.txt";
	    tmp = argParser.getOptionValue(miRNAFileNameFlag);
	    String miRNAFileName = tmp!=null ? (String)tmp : !species.equals("mm") ? null : (orchestra ? "/groups/neuroduo/martin/" : "/media/My Book/" ) + "Genomes/Mm/v9/miRNA.txt";
	    tmp = argParser.getOptionValue(miRNAFileNamePrecursorsFlag);
	    String miRNAFileNamePrecursors = null; //tmp!=null ? (String)tmp : (orchestra ? "/home/mh168" : "/home/mhemberg" ) + "/Projects/ChIP-Seq/TKData/mRNA/miRNA/landgrafTuschl_2007_C_mm_miRNA_precursors_names.txt";
	    boolean promoterLocationCategories = argParser.getOptionValue(promoterLocationCategoriesFlag)!=null;
	    tmp = argParser.getOptionValue(maxDivergentTSSDistanceFlag);
	    int maxDivergentTSSDistance = tmp!=null ? ((Integer)tmp).intValue() : promoterLocationCategories ? 10000 : 2000;
	    boolean mergeSegments = argParser.getOptionValue(mergeFlag)!=null;
	    boolean flatten = argParser.getOptionValue(flattenFlag)!=null;
	    tmp = argParser.getOptionValue(minDensFlag);
	    double minDens = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    //////////////////////////////////////////////////////////////////
	    int nChr = numberOfChromosomes(species) - 1;
	    int lVec[] = getLengthVec(lMin, lMax, deltaL), lVecCoeffs[] = getLengthVec(binSize, lMax, deltaL);
	    double fdrVec[] = getFDRVec(fdr, fdrMax, nfdr);
	    Pattern rnaFileNamePattern = Pattern.compile("hr(\\d)\\w");
	    Matcher matcher = rnaFileNamePattern.matcher(fileName);
	    int time = matcher.find() ? Integer.parseInt(matcher.group(1)) : 6;
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    TranscriptFinder tf[] = new TranscriptFinder[nChr];
	    ReducedModel reducedModel[] = new ReducedModel[nChr];
	    ReducedModel reducedModelRandom[] = new ReducedModel[nChr];
	    //Load the enhancers
	    TreeSet<Enhancer2> enhancers = parseEnhancerFile(enhancerFileName, chr);
	    //Load the lncRNAs
	    TreeSet<LncRNA> lncRNAs = AnalyzeTranscriptReadProfiles.loadLncRNA(lncRNAFileName, species);
	    TreeSet<MacroRNA> macroRNAs = AnalyzeTranscriptReadProfiles.loadMacroRNA(macroRNAFileName, species);
	    //Load the miRNAs
	    TreeSet<MiRNA> miRNAs = AnalyzeTranscriptReadProfiles.loadMiRNA(miRNAFileName, miRNAFileNamePrecursors, species);
	    //Load the peaks
	    Vector<Peak> peaks = peakFileName==null ? null : new Vector<Peak>(FindEnhancerSites.parsePeakFile(peakFileName, chr).values());
	    Vector<Peak> breakPointPeaks = breakPointsFileName==null ? peaks : new Vector<Peak>(FindEnhancerSites.parsePeakFile(breakPointsFileName, chr).values());
	    //Load the extended list of TSSs

	    //System.out.println("FDR\tn\ttranscribed\treads\tPPV\tSensitivity\t#unique\t#missed_g\t#g_spanned_by_multiple_t\t#non-coding_t\t#t_spanning_mutliple_g\tPeak_overlap");
	    for (int i=1; i<=nChr; i++) {
		int nTranscriptsFoundRandomization[] = new int[nReadsMC];
		if (i==chr || chr==-1) {
		    BufferedWriter bwDownSampling = nDownSample>0 ? new BufferedWriter(new FileWriter(saveFileName + DOWNSAMPLING_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".txt")) : null;
		    if (bwDownSampling!=null) {
			if (i==1) {
			    bwDownSampling.write("#Chr");
			    for (int m=1; m<=nDownSample; m++)
				bwDownSampling.write("\t" + format.format((double)m/nDownSample));
			    bwDownSampling.write("\n");
			}
			bwDownSampling.write((new Integer(i)).toString());
		    }
		    for (int m=1; m<=nDownSample; m++) {
			ABIRead[] rnaReadsForward = downSampleReads(ABIReads.loadMRNAReadsChr(fileName, true, i, readLength, false),(double)m/nDownSample);
			ABIRead[] rnaReadsReverse = downSampleReads(ABIReads.loadMRNAReadsChr(fileName, false, i, readLength, false),(double)m/nDownSample);
			//Keep track of how many transcripts were found when the reads were randomized
			for (int l=0; l<Math.max(1, nReadsMC); l++) {
			    if (nReadsMC>0) {
				rnaReadsForward = randomizeReads(rnaReadsForward);
				rnaReadsReverse = randomizeReads(rnaReadsReverse);
				System.out.println("Shuffled reads");
			    }
			    geneInfo[i-1] = new RefSeqGenes(i, orchestra, species);
			    geneInfo[i-1].setFactorReads(fileName, rnaReadsForward, 1);
			    geneInfo[i-1].setFactorReads(fileName, rnaReadsReverse, -1);
			    //if (!promoterLocationCategories) {
			    AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, fileName, rnaReadsForward, i, 1);
			    AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, rnaReadsForward, i, 1);
			    AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, fileName, rnaReadsForward, i);
			    AnalyzeTranscriptReadProfiles.setMiRNAFactorReads(miRNAs, fileName, rnaReadsReverse, i, -1);
			    AnalyzeTranscriptReadProfiles.setMaRNAFactorReads(macroRNAs, fileName, rnaReadsReverse, i, -1);
			    AnalyzeTranscriptReadProfiles.setLncRNAFactorReads(lncRNAs, fileName, rnaReadsReverse, i);
			    //}
			    tf[i-1] = new TranscriptFinder(reverseStrand ? rnaReadsReverse : rnaReadsForward, i, !reverseStrand, readLength, species, log);
			    tf[i-1].calculateHaarCoefficients(lVecCoeffs, binSize, saveInfoFileName);
			    geneInfo[i-1].setGeneIDsKeep(true, fileName, minDens);
			    geneInfo[i-1].setGeneIDsKeep(false, fileName, minDens);
			    for (int j=0; j<fdrVec.length; j++) {
				int bestIterations = tf[i-1].findRefinedReducedModel(lMin, lMax, binSize, fdrVec[j], mergeSegments, gmm, saveInfoFileName, geneInfo[i-1], lncRNAs, miRNAs, macroRNAs, enhancers, breakPointPeaks, peaks, maxDivergentTSSDistance, costFunction, -1);
				reducedModel[i-1] = tf[i-1].getReducedModel();
				//reducedModel[i-1].calculateLogEMALikelihood(log);
				/*if (promoterLocationCategories)
				    reducedModel[i-1].compare2Annotation(maxDivergentTSSDistance, geneInfo[i-1]);
				else
				reducedModel[i-1].compare2Annotation(maxDivergentTSSDistance, geneInfo[i-1], lncRNAs, miRNAs, macroRNAs, enhancers);*/
				//reducedModel[i-1].categorizeTranscribedRegions(promoterLocationCategories);
				//Find out how many of the transcripts have a peak near their start
				/*TreeMap<TranscribedRegion, Peak> transcriptsWithPeaks = peaks==null ? new TreeMap<TranscribedRegion, Peak>() : reducedModel[i-1].compare2Peaks(peaks, maxDivergentTSSDistance);
				reducedModel[i-1].removeGenes();
				for (int k=0; k<nMC; k++) {
				    reducedModelRandom[i-1] = reducedModel[i-1].getRandomModel(readLength, reverseStrand ? rnaReadsReverse : rnaReadsForward, log);
				    reducedModelRandom[i-1].compare2Annotation(maxDivergentTSSDistance, geneInfo[i-1], lncRNAs, miRNAs, macroRNAs, enhancers);
				    reducedModelRandom[i-1].categorizeTranscribedRegions(promoterLocationCategories);
				    reducedModelRandom[i-1].removeGenes();
				    }*/
			    }
			    if (saveFileName!=null && fdrVec.length<=1 && nReadsMC==0 && nDownSample==1) {
				String saveFile = saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".bed";
				//reducedModel[i-1].saveReducedModel(saveFile, (chr==1 || chr==-1) && fileName.contains("Plasmodium") ? saveFile.substring(saveFile.lastIndexOf("/") + 1, saveFile.lastIndexOf(".")) : null, chr==-1 && i>1);
				if (orchestra && chr!=-1 && !fileName.contains("Plasmodium")) //Put it in a directory on the www-site for easy upload to the genome browser
				    createWigFile(saveFileName, false, reverseStrand, chr, false);
				//Save the lengths of the different transcripts
				saveTranscriptProperties(reducedModel[i-1], binSize, peaks!=null ? peaks.firstElement().getFactorNames(false) : null, orchestra, saveFileName, promoterLocationCategories, reverseStrand, chr, conservation);			
				if (rand) {
				    saveFile = saveFileNameRandom + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".bed";
				    //reducedModelRandom[i-1].saveReducedModel(saveFile, (chr==1 || chr==-1) && fileName.contains("Plasmodium") ? saveFile.substring(saveFile.lastIndexOf("/") + 1, saveFile.lastIndexOf(".")) : null, chr==-1 && i>1);
				    saveTranscriptProperties(reducedModelRandom[i-1], binSize, peaks!=null ? peaks.firstElement().getFactorNames(false) : null, orchestra, saveFileNameRandom, promoterLocationCategories, reverseStrand, chr, conservation);			
				}
				if (sequence) { //Save the sequences for each transcript
				    Runtime runtime = cpc ? Runtime.getRuntime() : null;
				    extractSequences(reducedModel[i-1], promoterLocationCategories, runtime, saveFileName + "/" + SEQUENCE_DIR_NAME + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")), reverseStrand, chr, minAAs, orchestra);
				    if (rand)
					extractSequences(reducedModelRandom[i-1], promoterLocationCategories, runtime, saveFileNameRandom + "/" + SEQUENCE_DIR_NAME + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")), reverseStrand, chr, minAAs, orchestra);
				}
				//Save information about the Noncoding_AS
				//transcripts and their relative location to
				//their overlapping genes.
				if (!promoterLocationCategories)
				    saveNoncodingASInfo(reducedModel[i-1], saveFileName + "_" + NON_CODING_AS_INFO_SUFFIX + "_" + (reverseStrand ? "r" : "f") + "_" + i + ".txt");
				//Save statistics for the genes. Since the
				//number of classes of genes is fixed (unique,
				//split, multi and missed) we use the opposite
				//scheme than for the transcripts when saving
				//and use one file for each category with the
				//columns corresponding to properties instead
				//of the other way around.
				if (geneInfo[i-1].size()>0) {
				    //reducedModel[i - 1].saveGeneInfo(saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
				    //reducedModel[i - 1].saveGeneInfo(saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.MERGED_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.MERGED_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
				    //reducedModel[i - 1].saveGeneInfo(saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
				    //reducedModel[i - 1].saveGeneInfo(saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.MISSED_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.MISSED_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
				    if (rand) {
					//reducedModelRandom[i - 1].saveGeneInfo(saveFileNameRandom + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
					//reducedModelRandom[i - 1].saveGeneInfo(saveFileNameRandom + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.MERGED_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.MERGED_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
					//reducedModelRandom[i - 1].saveGeneInfo(saveFileNameRandom + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
					//reducedModelRandom[i - 1].saveGeneInfo(saveFileNameRandom + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + RefSeqGene.MISSED_TRANSCRIPT_GENE_STR + (chr!=-1 ? "_" + chr : "") + ".txt", fileName, RefSeqGene.MISSED_TRANSCRIPT_GENE_STR, minDens, chr==1 || chr==-1);
				    }
				}
			    }
			    else if (nReadsMC>0) { //Save info about the number of transcripts found
				nTranscriptsFoundRandomization[l] = reducedModel[i-1].numberOfTranscribedRegions();
				System.out.println("Found " + nTranscriptsFoundRandomization[l] + " transcripts from random model.");
			    }
			}
			if (nDownSample>0) {
			    System.out.println("Found " + reducedModel[i-1].numberOfTranscribedRegions() + " transcripts from down-sampled model.");
			    bwDownSampling.write("\t" + reducedModel[i-1].numberOfTranscribedRegions());
			}
		    }
		    tf[i-1] = null;
		    reducedModel[i-1] = null;
		    geneInfo[i-1] = null;
		    if (nDownSample>0) {
			bwDownSampling.write("\n");
			bwDownSampling.close();
		    }
		}
		if (nReadsMC>0) {
		    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName + RANDOM_READS_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".txt"));
		    for (int l=0; l<nReadsMC; l++)
			bw.write(nTranscriptsFoundRandomization[l] + "\n");
		    bw.close();
		}
	    }
	} 
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static ABIRead[] randomizeReads(ABIRead[] rnaReads) {
	//Use this method to randomize the location of the reads
	for (int i=0; i<rnaReads.length; i++)
	    rnaReads[i] = new ABIRead(rnaReads[i].getChromosome(), rnaReads[i].getStrand(), random.nextInt(RefSeqGenes.CHROMOSOME_LENGTH[rnaReads[i].getChromosome()]));
	Arrays.sort(rnaReads);
	return rnaReads;
    }

    public static void extractSequences(ReducedModel reducedModel, boolean promoterLocationCategories, Runtime runtime, String saveFileName, boolean reverseStrand, int chr, int minAAs, boolean orchestra) throws IOException {
	//Make sure that the directories exist
	File saveDir = new File(saveFileName.substring(0, saveFileName.lastIndexOf("/")));
	System.out.println("Creating directory " + saveDir.toString());
	if (!saveDir.mkdirs())
	    System.err.println("Error, failed to create directories to save: " + saveDir.toString());
	String[] arr = promoterLocationCategories ? TranscribedRegion.LOCATION_TYPES_ARRAY_STR : TranscribedRegion.TYPES_ARRAY_STR;
	for (int j=0; j<arr.length; j++) { 
	    //get one file for each category. This will allow us to split up the jobs for CPC speeding the whole thing up significantly
	    String seqSaveFileName = saveFileName + "_transcripts_" + arr[j] + "_" + (reverseStrand ? "r" : "f") + (chr==-1 ? "" : "_" + chr) + ".fa";
	    
	    //reducedModel.saveSequences(orchestra, seqSaveFileName, arr[j], minAAs);
	    if (runtime!=null) { //Run CPC to analyze if the sequences are
		String cpcOutDir = seqSaveFileName.substring(0, seqSaveFileName.lastIndexOf("/") + 1) + "CPC/" + (reverseStrand ? "r" : "f") + "_" + chr + "/" + arr[j] + "/";
		(new File(cpcOutDir)).mkdirs();
		String cpcCmd = "bsub -o /dev/null -q all_7d ./run_cpc.sh " + seqSaveFileName + " " + cpcOutDir + seqSaveFileName.substring(seqSaveFileName.lastIndexOf("/") + 1, seqSaveFileName.lastIndexOf(".")) + " " + cpcOutDir;
		System.out.println("Running CPC:\n" + cpcCmd);
		Process cpcProcess = runtime.exec(cpcCmd);
		//cpcProcess.waitFor();
	    }
	}
    }


    
    public static void saveGeneLatexTable(String geneLatexTableFileName, int totReads, int nUniqueGenes, int meanUniqueLengths, int totUniqueReads, int nSplitGenes, int meanSplitLengths, int totSplitReads, int nMergedGenes, int meanMergedLengths, int totMergedReads, int nMissedGenes, int meanMissedLength, int totMissedReads) { 
	DecimalFormat format = new DecimalFormat("###.###");
	try {
	    System.out.println("Saving summary statistics for genes to " + geneLatexTableFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(geneLatexTableFileName));
	    bw.write("\\begin{tabular}{lrcrr}\n\t\\textbf{Category} & \\multicolumn{2}{c}{\\textbf{\\# genes}} & \\textbf{$<$Length$>$} & \\textbf{Reads} \\\\\n\t\\hline\\\\\n");
	    double lenSum = 0;
	    int nReads = totUniqueReads + totMergedReads + totSplitReads + totMissedReads;
	    int nGenes = nUniqueGenes + nMergedGenes + nSplitGenes + nMissedGenes;
	    bw.write(RefSeqGene.UNIQUE_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nUniqueGenes + " & (" + format.format((double)nUniqueGenes/nGenes) + ") & " + meanUniqueLengths + " & " + format.format((double)totUniqueReads/totReads) + "\\\\\n");
	    bw.write(RefSeqGene.SPLIT_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nSplitGenes + " & (" + format.format((double)nSplitGenes/nGenes) + ") & " + meanSplitLengths + " & " + format.format((double)totSplitReads/totReads) + "\\\\\n");
	    bw.write(RefSeqGene.MERGED_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nMergedGenes + " & (" + format.format((double)nMergedGenes/nGenes) + ") & " + meanMergedLengths + " & " + format.format((double)totMergedReads/totReads) + "\\\\\n");
	    bw.write(RefSeqGene.MISSED_TRANSCRIPT_GENE_STR.replace("_", " ") + " & " + nMissedGenes + " & (" + format.format((double)nMissedGenes/nGenes) + ") & " + meanMissedLength + " & " + format.format((double)totMissedReads/totReads) + "\\\\\n");
	    bw.write("\\hline\n");
	    bw.write("Total & " + nGenes + " & & & " + format.format((double)nReads/totReads) + "\n");
	    bw.write("\\end{tabular}\n"); //footer
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving latex table " + e.toString());
	}
    }


    public static void saveTranscriptProperties(ReducedModel reducedModel, int binSize, String peakFileNames[], boolean orchestra, String saveFileName, boolean promoterLocationCategories, boolean reverseStrand, int chr, boolean conservation) { 
	//saveTranscriptProperty(reducedModel.getTranscribedRegionTypesLengths(), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_LENGTHS_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	//saveTranscriptProperty(reducedModel.getTranscribedRegionTypesDensities(), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_DENSITIES_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	//saveTranscriptProperty(reducedModel.getTranscribedRegionTypesWithinDensityStdDevs((int)Math.pow(2, binSize)), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_WITHIN_DENISTY_STD_DEV_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	if (peakFileNames!=null) {
	    //for (int i=0; i<peakFileNames.length; i++)
		//saveTranscriptProperty(reducedModel.getTranscribedRegionTypesPeaks(peakFileNames[i]), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + "_" + peakFileNames[i] + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	}
	if (conservation) {
	    //reducedModel.calculateConservation(orchestra);
	    //saveTranscriptProperty(reducedModel.getTranscribedRegionTypesConservation(), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_CONSERVATION_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	    //saveTranscriptProperty(reducedModel.getTranscribedRegionTypesWithinConservationStdDevs(binSize), promoterLocationCategories, saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + TRANSCRIPT_WITHIN_CONSERVATION_STD_DEV_SUFFIX + (chr!=-1 ? "_" + chr : "") + ".txt", chr==1 || chr==-1);
	}
    }

    public static void saveTranscriptProperty(TreeMap<String, TallyStore> type2data, boolean promoterLocationCategories, String saveFileName, boolean header) {
	//Use this method to save data of a specific transcript
	//property to a file. Each column represents a certain type of
	//transcript and they are padded with NaN to all have the same
	//number of rows which will make it easier to merge.
	try {
	    System.out.println("Saving to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    String[] arr = promoterLocationCategories ? TranscribedRegion.LOCATION_TYPES_ARRAY_STR : TranscribedRegion.TYPES_ARRAY_STR;
	    if (header) {
		bw.write("#");
		for (int j=0; j<arr.length; j++) 
		    bw.write(arr[j] + "\t");
		bw.write("\n");
	    }
	    int maxCount = 0; //Find the category with the largest no elements
	    for (TallyStore t : type2data.values())
		maxCount = maxCount>t.numberObs() ? maxCount : t.numberObs();
	    for (int i=0; i<maxCount; i++) {
		for (int j=0; j<arr.length; j++) {
		    if (type2data.containsKey(arr[j]) && i<type2data.get(arr[j]).numberObs() && !Double.isNaN(type2data.get(arr[j]).getArray()[i]) && !Double.isInfinite(type2data.get(arr[j]).getArray()[i]))
			bw.write(format.format(type2data.get(arr[j]).getArray()[i]) + "\t");
		    else
			bw.write("NaN\t");
		}
		bw.write("\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving " + e.toString());
	}
    }

    public static void saveNoncodingASInfo(ReducedModel reducedModel, String saveFileName) {
	//Save interesting info about Noncoding_AS transcripts wrt their overlapping genes.
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    System.out.println("Saving Noncoding_AS info to " + saveFileName);
	    //if (reducedModel.getChromosome()==1)
	    //	bw.write("#chr\tstart\tend\tstrand\tGeneID\tName\tgene_start\tgene_end\tPrev_exon_dist\tNext_exon_dist\tPrev_exon_no\tnExons\n");
	    /*for (TranscribedRegion t : reducedModel.getSegments(TranscribedRegion.NON_CODING_AS_STR)) {
		RefSeqGene g = t.getASGene();
		//Find out where we start wrt the exons of the overlapping gene
		int minExonDist = Integer.MAX_VALUE, minExonDistInd = -1;
		for (int i=0; i<g.numberOfExons(); i++) {
		    int dist = g.getExonStart(i) - t.getStart();
		    if (Math.abs(dist)<Math.abs(minExonDist)) {
			minExonDist = dist;
			minExonDistInd = i;
		    }
		}
		bw.write(reducedModel.getChromosome() + "\t" + t.getStart() + "\t" + t.getEnd() + "\t" + (t.getStrand() ? "1" : "0") + "\t" + g.getID() + "\t" +  g.getName() + "\t" + g.getStart() + "\t" + g.getEnd() + "\t" + minExonDist + "\t" + (t.getStart() + minExonDist) + "\t" + minExonDistInd + "\t" + g.numberOfExons() + "\n");
		}*/
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving " + e.toString());
	}
    }

    public static String printStats(double fdr, ReducedModel reducedModel, TreeMap<TranscribedRegion, Peak> transcriptsWithPeaks) { return printStats(fdr, reducedModel, transcriptsWithPeaks, null); }

    public static String printStats(double fdr, ReducedModel reducedModel, TreeMap<TranscribedRegion, Peak> transcriptsWithPeaks, Vector<Peak> peaks) {
	//n_regions, proportion transcribed, proportion of reads
	String str = format.format(fdr) + "\t" + reducedModel.numberOfTranscribedRegions() + "\t" + format.format(reducedModel.proportionTranscribed()) + "\t" + format.format(reducedModel.proportionOfReads()) + "\t";
	//PPV sensitivity
	//str += format.format(reducedModel.getPositivePredictiveValue()) + "\t";
	//n_unique, n_missed_genes, n_genes_w_multiple
	str += format.format(reducedModel.numberOfGenesWithOneTranscript()) + "\t" + format.format(reducedModel.numberOfGenesWithNTranscripts(0, false, false)) + "\t" + format.format(reducedModel.numberOfGenesWithNTranscripts(2, true, false)) + "\t";
	//n_nc_transcripts, n_transcripts_covering_multiple
	str += format.format(reducedModel.numberOfTranscriptsWithNGenes(0, false, false)) + "\t" + format.format(reducedModel.numberOfTranscriptsWithNGenes(2, true, false)) + "\t";
	//proportion with peaks
	if (peaks!=null)
	    str += format.format((double)transcriptsWithPeaks.size()/(peaks.size() - transcriptsWithPeaks.size() + reducedModel.numberOfTranscribedRegions()));
	else
	    str += format.format((double)transcriptsWithPeaks.size()/reducedModel.numberOfTranscribedRegions());
	System.out.println(str);
	return str;
    }

    public static void createWigFile(String saveFileName, boolean chip, boolean reverseStrand, int chr, boolean bayes) {
	String saveFile = saveFileName + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".bed";
	String name = saveFileName.substring(saveFileName.lastIndexOf("mRNA/") + 5) + BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f");
	String colorStr = CreateWigFiles.getColorString(saveFileName);
	int priority = CreateWigFiles.getPriority(saveFileName);
	String trackInfoStr = "track type=bigWig name=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" description=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" priority=" + priority + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=" + URL_NAME + name + (chr==-1 ? "" : "_" + chr) + ".bw";
	CreateWigFiles.addTrackInfo(trackInfoStr, TRACK_INFO_FILE_PATH + (saveFileName.contains("MeCP2") ? "MeCP2/" : saveFileName.substring(saveFileName.lastIndexOf("mRNA/") + 4, saveFileName.lastIndexOf("/") + 1)) + TRACK_INFO_FILE_NAME);
	//CreateWigFiles.createWWWLink(saveFile, "/www/" + WWW_DIR_NAME + (saveFileName.contains("MeCP2") ? "MeCP2/" : saveFileName.substring(saveFileName.lastIndexOf("mRNA/") + 5, saveFileName.lastIndexOf("/") + 1)));
	//createWigFile(saveFileName, name, chr);
    }

    /*public static void createWigFile(String saveFileName, int chr) { createWigFile(saveFileName, saveFileName.replace(".bed", ""), chr); }

    public static void createWigFile(String saveFileName, String name, int chr) {
	String colorStr = CreateWigFiles.getColorString(saveFileName);
	int priority = CreateWigFiles.getPriority(saveFileName);
	String trackInfoStr = "track type=bigWig name=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" description=\"" + name + (chr==-1 ? "" : "_" + chr) + "\" priority=" + priority + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=" + URL_NAME + name + (chr==-1 ? "" : "_" + chr) + ".bw";
	CreateWigFiles.addTrackInfo(trackInfoStr, TRACK_INFO_FILE_PATH + (saveFileName.contains("MeCP2") ? "MeCP2/" : "") + TRACK_INFO_FILE_NAME);
	CreateWigFiles.createWWWLink(saveFile, "/www/" + WWW_DIR_NAME + (saveFileName.contains("MeCP2") ? "MeCP2/" : ""));
	}*/

    public static int numberOfChromosomes(String species) {
	if (species.equals("hs") || species.equals("human"))
	    return 24;
	else if (species.equals("sc") || species.equals("cerevisiae"))
	    return 17;
	else if (species.equals("pf") || species.equals("plasmodium"))
	    return 15;
	return 21; //default is mouse
    }

    public static int[] getLengthVec(int lMin, int lMax, int deltaL) {
	int lVec[] = new int[(lMax-lMin)/deltaL];
	for (int i=0; i<(lMax-lMin)/deltaL; i++)
	    lVec[i] = lMin + i*deltaL;
	return lVec;
    }

    public static double[] getFDRVec(double fdrMin, double fdrMax, int nFDR) {
	//create nFDR points equally spaced on a log-scale btwn fdrMin and fdrMax
	if (nFDR<=1)
	    return new double[]{fdrMin};
	double fdrVec[] = new double[nFDR], deltaFDR = (Math.log10(fdrMax) - Math.log10(fdrMin))/(nFDR-1);
	for (int i=0; i<nFDR; i++)
	    fdrVec[i] = Math.pow(10, Math.log10(fdrMin) + deltaFDR*i);
	return fdrVec;
    }

    public static double[] getCompareVec(double compareMin, double compareMax, int nCompare) {
	if (nCompare<=1)
	    return new double[]{compareMin};
	double compareVec[] = new double[nCompare], delta = (compareMax - compareMin)/(nCompare - 1);
	for (int i=0; i<nCompare; i++)
	    compareVec[i] = compareMin + delta*i;
	return compareVec;
    }

    ///////////////////////////////////////////////////////



    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName) { return parseEnhancerFile(fileName, "Mm"); }
		  
    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName, String species) { return parseEnhancerFile(fileName, species, -1); }
		  
    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName, int chr) { return parseEnhancerFile(fileName, "Mm", chr); }

    public static TreeSet<Enhancer2> parseEnhancerFile(String fileName, String species, int chr) {
	//Use the subset files by JMG that contain chr, pos and TRUE|FALSE
	TreeSet<Enhancer2> enhancers = new TreeSet<Enhancer2>();
	if (fileName!=null) {
	    try {
		BufferedReader br = new BufferedReader(new FileReader(fileName));
		String line = br.readLine();
		//chr pos TRUE|FALSE
		Pattern enhancerPattern = Pattern.compile("^(\\d+)\\s+(\\d+)\\s+(\\w+)");
		while (line!=null) {
		    Matcher m = enhancerPattern.matcher(line);
		    if (m.find()) {
			int c = Integer.parseInt(m.group(1));
			if ((c==chr || chr==-1) && m.group(3).equals("TRUE")) 
			    enhancers.add(new Enhancer2(c, Integer.parseInt(m.group(2))));
		    }
		    line = br.readLine();
		}
		br.close();
	    }
	    catch (IOException e) {
		System.err.println("Error reading enhancers from " + fileName + ": " + e.toString());
	    }
	}
	System.out.println("Found " + enhancers.size() + " enhancers.");
	return enhancers;
    }

public static ABIRead[] downSampleReads(ABIRead[] reads, double keepFraction) {
    int nKeep = (int)Math.round((double)reads.length*keepFraction);
	//Permute all the indices, then take the nKeep first ones and
	//retain those reads
	int inds[] = new int[reads.length];
	for (int i=0; i<reads.length; i++)
	    inds[i] = i;
	for (int i=0; i<reads.length*10; i++) {
	    int ind1 = random.nextInt(reads.length), ind2 = random.nextInt(reads.length);
	    int tmp = inds[ind1];
	    inds[ind1] = inds[ind2];
	    inds[ind2] = tmp;
	}
	ABIRead rs[] = new ABIRead[nKeep];
	for (int i=0; i<nKeep; i++)
	    rs[i] = reads[inds[i]];
	Arrays.sort(rs);
	System.out.println("Sampled " + rs.length + " reads.");
	return rs;
    }

}

