//Use this class for a graph which contains genes and transcripts. Use
//one graph for each chromosome.


import edu.uci.ics.jung.graph.impl.*;
import edu.uci.ics.jung.exceptions.*;
import edu.uci.ics.jung.utils.*;

import java.util.*;
import java.io.*;
import java.text.DecimalFormat;
import java.util.regex.*;

public class GeneTranscriptsGraph {

    public static final double MIN_READS_PROPORTION = .5;

    //Use only one critical distance for all purposes
    protected int chr, cutOffDistance;
    protected boolean forward;
    protected RefSeqGenes geneInfo;
    //Stupid BipartiteGraph class does not work .....
    protected UndirectedSparseGraph graph = new UndirectedSparseGraph();
    //Keep track of the genes that are not uniquely assigned to transcripts
    protected TreeSet<RefSeqGene> mergedGenes = new TreeSet<RefSeqGene>(), splitGenes = new TreeSet<RefSeqGene>(), missedGenes = new TreeSet<RefSeqGene>(), uniqueGenes = new TreeSet<RefSeqGene>();
    protected TreeSet<Enhancer2> enhancers;


    protected Random random = new Random();
    public static DecimalFormat format = new DecimalFormat("###.#####");

    ///////////////////////////////////////////////////////

    public GeneTranscriptsGraph(RefSeqGenes gi, int c, boolean f, int d, TreeSet<TranscribedRegion> segments, TreeSet<LncRNA> lncRNAs, TreeSet<MiRNA> miRNAs, TreeSet<MacroRNA> macroRNAs, TreeSet<Enhancer2> ens) { 
	geneInfo = gi;
	chr = c; 
	forward = f;
	cutOffDistance = d;
	for (TranscribedRegion t : segments) 
	    addVertex(t);
	for (RefSeqGene g : geneInfo.getGenes()) 
	    addVertex(g);
	if (lncRNAs!=null)
	    addLncRNAInfo(lncRNAs);
	if (miRNAs!=null)
	    addMiRNAInfo(miRNAs);
	if (macroRNAs!=null)
	    addMacroRNAInfo(macroRNAs);
	enhancers = ens;
	if (enhancers!=null)
	  addEnhancerInfo(enhancers);
    }

    public GeneTranscriptsGraph(TreeSet<TranscribedRegion> segments, RefSeqGenes gi, boolean f, int maxDist) {
	chr = geneInfo.getChromosome();
	geneInfo = gi;
	forward = f;
	cutOffDistance = maxDist;
	for (TranscribedRegion t : segments) {
	    //if (t.getChromosome()==chr)
		graph.addVertex(t);
	}
	for (RefSeqGene g : geneInfo.getGenes())
	    graph.addVertex( g );
    }

    ///////////////////////////////////////////////////////

    public TreeSet<RefSeqGene> getSplitGenes() { return splitGenes; }

    public TreeSet<RefSeqGene> getMergedGenes() { return mergedGenes; }

    public TreeSet<RefSeqGene> getMissedGenes() { return missedGenes; }

    public TreeSet<RefSeqGene> getUniqueGenes() { return uniqueGenes; }

    public boolean addVertex(UndirectedSparseVertex v) {
	HashSet<UndirectedSparseGraph> verts = new HashSet<UndirectedSparseGraph>(graph.getVertices());
	if (!verts.contains(v)) {
	    graph.addVertex(v);
	    //Add connections
	    if (v instanceof RefSeqGene) {
		RefSeqGene g = (RefSeqGene)v;
		for (TranscribedRegion t : getTranscripts()) {
		    if (assignEdge(g, t)) {
			try {
			    graph.addEdge(new UndirectedSparseEdge(g, t));
			}
			catch (Exception e) { 
			    System.err.println("Failed to add edge between " + g.toString() + " and " + t.toString() + " because " + e.toString()); 
			}
		    }
		}
	    }
	    else if (v instanceof LncRNA) {
		LncRNA l = (LncRNA)v;
		for (TranscribedRegion t : getTranscripts()) {
		    if (l.overlaps(t.getStart(), t.getEnd())) {
			try {
			    graph.addEdge(new UndirectedSparseEdge(l, t));
			}
			catch (Exception e) { 
			    System.err.println("Failed to add edge between " + l.toString() + " and " + t.toString() + " because " + e.toString()); 
			}
		    }
		}
	    }
	    else if (v instanceof MiRNA) {
		MiRNA m = (MiRNA)v;
		for (TranscribedRegion t : getTranscripts()) {
		    if (m.overlaps(t.getStart(), t.getEnd())) {
			try {
			    graph.addEdge(new UndirectedSparseEdge(m, t));
			}
			catch (Exception e) { 
			    System.err.println("Failed to add edge between " + m.toString() + " and " + t.toString() + " because " + e.toString()); 
			}
		    }
		}
	    }
	    else if (v instanceof MacroRNA) {
		MacroRNA m = (MacroRNA)v;
		for (TranscribedRegion t : getTranscripts()) {
		    if (m.overlaps(t.getStart(), t.getEnd())) {
			try {
			    graph.addEdge(new UndirectedSparseEdge(m, t));
			}
			catch (Exception e) { 
			    System.err.println("Failed to add edge between " + m.toString() + " and " + t.toString() + " because " + e.toString()); 
			}
		    }
		}
	    }
	    else if (v instanceof Enhancer2) {
		Enhancer2 e = (Enhancer2)v;
		for (TranscribedRegion t : getTranscripts()) {
		    if (e.overlaps(-cutOffDistance + (forward ? t.getStart() : t.getEnd()), cutOffDistance + (forward ? t.getStart() : t.getEnd()))) { //should start near enhancer
			System.out.println("Found enhancer " + e.toString() + " near " + t.toString());
			try {
			    graph.addEdge(new UndirectedSparseEdge(e, t));
			}
			catch (Exception ex) { 
			    System.err.println("Failed to add edge between " + e.toString() + " and " + t.toString() + " because " + ex.toString()); 
			}
		    }
		}
	    }
	    return true;
	}
	return false;
    }

    public boolean assignEdge(RefSeqGene g, TranscribedRegion t) {
	//Return true if we should put an edge between the gene and the transcript
	if (g.getStrand()==forward && g.getFactorReads().length>0)
	    return g.overlaps(t.getStart(), t.getEnd());
	else { //To find divergent transcripts, we must allow for transcripts starting ahead of the AS gene as well
	    if (forward) //Gene is on reverse strand
		return g.overlaps(t.getStart() - cutOffDistance, t.getEnd());
	    else
		return g.overlaps(t.getStart(), t.getEnd() + cutOffDistance);
	}
    }

    public void addMiRNAInfo(TreeSet<MiRNA> miRNAs) {
	int n = graph.getVertices().size();
	for (MiRNA m : miRNAs) {
	    if (m.getChromosome()==chr) 
		addVertex(m);
	}
	//System.out.println("Addded " + (graph.getVertices().size() - n) + " miRNAs."); 
    }

    public void addMacroRNAInfo(TreeSet<MacroRNA> macroRNAs) {
	int n = graph.getVertices().size();
	for (MacroRNA m : macroRNAs) {
	    if (m.getChromosome()==chr) 
		addVertex(m);
	}
	//System.out.println("Addded " + (graph.getVertices().size() - n) + " miRNAs."); 
    }

    public void addEnhancerInfo(TreeSet<Enhancer2> enhancers) {
	int n = graph.getVertices().size();
	for (Enhancer2 e : enhancers) {
	    if (e.getChromosome()==chr) 
		addVertex(e);
	}
	System.out.println("Added " + (graph.getVertices().size() - n) + " enhancers."); 
    }

    public void addLncRNAInfo(TreeSet<LncRNA> lncRNAs) {
	int n = graph.getVertices().size();
	for (LncRNA l : lncRNAs) {
	    if (l.getChromosome()==chr)
		addVertex(l);
	}
	//System.out.println("Addded " + (graph.getVertices().size() - n) + " lincRNAs.");
    }

    /////////////////////////////////

    public void removeAllEdges() { graph.removeAllEdges(); }

    public void removeAllVertices() { graph.removeAllVertices(); }

    public void removeAll() { graph.removeAllEdges(); graph.removeAllVertices(); }

    public boolean contains(UndirectedSparseVertex v) { return graph.getVertices().contains(v); }

    public int numberOfGenesWithNTranscripts(int n, boolean ineq, boolean unique) {
	//Find the number of genes which are overlapped by n
	//transcripts. If the boolean is set, then we require that the
	//matching transcripts all have out-degree 1.
	int m = 0;
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(graph.getVertices())) {
	    if (v instanceof RefSeqGene) {
		if ((!ineq && v.inDegree()==n) || (ineq && v.inDegree()>n)) {
		    if (!unique)
			m++;
		    else {
			Iterator<UndirectedSparseVertex> it = v.getNeighbors().iterator();
			boolean uniqueTranscripts = true;
			while (it.hasNext()) 
			    uniqueTranscripts = uniqueTranscripts && it.next().numNeighbors()==1;
			m += uniqueTranscripts ? 1 : 0;
		    }
		}
	    }
	}
	return m;
    }
    
    public int numberOfGenesWithOneTranscript() {
	//Make sure that the the majority of the gene is covered by
	//only one transcript and that the transcript in turn only
	//covers the minority of other genes.
	int m = 0, nGenes = 0;
	for (TranscribedRegion t : getTranscripts()) {
	    if (t.getType().equals(TranscribedRegion.CODING_STR)) 
		m++;
	    /*if (v instanceof RefSeqGene) {
		RefSeqGene g = (RefSeqGene)v;
		//check how much overlap each transcript has
		TranscribedRegion t = g.getUniqueTranscript();
		m += t!=null && t.inDegree()==1 ? 1 : 0;
		nGenes++;
		}*/
	}
	//System.out.println("Found " + nGenes + " genes and " + m + " unique.");
	return m;
    }

    public static int overlap(RefSeqGene g, TranscribedRegion t) {
	int over = g.getStrand() ? Math.min(g.getEnd(), t.getEnd()) - Math.max(g.getStart(), t.getStart()) : Math.min(g.getStart(), t.getEnd()) - Math.max(g.getEnd(), t.getStart());
	//System.out.println("Checking overlap: " + g.toString() + "\t" + Math.min(g.getEnd(), t.getEnd()) + "-" + Math.max(g.getStart(), t.getStart()) + "=" + over + "\t" + t.toString());
	return over>0 ? over : 0;
    }

    public int numberOfTranscriptsWithOneGene() { return numberOfTranscriptsWithOneGene(true); }

    public int numberOfTranscriptsWithOneGene(boolean unique) { return numberOfTranscriptsWithNGenes(1, false, unique); }

    public int numberOfTranscriptsWithNGenes(int n, boolean ineq, boolean unique) {
	//Find the number of transcripts which are overlapped by n
	//genes. If the boolean unique is set, then we require that the
	//matching genes all have out-degree 1. If ineq is true, we
	//are interested in the case >=n.
	int m = 0;
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(graph.getVertices())) {
	    if (v instanceof TranscribedRegion) {
		if ((!ineq && v.inDegree()==n) || (ineq && v.inDegree()>n)) {
		    //System.out.println(v.toString() + " overlaps " + v.outDegree());
		    if (!unique)
			m++;
		    else {
			Iterator<UndirectedSparseVertex> it = v.getNeighbors().iterator();
			boolean uniqueTranscripts = true;
			while (it.hasNext()) 
			    uniqueTranscripts = uniqueTranscripts && it.next().numNeighbors()==1;
			m += uniqueTranscripts ? 1 : 0;
		    }
		}
	    }
	}
	//System.out.println(m + " transcripts overlap only one gene.");
	return m;
    }

    public void findGenesPerTranscript() { 
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(graph.getVertices())) {
	    if (v instanceof TranscribedRegion) {
		TranscribedRegion t = (TranscribedRegion)v;
		if (t.numNeighbors()==0)
		    System.out.println(t.toString() + " does not overlap any genes.");
		else {
		    Iterator<UndirectedSparseVertex> it = t.getNeighbors().iterator();
		    while (it.hasNext()) {
			RefSeqGene g = (RefSeqGene)it.next();
			System.out.println(t.toString() + " is overlapped for " + overlap(g, t) + "(" + format.format((double)overlap(g, t)/g.getLength()) + ") by " + g.toString());
		    }
		}
	    }
	}
    }

    public void findTranscriptsPerGene() {
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(graph.getVertices())) {
	    if (v instanceof RefSeqGene) {
		RefSeqGene g = (RefSeqGene)v;
		if (g.numNeighbors()==0)
		    System.out.println(g.toString() + " does not overlap any transcripts.");
		else {
		    Iterator<UndirectedSparseVertex> it = g.getNeighbors().iterator();
		    while (it.hasNext()) {
			TranscribedRegion t = (TranscribedRegion)it.next();
			//System.out.println((g.hasUniqueTranscript() ? "* " : "  ") + g.toString() + " is overlapped for " + overlap(g, t) + "(" + format.format((double)overlap(g, t)/t.getLength()) + ") by " + t.toString());
		    }
		}
	    }
	}
    }
    
    public TreeSet<TranscribedRegion> getTranscripts() { 
	TreeSet<TranscribedRegion> ts = new TreeSet<TranscribedRegion>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof TranscribedRegion)
		ts.add((TranscribedRegion)v);
	}
	return ts;
    }

    public Vector<RefSeqGene> getGenes(int geneIDs[]) {
	//only get the genes that are in the subset geneIDs
	Vector<RefSeqGene> genes = getGenes(), gs = new Vector<RefSeqGene>();
	if (geneIDs==null)
	    gs = genes;
	else {
	    for (RefSeqGene g : genes) {
		if (Arrays.binarySearch(geneIDs, g.getID())>=0)
		    gs.addElement(g);
	    }
	}
	return gs;
    }

    public Vector<RefSeqGene> getGenes() { 
	Vector<RefSeqGene> gs = new Vector<RefSeqGene>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof RefSeqGene)
		gs.addElement((RefSeqGene)v);
	}
	return gs;
    }

    public Vector<LncRNA> getLncRNAs() { 
	Vector<LncRNA> ls = new Vector<LncRNA>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof LncRNA)
		ls.addElement((LncRNA)v);
	}
	return ls;
    }

    public Vector<MiRNA> getMiRNAs(boolean precursors) { 
	Vector<MiRNA> ms = new Vector<MiRNA>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof MiRNA && (!precursors || ((MiRNA)v).hasPrecursor())) 
		    ms.addElement((MiRNA)v);
	}
	return ms;
    }

    public static boolean geneIsolated(RefSeqGenes geneInfo, RefSeqGene g, int isolatedDist) {
	if (isolatedDist>0)
	    return geneInfo.getNearestGeneDistance(g.getStart())<isolatedDist && geneInfo.getNearestGeneDistance(g.getEnd())<isolatedDist;
	return true;
    }

    public Vector<LncRNA> getLincRNAs() {
	Vector<LncRNA> lincRNAs = new Vector<LncRNA>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof LncRNA)
		lincRNAs.addElement((LncRNA)v);
	}
	return lincRNAs;
    }

    public Vector<MiRNA> getMiRNAs() {
	Vector<MiRNA> miRNAs = new Vector<MiRNA>();
	Vector<UndirectedSparseVertex> vs = new Vector<UndirectedSparseVertex>(graph.getVertices()); 
	for (UndirectedSparseVertex v : vs) {
	    if (v instanceof MiRNA)
		miRNAs.addElement((MiRNA)v);
	}
	return miRNAs;
    }


    public boolean overlapsLncRNA(TranscribedRegion t) { 
	for (LncRNA l : getLncRNAs()) {
	    if (l.overlaps(t.getStart(), t.getEnd()))
		return true;
	}
	return false;
    }

    public boolean overlapsMiRNA(TranscribedRegion t, boolean precursor) { 
	for (MiRNA m : getMiRNAs(precursor)) {
	    if (m.getStrand()==forward && m.overlaps(t.getStart(), t.getEnd()))
		return true;
	}
	return false;
    }

    public static boolean overlapsLncRNA(int chr, TreeSet<LncRNA> lncRNAs, TranscribedRegion t) { return overlappingLncRNA(chr, lncRNAs, t)!=null; }

    public static LncRNA overlappingLncRNA(int chr, TreeSet<LncRNA> lncRNAs, TranscribedRegion t) {
	if (lncRNAs==null || t==null)
	    return null;
	for (LncRNA lnc : lncRNAs) {
	    if (lnc.getChromosome()==chr) {
		if ((t.getStart()>lnc.getStart() && t.getStart()<lnc.getEnd()) || (t.getEnd()>lnc.getStart() && t.getEnd()<lnc.getEnd())) //Start or end of region inside lncRNA
		    return lnc;
	    }
	}
	return null;
    }
    
    public boolean nearEnhancer(RefSeqGenes geneInfo, TranscribedRegion t, int maxDist) { return nearEnhancer(this.chr, this.forward, geneInfo, this.enhancers, t, maxDist); }

    public static boolean nearEnhancer(int chr, boolean forward, RefSeqGenes geneInfo, TreeSet<Enhancer2> enhancers, TranscribedRegion t, int maxDist) {
	if (enhancers==null || t==null)
	    return false;
	for (Enhancer2 e : enhancers) {
	    RefSeqGene overlappingGene = geneInfo.getOverlappingGene(e.getCenter());
	    //chromosome, distance and make sure anti-sense strand for intragenic ones
	    if (e.getChromosome()==chr && Math.abs(t.getStart() - e.getCenter())<maxDist) {
		if (overlappingGene==null) {
		    //System.out.println("Found eRNA " + t.toString() + " near enhancer " + e.toString());
		    return true;
		}
		else if (overlappingGene.getStrand()==forward)
		    return true;
	    }
	}
	return false;
    }

    public void categorizeTranscribedRegions(boolean promoterLocationCategories) {
	//Go through all the transcripts and try to classify them
	for (TranscribedRegion t : getTranscripts()) {
	    TreeMap<RefSeqGene, Double> genes2reads = t.getOverlappingGenes(true);
	    if (isUniqueCoding(genes2reads, t)) {
		t.setType(TranscribedRegion.CODING_STR);
		uniqueGenes.add(t.getUniqueGene(true));
	    }
	    else if (isMultiGene(genes2reads, t)) {
		t.setType(TranscribedRegion.MULTI_GENE_STR);
		mergedGenes.addAll(genes2reads.keySet());
	    }
	    else if (isCodingFragment(genes2reads, t)) {
		t.setType(TranscribedRegion.CODING_FRAGMENT_STR);
		splitGenes.addAll(genes2reads.keySet());
	    }
	    else { //Non-coding, try to determine what type
		if (promoterLocationCategories) {
		    //Only need to find out if we are starting near a TSS
		    TreeMap<RefSeqGene, Double> genes2readsAS = t.getOverlappingGenes(false);
		    RefSeqGene asGene = getDivergent(genes2readsAS, t, true);
		    if (asGene!=null) {
			t.setType(TranscribedRegion.PROMOTER_PROXIMAL_NCRNA_STR);
			t.setDivergentGene(asGene);
		    }
		    else
			t.setType(TranscribedRegion.PROMOTER_DISTAL_NCRNA_STR);
		}
		else {
		    TreeMap<RefSeqGene, Double> genes2readsAS = t.getOverlappingGenes(false);
		    RefSeqGene asGene = getDivergent(genes2readsAS, t, true);
		    if (asGene!=null) {
			t.setType(TranscribedRegion.DIVERGENT_STR);
			t.setDivergentGene(asGene);
		    }
		    else {
			/*if (isERNA(t))
			  t.setType(TranscribedRegion.ERNA_STR);*/
			/*asGene = getDivergent(genes2readsAS, t, false);
			  if (asGene!=null) {
			  t.setType(TranscribedRegion.DIVERGENT_END_STR);
			  t.setASGene(asGene);
			  }
			  else {*/
			TreeMap<MacroRNA, Double> macroRNA2reads = t.getOverlappingMacroRNAs();
			if (isUniqueMacroRNA(macroRNA2reads, t) || isMacroRNAFragment(macroRNA2reads, t))
			    t.setType(TranscribedRegion.MACRO_RNA_STR);
			//else if (isMacroRNAFragment(macroRNA2reads, t))
			//  t.setType(TranscribedRegion.MACRO_RNA_FRAGMENT_STR);
			else {
			    TreeMap<LncRNA, Double> lncRNA2reads = t.getOverlappingLncRNAs();
			    if (isUniqueLncRNA(lncRNA2reads, t) || isLncRNAFragment(lncRNA2reads, t))
				t.setType(TranscribedRegion.LNC_RNA_STR);
			    //else if (isLncRNAFragment(lncRNA2reads, t))
			    //t.setType(TranscribedRegion.LNC_RNA_FRAGMENT_STR);
			    else if (overlapsMiRNA(t, true)) {
				t.setType(TranscribedRegion.MI_RNA_PRECURSOR_STR);
				//System.out.println("Found " + t.toString() + ", " + t.getType());
			    }
			    else if (overlapsMiRNA(t, false)) {
				t.setType(TranscribedRegion.MI_RNA_STR);
				//System.out.println("Found " + t.toString() + ", " + t.getType() + ", " + genes2reads.size() + ", " + genes2readsAS.size());
			    }
			    else if (nearEnhancer(geneInfo, t, cutOffDistance)) {
				t.setType(TranscribedRegion.ERNA_STR);
				System.out.println("Found " + t.toString() + ", " + t.getType());
			    }
			    else {
				asGene = getOverlappingAS(genes2readsAS, t);
				if (asGene!=null) {
				    t.setType(TranscribedRegion.NON_CODING_AS_STR);
				    t.setASGene(asGene);
				    System.out.println(t.toString() + " has type " + t.getType() + " and " + genes2readsAS.size() + " and AS is " + asGene.toString());
				}
				else
				    t.setType(TranscribedRegion.NON_CODING_OTHER_RNA_STR);
			    }
			}
		    }
		}
	    }
	    //System.out.println(t.toString() + " has type " + t.getType() + " and " + genes2reads.size());
	}
	for (RefSeqGene g : geneInfo.getGenes(true, forward)) { //Find out which genes were missed and store those in a list
	    if (!uniqueGenes.contains(g) && !mergedGenes.contains(g) && !splitGenes.contains(g))
		missedGenes.add(g);
	}
    }

    public boolean isDivergent(TreeMap<RefSeqGene, Double> genes2readsAS, TranscribedRegion t) { return isDivergent(genes2readsAS, t, true); }

    public boolean isDivergent(TreeMap<RefSeqGene, Double> genes2readsAS, TranscribedRegion t, boolean start) { return getDivergent(genes2readsAS, t, true)!=null; }

    public RefSeqGene getDivergent(TreeMap<RefSeqGene, Double> genes2readsAS, TranscribedRegion t, boolean start) {
	for (RefSeqGene g : genes2readsAS.keySet()) {
	    if (forward) {
		if (start) {
		    for (Integer s : g.getStarts()) { //Check all starts
			if (Math.abs(t.getStart() - s)<cutOffDistance)
			    return g;
		    }
		}
		else {
		    if (Math.abs(t.getStart() - g.getEnd())<cutOffDistance)
			return g;
		}
	    }
	    else {
		if (start) {
		    for (Integer s : g.getStarts()) { //Check all starts
			if (Math.abs(t.getEnd() - g.getStart())<cutOffDistance)
			    return g;
		    }
		}
		else {
		    if (Math.abs(t.getEnd() - g.getEnd())<cutOffDistance)
			return g;
		}
	    }
	}
	return null;
    }

    public static boolean isERNA(TranscribedRegion t) {
	Iterator it = t.getNeighbors().iterator();
	while (it.hasNext()) {
	    if (it.next() instanceof Enhancer2)
		return true;
	}
	return false;
    }

    public static boolean isOverlappingAS(TreeMap<RefSeqGene, Double> genes2readsAS, TranscribedRegion t) { return genes2readsAS.size()>0;    }

    public static RefSeqGene getOverlappingAS(TreeMap<RefSeqGene, Double> genes2readsAS, TranscribedRegion t) { return genes2readsAS.isEmpty() ? null : genes2readsAS.firstKey();    }
    
    public static boolean isUniqueCoding(TreeMap<RefSeqGene, Double> genes2reads, TranscribedRegion t) {
	//Determine if t overlaps a gene on the sense strand and
	//fulfils all the additional criteria for being a unique
	//transcript.
	if (genes2reads.size()==1) {
	    //Make sure that we cover at least half of the reads or
	    //half the length, and that no other transcripts cover the
	    //gene
	    RefSeqGene g = genes2reads.keySet().iterator().next();
	    if (genes2reads.get(g)>MIN_READS_PROPORTION)
		return g.numNeighbors()==1;
	    return false;
	}
	else if (genes2reads.size()>1 && genes2reads.size()<4) {
	    //Make sure that it includes most of the reads and only a fraction for the others
	    Double[] fractions = genes2reads.values().toArray(new Double[genes2reads.size()]);
	    Arrays.sort(fractions);
	    for (int i=0; i<fractions.length - 1; i++) {
		if (fractions[i]>MIN_READS_PROPORTION)
		    return false;
	    }
	    return fractions[fractions.length-1]>MIN_READS_PROPORTION;
	}
	return false;
    }

    public static boolean isUniqueLncRNA(TreeMap<LncRNA, Double> lncRNA2reads, TranscribedRegion t) {
	return lncRNA2reads.size()==1 && lncRNA2reads.values().iterator().next()>MIN_READS_PROPORTION;
    }

    public static boolean isUniqueMacroRNA(TreeMap<MacroRNA, Double> macroRNA2reads, TranscribedRegion t) {
	return macroRNA2reads.size()==1 && macroRNA2reads.values().iterator().next()>MIN_READS_PROPORTION;
    }

    public static boolean isMultiGene(TreeMap<RefSeqGene, Double> genes2reads, TranscribedRegion t) {
	//Check if multiple genes are overlapped
	int nOver = 0;
	for (Double f : genes2reads.values())
	    nOver += f>MIN_READS_PROPORTION ? 1 : 0;
	return nOver>1;
    }

    public static boolean isCodingFragment(TreeMap<RefSeqGene, Double> genes2reads, TranscribedRegion t) {
	//Check if at most one gene is overlapped and if that gene has
	//more fragments
	if (genes2reads.size()==1) 
	    return genes2reads.keySet().iterator().next().numNeighbors()>1 || genes2reads.values().iterator().next()<MIN_READS_PROPORTION;
	return false;
    }

    public static boolean isLncRNAFragment(TreeMap<LncRNA, Double> lncRNA2reads, TranscribedRegion t) {
	//Check if at most one gene is overlapped and if that gene has
	//more fragments
	if (lncRNA2reads.size()==1) 
	    return lncRNA2reads.keySet().iterator().next().numNeighbors()>1 || lncRNA2reads.values().iterator().next()>MIN_READS_PROPORTION;
	return false;
    }

    public static boolean isMacroRNAFragment(TreeMap<MacroRNA, Double> macroRNA2reads, TranscribedRegion t) {
	//Check if at most one gene is overlapped and if that gene has
	//more fragments
	if (macroRNA2reads.size()==1) 
	    return macroRNA2reads.keySet().iterator().next().numNeighbors()>1 || macroRNA2reads.values().iterator().next()>MIN_READS_PROPORTION;
	return false;
    }

}