//Use this method to scan for occurrences of specific motifs, eg AP1
//or CRE. Also score the motifs for PhastCons.
//
//Example:
//
//java -Xmx512m GenomeWideMotifs -m TGANTCA -C -o
//
//for ((i=1; i<=21; i++)) ; do bsub -q shared_2h -o /dev/null java -Xmx512m GenomeWideMotifs -o -m TGANTCA -C -c $i -s /home/mh168/Projects/ChIP-Seq/TKData/Athar/AP1Files/ap1_locs_$i.txt ; done
//./merge_files3.sh /home/mh168/Projects/ChIP-Seq/TKData/Athar/AP1Files/ap1 locs txt
//
//for ((i=1; i<=24; i++)) ; do bsub -q shared_2h -o /dev/null java -Xmx512m GenomeWideMotifs -o -m TGANTCA -C -c $i -s /home/mh168/Projects/ChIP-Seq/TKData/Athar/AP1Files/ap1_hg18_locs_$i.txt --hs ; done

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;


public class GenomeWideMotifs {

    public static DecimalFormat format = new DecimalFormat("###.######");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option verboseFlag = argParser.addBooleanOption('v', "verbose");
	    CmdLineParser.Option conservationFlag = argParser.addBooleanOption('C', "conservation");
	    CmdLineParser.Option hsFlag = argParser.addBooleanOption("hs");
	    CmdLineParser.Option hg19Flag = argParser.addBooleanOption("hg19");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option motifFileNameFlag = argParser.addStringOption('m', "motifFileName");
	    CmdLineParser.Option mismatchesFlag = argParser.addIntegerOption('M', "mismatches");
	    CmdLineParser.Option windowSizeFlag = argParser.addIntegerOption('w', "windowSize");
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean verbose = argParser.getOptionValue(verboseFlag)!=null;
	    boolean conservation = argParser.getOptionValue(conservationFlag)!=null;
	    boolean hs = argParser.getOptionValue(hsFlag)!=null;
	    boolean hg19 = argParser.getOptionValue(hg19Flag)!=null;
	    tmp = argParser.getOptionValue(mismatchesFlag);
	    int mismatches = tmp!=null ? ((Integer)tmp).intValue() : 0;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    tmp = argParser.getOptionValue(windowSizeFlag);
	    int windowSize = tmp!=null ? ((Integer)tmp).intValue() : 100;
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : "AP1Files/ap1_locs.txt";
	    tmp = argParser.getOptionValue(motifFileNameFlag);
	    String motifFileName = tmp!=null ? (String)tmp : "TGANTCA";
	    //////////////////////////////////////////////////////////////////
	    String species = hs ? "hs" : "mm";
	    Vector<String> motifs = new Vector<String>(), ms = new Vector<String>();
	    //Load the motifs/words
	    if (motifFileName.endsWith(".txt"))
		ms = FindMotifDistances.loadMotifs(motifFileName);
	    else
		ms.addElement(motifFileName);
	    for (String m : ms) 
		motifs.addAll(FindMotifDistances.getIUPACWords(m, mismatches));
	    Vector<Pattern> patterns = new Vector<Pattern>();
	    Vector<BufferedWriter> bws = new Vector<BufferedWriter>();
	    int longestMotif = 0;
	    for (String m : motifs) {
		patterns.addElement(Pattern.compile(m));
		bws.addElement(new BufferedWriter(new FileWriter(motifs.size()>1 ? saveFileName.replace(".txt", "_" + m.replace(".", "N") + ".txt") : saveFileName)));
		longestMotif = Math.max(longestMotif, m.length());
		System.out.println("Saving " + m + " locations to " + saveFileName);
	    }
	    int nChr = hs ? 24 : 21;
	    for (int i=1; i<=nChr; i++) {
		if (chr==i || chr==-1) {
		    FastaSequence2 fastaSeq = new FastaSequence2(species, hg19, i, orchestra);
		    Vector<PhastConsParser> phastCons = new Vector<PhastConsParser>();
		    if (conservation && !hg19) {
			for (int j=0; j<motifs.size(); j++) 
			    phastCons.addElement(new PhastConsParser(i, true, orchestra, species));
		    }
		    int pos = 0;
		    //The last base from the previous window
		    String prevSeq = "";
		    while (fastaSeq.hasMoreSequences()) {
			String seq = prevSeq + fastaSeq.getNextSequence(pos, windowSize).toUpperCase();
			for (int j=0; j<motifs.size(); j++) {
			    Matcher m = patterns.elementAt(j).matcher(seq);
			    while (m.find()) {
				int po = pos + m.start() - prevSeq.length();
				double pScores = 0, ps[] = null;
				if (conservation) {
				    ps = phastCons.elementAt(j).findScores(po - 1, po - 1 + motifs.elementAt(j).length());
				    for (int k=0; k<ps.length; k++)
					pScores += ps[k];
				}
				bws.elementAt(j).write(i + "\t" + po + (conservation ? "\t" + format.format(pScores/motifs.elementAt(j).length()) : "") + "\n");
				//System.out.println("Found " + i + "\t" + po + (conservation ? "\t" + Arrays.toString(ps) : "") + " at " + m.start());
			    }
			}
			pos += windowSize;
			prevSeq = seq.substring(seq.length() - longestMotif + 1);
		    }
		}
	    }
	    for (int j=0; j<motifs.size(); j++) 
		bws.elementAt(j).close();
	}
	catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////


}