import java.util.*;
import java.util.regex.*;
import java.io.*;

public class MRNAInfo {

    protected TreeSet genes = new TreeSet();
    protected int chr;

    public MRNAInfo(int c, boolean orchestra) {
	chr = c;
	readMRNAInfo(orchestra);
    }

    //////////////////////////////////////////////////////////////////

    public void readMRNAInfo(boolean orchestra) {
	String genomeLoc = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	try {
	    BufferedReader br = new BufferedReader(new FileReader(genomeLoc + "refSeqGenes/mrna_list.txt"));
	    String line = br.readLine();
	    //- gene geneID refseq - chr strand start end length n_coding n_intronic
	    Pattern mrnaPattern = Pattern.compile("\\d+\\s+([\\w|\\d]+)\\s+(\\d+)\\s+([X|N]M_\\d+\\.\\d)\\s+\\d+\\s+([X|Y|\\d]+)\\s+([+|-])\\s+(\\d+)\\s+(\\d+)");
	    while (line!=null) {
		Matcher m = mrnaPattern.matcher(line);
		if (m.find()) {
		    int c2;
		    if (m.group(4).equals("X"))
			c2 = 20;
		    else if (m.group(4).equals("Y"))
			c2 = 21;
		    else
			c2 = Integer.parseInt(m.group(4));
		    if (chr==c2) {
			boolean strand = m.group(5).equals("+");
			int start = strand ? Integer.parseInt(m.group(6)) : Integer.parseInt(m.group(7));
			int end = strand ? Integer.parseInt(m.group(7)) : Integer.parseInt(m.group(6));
			//check if a version of the gene already
			//exists. If so, we merge the two versions
			//into a super-gene which is contains the
			//union of the two versions
			int id = Integer.parseInt(m.group(2));
			MRNAInfoEntry g = getGene(id);
			if (g==null)
			    add(new MRNAInfoEntry(chr, start, end, strand, m.group(1), id, m.group(3)));
			else { //found version already parsed, add an alternative start and end site
			    if (strand) {
				g.setStart(Math.min(start, g.getStart()));
				g.setEnd(Math.max(end, g.getEnd()));
			    }
			    else {
				g.setStart(Math.max(start, g.getStart()));
				g.setEnd(Math.min(end, g.getEnd()));
			    }
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { 
	    System.err.println("Error reading mRNA info from " + genomeLoc + "refSeqGenes/mrna_list.txt\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
	//System.out.println("Found " + mRNAInfo.size() + " genes for chromosome " + c);
	addExonInfo(genomeLoc);
    }
    
    public void addExonInfo(String genomeLoc) {
	//Use this method to parse information about the location of
	//the exons for each gene.
	try {
	    BufferedReader br = new BufferedReader(new FileReader(genomeLoc + "Features/exon_list.txt"));
	    //# FeatureSORT	(Gene)	(GeneID)	RefSeq	Feature	(RefSeq_which_ftr)	(nFeatures)	(Chrom)	Orientation	(FeatureStart)	(FeatureStop)	Length
	    Pattern exonPattern = Pattern.compile("\\d+\\s+([\\d|\\w]+)\\s+(\\d+)\\s+\\w{2}_\\d+\\.\\d+\\s+EXON\\s+(\\d+)\\s+(\\d+)+\\s+([\\w|\\d]+)\\s+[+|-]\\s+(\\d+)\\s+(\\d+)\\s+\\d+");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = exonPattern.matcher(line);
		if (m.find()) {
		    int c;
		    if (m.group(5).equals("X")) { c = 20; }
		    else if (m.group(5).equals("Y")) { c = 21; }
		    else { c = Integer.parseInt(m.group(5)); }
		    if (c==chr) {
			int id = Integer.parseInt(m.group(2));
			//Find the corresponding gene
			MRNAInfoEntry gene = getGene(Integer.parseInt(m.group(2)));
			if (gene!=null) {
			    //System.out.println("Set " + Integer.parseInt(m.group(4)) + " exons for " + gene.getName());
			    gene.setNumberOfExons(Integer.parseInt(m.group(4)));
			    if (gene.getStrand()) //Add the exon info
				gene.addExon(Integer.parseInt(m.group(3)) - 1, Integer.parseInt(m.group(6)), Integer.parseInt(m.group(7)));
			    else
				gene.addExon(Integer.parseInt(m.group(3)) - 1, Integer.parseInt(m.group(7)), Integer.parseInt(m.group(6)));
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) { 
	    System.err.println("Error reading mRNA info from " + genomeLoc + "refSeqGenes/exon_list.txt");
	}
    }

    //////////////////////////////////////////////////////////////////

    public int getChromosome() { return chr; }

    public int size() { return genes.size(); }

    public void add(MRNAInfoEntry m) { genes.add(m); }

    public TreeSet getGenes() { return genes; }

    public MRNAInfoEntry getNearestGene(int pos) {
	int nearestDist = Integer.MAX_VALUE;
	MRNAInfoEntry nearestGene = null;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    MRNAInfoEntry m = (MRNAInfoEntry)it.next();
	    int dist = m.peakDistance(pos);
	    if (Math.abs(dist)<Math.abs(nearestDist)) {
		nearestDist = dist;
		nearestGene = m;
	    }
	}
	//System.out.println("Found distance " + nearestDist + " and gene " + nearestGene.toString() + " for peak " + pos);
	return nearestGene;
    }

    public void findNearbyPeaks(TreeSet peaks, int cutOff) {
	//Find the peaks which are within cutOff and not closer to any
	//other annotated gene.
	Iterator it = peaks.iterator();
	while (it.hasNext()) {
	    ChromosomeWindow peak = (ChromosomeWindow)it.next();
	    Iterator it2 = genes.iterator();
	    while (it2.hasNext()) {
		MRNAInfoEntry gene = (MRNAInfoEntry)it2.next();
		if (gene.getStrand()) {
		    if (gene.getStart() - cutOff<peak.getCenterOfMass() && gene.getEnd() + cutOff>peak.getCenterOfMass())
			gene.addNearbyPeak(peak);
		}
		else {
		    if (gene.getStart() + cutOff>peak.getCenterOfMass() && gene.getEnd() - cutOff<peak.getCenterOfMass())
			gene.addNearbyPeak(peak);
		}
	    }
	}
    }



    public MRNAInfoEntry getGene(String name) {
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    MRNAInfoEntry m = (MRNAInfoEntry)it.next();
	    String n = m.name;
	    if (n.toLowerCase().equals(name.toLowerCase()))
		return m;
	}
	return null;
    }

    public MRNAInfoEntry getGene(int id) {
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    MRNAInfoEntry m = (MRNAInfoEntry)it.next();
	    if (m.id==id)
		return m;
	}
	return null;
    }

    public int numberOfProximalOrInternal(int peakPos, int promoterLength) {
	//Use this method to find the number of genes which are
	//internal or proximal with respect to this peak.
	int nGenes = 0;
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    MRNAInfoEntry m = (MRNAInfoEntry)it.next();
	    nGenes += (m.peakInternal(peakPos) || m.isProximal(peakPos, promoterLength)) ? 1 : 0;
	}
	return nGenes;
    }

    public TreeMap alignTSS(int nUpstream, int nDownstream, ABIRead sortedReads[]) {
	//Use this method to find the distribution of reads in the
	//promoter regions of all genes.
	TreeMap histRelPos = new TreeMap();
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    MRNAInfoEntry gene = (MRNAInfoEntry)it.next();
	    histRelPos = gene.getReadsInPromoterRegion(sortedReads, nUpstream, nDownstream, histRelPos);
	}
	return histRelPos;
    }

    public boolean windowInPromoter(ChromosomeWindow win, int upstream, int downstream) {
	//Returns true if the window is completely within a promoter
	//region as defined by the range of upstream and downstream.
	Iterator it = genes.iterator();
	while (it.hasNext()) {
	    MRNAInfoEntry gene = (MRNAInfoEntry)it.next();
	    if (gene.getStrand()) {
		if (win.getStart()>gene.getStart()-upstream && win.getEnd()<gene.getStart()+downstream)
		    return true;
	    }
	    else {
		if (win.getStart()>gene.getStart()-downstream && win.getEnd()<gene.getStart()+upstream)
		    return true;
	    }
	}
	return false;
    }

}
