import java.util.*;

public class MRNAInfoEntry implements Comparable {

    //Use this class to represent one gene. start will always be
    //smaller than end and the boolean strand tells us from which
    //direction the gene will be transcribed.

    //The start is always the 5' end and will thus be larger than end
    //for genes on the - strand.
    public int chr, start, end, id, activityRegulated = 0, npas4RNAiRegulated = 0;
    public boolean strand;
    public String refSeq, name;
    //Keep track of the peaks which are close to this gene
    protected TreeSet nearbyPeaks = new TreeSet();
    //The coordinates are given with respect to the 5' end of the
    //gene. Thus genes on the negative strand will have higher values
    //for the start than for the stop.
    protected int[] exonStarts = null, exonStops = null;

    //////////////////////////////////////////////////////////////////

    public MRNAInfoEntry(int c, int s, int e, boolean st, String n, int i, String r) {
	chr = c;
	start = s;
	end = e;
	strand = st;
	name = n;
	refSeq = r;
	id = i;
    }

    public TreeSet findInternalPeaks(TreeSet regs) {
	TreeSet internal = new TreeSet();
	Iterator it = regs.iterator();
	while (it.hasNext()) {
	    ChromosomeWindow er = (ChromosomeWindow)it.next();
	    if (peakInternal(er.getCenter()))
		internal.add(er);
	}
	return internal;
    }

    public boolean peakInternal(int peakPos) { return strand ? peakPos>start && peakPos<end : peakPos<start && peakPos>end; }

    public boolean peakUpstream(int peakPos) { return strand ? peakPos<start : peakPos>end; }

    public boolean peakDownstream(int peakPos) { return strand ? peakPos>start : peakPos<end; }

    public TreeSet findProximalPeaks(TreeSet regs, int cutOff) {
	TreeSet proximal = new TreeSet();
	Iterator it = regs.iterator();
	while (it.hasNext()) {
	    ChromosomeWindow er = (ChromosomeWindow)it.next();
	    if (isProximal(er.getCenterOfMass(), cutOff))
		proximal.add(er);
	}
	return proximal;
    }

    public boolean isProximal(int peakPos, int cutOff) {
	//Proximal either to the 5' or 3' end
	if (strand)
	    return (start - peakPos>0 && start - peakPos<cutOff) || (peakPos - end>0 && peakPos - end<cutOff);
	else
	    return (end - peakPos>0 && end - peakPos<cutOff) || (peakPos - start>0 && peakPos - start<cutOff);
    }

    public TreeSet findNearestPeaks(TreeSet regs) {
	//Find those peaks which list this gene as their nearest one
	TreeSet set = new TreeSet();
	Iterator it = regs.iterator();
	while (it.hasNext()) {
	    ChromosomeWindow er = (ChromosomeWindow)it.next();
	    if (er.getNearestGene().equals(this)) {
		set.add(er);
		System.out.println(er.toString() + " is close to " + name);
	    }
	}
	return set;
    }

    public int peakDistance(int peakPos) {
	//Returns the shortest distance to the peak from either the
	//transcription start site or stop site, depending on which
	//one is smaller. Negative values for genes which are upstream
	//and positive values for genes which are downstream.
	int startDist = strand ? peakPos - start : end - peakPos;
	int endDist = strand ? end - peakPos : peakPos - start;
	return Math.abs(startDist)<Math.abs(endDist) ? startDist : endDist;
    }

    public boolean fivePrimeClosest(int peakPos) {
	return strand ? Math.abs(peakPos - start)<Math.abs(peakPos - end) : Math.abs(peakPos - end)<Math.abs(peakPos - start);
    }

    public int getStart() { return start; }
    
    public int getEnd() { return end; }

    public int getLength() { return Math.abs(start - end); }

    public void setStart(int s) { start = s; }
    
    public void setEnd(int e) { end = e; }

    public void setActivityRegulated(int a) { activityRegulated = a; }
    
    public int getActivityRegulated() { return activityRegulated; }

    public void setNpas4RNAiRegulated(int a) { npas4RNAiRegulated = a; }

    public int getNpas4RNAiRegulated() { return npas4RNAiRegulated; }

    //////////////////////////////////////////////////////////////////

    public TreeMap getReadsInPromoterRegion(ABIRead sortedReads[], int nUpstream, int nDownstream, TreeMap relPos) {
	//Use this method to search for reads which are located in the
	//promoter region of the gene.
	//Define the start and end as coming from the same direction,
	//regardless of strand.
	int promoterStart = strand ? start - nUpstream : end + nUpstream;
	int promoterEnd = strand ? start + nDownstream : end - nDownstream;
	//Find the reads which are inside this region
	int i = 0, s = strand ? promoterStart : promoterEnd, e = strand ? promoterEnd : promoterStart;
	while (i<sortedReads.length && sortedReads[i].getPosition()<s) { i++; }
	while (i<sortedReads.length && sortedReads[i].getPosition()<e) {
	    //The relative postion is calculated with respect to a fixed frame.
	    Integer relP = new Integer(strand ? sortedReads[i].getPosition() - start : end - sortedReads[i].getPosition());
	    for (int j=0; j<AnalyzeControls.READ_LENGTH; j++) {
		if (relP.intValue() + j<nUpstream + nDownstream) {
		    int nFound = relPos.containsKey(new Integer(relP.intValue() + j)) ? 1 + ((Integer)relPos.get(new Integer(relP.intValue() + j))).intValue() : 0;
		    relPos.put(new Integer(relP.intValue() + j), new Integer(nFound));
		}
	    }
	    i++;
	}
	return relPos;
    }

    public void addNearbyPeak(ChromosomeWindow p) { nearbyPeaks.add(p); }

    public TreeSet getNearbyPeaks() { return nearbyPeaks; }

    public boolean getStrand() { return strand; }

    public String getName() { return name; }

    public int getID() { return id; }

    public int getChromosome() { return chr; }

    public void setNumberOfExons(int n) {
	if (exonStarts==null || numberOfExons()<n) {
	    exonStarts = new int[n];
	    exonStops = new int[n];
	}
    }

    public void addExon(int i, int start, int stop) {
	exonStarts[i] = start;
	exonStops[i] = stop;
    }

    public int numberOfExons() { return exonStops.length; }

    public int getExonStart(int i) { 
	if (i<0) { return getStart(); }
	return exonStarts[i]; 
    }

    public int getExonStop(int i) { 
	if (i>=numberOfExons()) { return getEnd(); }
	return exonStops[i]; 
    }

    public boolean hasExonInfo() { return exonStops!=null; }

    //////////////////////////////////////////////////////////////////

    public int compareTo(Object o) {
	//Sort according to chromosomal position
	if (!(o instanceof MRNAInfoEntry))
	    System.err.println("ERROR comparing genes!");
	MRNAInfoEntry mi = (MRNAInfoEntry)o;
	if (chr<mi.chr)
	    return -1;
	else if (chr>mi.chr)
	    return 1;
	else {
	    if (getStart()<mi.getStart())
		return -1;
	    else if (getStart()>mi.getStart())
		return 1;
	    else
		return 0;
	}
    }

    public String toString() {
	return name + " " + chr + "(" + (strand ? "+" : "-") + ")" + start + "-" + end;
    }

    public static String getHeaderString() { return "Name\tRefSeq\tGeneID\tStart\tEnd\tStrand"; }

    public String geneInfo() { return name + "\t" + refSeq + "\t" + id + "\t" + start + "\t" + end + "\t" + (strand ? "+" : "-") + "\t"; }

    public boolean equals(Object o) {
	if (!(o instanceof MRNAInfoEntry))
	    System.err.println("ERROR comparing genes!");
	MRNAInfoEntry mi = (MRNAInfoEntry)o;
	return mi.chr==chr && mi.start==start && mi.end==end && mi.name.equals(name) && mi.id==id;
    }

    public int hashCode() { return refSeq.hashCode() + 13*start + 29*name.hashCode(); }

}
