//Use this class to represent one microRNA

import java.util.*;
import edu.uci.ics.jung.graph.impl.*;

public class MacroRNA extends UndirectedSparseVertex implements Comparable {

    //The start is always the 5' end and will thus be larger than end
    //for genes on the - strand.
    protected int chr, id, start, end;
    protected boolean strand, precursor;
    protected String name;
    //Keep track of the number of reads, can either be across coding
    //region or just at TSS
    protected TreeMap<String, ABIRead[]> factorReads = new TreeMap<String, ABIRead[]>();

    //////////////////////////////////////////////////////////////////

    public MacroRNA(int c, int s, int e, boolean st, String n) {
	chr = c;
	strand = st;
	start = s;
	end = e;
	name = n;
    }

    ///////////////////////////////////////////////////////

    public int getStart() { return start; }

    public int getEnd() { return end; }

    public int getLength() { return Math.abs(start - end); }

    //////////////////////////////////////////////////////////////////
    //Stuff for adding read info

    public void setFactorReads(String tfName, ABIRead[] reads) { factorReads.put(tfName, reads); }

    public ABIRead[] getFactorReads(String tfName) { return factorReads.containsKey(tfName) ? factorReads.get(tfName) : new ABIRead[]{}; }

    public double getReadDensity(String tfName) { return factorReads.containsKey(tfName) ? (double)getFactorReads(tfName).length/getLength() : 0; }

    public TreeSet<String> getFactors() { return new TreeSet<String>(factorReads.keySet()); }

    //////////////////////////////////////////////////////////////////

    public boolean getStrand() { return strand; }

    public int getChromosome() { return chr; }


    /////////////////////////////////

    public boolean overlaps(int st, int en) {
	//if (strand)
	//  return en>start && start<end;
	//return st<start && en>end;
	if (strand) {
	    if ((st>start && st<end) || (en>start && en<end)) //Start or en of region inside gene
		return true;
	    else if ((start>st && start<en) || (end>st && end<en)) //Start or en of exon inside region
		return true;
	}
	else {
	    if ((st<start && st>end) || (en<start && en>end)) //Start or en of region inside gene
		return true;
	    else if ((start>st && start<en) || (end>st && end<en)) //Start or en of exon inside region
		return true;
	}
	return false;
    }

    public boolean overlaps(Exon e) { return e.getChromosome()==chr && overlaps(e.getStart(), e.getEnd()); }

    public TreeSet<Exon> getOverlappingExons(TreeSet<Exon> exons) {
	TreeSet<Exon> overlappingExons = new TreeSet<Exon>();
	for (Exon e : exons) {
	    if (overlaps(e))
		overlappingExons.add(e);
	}
	return overlappingExons;
    }

    public TreeSet<ConservationIsland> getOverlappingPromoterCIs(ConservationIslands consIslands, int maxPromoterDist) {
	TreeSet<ConservationIsland> overlappingCIs = new TreeSet<ConservationIsland>();
	for (ConservationIsland c : consIslands.getConservationIslands()) {
	    if (chr==c.getChromosome() && ((strand && c.overlaps(start - maxPromoterDist, start)) || (!strand && c.overlaps(start, start + maxPromoterDist))))
		overlappingCIs.add(c);
	}
	return overlappingCIs;
    }

    public TreeSet<ConservationIsland> getOverlappingCIs(ConservationIslands consIslands) {
	TreeSet<ConservationIsland> overlappingCIs = new TreeSet<ConservationIsland>();
	for (ConservationIsland c : consIslands.getConservationIslands()) {
	    if (chr==c.getChromosome() && c.overlaps(strand ? start : end, strand ? end : start))
		overlappingCIs.add(c);
	}
	return overlappingCIs;
    }

    public int getID() { return hashCode(); }

    //////////////////////////////////////////////////////////////////
    //Stuff for the JUNG package
    
    //////////////////////////////////////////////////////////////////

    public int compareTo(Object o) {
	//Sort according to chromosomal position
	if (!(o instanceof MacroRNA))
	    System.err.println("ERROR comparing miRNA!");
	MacroRNA ma = (MacroRNA)o;
	if (chr<ma.chr)
	    return -1;
	else if (chr>ma.chr)
	    return 1;
	else {
	    if (getStart()<ma.getStart())
		return -1;
	    else if (getStart()>ma.getStart())
		return 1;
	    else
		return 0;
	}
    }

    public String toString() {
	return chr + "(" + (strand ? "+" : "-") + ")" + start + "-" + end;
    }

    public boolean equals(Object o) {
	if (!(o instanceof MacroRNA))
	    System.err.println("ERROR comparing MacroRNA!");
	MacroRNA ma = (MacroRNA)o;
	return ma.chr==chr && ma.start==start && ma.end==end;
    }

    public int hashCode() { return chr*29 + 5*end + 19*start; }

}
