//Use this method to sweep the parameter space for the
//FindTranscripts.java class. Search for the optimal values of L_min,
//L_max and FDr for a given chromosome. We want to allow for different
//cost functions.
//
//1. Optimality with respect to the annotated transcripts. Specify a
//cut-off (either relative or absolute density) for which genes to
//include. We are interested in monitoring the number of genes which
//are mapped by only one transcript, the number which are incorrectly
//split, the number which are incorrectly merged and the number which
//are missed.
//
//Example:
//
//java -Xmx1024m OptimalTranscriptParameters -c 1 -n ../SacCer/mRNA/c1_shot --species sc -L 16 --merge

import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import jargs.gnu.CmdLineParser;
import java.io.*;

public class OptimalTranscriptParameters {

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    public static void main(String[] args) {
	try {
	    //Handle the flags
	    //Flags used at the moment: 
	    CmdLineParser argParser = new CmdLineParser();
	    CmdLineParser.Option chromosomeFlag = argParser.addIntegerOption('c', "chromosome");	    
	    CmdLineParser.Option readLengthFlag = argParser.addIntegerOption('R', "readLength");	    
	    CmdLineParser.Option strandFlag = argParser.addBooleanOption('r', "reverseStrand");
	    CmdLineParser.Option orchestraFlag = argParser.addBooleanOption('o', "orchestra");
	    CmdLineParser.Option fileNameFlag = argParser.addStringOption('n', "fileName");
	    CmdLineParser.Option saveFileNameFlag = argParser.addStringOption('s', "saveFileName");
	    CmdLineParser.Option speciesFlag = argParser.addStringOption("species");
	    CmdLineParser.Option optimizationFlag = argParser.addStringOption("optimization");
	    //Flags for the simulated annealing procedure
	    CmdLineParser.Option tempFlag = argParser.addDoubleOption("temp");
	    CmdLineParser.Option stepsFlag = argParser.addIntegerOption("steps");
	    //Parameters for the transcript finding
	    CmdLineParser.Option lMinFlag = argParser.addIntegerOption('l', "lMin");
	    CmdLineParser.Option lMaxFlag = argParser.addIntegerOption('L', "lMax");
	    CmdLineParser.Option deltaLFlag = argParser.addIntegerOption("deltaL");
	    CmdLineParser.Option fdrMinFlag = argParser.addDoubleOption("fdrMin");
	    CmdLineParser.Option fdrMaxFlag = argParser.addDoubleOption("fdrMax");
	    //This is given in a log10-scale
	    CmdLineParser.Option nfdrFlag = argParser.addIntegerOption("nfdr");
	    CmdLineParser.Option mergeFlag = argParser.addBooleanOption("merge");
	    CmdLineParser.Option flattenFlag = argParser.addBooleanOption("flatten");
	    //include the compare highest expressed genes when testing
	    //how well the transcribed regions match the annotation.
	    CmdLineParser.Option compareFlag = argParser.addDoubleOption("compare");
	    CmdLineParser.Option minDensFlag = argParser.addDoubleOption("minDens");
	    //Use these flags for ChIP data. We need to have a
	    //fragment length and possibly a negative control data set
	    //(if available)
	    CmdLineParser.Option chipFlag = argParser.addBooleanOption("chip");
	    CmdLineParser.Option inputFileNameFlag = argParser.addStringOption('i', "inputFileName");
	    CmdLineParser.Option fragmentLengthFlag = argParser.addIntegerOption('f', "fragmentLength");
	    //////////////////////////////////////////////////////////////////
	    //Parse the flags
	    argParser.parse(args);
	    //The latest version of CmdLineParser seems to remove the
	    //value foreach flag when you've looked at it once. Thus
	    //we adjust our code to work with the new version.
	    Object tmp;
	    tmp = argParser.getOptionValue(chromosomeFlag);
	    int chr = tmp!=null ? ((Integer)tmp).intValue() : -1;
	    boolean orchestra = argParser.getOptionValue(orchestraFlag)!=null;
	    boolean chip = argParser.getOptionValue(chipFlag)!=null;
	    tmp = argParser.getOptionValue(fileNameFlag);
	    String fileName = tmp!=null ? (String)tmp + (chip ? "" : ".tsv") : "";
	    tmp = argParser.getOptionValue(inputFileNameFlag);
	    String inputFileName = tmp!=null ? (String)tmp : "";
	    tmp = argParser.getOptionValue(speciesFlag);
	    String species = tmp!=null ? ((String)tmp).toLowerCase() : "mm";
	    tmp = argParser.getOptionValue(optimizationFlag);
	    String optimization = tmp!=null ? ((String)tmp).toLowerCase() : "sa";
	    tmp = argParser.getOptionValue(saveFileNameFlag);
	    String saveFileName = tmp!=null ? (String)tmp : (chip ? fileName.substring(0, fileName.lastIndexOf("_")) : fileName.substring(0, fileName.lastIndexOf(".")));
	    boolean reverseStrand = argParser.getOptionValue(strandFlag)!=null;
	    tmp = argParser.getOptionValue(stepsFlag);
	    int maxSteps = tmp!=null ? ((Integer)tmp).intValue() : 10;
	    tmp = argParser.getOptionValue(readLengthFlag);
	    int readLength = tmp!=null ? ((Integer)tmp).intValue() : 35;
	    tmp = argParser.getOptionValue(lMinFlag);
	    int lMin = tmp!=null ? ((Integer)tmp).intValue() : 5;
	    tmp = argParser.getOptionValue(lMaxFlag);
	    int lMax = tmp!=null ? ((Integer)tmp).intValue() : 20;
	    tmp = argParser.getOptionValue(deltaLFlag);
	    int deltaL = tmp!=null ? ((Integer)tmp).intValue() : 1;
	    tmp = argParser.getOptionValue(tempFlag);
	    double temp = tmp!=null ? ((Double)tmp).doubleValue() : 1;
	    tmp = argParser.getOptionValue(fdrMinFlag);
	    double fdrMin = tmp!=null ? ((Double)tmp).doubleValue() : .01;
	    tmp = argParser.getOptionValue(fdrMaxFlag);
	    double fdrMax = tmp!=null ? ((Double)tmp).doubleValue() : .1;
	    tmp = argParser.getOptionValue(nfdrFlag);
	    int nfdr = tmp!=null ? ((Integer)tmp).intValue() : 10;
	    boolean mergeSegments = argParser.getOptionValue(mergeFlag)!=null;
	    boolean flatten = argParser.getOptionValue(flattenFlag)!=null;
	    tmp = argParser.getOptionValue(compareFlag);
	    double compare = tmp!=null ? ((Double)tmp).doubleValue() : -1;
	    tmp = argParser.getOptionValue(minDensFlag);
	    double minDens = tmp!=null ? ((Double)tmp).doubleValue() : .0286; //1/35
	    tmp = argParser.getOptionValue(fragmentLengthFlag);
	    int fragmentLength = tmp!=null ? ((Integer)tmp).intValue() : 120;
	    //////////////////////////////////////////////////////////////////
	    int nChr = FindTranscripts.numberOfChromosomes(species);
	    RefSeqGenes geneInfo[] = new RefSeqGenes[nChr];
	    TranscriptFinder tf[] = new TranscriptFinder[nChr];
	    ReducedModel reducedModel[] = new ReducedModel[nChr];
	    for (int i=1; i<=nChr; i++) {
		if (i==chr || chr==-1) {
		    geneInfo[i-1] = new RefSeqGenes(i, orchestra, species);
		    tf[i-1] = chip ? new ChIPRegionFinder(fileName, inputFileName, i, fragmentLength, species) : new TranscriptFinder(fileName, i, !reverseStrand, flatten, readLength, species);
		    geneInfo[i-1].setFactorReads(fileName, tf[i-1].getReads(), false, chip ? 0 : (reverseStrand ? -1 : 1));
		    //Find the set of genes that have high enough density for the comparison
		    double meanReadLength = ABIReads.meanReadLength(tf[i-1].getReads());
		    //minDens /= meanReadLength; //To get it to no reads without the extension
		    TreeMap<Double, Integer> exp2IdMap = FindTranscripts.findGeneDensities(geneInfo[i-1], reverseStrand, fileName);
		    int geneIDsKeep[] = FindTranscripts.getGenesWithDensityGreaterThan(minDens, exp2IdMap, geneInfo[i-1], !reverseStrand);
		    System.out.println("There are " + geneIDsKeep.length + " genes density above " + format.format(minDens));
		    //Calculate all the coefficients that we could potentially want to use
		    int lVec[] = FindTranscripts.getLengthVec(lMin, lMax, 1);
		    tf[i-1].calculateHaarCoefficients(lVec);
		    System.out.println("L_min\tL_max\tFDR\tn\t<log(L)>\ttranscribed\treads\tr-t\tgenes\tmissed genes\tnon-coding");
		    if (optimization.toLowerCase().equals("sa")) {
			//Choose random initial values for the parameters
			int li = lMin + random.nextInt((lMax - lMin)/2), bestLMin = li, liOld = li;
			int la = li + random.nextInt(lMax - li) + 1, bestLMax = la, laOld = la;
			double fdrVec[] = FindTranscripts.getFDRVec(fdrMin, fdrMax, nfdr);
			int fInd = random.nextInt(fdrVec.length), bestFDRInd = fInd, fOldInd = fInd;
			double cov = 0, bestCoverage = 0;
			HashMap<int[], Double> visitedStates = new HashMap<int[], Double>();
			int nSteps = 0;
			while (nSteps++<maxSteps) {
			    //Take a random step in parameter space
			    while ((la==laOld && li==liOld && fInd==fOldInd) || la==li) {
				li = randomStep(liOld, lMin, lMax - 2);
				la = randomStep(laOld, li + 1, lMax);
				fInd = randomStep(fOldInd, 0, fdrVec.length-1);
			    }
			    if (!visitedStates.containsKey(new int[]{li, la, fInd})) { //make sure that we do not revisit a state that we have already been 
				lVec = FindTranscripts.getLengthVec(li, la, 1);
				Vector<Integer> breakPoints = tf[i-1].findBreakpoints(lVec, fdrVec[fInd]);
				if (breakPoints.size()>1) {
				    reducedModel[i-1] = tf[i-1].findReducedModel(lVec, fdrVec[fInd], breakPoints, mergeSegments);
				    reducedModel[i-1].compare2Annotation(geneInfo[i-1], !reverseStrand, geneIDsKeep);
				    cov = (double)reducedModel[i-1].numberOfGenesWithOneTranscript()/geneIDsKeep.length;
				}
				else
				    cov = 0;
				visitedStates.put(new int[]{li, la, fInd}, cov);
				reducedModel[i-1].removeGenes();
			    }
			    else 
				cov = visitedStates.get(new int[]{li, la, fInd});
			    if (cov>bestCoverage) { //accept 
				bestLMin = li;
				bestLMax = la;
				bestFDRInd = fInd;
				bestCoverage = cov;
			    }
			    if (cov>bestCoverage || Math.exp(-(1 - cov)/temp)>random.nextDouble()) {
				//System.out.println("Coverage is " + format.format(cov) + " and prob of accepting is " + format.format(Math.exp(-(1 - cov)/temp)));
				liOld = li;
				laOld = la;
				fOldInd = fInd;
			    }
			    System.out.println(li + "\t" + la + "\t" + format.format(fdrVec[fInd]) + "\t" + reducedModel[i-1].numberOfTranscribedRegions() + "\t" + format.format(reducedModel[i-1].meanTranscribedRegionLength()) + "\t" + format.format(reducedModel[i-1].proportionTranscribed()) + "\t" + format.format(reducedModel[i-1].proportionOfReads()) + "\t" + format.format(reducedModel[i-1].proportionOfReads() - reducedModel[i-1].proportionTranscribed()) + "\t" + format.format(cov) + "\t" + format.format((double)reducedModel[i-1].numberOfGenesWithNTranscripts(0, false, false)/geneIDsKeep.length) + "\t" + format.format((double)reducedModel[i-1].numberOfTranscriptsWithNGenes(0, false, false)/reducedModel[i-1].numberOfTranscribedRegions()) + "\t");
			}
			System.out.println("Best solution is L_min = " + bestLMin + ", L_max = " + bestLMax + " FDR = " + format.format(fdrVec[bestFDRInd]) + " with coverage " + format.format(bestCoverage));
			lVec = FindTranscripts.getLengthVec(bestLMin, bestLMax, 1);
			tf[i-1].calculateHaarCoefficients(lVec);
			Vector<Integer> breakPoints = tf[i-1].findBreakpoints(lVec, fdrVec[bestFDRInd]);
			reducedModel[i-1] = tf[i-1].findReducedModel(lVec, fdrVec[bestFDRInd], breakPoints, mergeSegments);
			System.out.println(bestLMin + "\t" + bestLMax + "\t" + format.format(fdrVec[bestFDRInd]) + "\t" + reducedModel[i-1].numberOfTranscribedRegions() + "\t" + format.format(reducedModel[i-1].meanTranscribedRegionLength()) + "\t" + format.format(reducedModel[i-1].proportionTranscribed()) + "\t" + format.format(reducedModel[i-1].proportionOfReads()) + "\t" + format.format(reducedModel[i-1].proportionOfReads() - reducedModel[i-1].proportionTranscribed()) + "\t" + format.format(bestCoverage) + "\t" + format.format((double)reducedModel[i-1].numberOfGenesWithNTranscripts(0, false, false)/geneIDsKeep.length) + "\t" + format.format((double)reducedModel[i-1].numberOfTranscriptsWithNGenes(0, false, false)/reducedModel[i-1].numberOfTranscribedRegions()) + "\t");
		    }
		    if (saveFileName!=null && chr!=-1)
			tf[i-1].saveReducedModel(saveFileName + FindTranscripts.BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f") + (chr!=-1 ? "_" + chr : "") + ".bed", saveFileName.substring(saveFileName.lastIndexOf("/") + 1) + FindTranscripts.BED_FILE_SUFFIX + (reverseStrand ? "_r" : "_f"), species);
		}
	    }
	} catch (Exception e) {
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	} // end of try-catch
    } // end of main()

    //////////////////////////////////////////////////////////////////

    public static int randomStep(int r, int rMin, int rMax) {
	if (random.nextBoolean()) {
	    if (random.nextBoolean() && r>rMin)
		return r - 1;
	    else if (r<rMax)
		return r + 1;
	}
	return r;
    }

    //Use this version to take longer steps when in a region far away from the best solution
    public static int randomStep(int r, int rMin, int rMax, double relSol) {
	if (random.nextBoolean()) {
	    if (random.nextBoolean() && r>rMin)
		return r - 1;
	    else if (r<rMax)
		return r + 1;
	}
	return r;
    }


    public static double randomStep(double r, double dr, double rMin, double rMax) {
	if (random.nextBoolean()) {
	    if (random.nextBoolean() && r>rMin && r<rMax)
		return r - dr;
	    else if (r>rMin && r<rMax)
		return r + dr;
	}
	return r;
    }

}

