//Use this class for finding peaks. First method to be implemented is
//a very naive one which only searches for the location of the peaks.

import java.util.*;
import java.util.regex.*;
import java.io.*;
import org.apache.commons.math.distribution.PoissonDistributionImpl;
import org.apache.commons.math.MathException;
//import umontreal.iro.lecuyer.probdist.*;
import java.text.DecimalFormat;

public class PeakFinder {

    public static final String GENOME_LOC = "/media/exthd/Genomes/Mm/v9/";

    protected static Random random = new Random();
    public static DecimalFormat format = new DecimalFormat("###.##########");

    public static Pattern posReadPattern = Pattern.compile("chr([\\d|X|Y]+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
    //The groups are bead identifier, chr, strand, pos, mismatches
    public static Pattern rawPattern1 = Pattern.compile("([0-9]+_[0-9]+_[0-9]+_[F|R][0-9]+),([0-9|X|Y]+)_(-*?)([0-9]+)\\.([0-3])");
    //Use this pattern when we parse files which were randomly
    //generated. They typically have a shorter header and no barcode
    //info. The first group is only there to make sure that the
    //numbers for the other groups will match with rawPattern1.
    public static Pattern randPattern1 = Pattern.compile("(>),?([0-9|X|Y]+)_(-*?)([0-9]+)[\\.([0-3])]*?");
    //The second line contains information about the output
    public static Pattern rawPattern2 = Pattern.compile("T([0-3]+)");
    //For the files containing the codes
    public static Pattern codePattern = Pattern.compile("([0-3]+)\t([0-9]+)");

    public static final int[] CHROMOSOME_LENGTH = new int[]{197195432, 181748087, 159599783, 155630120, 152537259, 149517037, 152524553, 131738871, 124076172, 129993255, 121843856, 121257530, 120284312, 125194864, 103494974, 98319150, 95272651, 90772031, 61342430, 166650296, 15902555};
    public static final long TOTAL_CHROMOSOME_LENGTH = (long)(265489521.8*10);

    //////////////////////////////////////////////////////////////////

    public PeakFinder() { }

    //////////////////////////////////////////////////////////////////

    public static Vector findEnrichedRegions(int chr, int fragmentLength, String fileName, String inputFileName, boolean extend, int strand) {
	//Use this when analyzing the input data to establish the
	//background cluster size. In this version, we count the
	//number of reads in the ChIP experiment and down-sample the
	//input to the same amount.
	int nReads = parseFiles(chr, fileName, fragmentLength, extend, strand).length;
	ABIRead[] inputReads = parseFiles(chr, inputFileName, fragmentLength, extend, strand);
	inputReads = filterRandomReads(inputReads, nReads);
	Arrays.sort(inputReads);
	return getEnrichedRegions(chr, inputReads, fragmentLength);
    }

    public static Vector findEnrichedRegions(int chr, int fragmentLength, String inputFileName, boolean extend, int strand) {
	//Use this when analyzing the input data to establish the
	//background cluster size.
	ABIRead[] reads = parseFiles(chr, inputFileName, fragmentLength, extend, strand);
	System.out.println("Found " + reads.length + " unique reads for " + inputFileName);
	return getEnrichedRegions(chr, reads, fragmentLength);
    }

    public static Vector findEnrichedRegions(int chr, int fragmentLength, int minReads, double inputReadsRatio, String fileName, String inputFileName, boolean extend, int strand) {
	return findEnrichedRegions(chr, fragmentLength, minReads, inputReadsRatio, fileName, inputFileName, fileName + "_ER.txt", extend, strand);
    }

    public static Vector findEnrichedRegions(int chr, int fragmentLength, int minReads, double inputReadsRatio, String fileName, String inputFileName, String saveFileName, boolean extend, int strand) {
	//Use this method to find enriched regions in the same manner
	//as described by Johnson et al (2007). The basic idea is to
	//find a cluster of at least minReads. Furthermore, we assume
	//that the fragments have length given by
	//fragmentLength. Next, we compare the regions with a
	//sub-sample of the input (to make sure that we have the same
	//number of reads in both samples).
	ABIRead[] reads = parseFiles(chr, fileName, fragmentLength, extend, strand);
	ABIRead[] inputReads = parseFiles(chr, inputFileName, reads.length, fragmentLength, extend, strand);
	//ABIRead[] inputReads = parseFiles(inputFileName, fragmentLength);
	return getEnrichedRegions(chr, reads, inputReads, fragmentLength, minReads, inputReadsRatio);
    }

    public static ABIRead[] parseFiles(int chr, String fileName, int nSample, int fragmentLength, boolean extend, int strand) {
	return parseFiles(chr, fileName, 3, true, nSample, fragmentLength, extend, strand);
    }

    public static ABIRead[] parseFiles(int chr, String fileName, int fragmentLength, boolean extend, int strand) {
	return parseFiles(chr, fileName, 3, true, -1, fragmentLength, extend, strand);
    }

    public static ABIRead[] parseFiles(int chr, String fileName, int nMismatches, boolean unique, int nSample, int fragmentLength, boolean extend, int strand) {
	try {
	    int nReadsForward = 0, nReadsReverse = 0;
	    BufferedReader br = new BufferedReader(new FileReader(fileName + "_" + chr));
	    String line = br.readLine();
	    boolean randFile = fileName.indexOf("rand")!=-1 || fileName.indexOf("_E")!=-1;
	    //Go over once just to count the number of reads for this
	    //chromosome
	    while (line!=null) {
		Matcher m = randFile ? PeakFinder.randPattern1.matcher(line) : PeakFinder.rawPattern1.matcher(line);
		if (m.find()) {
		    boolean forward = !m.group(3).equals("-");
		    int nMisses = !randFile || m.groupCount()==5 ? Integer.parseInt(m.group(5)) : 0;
		    if (nMisses<=nMismatches) {
			if (Integer.parseInt(m.group(2))==chr) {
			    if (forward)
				nReadsForward++;
			    else
				nReadsReverse++;
			}
		    }
		    line = br.readLine();
		}
		line = br.readLine(); //Get the sequence information
	    }
	    br.close();
	    //Go through again and store them. We want to use an
	    //ordinary array to save memory, going through the file is
	    //not that slow.
	    return parseFile(chr, fileName, nMismatches, nReadsForward, nReadsReverse, unique, nSample, fragmentLength, extend, strand);
	}
	catch (IOException e) {
	    System.err.println("Error, could not open file: " + fileName);
	}
	return null;
    }

    protected static ABIRead[] parseFile(int chr, String fileName, int nMismatches, int nReadsForward, int nReadsReverse, boolean unique, int nSample, int fragmentLength, boolean extend, int strand) {
	//Use an int to determine which strand to return. Negative
	//means reverse strand, positive forward strand and zero means
	//both.
	ABIRead readsForward[] = new ABIRead[nReadsForward], readsReverse[] = new ABIRead[nReadsReverse], reads[] = new ABIRead[nReadsForward + nReadsReverse], returnReads[] = null;
	try {
	    int i = 0, j = 0, k = 0;
	    BufferedReader br = new BufferedReader(new FileReader(fileName + "_" + chr));
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = PeakFinder.rawPattern1.matcher(line);
		if (m.find()) {
		    boolean forward = !m.group(3).equals("-");
		    int nMisses = Integer.parseInt(m.group(5)), pos = Integer.parseInt(m.group(4));
		    if (Integer.parseInt(m.group(2))==chr && nMisses<=nMismatches) {
			if (forward) { //Use 1-based index when saving
			    readsForward[i++] = new ABIRead(chr, true, pos + 1 + (extend ? 0 : fragmentLength/2), extend ? fragmentLength : AnalyzeControls.READ_LENGTH);
			    reads[k++] = new ABIRead(chr, true, pos + 1 + (extend ? 0 : fragmentLength/2), extend ? fragmentLength : AnalyzeControls.READ_LENGTH);
			}
			else {
			    readsReverse[j++] = new ABIRead(chr, false, pos + 1 - (extend ? fragmentLength : fragmentLength/2 + AnalyzeControls.READ_LENGTH), extend ? fragmentLength : AnalyzeControls.READ_LENGTH);
			    reads[k++] = new ABIRead(chr, false, pos + 1 - (extend ? fragmentLength : fragmentLength/2) - AnalyzeControls.READ_LENGTH, extend ? fragmentLength : AnalyzeControls.READ_LENGTH);
			}
			line = br.readLine();
		    }
		    else //Skip the next line
			line = br.readLine();
		}
		line = br.readLine();
	    }
	    br.close();
	    /*if (readsForward.length>0) {
		readsForward = findUniqueReads(readsForward);
		Arrays.sort(readsForward);
		savePositionsReads(readsForward, fileName + "_F_" + (extend ? "E_" : "") + fragmentLength + "_chr" + (new Integer(chr)).toString() + ".bed", true, fragmentLength, chr, true);
	    }
	    if (readsReverse.length>0) {
		readsReverse = findUniqueReads(readsReverse);
		Arrays.sort(readsReverse);
		savePositionsReads(readsReverse, fileName + "_R_" + (extend ? "E_" : "") + fragmentLength + "_chr" + (new Integer(chr)).toString() + ".bed", true, fragmentLength, chr, true);
		}*/
	    if (strand<0) { returnReads = readsReverse; }
	    else if (strand>0) { returnReads = readsForward; }
	    else { returnReads = reads; }
	    if (returnReads.length>0) {
		Arrays.sort(returnReads);
		//System.out.println("Found " + reads.length + " reads for " + fileName);
		if (unique) {
		    returnReads = findUniqueReads(returnReads);
		    if (nSample<returnReads.length && nSample>0)
			returnReads = filterRandomReads(returnReads, nSample);
		    Arrays.sort(returnReads);
		    //System.out.println("Found " + reads.length + " unique reads for " + fileName);
		}
	    }
	    else
		System.err.println("Warning: No reads found!");
	}
	catch (IOException e) {
	    
	}
	return returnReads;
    }

    protected static Vector getEnrichedRegions(int chr, ABIRead[] reads, ABIRead[] inputReads, int fragmentLength, int minReads, double inputReadsRatio) {
	//Returns the number of unique positions with at least one read
	Vector enrichedRegions = getEnrichedRegions(chr, reads, fragmentLength, minReads);
	TreeSet removeSet = new TreeSet();
	//Compare with the input set. This is also assumed to
	//consist of unique and sorted elements. Furthermore, we
	//assume that it has been reduced so that it has the same
	//size as the ChIP sample to make the comparison more
	//direct.
	Iterator it = enrichedRegions.iterator();
	int i = 0, len = reads[0].getLength();
	while (it.hasNext()) {
	    EnrichedRegion e = (EnrichedRegion)it.next();
	    //Find the corresponding region in the input data
	    while (i<inputReads.length && inputReads[i].getPosition()<e.getStart())
		i++;
	    int nInputReads = 0; //Count the number of reads in the input data
	    while (i<inputReads.length && inputReads[i].getPosition()<e.getEnd() + len) {
		nInputReads++;
		i++;
	    }
	    e.setInputReads(nInputReads);
	    double ratio = nInputReads==0 ? e.numberOfReads() : ((double)e.numberOfReads())/nInputReads;
	    if (inputReadsRatio>ratio)
		removeSet.add(e);
	}
	it = removeSet.iterator(); //Remove the ones with too many input reads
	while (it.hasNext()) 
	    enrichedRegions.remove((EnrichedRegion)it.next());
	return enrichedRegions;
    }

    public static Vector getEnrichedRegions(int chr, ABIRead[] reads, int fragmentLength) {
	return getEnrichedRegions(chr, reads, fragmentLength, 0);
    }

    protected static Vector getEnrichedRegions(int chr, ABIRead[] reads, int fragmentLength, int minReads) {
	//Use this to find all the clusters for a single data set
	Vector enrichedRegions = new Vector();
	EnrichedRegion er = new EnrichedRegion(reads[0], chr, reads[0].getLength());
	for (int i=1; i<reads.length; i++) {
	    if (reads[i].getPosition() - reads[i-1].getPosition()<reads[i].getLength()) 
		er.add(reads[i]);
	    else {
		if (er.numberOfReads()>minReads)
		    enrichedRegions.add(er);
		er = new EnrichedRegion(reads[i], chr, reads[i].getLength());
	    }
	}
	/*if (er.numberOfReads()>minReads)
	    enrichedRegions.add(er); //Add the last region
	int nReads = 0;
	Iterator it = enrichedRegions.iterator();
	while (it.hasNext())
	    nReads += ((EnrichedRegion)it.next()).numberOfReads();
	    System.out.println("Found " + nReads + " reads in the regions.");*/
	Collections.sort(enrichedRegions);
	return enrichedRegions;
    }

    public static Vector getEnrichedRegions(int fragmentLength, String fileName, boolean extend, int strand) { return getEnrichedRegions(fragmentLength, fileName, 8, extend, strand); }

    public static Vector getEnrichedRegions(int fragmentLength, String fileName, int minClusterSize, boolean extend, int strand) {
	Vector regions = new Vector();
	for (int i=0; i<21; i++)
	    regions.addAll(getEnrichedRegions(i+1, fragmentLength, fileName, minClusterSize, extend, strand));
	//Add the sequence info as well
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fileName + "_seqs.fa"));
	    Pattern posPattern = Pattern.compile(">chr(\\d+)_(\\d+)-(\\d+)");
	    String line = br.readLine(), seq = "";
	    Matcher m = posPattern.matcher(line); //Find the cluster which this sequence belongs to
	    int center = 0, chr = 0, nFound = 0;
	    EnrichedRegion er = null;
	    if (m.find()) {
		center = (Integer.parseInt(m.group(2)) + Integer.parseInt(m.group(3)))/2; 
		chr = Integer.parseInt(m.group(1));
		nFound++;
		er = findRegion(regions, chr, center);
	    }
	    while (line!=null) {
		m = posPattern.matcher(line); //Find the cluster which this sequence belongs to
		if (m.find()) {
		    er.setSequence(seq);
		    center = (Integer.parseInt(m.group(2)) + Integer.parseInt(m.group(3)))/2;
		    chr = Integer.parseInt(m.group(1));
		    er = findRegion(regions, chr, center);
		    nFound++;
		    seq = "";
		}
		else
		    seq += line;
		line = br.readLine();
	    }
	    //Fix the last region
	    if (er.getSequence()!=null)
		System.out.println(er.toString() + " already has sequence!");
	    er.setSequence(seq);
	    br.close();
	    System.out.println("Found " + regions.size() + " peaks and " + nFound);
	}
	catch (IOException e) { 
	    System.err.println("Error, could not load sequences!");
	}
	return regions;
    }

    public static Vector getEnrichedRegions(int c, int fragmentLength, String fileName, int minClusterSize, boolean extend, int strand) {
	//Use this method when loading a set of saved enriched regions
	ABIRead reads[] = parseFiles(c, fileName + "_E", fragmentLength, extend, strand);
	//Create the enriched regions. Easy since we only have to
	//consider the groups of fragments which are overlapping
	Vector regions = new Vector();
	if (reads.length>0) {
	    EnrichedRegion er = new EnrichedRegion(reads[0], c, reads[0].getLength());
	    for (int i=1; i<reads.length; i++) {
		if (er.overlaps(reads[i]) || er.numberOfReads()<minClusterSize) 
		    er.add(reads[i]);
		else {
		    regions.add(er);
		    er = new EnrichedRegion(reads[i], c, reads[i].getLength());
		}
	    }
	    regions.add(er);
	}
	return regions;
    }

    protected static EnrichedRegion findRegion(Vector regions, int chr, int center) {
	for (int i=0; i<regions.size(); i++) {
	    EnrichedRegion e = (EnrichedRegion)regions.elementAt(i);
	    if (e.getChromosome()==chr && e.getStart()<center && e.getStart() + e.getLength()>center)
		return e;
	}
	System.out.println("Warning, could not find region at chr " + chr + ": " + center);
	return null;
    }

    //////////////////////////////////////////////////////////////////
    //Use the functions below to find the significant peaks

    public static TreeSet findSignificantClusters(int c, int fragmentLength, double pValue, String fileName, String inputFileName, String GENOME_LOC, boolean extend, int strand) {
	Vector regs = findEnrichedRegions(c, fragmentLength, fileName, extend, strand);
	int nInputReads = findInputReads(c, fragmentLength, regs, inputFileName, extend, strand);
	return findSignificantClusters(c, fragmentLength, pValue, regs, nInputReads, null, GENOME_LOC);
    }

    public static TreeSet findSignificantClusters(int c, int fragmentLength, double pValue, Vector regs, int nInputReads, String fileName, String GENOME_LOC) {
	int nBasepairs = 0, nUniqueBasePairs = 0, nWildcards = 0, totReads = 0;
	TreeSet saveRegions = new TreeSet();
	//count the number of reads
	for (int i=0; i<regs.size(); i++) { totReads += ((EnrichedRegion)regs.elementAt(i)).numberOfReads(); }
	String fName = GENOME_LOC + "chr" + AnalyzeControls.getChromosomeNumber(c) + "_stats_" + AnalyzeControls.READ_LENGTH + ".txt";
	System.out.println("Looking for " + fName);
	try {
	    BufferedReader br = new BufferedReader(new FileReader(fName));
	    Pattern statsPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)");
	    String line = br.readLine();
	    Matcher m = statsPattern.matcher(line);
	    if (m.find()) {
		nBasepairs = Integer.parseInt(m.group(1));
		nUniqueBasePairs = Integer.parseInt(m.group(2));
		nWildcards = Integer.parseInt(m.group(3));
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error reading " + fName + "\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
	double reads2SitesRatio = (double)totReads/(double)nUniqueBasePairs, inputReads2SitesRatio = (double)nInputReads/(double)nUniqueBasePairs;	//Find out how many potential binding sites there are
	//Find the clusters and calculate the effective cluster-size
	//based on the potential number of reads which can be mapped
	//to the region.
	TreeMap clusterSizes = new TreeMap(), effectiveClusterSizes = new TreeMap();
	fName = GENOME_LOC + "chr" + AnalyzeControls.getChromosomeNumber(c) + "_" + AnalyzeControls.READ_LENGTH + ".bed";
	try {
	    Iterator it = regs.iterator(); //Assume that the regions are sorted
	    //open the file containing the mappability index
	    BufferedReader br = new BufferedReader(new FileReader(fName));
	    Pattern bedPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+([\\d|\\.]+)");
	    String line = "";
	    Matcher m;
	    int start = 0, end = 0, nFound = 0, nReadsInPeaks = 0;
	    double u = 0, up = 0;
	    while (it.hasNext()) {
		//The loops and ifs in here could use some cleaning
		//up, but for now we leave it as is since it works...
		EnrichedRegion er = (EnrichedRegion)it.next();
		if (er.numberOfReads()>1) {
		    //Calculate the number of reads which can be mapped into
		    //this region
		    while (end<er.getStart() + er.getFragmentLength()) {
			line = br.readLine();
			if (line==null)
			    break;
			m = bedPattern.matcher(line);
			if (m.find()) {
			    start = Integer.parseInt(m.group(1));
			    end = Integer.parseInt(m.group(2));
			    u = Double.parseDouble(m.group(3));
			}
		    }
		    int potReads = 0;
		    boolean insideRegion = true;
		    while (insideRegion) {
			int i = 0;
			while (insideRegion && start + i<end) { //Found the mappability index corresponding to this region
			    if (start + i==er.getStart() + er.getFragmentLength()) {
				potReads += AnalyzeControls.READ_LENGTH*u;
				up = u;
				i++;
			    }
			    else if (start + i>er.getStart() + er.getFragmentLength()) {
				int tmp = u-up>=0 ? 1 : 0;
				tmp *= start + i>er.getStart() && er.getStart() + er.getLength()>start + i ? 2 : 1;
				potReads += tmp;
				up = u;
				i++;
				insideRegion = start + i<er.getStart() + er.getLength() - er.getFragmentLength();
			    }
			    else { i++; }
			}
			if (insideRegion) {
			    line = br.readLine();
			    if (line==null) { break; }
			    m = bedPattern.matcher(line);
			    if (m.find()) {
				start = Integer.parseInt(m.group(1));
				end = Integer.parseInt(m.group(2));
				up = u;
				u = Double.parseDouble(m.group(3));
			    }
			}
		    }
		    int maxReads = er.getLength()*2;
		    if (potReads<=0) {
			System.err.println("Error for region " + er.toString());
			break;
		    }
		    er.setEffectiveNumberOfReads(er.numberOfReads()*maxReads/potReads);
		    er.setPValue(calculatePValue(er.numberOfReads(), potReads, reads2SitesRatio/2));
		    er.setInputPValue(calculatePValue(Math.max(er.numberOfInputReads(), 2), potReads, inputReads2SitesRatio/2));
		    //if (er.getPValue()<pValue && er.getInputPValue()>inputPValue && er.getPValue()/er.getInputPValue()<pValue && er.numberOfReads()>minClustSize) {
		    if (er.getPValue()/er.getInputPValue()<pValue) {
			//System.out.println("Start " + er.getStart() + "-" + (er.getStart() + er.getLength()) + " has " + er.numberOfReads() + ", " + er.numberOfInputReads() + " with " + potReads + " potential, " + maxReads + " maximum, p-val = " + format.format(er.getPValue()) + " and input P-val " + format.format(er.getInputPValue()) + ", quality = " + format.format(er.getPeakQualityScore()));
			nFound++;
			nReadsInPeaks += er.numberOfReads();
			saveRegions.add(er);
		    }
		}
		else 
		    er.setEffectiveNumberOfReads(1);
		//Save for histogram
		Integer key = new Integer(er.numberOfReads());
		Integer val = clusterSizes.containsKey(key) ? (Integer)clusterSizes.get(key) : new Integer(0);
		clusterSizes.put(key, new Integer(val.intValue() + 1));
		//Store the effective cluster size
		key = new Integer(er.effectiveNumberOfReads());
		val = effectiveClusterSizes.containsKey(key) ? (Integer)effectiveClusterSizes.get(key) : new Integer(0);
		effectiveClusterSizes.put(key, new Integer(val.intValue() + 1));
	    }
	    br.close();
	    System.out.println("Found " + nFound + " and " + format.format((double)nReadsInPeaks/totReads) + " in peaks.");
	}
	catch (IOException e) {
	    System.err.println("Error reading " + fName + "\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
	if (fileName!=null) {
	    //Save the histograms. Use Matlab-file compile_peak_dist.m to
	    //combine the results for the different chromosomes.
	    try {
		String clusterFileName = fileName + "_" + c + "_peak_dist.txt";
		BufferedWriter bw = new BufferedWriter(new FileWriter(clusterFileName));
		Iterator it = clusterSizes.keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    bw.write(key + "\t" + clusterSizes.get(key) + "\n");
		}
		bw.close();
		String effectiveClusterFileName = fileName + "_" + c + "_eff_peak_dist.txt";
		bw = new BufferedWriter(new FileWriter(effectiveClusterFileName));
		it = effectiveClusterSizes.keySet().iterator();
		while (it.hasNext()) {
		    Integer key = (Integer)it.next();
		    bw.write(key + "\t" + effectiveClusterSizes.get(key) + "\n");
		}
		bw.close();
	    }
	    catch (IOException e) { }
	}
	return saveRegions;
    }
    
    public static double calculatePValue(int nReads, int potReads, double reads2SitesRatio) {
	//Use this method to calculate the p-value for observing a
	//given number of reads in a certain region where potReads can
	//be mapped. totReads is the total number of reads and
	//totSites is the total number of sites which a read can be
	//mapped to. The null-distribution is assumed to be Poisson
	//and the parameter is potReads*totReads/totSites.
	double lambda = ((double)potReads)*reads2SitesRatio;
	//return PoissonDist.barF(lambda, nReads);
	try {
	    return 1 - (new PoissonDistributionImpl(lambda)).cumulativeProbability(nReads);
	}
	catch (MathException e) { }
	return 0;
    }

    public static int findInputReads(int c, int fragmentLength, Vector regs, String inputFileName, boolean extend) {
	ABIRead inputReads[] = parseFiles(c, inputFileName, fragmentLength, extend, 0); //Find the input reads
	Iterator it = regs.iterator();
	int i = 0;
	while (it.hasNext()) {
	    EnrichedRegion e = (EnrichedRegion)it.next();
	    //Find the corresponding region in the input data
	    while (i<inputReads.length && inputReads[i].getPosition()<e.getStart())
		i++;
	    int nInputReads = 0; //Count the number of reads in the input data
	    while (i<inputReads.length && inputReads[i].getPosition()<e.getEnd() + e.getFragmentLength()) {
		nInputReads++;
		i++;
	    }
	    e.setInputReads(nInputReads);
	}
	return inputReads.length;
    }

    public static int findInputReads(int c, int fragmentLength, Vector regs, String inputFileName, boolean extend, int strand) {
	//Use this when we only want to look at one strand
	ABIRead inputReads[] = parseFiles(c, inputFileName, fragmentLength, extend, strand); //Find the input reads
	Iterator it = regs.iterator();
	int i = 0;
	while (it.hasNext()) {
	    EnrichedRegion e = (EnrichedRegion)it.next();
	    //Find the corresponding region in the input data
	    while (i<inputReads.length && inputReads[i].getPosition()<e.getStart())
		i++;
	    int nInputReads = 0; //Count the number of reads in the input data
	    while (i<inputReads.length && inputReads[i].getPosition()<e.getEnd() + e.getFragmentLength()) {
		nInputReads++;
		i++;
	    }
	    e.setInputReads(nInputReads);
	}
	return inputReads.length;
    }


    //////////////////////////////////////////////////////////////////

    public static ABIRead[] findUniqueReads(ABIRead sortedReads[]) {
	Arrays.sort(sortedReads); //just to make sure
	int nDoubles = 0, j = 1;
	for (int i=1; i<sortedReads.length; i++)
	    nDoubles += sortedReads[i].equals(sortedReads[i-1]) ? 1 : 0;
	System.out.println("Found " + nDoubles + " non-unique reads.");
	ABIRead reads[] = new ABIRead[sortedReads.length - nDoubles];
	reads[0] = sortedReads[0];
	for (int i=1; i<sortedReads.length; i++) {
	    if (!sortedReads[i].equals(sortedReads[i-1]))
		reads[j++] = sortedReads[i];
	    if (j>=reads.length) { break; }
	}
	return reads;
    }

    protected static ABIRead[] filterRandomReads(ABIRead uniqueReads[], int nKeep) {
	ABIRead reads[] = new ABIRead[nKeep];
	//Create an array with the numbers 0...reads.length-1 in
	//random order and then take the first nKeep of these
	//elements.
	int randVec[] = new int[uniqueReads.length];
	for (int i=0; i<randVec.length; i++) { randVec[i] = i; }
	for (int i=0; i<randVec.length*10; i++) { //swap two elements
	    int ind1 = random.nextInt(randVec.length), ind2 = random.nextInt(randVec.length), tmp = randVec[ind1];
	    randVec[ind1] = randVec[ind2];
	    randVec[ind2] = tmp;
	}
	for (int i=0; i<reads.length; i++) { reads[i] = uniqueReads[randVec[i]]; }
	Arrays.sort(reads);
	return reads;
    }

    protected static void savePositionsReads(ABIRead[] reads, String posReadsFileName, boolean unique, int fragmentLength, int chr, boolean header) {
	//Returns the number of unique positions with at least one read
	int nConstReads = 0, nDouble = 0;
	try {
	    Arrays.sort(reads);
	    if (unique) {
		reads = PeakFinder.findUniqueReads(reads);
		Arrays.sort(reads);
	    }
	    //Use a sliding window to count the number of reads for
	    //each position
	    LinkedList slidingWindowReads = new LinkedList();
	    //Initialize
	    for (int i=0; i<reads[0].getLength(); i++) 
		slidingWindowReads.addFirst(new Integer(0));
	    int slidingWindowStart = reads[0].getPosition(), readStart = 0, j = 0, prevReads = -1;
	    Vector starts = new Vector(), ends = new Vector(), readCount = new Vector();
	    starts.addElement(new Integer(reads[0].getPosition()));
	    //Write to a new file where the first column is the
	    //position, the second column is the number of reads
	    while (j<reads.length) {
		readStart = reads[j++].getPosition();
		if (readStart==slidingWindowStart) { //Same pos as before
		    //if (!unique || prevReads==-1) 
		    //System.out.println("Found duplicate reads at " + readStart);
		    slidingWindowReads = addRead(slidingWindowReads, reads[j].getLength());
		    //nDouble++;
		}
		else if (readStart>slidingWindowStart) { //Move window forward
		    while (readStart>slidingWindowStart) {
			if (prevReads==-1) //Initialization
			    prevReads = ((Integer)slidingWindowReads.getFirst()).intValue();
			//Remove the first entry of the window and save it
			if (((Integer)slidingWindowReads.getFirst()).intValue()!=prevReads) {
			    int r = ((Integer)slidingWindowReads.removeFirst()).intValue();
			    if (prevReads>0) {
				ends.addElement(new Integer(slidingWindowStart-1));
				readCount.addElement(new Integer(prevReads));
				nConstReads++;
			    }
			    if (r>0)
				starts.addElement(new Integer(slidingWindowStart));
			    prevReads = r;
			}
			else 
			    slidingWindowReads.removeFirst();
			slidingWindowStart++;
			//Add an empty element at the end to keep the size
			slidingWindowReads.addLast(new Integer(0));
		    }
		    slidingWindowReads = addRead(slidingWindowReads, reads[0].getLength());
		}
		else
		    System.err.println("Error! found read starting earlier than window at " + reads[j-1].getPosition());
	    }
	    int lastWindowCount = slidingWindowStart;
	    while (!slidingWindowReads.isEmpty()) { //Write the last window
		if (((Integer)slidingWindowReads.getFirst()).intValue()!=prevReads) {
		    int r = ((Integer)slidingWindowReads.removeFirst()).intValue();
		    if (prevReads>0) {
			ends.addElement(new Integer(lastWindowCount-1));
			readCount.addElement(new Integer(prevReads));
			nConstReads++;
		    }
		    if (r>0)
			starts.addElement(new Integer(lastWindowCount));
		    prevReads = r;
		}
		else
		    slidingWindowReads.removeFirst();
		lastWindowCount++;
	    }
	    ends.addElement(new Integer(lastWindowCount - 1));
	    readCount.addElement(new Integer(prevReads));
	    nConstReads++;
	    BufferedWriter bw = null; //Save as a .bed file
	    //System.out.println("Saving with " + readCount.size() + " to " + posReadsFileName);
	    bw = new BufferedWriter(new FileWriter(posReadsFileName));
	    if (header) {
		if (chr==20) 
		    bw.write("browser position chrX:3000000-4000000");
		else if (chr==21)
		    bw.write("browser position chrY:3000000-4000000");
		else 
		    bw.write("browser position chr" + chr + ":3000000-4000000\n");
		//Header
		String colorStr = "", name = posReadsFileName;
		colorStr = posReadsFileName.matches("KCl") ? "255," : "0,";
		colorStr += posReadsFileName.matches("input") ? "255," : "0,";
		colorStr += posReadsFileName.matches("MeCP2") ? "127" : "0";
		name += unique ? "_singletons" : "";
		bw.write("track type=wiggle_0 name=\"" + name + "\" description=\"" + name + "\" visibility=full autoScale=off viewLimits=0:10 color=\"" + colorStr + "\" maxHeightPixels=70:40:10 priority=10\n");
	    }
	    for (int i=0; i<starts.size(); i++) {
		if (chr==20)
		    bw.write("chrX\t" + (((Integer)starts.elementAt(i)).intValue()-1) + "\t" + (((Integer)ends.elementAt(i)).intValue()) + "\t" + (((Integer)readCount.elementAt(i)).intValue()) + "\n");
		else if (chr==21)
		    bw.write("chrY\t" + (((Integer)starts.elementAt(i)).intValue()-1) + "\t" + (((Integer)ends.elementAt(i)).intValue()) + "\t" + (((Integer)readCount.elementAt(i)).intValue()) + "\n");
		else 
		    bw.write("chr" + chr + "\t" + (((Integer)starts.elementAt(i)).intValue()-1) + "\t" + (((Integer)ends.elementAt(i)).intValue()) + "\t" + (((Integer)readCount.elementAt(i)).intValue()) + "\n");
	    }
	    bw.close();
	}
	catch(IOException e) {
	    System.err.println("Error saving to " + posReadsFileName + "\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
    }

    protected static LinkedList addRead(LinkedList l, int fragmentLength) {
	for (int i=0; i<fragmentLength; i++)
	    l.set(i, new Integer(((Integer)l.get(i)).intValue() + 1));
	return l;
    }

    
}

