//Use this class for finding peaks. All methods in the class are
//static and can thus be called by outside classes. The basic
//functions here are:
//
//parseFiles - load the files with the reads and parse the contents
//
//findEnrichedRegions - find those peaks which have more than expected reads
//

import java.util.*;
import java.util.regex.*;
import java.io.*;
import org.apache.commons.math.distribution.PoissonDistributionImpl;
import org.apache.commons.math.MathException;
//import umontreal.iro.lecuyer.probdist.*;
import java.text.DecimalFormat;
import cern.jet.math.Bessel;

public class PeakFinder2 {

    protected static Random random = new Random();
    public static DecimalFormat format = new DecimalFormat("#####.#####");

    public static Pattern posReadPattern = Pattern.compile("chr([\\d|X|Y]+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
    //The groups are bead identifier, chr, strand, pos, mismatches and quality score
    public static Pattern rawPattern1 = Pattern.compile(">([0-9]+_[0-9]+_[0-9]+_[F|R][0-9]+),([0-9]+)_(-*?)([0-9]+)\\.([0-3])(?:,([0-9]+\\.[0-9]+))?");
    //Use this pattern when we parse files which were randomly
    //generated. They typically have a shorter header and no barcode
    //info. The first group is only there to make sure that the
    //numbers for the other groups will match with rawPattern1.
    public static Pattern randPattern1 = Pattern.compile("(>),?([0-9|X|Y]+)_(-*?)([0-9]+)[\\.([0-3])]*?");
    //The second line contains information about the output
    public static Pattern rawPattern2 = Pattern.compile("T([0-3]+)");
    //For the files containing the codes
    public static Pattern codePattern = Pattern.compile("([0-3]+)\t([0-9]+)");
    //For files which have been pre-processed and the sequence has
    //already been extracted
    public static Pattern processedPattern = Pattern.compile(">(\\d+)([+|-])");


    public static final int[] CHROMOSOME_LENGTH = new int[]{197195432, 181748087, 159599783, 155630120, 152537259, 149517037, 152524553, 131738871, 124076172, 129993255, 121843856, 121257530, 120284312, 125194864, 103494974, 98319150, 95272651, 90772031, 61342430, 166650296, 15902555};
    public static final int[] CHROMOSOME_LENGTH_HS = new int[]{247249724, 242951154, 199501832, 191273068, 180857871, 170899997, 158821429, 146274831, 140273257, 135374743, 134452390, 132349540, 114142986, 106368591, 100338921, 88827260, 78774748, 76117159, 63811657, 62435970, 46944329, 49691438, 154913759, 57772959};
    public static final int[] CHROMOSOME_LENGTH_SC = new int[]{230208, 813178, 316617, 1531919, 576869, 270148, 1090947, 562643, 439885, 745741, 666454, 1078175, 924429, 784333, 1091289, 948062, 85779};
    public static final long TOTAL_CHROMOSOME_LENGTH = (long)(265489521.8*10);
    public static final long TOTAL_CHROMOSOME_LENGTH_HS = (long)(3.080419613e+09);
    public static final String INPUT_GC_BIAS_FILE_NAME = "inputKCl_GC_bias_";
    public static final int MIN_THRESHOLD_HEIGHT = 3;

    //The loci in this file should not be considered since they
    //correspond to ribosomes or other suspiciously behaving regions
    public static final String JUNK_FILE_NAME = "mm9_Junk_loci.txt";

    protected Mappability map;
    protected int chrPos, lastReadPos, readsInd = 0, inputInd = 0, chr, strand;
    protected ABIRead[] reads, inputReads;
    protected String GENOME_LOC;

    //////////////////////////////////////////////////////////////////

    public PeakFinder2() { }

    /*public PeakFinder2(boolean orchestra, int c, String fileName, String inputFileName, int fragmentLength, boolean extend, int str) {
	GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	chr = c;
	strand = str;
	map = new Mappability(GENOME_LOC, chr);
	chrPos = 0;
	lastReadPos = map.getNumberOfBases();
	reads = parseFiles(chr, fileName, fragmentLength, extend, strand);
	if (inputFileName!=null)
	    inputReads = parseFiles(chr, inputFileName, fragmentLength, extend, strand);
	    }*/

    /*public PeakFinder2(boolean orchestra, int c, ABIRead r[], ABIRead input[]) {
	//Use this when we want to avoid parsing the file multiple times. 
	GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	chr = c;
	map = new Mappability(GENOME_LOC, chr);
	chrPos = 0;
	lastReadPos = map.getNumberOfBases();
	reads = r;
	inputReads = input;
	}*/

    //////////////////////////////////////////////////////////////////


    public double getReads2SitesRatio() { return ((double)reads.length)/map.getNumberOfUniqueBases(); }

    public double getInputReads2SitesRatio() { return ((double)inputReads.length)/map.getNumberOfUniqueBases(); }

    public static ABIRead[] parseFilesWithSequence(int chr, String fileName, int fragmentLength, boolean extend, int strand) { return parseFilesWithSequence(chr, fileName, 3, true, -1, fragmentLength, extend, strand); }

    public static ABIRead[] parseFilesWithSequence(int chr, String fileName, int nMismatches, boolean unique, int nSample, int fragmentLength, boolean extend, int strand) {	//Use this method to analyze the reads which have already been
	//pre-processed and the sequences already extracted.
	int nForwardReads = 0, nReverseReads = 0;
	ABIRead r[] = null;
	try {
	    //count the number of reads on each strand
	    BufferedReader br = new BufferedReader(new FileReader(getProcessedReadsFileName(fileName, chr, extend, fragmentLength)));
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = processedPattern.matcher(line);
		if (m.find()) {
		    if (m.group(2).equals("+")) { nForwardReads++; }
		    else { nReverseReads++; }
		    line = br.readLine(); //Skip line with sequence info
		}
		line = br.readLine();
	    }
	    br.close();
	    if (strand<0) 
		r = new ABIRead[nReverseReads];
	    else if (strand>0) 
		r = new ABIRead[nForwardReads];
	    else 
		r = new ABIRead[nForwardReads + nReverseReads];
	    br = new BufferedReader(new FileReader(getProcessedReadsFileName(fileName, chr, extend, fragmentLength)));
	    line = br.readLine();
	    int i = 0;
	    while (line!=null) {
		Matcher m = processedPattern.matcher(line);
		if (m.find()) {
		    if (m.group(2).equals("+") && strand>0) { 
			r[i] = new ABIRead(chr, true, Integer.parseInt(m.group(1)), fragmentLength); 
			r[i++].setSequence(br.readLine());
		    }
		    else if (m.group(2).equals("-") && strand<0) { 
			r[i] = new ABIRead(chr, false, Integer.parseInt(m.group(1)), fragmentLength); 
			r[i++].setSequence(br.readLine());
		    }
		    else { 
			r[i] = new ABIRead(chr, m.group(2).equals("+"), Integer.parseInt(m.group(1)), fragmentLength); 
			r[i++].setSequence(br.readLine());
		    }
		}
		line = br.readLine();
	    }
	    if (r.length>0) {
		Arrays.sort(r);
		if (nSample<r.length && nSample>0)
		    r = sampleReads(r, nSample);
		if (unique) { r = findUniqueReads(r); }
		Arrays.sort(r);
	    }
	    else
		System.err.println("Warning: No reads found!");
	}
	catch (IOException e) {

	}
	return r;
    }

    public static ABIRead[] parseFiles(int chr, String fileName, int fragmentLength, boolean extend, int strand) { return parseFiles(chr, fileName, fragmentLength, extend, strand, 0, null); }

    public static ABIRead[] parseFiles(int chr, String fileName, int fragmentLength, boolean extend, int strand, double minQualityScore) {
	return parseFiles(chr, fileName, 3, true, -1, fragmentLength, extend, strand, minQualityScore, null);
    }

    public static ABIRead[] parseFiles(int chr, String fileName, int fragmentLength, boolean extend) {
	return parseFiles(chr, fileName, 3, true, -1, fragmentLength, extend, 0, 10, null);
    }

    public static ABIRead[] parseFiles(int chr, String fileName, int fragmentLength, boolean extend, int strand, double minQualityScore, String statsSaveFileName) {
	return parseFiles(chr, fileName, 3, true, -1, fragmentLength, extend, strand, minQualityScore, statsSaveFileName);
    }

    public static ABIRead[] parseFiles(int chr, String fileName, int nMismatches, boolean unique, int nSample, int fragmentLength, boolean extend, int strand) {return parseFiles(chr, fileName, nMismatches, unique, nSample, fragmentLength, extend, strand, 0, null); }

    public static ABIRead[] parseFiles(int chr, String fileName, int nMismatches, boolean unique, int nSample, int fragmentLength, boolean extend, int strand, double minQualityScore, String statsSaveFileName) { return parseFiles(chr, fileName, nMismatches, unique, nSample, fragmentLength, extend, strand, 0, statsSaveFileName, false); }

    public static ABIRead[] parseFiles(int chr, String fileName, int nMismatches, boolean unique, int nSample, int fragmentLength, boolean extend, int strand, double minQualityScore, String statsSaveFileName, boolean allowRepeats) {
	try {
	    TreeSet<TranscribedRegion2> repeatRNAs = !allowRepeats ? EstimateGenicTranscription3.parseRepeatFile(EstimateGenicTranscription3.REPEAT_CHIP_FILE_MM, "mm", chr) : new TreeSet<TranscribedRegion2>();
	    int nReadsForward = 0, nReadsReverse = 0, nQualityLost = 0, nMismatchesCount[] = new int[4];
	    BufferedReader br = new BufferedReader(new FileReader(fileName + "_" + chr));
	    String line = br.readLine();
	    //Go over once just to count the number of reads for this
	    //chromosome
	    while (line!=null) {
		Matcher m = rawPattern1.matcher(line);
		if (m.find()) {
		    boolean forward = !m.group(3).equals("-");
		    int nMisses = Integer.parseInt(m.group(5)), pos = Integer.parseInt(m.group(4));
		    //System.out.println(line + "\t" + m.groupCount() + "\t" + m.group(5));
		    double qualityScore = minQualityScore<0 || m.group(6)==null ? minQualityScore + 1 : Double.parseDouble(m.group(6)); //Don't bother looking if limit is negative
		    //System.out.println(m.group(6) + "\t" + qualityScore);
		    if (nMisses<=nMismatches && Integer.parseInt(m.group(2))==chr) {
			if (!ABIReads.overlapsTranscript(chr, pos - (forward ? 0 : fragmentLength), pos + (forward ? fragmentLength : 0), repeatRNAs)) {
			    if (qualityScore>minQualityScore) {
				if (forward) { nReadsForward++; }
				else { nReadsReverse++; }
				nMismatchesCount[nMisses]++;
			    }
			    else
				nQualityLost++;
			}
		    }
		    line = br.readLine();
		}
		line = br.readLine(); //Get the sequence information
	    }
	    br.close();
	    //Go through again and store them. We want to use an
	    //ordinary array to save memory, going through the file is
	    //not that slow.
	    if (strand<0)
		return parseFile(chr, fileName, nMismatches, nReadsReverse, unique, nSample, fragmentLength, extend, strand, minQualityScore, statsSaveFileName);
	    else if (strand>0)
		return parseFile(chr, fileName, nMismatches, nReadsForward, unique, nSample, fragmentLength, extend, strand, minQualityScore, statsSaveFileName);
	    else
		return parseFile(chr, fileName, nMismatches, nReadsForward + nReadsReverse, unique, nSample, fragmentLength, extend, strand, minQualityScore, statsSaveFileName);
	}
	catch (IOException e) {
	    System.err.println("Error, could not open file: " + fileName);
	}
	return null;
    }

    protected static ABIRead[] parseFile(int chr, String fileName, int nMismatches, int nReads, boolean unique, int nSample, int fragmentLength, boolean extend, int strand, double minQualityScore, String statsSaveFileName) { return parseFile(chr, fileName, nMismatches, nReads, unique, nSample, fragmentLength, extend, strand, minQualityScore, statsSaveFileName, false); }

    protected static ABIRead[] parseFile(int chr, String fileName, int nMismatches, int nReads, boolean unique, int nSample, int fragmentLength, boolean extend, int strand, double minQualityScore, String statsSaveFileName, boolean allowRepeats) {
	//Use an int to determine which strand to return. Negative
	//means reverse strand, positive forward strand and zero means
	//both.
	ABIRead reads[] = new ABIRead[nReads];
	TreeSet<TranscribedRegion2> repeatRNAs = !allowRepeats ? EstimateGenicTranscription3.parseRepeatFile(EstimateGenicTranscription3.REPEAT_CHIP_FILE_MM, "mm", chr) : new TreeSet<TranscribedRegion2>();
	try {
	    int i = 0, nQualityLost = 0, nMismatchesCount[] = new int[4];
	    BufferedReader br = new BufferedReader(new FileReader(fileName + "_" + chr));
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = rawPattern1.matcher(line);
		if (m.find()) {
		    boolean forward = !m.group(3).equals("-");
		    int nMisses = Integer.parseInt(m.group(5)), pos = Integer.parseInt(m.group(4));
		    //System.out.println(line + "\t" + m.groupCount() + "\t" + m.group(5));
		    double qualityScore = minQualityScore<0 || m.group(6)==null ? minQualityScore + 1 : Double.parseDouble(m.group(6)); //Don't bother looking if limit is negative
		    //System.out.println(m.group(6) + "\t" + qualityScore);
		    if (Integer.parseInt(m.group(2))==chr && nMisses<=nMismatches) {
			if (!ABIReads.overlapsTranscript(chr, pos - (forward ? 0 : fragmentLength), pos + (forward ? fragmentLength : 0), repeatRNAs)) {
			    if (qualityScore>minQualityScore) {
				if (strand==0) {
				    if (forward)
					reads[i++] = new ABIRead(chr, true, pos + 1 + (extend ? 0 : fragmentLength/2), extend ? fragmentLength : AnalyzeControls.READ_LENGTH, qualityScore, nMisses);
				    else
					reads[i++] = new ABIRead(chr, false, pos + 1 - (extend ? fragmentLength : fragmentLength/2), extend ? fragmentLength : AnalyzeControls.READ_LENGTH, qualityScore, nMisses);
				}
				else if (forward && strand>0) //Use 1-based index when saving
				    reads[i++] = new ABIRead(chr, true, pos + 1 + (extend ? 0 : fragmentLength/2), extend ? fragmentLength : AnalyzeControls.READ_LENGTH, qualityScore, nMisses);
				else if (!forward && strand<0)
				    reads[i++] = new ABIRead(chr, false, pos + 1 - (extend ? fragmentLength : fragmentLength/2) - AnalyzeControls.READ_LENGTH, extend ? fragmentLength : AnalyzeControls.READ_LENGTH, qualityScore, nMisses);
			    }
			    else
				nQualityLost++;
			}
			line = br.readLine();
		    }
		    else //Skip the next line
			line = br.readLine();
		}
		line = br.readLine();
	    }
	    br.close();
	    if (reads.length>0) {
		Arrays.sort(reads);
		if (unique) { reads = findUniqueReads(reads); }
		if (nSample<reads.length && nSample>0)
		    reads = sampleReads(reads, nSample);
		Arrays.sort(reads);
	    }
	    else
		System.err.println("Warning: No reads found!");
	    if (statsSaveFileName!=null) { //Save the basic statistics
		try { //Not necessary, but want to catch write errors separately
		    BufferedWriter bw = new BufferedWriter(new FileWriter(statsSaveFileName));
		    if (chr==1)
			bw.write("#Tot reads\t-Quality\tUnique\t0 mismatch\t1 mismatch\t2 mismatch\t3 mismatch\n");
		    for (int j=0; j<reads.length; j++) //Count the number of mismatches after the QPCR filtering
			nMismatchesCount[reads[j].numberOfMismatches()]++;
		    bw.write(nReads + "\t" + (nReads - nQualityLost) + "\t" + reads.length + "\t" + nMismatchesCount[0] + "\t" + nMismatchesCount[1] + "\t" + nMismatchesCount[2] + "\t" + nMismatchesCount[3] + "\n");
		    bw.close();
		}
		catch (IOException e) {
		    System.err.println("Error saving to " + statsSaveFileName);
		}
	    }
	}
	catch (IOException e) {
	    System.err.println("Failed to open " + fileName + "_" + chr);
	}
	return reads;
    }

    public static ABIRead[] parseBroadFiles(int chr, String fileName, int fragmentLength, boolean extend) { return parseBroadFiles(chr, fileName, fragmentLength, extend, fileName + "_read_stats_" + chr + (extend ? "_E" + fragmentLength : "") + ".txt"); } 

    public static ABIRead[] parseBroadFiles(int chr, String fileName, int fragmentLength, boolean extend, String statsSaveFileName) {
	ABIRead reads[] = null;
	try {
	    TreeSet<TranscribedRegion2> repeatRNAs = EstimateGenicTranscription3.parseRepeatFile(EstimateGenicTranscription3.REPEAT_CHIP_FILE_MM, "mm", chr);
	    //First count the number of reads
	    int nQualityLost = 0, nMismatchesCount[] = new int[4];
	    BufferedReader br = new BufferedReader(new FileReader(fileName));
	    String line = br.readLine();
	    //chr start end strand something mismatches seq
	    Pattern readPattern = Pattern.compile("^chr([X|Y|\\d]+)\\s+(\\d+)\\s+(\\d+)\\s+([+|-]\\s+.+\\s+(\\d))");
	    int nReads = 0;
	    LinkedHashSet<ABIRead> rs = new LinkedHashSet<ABIRead>();
	    while (line!=null) {
		Matcher m = readPattern.matcher(line);
		if (m.find()) {
		    boolean forward = m.group(4).equals("+");
		    int c, pos = Integer.parseInt(m.group(forward ? 2 : 3)), nMisses = Integer.parseInt(m.group(5));
		    if (nMisses>3) 
			nQualityLost++;
		    else {
			if (m.group(1).equals("X"))
			    c = 20;
			else if (m.group(1).equals("Y"))
			    c = 21;
			else
			    c = Integer.parseInt(m.group(1));
			if ((chr==-1 || c==chr) && !ABIReads.overlapsTranscript(c, pos - (forward ? 0 : fragmentLength), pos + (forward ? fragmentLength : 0), repeatRNAs)) {
			    rs.add(new ABIRead(c, forward, pos, fragmentLength));
			    nMismatchesCount[nMisses]++;
			}
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	    //Now we can store them
	    reads = new ABIRead[rs.size()];
	    int i = 0;
	    for (ABIRead r : rs)
		reads[i++] = r;
	    Arrays.sort(reads);
	    if (statsSaveFileName!=null) { //Save the basic statistics
		try { //Not necessary, but want to catch write errors separately
		    BufferedWriter bw = new BufferedWriter(new FileWriter(statsSaveFileName));
		    if (chr==1)
			bw.write("#Tot reads\t-Quality\tUnique\t0 mismatch\t1 mismatch\t2 mismatch\t3 mismatch\n");
		    for (int j=0; j<reads.length; j++) //Count the number of mismatches after the QPCR filtering
			nMismatchesCount[reads[j].numberOfMismatches()]++;
		    bw.write(reads.length + "\t" + (reads.length - nQualityLost) + "\t" + reads.length + "\t" + nMismatchesCount[0] + "\t" + nMismatchesCount[1] + "\t" + nMismatchesCount[2] + "\t" + nMismatchesCount[3] + "\n");
		    bw.close();
		}
		catch (IOException e) {
		    System.err.println("Error saving to " + statsSaveFileName);
		}
	    }
	}
	catch (IOException e) {
	    System.err.println("Failed to open " + fileName + " : " + e.toString());
	}
	return reads;
    }

    public static int numberOfReads(int chr, String fileName, int nMismatches, int strand, double minQualityScore) {
	try {
	    int nReadsForward = 0, nReadsReverse = 0;
	    BufferedReader br = new BufferedReader(new FileReader(fileName + "_" + chr));
	    String line = br.readLine();
	    //Go over once just to count the number of reads for this
	    //chromosome
	    while (line!=null) {
		Matcher m = rawPattern1.matcher(line);
		if (m.find()) {
		    boolean forward = !m.group(3).equals("-");
		    int nMisses = Integer.parseInt(m.group(5));
		    double qualityScore = minQualityScore<0 || m.group(6)==null ? minQualityScore + 1 : Double.parseDouble(m.group(6)); //Don't bother looking if limit is negative
		    if (nMisses<=nMismatches && Integer.parseInt(m.group(2))==chr && qualityScore>minQualityScore) {
			if (forward) { nReadsForward++; }
			else { nReadsReverse++; }
		    }
		    line = br.readLine();
		}
		line = br.readLine(); //Get the sequence information
	    }
	    br.close();
	    //Go through again and store them. We want to use an
	    //ordinary array to save memory, going through the file is
	    //not that slow.
	    if (strand<0)
		return nReadsReverse;
	    else if (strand>0)
		return nReadsForward;
	    else
		return nReadsForward + nReadsReverse;
	}
	catch (IOException e) {
	    System.err.println("Error, could not open file: " + fileName);
	}
	return -1;
    }

    public static int numberOfReads(String fileName, int nMismatches, int strand, double qualityScore) {
	int nReads = 0;
	for (int i=0; i<21; i++)
	    nReads += numberOfReads(i + 1, fileName, nMismatches, strand, qualityScore);
	return nReads;
    }


    //////////////////////////////////////////////////////////////////
    //Use a window of size 2*fragmentLength to scan through the genome
    //and pick out the ones which have significantly more reads than
    //the input. In particular, we will be taking the mappability of
    //the genome into consideration.

    public static TreeSet findEnrichedRegions(String fileName, String inputFileName, int fragmentLength, int windowSize, boolean all, boolean extend, int strand, double pValue, int chr, boolean orchestra) {
	return findEnrichedRegions(parseFiles(chr, fileName, fragmentLength, extend, strand), parseFiles(chr, inputFileName, fragmentLength, extend, strand), windowSize, all, pValue, chr, strand, orchestra);
    }

    public static TreeSet findEnrichedRegions(ABIRead reads[], ABIRead inputReads[], int windowSize, double pValue, int chr, int strand, boolean orchestra) {
	return findEnrichedRegions(reads, inputReads, windowSize, false, pValue, chr, strand, orchestra);
    }

    public static TreeSet findEnrichedRegions(ABIRead reads[], ABIRead inputReads[], int windowSize, boolean all, double pValue, int chr, int strand, boolean orchestra) {
	//Instead of using the method by Johnson et al (2007) where we
	//zoom in on the clusters with large number of reads, we scan
	//through the genome with a fixed-size window. This should be
	//better in the sense that it does not introduce any bias in
	//the selection of regions. Moreover, scanning the genome is
	//not going to be that slow since we can automatically skip
	//regions with few basepairs.
	String GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	Mappability mappability = null; //new Mappability(GENOME_LOC, chr);
	//Calculate the parameters for the Poisson distribution for
	//the reads and the input
	double reads2SitesRatio = ((double)reads.length)/mappability.getNumberOfUniqueBases(), inputReads2SitesRatio = ((double)inputReads.length)/mappability.getNumberOfUniqueBases();
	//Calculate the minimum number of reads which we must have in
	//a window if there are no input reads and only half of the
	//positions are mappable. This should provide an approximate
	//lower bound on the number of reads required in the window.
	TreeSet regs = new TreeSet();
	int readsInd = 0, inputInd = 0;
	int chrPos = 0/*(reads[0]/windowSize)*windowSize*/, lastReadPos = mappability.getNumberOfBases() /*Math.max(reads[reads.length-1], inputReads[inputReads.length-1])*/;
	while (chrPos<lastReadPos) {
	    ChromosomeWindow reg = new ChromosomeWindow(chr, chrPos, chrPos + windowSize); //Find the reads in the window
	    while (readsInd<reads.length && reads[readsInd].getPosition()<chrPos + windowSize) 
		reg.add(reads[readsInd++]);
	    //See how many input reads there are in this region
	    while (inputInd<inputReads.length && inputReads[inputInd].getPosition()<chrPos + windowSize) { reg.addInputRead(); inputInd++; }
	    //Calculate how many reads can be mapped to this region
	    //reg.setPotentialNumberOfReads(mappability.numberOfPotentialReads(chrPos, windowSize, strand==0));
	    reg.setPotentialNumberOfReads(mappability.numberOfPotentialReadsNextWindow());
	    if (reg.numberOfReads()>0) {
		//Calculate the p-value based on a Poisson-distribution
		//System.out.println(reg.toString() + " has " + reg.numberOfPotentialReads() + " and lambda " + reads2SitesRatio);
		if (reg.numberOfReads()>0 && reg.numberOfPotentialReads()>0) {
		    reg.setLogPValue(calculateLogPValue(reg.numberOfReads(), reg.numberOfPotentialReads(), reads2SitesRatio));
		    reg.setLogInputPValue(calculateLogPValue(reg.numberOfInputReads(), reg.numberOfPotentialReads(), inputReads2SitesRatio));
		}
		if (reg.getPValueRatio()<pValue || all)
		    regs.add(reg);
	    }
	    else if (all) {
		//reg.setPotentialNumberOfReads(mappability.numberOfPotentialReads(chrPos, windowSize, strand==0));
		reg.setLogPValue(0);
		if (reg.numberOfInputReads()>0 && reg.numberOfPotentialReads()>0) 
		    reg.setLogInputPValue(calculateLogPValue(reg.numberOfInputReads(), reg.numberOfPotentialReads(), inputReads2SitesRatio));
		else
		    reg.setLogInputPValue(0);
		regs.add(reg);
	    }
	    chrPos += windowSize;
	}
	return regs;
    }

    public static TreeMap addObservation2Histogram(int obs, TreeMap hist) {
	Integer key = new Integer(obs);
	Integer val = hist.containsKey(key) ? (Integer)hist.get(key) : new Integer(0);
	hist.put(key, new Integer(val.intValue() + 1));
	return hist;
    }

    public static String printHistogram(TreeMap hist) {
	String str = "";
	Iterator it = hist.keySet().iterator();
	while (it.hasNext()) {
	    Integer key = (Integer)it.next();
	    str += key.toString() + "\t" + ((Integer)hist.get(key)).toString() + "\n";
	}
	return str;
    }

    public static TreeSet findEnrichedRegionsScanningWindow(ABIRead reads[], ABIRead inputReads[], int windowSize, int nWindowStep, String fileStart, double fdr, boolean gcCorrection, int chr, int strand, boolean orchestra) {
	//Use this method to extract a window with fixed size
	String GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	Mappability mappability = null; //new Mappability(GENOME_LOC, chr, windowSize);
	FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(chr));
	//Get the cut-offs for the differences in number of reads from a file
	FDRCutOffs cutOffs = new FDRCutOffs(fileStart, fdr, true);
	int fragmentLength = reads[0].getLength(), j = 0, nDiscarded = 0;
	int readsInd = 0, inputReadsInd = 0, nWindowsFound = 0;
	double reads2SitesRatio = 0, inputReads2SitesRatio = 0;
	ChromosomeWindow win = new ChromosomeWindow(chr, 0, windowSize);
	if (gcCorrection) { win.setSequence(fastaSeq.getNextSequence(1, windowSize)); }
	TreeSet significantWins = new TreeSet();
	while (inputReadsInd<inputReads.length && readsInd<reads.length && win.getEnd()<CHROMOSOME_LENGTH[chr-1]) {
	    // Remove reads shifted out of the window
	    win.shiftRight(nWindowStep);
	    //Add the reads to the window
	    while (readsInd<reads.length && reads[readsInd].getPosition()<=win.getStart() + windowSize - reads[readsInd].getLength()) //Found read going into the window
		win.add(reads[readsInd++]);
	    while (inputReadsInd<inputReads.length && inputReads[inputReadsInd].getPosition()<=win.getStart() + windowSize - inputReads[inputReadsInd].getLength()) //Found input read going into the window
		win.addInputRead(inputReads[inputReadsInd++]);
	    if (gcCorrection) { 
		win.addSequence(fastaSeq.getNextSequence(win.getEnd(), nWindowStep)); //update the sequence
	    }
	    //Calculate the potential number of reads
	    win.setPotentialNumberOfReads(mappability.numberOfPotentialReadsNextWindow(nWindowStep));
	    //if (win.getStart()>5000000) { break; }
	    if (win.numberOfPotentialReads()>0) { //Mappable window
		//Find if the difference in terms of number of reads is significant
		int zCutOff = cutOffs.getDiffThreshold(win.numberOfPotentialReads(), win.gcContent());
		//Find the maximum number of input-reads allowed
		int maxInputReads = cutOffs.getMaxInput(win.numberOfPotentialReads(), win.gcContent());
		int z = win.numberOfReads() - win.numberOfInputReads();
		if (z>=zCutOff && win.numberOfInputReads()<=maxInputReads) {
		    //System.out.println("For " + win.toString() + ", the cut-off is " + zCutOff);
		    win.setLambda(reads2SitesRatio);
		    win.setDetectionLimitDistance(z-zCutOff);
		    //if (win.getDetectionLimitDistance()>0)
			significantWins.add(win);
		}
		else if (z>=zCutOff && win.numberOfInputReads()>maxInputReads) {
		    System.out.println("Found significant window with too many inputs at " + win.getStart());
		  nDiscarded++;
		}
	    }
	    win = new ChromosomeWindow(win);
	}
	System.out.println("Found " + significantWins.size() + " significant windows and discarded " + nDiscarded);
	//Go through the list of significant windows and merge the
	//ones which are adjacent
	TreeSet wins2merge = new TreeSet(), regs = new TreeSet();
	Iterator it = significantWins.iterator();
	if (it.hasNext()) {
	    ChromosomeWindow prevWin = (ChromosomeWindow)it.next();
	    wins2merge.add(prevWin);
	    while (it.hasNext()) {
		ChromosomeWindow w = (ChromosomeWindow)it.next();
		if (prevWin.overlaps(w.getStart()))
		    wins2merge.add(w);
		else {
		    if (wins2merge.size()>1) { //Merge the windows
			ChromosomeWindow newWin = mergeWindows(wins2merge, reads2SitesRatio, inputReads2SitesRatio, gcCorrection, nWindowStep);
			int maxLimit = 0; //Use the max of the merged windows
			Iterator it2 = wins2merge.iterator();
			while (it2.hasNext()) {
			    ChromosomeWindow mw = (ChromosomeWindow)it2.next();
			    maxLimit = maxLimit>mw.getDetectionLimitDistance() ? maxLimit : mw.getDetectionLimitDistance();
			}
			newWin.setDetectionLimitDistance(maxLimit); 
			regs.add(newWin);
			//System.out.println("Added merged " + wins2merge.size() + " windows " + newWin.toString());
		    }
		    else
			regs.add((ChromosomeWindow)wins2merge.first()); //just add the window
		    wins2merge = new TreeSet();
		    wins2merge.add(w);
		}
		prevWin = w;
	    }
	    //Add the last window
	    ChromosomeWindow newWin = mergeWindows(wins2merge, reads2SitesRatio, inputReads2SitesRatio, gcCorrection, nWindowStep);
	    int maxLimit = 0; //Use the max of the merged windows
	    Iterator it2 = wins2merge.iterator();
	    while (it2.hasNext()) {
		ChromosomeWindow mw = (ChromosomeWindow)it2.next();
		maxLimit = maxLimit>mw.getDetectionLimitDistance() ? maxLimit : mw.getDetectionLimitDistance();
	    }
	    newWin.setDetectionLimitDistance(maxLimit); 
	    //////////////////////////////////////////////////////////////////
	    //Trim the windows by only considering the parts which are
	    //above the background threshold. 
	    TreeSet removeSet = new TreeSet();
	    it = regs.iterator();
	    while (it.hasNext()) {
		ChromosomeWindow w = (ChromosomeWindow)it.next();
		int threshold = Math.max(MIN_THRESHOLD_HEIGHT, (int)Math.round((double)w.numberOfPotentialReads()*reads2SitesRatio*w.getGCMultiplier()));
		int upCrossings[] = w.getCrossings(threshold, true), downCrossings[] = w.getCrossings(threshold, false);
		//System.out.println(w.toString() + "\n has threshold " + threshold + " and " + upCrossings.length + " up crossings and " + downCrossings.length + " down.");
		//for (int i=0; i<upCrossings.length; i++)
		//  System.out.println("up: " + upCrossings[i]);
		//for (int i=0; i<downCrossings.length; i++)
		//  System.out.println("down: " + downCrossings[i]);
	    }
	    it = removeSet.iterator();
	    //while (it.hasNext())
	    //regs.remove((ChromosomeWindow)it.next());
	    System.out.println("After merging, there are " + regs.size()); // + ", removed " + removeSet.size() + " redundant peaks.");
	}
	return regs;
    }

    public static int getCutOff(TreeMap cutOffs, int mappability, int gcContent, int nMappabilityBins, int nGCBins) {
	int mappabilityBinSize = cutOffs.size(), gcBinSize = ((TreeMap)cutOffs.get(cutOffs.firstKey())).size();
	int mappabilityBinIndex = mappability/nMappabilityBins;
	if (mappabilityBinIndex==nMappabilityBins) { mappabilityBinIndex--; } //Adjust for the last bin
	int gcBinIndex = gcContent/gcBinSize;
	if (gcBinIndex==nGCBins) { gcBinIndex--; } //Adjust for the last bin
	return ((Integer)((TreeMap)cutOffs.get(new Integer(mappabilityBinIndex))).get(new Integer(gcBinIndex))).intValue();
    }

    /*public static TreeSet findEnrichedRegionsScanningWindow(ABIRead reads[], ABIRead inputReads[], int windowSize, int nWindowStep, int minMappability, double fdr, boolean gcCorrection, int chr, int strand, boolean orchestra) {
	//Use this method to extract a window with fixed size
	String GENOME_LOC = orchestra ? "/groups/neuroduo/martin/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	Mappability mappability = new Mappability(GENOME_LOC, chr, windowSize);
	FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(chr));
	GCBiasData gcBias = new GCBiasData(INPUT_GC_BIAS_FILE_NAME + windowSize + ".txt");
	//Calculate the parameters for the Poisson distribution for
	//the reads and the input
	int fragmentLength = reads[0].getLength(), j = 0;
	int readsInd = 0, inputReadsInd = 0, nWindowsFound = 0;
	double meanWinSize = 0, meanChipReads = 0, meanInputReads = 0;
	ChromosomeWindow win = new ChromosomeWindow(chr, 0, windowSize);
	if (gcCorrection) { win.setSequence(fastaSeq.getNextSequence(1, windowSize)); }
	TreeSet wins = new TreeSet(), significantWins = new TreeSet();
	TreeMap zHistogram = new TreeMap(), mappabilityCountHistogram = new TreeMap();
	//System.out.println(win.toString() + "\n" + win.getSequence().length() + ", " + win.getLength());
	try {
	    BufferedWriter windowDistWriter = new BufferedWriter(new FileWriter("win_dist_" + chr + ".txt"));
	while (inputReadsInd<inputReads.length && readsInd<reads.length && win.getEnd()<CHROMOSOME_LENGTH[chr-1]) {
	    // Remove reads shifted out of the window
	    win.shiftRight(nWindowStep);
	    //Add the reads to the window
	    while (readsInd<reads.length && reads[readsInd].getPosition()<=win.getStart() + windowSize - reads[readsInd].getLength()) //Found read going into the window
		win.add(reads[readsInd++]);
	    while (inputReadsInd<inputReads.length && inputReads[inputReadsInd].getPosition()<=win.getStart() + windowSize - inputReads[inputReadsInd].getLength()) //Found input read going into the window
		win.addInputRead(inputReads[inputReadsInd++]);
	    if (gcCorrection) { 
		win.addSequence(fastaSeq.getNextSequence(win.getEnd(), nWindowStep)); //update the sequence
		win.setGCMultiplier(gcBias.calculateGCMultiplier(win.getSequence())); //Calculate the GC-bias
	    }
	    //Calculate the potential number of reads
	    win.setPotentialNumberOfReads(mappability.numberOfPotentialReadsNextWindow(nWindowStep));
	    //System.out.println(win.toString() + "\n" + win.getSequence().length() + ", " + win.getLength());
	    //if (win.getStart()>5000000) { break; }
	    if (win.numberOfPotentialReads()>minMappability) { //Mappable window
		windowDistWriter.write(win.numberOfInputReads(true) + "\t" + win.numberOfReads() + "\t" + win.numberOfPotentialReads() + "\t" + format.format(win.getGCMultiplier()) + "\n"); 
		int z = win.calculateEffectiveReadDiff(fragmentLength);
		Integer key = new Integer(z);
		int nFound = zHistogram.containsKey(key) ? ((Integer)zHistogram.get(key)).intValue() : 0;
		zHistogram.put(key, new Integer(++nFound));
		//Count the number of windows with a given mappability index
		key = new Integer(win.numberOfPotentialReads());
		nFound = mappabilityCountHistogram.containsKey(key) ? ((Integer)mappabilityCountHistogram.get(key)).intValue() : 0;
		mappabilityCountHistogram.put(key, new Integer(++nFound));
		nWindowsFound++;
		meanChipReads += win.numberOfReads();
		meanInputReads += win.numberOfInputReads();
		meanWinSize += win.numberOfPotentialReads();
		//Can't store the windows due to lack of
		//memory. Instead make another pass once we have
		//established the threshold
		//wins.add(win);
	    }
	    win = new ChromosomeWindow(win);
	}
	windowDistWriter.close();
	}
	catch (IOException e) {
	    
	}
	meanWinSize /= nWindowsFound;
	//Number of reads per window is different than number of reads
	//per base since we must allow for some doubel counting when
	//the windows are overlapping
	meanChipReads /= (nWindowsFound);
	meanInputReads /= (nWindowsFound);
	//System.out.println("Number of windows binned by mappability is:");
	Iterator mapIt = mappabilityCountHistogram.keySet().iterator();
	while (mapIt.hasNext()) {
	    Integer key = (Integer)mapIt.next();
	    double ratio = (double)((Integer)mappabilityCountHistogram.get(key)).intValue()/nWindowsFound;
	    //System.out.println(key + "\t" + mappabilityCountHistogram.get(key) + "\t" + format.format(ratio) +"\t" + format.format((1-Math.exp(-meanChipReads*240.0/key.intValue()))/ratio));
	}
	double reads2SitesRatio = (double)reads.length/(2*mappability.getNumberOfUniqueBases());
	double inputReads2SitesRatio = (double)inputReads.length/(2*mappability.getNumberOfUniqueBases());
	System.out.println("Poisson-ratios are " + format.format(reads2SitesRatio) + " for ChIP and " + format.format(inputReads2SitesRatio) + " for " + reads.length + " reads and " + inputReads.length + " input reads at " + (2*mappability.getNumberOfUniqueBases()));
	//Now we can determine the cut-off required to be considered
	//significant difference
	int zCutOff = calculateFDRCutOff(fdr, zHistogram, nWindowsFound, meanChipReads, meanInputReads);
	System.out.println("the limit is " + zCutOff + " for " + nWindowsFound + " with mean size " + format.format(meanWinSize));
	//Go through all the windows and determine which ones are significant
	mappability = new Mappability(GENOME_LOC, chr, windowSize);
	fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(chr));
	win = new ChromosomeWindow(chr, 0, windowSize);
	win.setLambda(reads2SitesRatio);
	if (gcCorrection) { win.setSequence(fastaSeq.getNextSequence(1, windowSize)); }
	inputReadsInd = 0;
	readsInd = 0;
	while (inputReadsInd<inputReads.length && readsInd<reads.length && win.getEnd()<CHROMOSOME_LENGTH[chr-1]) {
	    // Remove reads shifted out of the window
	    win.shiftRight(nWindowStep);
	    //Add the reads to the window
	    while (readsInd<reads.length && reads[readsInd].getPosition()<=win.getStart() + windowSize - reads[readsInd].getLength()) //Found read going into the window
		win.add(reads[readsInd++]);
	    while (inputReadsInd<inputReads.length && inputReads[inputReadsInd].getPosition()<=win.getStart() + windowSize - inputReads[inputReadsInd].getLength()) //Found input read going into the window
		win.addInputRead(inputReads[inputReadsInd++]);
	    if (gcCorrection) { 
		win.addSequence(fastaSeq.getNextSequence(win.getEnd(), nWindowStep)); //update the sequence
		win.setGCMultiplier(gcBias.calculateGCMultiplier(win.getSequence())); //Calculate the GC-bias
	    }
	    //Calculate the potential number of reads
	    win.setPotentialNumberOfReads(mappability.numberOfPotentialReadsNextWindow(nWindowStep));
	    if (win.numberOfPotentialReads()>minMappability) { //Mappable window
		if (win.calculateEffectiveReadDiff(fragmentLength)>=zCutOff) {
		    win.setLambda(reads2SitesRatio);
		    win.setDetectionLimitDistance(fragmentLength, zCutOff);
		    if (win.numberOfInputReads()>6)
			System.out.println("accepted window with more than 6 input.");
		    if (win.getDetectionLimitDistance()>0)
			significantWins.add(win);
		}
	    }
	    win = new ChromosomeWindow(win);
	}
	System.out.println("Found " + significantWins.size() + " significant windows out of " + nWindowsFound + " and the limit is " + zCutOff);
	//Go through the list of significant windows and merge the
	//ones which are adjacent
	TreeSet wins2merge = new TreeSet(), regs = new TreeSet();
	Iterator it = significantWins.iterator();
	if (it.hasNext()) {
	    ChromosomeWindow prevWin = (ChromosomeWindow)it.next();
	    wins2merge.add(prevWin);
	    while (it.hasNext()) {
		ChromosomeWindow w = (ChromosomeWindow)it.next();
		if (prevWin.overlaps(w.getStart()))
		    wins2merge.add(w);
		else {
		    if (wins2merge.size()>1) { //Merge the windows
			ChromosomeWindow newWin = mergeWindows(wins2merge, gcBias, reads2SitesRatio, inputReads2SitesRatio, gcCorrection, nWindowStep);
			newWin.setGCMultiplier(gcBias.calculateGCMultiplier(newWin.getSequence())); //Calculate the GC-bias
			int maxLimit = 0; //Use the max of the merged windows
			Iterator it2 = wins2merge.iterator();
			while (it2.hasNext()) {
			    ChromosomeWindow mw = (ChromosomeWindow)it2.next();
			    maxLimit = maxLimit>mw.getDetectionLimitDistance() ? maxLimit : mw.getDetectionLimitDistance();
			}
			newWin.setDetectionLimitDistance(maxLimit); 
			regs.add(newWin);
			//System.out.println("Added merged " + wins2merge.size() + " windows " + newWin.toString());
		    }
		    else
			regs.add((ChromosomeWindow)wins2merge.first()); //just add the window
		    wins2merge = new TreeSet();
		    wins2merge.add(w);
		}
		prevWin = w;
	    }
	    //Add the last window
	    ChromosomeWindow newWin = mergeWindows(wins2merge, gcBias, reads2SitesRatio, inputReads2SitesRatio, gcCorrection, nWindowStep);
	    newWin.setGCMultiplier(gcBias.calculateGCMultiplier(newWin.getSequence())); //Calculate the GC-bias
	    int maxLimit = 0; //Use the max of the merged windows
	    Iterator it2 = wins2merge.iterator();
	    while (it2.hasNext()) {
		ChromosomeWindow mw = (ChromosomeWindow)it2.next();
		maxLimit = maxLimit>mw.getDetectionLimitDistance() ? maxLimit : mw.getDetectionLimitDistance();
	    }
	    newWin.setDetectionLimitDistance(maxLimit); 
	    //////////////////////////////////////////////////////////////////
	    //Trim the windows by only considering the parts which are
	    //above the background threshold. 
	    TreeSet removeSet = new TreeSet();
	    it = regs.iterator();
	    while (it.hasNext()) {
		ChromosomeWindow w = (ChromosomeWindow)it.next();
		int threshold = Math.max(MIN_THRESHOLD_HEIGHT, (int)Math.round((double)w.numberOfPotentialReads()*reads2SitesRatio*w.getGCMultiplier()));
		int upCrossings[] = w.getCrossings(threshold, true), downCrossings[] = w.getCrossings(threshold, false);
		//System.out.println(w.toString() + "\n has threshold " + threshold + " and " + upCrossings.length + " up crossings and " + downCrossings.length + " down.");
		//for (int i=0; i<upCrossings.length; i++)
		//  System.out.println("up: " + upCrossings[i]);
		//for (int i=0; i<downCrossings.length; i++)
		//  System.out.println("down: " + downCrossings[i]);
	    }
	    it = removeSet.iterator();
	    /*while (it.hasNext())
	      regs.remove((ChromosomeWindow)it.next());
	      System.out.println("After merging, there are " + regs.size()/* + ", removed " + removeSet.size() + " redundant peaks.");
	}
	return regs;
    }*/

    public static ChromosomeWindow mergeWindows(TreeSet wins, double reads2SitesRatio, double inputReads2SitesRatio, boolean gcCorrection, int nWindowStep) {
	//Use this method to merge a set of consequtive windows into one larger
	Iterator it = wins.iterator();
	ChromosomeWindow win = (ChromosomeWindow)it.next(), newWin = new ChromosomeWindow(win);
	//If the windows are merged pairwise, keeping track of the
	//potential number of reads becomes complicated and thus we
	//will do that manually.
	int potReads = win.numberOfPotentialReads(), prevWinPotReads = potReads, prevWinStart = win.getStart();
	while (it.hasNext()) {
	    win = (ChromosomeWindow)it.next();
	    newWin = newWin.merge(win);
	    int nWindowsSeparating = (win.getStart() - prevWinStart)/nWindowStep;
	    int diff = win.numberOfPotentialReads()-(prevWinPotReads-(nWindowsSeparating - 1)*nWindowStep*2);
	    potReads +=  diff>0 ? diff : 0;
	    prevWinPotReads = win.numberOfPotentialReads();
	    prevWinStart = win.getStart();
	}
	newWin.setPotentialNumberOfReads(potReads);
	newWin.setLogPValue(calculateLogPValue(newWin, reads2SitesRatio, gcCorrection));
	newWin.setLogInputPValue(calculateLogInputPValue(newWin, inputReads2SitesRatio, gcCorrection));
	return newWin;
    }
    
    public static int calculateFDRCutOff(double fdr, TreeMap zHistogram, int nWindows, double chipPoisson, double inputPoisson) {
	//Use this method to calculate the cut-off corresponding to a
	//given FDR ratio. Find the smallest key for which
	//zHistogram.get(key)/nWindows<fdr.
	System.out.println("kc = " + chipPoisson + ", ki = " + inputPoisson);
	Iterator it = zHistogram.keySet().iterator();
	double ratios[] = new double[zHistogram.keySet().size()], theoreticalCDF[] = new double[zHistogram.keySet().size()], theoretical2empirical[] = new double[zHistogram.keySet().size()];
	int keys[] = new int[zHistogram.keySet().size()], i = 0, minDiff = -1;
	while (it.hasNext()) {
	    Integer key = (Integer)it.next();
	    if (i>0)
		ratios[i] = ((double)((Integer)zHistogram.get(key)).intValue())/nWindows + ratios[i-1];
	    else
		ratios[i] = ((double)((Integer)zHistogram.get(key)).intValue())/nWindows;
	    keys[i] = key.intValue();
	    theoreticalCDF[i] = Math.min(1, skellamCDF(keys[i], chipPoisson, inputPoisson));
	    //System.out.println(keys[i] + "\t" + format.format(ratios[i]) + "\t" + ((Integer)zHistogram.get(key)).intValue() + "\t" + Math.round(skellamPDF(keys[i], chipPoisson, inputPoisson)*(double)nWindows) + "\t" + format.format(theoreticalCDF[i]) + "\t" + format.format((1-theoreticalCDF[i])/(1-ratios[i])));
	    //Calculate the ratio between the null-theoretical and empirical CDFs
	    theoretical2empirical[i] = (1-theoreticalCDF[i])/(1-ratios[i]);
	    if (keys[i]>0 &&theoretical2empirical[i]<fdr && minDiff==-1) 
		minDiff = key.intValue();
	    i++;
	}
	return minDiff;
    }

    public static double skellamPDF(int x, double k1, double k2) {
	//Calculate the pdf for the Skellam distribution
	return Math.exp(-(k1 + k2))*Math.pow(k1/k2, ((double)x)/2)*besselin(x, 2*Math.sqrt(k1*k2));
    }

    public static double besseli0(double x) {
	//use the implementation from NR
	double ax = Math.abs(x);
	if (ax<3.75) {
	    double y = x/3.75;
	    y *= y;
	    return 1 + y*(3.5156229 + y*(3.0899424 + y*(1.2067492 + y*(0.2659732 + y*(0.360768e-1 + y*0.45813e-2)))));
	}
	else {
	    double y = 3.75/ax;
	    return (Math.exp(ax)/Math.sqrt(ax))*(0.39894228 + y*(0.1328592e-1 + y*(0.225319e-2 + y*(-0.157565e-2 + y*(0.916281e-2 + y*(-0.2057706e-1 + y*(0.2635537e-1 + y*(-0.1647633e-1 + y*0.392377e-2))))))));
	}
    }

    public static double besseli1(double x) {
	double ax = Math.abs(x), a;
	if (ax<3.75) {
	    double y = x/3.75;
	    y *= y;
	    a = ax*(0.5 + y*(0.87890594 + y*(0.51498869 + y*(0.15084934 + y*(0.2658733e-1 + y*(0.301532e-2 + y*0.32411e-3))))));
	}
	else {
	    double y = 3.75/ax;
	    a = 0.2282967e-1 + y*(-0.2895312e-1 + y*(0.1787654e-1 - y*0.420059e-2));
	    a = 0.39894228 + y*(-0.3988024e-1 + y*(-0.362018e-2 + y*(0.163801e-2 + y*(-0.1031555e-1 + y*a))));
	    a *= (Math.exp(ax)/Math.sqrt(ax));
	}
	return x<0 ? -a : a;
    }
		
    public static double besselin(int n, double x) {
	//Modified bessel function of order n using the
	//implementation suggested in Numerical Recipies.
	n = Math.abs(n); //Symmetry
	double ACC = 40, BIGNO = 1e10, BIGNI = 1e-10;
	if (n==0)
	    return besseli0(x);
	else if (n==1)
	    return besseli1(x);
	double tox = 2/Math.abs(x), ans = 0, bip = 0, bi = 1;
	for (int j=2*(n + (int)Math.sqrt(ACC*n)); j>0; j--) {
	    double bim = bip + j*tox*bi;
	    bip = bi;
	    bi = bim;
	    if (Math.abs(bi) > BIGNO) { //Renormalize to prevent overflow
		ans *= BIGNI;
		bi *= BIGNI;
		bip *= BIGNI;
	    }
	    if (j==n)
		ans = bip;
	}
	ans *= besseli0(x)/bi; //Normalize
	//return x < 0 && (n & 1)>0 ? -ans : ans;
	return x < 0 && (n % 2)>0 ? -ans : ans;
    }

    public static double skellamCDF(int x, double k1, double k2) {
	double threshold = 1e-10;
	double diff = k1 - k2;
	int start = -10; //(int)(diff - 8*Math.abs(diff));
	double p = skellamPDF(start, k1, k2);
	//System.out.println("Calculating CDF for " + x + ", " + format.format(k1) + ", " + format.format(k2) + " with start = " + start + " and p = " + format.format(p));
	while (p>threshold)
	    p = skellamPDF(--start, k1, k2);
	while (start<x)
	    p += skellamPDF(++start, k1, k2);
	//System.out.println("s = " + start + ", p = " + p);
	return p;
    }

    //////////////////////////////////////////////////////////////////

    public boolean hasMoreEnrichedRegions() { return chrPos<lastReadPos; }
    
    public ChromosomeWindow findNextEnrichedRegion(int windowSize) {
	//use this method for the non-static case
	//Calculate the minimum number of reads which we must have in
	//a window if there are no input reads and only half of the
	//positions are mappable. This should provide an approximate
	//lower bound on the number of reads required in the window.
	if (chrPos<lastReadPos) {
	    ChromosomeWindow reg = new ChromosomeWindow(chr, chrPos, chrPos + windowSize); //Find the reads in the window
	    while (readsInd<reads.length && reads[readsInd].getPosition()<chrPos + windowSize) 
		reg.add(reads[readsInd++]);
	    //See how many input reads there are in this region
	    while (inputInd<inputReads.length && inputReads[inputInd].getPosition()<chrPos + windowSize) { reg.addInputRead(); inputInd++; }
	    //Calculate how many reads can be mapped to this region
	    //reg.setPotentialNumberOfReads(map.numberOfPotentialReads(chrPos, windowSize, strand==0));
	    reg.setLogPValue(reg.numberOfReads()>0 && reg.numberOfPotentialReads()>0 ? calculateLogPValue(reg.numberOfReads(), reg.numberOfPotentialReads(), getReads2SitesRatio()) : 0);
		reg.setLogInputPValue(reg.numberOfInputReads()>0 && reg.numberOfPotentialReads()>0 ? calculateLogPValue(reg.numberOfInputReads(), reg.numberOfPotentialReads(), getInputReads2SitesRatio()) : 0);	    
	    chrPos += windowSize;
	    return reg;
	}
	return null;
    }

    //////////////////////////////////////////////////////////////////

    /*public static double calculatePValue(ChromosomeWindow win, double reads2SitesRatio) { return calculatePValue(win, reads2SitesRatio, false); }

    public static double calculatePValue(ChromosomeWindow win, double reads2SitesRatio, boolean gcCompensate) {
	if (win.numberOfPotentialReads()>0) {
	    if (win.numberOfReads()>0)
		return calculatePValue((int)Math.round((double)win.numberOfReads()*(gcCompensate ? win.getGCMultiplier() : 1.0)), win.numberOfPotentialReads(), reads2SitesRatio);
	    return 1;// - math.exp(((double)win.numberOfPotentialReads())*reads2SitesRatio);
	}
	return 1;
    }

    public static double calculateInputPValue(ChromosomeWindow win, double reads2SitesRatio) { return calculateInputPValue(win, reads2SitesRatio, false); }

    public static double calculateInputPValue(ChromosomeWindow win, double reads2SitesRatio, boolean gcCompensate) {
	if (win.numberOfPotentialReads()>0) {
	    if (win.numberOfInputReads()>0)
		return calculatePValue((int)Math.round((double)win.numberOfInputReads()*(gcCompensate ? win.getGCMultiplier() : 1)), win.numberOfPotentialReads(), reads2SitesRatio);
	    return 1;// - math.exp(((double)win.numberOfPotentialReads())*reads2SitesRatio);
	}
	return 1;
	}*/

    public static double calculatePValue(int nReads, int potReads, double reads2SitesRatio) {
	//Use this method to calculate the p-value for observing a
	//given number of reads in a certain region where potReads can
	//be mapped. totReads is the total number of reads and
	//totSites is the total number of sites which a read can be
	//mapped to. The null-distribution is assumed to be Poisson
	//and the parameter is potReads*totReads/totSites.
	double lambda = ((double)potReads)*reads2SitesRatio;
	//return PoissonDist.barF(lambda, nReads);
	/*try {
	    return 1 - (new PoissonDistributionImpl(lambda)).cumulativeProbability(nReads);
	}
	catch (Exception e) { 
	    System.err.println("Error calculating P-value: " + e.toString());
	    System.err.println("With arguments nReads = " + nReads + ", potReads = " + potReads + " and reads2SitesRatio = " + reads2SitesRatio);
	    return 1;
	    }*/
	//Try calculating in log-scale instead to get better stability
	double sum = 1, prevTerm = 1;
	for (int i=1; i<=nReads; i++) { 
	    prevTerm *= (lambda/i);
	    sum += prevTerm;
	}
	//Use an approximation from the Taylor expansion, should be
	//accurate since we are dealing with very small numbers
	//return -lambda + Math.log(sum);
	return 1 - Math.exp(-lambda + Math.log(sum));
    }

    public static double calculateLogPValue(ChromosomeWindow win, double reads2SitesRatio) { return calculateLogPValue(win, reads2SitesRatio, false); }

    public static double calculateLogPValue(ChromosomeWindow win, double reads2SitesRatio, boolean gcCompensate) {
	return calculateLogPValue((int)Math.round((double)win.numberOfReads()*(gcCompensate ? win.getGCMultiplier() : 1.0)), win.numberOfPotentialReads(), reads2SitesRatio);
    }

    public static double calculateLogInputPValue(ChromosomeWindow win, double reads2SitesRatio) { return calculateLogInputPValue(win, reads2SitesRatio, false); }

    public static double calculateLogInputPValue(ChromosomeWindow win, double reads2SitesRatio, boolean gcCompensate) {
	return calculateLogPValue((int)Math.round((double)win.numberOfInputReads()*(gcCompensate ? win.getGCMultiplier() : 1)), win.numberOfPotentialReads(), reads2SitesRatio);
    }

    public static double calculateLogPValue(int nReads, int potReads, double reads2SitesRatio) {
	//if (nReads<30) {
	    double logPVal = Math.log(calculatePValue(nReads, potReads, reads2SitesRatio));
	    if (logPVal!=Double.NEGATIVE_INFINITY)
		return logPVal;
	    //}
	//Use the approximation that p-val goes as exp(-x)
	//asymptotically for the Poisson distribution.
	return -nReads;
    }


    public static double calculatePeakObservationProbability(double pValue, ChromosomeWindow win, double reads2SitesRatio) { return calculatePeakObservationProbability(pValue, win.numberOfReads(), win.numberOfPotentialReads(), reads2SitesRatio); }

    public static double calculatePeakObservationProbability(double pValue, int nReads, int potReads, double reads2SitesRatio) {
	//Calculate the number of reads required to observe this peak
	//(ie make the CDF drop below the input-modified
	//p-value). Assuming that the enrichment is given by the
	//number of reads observed, we can calculate the probability
	//of observing the peak using the Poisson model.
	double lambda = ((double)potReads)*reads2SitesRatio;
	int nRequired = 0;
	double cdf = 0;
	while (1-cdf>pValue) { cdf += (new PoissonDistributionImpl(lambda)).probability(nRequired++); }
	//Find out the prob of observing more than nRequired for a
	//Poisson distribution with parameter nReads.
	try {
	    double p = 1 - (new PoissonDistributionImpl(nReads)).cumulativeProbability(nRequired);
	    if (p==0)
		System.out.println("For peak the potential no reads is " + potReads + ", the required number of reads is " + nRequired + " and the number found is " + nReads);
	    return p;
	}
	catch (Exception e) { 
	    System.err.println("Error calculating P-value: " + e.toString());
	    System.err.println("With arguments nReads = " + nReads + ", nRequired = " + nRequired + " and reads2SitesRatio = " + reads2SitesRatio);
	    return 0;
	}
    }

    public static int calculateDetectionLimitDistance(ChromosomeWindow win, double pValue, double reads2SitesRatio) {
	//Find out how many reads need to be removed in order to get
	//down to the detection limit.
	double inputPVal = win.getLogInputPValue(), pVal = win.getLogPValue();
	int potReads = win.numberOfPotentialReads();
	int reads = win.numberOfReads();
	while (pVal-inputPVal<pValue) 
	    //remove a read
	    pVal = calculateLogPValue(--reads, potReads, reads2SitesRatio);
	//System.out.println("Detection limit is " + (win.numberOfReads() - reads) + " for site with " + win.numberOfInputReads() + " and " + potReads + " potential.\nInput p-val  = " + format.format(inputPVal) + ", p-val = " + format.format(pVal) + " and original " + format.format(win.getLogPValue()) + " at\n " + win.toString());
	return win.numberOfReads() - reads + 1;
    }
    
    //////////////////////////////////////////////////////////////////

    public static ABIRead[] findUniqueReads(ABIRead sortedReads[]) {
	Arrays.sort(sortedReads); //just to make sure
	int nDoubles = 0, j = 1;
	for (int i=1; i<sortedReads.length; i++)
	    nDoubles += sortedReads[i].equals(sortedReads[i-1]) ? 1 : 0;
	//System.out.println("Found " + nDoubles + " non-unique reads.");
	ABIRead reads[] = new ABIRead[sortedReads.length - nDoubles];
	reads[0] = sortedReads[0];
	for (int i=1; i<sortedReads.length; i++) {
	    if (!sortedReads[i].equals(sortedReads[i-1]))
		reads[j++] = sortedReads[i];
	    if (j>=reads.length) { break; }
	}
	return reads;
    }

    public static ABIRead[] sampleReads(ABIRead uniqueReads[], int nKeep) { return sampleReads(uniqueReads, nKeep, false); }

    public static ABIRead[] sampleReads(ABIRead uniqueReads[], int nKeep, boolean withReplacement) {
	ABIRead reads[] = new ABIRead[nKeep];
	if (withReplacement) {
	    for (int i=0; i<reads.length; i++) 
		reads[i] = uniqueReads[random.nextInt(uniqueReads.length)];
	}
	else {
	    //Create an array with the numbers 0...reads.length-1 in
	    //random order and then take the first nKeep of these
	    //elements.
	    int randVec[] = new int[uniqueReads.length];
	    for (int i=0; i<randVec.length; i++) { randVec[i] = i; }
	    for (int i=0; i<randVec.length*10; i++) { //swap two elements
		int ind1 = random.nextInt(randVec.length), ind2 = random.nextInt(randVec.length), tmp = randVec[ind1];
		randVec[ind1] = randVec[ind2];
		randVec[ind2] = tmp;
	    }
	    for (int i=0; i<reads.length; i++) { reads[i] = uniqueReads[randVec[i]]; }
	    //System.out.println("There were originally " + uniqueReads.length + " and now there are " + reads.length);
	}
	Arrays.sort(reads);
	return reads;
    }

    //////////////////////////////////////////////////////////////////

    public static void saveEnrichedRegions(Vector regs, String saveFileName) {
	//Use this method to save all the enriched regions extracted
	//from the genome.

    }

    public static void savePositionsReads(ABIRead[] reads, String posReadsFileName, int chr, boolean header, double normalizationFactor) { savePositionsReads(reads, posReadsFileName, chr, "mm", header, normalizationFactor, false); }

    public static void savePositionsReads(ABIRead[] reads, ABIRead[] inputReads, String posReadsFileName, int chr, boolean header, double normalizationFactor, double inputNormalizationFactor) { savePositionsReads(reads, inputReads, posReadsFileName, chr, header, normalizationFactor, inputNormalizationFactor, false); }

    public static void savePositionsReads(ABIRead[] reads, ABIRead[] inputReads, String posReadsFileName, int chr, boolean header, double normalizationFactor, double inputNormalizationFactor, boolean append) {
	System.out.println("Found " + reads.length + " reads with norm-factor " + format.format(normalizationFactor) + " and unit " + format.format(normalizationFactor - inputNormalizationFactor));
	System.out.println("Found " + inputReads.length + " reads with norm-factor " + format.format(inputNormalizationFactor));
	try {
	    Arrays.sort(reads);
	    Arrays.sort(inputReads);
	    //Use a sliding window to count the number of reads for
	    //each position
	    LinkedList slidingWindowReads = new LinkedList(), slidingWindowInputReads = new LinkedList();
	    //Initialize
	    for (int i=0; i<reads[0].getLength(); i++) {
		slidingWindowReads.addFirst(new Integer(0));
		slidingWindowInputReads.addFirst(new Integer(0));
	    }
	    int readStart = 0, j = 0, prevReads = Integer.MAX_VALUE;
	    int inputReadStart = 0, k = 0, prevInputReads = 0;
	    int slidingWindowStart = Math.min(reads[0].getPosition(), inputReads[0].getPosition());
	    Vector starts = new Vector(), ends = new Vector(), readCount = new Vector(), inputReadCount = new Vector();
	    starts.addElement(new Integer(slidingWindowStart));
	    //Write to a new file where the first column is the
	    //position, the second column is the number of reads
	    while (j<reads.length || k<inputReads.length) {
		readStart = j<reads.length ? reads[j].getPosition() : -Integer.MAX_VALUE;
		inputReadStart = k<inputReads.length ? inputReads[k].getPosition() : -Integer.MAX_VALUE;
		if (readStart==slidingWindowStart) { //Same pos as before
		    slidingWindowReads = addRead(slidingWindowReads, reads[j].getLength());
		}
		if (inputReadStart==slidingWindowStart) { //Same pos as before
		    slidingWindowInputReads = addRead(slidingWindowInputReads, reads[j].getLength());
		}
		if (readStart>slidingWindowStart && inputReadStart>slidingWindowStart) { //Move window forward
		    while (readStart>slidingWindowStart && inputReadStart>slidingWindowStart) {
			if (prevReads==Integer.MAX_VALUE)
			    prevReads = ((Integer)slidingWindowReads.getFirst()).intValue();
			if (prevInputReads==Integer.MAX_VALUE) //Initialization
			    prevInputReads = ((Integer)slidingWindowReads.getFirst()).intValue();
			//Remove the first entry of the window and save it
			if (((Integer)slidingWindowReads.getFirst()).intValue()!=prevReads || ((Integer)slidingWindowInputReads.getFirst()).intValue()!=prevInputReads) {
			    int r = ((Integer)slidingWindowReads.removeFirst()).intValue(), ir = ((Integer)slidingWindowInputReads.removeFirst()).intValue();
			    if (prevReads!=Integer.MAX_VALUE || prevInputReads!=Integer.MAX_VALUE) {
				ends.addElement(new Integer(slidingWindowStart-1));
				readCount.addElement(new Integer(prevReads));
				inputReadCount.addElement(new Integer(prevInputReads));
			    }
			    if (r!=Integer.MAX_VALUE || ir!=Integer.MAX_VALUE)
				starts.addElement(new Integer(slidingWindowStart));
			    prevReads = r;
			    prevInputReads = ir;
			}
			else {
			    slidingWindowReads.removeFirst();
			    slidingWindowInputReads.removeFirst();
			}
			slidingWindowStart++;
			//Add an empty element at the end to keep the size
			slidingWindowReads.addLast(new Integer(0));
			slidingWindowInputReads.addLast(new Integer(0));
		    }
		    if (readStart>slidingWindowStart)
			slidingWindowReads = addRead(slidingWindowReads, reads[0].getLength());
		    if (inputReadStart>slidingWindowStart)
			slidingWindowInputReads = addRead(slidingWindowInputReads, inputReads[0].getLength());
		}
		//else
		//  System.err.println("Error! found read starting earlier than window at " + reads[j].getPosition());
		if (j<reads.length && reads[j].getPosition()<=slidingWindowStart)
		    j++; 
		if (k<inputReads.length && inputReads[k].getPosition()<=slidingWindowStart)
		    k++;
		if (j>=reads.length || k>=inputReads.length)
		    break;
	    }
	    int lastWindowCount = slidingWindowStart;
	    while (!slidingWindowReads.isEmpty()) { //Write the last window
		if (((Integer)slidingWindowReads.getFirst()).intValue()!=prevReads) {
		    int r = ((Integer)slidingWindowReads.removeFirst()).intValue(), ir = ((Integer)slidingWindowInputReads.removeFirst()).intValue();
		    ends.addElement(new Integer(lastWindowCount-1));
		    readCount.addElement(new Integer(prevReads));
		    inputReadCount.addElement(new Integer(prevInputReads));
		    starts.addElement(new Integer(lastWindowCount));
		    prevReads = r;
		    prevInputReads = ir;
		}
		else {
		    slidingWindowReads.removeFirst();
		    slidingWindowInputReads.removeFirst();
		}
		lastWindowCount++;
	    }
	    ends.addElement(new Integer(lastWindowCount - 1));
	    readCount.addElement(new Integer(prevReads));
	    inputReadCount.addElement(new Integer(prevInputReads));
	    //System.out.println("Saving with " + readCount.size() + " to " + posReadsFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(posReadsFileName, append));
	    if (header) {
		if (chr==20) 
		    bw.write("browser position chrX:3000000-4000000");
		else if (chr==21)
		    bw.write("browser position chrY:3000000-4000000");
		else 
		    bw.write("browser position chr" + chr + ":3000000-4000000\n");
		//Header
		String colorStr = "", name = posReadsFileName.substring(posReadsFileName.lastIndexOf("/") + 1, posReadsFileName.lastIndexOf("."));
		colorStr = posReadsFileName.matches("KCl") ? "255,0,0" : "0,0,0";
		colorStr = posReadsFileName.matches("un") ? "0,255,0" : colorStr;
		//bw.write("track type=wiggle_0 name=\"" + name + "\" description=\"" + name + "\" visibility=hide db=mm9 autoScale=off viewLimits=0:2 color=\"" + colorStr + "\" maxHeightPixels=70:40:10 priority=" + chr + "\n");
		bw.write("track type=bedGraph name=\"" + name + "\" description=\"" + name + "\" visibility=hide db=mm9 autoScale=off viewLimits=0:2 color=\"" + colorStr + "\" maxHeightPixels=70:40:10 priority=" + chr + "\n");
	    }
	    for (int i=0; i<starts.size(); i++) {
		double height = -(normalizationFactor*(double)((Integer)readCount.elementAt(i)).intValue() - inputNormalizationFactor*(double)((Integer)inputReadCount.elementAt(i)).intValue());
		if (height!=0) {
		    if (chr==20)
			bw.write("chrX\t" + Math.min(CHROMOSOME_LENGTH[19], (((Integer)starts.elementAt(i)).intValue()-1)) + "\t" + Math.min(CHROMOSOME_LENGTH[19], (((Integer)ends.elementAt(i)).intValue())) + "\t" + format.format(height) + "\n");
		    else if (chr==21)
			bw.write("chrY\t" + Math.min(CHROMOSOME_LENGTH[20], (((Integer)starts.elementAt(i)).intValue()-1)) + "\t" + Math.min(CHROMOSOME_LENGTH[20], (((Integer)ends.elementAt(i)).intValue())) + "\t" + format.format(height) + "\n");
		    else 
			bw.write("chr" + chr + "\t" + Math.min(CHROMOSOME_LENGTH[chr-1], (((Integer)starts.elementAt(i)).intValue()-1)) + "\t" + Math.min(CHROMOSOME_LENGTH[chr-1], (((Integer)ends.elementAt(i)).intValue())) + "\t" + format.format(height) + "\n");
		}
	    }
	    bw.close();
	}
	catch(IOException e) {
	    System.err.println("Error saving to " + posReadsFileName + "\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
    }

    public static void savePositionsReads(ABIRead[] reads, String posReadsFileName, int chr, String species, boolean header, double normalizationFactor, boolean append) { savePositionsReads(reads, posReadsFileName, chr, species, header, normalizationFactor, append, false); }

    public static void savePositionsReads(ABIRead[] reads, String posReadsFileName, int chr, String species, boolean header, double normalizationFactor, boolean append, boolean neg) {
	int nConstReads = 0, nDouble = 0;
	try {
	    Arrays.sort(reads);
	    //Use a sliding window to count the number of reads for
	    //each position
	    LinkedList slidingWindowReads = new LinkedList();
	    //Initialize
	    for (int i=0; i<reads[0].getLength(); i++) 
		slidingWindowReads.addFirst(new Integer(0));
	    int readStart = 0, j = 0, prevReads = -1;
	    int slidingWindowStart = reads[0].getPosition();
	    Vector starts = new Vector(), ends = new Vector(), readCount = new Vector();
	    starts.addElement(new Integer(reads[0].getPosition()));
	    //Write to a new file where the first column is the
	    //position, the second column is the number of reads
	    while (j<reads.length) {
		readStart = reads[j].getPosition();
		if (readStart==slidingWindowStart) { //Same pos as before
		    //if (!unique || prevReads==-1) 
		    //System.out.println("Found duplicate reads at " + readStart);
		    slidingWindowReads = addRead(slidingWindowReads, reads[j].getLength());
		    //nDouble++;
		}
		else if (readStart>slidingWindowStart) { //Move window forward
		    while (readStart>slidingWindowStart) {
			if (prevReads==-1) //Initialization
			    prevReads = ((Integer)slidingWindowReads.getFirst()).intValue();
			//Remove the first entry of the window and save it
			if (((Integer)slidingWindowReads.getFirst()).intValue()!=prevReads) {
			    int r = ((Integer)slidingWindowReads.removeFirst()).intValue();
			    if (prevReads>0) {
				ends.addElement(new Integer(slidingWindowStart-1));
				readCount.addElement(new Integer(prevReads));
				nConstReads++;
			    }
			    if (r>0)
				starts.addElement(new Integer(slidingWindowStart));
			    prevReads = r;
			}
			else 
			    slidingWindowReads.removeFirst();
			slidingWindowStart++;
			//Add an empty element at the end to keep the size
			slidingWindowReads.addLast(new Integer(0));
		    }
		    slidingWindowReads = addRead(slidingWindowReads, reads[0].getLength());
		}
		else
		    System.err.println("Error! found read starting earlier than window at " + reads[j].getPosition());
		j++;
	    }
	    int lastWindowCount = slidingWindowStart;
	    while (!slidingWindowReads.isEmpty()) { //Write the last window
		if (((Integer)slidingWindowReads.getFirst()).intValue()!=prevReads) {
		    int r = ((Integer)slidingWindowReads.removeFirst()).intValue();
		    if (prevReads>0) {
			ends.addElement(new Integer(lastWindowCount-1));
			readCount.addElement(new Integer(prevReads));
			nConstReads++;
		    }
		    if (r>0)
			starts.addElement(new Integer(lastWindowCount));
		    prevReads = r;
		}
		else
		    slidingWindowReads.removeFirst();
		lastWindowCount++;
	    }
	    ends.addElement(new Integer(lastWindowCount - 1));
	    readCount.addElement(new Integer(prevReads));
	    nConstReads++;
	    //System.out.println("Saving with " + readCount.size() + " to " + posReadsFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(posReadsFileName, append));
	    if (header) {
		if (chr==20) 
		    bw.write("browser position chrX:3000000-4000000");
		else if (chr==21)
		    bw.write("browser position chrY:3000000-4000000");
		else 
		    bw.write("browser position chr" + chr + ":3000000-4000000\n");
		//Header
		String colorStr = "", name = posReadsFileName.substring(posReadsFileName.lastIndexOf("/") + 1, posReadsFileName.lastIndexOf("."));
		colorStr = posReadsFileName.matches("KCl") ? "255,0,0" : "0,0,0";
		colorStr = posReadsFileName.matches("un") ? "0,255,0" : colorStr;
		//bw.write("track type=wiggle_0 name=\"" + name + "\" description=\"" + name + "\" visibility=hide db=mm9 autoScale=off viewLimits=0:2 color=\"" + colorStr + "\" maxHeightPixels=70:40:10 priority=" + chr + "\n");
		bw.write("track type=bedGraph name=\"" + name + "\" description=\"" + name + "\" visibility=hide db=mm9 autoScale=off viewLimits=0:2 color=\"" + colorStr + "\" maxHeightPixels=70:40:10 priority=" + chr + "\n");
	    }
	    int chrLen[] = species.equals("hs") ? RefSeqGenes.CHROMOSOME_LENGTH_HS : RefSeqGenes.CHROMOSOME_LENGTH;
	    for (int i=0; i<starts.size(); i++) {
		if ((species.equals("mm") && chr==20) || (species.equals("hs") && chr==23))
		    bw.write("chrX\t" + Math.min(chrLen[chr-1], (((Integer)starts.elementAt(i)).intValue()-1)) + "\t" + Math.min(chrLen[chr-1], (((Integer)ends.elementAt(i)).intValue())) + "\t" + (neg ? "-" : "") + format.format(normalizationFactor*(double)((Integer)readCount.elementAt(i)).intValue()) + "\n");
		else if ((species.equals("mm") && chr==21) || (species.equals("hs") && chr==24))
		    bw.write("chrY\t" + Math.min(chrLen[chr-1], (((Integer)starts.elementAt(i)).intValue()-1)) + "\t" + Math.min(chrLen[chr-1], (((Integer)ends.elementAt(i)).intValue())) + "\t" + (neg ? "-" : "") + format.format(normalizationFactor*(double)((Integer)readCount.elementAt(i)).intValue()) + "\n");
		else 
		    bw.write("chr" + chr + "\t" + Math.min(chrLen[chr-1], (((Integer)starts.elementAt(i)).intValue()-1)) + "\t" + Math.min(chrLen[chr-1], (((Integer)ends.elementAt(i)).intValue())) + "\t" + (neg ? "-" : "") + format.format(normalizationFactor*(double)((Integer)readCount.elementAt(i)).intValue()) + "\n");
	    }
	    bw.close();
	}
	catch(IOException e) {
	    System.err.println("Error saving to " + posReadsFileName + "\n" + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
    }

    protected static LinkedList addRead(LinkedList l, int fragmentLength) {
	for (int i=0; i<fragmentLength; i++)
	    l.set(i, new Integer(((Integer)l.get(i)).intValue() + 1));
	return l;
    }

    protected static LinkedList addRead(LinkedList l, int fragmentLength, boolean input) {
	for (int i=0; i<fragmentLength; i++)
	    l.set(i, new Integer(((Integer)l.get(i)).intValue() + (input ? -1 : 1)));
	return l;
    }

    /*public static TreeMap getReadsHistogram(int binSize, ABIRead r[], int chr, boolean orchestra) {
	Mappability mappability = new Mappability();
	for (int i=0; i<r.length; i++) {
	    
	}
	}*/

    public static String getProcessedReadsFileName(String fileName, int chr, boolean extend, int fragmentLength) { return fileName + "_reads_" + chr + "_" + (extend ? "E" : "") + fragmentLength + ".fa"; }

    public void saveReadsAndSequenceData(String saveFileName) {
	//Use this method to parse all the reads, sort them, extract
	//the sequence info and save to a file. The file has fasta format:
	//
	//>pos\tstrand\n
	//sequence\n
	//
	//The position gives the 5' end for positive reads and 3' end
	//for negative reads and the sequence is always with respect
	//to the positive strand.
	
	FastaSequence2 fastaSeq = new FastaSequence2(GENOME_LOC + AnalyzeControls.getChromosomeString(chr));
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    bw.write(">" + reads[0].getPosition() + (reads[0].getStrand() ? "+" : "-") + "\n");	    
	    reads[0].setSequence(fastaSeq.getNextSequence(reads[0].getPosition(), reads[0].getLength()));
	    bw.write(reads[0].getSequence() + "\n");
	    for(int i=1; i<reads.length; i++) {
		bw.write(">" + reads[i].getPosition() + (reads[i].getStrand() ? "+" : "-") + "\n");	    
		int overlap = reads[i-1].getLength() - (reads[i].getPosition() - reads[i-1].getPosition());
		if (overlap<=0) //No overlap
		    reads[i].setSequence(fastaSeq.getNextSequence(reads[i].getPosition(), reads[i].getLength()));
		else { //Take the part which has already been read from
		    //the previous read
		    reads[i].setSequence(reads[i-1].getSequence().substring(reads[i-1].getLength() - overlap) + fastaSeq.getNextSequence(reads[i].getPosition() + overlap, reads[i].getLength() - overlap));
		}
		bw.write(reads[i].getSequence() + "\n");
	    }
	}
	catch (IOException e) {
	    
	}
    }

    public static int[][] parseJunkFile() {
	//Use this method to find the regions which appear to contain
	//mostly junk
	int regions[][] = null;
	try {
	    BufferedReader br = new BufferedReader(new FileReader(JUNK_FILE_NAME));
	}
	catch (IOException e) {
	    System.err.println("Error parsing junk file.");
	}
	return regions;
    }
    
}

