//Use this class for finding peaks. All methods in the class are
//static and can thus be called by outside classes. The basic
//functions here are:
//
//parseFiles - load the files with the reads and parse the contents
//
//findEnrichedRegions - find those peaks which have more than expected reads
//
//Compared with version 2, this version picks the peaks which seem the
//most promising ones, extends a fixed window and compares with the input.

import java.util.*;
import java.util.regex.*;
import java.io.*;
import org.apache.commons.math.distribution.PoissonDistributionImpl;
import org.apache.commons.math.MathException;
//import umontreal.iro.lecuyer.probdist.*;
import java.text.DecimalFormat;

public class PeakFinder3 extends PeakFinder2 {

    protected static Random random = new Random();
    public static DecimalFormat format = new DecimalFormat("###.##########");

    public static Pattern posReadPattern = Pattern.compile("chr([\\d|X|Y]+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
    //The groups are bead identifier, chr, strand, pos, mismatches
    public static Pattern rawPattern1 = Pattern.compile("([0-9]+_[0-9]+_[0-9]+_[F|R][0-9]+),([0-9|X|Y]+)_(-*?)([0-9]+)\\.([0-3])");
    //Use this pattern when we parse files which were randomly
    //generated. They typically have a shorter header and no barcode
    //info. The first group is only there to make sure that the
    //numbers for the other groups will match with rawPattern1.
    public static Pattern randPattern1 = Pattern.compile("(>),?([0-9|X|Y]+)_(-*?)([0-9]+)[\\.([0-3])]*?");
    //The second line contains information about the output
    public static Pattern rawPattern2 = Pattern.compile("T([0-3]+)");
    //For the files containing the codes
    public static Pattern codePattern = Pattern.compile("([0-3]+)\t([0-9]+)");

    public static final int[] CHROMOSOME_LENGTH = new int[]{197195432, 181748087, 159599783, 155630120, 152537259, 149517037, 152524553, 131738871, 124076172, 129993255, 121843856, 121257530, 120284312, 125194864, 103494974, 98319150, 95272651, 90772031, 61342430, 166650296, 15902555};
    public static final long TOTAL_CHROMOSOME_LENGTH = (long)(265489521.8*10);

    protected Mappability map;
    protected int chrPos, lastReadPos, readsInd = 0, inputInd = 0, chr, strand;
    protected ABIRead[] reads, inputReads;

    //////////////////////////////////////////////////////////////////

    public PeakFinder3() { }

    public PeakFinder3(boolean orchestra, int c, String fileName, String inputFileName, int fragmentLength, boolean extend, int str) {
	String GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	chr = c;
	strand = str;
	map = new Mappability(GENOME_LOC, chr);
	chrPos = 0;
	lastReadPos = map.getNumberOfBases();
	reads = parseFiles(chr, fileName, fragmentLength, extend, strand);
	inputReads = parseFiles(chr, inputFileName, fragmentLength, extend, strand);
    }

    public PeakFinder3(boolean orchestra, int c, ABIRead r[], ABIRead input[]) {
	//Use this when we want to avoid parsing the file multiple times. 
	String GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	chr = c;
	map = new Mappability(GENOME_LOC, chr);
	chrPos = 0;
	lastReadPos = map.getNumberOfBases();
	reads = r;
	inputReads = input;
    }

    //////////////////////////////////////////////////////////////////

    public double getReads2SitesRatio() { return ((double)reads.length)/map.getNumberOfUniqueBases(); }

    public double getInputReads2SitesRatio() { return ((double)inputReads.length)/map.getNumberOfUniqueBases(); }

    public static ABIRead[] parseFiles(int chr, String fileName, int fragmentLength, boolean extend, int strand) {
	return parseFiles(chr, fileName, 3, true, -1, fragmentLength, extend, strand);
    }

    public static ABIRead[] parseFiles(int chr, String fileName, int nMismatches, boolean unique, int nSample, int fragmentLength, boolean extend, int strand) {
	try {
	    int nReadsForward = 0, nReadsReverse = 0;
	    BufferedReader br = new BufferedReader(new FileReader(fileName + "_" + chr));
	    String line = br.readLine();
	    //Go over once just to count the number of reads for this
	    //chromosome
	    while (line!=null) {
		Matcher m = rawPattern1.matcher(line);
		if (m.find()) {
		    boolean forward = !m.group(3).equals("-");
		    int nMisses = Integer.parseInt(m.group(5));
		    if (nMisses<=nMismatches && Integer.parseInt(m.group(2))==chr) {
			if (forward) { nReadsForward++; }
			else { nReadsReverse++; }
		    }
		    line = br.readLine();
		}
		line = br.readLine(); //Get the sequence information
	    }
	    br.close();
	    //Go through again and store them. We want to use an
	    //ordinary array to save memory, going through the file is
	    //not that slow.
	    if (strand<0)
		return parseFile(chr, fileName, nMismatches, nReadsReverse, unique, nSample, fragmentLength, extend, strand);
	    else if (strand>0)
		return parseFile(chr, fileName, nMismatches, nReadsForward, unique, nSample, fragmentLength, extend, strand);
	    else
		return parseFile(chr, fileName, nMismatches, nReadsForward + nReadsReverse, unique, nSample, fragmentLength, extend, strand);
	}
	catch (IOException e) {
	    System.err.println("Error, could not open file: " + fileName);
	}
	return null;
    }

    protected static ABIRead[] parseFile(int chr, String fileName, int nMismatches, int nReads, boolean unique, int nSample, int fragmentLength, boolean extend, int strand) {
	//Use an int to determine which strand to return. Negative
	//means reverse strand, positive forward strand and zero means
	//both.
	ABIRead reads[] = new ABIRead[nReads];
	try {
	    int i = 0;
	    BufferedReader br = new BufferedReader(new FileReader(fileName + "_" + chr));
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = PeakFinder.rawPattern1.matcher(line);
		if (m.find()) {
		    boolean forward = !m.group(3).equals("-");
		    int nMisses = Integer.parseInt(m.group(5)), pos = Integer.parseInt(m.group(4));
		    if (Integer.parseInt(m.group(2))==chr && nMisses<=nMismatches) {
			if (forward && strand>=0)  //Use 1-based index when saving
			    reads[i++] = new ABIRead(chr, true, pos + 1 + (extend ? 0 : fragmentLength/2), extend ? fragmentLength : AnalyzeControls.READ_LENGTH);
			else if (!forward && strand<=0)
			    reads[i++] = new ABIRead(chr, false, pos + 1 - (extend ? fragmentLength : fragmentLength/2) - AnalyzeControls.READ_LENGTH, extend ? fragmentLength : AnalyzeControls.READ_LENGTH);
			line = br.readLine();
		    }
		    else //Skip the next line
			line = br.readLine();
		}
		line = br.readLine();
	    }
	    br.close();
	    if (reads.length>0) {
		Arrays.sort(reads);
		if (nSample<reads.length && nSample>0)
		    reads = sampleReads(reads, nSample);
		if (unique) { reads = findUniqueReads(reads); }
		Arrays.sort(reads);
	    }
	    else
		System.err.println("Warning: No reads found!");
	}
	catch (IOException e) {
	    
	}
	return reads;
    }

    public static int numberOfReads(int chr, String fileName, int nMismatches, int strand) {
	try {
	    int nReadsForward = 0, nReadsReverse = 0;
	    BufferedReader br = new BufferedReader(new FileReader(fileName + "_" + chr));
	    String line = br.readLine();
	    //Go over once just to count the number of reads for this
	    //chromosome
	    while (line!=null) {
		Matcher m = rawPattern1.matcher(line);
		if (m.find()) {
		    boolean forward = !m.group(3).equals("-");
		    int nMisses = Integer.parseInt(m.group(5));
		    if (nMisses<=nMismatches && Integer.parseInt(m.group(2))==chr) {
			if (forward) { nReadsForward++; }
			else { nReadsReverse++; }
		    }
		    line = br.readLine();
		}
		line = br.readLine(); //Get the sequence information
	    }
	    br.close();
	    //Go through again and store them. We want to use an
	    //ordinary array to save memory, going through the file is
	    //not that slow.
	    if (strand<0)
		return nReadsReverse;
	    else if (strand>0)
		return nReadsForward;
	    else
		return nReadsForward + nReadsReverse;
	}
	catch (IOException e) {
	    System.err.println("Error, could not open file: " + fileName);
	}
	return -1;
    }

    public static int numberOfReads(String fileName, int nMismatches, int strand) {
	int nReads = 0;
	for (int i=0; i<21; i++)
	    nReads += numberOfReads(i + 1, fileName, nMismatches, strand);
	return nReads;
    }


    //////////////////////////////////////////////////////////////////
    //Use a window of size 2*fragmentLength to scan through the genome
    //and pick out the ones which have significantly more reads than
    //the input. In particular, we will be taking the mappability of
    //the genome into consideration.

    public static TreeSet findEnrichedRegions(String fileName, String inputFileName, int fragmentLength, int windowSize, boolean all, boolean extend, int strand, double pValue, int chr, boolean orchestra) {
	return findEnrichedRegions(parseFiles(chr, fileName, fragmentLength, extend, strand), parseFiles(chr, inputFileName, fragmentLength, extend, strand), windowSize, all, pValue, chr, strand, orchestra);
    }

    public static TreeSet findEnrichedRegions(ABIRead reads[], ABIRead inputReads[], int windowSize, double pValue, int chr, int strand, boolean orchestra) {
	return findEnrichedRegions(reads, inputReads, windowSize, false, pValue, chr, strand, orchestra);
    }

    public static TreeSet findEnrichedRegionsFixedWidth(ABIRead reads[], ABIRead inputReads[], int windowSize, boolean all, double pValue, int chr, int strand, boolean orchestra) {
	//Instead of using the method by Johnson et al (2007) where we
	//zoom in on the clusters with large number of reads, we scan
	//through the genome with a fixed-size window. This should be
	//better in the sense that it does not introduce any bias in
	//the selection of regions. Moreover, scanning the genome is
	//not going to be that slow since we can automatically skip
	//regions with few basepairs.
	String GENOME_LOC = orchestra ? "/home/mh168/Genomes/Mm/v9/" : "/media/exthd/Genomes/Mm/v9/";
	Mappability mappability = new Mappability(GENOME_LOC, chr);
	//Calculate the parameters for the Poisson distribution for
	//the reads and the input
	double reads2SitesRatio = ((double)reads.length)/mappability.getNumberOfUniqueBases(), inputReads2SitesRatio = ((double)inputReads.length)/mappability.getNumberOfUniqueBases();
	//Calculate the minimum number of reads which we must have in
	//a window if there are no input reads and only half of the
	//positions are mappable. This should provide an approximate
	//lower bound on the number of reads required in the window.
	int minReads = (int)(inputReads2SitesRatio*windowSize) + 1, j = 0;
	ChromosomeWindow win = new ChromosomeWindow(reads[0], chr);

	TreeSet regs = new TreeSet();
	for (int i=0; i<reads.length; i++) {
	    if (reads[i].getPosition() - reads[i-1].getPosition()<reads[i].getLength()) 
		win.add(reads[i]);
	    else { //Finished adding reads to the window
		win.trimWindow(windowSize);
		if (win.numberOfReads()>minReads) {
		    //Set the input reads
		    while (j<inputReads.length && inputReads[j].getPosition()<win.getEnd()) {
			if (inputReads[j].getPosition()>win.getStart()) { win.addInputRead(); j++; }
		    }
		    //Calculate how many reads can be mapped to this region
		    win.setPotentialNumberOfReads(mappability.numberOfPotentialReads(win.getStart(), win.getLength(), strand==0));
		    if (win.numberOfPotentialReads()>0) {
			//Calculate the p-value
			win.setPValue(calculatePValue(win, reads2SitesRatio));
			win.setInputPValue(calculateInputPValue(win, reads2SitesRatio));
			if (win.getPValueRatio()<pValue)
			    regs.add(win);
		    }
		}
		win = new ChromosomeWindow(reads[i], chr);
	    }
	}
	return regs;
    }

    //////////////////////////////////////////////////////////////////

    
}

