//Represent the reduced model from the Haar wavelets. Can be used both
//for ChIP and RNA seq.

import java.io.*;
import java.util.*;
import java.text.DecimalFormat;
import edu.uci.ics.jung.graph.impl.*;

public class ReducedModel {

    protected TreeSet<TranscribedRegion> segments = new TreeSet<TranscribedRegion>();
    //Keep track of the transcribed regions and how they overlap annotated genes
    protected UndirectedSparseGraph genesTranscriptsGraph = new UndirectedSparseGraph();

    protected String species;
    protected int chr, nReads;
    private boolean forward, chip;

    public static DecimalFormat format = new DecimalFormat("###.#####");

    public ReducedModel(ABIRead rs[], Vector<Integer> breakPoints, int c, int maxReadLength, int lVec[], double cutOffDens[], boolean mergeSegments, boolean log, String sp, boolean ch) {
	//Use this constructor for the FDR approach to keeping
	//transcripts. The idea is to only keep those transcripts
	//whose density is relevant for a given fdr
	//threshold. Calculate densities for several length scales and
	//use the one which is closest to the length. 
	chr = c;
	chip = ch;
	species = sp;
	createSegments(rs, null, breakPoints, maxReadLength, lVec, cutOffDens, mergeSegments, log);
    }

    public ReducedModel(ABIRead rs[], ABIRead irs[], Vector<Integer> breakPoints, int c, int maxReadLength, int lVec[], double cutOffDens[], boolean mergeSegments, boolean log, String sp, boolean ch) {
	//Use this constructor for the FDR approach to keeping
	//transcripts. The idea is to only keep those transcripts
	//whose density is relevant for a given fdr
	//threshold. Calculate densities for several length scales and
	//use the one which is closest to the length. 
	chr = c;
	chip = ch;
	species = sp;
	createSegments(rs, irs, breakPoints, maxReadLength, lVec, cutOffDens, mergeSegments, log);
    }

    //////////////////////////////////////////////////////////////////

    public void createSegments(ABIRead[] reads, ABIRead[] inputReads, Vector<Integer> breakPoints, int maxReadLength, int[] lVec, double[] cutOffDens, boolean mergeSegments, boolean log) {
	//Start with one segment, the one which reduces the error the most
	TranscribedRegion seg = null;//new TranscribedRegion(breakPoints.elementAt(0), breakPoints.elementAt(1), reads, inputReads, chip, chr, maxReadLength, log);
	if (densityCutOff(seg.getLevel(), seg.getLength(), lVec, cutOffDens)) {
	    segments.add(seg);
	    //System.out.println("Added segment " + seg.toString());
	}
	nReads += seg.numberOfReads();
	for (int i=1; i<breakPoints.size()-1; i++) {
	    //seg = new TranscribedRegion(breakPoints.elementAt(i), breakPoints.elementAt(i+1), reads, inputReads, chip, chr, maxReadLength, log);
	    if (densityCutOff(seg.getLevel(), seg.getLength(), lVec, cutOffDens)) {
		segments.add(seg);
		//System.out.println("Added segment " + seg.toString());
	    }
	    nReads += seg.numberOfReads();
	}
	if (mergeSegments) { //Merge the segments which are adjacent
	    TreeSet<TranscribedRegion> mergedSegs = new TreeSet<TranscribedRegion>(), removedSegs = new TreeSet<TranscribedRegion>(), segs2merge = new TreeSet<TranscribedRegion>();
	    Iterator<TranscribedRegion> it = segments.iterator();
	    if (it.hasNext()) {
		TranscribedRegion prevSeg = it.next();
		while (it.hasNext()) {
		    TranscribedRegion currSeg = it.next();
		    if (prevSeg.getEnd()==currSeg.getStart()) {
			segs2merge.add(currSeg);
			removedSegs.add(prevSeg);
			removedSegs.add(currSeg);
		    }
		    else {
			if (segs2merge.size()>0)
			    mergedSegs.add(new TranscribedRegion(segs2merge, reads, inputReads, maxReadLength, log));
			segs2merge = new TreeSet<TranscribedRegion>();
		    }
		    prevSeg = currSeg;
		}
	    }
	    segments.removeAll(removedSegs);
	    segments.addAll(mergedSegs);
	}
    }

    //////////////////////////////////////////////////////////////////

    public boolean densityCutOff(double dens, int len, int lVec[], double cutOffs[]) {
	int i = 0;
	while (i<lVec.length && Math.pow(2, lVec[i])<len)
	    i++;
	return dens>cutOffs[i==lVec.length ? i-1 : i];
    }

    public TranscribedRegion findTranscribedRegionEnding(TranscribedRegion segToAdd) {
	TreeSet<TranscribedRegion> headSet = new TreeSet<TranscribedRegion>(segments.headSet(segToAdd));
	if (headSet!=null && headSet.size()>0)
	    return headSet.last().getEnd()==segToAdd.getStart() ? headSet.last() : null;
	return null;
    }

    public TranscribedRegion findTranscribedRegionStarting(TranscribedRegion segToAdd) {
	TreeSet<TranscribedRegion> tailSet = new TreeSet<TranscribedRegion>(segments.tailSet(segToAdd));
	if (tailSet!=null && tailSet.size()>0)
	    return tailSet.first().getStart()==segToAdd.getEnd() ? tailSet.first() : null;
	return null;
    }

    public int numberOfTranscribedRegions() { return segments.size(); }

    public double meanTranscribedRegionLength() { return meanTranscribedRegionLength(true); }

    public double meanTranscribedRegionLength(boolean log) {
	double len = 0;
	for (TranscribedRegion reg : segments)
	    len += log ? Math.log(reg.getLength()) : reg.getLength();
	return len/segments.size();
    }

    public double proportionTranscribed() { 
	double len = 0;
	for (TranscribedRegion reg : segments)
	    len += reg.getLength();
	return len/TranscriptFinder.numberOfBasePairs(chr, species);
    }

    public double numberOfReads() {
	//returns the number of reads in transcribed regions
	int n = 0;
	for (TranscribedRegion tr : segments)
	    n += tr.numberOfReads();
	return n;
    }

    public double proportionOfReads() { return (double)numberOfReads()/nReads; }

    public double minTranscribedRegionDensity() {
	double m = 100000;
	for (TranscribedRegion t : segments) 
	    m = t.getLevel()>0 ? Math.min(m, t.getLevel()) : m;
	return m;
    }

    ///////////////////////////////////////////////////////

    public void addGenes(RefSeqGenes geneInfo, boolean forward, int geneIDsKeep[]) {
	for (RefSeqGene g : geneInfo.getGenes()) {
	    if (g.getStrand()==forward && Arrays.binarySearch(geneIDsKeep, g.getID())>-1) {
		HashSet<UndirectedSparseGraph> verts = new HashSet<UndirectedSparseGraph>(genesTranscriptsGraph.getVertices());
		if (!verts.contains(g))
		    genesTranscriptsGraph.addVertex(g);
	    }
	}
    }

    public void removeGenes() { genesTranscriptsGraph.removeAllVertices(); }

    public void compare2Annotation(RefSeqGenes geneInfo, boolean forward, int geneIDsKeep[]) {
	//Build a graphical model between overlapping genes and
	//transcripts to determine how well we have managed to map one
	//transcript per gene.
	if (genesTranscriptsGraph!=null) {
	    genesTranscriptsGraph.removeAllEdges();
	    //genesTranscriptsGraph.removeAllVertices();
	}
	//genesTranscriptsGraph = new UndirectedSparseGraph();
	if (genesTranscriptsGraph.numVertices()==0) {
	    for (TranscribedRegion t : segments) 
		genesTranscriptsGraph.addVertex(t);
	}
	for (RefSeqGene g : geneInfo.getGenes()) {
	    if (g.getStrand()==forward && Arrays.binarySearch(geneIDsKeep, g.getID())>-1) {
		HashSet<UndirectedSparseGraph> verts = new HashSet<UndirectedSparseGraph>(genesTranscriptsGraph.getVertices());
		if (!verts.contains(g))
		    genesTranscriptsGraph.addVertex(g);
		Iterator<TranscribedRegion> it = segments.iterator();
		if (it.hasNext()) {
		    TranscribedRegion t = it.next();
		    while (it.hasNext() && !(t.getStart()>g.getStart() && t.getStart()>g.getEnd())) {
			if (g.overlaps(t.getStart(), t.getEnd())) {
			    //System.out.println(g.toString() + " is overlapped by " + t.toString());
			    genesTranscriptsGraph.addEdge(new UndirectedSparseEdge(g, t));
			}
			t = it.next();
		    }
		}
	    }
	}
	//System.out.println("There are " + geneIDsKeep.length + " genes, " + segments.size() + " transcribed regions and " + genesTranscriptsGraph.numEdges() + " edges.");
	//Check the out-degree of the transcribed regions with respect
	//to the annotated genes.
	int degrees[][] = new int[2][10];
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(genesTranscriptsGraph.getVertices())) {
	    if (v instanceof RefSeqGene)
		degrees[0][Math.min(v.outDegree(), degrees[0].length-1)]++;
	    else
		degrees[1][Math.min(v.outDegree(), degrees[1].length-1)]++;
	}
	//return degrees;
    }

    public int numberOfGenesWithOneTranscript() { return numberOfGenesWithOneTranscript(true); }

    public int numberOfGenesWithOneTranscript(boolean unique) { return numberOfGenesWithNTranscripts(1, false, unique); }

    public int numberOfGenesWithNTranscripts(int n, boolean ineq, boolean unique) {
	//Find the number of genes which are overlapped by n
	//transcripts. If the boolean is set, then we require that the
	//matching transcripts all have out-degree 1.
	int m = 0;
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(genesTranscriptsGraph.getVertices())) {
	    if (v instanceof RefSeqGene) {
		if ((!ineq && v.inDegree()==n) || (ineq && v.inDegree()>n)) {
		    if (!unique)
			m++;
		    else {
			Iterator<UndirectedSparseVertex> it = v.getNeighbors().iterator();
			boolean uniqueTranscripts = true;
			while (it.hasNext()) 
			    uniqueTranscripts = uniqueTranscripts && it.next().numNeighbors()==1;
			m += uniqueTranscripts ? 1 : 0;
		    }
		}
	    }
	}
	return m;
    }

    public int numberOfTranscriptsWithOneGene() { return numberOfTranscriptsWithOneGene(true); }

    public int numberOfTranscriptsWithOneGene(boolean unique) { return numberOfTranscriptsWithNGenes(1, false, unique); }

    public int numberOfTranscriptsWithNGenes(int n, boolean ineq, boolean unique) {
	//Find the number of transcripts which are overlapped by n
	//genes. If the boolean is set, then we require that the
	//matching genes all have out-degree 1. If ineq is true, we
	//are interested in the case >n.
	int m = 0;
	for (UndirectedSparseVertex v : new Vector<UndirectedSparseVertex>(genesTranscriptsGraph.getVertices())) {
	    if (v instanceof TranscribedRegion) {
		if ((!ineq && v.inDegree()==n) || (ineq && v.inDegree()>n)) {
		    if (!unique)
			m++;
		    else {
			Iterator<UndirectedSparseVertex> it = v.getNeighbors().iterator();
			boolean uniqueTranscripts = true;
			while (it.hasNext()) 
			    uniqueTranscripts = uniqueTranscripts && it.next().numNeighbors()==1;
			m += uniqueTranscripts ? 1 : 0;
		    }
		}
	    }
	}
	return m;
    }

    ///////////////////////////////////////////////////////

    public TreeMap<Integer, Vector<RefSeqGene>> findTranscriptsPerGene(RefSeqGenes geneInfo, boolean forward) { return findTranscriptsPerGene(geneInfo, forward, null); }

    public TreeMap<Integer, Vector<RefSeqGene>> findTranscriptsPerGene(RefSeqGenes geneInfo, boolean forward, int geneIDsKeep[]) {
	TreeMap<Integer, Vector<RefSeqGene>> transcriptsPerGene = new TreeMap<Integer, Vector<RefSeqGene>>();
	//What proportion of genes are overlapped by a single transcript
	Iterator<TranscribedRegion> it = segments.iterator();
	TranscribedRegion t = it.next();
	for (RefSeqGene g : geneInfo.getGenes()) {
	    if (g.getStrand()==forward && (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>-1)) {
		TreeSet<TranscribedRegion> regionsOverlappingGene = new TreeSet<TranscribedRegion>();
		while (it.hasNext() && !(t.getStart()>g.getStart() && t.getStart()>g.getEnd())) {
		    if (g.overlaps(t.getStart(), t.getEnd())) {
			//System.out.println(g.toString() + " is overlapped by " + t.toString());
			regionsOverlappingGene.add(t);
		    }
		    t = it.next();
		}
		if (!transcriptsPerGene.containsKey(regionsOverlappingGene.size()))
		    transcriptsPerGene.put(regionsOverlappingGene.size(), new Vector<RefSeqGene>());
		transcriptsPerGene.get(regionsOverlappingGene.size()).addElement(g);
	    }
	}
	return transcriptsPerGene;
    }

    public TreeMap<Integer, Vector<TranscribedRegion>> findGenesPerTranscript(RefSeqGenes geneInfo, boolean forward) { return findGenesPerTranscript(geneInfo, forward); }

    public TreeMap<Integer, Vector<TranscribedRegion>> findGenesPerTranscript(RefSeqGenes geneInfo, boolean forward, int geneIDsKeep[]) {
	TreeMap<Integer, Vector<TranscribedRegion>> genesPerTranscript = new TreeMap<Integer, Vector<TranscribedRegion>>();
	//Find out what proportion of transcripts overlap known genes
	Iterator<RefSeqGene> itGene = geneInfo.getGenes().iterator();
	RefSeqGene g = itGene.next();
	for (TranscribedRegion tr : segments) {
	    TreeSet<RefSeqGene> genesOverlappingRegion = new TreeSet<RefSeqGene>();
	    while (itGene.hasNext() && !(g.getEnd()>tr.getEnd() && g.getStart()>tr.getEnd())) {
		if (g.getStrand()==forward && (geneIDsKeep==null || Arrays.binarySearch(geneIDsKeep, g.getID())>-1)) {
		    if (g.overlaps(tr.getStart(), tr.getEnd())) {
			genesOverlappingRegion.add(g);
			//System.out.println(tr.toString() + " is overlapped by " + g.toString());
		    }
		}
		g = itGene.next();
	    }
	    if (!genesPerTranscript.containsKey(genesOverlappingRegion.size()))
		genesPerTranscript.put(genesOverlappingRegion.size(), new Vector<TranscribedRegion>());
	    genesPerTranscript.get(genesOverlappingRegion.size()).addElement(tr);
	}
	return genesPerTranscript;
    }

    ///////////////////////////////////////////////////////

    public String getBedFileString() { return getBedFileString(false, null, "mm"); }

    public String getBedFileString(boolean reverse, String desc, String species) {
	String str = desc!=null ? "browser position chr1:" + segments.first().getStart() + "-" + segments.first().getEnd() + "\ntrack type=wiggle_0 name=\"" + desc + "\" description=\"" + desc + "\" visibility=dense autoScale=off viewLimits=0:5 color=\"255,127,0\" maxHeightPixels=70:40:10 priority=1\n" : "";
	for (TranscribedRegion t : segments)
	    str += t.getBedFileString(!chip && reverse, species) + "\n";
	return str;
    }

    //////////////////////////////////////////////////////////////////

    public String toString() {
	String str = "";
	for (TranscribedRegion t : segments)
	    str += t.toString() + "\n";
	return str;
    }

}

