//Represent the reduced model from the Haar wavelets. Can be used both
//for ChIP and RNA seq.

import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;
import edu.uci.ics.jung.graph.impl.*;
import umontreal.iro.lecuyer.probdist.NormalDist;
import org.apache.commons.math.special.Gamma;
import umontreal.iro.lecuyer.stat.*;

public class ReducedModel2 implements Cloneable {

    public static final String START_HIST_SUFFIX = "_startProfile";
    public static final String END_HIST_SUFFIX = "_endProfile";
    public static final String BIN_HIST_SUFFIX = "_binProfile";
    public static final String ASSOCIATION_STRENGTH_SUFFIX = "association_strengths";

    public static final double MAX_OTHER_GENE_OVERLAP = .2;
    public static final int MODEL_MIN_BIN_SIZE = 20;
    public static final int N_MODEL_BINS = 100;
    public static final int MIN_READ_NUMBER = 5;
    public static final int MIN_TRANSCRIPT_LENGTH = 200;
    public static final int MAX_REPEAT_REGION_LENGTH = 100000;
    public static final double MAPPABILITY_GAP_THRESHOLD = .2;
    public static final double RANDOM_MAPPABILITY_THRESHOLD = .8;

    protected TreeSet<TranscribedRegion2> segments = new TreeSet<TranscribedRegion2>();
    //Keep track of the transcribed regions and how they overlap annotated genes
    protected GeneTranscriptsGraph2 genesTranscriptsGraph;

    protected String species;
    protected int chr, nReads;
    private boolean forward, chip;

    //For keeping track of the segment density distribution
    protected NormalDist gaussian3[] = new NormalDist[3], gaussian2[] = new NormalDist[2];
;
    protected double weights3[] = new double[3], weight, gaussLogLike[] = new double[3];
    protected double likelihood = Double.POSITIVE_INFINITY, likelihoodSigma = Double.POSITIVE_INFINITY, modifiedBIC, bic;

    protected Mappability mappability = null;
    
    /////////////////////////////////

    public static DecimalFormat format = new DecimalFormat("###.#######");
    public static Random random = new Random(42);

    ///////////////////////////////////////////////////////

    public ReducedModel2(ABIRead rs[], Vector<Integer> breakPoints, int c, int maxReadLength, int lVec[], double cutOffDens[], boolean mergeSegments, boolean log, String sp, boolean ch) {
	//Use this constructor for the FDR approach to keeping
	//transcripts. The idea is to only keep those transcripts
	//whose density is relevant for a given fdr
	//threshold. Calculate densities for several length scales and
	//use the one which is closest to the length. 
	chr = c;
	chip = ch;
	species = sp;
	forward = rs[0].getStrand();
	if (breakPoints.size()>1)
	    segments = createSegments(rs, breakPoints, maxReadLength, lVec, cutOffDens, mergeSegments, log);
    }

    public ReducedModel2(ABIRead rs[], Vector<Integer> breakPoints, int c, int maxReadLength, int lVec[], boolean mergeSegments, boolean log, int binSize, String sp, boolean ch, double fdr, String saveInfoFileName) {
	//Use this constructor for the GMM
	chr = c;
	chip = ch;
	if (!chip)
	    forward = rs[0].getStrand();
	species = sp;
	if (breakPoints.size()>1)
	    segments = createSegments(rs, breakPoints, maxReadLength, lVec, mergeSegments, log, binSize, fdr, saveInfoFileName);
    }

    public ReducedModel2(int c, boolean f, String sp) {
	//Use this constructor when parsing from a .bed file
	chr = c;
	chip = false;
	forward = f;
	species = sp;
    }
    
    public ReducedModel2(String bedFileName, int c, boolean f, String sp) {
	//Use this constructor when parsing from a .bed file
	chr = c;
	forward = f;
	species = sp;
	parseBedFile(bedFileName);
    }

    public ReducedModel2(TreeSet<TranscribedRegion2> segs, int c, boolean f, String sp) {
	//Use this constructor when parsing from a .bed file
	chr = c;
	forward = f;
	species = sp;
	segments = segs;
    }

    public ReducedModel2(String bedFileName, ABIRead[] reads, String rnaName, int c, boolean f, String sp) {
	//Use this constructor when parsing from a .bed file
	chr = c;
	forward = f;
	species = sp;
	parseBedFile(bedFileName, reads, rnaName);
    }

    public void parseBedFile(String bedFileName) { parseBedFile(bedFileName, null); }

    public void parseBedFile(String bedFileName, ABIRead[] reads) { parseBedFile(bedFileName, reads, null); }

    public void parseBedFile(String bedFileName, ABIRead[] reads, String rnaName) {
	int maxReadLength = 0;
	if (reads!=null) {
	    for (int i=0; i<reads.length; i++) 
		maxReadLength = maxReadLength>=reads[i].getLength() ? maxReadLength : reads[i].getLength();
	}
	try {
	    System.out.println("Parsing transcripts from " + bedFileName);
	    BufferedReader br = new BufferedReader(new FileReader(bedFileName));
	    Pattern bedPattern = Pattern.compile("chr([X|Y|\\d]+)\\s+(\\d+)\\s+(\\d+)\\s+([\\d|\\.|-]+)");
	    String line = br.readLine();
	    while (line!=null) {
		Matcher m = bedPattern.matcher(line);
		if (m.find()) {
		    int ch = 0;
		    if (m.group(1).equals("X")) 
			ch = 20;
		    else if (m.group(1).equals("Y")) 
			ch = 21;
		    else
			ch = Integer.parseInt(m.group(1));
		    if (chr==ch) {
			if (reads==null)
			    segments.add(new TranscribedRegion2(Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)), forward, chip, chr, Double.parseDouble(m.group(4))));
			else
			    segments.add(new TranscribedRegion2(Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)), forward, chip, chr, reads, maxReadLength, true, rnaName));
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	}
	catch (IOException e) {
	    System.err.println("Error trying to open " + bedFileName + " :" + e.toString());
	}
    }

    ///////////////////////////////////////////////////////

    public ReducedModel2 getRandomModel(int maxReadLength, ABIRead reads[], boolean log){
	//Use this method to create a random model that has the same
	//lengths as the true transcripts.
	ReducedModel2 randModel = new ReducedModel2(chr, forward, species);
	//Create a set of random transcripts and throw away those with
	//low mappability
	TreeMap<TranscribedRegion2, TranscribedRegion2> rand2real = new TreeMap<TranscribedRegion2, TranscribedRegion2>();
	while (rand2real.size()<numberOfTranscribedRegions()) {
	    for (TranscribedRegion2 t : segments) {
		if (!rand2real.containsValue(t)) 
		    rand2real.put(new TranscribedRegion2(random, t, reads, maxReadLength, log), t);
	    }
	    //Make sure that regions do not overlap and that they have
	    //OK mappability
	    TreeSet<TranscribedRegion2> randSegs = new TreeSet<TranscribedRegion2>(rand2real.values()), removeRegs = new TreeSet<TranscribedRegion2>();
	    TranscribedRegion2 prevReg = null;
	    for (TranscribedRegion2 t : randSegs) {
		if (prevReg!=null && t.getStart()<prevReg.getEnd())
		    removeRegs.add(t);
		prevReg = t;
	    }
	    for (TranscribedRegion2 t : removeRegs) 
		randSegs.remove(t);
	    mappability = new Mappability(true, chr, maxReadLength + 1, maxReadLength, species);
	    for (TranscribedRegion2 t : randSegs) {
		int map = 0;
		mappability.moveWindowStartTo(t.getStart());
		for (int i=0; i<t.getLength(); i++)
		    map += mappability.numberOfPotentialReadsNextWindow()/2;
		if (map<RANDOM_MAPPABILITY_THRESHOLD*(double)t.getLength())
		    randSegs.add(t);
	    }
	    for (TranscribedRegion2 t : removeRegs) 
		randSegs.remove(t);
	    //Now remove the keys that are in removeRegs from the rand2real map
	    for (TranscribedRegion2 t : randSegs) {
		if (rand2real.containsKey(t))
		    rand2real.remove(t);
	    }
	}
	for (TranscribedRegion2 t : rand2real.keySet()) 
	    randModel.addSegment(t);
	return randModel;
    }


    //////////////////////////////////////////////////////////////////

    public int getChromosome() { return chr; }

    public boolean getStrand() { return forward; }

    public Object clone() {
	ReducedModel2 r = new ReducedModel2(chr, forward, species);
	r.nReads = nReads;
	for (TranscribedRegion2 t : segments) //Copy the segments
	    r.segments.add((TranscribedRegion2)t.clone());
	//System.out.println("Reduced model has " + segments.size() + " transcribed regions.");
	return r;
    }

      

    public TreeSet<TranscribedRegion2> createSegments(ABIRead[] reads, Vector<Integer> breakPoints, int maxReadLength, int[] lVec, double[] cutOffDens, boolean mergeSegments, boolean log) {
	TreeSet<TranscribedRegion2> segments = new TreeSet<TranscribedRegion2>();
	//Start with one segment, the one which reduces the error the most
	TranscribedRegion2 seg = new TranscribedRegion2(breakPoints.elementAt(0), breakPoints.elementAt(1), forward, reads, chip, chr, maxReadLength, log);
	if (densityCutOff(seg.getLevel(), seg.getLength(), lVec, cutOffDens)) {
	    segments.add(seg);
	    //System.out.println("Added segment " + seg.toString());
	}
	nReads += seg.numberOfReads();
	for (int i=1; i<breakPoints.size()-1; i++) {
	    seg = new TranscribedRegion2(breakPoints.elementAt(i), breakPoints.elementAt(i+1), forward, reads, chip, chr, maxReadLength, log);
	    if (densityCutOff(seg.getLevel(), seg.getLength(), lVec, cutOffDens)) {
		segments.add(seg);
		//System.out.println("Added segment " + seg.toString());
	    }
	    nReads += seg.numberOfReads();
	}
	return mergeSegments ? merge(segments, reads, maxReadLength, log) : segments;
    }

    public TreeSet<TranscribedRegion2> merge(ABIRead reads[], int maxReadLength, boolean log) { return merge(segments, reads, maxReadLength, log); }

    public TreeSet<TranscribedRegion2> merge(TreeSet<TranscribedRegion2> segments, ABIRead reads[], int maxReadLength, boolean log) {
	//mappability = species.equals("mm") ? new Mappability(true, chr, maxReadLength + 1, maxReadLength) : null;
	mappability = new Mappability(true, chr, maxReadLength + 1, maxReadLength, species);
	TreeSet<TranscribedRegion2> mergedSegs = new TreeSet<TranscribedRegion2>(), removedSegs = new TreeSet<TranscribedRegion2>(), segs2merge = new TreeSet<TranscribedRegion2>();
	Iterator<TranscribedRegion2> it = segments.iterator();
	if (it.hasNext()) {
	    TranscribedRegion2 prevSeg = it.next();
	    while (it.hasNext()) {
		TranscribedRegion2 currSeg = it.next();
		int dist = Math.abs(forward ? currSeg.getStart() - prevSeg.getEnd() : currSeg.getEnd() - prevSeg.getStart());
		if (dist<=MIN_TRANSCRIPT_LENGTH || (mappability!=null && mappabilityGap(forward ? prevSeg.getEnd() : currSeg.getEnd(), forward ? currSeg.getStart() : prevSeg.getStart()))) {
		    segs2merge.add(prevSeg);
		    segs2merge.add(currSeg);
		    removedSegs.add(prevSeg);
		    removedSegs.add(currSeg);
		}
		else {
		    if (segs2merge.size()>0)
			mergedSegs.add(new TranscribedRegion2(segs2merge, reads, maxReadLength, log));
		    segs2merge = new TreeSet<TranscribedRegion2>();
		}
		prevSeg = currSeg;
	    }
	}
	if (segs2merge.size()>0)
	    mergedSegs.add(new TranscribedRegion2(segs2merge, reads, maxReadLength, log));
	segments.removeAll(removedSegs);
	segments.addAll(mergedSegs);
	return segments;
    }

    public boolean mappabilityGap(int gapStart, int gapEnd) { 
	//Find out if there are repeat-regions between the
	//transcripts. If so, we merge the adjacent
	//transcripts
	int gapMappability = 0;
	if (mappability!=null && gapEnd - gapStart<MAX_REPEAT_REGION_LENGTH) {
	    mappability.moveWindowStartTo(gapStart - 1);
	    for (int i=gapStart; i<gapEnd; i++)
		gapMappability += mappability.numberOfPotentialReadsNextWindow()/2;
	    //System.out.println("Between " + gapStart + " and " + gapEnd + " there are " + gapMappability + " out of a potential " + (gapEnd - gapStart));
	    if ((double)gapMappability/(gapEnd - gapStart)<MAPPABILITY_GAP_THRESHOLD) 
		System.out.println("Merging gap between " + gapStart + " and " + gapEnd + " with " + gapMappability + " out of a potential " + (gapEnd - gapStart));
	    return (double)gapMappability/(gapEnd - gapStart)<MAPPABILITY_GAP_THRESHOLD;
		
	}
	return false;
    }

    /////////////////////////////////

    public TreeSet<TranscribedRegion2> createSegments(ABIRead[] reads, Vector<Integer> breakPoints, int maxReadLength, int[] lVec, boolean mergeSegments, boolean log, int binSize, double fdr, String saveInfoFileName) {
	//Use this version when we want to determine the cut-off based
	//on the segment partitioning. We work under the assumption
	//that it is bimodal and we try to find the local minimum.
	int nNonZero = 0;
	TreeSet<TranscribedRegion2> segments = new TreeSet<TranscribedRegion2>();
	segments.add(new TranscribedRegion2(breakPoints.elementAt(0), breakPoints.elementAt(1), forward, reads, chip, chr, maxReadLength, log));
	nNonZero += segments.last().getLevel()>0 ? 1 : 0;
	nReads += segments.last().numberOfReads();
	for (int i=1; i<breakPoints.size()-1; i++) {
	    segments.add(new TranscribedRegion2(breakPoints.elementAt(i), breakPoints.elementAt(i+1), forward, reads, chip, chr, maxReadLength, log));
	    nNonZero += segments.last().getLevel()>0 ? 1 : 0;
	    nReads += segments.last().numberOfReads();
	}
	try {
	    if (saveInfoFileName!=null) {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveInfoFileName + "_segs_" + chr + ".txt"));
		for (TranscribedRegion2 s : segments) 
		    bw.write("chr" + chr + "\t" + s.getStart() + "\t" + s.getEnd() + "\t" + format.format(s.getLevel()) + "\n");
		bw.close();
		System.out.println("Saved " + segments.size() + " segments to " + saveInfoFileName + "_segs_" + chr + ".txt");
	    }
	}
	catch (IOException e) { }	    
	if (nNonZero>0) {
	    double densVec[] = new double[nNonZero];
	    for (TranscribedRegion2 t : segments) {
		if (t.getLevel()>0)
		    densVec[--nNonZero] = Math.log(t.getLevel());
	    }
	    //Now try to determine the distribution of the densities
	    //Try to fit using a single gaussian
	    NormalDist gaussian = NormalDist.getInstanceFromMLE(densVec, densVec.length);
	    //calculate the likelihood for the fit
	    for (int i=0; i<densVec.length; i++)
		gaussLogLike[0] += -Math.log(gaussian.getSigma()) - (gaussian.getMu() - densVec[i])*(gaussian.getMu() - densVec[i])/(2*gaussian.getSigma()*gaussian.getSigma());
	    //System.out.println("logLike = " + format.format(gaussLogLike[0]) + "\ng = " + gaussian.toString());
	    //Now try fitting using two gaussians using a ML-scheme
	    weight = findGMM2(densVec, gaussian, gaussian2);
	    //calculate the log-like and compare
	    gaussLogLike[1] = calculateGMMLikelihood2(gaussian2, densVec, weight);
	    //System.out.println("w = " + weight + ", logLike2 = " + format.format(gaussLogLike[1]));
	    //System.out.println("g0 = " + format.format(weight) + "*normpdf(x, " + format.format(gaussian2[0].getMu()) + ", " + format.format(gaussian2[0].getSigma()) + ");");
	    //System.out.println("g1 = " + format.format(1-weight) + "*normpdf(x, " + format.format(gaussian2[1].getMu()) + ", " + format.format(gaussian2[1].getSigma()) + ");");
	    //Try three components
	    weights3 = findGMMN(densVec, gaussian, gaussian3);
	    double intersection = findGaussianIntersection(gaussian2[0], gaussian2[1], weight, 1-weight);//findGaussianIntersection(gaussian3[0], gaussian3[1], weights3[0], weights3[1]);
	    //calculate the log-like and compare
	    gaussLogLike[2] = calculateGMMLikelihoodN(gaussian3, densVec, weights3);
	    System.out.println("w3 = " + Arrays.toString(weights3) + "\n" + "logLikes = " + Arrays.toString(gaussLogLike));
	    for (int j=0; j<gaussian3.length; j++)
		System.out.println("g" + j + " = " + format.format(weights3[j]) + "*normpdf(x, " + format.format(gaussian3[j].getMu()) + ", " + format.format(gaussian3[j].getSigma()) + ");");
	    //Remove those segments that are more likely to belong to the
	    //distribution with low mean
	    TreeSet<TranscribedRegion2> removeSegs = new TreeSet<TranscribedRegion2>();
	    double maxLevelRemoved = 0;
	    for (TranscribedRegion2 t : segments) {
		if (removeSegment(t, fdr)) {
		    removeSegs.add(t);
		    maxLevelRemoved = t.getLevel()>maxLevelRemoved ? t.getLevel() : maxLevelRemoved;
		    //if (t.getLevel()>0)
		    //  System.out.println("removing " + t.toString() + " with prob " + format.format(calculateComponentProbability(gaussian2, weight, Math.log(t.getLevel()))));
		}
	    }
	    //for (double d=-6.0; d<4.0; d++)
	    //  System.out.println(d + "\t" + format.format(calculateComponentProbability(gaussian3, weights3, d)[0]));
	    System.out.println("Max level removed is " + format.format(maxLevelRemoved) + " (" + format.format(Math.log(maxLevelRemoved)) + ")");
	    segments.removeAll(removeSegs);
	    //Calculate the likelihood before we merge the segments
	    this.segments = segments;
	    //calculateLogEMALikelihood(log, binSize);
	    //calculateBIC(true, binSize, log, true);
	    //calculateBIC(false, binSize, log, false);
	    if (mergeSegments)
		segments = merge(segments, reads, maxReadLength, log);
	    try {
		if (saveInfoFileName!=null) {
		    TreeMap<Double, Integer> resHist = new TreeMap<Double, Integer>();
		    for (TranscribedRegion2 s : segments) {
			double m = s.calculateLogLogLevel(), sd = s.calculateLogLogStdDev(m);
			//resHist = s.getLogLogResiduals(m, sd, resHist);
		    }
		    BufferedWriter bw = new BufferedWriter(new FileWriter(saveInfoFileName + "_res_" + chr + ".txt"));
		    for (Double r : resHist.keySet())
			bw.write(format.format(r) + "\t" + resHist.get(r) + "\n");
		    bw.close();
		}
	    }
	    catch (IOException e) { }	    
	}
	return segments;
    }

    public TreeSet<TranscribedRegion2> createMoreSegments(int nIterations, ABIRead[] reads, Vector<Integer> breakPoints, int maxReadLength, int[] lVec, boolean mergeSegments, boolean log, int binSize, double fdr, String saveInfoFileName) {
	//Use this method when we are running the second pass of the
	//algorithm. We use the same strategy as before, but now we
	//must make sure that the new segments do not span existing ones.
	int nNonZero = 0;
	TreeSet<TranscribedRegion2> segments = new TreeSet<TranscribedRegion2>();
	TranscribedRegion2 seg = new TranscribedRegion2(breakPoints.elementAt(0), breakPoints.elementAt(1), forward, reads, chip, chr, maxReadLength, log);
	if (!segmentOverlaps(seg)) {
	    segments.add(seg);
	    nNonZero += segments.last().getLevel()>0 ? 1 : 0;
	    nReads += segments.last().numberOfReads();
	}
	for (int i=1; i<breakPoints.size()-1; i++) {
	    seg = new TranscribedRegion2(breakPoints.elementAt(i), breakPoints.elementAt(i+1), forward, reads, chip, chr, maxReadLength, log);
	    if (!segmentOverlaps(seg)) {
		segments.add(seg);
		nNonZero += segments.last().getLevel()>0 ? 1 : 0;
		nReads += segments.last().numberOfReads();
	    }
	}
	try {
	    if (saveInfoFileName!=null) {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveInfoFileName + "_segs" + nIterations + "_" + chr + ".txt"));
		for (TranscribedRegion2 s : segments) 
		    bw.write("chr" + chr + "\t" + s.getStart() + "\t" + s.getEnd() + "\t" + format.format(s.getLevel()) + "\n");
		bw.close();
		System.out.println("Saved " + segments.size() + " segments to " + saveInfoFileName + "_segs" + nIterations + "_" + chr + ".txt");
	    }
	}
	catch (IOException e) { }	    
	if (nNonZero>1) {
	    double densVec[] = new double[nNonZero];
	    Tally densTally = new Tally();
	    for (TranscribedRegion2 t : segments) {
		if (t.getLevel()>0) {
		    densVec[--nNonZero] = Math.log(t.getLevel());
		    densTally.add(Math.log(t.getLevel()));
		}
	    }
	    //Now try to determine the distribution of the densities
	    //Arrays.sort(densVec);
	    //Try to fit using a single gaussian
	    try {
		NormalDist gaussian = NormalDist.getInstanceFromMLE(densVec, densVec.length);
		//calculate the likelihood for the fit
		for (int i=0; i<densVec.length; i++)
		    gaussLogLike[0] += -Math.log(gaussian.getSigma()) - (gaussian.getMu() - densVec[i])*(gaussian.getMu() - densVec[i])/(2*gaussian.getSigma()*gaussian.getSigma());
		//System.out.println("logLike = " + format.format(gaussLogLike[0]) + "\ng = " + gaussian.toString());
		//Now try fitting using two gaussians using a ML-scheme
		double weight = findGMM2(densVec, gaussian, gaussian2);
		//calculate the log-like and compare
		gaussLogLike[1] = calculateGMMLikelihood2(gaussian2, densVec, weight);
		//System.out.println("w = " + format.format(weight) + ", logLike2 = " + format.format(gaussLogLike[1]));
		//System.out.println("g0 = " + format.format(weight) + "*normpdf(x, " + format.format(gaussian2[0].getMu()) + ", " + format.format(gaussian2[0].getSigma()) + ");");
		//System.out.println("g1 = " + format.format(1-weight) + "*normpdf(x, " + format.format(gaussian2[1].getMu()) + ", " + format.format(gaussian2[1].getSigma()) + ");");
		//Try three components
		weights3 = findGMMN(densVec, gaussian, gaussian3);
		double intersection = findGaussianIntersection(gaussian2[0], gaussian2[1], weight, 1-weight);
		//calculate the log-like and compare
		gaussLogLike[2] = calculateGMMLikelihoodN(gaussian3, densVec, weights3);
		System.out.println("w3 = " + Arrays.toString(weights3) + "\n" + "logLikes = " + Arrays.toString(gaussLogLike));
		for (int j=0; j<gaussian3.length; j++)
		    System.out.println("g" + j + " = " + format.format(weights3[j]) + "*normpdf(x, " + format.format(gaussian3[j].getMu()) + ", " + format.format(gaussian3[j].getSigma()) + ");");
		//Remove those segments that are more likely to belong to the
		//distribution with low mean
		TreeSet<TranscribedRegion2> removeSegs = new TreeSet<TranscribedRegion2>();
		double maxLevelRemoved = 0;
		for (TranscribedRegion2 t : segments) {
		    if (removeSegment(t, fdr)) {
			removeSegs.add(t);
			maxLevelRemoved = t.getLevel()>maxLevelRemoved ? t.getLevel() : maxLevelRemoved;
			//if (t.getLevel()>0)
			//System.out.println("removing " + t.toString() + " with prob " + format.format(calculateComponentProbability(gaussian2, weight, Math.log(t.getLevel()))));
		    }
		}
		//for (double d=-6.0; d<4.0; d++)
		//  System.out.println(d + "\t" + format.format(calculateComponentProbability(gaussian3, weights3, d)[0]));
		System.out.println("Max level removed is " + format.format(maxLevelRemoved) + " (" + format.format(Math.log(maxLevelRemoved)) + "), " + removeSegs.size() + " out of " + segments.size());
		segments.removeAll(removeSegs);
		for (TranscribedRegion2 t : this.segments)
		    segments.add(t);
		if (mergeSegments)
		    segments = merge(segments, reads, maxReadLength, log);
		this.segments = segments;
	    }
	    catch (Exception e) {
		System.err.println("There was an error trying to create more segments: " + e.toString());
	    }
	}
	return segments;
    }
	
    public TreeSet<TranscribedRegion2> getSegments() { return segments; }

    public TreeSet<TranscribedRegion2> getSegments(String type) { 
	TreeSet<TranscribedRegion2> typeSegs = new TreeSet<TranscribedRegion2>();
	for (TranscribedRegion2 t : segments) {
	    if (t.getType()!=null && t.getType().equals(type))
		typeSegs.add(t);
	}
	return typeSegs;
    }

    protected boolean removeSegment(TranscribedRegion2 t, double fdr, double intersection) {
	if (t.getLevel()==0/* || t.numberOfReads()<MIN_READ_NUMBER || t.getLength()<MIN_TRANSCRIPT_LENGTH*/ || Math.log(t.getLevel())<gaussian3[0].getMu())
	    return true;
	//As an alternative strategy, we simply choose the
	//intersection between the first and second Gaussian as the
	//cut-off point
	return t.getLevel()<intersection;
    }

    protected boolean removeSegment(TranscribedRegion2 t, double fdr) {
	if (t.getLevel()==0 || t.numberOfReads()<MIN_READ_NUMBER || t.getLength()<MIN_TRANSCRIPT_LENGTH || Math.log(t.getLevel())<gaussian3[0].getMu())
	    return true;
	if (gaussLogLike[2]<gaussLogLike[1]) {
	    double p[] = calculateComponentProbability(gaussian3, weights3, Math.log(t.getLevel()));
	    //System.out.println("p = " + format.format(p[0]) + " for " + t.toString());
	    return ((1 - p[0])<1 - fdr) && Math.log(t.getLevel())<gaussian3[1].getMu() + gaussian3[1].getSigma();
	}
	double p = calculateComponentProbability(gaussian2, weight, Math.log(t.getLevel()));
	//System.out.println("p = " + format.format(p) + " for " + t.toString());
	return ((1 - p)<1 - fdr) && Math.log(t.getLevel())<gaussian2[1].getMu() + gaussian2[1].getSigma();
    }

    protected double findGaussianIntersection(NormalDist g1, NormalDist g2, double w1, double w2) {
	//Find the point where the two Gaussians intersect.
	if (g1.getMu()>g2.getMu()) { //make sure that g1<g2
	    NormalDist tmp = g1;
	    g1 = g2;
	    g2 = tmp;
	    double wtmp = w1;
	    w1 = w2;
	    w2 = wtmp;
	}
	//The quadratic equation that we need to solve is -s2^2(x-m1)^2 + s1^2(x-m2)^2 = 2s1^2s2^2log(s1w2/s2w1)
	//Try to get both solutions and choose the one which is in btwn m1 and m2
	//Putting the above eq on normal form: ax^2 + bx + c = 0
	//
	//a = s1^2-s2^2
	//b = 2(s2^2*m1 - s1^2*m2)
	//c = s1^2*m2^2 - s2^2*m1^2 - 2s1^2s2^2log(s1w2/s2w1)
	double a = g1.getSigma()*g1.getSigma() - g2.getSigma()*g2.getSigma();
	double b = 2*(g2.getSigma()*g2.getSigma()*g1.getMu() - g1.getSigma()*g1.getSigma()*g2.getMu());
	double c = g1.getSigma()*g1.getSigma()*g2.getMu()*g2.getMu() - g2.getSigma()*g2.getSigma()*g1.getMu()*g1.getMu() - 2*g1.getSigma()*g1.getSigma()*g2.getSigma()*g2.getSigma()*Math.log(g1.getSigma()*w2/(g2.getSigma()*w1));
	double r1 = (-b + Math.sqrt(b*b - 4*a*c))/(2*a);
	double r2 = (-b - Math.sqrt(b*b - 4*a*c))/(2*a);
	boolean r1Between = r1>g1.getMu() && r1<g2.getMu();
	boolean r2Between = r2>g1.getMu() && r2<g2.getMu();
	System.out.println("roots are " + format.format(r1) + " and " + format.format(r2) + " with m1 = " + format.format(g1.getMu()) + ", m2 = " + format.format(g2.getMu()));
	if (r1Between && !r2Between)
	    return r1;
	else if (!r1Between && r2Between)
	    return r2;
	else {
	    System.err.println("Error, roots are " + format.format(r1) + " and " + format.format(r2) + " with m1 = " + format.format(g1.getMu()) + ", m2 = " + format.format(g2.getMu()));
	    return (g1.getMu() + g2.getMu())/2;
	}
    }

    public void removeShortSegments() {
	//Prune away any short segments. These are typically the
	//result of having introduced extra breakpoints for the RNAPII peaks
	TreeSet<TranscribedRegion2> removeSegs = new TreeSet<TranscribedRegion2>();
	for (TranscribedRegion2 t : segments) {
	    if (t.getLength()<MIN_TRANSCRIPT_LENGTH || t.numberOfReads()<MIN_READ_NUMBER)
		removeSegs.add(t);
	}
	for (TranscribedRegion2 t : removeSegs)
	    segments.remove(t);
    }

    protected double findGMM2(double[] densVec, NormalDist gaussian, NormalDist gaussians[]) {
	//Use this to obtain a GMM using the EM-algorithm for 2 components.
	double weight = .5;
	gaussians[0] = new NormalDist(gaussian.getMu() - gaussian.getSigma(), gaussian.getSigma());
	gaussians[1] = new NormalDist(gaussian.getMu() + gaussian.getSigma(), gaussian.getSigma());
	double logLikePrev = 0, logLike = calculateGMMLikelihood2(gaussians, densVec, weight); //Calculate the likelihood
	while (Math.abs(logLikePrev-logLike)>.01) {
	    //calculate the probability of each point belonging to the first component
	    double p1[] = calculateComponentProbability(gaussians, weight, densVec);
	    //Calculate new values for the parameters
	    double sumP1 = 0, newMu1 = 0, newSigma1 = 0, newMu2 = 0, newSigma2 = 0;
	    for (int i=0; i<p1.length; i++) { 
		sumP1 += p1[i];
		newMu1 += p1[i]*densVec[i];
		newMu2 += (1 - p1[i])*densVec[i];
	    }
	    newMu1 /= sumP1;
	    newMu2 /= (densVec.length - sumP1);
	    for (int i=0; i<p1.length; i++) {
		newSigma1 += p1[i]*(densVec[i] - newMu1)*(densVec[i] - newMu1);
		newSigma2 += (1 - p1[i])*(densVec[i] - newMu2)*(densVec[i] - newMu2);
	    }
	    weight = sumP1/densVec.length;
	    //System.out.println("w = " + weight + ", p1 = " + sumP1 + " of " + densVec.length);
	    //System.out.println("nS1 = " + newSigma1 + ", nS2 = " + newSigma2);
	    try {
		gaussians[0].setParams(newMu1, Math.sqrt(newSigma1/sumP1));
	    }
	    catch (java.lang.IllegalArgumentException e) {
		System.err.println("Error, trying to assign non-positive sigma " + Math.sqrt(newSigma1/sumP1));
		break;
	    }
	    try {
		gaussians[1].setParams(newMu2, Math.sqrt(newSigma2/(densVec.length - sumP1)));
	    }
	    catch (java.lang.IllegalArgumentException e) {
		System.err.println("Error, trying to assign non-positive sigma " + Math.sqrt(newSigma2/(densVec.length - sumP1)));
		break;
	    }
	    logLikePrev = logLike;
	    logLike = calculateGMMLikelihood2(gaussians, densVec, weight); //Calculate the likelihood
	    //System.out.println("logLike = " + format.format(logLike) + " w = " + format.format(weight) + "\ng0 = " + gaussians[0].toString() + "\ng1 = " + gaussians[1].toString());
	}
	//Make sure that the one with lower mean comes first
	if (gaussians[0].getMu()>gaussians[1].getMu()) {
	    NormalDist tmp = gaussians[0];
	    gaussians[0] = gaussians[1];
	    gaussians[1] = tmp;
	}
	return weight;
    }

    protected double[] findGMMN(double[] densVec, NormalDist gaussian, NormalDist gaussians[]) {
	//Use this to obtain a GMM using the EM-algorithm for N components.
	int n = gaussians.length;
	double weights[] = new double[n];
	for (int i=0; i<n; i++) { weights[i] = 1.0/n; }
	for (int i=0; i<n/2; i++) 
	    gaussians[i] = new NormalDist(gaussian.getMu() - (i+1)*gaussian.getSigma(), gaussian.getSigma());
	if (n % 2 == 1)
	    gaussians[n/2] = new NormalDist(gaussian.getMu(), gaussian.getSigma());
	for (int i=n/2+1; i<n; i++) 
	    gaussians[i] = new NormalDist(gaussian.getMu() + i*gaussian.getSigma(), gaussian.getSigma());
	double logLikePrev = 0, logLike = calculateGMMLikelihoodN(gaussians, densVec, weights); //Calculate the likelihood
	while (Math.abs(logLikePrev-logLike)>1) {
	    //calculate the probability of each point belonging to each of the components
	    double p[][] = calculateComponentProbability(gaussians, weights, densVec);
	    //Calculate new values for the parameters
	    double[] newWeights = new double[n], newMu = new double[n], newSigma = new double[n], sumP = new double[n];
	    double newWeightSum = 0;
	    for (int j=0; j<n; j++) {
		for (int i=0; i<p.length; i++) {
		    sumP[j] += p[i][j];
		    newMu[j] += p[i][j]*densVec[i];
		}
		newWeights[j] = sumP[j]/p.length;
		newMu[j] /= sumP[j];
		newWeightSum += newWeights[j];
	    }
	    //System.out.println("mu = " + Arrays.toString(newMu));
	    //System.out.println("w = " + Arrays.toString(newWeights));
	    //Now we can estimate the new variance
	    for (int j=0; j<n; j++) {
		for (int i=0; i<p.length; i++) 
		    newSigma[j] += p[i][j]*(densVec[i] - newMu[j])*(densVec[i] - newMu[j]);
		weights[j] = newWeights[j]/newWeightSum; //Make sure that the weights are normalized
		try {
		    gaussians[j].setParams(newMu[j], Math.sqrt(newSigma[j]/sumP[j])); //update the Gaussians
		}
		catch (java.lang.IllegalArgumentException e) {
		    System.err.println("Error, trying to assign non-positive sigma " + newSigma[j]);
		    break;
		}
	    }
	    logLikePrev = logLike;
	    logLike = calculateGMMLikelihoodN(gaussians, densVec, weights); //Calculate the likelihood
	}
	return weights;
    }

    public static double[][] calculateComponentProbability(NormalDist[] gaussians, double[] weights, double[] densVec) {
	double p[][] = new double[densVec.length][gaussians.length];
	for (int i=0; i<p.length; i++) 
	    p[i] = calculateComponentProbability(gaussians, weights, densVec[i]); 
	return p;
    }

    public static double[] calculateComponentProbability(NormalDist[] gaussians, double[] weights, double dens) {
	double p[] = new double[gaussians.length], normFactor = 0;
	for (int i=0; i<p.length; i++) 
	    normFactor += weights[i]*gaussians[i].density(dens);
	for (int i=0; i<p.length; i++) 
	    p[i] = weights[i]*gaussians[i].density(dens)/normFactor;
	return p;
    }

    public static double[] calculateComponentProbability(NormalDist[] gaussians, double weight, double[] densVec) {
	double p1[] = new double[densVec.length];
	for (int i=0; i<p1.length; i++)
	    p1[i] = calculateComponentProbability(gaussians, weight, densVec[i]); 
	return p1;
    }

    //The probability of belonging to the first Gaussian
    public static double calculateComponentProbability(NormalDist[] gaussians, double weight, double dens) { return weight*gaussians[0].density(dens)/(weight*gaussians[0].density(dens) + (1 - weight)*gaussians[1].density(dens)); }

    public static double calculateGMMLikelihood2(NormalDist[] gaussians, double[] densVec, double weight) {
	double logLike = 0;
	double p1[] = calculateComponentProbability(gaussians, weight, densVec); //probability of point belonging to each component
	for (int i=0; i<densVec.length; i++) {
	    logLike += p1[i]*(Math.log(weight) - Math.log(Math.sqrt(2*Math.PI)) - Math.log(gaussians[0].getSigma()) - (gaussians[0].getMu() - densVec[i])*(gaussians[0].getMu() - densVec[i])/(2*gaussians[0].getSigma()*gaussians[0].getSigma()));
	    logLike += (1 - p1[i])*(Math.log(1 - weight) - Math.log(Math.sqrt(2*Math.PI)) - Math.log(gaussians[1].getSigma()) - (gaussians[1].getMu() - densVec[i])*(gaussians[1].getMu() - densVec[i])/(2*gaussians[1].getSigma()*gaussians[1].getSigma()));
	}
	return logLike;
    }

    public static double calculateGMMLikelihoodN(NormalDist[] gaussians, double[] densVec, double[] weights) {
	double logLike = 0;
	double p[][] = calculateComponentProbability(gaussians, weights, densVec); //probability of point belonging to each component
	for (int i=0; i<densVec.length; i++) {
	    for (int j=0; j<gaussians.length; j++) 
		logLike += p[i][j]*(Math.log(weights[j]) - Math.log(gaussians[j].getSigma()) - Math.log(Math.sqrt(2*Math.PI)) - (gaussians[j].getMu() - densVec[i])*(gaussians[j].getMu() - densVec[i])/(2*gaussians[j].getSigma()*gaussians[j].getSigma()));
	}
	return logLike;
    }

    //////////////////////////////////////////////////////////////////
    //Use these methods for the Bayesian version

    public void addSegment(int start, int end, double l, double std) {
	segments.add(new TranscribedRegion2(start, end, forward, chip, chr, l, std));
    }

    public void addSegment(TranscribedRegion2 t) { segments.add(t); }

    //////////////////////////////////////////////////////////////////

    /////////////////////////////////

    public boolean densityCutOff(double dens, int len, int lVec[], double cutOffs[]) {
	int i = 0;
	while (i<lVec.length && Math.pow(2, lVec[i])<len)
	    i++;
	return dens>cutOffs[i==lVec.length ? i-1 : i];
    }

    public TranscribedRegion2 findTranscribedRegionEnding(TranscribedRegion2 segToAdd) {
	TreeSet<TranscribedRegion2> headSet = new TreeSet<TranscribedRegion2>(segments.headSet(segToAdd));
	if (headSet!=null && headSet.size()>0)
	    return headSet.last().getEnd()==segToAdd.getStart() ? headSet.last() : null;
	return null;
    }

    public TranscribedRegion2 findTranscribedRegionStarting(TranscribedRegion2 segToAdd) {
	TreeSet<TranscribedRegion2> tailSet = new TreeSet<TranscribedRegion2>(segments.tailSet(segToAdd));
	if (tailSet!=null && tailSet.size()>0)
	    return tailSet.first().getStart()==segToAdd.getEnd() ? tailSet.first() : null;
	return null;
    }

    public int numberOfTranscribedRegions() { return segments.size(); }

    public int numberOfTranscribedRegions(String type) { 
	if (type==null)
	    return numberOfTranscribedRegions();
	int n = 0;
	for (TranscribedRegion2 t : segments)
	    n += type.equals(t.getType()) ? 1 : 0;
	return n;
    }

    public int numberOfReads(String type, String factor) { 
	if (type==null)
	    return numberOfReads(factor);
	int n = 0;
	for (TranscribedRegion2 t : segments)
	    n += type.equals(t.getType()) ? t.getFactorReads(factor).length : 0;
	return n;
    }

    public int getLength() { 
	int n = 0;
	for (TranscribedRegion2 t : segments)
	    n += t.getLength();
	return n;
    }

    public int getLength(String type) { 
	if (type==null)
	    return getLength();
	int n = 0;
	for (TranscribedRegion2 t : segments)
	    n += type.equals(t.getType()) ? t.getLength() : 0;
	return n;
    }

    public int numberOfTranscribedRegions(String type, boolean proximal, int proximalDist, RefSeqUCSCEnsemblGenes geneInfo, int geneIDsKeep[]) { return getSegments(type, proximal, proximalDist, geneInfo, geneIDsKeep).size(); }

    public int numberOfReads(String type, boolean proximal, int proximalDist, RefSeqUCSCEnsemblGenes geneInfo, int geneIDsKeep[], String factor) { 
	TreeSet<TranscribedRegion2> segs = getSegments(type, proximal, proximalDist, geneInfo, geneIDsKeep); 
	int n = 0;
	for (TranscribedRegion2 t : segs)
	    n += t.getFactorReads(factor).length;
	return n;
    }
    
    public int getLength(String type, boolean proximal, int proximalDist, RefSeqUCSCEnsemblGenes geneInfo, int geneIDsKeep[]) { 
	TreeSet<TranscribedRegion2> segs = getSegments(type, proximal, proximalDist, geneInfo, geneIDsKeep); 
	int n = 0;
	for (TranscribedRegion2 t : segs)
	    n += t.getLength();
	return n;
    }

    public TreeSet<TranscribedRegion2> getSegments(String type, boolean proximal, int proximalDist, RefSeqUCSCEnsemblGenes geneInfo, int geneIDsKeep[]) { 
	//Use this method for non-coding transcripts if we want to
	//determine if they are close to the start of an annotated
	//gene.
	if (type==null)
	    return getSegments();
	TreeSet<TranscribedRegion2> segs = new TreeSet<TranscribedRegion2>();
	for (TranscribedRegion2 t : segments) {
	    if (type.equals(t.getType())) {
		int start = t.getStrand() ? t.getStart() : t.getEnd();
		if (Math.abs(geneInfo.getNearestTSS(start, !t.getStrand(), geneIDsKeep) - start)<proximalDist==proximal)
		    segs.add(t);
	    }
	}
	return segs;
    }

    public double getPositivePredictiveValue() {
	//PPV = TP/(TP + FP)
	//TP - gene & transcript
	//FP - no gene & transcript
	//Do not penalize for transcripts that are not covering any known genes.
	return (double)(numberOfGenesWithOneTranscript())/(numberOfTranscribedRegions() - numberOfTranscriptsWithNGenes(0, false, false));
    }

    public double getSpecificity(int[] geneIDsKeep) {
	//Specificity = TN/(TN + FP), where
	//TN - no gene & no transcript
	//FP - no gene & transcript
	return 0; //can't figure out how to define TN
    }

    public double getSensitivity(int[] geneIDsKeep) {
	//Specificity = TP/(TP + FN), where
	//TN - no gene & no transcript
	//FN - gene & no transcript
	return geneIDsKeep!=null ? (double)(numberOfGenesWithOneTranscript())/geneIDsKeep.length : 0;
    }

    public double meanTranscribedRegionLength() { return meanTranscribedRegionLength(true); }

    public double meanTranscribedRegionLength(boolean log) {
	double len = 0;
	for (TranscribedRegion2 reg : segments)
	    len += log ? Math.log(reg.getLength()) : reg.getLength();
	return len/segments.size();
    }

    public double proportionTranscribed() { 
	double len = 0;
	for (TranscribedRegion2 reg : segments)
	    len += reg.getLength();
	return len/TranscriptFinder.numberOfBasePairs(chr, species);
    }

    public int numberOfReads(String factor) {
	//returns the number of reads in transcribed regions
	int n = 0;
	for (TranscribedRegion2 tr : segments)
	    n += tr.getFactorReads(factor).length;
	return n;
    }

    public double proportionOfReads(String factor) { return (double)numberOfReads(factor)/nReads; }

    public double minTranscribedRegionDensity(boolean log) {
	double m = 100000;
	for (TranscribedRegion2 t : segments) 
	    m = t.getLevel()>0 ? Math.min(m, t.getLevel()) : m;
	return log ? Math.exp(m) - 1 : m;
    }

    ///////////////////////////////////////////////////////

    public void removeGenes() { if (genesTranscriptsGraph!=null) { genesTranscriptsGraph.removeAllVertices(); } }

    public void removeAll() { if (genesTranscriptsGraph!=null) { genesTranscriptsGraph.removeAll(); } }

    public void refineTranscribedRegionBoundaries(ABIRead reads[], int binSize, int lMin, int maxReadLength, boolean log) {
	//Use this method to get better precision for the boundaries
	//of the called regions. This is necessary if we have used a
	//binSize>0 and thereby coarse-graining the 
	double normCoeff = 1/Math.sqrt(Math.pow(2, lMin + 1));
	//Calculate the coefficient corresponding to having only
	//one read in the window. We would like to avoid this
	//situation since we consider it to be below the noise
	//limit.
	double wThreshold = TranscriptFinder.calculateHaarCoefficientNReads(binSize, (int)Math.pow(2, binSize)/maxReadLength, maxReadLength, log, normCoeff); 
	TreeSet<TranscribedRegion2> segments2remove = new TreeSet<TranscribedRegion2>();
	for (TranscribedRegion2 t : segments) {
	    if (!t.refineBoundaries(reads, lMin, binSize, maxReadLength, log, wThreshold, normCoeff))
		segments2remove.add(t);
	}
	segments.removeAll(segments2remove);
    }

    ///////////////////////////////////////////////////////

    public boolean segmentOverlaps(int pos) {
	for (TranscribedRegion2 t : segments) {
	    if (t.getStart()<=pos && t.getEnd()>=pos)
		return true;
	}
	return false;
    }

    public boolean segmentOverlaps(TranscribedRegion2 seg) { return segmentOverlaps(seg.getStart(), seg.getEnd()); }

    public boolean segmentOverlaps(int s, int e) {
	for (TranscribedRegion2 t : segments) {
	    if (t.getStart()<=s && t.getEnd()>=e)
		return true;
	}
	return false;
    }

    public boolean segmentAdjecent(TranscribedRegion2 seg) { return segmentAdjecent(seg.getStart(), seg.getEnd()); }

    public boolean segmentAdjecent(int s, int e) {
	for (TranscribedRegion2 t : segments) {
	    if (t.getStart()==e || t.getEnd()==s) {
		System.out.println(t.toString() + " is adjacent to " + s + " - " + e);
		return true;
	    }
	}
	return false;
    }

    public void compare2Annotation(int maxDivergentTSSDistance, RefSeqUCSCEnsemblGenes geneInfo) { 
	//Build a graphical model between overlapping genes and
	//transcripts to determine how well we have managed to map one
	//transcript per gene.
	//if (genesTranscriptsGraph!=null) 
	//  genesTranscriptsGraph.removeAll();
	//else
	genesTranscriptsGraph = new GeneTranscriptsGraph2(segments, geneInfo, forward, maxDivergentTSSDistance);
    }

    public void compare2Annotation(int maxDivergentTSSDistance, RefSeqUCSCEnsemblGenes geneInfo, TreeSet<LncRNA> lncRNAs, TreeSet<MiRNA> miRNAs, TreeSet<MacroRNA> macroRNAs, TreeSet<Enhancer2> enhancers) {
	genesTranscriptsGraph = new GeneTranscriptsGraph2(geneInfo, chr, forward, maxDivergentTSSDistance, segments, lncRNAs, miRNAs, macroRNAs, enhancers);
    }

    public void categorizeTranscribedRegions() { genesTranscriptsGraph.categorizeTranscribedRegions(); }

    public void categorizeUnannotatedTranscribedRegions(int maxDist, boolean allowERNA) { genesTranscriptsGraph.categorizeUnannotatedTranscribedRegions(maxDist, allowERNA); }
    
    public TreeMap<String, Integer> getTypeCount() { return getTypeCount(null); }

    public TreeMap<String, Integer> getTypeCount(TreeMap<String, Integer> typeCount) {
	if (typeCount==null)
	    typeCount = new TreeMap<String, Integer>();
	for (TranscribedRegion2 t : segments) 
	    typeCount.put(t.getType(), typeCount.containsKey(t.getType()) ? 1 + typeCount.get(t.getType()) : 1);
	return typeCount;
    }

    public TreeMap<RefSeqUCSCEnsemblGene, TranscribedRegion2> findCodingTranscript(TreeMap<RefSeqUCSCEnsemblGene, TranscribedRegion2> dists) {
	//Calculate the distances from the starts and ends to the
	//annotated starts and ends for the genes.
	if (dists==null)
	    dists = new TreeMap<RefSeqUCSCEnsemblGene, TranscribedRegion2>();
	for (TranscribedRegion2 t : segments) {
	    if (t.getType().equals(TranscribedRegion2.CODING_STR)) {
		TreeMap<RefSeqUCSCEnsemblGene, Double> genes2reads = t.getOverlappingGenes(true);
		for (RefSeqUCSCEnsemblGene g : genes2reads.keySet()) {
		    if (genes2reads.get(g)>GeneTranscriptsGraph.MIN_READS_PROPORTION) {
			dists.put(g, t);
		    }
		}
	    }
	}
	return dists;
    }

    public TreeMap<String, TreeSet<TranscribedRegion2>> getTranscribedRegionTypes() {
	TreeMap<String, TreeSet<TranscribedRegion2>> types2segments = new TreeMap<String, TreeSet<TranscribedRegion2>>();
	return types2segments;
    }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesLengths() { return getTranscribedRegionTypesLengths(null); }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesLengths(TreeMap<String, TallyStore> segmentTypeLengths) { return getTranscribedRegionTypesLengths(segmentTypeLengths, segments); }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesLengths(TreeMap<String, TallyStore> segmentTypeLengths, TreeSet<TranscribedRegion2> segs) {
	//get a data structure containing the lengths of the transcribed regions by type
	if (segmentTypeLengths==null)
	    segmentTypeLengths = new TreeMap<String, TallyStore>();
	for (TranscribedRegion2 t : segs) {
	    if (!segmentTypeLengths.containsKey(t.getType()))
		segmentTypeLengths.put(t.getType(), new TallyStore(t.getType()));
	    segmentTypeLengths.get(t.getType()).add(t.getLength());
	}
	return segmentTypeLengths;
    }

    public TreeMap<String, Integer> getTranscribedRegionTypesReads() { return getTranscribedRegionTypesReads(null); }

    public TreeMap<String, Integer> getTranscribedRegionTypesReads(TreeMap<String, Integer> segmentTypeReads) {
	//get a data structure containing the number of reads of the transcribed regions by type
	if (segmentTypeReads==null)
	    segmentTypeReads = new TreeMap<String, Integer>();
	for (TranscribedRegion2 t : segments) 
	    segmentTypeReads.put(t.getType(), segmentTypeReads.containsKey(t.getType()) ? segmentTypeReads.get(t.getType()) + t.numberOfReads() : t.numberOfReads());
	return segmentTypeReads;
    }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesDensities() { return getTranscribedRegionTypesDensities(null); }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesDensities(TreeMap<String, TallyStore> segmentTypeDensities) {
	//get a data structure containing the number of reads of the transcribed regions by type
	if (segmentTypeDensities==null)
	    segmentTypeDensities = new TreeMap<String, TallyStore>();
	for (TranscribedRegion2 t : segments) {
	    if (!segmentTypeDensities.containsKey(t.getType()))
		segmentTypeDensities.put(t.getType(), new TallyStore());
	    segmentTypeDensities.get(t.getType()).add(t.getDensity());
	}
	return segmentTypeDensities;
    }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesWithinDensityStdDevs(int binSize) { return getTranscribedRegionTypesWithinDensityStdDevs(binSize, null); }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesWithinDensityStdDevs(int binSize, TreeMap<String, TallyStore> segmentTypeWithinDensityStdDevs) {
	//get a data structure containing the number of reads of the transcribed regions by type
	if (segmentTypeWithinDensityStdDevs==null)
	    segmentTypeWithinDensityStdDevs = new TreeMap<String, TallyStore>();
	for (TranscribedRegion2 t : segments) {
	    if (!segmentTypeWithinDensityStdDevs.containsKey(t.getType()))
		segmentTypeWithinDensityStdDevs.put(t.getType(), new TallyStore());
	    segmentTypeWithinDensityStdDevs.get(t.getType()).add(t.calculateInSegmentStdDev(binSize));
	}
	return segmentTypeWithinDensityStdDevs;
    }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesPeaks(String factor) { return getTranscribedRegionTypesPeaks(factor, null); }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesPeaks(String factor, TreeMap<String, TallyStore> segmentTypePeaks) {
	//get a data structure containing the number of reads of the transcribed regions by type
	System.out.println("Finding peaks for " + factor);
	if (segmentTypePeaks==null)
	    segmentTypePeaks = new TreeMap<String, TallyStore>();
	for (TranscribedRegion2 t : segments) {
	    for (Peak p : t.getPeaks()) {
		if (p.hasFactor(factor)) {
		    if (!segmentTypePeaks.containsKey(t.getType()))
			segmentTypePeaks.put(t.getType(), new TallyStore());
		    segmentTypePeaks.get(t.getType()).add(1.0);
		    break;
		}
	    }
	}
	return segmentTypePeaks;
    }

    public int numberOfMergedGenes() { return genesTranscriptsGraph.getMergedGenes().size(); }

    public int numberOfSplitGenes() { return genesTranscriptsGraph.getSplitGenes().size(); }

    public void saveGeneInfo(String saveFileName, String rnaFileName, String geneType, double minDens, boolean header) {
	//Use this method to save the relevant information about the
	//coverage of the expressed genes.
	try {
	    System.out.println("Saving gene info to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    TreeSet<RefSeqUCSCEnsemblGene> geneSet = new TreeSet<RefSeqUCSCEnsemblGene>();
	    if (geneType.equals(RefSeqUCSCEnsemblGene.UNIQUE_TRANSCRIPT_GENE_STR))
		geneSet = genesTranscriptsGraph.getUniqueGenes();
	    else if (geneType.equals(RefSeqUCSCEnsemblGene.MERGED_TRANSCRIPT_GENE_STR))
		geneSet = genesTranscriptsGraph.getMergedGenes();
	    else if (geneType.equals(RefSeqUCSCEnsemblGene.SPLIT_TRANSCRIPT_GENE_STR))
		geneSet = genesTranscriptsGraph.getSplitGenes();
	    else if (geneType.equals(RefSeqUCSCEnsemblGene.MISSED_TRANSCRIPT_GENE_STR))
		geneSet = genesTranscriptsGraph.getMissedGenes();
	    else
		System.err.println("Error, can't find any genes with type " + geneType);
	    if (header)
		bw.write("#ID\tReads\tLength\n");
	    for (RefSeqUCSCEnsemblGene g : geneSet) {
		if (g.getReadDensity(rnaFileName)>minDens)
		    bw.write(g.getID() + "\t" + g.getFactorReads(rnaFileName).length + "\t" + g.getLength() + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving gene info " + e.toString());
	}
    }

    ///////////////////////////////////////////////////////

    public void calculateConservation(boolean orchestra) {
	int longest = 0;
	for (TranscribedRegion2 t : getSegments())
	    longest = longest>t.getLength() ? longest : t.getLength();
	PhastConsParser verterbateCons = new PhastConsParser(chr, false, orchestra, longest);
	for (TranscribedRegion2 t : getSegments()) {
	    t.setConservation(verterbateCons.findScores(t.getStart(), t.getEnd()));
	    System.out.println(t.toString() + "\t" + t.getType() + "\t" + t.getAverageConservation());
	}
    }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesConservation() { return getTranscribedRegionTypesConservation(null); }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesConservation(TreeMap<String, TallyStore> segmentTypeConservation) {
	//get a data structure containing the number of reads of the transcribed regions by type
	if (segmentTypeConservation==null)
	    segmentTypeConservation = new TreeMap<String, TallyStore>();
	for (TranscribedRegion2 t : segments) {
	    if (!segmentTypeConservation.containsKey(t.getType()))
		segmentTypeConservation.put(t.getType(), new TallyStore());
	    segmentTypeConservation.get(t.getType()).add(t.getAverageConservation());
	}
	return segmentTypeConservation;
    }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesWithinConservationStdDevs(int binSize) { return getTranscribedRegionTypesWithinConservationStdDevs(binSize, null); }

    public TreeMap<String, TallyStore> getTranscribedRegionTypesWithinConservationStdDevs(int binSize, TreeMap<String, TallyStore> segmentTypeConservation) {
	//get a data structure containing the number of reads of the transcribed regions by type
	if (segmentTypeConservation==null)
	    segmentTypeConservation = new TreeMap<String, TallyStore>();
	for (TranscribedRegion2 t : segments) {
	    if (!segmentTypeConservation.containsKey(t.getType()))
		segmentTypeConservation.put(t.getType(), new TallyStore());
	    segmentTypeConservation.get(t.getType()).add(t.getConservationStdDev((int)Math.pow(2.0, binSize)));
	}
	return segmentTypeConservation;
    }

    ///////////////////////////////////////////////////////

    public TallyStore getUniqueCodingDist(boolean start) {
	//Calculate the distance to the start/end for all the transcripts that uniquely 
	return null;
    }

    
    //public int numberOfGenesWithOneTranscript(boolean unique) { return numberOfGenesWithNTranscripts(1, false, unique); }

    public int numberOfGenesWithNTranscripts(int n, boolean ineq, boolean unique) { return genesTranscriptsGraph.numberOfGenesWithNTranscripts(n, ineq, unique); }

    public int numberOfGenesWithOneTranscript() { return genesTranscriptsGraph.numberOfGenesWithOneTranscript(); }

    public int numberOfTranscriptsWithOneGene() { return numberOfTranscriptsWithOneGene(true); }

    public int numberOfTranscriptsWithOneGene(boolean unique) { return numberOfTranscriptsWithNGenes(1, false, unique); }

    public int numberOfTranscriptsWithNGenes(int n, boolean ineq, boolean unique) { return genesTranscriptsGraph.numberOfTranscriptsWithNGenes(n, ineq, unique); }

    /////////////////////////////////

    public TreeMap<TranscribedRegion2, Peak> compare2Peaks(Vector<Peak> peaks) { return chip ? compare2Peaks(peaks, 0) : compare2PeaksStrand(peaks, 0); }

    public TreeMap<TranscribedRegion2, Peak> compare2PeaksStrand(Vector<Peak> peaks, int windowSize) {
	//Find all peaks that are within windowsize bps of the start of the transcript
	TreeMap<TranscribedRegion2, Peak> overlappingPeaks = new TreeMap<TranscribedRegion2, Peak>();
	for (TranscribedRegion2 t : segments) {
	    for (Peak p : peaks) {
		if (forward) {
		    if (p.overlaps(t.getStart()) || Math.abs(t.getStart() - p.getCenter())<windowSize) {
			overlappingPeaks.put(t, p);
			t.addPeak(p);
		    }
		}
		else {
		    if (p.overlaps(t.getEnd()) || Math.abs(t.getEnd() - p.getCenter())<windowSize) {
			overlappingPeaks.put(t, p);
			t.addPeak(p);
		    }
		}
	    }
	}
	return overlappingPeaks;
    }

    public TreeMap<TranscribedRegion2, Peak> compare2Peaks(Vector<Peak> peaks, int windowSize) {
	//Find all peaks that are within windowsize bps of the start of the transcript
	if (!chip)
	    return compare2PeaksStrand(peaks, windowSize);
	TreeMap<TranscribedRegion2, Peak> overlappingPeaks = new TreeMap<TranscribedRegion2, Peak>();
	for (TranscribedRegion2 t : segments) {
	    for (Peak p : peaks) {
		if (p.overlaps(t.getStart()) || Math.abs(t.getStart() - p.getCenter())<windowSize) {
		    overlappingPeaks.put(t, p);
		    t.addPeak(p);
		}
		else if (p.overlaps(t.getEnd()) || Math.abs(t.getEnd() - p.getCenter())<windowSize) {
		    overlappingPeaks.put(t, p);
		    t.addPeak(p);
		}
	    }
	}
	return overlappingPeaks;
    }    

    ///////////////////////////////////////////////////////

    public void findGenesPerTranscript() { genesTranscriptsGraph.findGenesPerTranscript(); }

    public void findTranscriptsPerGene() { genesTranscriptsGraph.findTranscriptsPerGene(); }

    ///////////////////////////////////////////////////////
    //Methods for investigating the read-densities across the transcribed regions.

    public void extractPhastConsProfiles(int windowSize, int nBins, String saveFileStart, boolean promoterLocationCategories, TreeMap<String, Integer> typeCount, boolean orchestra) {
	//For now only look near the start and ends
	int nFixedBins = 2*windowSize + 1, j = 0;
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	//Histograms, [type][pos]
	double[][] startHist = new double[arr.length][nFixedBins], endHist = new double[arr.length][nFixedBins], binHist = new double[arr.length][nBins];
	for (int i=0; i<arr.length; i++) {
	    if (typeCount.containsKey(arr[i])) {
		startHist[i] = extractAveragePhastConsScore(arr[i], typeCount.get(arr[i]), windowSize, true, orchestra);
		endHist[i] = extractAveragePhastConsScore(arr[i], typeCount.get(arr[i]), windowSize, false, orchestra);
	    }
	}
	saveProfile(startHist, saveFileStart + "_cons" + START_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
	saveProfile(endHist, saveFileStart + "_cons" + END_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
	//saveProfile(binHist, saveFileStart + "_" + fileName.substring(fileName.lastIndexOf("/") + 1) + BIN_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", nBins, true);
    }

    public double[] extractAveragePhastConsScore(String type, int n, int windowSize, boolean start, boolean orchestra) {
	int nFixedBins = 2*windowSize + 1, j = 0;
	double scores[][] = new double[n][nFixedBins];
	PhastConsParser phastConsScores = new PhastConsParser(chr, true, orchestra, nFixedBins);
	for (TranscribedRegion2 t : getSegments(type)) {
	    int s = start ? (forward ? t.getStart() : t.getEnd()) : (forward ? t.getEnd() : t.getStart());
	    if (forward)
		scores[j++] = phastConsScores.findScores(s - windowSize, s + windowSize + 1);
	    else
		scores[j++] = FindReadsAcrossFeature.flipArray(phastConsScores.findScores(s - windowSize, s + windowSize + 1));
	    //Sum up the scores and store for each transcript
	    double phastConsSum = 0;
	    for (int i=0; i<scores[j-1].length; i++)
		phastConsSum += scores[j-1][i];
	    t.setConservation(phastConsSum);
	}
	return FindReadsAcrossFeature.getConservationHistogram(scores);
    }

    ////////////////////////////////////////////

    public void extractGCProfiles(int windowSize, int nBins, String saveFileStart, boolean promoterLocationCategories, TreeMap<String, Integer> typeCount, boolean orchestra) {
	//For now only look near the start and ends
	int nFixedBins = 2*windowSize + 1, j = 0;
	extractSequences(null, windowSize, orchestra);
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	//Histograms, [type][pos]
	double[][] startHist = new double[arr.length][nFixedBins], startHistCpG = new double[arr.length][nFixedBins], endHist = new double[arr.length][nFixedBins], binHist = new double[arr.length][nBins];
	for (int i=0; i<arr.length; i++) {
	    if (typeCount.containsKey(arr[i])) {
		startHist[i] = extractAverageGCProfile(arr[i], typeCount.get(arr[i]), windowSize, true, orchestra, saveFileStart + "_" + arr[i] + "_start_" + (!forward ? "r" : "f") + "_" + chr + ".fa");
		endHist[i] = extractAverageGCProfile(arr[i], typeCount.get(arr[i]), windowSize, false, orchestra, saveFileStart + "_" + arr[i] + "_end_" + (!forward ? "r" : "f") + "_" + chr + ".fa");
		//Calculate the CpG profile
		startHistCpG[i] = extractAverageCpGProfile(arr[i], typeCount.get(arr[i]), windowSize, true, orchestra, saveFileStart + "_" + arr[i] + "_start_" + (!forward ? "r" : "f") + "_" + chr + ".fa");
	    }
	}
	saveProfile(startHist, saveFileStart + "_GC" + START_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
	saveProfile(endHist, saveFileStart + "_GC" + END_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
	saveProfile(startHistCpG, saveFileStart + "_CpG" + START_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
	//saveProfile(binHist, saveFileStart + "_" + fileName.substring(fileName.lastIndexOf("/") + 1) + BIN_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", nBins, true);
    }

    public double[] extractAverageGCProfile(String type, int n, int windowSize, boolean start, boolean orchestra, String saveFileStart) {
	double gcHist[] = new double[FindReadsAcrossFeature.numberOfBins(windowSize, windowSize, 1)];
	if (!start) {
	    int nFixedBins = 2*windowSize + 1, j = 0, startVec[] = new int[n];
	    for (TranscribedRegion2 t : getSegments(type)) 
		startVec[j++] = start ? (forward ? t.getStart() : t.getEnd()) : (forward ? t.getEnd() : t.getStart());
	    return FindReadsAcrossFeature.getGCHistogram(chr, orchestra, startVec, windowSize, windowSize, forward, saveFileStart);
	}
	try {
	    BufferedWriter bw = saveFileStart==null ? null : new BufferedWriter(new FileWriter(saveFileStart));
	    if (bw!=null)
		System.out.println("Saving sequence to " + saveFileStart);
	    String seqs[] = new String[n];
	    int j = 0;
	    for (TranscribedRegion2 t : getSegments(type)) {
		seqs[j] = t.getSequence(start);
		int nGC = 0;
		if (seqs[j]!=null) {
		    char seq[] = seqs[j].toCharArray();
		    for (int i=0; i<seq.length; i++) {
			if (seq[i]=='c' || seq[i]=='C' || seq[i]=='g' || seq[i]=='G') {
			    gcHist[i] += 1.0/n;
			    nGC++;
			}
		    }
		    if (bw!=null)
			bw.write(">chr" + chr + ":" + (forward ? t.getStart() : t.getEnd()) + "\n" + (new String(seq)) + "\n");
		}
		t.setGCContent((double)nGC/(2*windowSize + 1));
		j++;
	    }
	    if (bw!=null)
		bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving sequences " + e.toString());
	}
	return gcHist;
    }


    public double[] extractAverageCpGProfile(String type, int n, int windowSize, boolean start, boolean orchestra, String sequenceFileName) {
	//Calculate the normalized frequency of CpGs as in the paper
	//by Saxonov et al (2006). This is intended to be done after
	//the GC calculation and hence we first try to read the
	//sequences from file.
	double cpgHist[] = new double[FindReadsAcrossFeature.numberOfBins(windowSize, windowSize, 1)], gcHist[] = new double[FindReadsAcrossFeature.numberOfBins(windowSize, windowSize, 1)];
	String seqs[] = new String[n];
	int j = 0;
	for (TranscribedRegion2 t : getSegments(type)) {
	    seqs[j] = t.getSequence(start);
	    if (seqs[j]!=null) {
		char seq[] = seqs[j].toCharArray();
		for (int i=0; i<seq.length; i++) {
		    if (seq[i]=='c' || seq[i]=='C' || seq[i]=='g' || seq[i]=='G') 
			gcHist[i] += 1.0/n;
		}
	    }
	}
	for (TranscribedRegion2 t : getSegments(type)) {
	    seqs[j] = t.getSequence(start);
	    int nCpG = 0;
	    if (seqs[j]!=null) {
		char seq[] = seqs[j].toCharArray();
		for (int i=1; i<seq.length; i++) {
		    //Count the number of CpGs and normalize by the
		    //expected CpG which is defined as (GC/2)^2
		    if (((seq[i-1]=='c' || seq[i-1]=='C') && (seq[i]=='g' || seq[i]=='G'))) {// || ((seq[i-1]=='g' || seq[i-1]=='G') && (seq[i]=='c' || seq[i]=='C'))) {
			cpgHist[i] += (1.0/n)/(((gcHist[i]+gcHist[i-1])/4)*((gcHist[i]+gcHist[i-1])/4)); //Normalize by the average GC-content at this dinucleotide
			nCpG++;
		    }
		}
	    }
	    t.setCpGContent((double)nCpG/(2*windowSize), t.getGCContent());
	    j++;
	}
	if (type.equals(TranscribedRegion2.CODING_STR) || type.equals(TranscribedRegion2.DIVERGENT_STR))
	    System.out.println(format.format((gcHist[windowSize]+gcHist[windowSize-1])/2) + "\t" + format.format(cpgHist[windowSize]));
	return cpgHist;
    }

    public static double normalizedCpG(String sequence) {
	int nGC = 0, nCpG = 0;
	char seq[] = sequence.toCharArray();
	for (int i=0; i<seq.length; i++) {
	    if (seq[i]=='c' || seq[i]=='C' || seq[i]=='g' || seq[i]=='G') {
		nGC++;
		if (i>0 && ((seq[i-1]=='c' || seq[i-1]=='C') && (seq[i]=='g' || seq[i]=='G'))) 
		    nCpG++;
	    }
	}
	double gc = (double)nGC/seq.length;
	return ((double)nCpG/seq.length)/((gc/2)*(gc/2));
    }

    public void saveTranscriptPromoterGC(String saveFileName, boolean header) {
	try {
	    System.out.println("Saving promoter GC to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    //chr strand start end RNA GC CpG type
	    if (header) 
		bw.write("#chr\tstrand\tstart\tend\tRNA\tGC\tCpG\tcons\ttype\tGeneID\n");
	    for (TranscribedRegion2 t : getSegments()) {
		bw.write(chr + "\t" + (forward ? "1" : "0") + "\t" + t.getStart() + "\t" + t.getEnd() + "\t" + format.format(t.getDensity()) + "\t" + format.format(t.getGCContent()) + "\t" + format.format(t.getCpGContent()) + "\t" + format.format(t.getConservation()) + "\t");
		RefSeqUCSCEnsemblGene g = null;
		if (t.getType().equals(TranscribedRegion2.CODING_STR))
		    g = t.getUniqueGene(true);
		else if (t.getType().equals(TranscribedRegion2.DIVERGENT_STR) || t.getType().equals(TranscribedRegion2.DIVERGENT_END_STR))
		    g = t.getDivergentGene(t.getType().equals(TranscribedRegion2.DIVERGENT_STR));
		else if (t.getType().equals(TranscribedRegion2.NON_CODING_AS_STR))
		    g = t.getUniqueGene(false);
		bw.write(t.getType() + "\t" + (g!=null ? g.getID() : "0") + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
		System.err.println("Error saving reads info " + e.toString());
	}
    }

    ////////////////////////////////////////////

    public void extractBinProfiles(int binSize, int readLength, String saveFileName) {

    }

    public void extractSequences(boolean orchestra) { extractSequences(null, 0, orchestra); }

    public void extractSequences(String saveFileStart, boolean orchestra) { extractSequences(saveFileStart, 0, orchestra); }

    public void extractSequences(String saveFileStart, int startWin, boolean orchestra) {
	try {
	    FastaSequence2 fastaReader = new FastaSequence2(chr, orchestra);
	    String sequenceSaveFileName = saveFileStart==null ? null : saveFileStart + (forward ? "f" : "r") + "_" + chr + ".fa";
	    BufferedWriter bw = null;
	    if (sequenceSaveFileName!=null) { 
		System.out.println("Saving sequences to " + sequenceSaveFileName);
		bw = new BufferedWriter(new FileWriter(sequenceSaveFileName));
	    }
	    for (TranscribedRegion2 t : getSegments()) {
		if (startWin>0)
		    t.setSequence(forward ? fastaReader.getSequence(t.getStart() - startWin, 2*startWin+1) : FastaSequence2.complement(FastaSequence2.reverse(fastaReader.getSequence(t.getEnd() - startWin, 2*startWin+1))), true); 
		else
		    t.setSequence(forward ? fastaReader.getSequence(t.getStart(), t.getLength()) : FastaSequence2.complement(FastaSequence2.reverse(fastaReader.getSequence(t.getStart(), t.getLength())))); 
		if (sequenceSaveFileName!=null) 
		    bw.write(t.getFastaStr() + "\n");
	    }
	    if (sequenceSaveFileName!=null)
		bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving sequences " + e.toString());
	}
    }

    public void setFactorReads(String tfName, ABIRead[] reads) { setFactorReads(tfName, reads, -1, true, true); }

    public void setFactorReads(String tfName, ABIRead[] reads, int windowSize, boolean start, boolean upstream) {
	for (TranscribedRegion2 t : segments) {
	    if (windowSize>0) {
		int s = forward ? (start ? t.getStart() : t.getEnd()) : (start ? t.getEnd() : t.getStart());
		s -= upstream ? (forward ? windowSize : 0) : (forward ? 0 : windowSize);
		int startInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, s));
		startInd = startInd<0 ? -startInd-1 : startInd;
		int e = forward ? (start ? t.getStart() : t.getEnd()) : (start ? t.getEnd() : t.getStart());
		e += upstream ? (forward ? 0 : windowSize + 1) : (forward ? windowSize + 1 : 0);
		int endInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, e));
		endInd = endInd<0 ? -endInd-1 : endInd;
		ABIRead rs[] = new ABIRead[endInd - startInd];
		for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
		t.setFactorReads(tfName + (upstream ? "_u" : "_d"), rs);
	    }
	    else {
		int startInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, t.getStart()));
		startInd = startInd<0 ? -startInd-1 : startInd;
		int endInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, t.getEnd()));
		endInd = endInd<0 ? -endInd-1 : endInd;
		ABIRead rs[] = new ABIRead[endInd - startInd];
		for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
		t.setFactorReads(tfName, rs);
		if (tfName.contains("DNAse") && (t.getType().equals(TranscribedRegion2.MI_RNA_STR) || t.getType().equals(TranscribedRegion2.MACRO_RNA_STR)))
		    System.out.println("found " + rs.length + " reads for " + tfName + " for " + t.toString());
	    }
	}
    }

    public void saveTranscriptFactorDensities(TreeMap<String, Integer> normFactors, String inputFileName, String saveFileName, boolean header) {
	try {
	    System.out.println("Saving factor densities to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    //chr strand start end [factor_dens] type
	    if (header) {
		bw.write("#chr\tstrand\tstart\tend\tRNA\t");
		for (String fName : normFactors.keySet()) {
		    if (!fName.equals(inputFileName))
			bw.write(fName.substring(fName.lastIndexOf("/")+1) + "\t");
		}
		bw.write("type\tGeneID\n");
	    }
	    double inputNormFactor = (double)CreateBedFiles.READS_NORMALIZATION_FACTOR/normFactors.get(inputFileName);
	    for (TranscribedRegion2 t : getSegments()) {
		bw.write(chr + "\t" + (forward ? "1" : "0") + "\t" + t.getStart() + "\t" + t.getEnd() + "\t" + format.format(t.getDensity()) + "\t");
		for (String fName : normFactors.keySet()) {
		    if (!fName.equals(inputFileName)) {
			if (fName.contains("RNA") || fName.contains("rna"))
			    bw.write(format.format(((double)t.getFactorReads(fName).length)/t.getLength()) + "\t");
			else
			    bw.write(format.format(((double)t.getFactorReads(fName).length*CreateBedFiles.READS_NORMALIZATION_FACTOR/normFactors.get(fName) - (double)t.getFactorReads(inputFileName).length*inputNormFactor)/t.getLength()) + "\t");
		    }
		}
		RefSeqUCSCEnsemblGene g = null;
		if (t.getType().equals(TranscribedRegion2.CODING_STR))
		    g = t.getUniqueGene(true);
		else if (t.getType().equals(TranscribedRegion2.DIVERGENT_STR) || t.getType().equals(TranscribedRegion2.DIVERGENT_END_STR))
		    g = t.getDivergentGene(t.getType().equals(TranscribedRegion2.DIVERGENT_STR));
		else if (t.getType().equals(TranscribedRegion2.NON_CODING_AS_STR))
		    g = t.getUniqueGene(false);
		bw.write(t.getType() + "\t" + (g!=null ? g.getID() : "0") + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
		System.err.println("Error saving reads info " + e.toString());
	}
    }

    public void saveTranscriptFactorDensities(Vector<String> fileNames, String saveFileName, boolean header) {
	//Use this for RNA when we are not normalizing
	try {
	    System.out.println("Saving factor densities to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    //chr strand start end [factor_dens] type
	    if (header) {
		bw.write("#chr\tstrand\tstart\tend\tRNA\t");
		for (String fName : fileNames) 
		    bw.write(fName.substring(fName.lastIndexOf("/")+1,fName.lastIndexOf(".")) + "_f" + "\t" + fName.substring(fName.lastIndexOf("/")+1,fName.lastIndexOf(".")) + "_r" + "\t");
		bw.write("type\tGeneID\n");
	    }
	    for (TranscribedRegion2 t : getSegments()) {
		bw.write(chr + "\t" + (forward ? "1" : "0") + "\t" + t.getStart() + "\t" + t.getEnd() + "\t" + format.format(t.getDensity()) + "\t");
		for (String fName : fileNames) {
		    if (fName.contains("RNA") || fName.contains("rna")) {
			bw.write(format.format(((double)t.getFactorReads(fName.substring(0, fName.lastIndexOf(".")) + "_f").length)/t.getLength()) + "\t");
			bw.write(format.format(((double)t.getFactorReads(fName.substring(0, fName.lastIndexOf(".")) + "_r").length)/t.getLength()) + "\t");
		    }
		}
		RefSeqUCSCEnsemblGene g = null;
		if (t.getType().equals(TranscribedRegion2.CODING_STR))
		    g = t.getUniqueGene(true);
		else if (t.getType().equals(TranscribedRegion2.DIVERGENT_STR) || t.getType().equals(TranscribedRegion2.DIVERGENT_END_STR))
		    g = t.getDivergentGene(t.getType().equals(TranscribedRegion2.DIVERGENT_STR));
		else if (t.getType().equals(TranscribedRegion2.NON_CODING_AS_STR))
		    g = t.getUniqueGene(false);
		bw.write(t.getType() + "\t" + (g!=null ? g.getID() : "0") + "\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
		System.err.println("Error saving reads info " + e.toString());
	}
    }

    public void extractReadsProfiles(TreeMap<String, Integer> nTotalReads, String inputFileName, int windowSize, int nBins, String saveFileStart, boolean promoterLocationCategories, TreeMap<String, Integer> typeCount, int startWinSize) { extractReadsProfiles(nTotalReads, inputFileName, windowSize, nBins, saveFileStart, promoterLocationCategories, typeCount, 0, startWinSize, -1, 0); }

    public void extractReadsProfiles(TreeMap<String, Integer> nTotalReads, String inputFileName, int windowSize, int nBins, String saveFileStart, boolean promoterLocationCategories, TreeMap<String, Integer> typeCount, int selectHighLow, int startWinSize, int characteristicLength, double peakThres) {
	//Get the distribution for start, end and across the length for all types
	double qualityScore = 10;
	int fragmentLength = 120;
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	ABIRead inputReads[] = inputFileName!=null ? PeakFinder2.parseFiles(chr, inputFileName, fragmentLength, true, 0, qualityScore) : new ABIRead[0];
	if (inputFileName!=null) {
	    setFactorReads(inputFileName, inputReads);
	    setFactorReads(inputFileName + "_start_w_" + startWinSize, inputReads, startWinSize, true, true);
	    setFactorReads(inputFileName + "_start_w_" + startWinSize, inputReads, startWinSize, true, false);
	}
	double inputNormFactor = inputFileName==null ? 0 : ((double)inputReads.length)/nTotalReads.get(inputFileName);
	int nFixedBins = 2*windowSize + 1, j = 0;
	//Histograms, [fileName][type][pos]
	int nTFs = inputFileName==null ? nTotalReads.size() : nTotalReads.size() - 1;
	double[][][] startHist = new double[nTFs][arr.length][nFixedBins], endHist = new double[nTFs][arr.length][nFixedBins], binHist = new double[nTFs][arr.length][nBins];
	for (String fileName : nTotalReads.keySet()) {
	    if (!fileName.equals(inputFileName)) {
		ABIRead reads[] = PeakFinder2.parseFiles(chr, fileName, fragmentLength, true, 0, qualityScore);;
		System.out.println("Loaded " + reads.length + " reads for " + fileName + " with ind " + j);
		setFactorReads(fileName, reads);
		setFactorReads(fileName + "_start_w_" + startWinSize, reads, startWinSize, true, true);
		setFactorReads(fileName + "_start_w_" + startWinSize, reads, startWinSize, true, false);
		double normFactor = ((double)reads.length)/nTotalReads.get(fileName);
		calculateAssociationStrengths(reads, normFactor, inputReads, inputNormFactor, fileName, windowSize, characteristicLength, peakThres);
		for (int i=0; i<arr.length; i++) {
		    startHist[j][i] = extractAverageReadsProfile(reads, normFactor, inputReads, inputNormFactor, true, windowSize, arr[i], selectHighLow);
		    endHist[j][i] = extractAverageReadsProfile(reads, normFactor, inputReads, inputNormFactor, false, windowSize, arr[i], selectHighLow);
		    //binHist[j][i] = extractAverageReadsProfile(reads, normFactor, inputReads, inputNormFactor, nBins, arr[i]);
		}
		j++;
	    }
	}
	j = 0;
	for (String fileName : nTotalReads.keySet()) {
	    if (!fileName.equals(inputFileName)) {
		saveProfile(startHist[j], saveFileStart + "_" + fileName.substring(fileName.lastIndexOf("/") + 1) + START_HIST_SUFFIX + "_" + (selectHighLow==0 ? "" : selectHighLow>0 ? "high_" : "low_") + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
		saveProfile(endHist[j], saveFileStart + "_" + fileName.substring(fileName.lastIndexOf("/") + 1) + END_HIST_SUFFIX + "_" + (selectHighLow==0 ? "" : selectHighLow>0 ? "high_" : "low_") + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
		//saveProfile(binHist[j], saveFileStart + "_" + fileName.substring(fileName.lastIndexOf("/") + 1) + BIN_HIST_SUFFIX + "_" + (selectHighLow==0 ? "" : selectHighLow>0 ? "high_" : "low_") + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, nBins, true, typeCount);
		j++;
	    }
	}
    }

    public void extractReadsProfiles(Vector<String> fileNames, boolean plus, int maxReadLength, int windowSize, int nBins, boolean log, String saveFileStart, boolean promoterLocationCategories, TreeMap<String, Integer> typeCount, int startWinSize) {
	//Get the distribution for start, end and across the length
	//for all types. Use this for RNA experiments
	int nFixedBins = 2*windowSize + 1, j = 0;
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	//Histograms, [fileName][type][pos]
	double[][][] startHist = new double[fileNames.size()][arr.length][nFixedBins], endHist = new double[fileNames.size()][arr.length][nFixedBins], binHist = new double[fileNames.size()][arr.length][nBins];
	String repeatFileName = species.toLowerCase().equals("mm") ? EstimateGenicTranscription3.REPEAT_RNA_FILE_MM : EstimateGenicTranscription3.REPEAT_RNA_FILE_HS_36;
	for (String fileName : fileNames) {
	    ABIRead reads[] = ABIReads.loadMRNAReadsChr(fileName, plus, chr, maxReadLength, false, repeatFileName);
	    System.out.println("Loaded " + reads.length + " reads for " + fileName + " with ind " + j);
	    setFactorReads(fileName.substring(0, fileName.lastIndexOf(".")) + "_" + (plus ? "f" : "r"), reads);
	    setFactorReads(fileName.substring(0, fileName.lastIndexOf(".")) + "_" + (plus ? "f" : "r") + "_start_w_" + startWinSize, reads, startWinSize, true, true);
	    setFactorReads(fileName.substring(0, fileName.lastIndexOf(".")) + "_" + (plus ? "f" : "r") + "_start_w_" + startWinSize, reads, startWinSize, true, false);
	    for (int i=0; i<arr.length; i++) {
		startHist[j][i] = extractAverageReadsProfile(reads, true, windowSize, maxReadLength, arr[i], log);
		endHist[j][i] = extractAverageReadsProfile(reads, false, windowSize, maxReadLength, arr[i], log);
		binHist[j][i] = extractAverageReadsProfile(reads, nBins, arr[i], log);
	    }
	    j++;
	}
	j = 0;
	for (String fileName : fileNames) {
	    saveProfile(startHist[j], saveFileStart + "_" + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")) + "_" + (plus ? "f" : "r") + START_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
	    saveProfile(endHist[j], saveFileStart + "_" + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")) + "_" + (plus ? "f" : "r") + END_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
	    saveProfile(binHist[j], saveFileStart + "_" + fileName.substring(fileName.lastIndexOf("/") + 1, fileName.lastIndexOf(".")) + "_" + (plus ? "f" : "r") + BIN_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, nBins, true, typeCount);
	    j++;
	}
    }

    public void extractReadsProfiles(ABIRead[] reads, int windowSize, int maxReadLength, int nBins, String fileName, boolean promoterLocationCategories, boolean log, TreeMap<String, Integer> typeCount) {
	//Get the distribution for start, end and across the length
	//for all types. Use this method for the RNA reads.
	int nFixedBins = 2*windowSize + 1;
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	//Histograms, [fileName][type][pos]
	double[][] startHist = new double[arr.length][nFixedBins], endHist = new double[arr.length][nFixedBins], binHist = new double[arr.length][nBins];
	for (int i=0; i<arr.length; i++) {
	    startHist[i] = extractAverageReadsProfile(reads, true, windowSize, maxReadLength, arr[i], log);
	    endHist[i] = extractAverageReadsProfile(reads, false, windowSize, maxReadLength, arr[i], log);
	    binHist[i] = extractAverageReadsProfile(reads, nBins, arr[i], log);
	}
	saveProfile(startHist, fileName + START_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
	saveProfile(endHist, fileName + END_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, windowSize, false, typeCount);
	saveProfile(binHist, fileName + BIN_HIST_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt", promoterLocationCategories, nBins, true, typeCount);
    }

    public static void saveProfile(double[][] hist, String saveFileName, boolean promoterLocationCategories, int windowSize, boolean bins) { saveProfile(hist, saveFileName, promoterLocationCategories, windowSize, bins, null); }

    public static void saveProfile(double[][] hist, String saveFileName, boolean promoterLocationCategories, int windowSize, boolean bins, TreeMap<String, Integer> typeCount) {
	System.out.println("Saving profile to " + saveFileName);
	try {
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
	    String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	    if (bins)
		bw.write("#Bin");
	    else
		bw.write("#Pos");
	    for (int i=0; i<arr.length; i++) 
		bw.write("\t" + arr[i]);
	    bw.write("\n");
	    if (typeCount!=null) {//Save the number of transcripts for each category
		bw.write("NaN");
		for (int i=0; i<arr.length; i++)
		    bw.write("\t" + (typeCount.containsKey(arr[i]) ? typeCount.get(arr[i]) : 0));
		bw.write("\n");
	    }
	    for (int k=0; k<(bins ? windowSize : 2*windowSize + 1); k++) {
		bw.write(new Integer((bins ? k+1 : k-windowSize)).toString());
		for (int i=0; i<arr.length; i++)
		    bw.write("\t" + format.format(hist[i][k]));
		bw.write("\n");
	    }
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving read profiles " + e.toString());
	}
    }

    public double[][][] extractReadsProfiles(TreeMap<String, Integer> nTotalReads, String inputFileName, boolean promoterLocationCategories, int windowSize, boolean tss) {
	//Get the distribution for start or end for all types
	double qualityScore = 10;
	int fragmentLength = 120;
	ABIRead inputReads[] = PeakFinder2.parseFiles(chr, inputFileName, fragmentLength, true, 0, qualityScore);
	setFactorReads(inputFileName, inputReads);
	int nFixedBins = 2*windowSize + 1, j = 0;
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	//Histograms, [fileName][type][pos]
	double[][][] hist = new double[nTotalReads.size() - 1][arr.length][nFixedBins];
	for (String fileName : nTotalReads.keySet()) {
	    if (!fileName.equals(inputFileName)) {
		ABIRead reads[] = PeakFinder2.parseFiles(chr, fileName, fragmentLength, true, 0, qualityScore);;
		System.out.println("Loaded " + reads.length + " reads for " + fileName);
		for (int i=0; i<arr.length; i++) 
		    hist[j][i] = extractAverageReadsProfile(reads, ((double)reads.length)/nTotalReads.get(fileName), inputReads, ((double)inputReads.length)/nTotalReads.get(inputFileName), true, windowSize, arr[i]);
		//Also save the reads for the entire transcript so
		//that we can create a file with densities for
		//each factor
		setFactorReads(fileName, reads);
		j++;
	    }
	}
	return hist;
    }

    public double[][] extractReadsProfiles(ABIRead reads[], int windowSize, boolean tss, int maxReadLength, boolean log, TreeMap<String, Integer> typeCount, boolean promoterLocationCategories) {
	//Get the distribution for start or end for all types
	int nFixedBins = 2*windowSize + 1;
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	//Histograms, [fileName][type][pos]
	double[][] hist = new double[arr.length][nFixedBins];
	for (int i=0; i<arr.length; i++) 
	    hist[i] = extractAverageReadsProfile(reads, tss, windowSize, maxReadLength, arr[i], log);
	return hist;
    }


    public double[][][] extractReadsProfiles(TreeMap<String, Integer> nTotalReads, String inputFileName, boolean promoterLocationCategories, int nBins) {
	//Get the distribution across the length for all types
	double qualityScore = 10;
	int fragmentLength = 120, j = 0;
	ABIRead inputReads[] = PeakFinder2.parseFiles(chr, inputFileName, fragmentLength, true, 0, qualityScore);
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	//Histograms, [fileName][type][pos]
	double[][][] hist = new double[nTotalReads.size() - 1][arr.length][nBins];
	for (String fileName : nTotalReads.keySet()) {
	    if (!fileName.equals(inputFileName)) {
		ABIRead reads[] = PeakFinder2.parseFiles(chr, fileName, fragmentLength, true, 0, qualityScore);;
		System.out.println("Loaded " + reads.length + " reads for " + fileName);
		for (int i=0; i<arr.length; i++) 
		    hist[j][i] = extractAverageReadsProfile(reads, ((double)reads.length)/nTotalReads.get(fileName), inputReads, ((double)inputReads.length)/nTotalReads.get(inputFileName), nBins, arr[i]);
		j++;
	    }
	}
	return hist;
    }

    public double[][] extractReadsProfiles(ABIRead[] reads, int nBins, int maxReadLength, boolean log, TreeMap<String, Integer> typeCount, boolean promoterLocationCategories) {
	//Get the distribution across the length for all types
	String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	//Histograms, [type][pos]
	double[][] hist = new double[arr.length][nBins];
	for (int i=0; i<arr.length; i++) 
	    hist[i] = extractAverageReadsProfile(reads, nBins, arr[i], log);
	return hist;
    }

    public double[][] extractReadsProfile(ABIRead reads[], double normFactor, ABIRead inputReads[], double normFactorInput, boolean start, int windowSize, String type, int selectHighLow) {
	//selectHighLow = 0 means all transcripts, -1 means low half
	//and +1 means high half.
	int n = numberOfTranscribedRegions(type);
	double hist[][] = new double[selectHighLow==0 ? n : n/2][2*windowSize + 1], medianExpression = -1;
	if (selectHighLow!=0 && n>0) { //Find the median expression
	    double expressions[] = new double[n];
	    int j = 0;
	    for (TranscribedRegion2 t : segments) { 
		if (t.getType().equals(type)) 
		    expressions[j++] = t.getLevel();
	    }
	    Arrays.sort(expressions);
	    medianExpression = expressions[n/2];
	}
	int i = 0, fragmentLength = 120;
	for (TranscribedRegion2 t : segments) { //Extract the profile surrounding the start/end
	    if (t==null || t.getType().equals(type)) {
		if (selectHighLow==0 || (selectHighLow<0 && t.getLevel()<medianExpression) || (selectHighLow>0 && t.getLevel()>medianExpression)) {
		    int pos = start ? (forward ? t.getStart() : t.getEnd()) : (forward ? t.getEnd() : t.getStart());
		    hist[i++] = FindReadsAcrossFeature.getNormalizedHistogram(reads, inputReads, chr, pos, forward, windowSize, windowSize, fragmentLength, normFactor, normFactorInput);
		}
	    }
	}
	return hist;
    }

    public double[][] extractReadsProfile(ABIRead reads[], boolean start, int windowSize, int maxReadLength, String type, boolean log) {
	double hist[][] = new double[numberOfTranscribedRegions(type)][2*windowSize + 1];
	int i = 0;
	for (TranscribedRegion2 t : segments) { //Extract the profile surrounding the start/end
	    if (t==null || t.getType().equals(type)) {
		int pos = start ? t.getStart() : t.getEnd();
		hist[i++] = TranscriptFinder.calculateHistogram(pos - windowSize, pos + windowSize + 1, reads, chr, maxReadLength, log);
	    }
	}
	return hist;
    }

    public double[] extractAverageReadsProfile(ABIRead reads[], double normFactor, ABIRead inputReads[], double normFactorInput, boolean start, int windowSize, String type, int selectHighLow) { return FindReadsAcrossFeature.findMeanReadHistograms(extractReadsProfile(reads, normFactor, inputReads, normFactorInput, start, windowSize, type, selectHighLow), 2*windowSize+1)[0]; }

    public double[] extractAverageReadsProfile(ABIRead reads[], double normFactor, ABIRead inputReads[], double normFactorInput, boolean start, int windowSize, String type) { return FindReadsAcrossFeature.findMeanReadHistograms(extractReadsProfile(reads, normFactor, inputReads, normFactorInput, start, windowSize, type, 0), 2*windowSize+1)[0]; }

    public double[] extractAverageReadsProfile(ABIRead reads[], boolean start, int windowSize, int maxReadLength, String type, boolean log) { return FindReadsAcrossFeature.findMeanReadHistograms(extractReadsProfile(reads, start, windowSize, maxReadLength, type, log), 2*windowSize+1)[0]; }

    public double[][] extractReadsProfile(ABIRead reads[], double normFactor, ABIRead inputReads[], double normFactorInput, int nBins, String type) { return extractReadsProfile(reads, normFactor, inputReads, normFactorInput, nBins, type, 0); }

    public double[][] extractReadsProfile(ABIRead reads[], double normFactor, ABIRead inputReads[], double normFactorInput, int nBins, String type, int selectHighLow) {
	double hist[][] = new double[numberOfTranscribedRegions(type)][nBins];
	int i = 0, fragmentLength = 120;
	for (TranscribedRegion2 t : segments) { 
	    if (t==null || t.getType().equals(type)) {
		int binSize = (int)Math.round((double)t.getLength()/nBins), start = t.getStart();
		//for (int j= forward ? 1 : nBins-1; j<nBins && j>=0; j+= forward ? 1 : -1) 
		//  hist[i][j] = (((double)FindReadsAcrossFeature.numberOfReads(chr, start + binSize*(j-1), start + binSize*j, fragmentLength, reads))*normFactor - ((double)FindReadsAcrossFeature.numberOfReads(chr, start + binSize*(j-1), start + binSize*j, fragmentLength, inputReads))*normFactorInput)/binSize;
		for (int j=0; j<nBins; j++)
		    hist[i][(forward ? j : nBins - j - 1)] = (((double)FindReadsAcrossFeature.numberOfReads(chr, start + binSize*j, start + binSize*(j+1), fragmentLength, reads))*normFactor - ((double)FindReadsAcrossFeature.numberOfReads(chr, start + binSize*j, start + binSize*(j+1), fragmentLength, inputReads))*normFactorInput)/binSize;
	    }
	    i++;
	}
	return hist;
    }

    public double[][] extractReadsProfile(ABIRead reads[], int nBins, String type, boolean log) {
	double hist[][] = new double[numberOfTranscribedRegions(type)][nBins];
	int i = 0;
	for (TranscribedRegion2 t : segments) { 
	    if (t==null || t.getType().equals(type) && t.getLength()>nBins) {
		int binSize = (int)Math.round((double)t.getLength()/nBins), start = t.getStart();
		//for (int j= (forward ? 1 : nBins-1); j<nBins && j>=0; j+= (forward ? 1 : -1)) {
		for (int j=0; j<nBins; j++) {
		    int n = FindReadsAcrossFeature.numberOfReads(chr, start + binSize*j, start + binSize*(j+1), reads);
		    hist[i][(forward ? j : nBins - j - 1)] = log ? Math.log(1 + (double)n/binSize) : ((double)n)/binSize;
		}
		i++;
	    }
	}
	return hist;
    }

    public double[] extractAverageReadsProfile(ABIRead reads[], double normFactor, ABIRead inputReads[], double normFactorInput, int nBins, String type) { return FindReadsAcrossFeature.findMeanReadHistograms(extractReadsProfile(reads, normFactor, inputReads, normFactorInput, nBins, type), nBins)[0]; }

    public double[] extractAverageReadsProfile(ABIRead reads[], int nBins, String type, boolean log) { return FindReadsAcrossFeature.findMeanReadHistograms(extractReadsProfile(reads, nBins, type, log), nBins)[0]; }

    ///////////////////////////////////////////////////////

    public void calculateAssociationStrengths(ABIRead[] reads, double normFactor, ABIRead[] inputReads, double inputNormFactor, String fileName, int windowSize, int characteristicLength, double peakThres){
	//Use this method to calculate the weighted read sum for each
	//TF in windowSize. Use an exponential weight with
	//characteristic length characteristicLength. This is inspired
	//by the Ouyang et al (2009) paper.
	int fragmentLength = 120;
	if (characteristicLength>0) {
	    for (TranscribedRegion2 t : segments) {
		int h[] = FindReadsAcrossFeature.getHistogram(reads, null, chr, t.getStrand() ? t.getStart() : t.getEnd(), t.getStrand(), windowSize, windowSize, fragmentLength);
		int hInput[] = FindReadsAcrossFeature.getHistogram(inputReads, null, chr, t.getStrand() ? t.getStart() : t.getEnd(), t.getStrand(), windowSize, windowSize, fragmentLength);
		double hd[] = new double[h.length];
		for (int j=0; j<h.length; j++) 
		    hd[j] = ((double)h[j])*normFactor - ((double)hInput[j])*inputNormFactor;
		t.calculateAssociationStrength(fileName, hd, windowSize, windowSize, characteristicLength, peakThres);
		
	    }
	}
    }

    public void saveAssociationStrengths(String saveFileStart, boolean promoterLocationCategories, Vector<String> tfs) {
	try {
	    String[] arr = promoterLocationCategories ? TranscribedRegion2.LOCATION_TYPES_ARRAY_STR : TranscribedRegion2.TYPES_ARRAY_STR;
	    for (int i=0; i<arr.length; i++) {
		String saveFileName = saveFileStart + arr[i] + "_"  + ASSOCIATION_STRENGTH_SUFFIX + "_" + (!forward ? "r" : "f") + "_" + chr + ".txt";
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		System.out.println("Saving association strengths to " + saveFileName);
		if (chr==1) {
		    bw.write("#chr\tstart\tend\tstrand");
		    for (String s : tfs)
			bw.write("\t" + s.substring(s.lastIndexOf("/") + 1));
		    bw.write("\n");
		}
		for (TranscribedRegion2 t : segments) {
		    if (t.getType().equals(arr[i])) {
			bw.write(chr + "\t" + "\t" + t.getStart() + "\t" + t.getEnd() + "\t" + (t.getStrand() ? "1" : "0"));
			for (String s : tfs)
			    bw.write("\t" + format.format(t.getAssociationStrength(s)));
			bw.write("\n");
		    }
		}
		bw.close();
	    }
	}
	catch (IOException e) {
	    System.err.println("Error saving association strengths " + e.toString());
	}
    }

    ///////////////////////////////////////////////////////
    
    public void findTranscriptForests(int maxDist, String saveFileName) {
	//Use this method to find regions with dense
	//transcription. For each region we want to record the
	//following information:
	//
	//chr  start  end  nForward  nReverse  critDist
	//
    }


    ///////////////////////////////////////////////////////

    public String getBedFileString() { return getBedFileString(false, null, "mm", false); }

    public String getBedFileString(boolean reverse, String desc, String species, boolean hg19) {
	String str = desc!=null && false ? "browser position chr1:" + segments.first().getStart() + "-" + segments.first().getEnd() + "\ntrack type=wiggle_0 name=\"" + desc + "\" description=\"" + desc + "\" visibility=dense autoScale=off viewLimits=0:5 color=\"255,127,0\" maxHeightPixels=70:40:10 priority=1\n" : "";
	for (TranscribedRegion2 t : segments)
	    str += t.getBedFileString(!chip && reverse, species, hg19) + "\n";
	return str;
    }

    
    //////////////////////////////////////////////////////////////////
    //Use these functions for saving the reduced model in a .bed-file format

    public void saveReducedModel(String saveFileName, boolean append) { saveReducedModel(saveFileName, null, append, false); }

    public void saveReducedModel(String saveFileName, String desc, boolean append, boolean hg19) {
	try {
	    System.out.println("Saving model with " + numberOfTranscribedRegions() + " regions to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName, append));
	    bw.write(getBedFileString(!forward, chr==1 ? desc : null, species, hg19));
	    bw.close();
	}
	catch (IOException e) {
	    System.err.println("Error saving: " + e.toString());
	}
    }

    public void saveSequences(boolean orchestra, String saveFileName) { saveSequences(orchestra, saveFileName, null, -1); }

    public void saveSequences(boolean orchestra, String saveFileName, String type, int minAAs) {
	try {
	    System.out.println("Saving sequences to " + saveFileName);
	    BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName)), aabws[] = new BufferedWriter[3];
	    if (minAAs>=0) {
		File dirName = new File(saveFileName.substring(0, saveFileName.lastIndexOf("/")) + FindTranscripts2.AA_DIR_NAME);
		dirName.mkdirs();
		for (int i=0; i<3; i++)
		    aabws[i] = new BufferedWriter(new FileWriter(saveFileName.replace(FindTranscripts2.SEQUENCE_DIR_NAME, FindTranscripts2.SEQUENCE_DIR_NAME + FindTranscripts2.AA_DIR_NAME).replace("transcripts", "aas_" + i)));
	    }
	    FastaSequence2 fastaSeq = new FastaSequence2(chr, orchestra);
	    TreeSet<TranscribedRegion2> segs = type==null ? segments : getSegments(type); 
	    for (TranscribedRegion2 t : segs) {
		String header = ">chr" + chr + ":" + t.getStart() + "-" + t.getEnd() + "(" + (forward ? "+" : "-") + ")" + "_" + t.getType();
		String seq = fastaSeq.getSequence(t.getStart(), t.getLength(), !forward);
		bw.write(header + "\n" + seq + "\n");
		if (minAAs>0) {
		    //Get the AA sequence for all three different reading frames
		    for (int i=0; i<3; i++) {
			//We want to split the sequences when we find
			//a stop-codon. Information about this should
			//be stored in the header
			String aaSeqs[] = (FastaSequence2.convert2AAs(seq, i)).split("\\*");
			for (int j=0; j<aaSeqs.length; j++) {
			    if (aaSeqs[j].length()>=minAAs)
				aabws[i].write(header + "_frag_" + j + "\n" + aaSeqs[j] + "\n");
			}
		    }
		}
	    }
	    bw.close();
	    if (minAAs>0) {
		for (int i=0; i<3; i++)
		    aabws[i].close();
	    }
	}
	catch (IOException e) {
	    System.err.println("Error saving: " + e.toString());
	}
    }

    public TranscribedRegion2 getNearestTranscript(int p, boolean start) { return getNearestTranscript(segments, p, start, forward); }

    public static TranscribedRegion2 getNearestTranscript(TreeSet<TranscribedRegion2> segs, int p, boolean start, boolean forward) { 
	//Use this method to find the transcript which has its start (or end) closest to p
	if (segs.size()==0)
	    return null;
	TranscribedRegion2 nearest = segs.first();
	int minDist = Math.abs(p - (start ? (forward ? nearest.getStart() : nearest.getEnd()) : (forward ? nearest.getEnd() : nearest.getStart())));
	for (TranscribedRegion2 t : segs) {
	    int dist = Math.abs(p - (start ? (forward ? t.getStart() : t.getEnd()) : (forward ? t.getEnd() : t.getStart())));
	    if (dist<minDist) {
		nearest = t;
		minDist = dist;
	    }
	}
	return nearest;
    }

    public TranscribedRegion2 getOverlappingSegment(int pos) { return getOverlappingSegment(segments, pos); }

    public static TranscribedRegion2 getOverlappingSegment(TreeSet<TranscribedRegion2> segs, int pos) {
	if (segs.size()==0)
	    return null;
	for (TranscribedRegion2 t : segs) {
	    if (t.getStart()<=pos && t.getEnd()>=pos)
		return t;
	}
	return null;
    }

    public TranscribedRegion2 getOverlappingSegment(int start, int end) { return getOverlappingSegment(segments, start, end); }

    public static TranscribedRegion2 getOverlappingSegment(TreeSet<TranscribedRegion2> segs, int start, int end) {
	if (segs.size()==0)
	    return null;
	for (TranscribedRegion2 t : segs) {
	    if ((t.getStart()<=start && t.getEnd()>=start) || (t.getStart()<=end && t.getEnd()>=end) || (t.getStart()>=start && t.getEnd()<=end))
		return t;
	}
	return null;
    }

    //////////////////////////////////////////////////////////////////

    public String toString() {
	String str = "#Transcripts = " + segments.size() + "\n";
	for (TranscribedRegion2 t : segments)
	    str += t.toString() + "\n";
	return str;
    }

}

