//Use this method to represent the genes from the collection of
//RefSeq, UCSC and Ensembl genes. They only contain info about start,
//stop and strand.

import java.util.*;
import edu.uci.ics.jung.graph.impl.*;

public class RefSeqUCSCEnsemblGene extends UndirectedSparseVertex implements Comparable {

    //Use this class to represent one gene. start will always be
    //smaller than end and the boolean strand tells us from which
    //direction the gene will be transcribed.

    public static final String UNIQUE_TRANSCRIPT_GENE_STR = "Unique";
    public static final String MERGED_TRANSCRIPT_GENE_STR = "Merged";
    public static final String SPLIT_TRANSCRIPT_GENE_STR = "Split";
    public static final String MISSED_TRANSCRIPT_GENE_STR = "Missed";

    //The start is always the 5' end and will thus be larger than end
    //for genes on the - strand.
    protected int chr, id, maxStart, maxEnd;
    protected boolean strand, nonCoding = false;
    protected String name;
    //Keep track of the number of reads, can either be across coding
    //region or just at TSS
    protected TreeMap<String, ABIRead[]> factorReads = new TreeMap<String, ABIRead[]>();

    //////////////////////////////////////////////////////////////////

    public RefSeqUCSCEnsemblGene(int c, int s, int e, boolean st, String n, int i, boolean nc) {
	chr = c;
	maxStart = s;
	maxEnd = e;
	strand = st;
	name = n;
	id = i;
	nonCoding = nc;
    }

    ///////////////////////////////////////////////////////

    public int peakDistance(int peakPos) {
	//Returns the shortest distance to the peak from either the
	//transcription start site or stop site, depending on which
	//one is smaller. Negative values for genes which are upstream
	//and positive values for genes which are downstream.
	int startDist = strand ? peakPos - maxStart : maxEnd - peakPos;
	int endDist = strand ? maxEnd - peakPos : peakPos - maxStart;
	return Math.abs(startDist)<Math.abs(endDist) ? startDist : endDist;
    }

    public int getMinTSSDistance(int pos) { return maxStart - pos; }

    public int getMinTESDistance(int pos) { return maxEnd - pos; }

    public int getStart() { return maxStart; }

    public int getEnd() { return maxEnd; }

    public int getLength() { return Math.abs(maxStart - maxEnd); }

    //////////////////////////////////////////////////////////////////
    //Stuff for adding read info

    public void setFactorReads(String tfName, ABIRead[] reads) { factorReads.put(tfName, reads); }

    public ABIRead[] getFactorReads(String tfName) { return factorReads.containsKey(tfName) ? factorReads.get(tfName) : null; }

    public ABIRead[] getFactorReads() { return factorReads.size()>0 ? factorReads.get(factorReads.firstKey()) : null; }


    public TreeSet<String> getFactors() { return new TreeSet<String>(factorReads.keySet()); }

    public double getReadDensity(String tfName) { return (double)getFactorReads(tfName).length/getLength(); }

    //////////////////////////////////////////////////////////////////

    public boolean getStrand() { return strand; }

    public String getName() { return name; }

    public int getID() { return id; }

    public int getChromosome() { return chr; }

    public boolean isNonCoding() { return nonCoding; }

    /////////////////////////////////


    /////////////////////////////////

    public boolean overlaps(TreeSet<RefSeqUCSCEnsemblGene> genes) {
	//Returns true if we overlap at least one of genes
	for (RefSeqUCSCEnsemblGene g : genes) {
	    if (overlaps(g))
		return true;
	}
	return false;
    }

    public boolean overlaps(RefSeqUCSCEnsemblGene g) {
	if (g.getStrand())
	    return overlaps(g.getStart(), g.getEnd());
	else
	    return overlaps(g.getEnd(), g.getStart());
    }

    public boolean overlaps(Exon e) { return e.getChromosome()==chr && overlaps(e.getStart(), e.getEnd()); }

    public boolean overlaps(ConservationIsland c) { return c.getChromosome()==chr && overlaps(c.getStart(), c.getEnd()); }

    public boolean overlaps(int start, int end) {
	if (strand) {
	    if ((start>=maxStart && start<=maxEnd) || (end>=maxStart && end<=maxEnd) || (maxStart>=start && maxEnd<=end)) //Start or end of region inside gene
		return true;
	}
	else {
	    if ((start<=maxStart && start>=maxEnd) || (end<=maxStart && end>=maxEnd) || (start<=maxEnd && end>=maxStart)) //Start or end of region inside gene
		return true;
	}
	return false;
    }

    public TreeSet<Exon> getOverlappingExons(TreeSet<Exon> exons) {
	TreeSet<Exon> overlappingExons = new TreeSet<Exon>();
	for (Exon e : exons) {
	    if (overlaps(e))
		overlappingExons.add(e);
	}
	return overlappingExons;
    }

    public TreeSet<ConservationIsland> getOverlappingPromoterCIs(ConservationIslands consIslands, int maxPromoterDist) {
	TreeSet<ConservationIsland> overlappingCIs = new TreeSet<ConservationIsland>();
	for (ConservationIsland c : consIslands.getConservationIslands()) {
	    if (chr==c.getChromosome()) {
		if ((strand && c.overlaps(maxStart - maxPromoterDist, maxStart)) || (!strand && c.overlaps(maxStart, maxStart + maxPromoterDist)))
		    overlappingCIs.add(c);
	    }
	}
	return overlappingCIs;
    }

    public TreeSet<ConservationIsland> getOverlappingCIs(ConservationIslands consIslands) {
	TreeSet<ConservationIsland> overlappingCIs = new TreeSet<ConservationIsland>();
	for (ConservationIsland c : consIslands.getConservationIslands()) {
	    if (chr==c.getChromosome() && c.overlaps(strand ? maxStart : maxEnd, strand ? maxEnd : maxStart))
		overlappingCIs.add(c);
	}
	return overlappingCIs;
    }

    //////////////////////////////////////////////////////////////////
    //Stuff for the JUNG package
    
    public Vector<Enhancer> getNearbyEnhancers() { return new Vector<Enhancer>(getNeighbors()); }

    public boolean hasUniqueTranscript() { return getUniqueTranscript()!=null; }

    public TranscribedRegion2 getUniqueTranscript() {
	if (numNeighbors()==1) { //one overlapping
	    TranscribedRegion2 mainOverlappingRegion = (TranscribedRegion2)getNeighbors().iterator().next();
	    if (mainOverlappingRegion.numNeighbors()==1 /*&& (double)ReducedModel.overlap(this, mainOverlappingRegion)>.5*/)
		return mainOverlappingRegion; //One transcript covering more than half the gene and no other genes
	    else { //make sure that overlap of other genes is small
		Iterator<UndirectedSparseVertex> it2 = mainOverlappingRegion.getNeighbors().iterator();
		while (it2.hasNext()) {
		    UndirectedSparseVertex v = it2.next();
		    if (v instanceof RefSeqGene) {
			RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
			if (!this.equals(g) && (double)GeneTranscriptsGraph2.overlap(g, mainOverlappingRegion)/g.getLength()>.5)//ReducedModel.MAX_OTHER_GENE_OVERLAP)
			    return null;
		    }
		}
		return mainOverlappingRegion;
	    }
	}
	else { //more than one transcript overlapping. make sure that one of them covers the majority
	    Iterator<UndirectedSparseVertex> it = getNeighbors().iterator();
	    boolean uniqueTranscripts = true;
	    TranscribedRegion2 mainOverlappingRegion = null;
	    while (it.hasNext()) {
		TranscribedRegion2 t = (TranscribedRegion2)it.next();
		if ((double)GeneTranscriptsGraph2.overlap(this, t)/getLength()>.5)  //transcript mostly overlapping
		    mainOverlappingRegion = t;
	    }
	    if (mainOverlappingRegion!=null) { //make sure that it does not cover much of any other gene
		Iterator<UndirectedSparseVertex> it2 = mainOverlappingRegion.getNeighbors().iterator();
		while (it2.hasNext()) {
		    UndirectedSparseVertex v = it2.next();
		    if (v instanceof RefSeqGene) {
			RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
			if (!this.equals(g) && (double)GeneTranscriptsGraph2.overlap(g, mainOverlappingRegion)/g.getLength()>.5)//ReducedModel.MAX_OTHER_GENE_OVERLAP)
			    return null;
		    }
		}
		return mainOverlappingRegion;
	    }
	    else
		return null;
	}
    }

    //////////////////////////////////////////////////////////////////

    public int compareTo(Object o) {
	//Sort according to chromosomal position
	if (!(o instanceof RefSeqUCSCEnsemblGene))
	    System.err.println("ERROR comparing genes!");
	RefSeqUCSCEnsemblGene mi = (RefSeqUCSCEnsemblGene)o;
	if (chr<mi.chr)
	    return -1;
	else if (chr>mi.chr)
	    return 1;
	else {
	    //sort in descending order for negative strand
	    int s = getStrand() ? getStart() : -getStart();
	    int miS = mi.getStrand() ? mi.getStart() : -mi.getStart();
	    if (s<miS)
		return -1;
	    else if (s>miS)
		return 1;
	    else
		return 0;
	}
    }

    public String toString() {
	return name + " " + chr + ":" + (strand ? maxStart + "-" + maxEnd : maxEnd + "-" + maxStart) + "(" + (strand ? "+" : "-") + ")";
    }

    public static String getHeaderString() { return "Name\tRefSeq\tGeneID\tStart\tEnd\tStrand"; }

    public boolean equals(Object o) {
	if (!(o instanceof RefSeqUCSCEnsemblGene))
	    System.err.println("ERROR comparing genes!");
	RefSeqUCSCEnsemblGene mi = (RefSeqUCSCEnsemblGene)o;
	return mi.chr==chr && mi.maxStart==maxStart && mi.maxEnd==maxEnd && mi.strand==strand;
    }

    public int hashCode() { return 93*chr + 53*maxStart + 29*maxEnd; }

}
