//Keep track of the different segments which are part of the ReducedModel.

import java.io.*;
import java.util.*;
import java.text.DecimalFormat;
import edu.uci.ics.jung.graph.impl.*;


public class TranscribedRegion2 extends UndirectedSparseVertex implements Comparable, Cloneable {

    //Defines the different types. Using Enums does not make life a
    //whole lot easier, so we stick with Strings
    public static final String CODING_STR = "Annotated_coding_gene";
    public static final String MULTI_GENE_STR = "Multiple_annotated_genes";
    public static final String CODING_FRAGMENT_STR = "Fragment_of_annotated_gene";
    public static final String NON_CODING_ANNOTATED_STR = "Annotated_non-coding_gene";
    public static final String NON_CODING_REPEAT_STR = "Annotated_non-coding_repeat_gene";
    public static final String DIVERGENT_STR = "Promoter-upstream";
    public static final String DIVERGENT_END_STR = "Divergent_end";
    public static final String NON_CODING_AS_STR = "Non-promoter-upstream_AS";
    public static final String LNC_RNA_STR = "lncRNA";
    public static final String LNC_RNA_FRAGMENT_STR = "lncRNA_fragment";
    public static final String MI_RNA_STR = "miRNA";
    public static final String PI_RNA_STR = "piRNA";
    public static final String MI_RNA_PRECURSOR_STR = "miRNA_precursor";
    public static final String ERNA_STR = "eRNA";
    public static final String MACRO_RNA_STR = "macroRNA";
    public static final String MACRO_RNA_FRAGMENT_STR = "macroRNA_fragment";
    public static final String RFAM_REGULATORY_REGION_STR = "Rfam_regulatory_region";
    public static final String RFAM_NCRNA_STR = "Rfam_ncRNA";
    public static final String FANTOM_NCRNA_STR = "Fantom_ncRNA";
    public static final String NREB_AS_STR = "NREB_AS";
    public static final String NON_CODING_OTHER_RNA_STR = "Noncoding_other";

    public static final String[] TYPES_ARRAY_STR = new String[]{CODING_STR, MULTI_GENE_STR, CODING_FRAGMENT_STR, NON_CODING_ANNOTATED_STR, DIVERGENT_STR/*, DIVERGENT_END_STR*/, NON_CODING_AS_STR, LNC_RNA_STR/*, LNC_RNA_FRAGMENT_STR, MI_RNA_PRECURSOR_STR*/, MI_RNA_STR, ERNA_STR, MACRO_RNA_STR/*, MACRO_RNA_FRAGMENT_STR*/, NON_CODING_OTHER_RNA_STR};

    //Types based only on location
    public static final int PROXIMAL_CUT_OFF_DIST = 10000;
    public static final String PROMOTER_PROXIMAL_NCRNA_STR = "Proximal";
    public static final String PROMOTER_DISTAL_NCRNA_STR = "Distal";
    
    public static final String[] LOCATION_TYPES_ARRAY_STR = new String[]{CODING_STR, MULTI_GENE_STR, CODING_FRAGMENT_STR, NON_CODING_ANNOTATED_STR, PROMOTER_PROXIMAL_NCRNA_STR, PROMOTER_DISTAL_NCRNA_STR};
    
    //protected TranscribedRegion previousTranscribedRegion, nextTranscribedRegion;
    protected double level, stdDev, hist[], conservation[];
    protected int start, end, chr, nReads;
    protected boolean chip, boundariesAdjusted = false, forward;
    //Keep track of what type of transcript this is. Possibilities are:
    //coding, coding-fragment, multi-gene, non-coding AS, divergent, eRNA, lncRNA, non-coding other
    protected String type, sequence, startSequence;
    protected double gcContent, cpgContent, phastConsSum;
    //Keep track of the peaks that are close to the start of the transcript
    protected TreeSet<Peak> startPeaks = new TreeSet<Peak>();
    //Keep track of the reads for various factors
    protected TreeMap<String, ABIRead[]> factorReads = new TreeMap<String, ABIRead[]>();
    protected TreeMap<String, Double> associationStrengths = new TreeMap<String, Double>();
    //Keep track of the neighboring genes
    protected RefSeqUCSCEnsemblGene divergentGene = null, asGene = null;
    //Keep track of neighboring transcripts
    protected TranscribedRegion2 tail2head = null, head2head = null, tail2tail = null;
    protected TranscribedRegion2 overlap5p = null, overlap3p = null;
    protected TreeSet<TranscribedRegion2> overlapping = new TreeSet<TranscribedRegion2>();

    public static DecimalFormat format = new DecimalFormat("###.#####");

    public TranscribedRegion2(int s, int e, boolean f, ABIRead reads[], boolean ch, int c, int maxReadLength, boolean log) {
	start = s;
	end = e;
	forward = f;
	chr = c;
	chip = ch;
	hist = getHistogram(reads, chr, maxReadLength, log);
	calculateLevelAndStdDev(hist, log);
	//System.out.println("First bin is " + format.format(hist[0]) + ", mid hist is " + format.format(hist[hist.length/2]) + " at level " + format.format(getLevel()) + " with std dev " + format.format(getStdDev()));
    }

    public TranscribedRegion2(TreeSet<TranscribedRegion2> trs, ABIRead reads[], int maxReadLength, boolean log) {
	//Use this constructor when we are merging several regions
	forward = trs.last().forward;
	start = forward ? trs.first().getStart() : trs.last().getEnd();
	end = forward ? trs.last().getEnd() : trs.first().getStart();
	chr = trs.first().chr;
	chip = trs.first().chip;
	startPeaks = forward ? trs.first().startPeaks : trs.last().startPeaks;
	boundariesAdjusted = trs.first().boundariesAdjusted;
	if (start>end) {
	    System.err.println("Error trying illegal merge.");
	    for (TranscribedRegion2 t : trs)
		System.err.println(t.toString());
	}
	hist = getHistogram(reads, chr, maxReadLength, log);
	calculateLevelAndStdDev(hist, log);
    }

    /////////////////////////////////

    public TranscribedRegion2(int s, int e, boolean f, ABIRead reads[], ABIRead inputReads[], boolean ch, int c, int maxReadLength, boolean log) {
	start = s;
	end = e;
	forward = f;
	chr = c;
	chip = ch;
	calculateLevel(reads, inputReads, chr, maxReadLength, log);
    }

    public TranscribedRegion2(TreeSet<TranscribedRegion2> trs, ABIRead reads[], ABIRead inputReads[], int maxReadLength, boolean log) {
	//Use this constructor when we are merging several regions
	start = trs.first().getStart();
	end = trs.last().getEnd();
	forward = trs.last().forward;
	chr = trs.first().chr;
	chip = trs.first().chip;
	boundariesAdjusted = trs.first().boundariesAdjusted;
	startPeaks = trs.first().startPeaks;
	calculateLevel(reads, inputReads, chr, maxReadLength, log);
    }

    //To be used for the Bayesian stuff
    public TranscribedRegion2(int s, int e, boolean f, boolean ch, int c, double l, double std) {
	start = s;
	end = e;
	forward = f;
	chr = c;
	chip = ch;
	level = l;
	stdDev = std;
    }

    public TranscribedRegion2(int s, int e, boolean f, boolean ch, int c, double l) {
	start = s;
	end = e;
	forward = f;
	chr = c;
	chip = ch;
	level = l;
    }

    public TranscribedRegion2(int s, int e, boolean f, boolean ch, int c, ABIRead[] reads, int maxReadLength, boolean log) {
	start = s;
	end = e;
	forward = f;
	chr = c;
	chip = ch;
	hist = getHistogram(reads, chr, maxReadLength, log);
	calculateLevelAndStdDev(hist, log);
    }

        public TranscribedRegion2(int s, int e, boolean f, boolean ch, int c, ABIRead[] reads, int maxReadLength, boolean log, String rnaName) {
	start = s;
	end = e;
	forward = f;
	chr = c;
	chip = ch;
	hist = getHistogram(reads, chr, maxReadLength, log);
	calculateLevelAndStdDev(hist, log);
	if (rnaName!=null)
	    setFactorReadsSearch(rnaName, reads);
    }

    //copy-constructor
    public TranscribedRegion2(TranscribedRegion2 t) {
	level = t.level;
	stdDev = t.stdDev;
	start = t.start;
	end = t.end;
	forward = t.forward;
	chr = t.chr;
	nReads = t.nReads;
	chip = t.chip;
	hist = t.hist;
	boundariesAdjusted = t.boundariesAdjusted;
	startPeaks = t.startPeaks;
	type = t.type;
	sequence = t.sequence;
    }

    public TranscribedRegion2(Random random, TranscribedRegion2 t, ABIRead[] reads, int maxReadLength, boolean log) {
	//Use this constructor to create a randomly located segment on
	//the same chr and strand and length as t.
	end = RefSeqUCSCEnsemblGenes.CHROMOSOME_LENGTH[t.getChromosome()-1] + 1;
	while (end>RefSeqUCSCEnsemblGenes.CHROMOSOME_LENGTH[t.getChromosome()-1]) {
	    start = random.nextInt(RefSeqUCSCEnsemblGenes.CHROMOSOME_LENGTH[t.getChromosome()-1]);
	    end = start + t.getLength();
	}
	forward = t.forward;
	chr = t.chr;
	hist = getHistogram(reads, chr, maxReadLength, log);
	calculateLevelAndStdDev(hist, log);
    }

    //////////////////////////////////////////////////////////////////

    public Object clone() {
	TranscribedRegion2 t = new TranscribedRegion2(this);
	return t;
    }

    public void setEnd(int e) { end = e; }

    public void setStart(int s) { start = s; }

    public int getEnd() { return end; }

    public int getStart() { return start; }

    public int getLength() { return getEnd() - getStart(); }

    public double getLevel() { return level; }

    public double getDensity() { return (double)nReads/getLength(); }

    public double getStdDev() { return stdDev; }

    public void setType(String t) { type = t; }

    public String getType() { return type; }

    public void setSequence(String s) { setSequence(s, false); }
    
    public String getSequence() { return getSequence(false); }

    public void setSequence(String s, boolean start) { 
	if(start)
	    startSequence = s;
	sequence = s;
    }
    
    public String getSequence(boolean start) { return start ? startSequence : sequence; }

    public void setGCContent(double g) { gcContent = g; }

    public void setCpGContent(double c, double g) { cpgContent = c/((g/2)*(g/2)); }

    public double getGCContent() { return gcContent; }

    public double getCpGContent() { return cpgContent; }

    public void setConservation(double c) { phastConsSum = c; }
    
    public double getConservation() { return phastConsSum; }

    public String getFastaStr() { return ">chr" + chr + ":" + start + "-" + end + "(" + ( forward ? "+" : "-" ) + ")" + "," + type + "\n" + sequence; }

    public int getChromosome() { return chr; }

    public boolean getStrand() { return forward; }

    public static String[] getTypesStr() { return TYPES_ARRAY_STR; }

    public TreeSet<Peak> getPeaks() { return startPeaks; }

    public void addPeak(Peak p) { startPeaks.add(p); }

    public void setDivergentGene(RefSeqUCSCEnsemblGene g) { divergentGene = g; }

    public void setASGene(RefSeqUCSCEnsemblGene g) { asGene = g; }

    public RefSeqUCSCEnsemblGene getDivergentGene() { return divergentGene; }

    public RefSeqUCSCEnsemblGene getASGene() { return asGene; }


    public boolean overlaps(TranscribedRegion2 s) { return s.getChromosome()==chr && overlaps(s.getStart(), s.getEnd()); }

    public boolean overlaps(int s, int e) {
	if ((s>start && s<end) || (e>start && e<end)) //Start or end of region inside gene
	    return true;
	else if ((start>s && start<e) || (end>s && end<e)) //Start or end of exon inside region
	    return true;
	return false;
    }

    public boolean overlaps(int p) { return p>getStart() && p<getEnd(); }

    public boolean overlaps(Exon e) { return e.getChromosome()==chr && overlaps(e.getStart(), e.getEnd()); }

    public TreeSet<Exon> getOverlappingExons(TreeSet<Exon> exons) {
	TreeSet<Exon> overlappingExons = new TreeSet<Exon>();
	for (Exon e : exons) {
	    if (overlaps(e))
		overlappingExons.add(e);
	}
	return overlappingExons;
    }

    public int numberOfReads() { return nReads; }

    protected double[] getHistogram(ABIRead reads[], int chr, int maxReadLength) { return getHistogram(reads, chr, maxReadLength); }

    protected double[] getHistogram(ABIRead reads[], int chr, int maxReadLength, boolean log) {
	int startInd = Arrays.binarySearch(reads, new ABIRead(chr, reads[0].getStrand(), start - maxReadLength));
	startInd = startInd<0 ? -startInd-1 : startInd;
	int endInd = Arrays.binarySearch(reads, new ABIRead(chr, reads[0].getStrand(), end));
	endInd = endInd<0 ? -endInd-1 : endInd;	
	nReads = endInd - startInd;
	return /*chip ? ChIPRegionFinder.calculateHistogram(start, end, reads, chr, maxReadLength) :*/ TranscriptFinder.calculateHistogram(start, end, reads, chr, maxReadLength);
    }

    public double calculateLevel(ABIRead reads[], int chr, int maxReadLength) { return calculateLevel(reads, chr, maxReadLength, true); }

    public double calculateLevel(ABIRead reads[], int chr, int maxReadLength, boolean log) { return calculateAverageReadLevel(getHistogram(reads, chr, maxReadLength, log), log); }

    public double calculateLogLogLevel() {
	double n = 0;
	for (int i=0; i<hist.length; i++) 
	    n += Math.log(1 + Math.log(hist[i] + 1));
	return n/hist.length;
    }

    public double calculateLogLogStdDev(double loglogLevel) {
	double n2 = 0;
	for (int i=0; i<hist.length; i++) 
	    n2 += (Math.log(1 + Math.log(hist[i] + 1)) - loglogLevel)*(Math.log(1 + Math.log(hist[i] + 1)) - loglogLevel);
	return Math.sqrt(n2/(hist.length - 1));
    }

    public double calculateLevelAndStdDev(ABIRead reads[], int chr, int maxReadLength) { return calculateLevelAndStdDev(reads, chr, maxReadLength, true); }

    public double calculateLevelAndStdDev(ABIRead reads[], int chr, int maxReadLength, boolean log) {
	double hist[] = getHistogram(reads, chr, maxReadLength, log);
	level = calculateAverageReadLevel(hist, log);
	stdDev = calculateStdDevReadLevel(hist, level, log);
	return level;
    }

    public double calculateLevelAndStdDev(double hist[], boolean log) {
	level = calculateAverageReadLevel(hist, log);
	stdDev = calculateStdDevReadLevel(hist, level, log);
	return level;
    }

    public double calculateLevel(ABIRead reads[], ABIRead inputReads[], int chr, int maxReadLength) { return calculateLevel(reads, inputReads, chr, maxReadLength, true); }

    public double calculateLevel(ABIRead reads[], ABIRead inputReads[], int chr, int maxReadLength, boolean log) {
	int startInd = Arrays.binarySearch(reads, new ABIRead(chr, reads[0].getStrand(), start - maxReadLength));
	startInd = startInd<0 ? -startInd-1 : startInd;
	int endInd = Arrays.binarySearch(reads, new ABIRead(chr, reads[0].getStrand(), end));
	endInd = endInd<0 ? -endInd-1 : endInd;	
	nReads = endInd - startInd;
	if (inputReads!=null) {

	}
	double hist[] = /*chip ? ChIPRegionFinder.calculateHistogram(start, end, reads, chr, maxReadLength) : */TranscriptFinder.calculateHistogram(start, end, reads, chr, maxReadLength);
	level = calculateAverageReadLevel(hist, log);
	return level;
    }

    public double calculateInSegmentStdDev(int binSize) { return calculateInSegmentStdDev(hist, binSize); }

    public static double calculateInSegmentStdDev(double hist[], int binSize) {
	//Calculate the std_dev of adjacent bins in the histogram
	if (hist.length<2*binSize)
	    return Double.NaN;
	int i = 1;
	double v = 0, m = 0, mp = 0;
	for (int j=(i-1)*binSize; j<i*binSize; j++)
	    mp += hist[j]/binSize;
	i++;
	while (i*binSize<hist.length) {
	    m = 0;
	    for (int j=(i-1)*binSize; j<i*binSize; j++)
		m += hist[j]/binSize;
	    v += (m - mp)*(m - mp);
	    mp = m;
	    i++;
	}
	return i>2 ? Math.sqrt(v/(i-2)) : 0;
    }

    public static double calculateLogLikelihood(double hist[], double mean, double stdDev, boolean loglog) {
	//Assuming that the heights of the histogram are
	//(log-)log-normally distributed wrt to the mean and std dev
	double l = 0;
	for (int i=0; i<hist.length; i++) 
	    l += -Math.log(Math.sqrt(2*Math.PI)) - Math.log(stdDev) - (Math.log(1 + hist[i]) - mean)*(Math.log(1 + hist[i]) - mean)/(2*stdDev*stdDev);
	return l;
    }
    
    public TreeMap<Integer, Integer> findBinCount() { return findBinCount(0); }

    public TreeMap<Integer, Integer> findBinCount(int binSize) {
	int bin = (int)Math.pow(2, binSize);
	TreeMap<Integer, Integer> c = new TreeMap<Integer, Integer>();
	for (int i=0; i<hist.length/bin; i++) {
	    double h = hist[i*bin];
	    for (int j=1; j<bin; j++)
		h += hist[i*bin+j];
	    c.put((int)h, c.containsKey((int)h) ? c.get((int)h) + 1 : 1);
	}
	return c;
    }

    public double getSquaredDeviation(double meanLevel, boolean log) { return getSquaredDeviation(meanLevel, 0, log); }

    public double getSquaredDeviation(double meanLevel, int binSize, boolean log) {
	int bin = (int)Math.pow(2, binSize);
	double s = 0;
	for (int i=0; i<hist.length/bin; i++) {
	    double h = hist[i*bin];
	    for (int j=1; j<bin; j++)
		h += log ? Math.log(hist[i*bin+j] + 1) : hist[i*bin+j];
	    s += (h - meanLevel)*(h - meanLevel);
	}
	return s;
    }

    /////////////////////////////////

    public boolean refineBoundaries(ABIRead reads[], int lMin, int binSize, int maxReadLength, boolean log, double wThreshold, double normCoeff) {
	//Use this method to improve the resolution for the start and
	//end of the transcribed region. Search for a 'step' in the
	//2^binSize first and last bps of the transcript.
	if (!boundariesAdjusted) {
	    int bin = (int)Math.pow(2, binSize), len = (int)Math.pow(2, lMin), oldStart = start, oldEnd = end;
	    ABIRead startReads[] = FindReadsAcrossFeature.getReads(chr, start-bin, start+bin, reads);
	    if (startReads.length>0) {
		//Calculate the Haar coefficients
		double readsHist[] = TranscriptFinder.calculateHistogram(startReads, maxReadLength, len, 0);
		double w[] = new double[readsHist.length];
		int maxCoeffInd = 0, histStart = startReads[0].getPosition() - len;
		//Calculate the first coefficient
		w[0] = TranscriptFinder.calculateHaarCoefficient(lMin, 0, readsHist, log, normCoeff);
		w[1] = TranscriptFinder.calculateHaarCoefficientInc(lMin, 1, readsHist, w[0], log, normCoeff);
		for (int j=2; j<readsHist.length; j++) { //subsequent coefficients can be calculated incrementally and hence significantly faster
		    w[j] = TranscriptFinder.calculateHaarCoefficientInc(lMin, j, readsHist, w[j-1], log, normCoeff);
		    if (w[j-1]>w[j-2] && w[j-1]>=w[j]) {
			maxCoeffInd = j;
			if (w[j-1]>wThreshold)
			    break;
		    }
		}
		//System.out.println("Moving start from " + start + " to " + (histStart + maxCoeffInd));
		start = histStart + maxCoeffInd; //Find the largest one
	    }
	    else
		//System.err.println("Error! no end reads for " + toString());
		start += bin;
	    //do the same for the end
	    ABIRead endReads[] = FindReadsAcrossFeature.getReads(chr, end-bin, end+bin, reads);
	    if (endReads.length>0) {
		double[] readsHist = TranscriptFinder.calculateHistogram(endReads, maxReadLength, len, 0);
		double w[] = new double[readsHist.length];
		int maxCoeffInd = 0, histStart = endReads[0].getPosition() - len;
		w[0] = TranscriptFinder.calculateHaarCoefficient(lMin, 0, readsHist, log, normCoeff);
		w[1] = TranscriptFinder.calculateHaarCoefficientInc(lMin, 1, readsHist, w[0], log, normCoeff);
		for (int j=2; j<readsHist.length; j++)  //subsequent coefficients can be calculated incrementally and hence significantly faster
		    w[j] = TranscriptFinder.calculateHaarCoefficientInc(lMin, j, readsHist, w[j-1], log, normCoeff);
		//Select end-point by going in the reverse direction
		for (int j=w.length-3; j>=0; j--) {
		    if (w[j+1]<w[j+2] && w[j+1]<=w[j]) {
			maxCoeffInd = j;
			if (w[j+1]<-wThreshold)
			    break;
		    }
		}
		//System.out.println("Moving end from " + end + " to " + (histStart + maxCoeffInd));
		end = histStart + maxCoeffInd;
	    }
	    else {
		//System.err.println("Error! no end reads for " + toString());
		end -= bin; //maximum shrink
	    }
	    //Update the properties of the transcript
	    if (start<end) {
		hist = getHistogram(reads, chr, maxReadLength, log);
		calculateLevelAndStdDev(hist, log);
		boundariesAdjusted = true;
	    }
	    else
		return false;
	    //System.err.println("Error, start greater than end for " + toString() + " with old values " + oldStart + " - " + oldEnd);
	}
	return true;
    }

    ////////////////////////////////////////////

    public String getBedFileString() { return getBedFileString(true, "mm", false); }

    public String getBedFileString(boolean reverse, String species, boolean hg19) {
	String chrStr = "chr" + chr;
	if (species.equals("mm")) {
	    if (chr==20)
		chrStr = "chrX";
	    else if (chr==21)
		chrStr = "chrY";
	}
	else if (species.equals("hs")) {
	    if (chr==23)
		chrStr = "chrX";
	    else if (chr==24)
		chrStr = "chrY";
	}
	String lStr = !Double.isNaN(level) && !Double.isInfinite(level) /*&& level>stdDev*/ ? format.format(reverse ? -level : level) : "0";
	return chrStr + "\t" + Math.max(1, start) + "\t" + Math.min(RefSeqUCSCEnsemblGenes.getChromosomeLength(chr, species, hg19), end) + "\t" + lStr + "\t" + EstimateGenicTranscription3.getCategoryNumber(getType());
    }

    //////////////////////////////////////////////////////////////////

    public static double calculateAverageReadLevel(double hist[]) { return calculateAverageReadLevel(hist, true); }

    public static double calculateAverageReadLevel(double hist[], boolean log) { 
	double n = 0;
	//System.out.println(Arrays.toString(hist));
	for (int i=0; i<hist.length; i++)
	    n += log ? Math.log(hist[i] + 1) : hist[i];
	//n += hist[i];
	//return (log ? Math.log(n + 1) : n)/hist.length;
	//System.out.print(format.format(n/hist.length) + "\t");
	return n/hist.length;
    }

    public static double calculateStdDevReadLevel(double hist[]) { return calculateStdDevReadLevel(hist, calculateAverageReadLevel(hist)); }

    public static double calculateStdDevReadLevel(double hist[], double mean) { return calculateStdDevReadLevel(hist, mean, true); }

    public static double calculateStdDevReadLevel(double hist[], double mean, boolean log) {
	double n2 = 0;
	for (int i=0; i<hist.length; i++)
	    n2 += log ? (Math.log(hist[i] + 1) - mean)*(Math.log(hist[i] + 1) - mean) : (hist[i] - mean)*(hist[i] - mean);
	return Math.sqrt(n2/(hist.length - 1));
    }

    public TreeMap<RefSeqUCSCEnsemblGene, Double> getOverlappingGenes(boolean sense) {
	//Returns a list of genes on the same or opposite strand that
	//are overlapped by this transcript. The value in the map is
	//the proportion of reads covering the gene that are also
	//included in the TranscribedRegion.
	TreeMap<RefSeqUCSCEnsemblGene, Double> genes2reads = new TreeMap<RefSeqUCSCEnsemblGene, Double>();
	Iterator it = getNeighbors().iterator();
	//System.out.println(toString() + " has " + getNeighbors().size());
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof RefSeqUCSCEnsemblGene) {
		RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
		if ((!sense && g.getStrand()!=forward) || (sense && g.getStrand()==forward)) {
		    if (sense) {
			//Find out how many of the reads found in the gene are also overlapping the transcript
			TreeSet<String> factors = g.getFactors();
			if (factors.size()>1) {
			    System.err.println("Error! found more than one factor!");
			    break;
			}
			ABIRead[] geneReads = g.getFactorReads(factors.first());
			double frac = (double)ABIReads.getReads(chr, getStart(), getEnd(), geneReads).length/geneReads.length;
			genes2reads.put(g, Double.isNaN(frac) ? 0 : frac);
			//System.out.println("Found " + format.format(frac) + " from " + toString() + " overlapping " + g.toString());
		    }
		    else
			genes2reads.put(g, 0.0);
		}
	    }
	}
	return genes2reads;
    }

    public TreeMap<RefSeqUCSCEnsemblGene, Double> getOverlappingGeneBps() {
	//Returns a list of genes on the same or opposite strand that
	//are overlapped by this transcript. The value in the map is
	//the proportion of bps covering the gene that are also
	//included in the TranscribedRegion.
	TreeMap<RefSeqUCSCEnsemblGene, Double> genes2bps = new TreeMap<RefSeqUCSCEnsemblGene, Double>();
	Iterator it = getNeighbors().iterator();
	//System.out.println(toString() + " has " + getNeighbors().size());
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof RefSeqUCSCEnsemblGene) {
		RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
		//Find out how many of the bps found in the
		//gene are also overlapping the transcript
		double frac = 0;
		if (g.getStrand()) 
		    frac = ((double)(Math.min(g.getEnd(), getEnd()) - Math.max(g.getStart(), getStart())))/g.getLength();
		else
		    frac = ((double)(Math.min(g.getStart(), getEnd()) - Math.max(g.getEnd(), getStart())))/g.getLength();
		genes2bps.put(g, Double.isNaN(frac) ? 0 : frac);
		//System.out.println("Found " + format.format(frac) + " from " + toString() + " overlapping " + g.toString());
	    }
	}
	return genes2bps;
    }

    public RefSeqUCSCEnsemblGene getUniqueGene(boolean sense) {
	boolean geneStrand = forward ? (sense ? true : false) : (sense ? false : true);
	if (numNeighbors()==1) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)getNeighbors().iterator().next();
	    if (v instanceof RefSeqUCSCEnsemblGene) {
		RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
		return g.getStrand()==geneStrand ? g : null;
	    }
	}
	Iterator it = getNeighbors().iterator();
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof RefSeqUCSCEnsemblGene) {
		RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
		if (g.getStrand()==geneStrand) {
		    if (sense) {
			//Find out how many of the reads found in the gene are also overlapping the transcript
			TreeSet<String> factors = g.getFactors();
			if (factors.size()>1) {
			    System.err.println("Error! found more than one factor!");
			    break;
			}
			ABIRead[] geneReads = g.getFactorReads(factors.first());
			if ((double)ABIReads.getReads(chr, getStart(), getEnd(), geneReads).length/geneReads.length>GeneTranscriptsGraph.MIN_READS_PROPORTION)
			    return g;
		    }
		    else
			return g;
		}
	    }
	}
	return null;
    }

    public TreeSet<RefSeqUCSCEnsemblGene> getGenes(boolean sense) {
	//Returns a list of genes on the same or opposite strand that
	//are overlapped by this transcript. The value in the map is
	//the proportion of reads covering the gene that are also
	//included in the TranscribedRegion.
	TreeSet<RefSeqUCSCEnsemblGene> genes = new TreeSet<RefSeqUCSCEnsemblGene>();
	Iterator it = getNeighbors().iterator();
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof RefSeqUCSCEnsemblGene) {
		RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
		if ((!sense && g.getStrand()!=forward) || (sense && g.getStrand()==forward)) 
		    genes.add(g);
	    }
	}
	return genes;
    }
    public RefSeqUCSCEnsemblGene getDivergentGene(boolean tss) {
	//Find the gene which the transcript is divergent with respect to.
	Iterator it = getNeighbors().iterator();
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof RefSeqUCSCEnsemblGene) {
		RefSeqUCSCEnsemblGene g = (RefSeqUCSCEnsemblGene)v;
		if (g.getStrand()!=forward)
		    return g;
	    }
	}
	return null;
    }


    public boolean matchesUniquely(TranscribedRegion2 t) {
	//Return true if the two transcripts appear to be equal
	if (chr==t.getChromosome() && forward==t.getStrand() && type.equals(t.getType())) {
	    //if they both belong to the same coding category, then we
	    //want to make sure that they refer to the same gene(s)
	    if (type.equals(CODING_STR)) {
		RefSeqUCSCEnsemblGene g1 = getUniqueGene(true), g2 = t.getUniqueGene(true);
		return g1!=null && g2!=null && g1.equals(g2);
	    }
	    else if (type.equals(CODING_FRAGMENT_STR)) {
		RefSeqUCSCEnsemblGene g1 = getUniqueGene(true), g2 = t.getUniqueGene(true);
		if (g1!=null && g2!=null && g1.equals(g2)) 
		    //Usually there is more than one fragment, so we want to make sure that they also overlap physically
		    return overlaps(t);
	    }
	    else if (type.equals(MULTI_GENE_STR)) {
		//Make sure that all the genes are the same
		TreeSet<RefSeqUCSCEnsemblGene> genes1 = getGenes(true);
		TreeSet<RefSeqUCSCEnsemblGene> genes2 = t.getGenes(true);
		if (genes2.size()!=genes1.size())
		    return false;
		for (RefSeqUCSCEnsemblGene g : genes1) {
		    if (!genes2.contains(g))
			return false;
		}
		return true;
	    }
	    else 
		System.err.println("Don't know how to compare transcripts with type " + type);
	}
	return false;
    }

    ///////////////////////////////////////////////////////

    public void setConservation(double cons[]) { conservation = cons; }

    public double getAverageConservation() {
	double a = 0;
	for (int i=0; i<conservation.length; i++)
	    a += conservation[i];
	return a/getLength();
    }

    public double getConservationStdDev(int binSize) {
	//Calculate the std_dev of adjacent bins in the histogram
	if (conservation.length<2*binSize)
	    return Double.NaN;
	int i = 1;
	double v = 0, m = 0, mp = 0;
	for (int j=(i-1)*binSize; j<i*binSize; j++)
	    mp += conservation[j]/binSize;
	i++;
	while (i*binSize<hist.length) {
	    m = 0;
	    for (int j=(i-1)*binSize; j<i*binSize; j++)
		m += conservation[j]/binSize;
	    v += (m - mp)*(m - mp);
	    mp = m;
	    i++;
	}
	return i>2 ? Math.sqrt(v/(i-2)) : 0;
    }

    public boolean isAnnotated() { return type.equals(CODING_STR) || type.equals(CODING_FRAGMENT_STR) || type.equals(MULTI_GENE_STR); }

    //////////////////////////////////////////////////////////////////

    public void setFactorReadsSearch(String tfName, ABIRead[] reads) { 
	int startInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, getStart()));
	startInd = startInd<0 ? -startInd-1 : startInd;
	int endInd = Arrays.binarySearch(reads, new ABIRead(chr, forward, getEnd()));
	endInd = endInd<0 ? -endInd-1 : endInd;
	ABIRead rs[] = new ABIRead[endInd - startInd];
	for (int i=startInd; i<endInd; i++) { rs[i-startInd] = reads[i]; }
	setFactorReads(tfName, rs);
    }

    public void setFactorReads(String tfName, ABIRead[] reads) { factorReads.put(tfName, reads); }

    public ABIRead[] getFactorReads(String tfName) { return factorReads.containsKey(tfName) ? factorReads.get(tfName) : new ABIRead[0]; }

    public TreeSet<String> getFactors() { return new TreeSet<String>(factorReads.keySet()); }

    public double getReadDensity(String tfName) { return (double)getFactorReads(tfName).length/getLength(); }

    public double calculateAssociationStrength(String factorName, double[] hist, int upstream, int downstream, int characteristicLength, double peakThres) {
	//Use this method to calculalate the weighted sum of reads for
	//each TF near the promoter.
	double s = 0;
	for (int i=0; i<hist.length; i++) {
	    if (hist[i]>peakThres) {
		s += hist[i]*Math.exp(-Math.abs(i - upstream)/characteristicLength);
		//System.out.println("Found peak at " + (i - upstream) + " for " + factorName + " in " + toString());
	    }
	}
	associationStrengths.put(factorName, s);
	return s;
    }

    public double getAssociationStrength(String factorName) { 
	if (associationStrengths.containsKey(factorName))
	    return associationStrengths.get(factorName); 
	return 0;
    }

    ///////////////////////////////////////////////////////

    public int setHead2HeadTranscript(Vector transcripts) { return setHead2HeadTranscript(getTreeSet(transcripts)); }

    public int setHead2HeadTranscript(TreeSet<TranscribedRegion2> transcripts) {
	int minDist = Integer.MAX_VALUE;
	for (TranscribedRegion2 t : transcripts) {
	    int md = Integer.MAX_VALUE;
	    if (t.getChromosome()==getChromosome() && !equals(t) && t.getStrand()!=getStrand() && !overlaps(t)) {
		if (getStrand() && getStart()>t.getEnd()) 
		    md = Math.abs(getStart() - t.getEnd());
		else if (!getStrand() && getEnd()<t.getStart())
		    md = Math.abs(t.getStart() - getEnd());
	    }
	    if (md<minDist) {
		head2head = t;
		minDist = md;
	    }
		    
	}
	return minDist;
    }

    public TranscribedRegion2 getHead2HeadTranscript() { return head2head; }

    public int getHead2HeadDistance() { return head2head==null ? Integer.MAX_VALUE : getStrand() ? getStart() - head2head.getEnd() : head2head.getStart() - getEnd(); }

    public int setTail2HeadTranscript(Vector transcripts) { return setTail2HeadTranscript(getTreeSet(transcripts)); }

    public int setTail2HeadTranscript(TreeSet<TranscribedRegion2> transcripts) {
	int minDist = Integer.MAX_VALUE;
	for (TranscribedRegion2 t : transcripts) {
	    int md = Integer.MAX_VALUE;
	    if (t.getChromosome()==getChromosome() && !equals(t) && t.getStrand()==getStrand() && !overlaps(t)) {
		if (getStrand() && getEnd()<t.getStart()) 
		    md = Math.abs(t.getStart() - getEnd());
		else if (!getStrand() && t.getEnd()<getStart())
		    md = Math.abs(getStart() - t.getEnd());
	    }
	    if (md<minDist) {
		tail2head = t;
		minDist = md;
	    }
	}
	return minDist;
    }

    public TranscribedRegion2 getTail2HeadTranscript() { return tail2head; }

    public int getTail2HeadDistance() { return tail2head==null ? Integer.MAX_VALUE : getStrand() ? tail2head.getStart() - getEnd() : getStart() - tail2head.getEnd(); }

    public int setTail2TailTranscript(Vector transcripts) { return setTail2TailTranscript(getTreeSet(transcripts)); }

    public int setTail2TailTranscript(TreeSet<TranscribedRegion2> transcripts) {
	int minDist = Integer.MAX_VALUE;
	for (TranscribedRegion2 t : transcripts) {
	    int md = Integer.MAX_VALUE;
	    if (t.getChromosome()==getChromosome() && !equals(t) && t.getStrand()!=getStrand() && !overlaps(t)) {
		if (getStrand() && getEnd()<t.getStart()) 
		    md = Math.abs(t.getStart() - getEnd());
		else if (!getStrand() && t.getEnd()<getStart())
		    md = Math.abs(getStart() - t.getEnd());
	    }
	    if (md<minDist) {
		tail2tail = t;
		minDist = md;
	    }
	}
	return minDist;
    }

    public TranscribedRegion2 getTail2TailTranscript() { return tail2tail; }

    public int getTail2TailDistance() { return tail2tail==null ? Integer.MAX_VALUE : getStrand() ? tail2tail.getStart() - getEnd() : getStart() - tail2tail.getEnd(); }

    public void setOverlapping5p(Vector transcripts) { setOverlapping5p(getTreeSet(transcripts)); }

    public void setOverlapping5p(TreeSet<TranscribedRegion2> transcripts) {
	for (TranscribedRegion2 t : transcripts) {
	    if (t.getChromosome()==getChromosome() && !equals(t) && t.getStrand()!=getStrand()) {
		if (getStrand() && t.overlaps(getStart())) {
		    overlap5p = t;
		    break;
		}
		else if (!getStrand() && t.overlaps(getEnd())) {
		    overlap5p = t;
		    break;
		}
	    }
	}
    }

    public TranscribedRegion2 getOverlapping5p() { return overlap5p; }

    public void setOverlapping3p(Vector transcripts) { setOverlapping3p(getTreeSet(transcripts)); }

    public void setOverlapping3p(TreeSet<TranscribedRegion2> transcripts) {
	for (TranscribedRegion2 t : transcripts) {
	    if (t.getChromosome()==getChromosome() && !equals(t) && t.getStrand()!=getStrand()) {
		if (getStrand() && t.overlaps(getEnd())) {
		    overlap3p = t;
		    break;
		}
		else if (!getStrand() && t.overlaps(getStart())) {
		    overlap3p = t;
		    break;
		}
	    }
	}
    }

    public TranscribedRegion2 getOverlapping3p() { return overlap3p; }

    public void setFullyOverlapping(Vector transcripts) { setFullyOverlapping(getTreeSet(transcripts)); }

    private TreeSet<TranscribedRegion2> getTreeSet(Vector transcripts) {
	//utility function for making things easier to work with
	//matlab where we cannot create typed containers.
	TreeSet<TranscribedRegion2> ts = new TreeSet<TranscribedRegion2>();
	Iterator it = transcripts.iterator();
	while (it.hasNext())
	    ts.add((TranscribedRegion2)it.next());
	return ts;
    }

    public void setFullyOverlapping(TreeSet<TranscribedRegion2> transcripts) {
	for (TranscribedRegion2 t : transcripts) {
	    if (t.getChromosome()==getChromosome() && !equals(t) && t.getStrand()!=getStrand()) {
		if (getStrand() && overlaps(t)) 
		    overlapping.add(t);
		else if (!getStrand() && overlaps(t)) 
		    overlapping.add(t);
	    }
	}
    }

    public TreeSet<TranscribedRegion2> getFullyOverlapping() { return overlapping; }

    public int getNearestTranscriptDistance() {
	//Return the distance to the nearest transcript. If there is
	//an overlapping transcript, the distance is zero.
	if (overlap5p!=null || overlap3p!=null || overlapping.size()>0)
	    return 0;
	return Math.min(getHead2HeadDistance(), Math.min(getTail2HeadDistance(), getTail2TailDistance()));
    }

    public TreeMap<LncRNA, Double> getOverlappingLncRNAs() {
	//Returns a list of LncRNA that are overlapped by this
	//transcript. The value in the map is the proportion of reads
	//covering the lncRNA that are also included in the
	//TranscribedRegion.
	TreeMap<LncRNA, Double> lncRNA2reads = new TreeMap<LncRNA, Double>();
	Iterator it = getNeighbors().iterator();
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof LncRNA) {
		LncRNA l = (LncRNA)v;
		TreeSet<String> factors = l.getFactors();
		if (factors.size()>1) {
		    System.err.println("Error! found more than one factor!");
		    break;
		}
		ABIRead[] lncRNAReads = l.getFactorReads(factors.first());
		lncRNA2reads.put(l, (double)ABIReads.getReads(chr, getStart(), getEnd(), lncRNAReads).length/lncRNAReads.length);
	    }
	}
	return lncRNA2reads;
    }

    public TreeMap<MacroRNA, Double> getOverlappingMacroRNAs() {
	//Returns a list of MacroRNAs that are overlapped by this
	//transcript. The value in the map is the proportion of reads
	//covering the MacroRNA that are also included in the
	//TranscribedRegion.
	TreeMap<MacroRNA, Double> macroRNA2reads = new TreeMap<MacroRNA, Double>();
	Iterator it = getNeighbors().iterator();
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof MacroRNA) {
		MacroRNA m = (MacroRNA)v;
		if (m.getStrand()==forward) {
		    TreeSet<String> factors = m.getFactors();
		    if (factors.size()>1) {
			System.err.println("Error! found more than one factor!");
			break;
		    }
		    ABIRead[] macroRNAreads = m.getFactorReads(factors.first());
		    macroRNA2reads.put(m, (double)ABIReads.getReads(chr, getStart(), getEnd(), macroRNAreads).length/macroRNAreads.length);
		}
	    }
	}
	return macroRNA2reads;
    }

    public LncRNA getUniqueLncRNA() {
	if (numNeighbors()==1 && getNeighbors().iterator().next() instanceof LncRNA) 
	    return (LncRNA)getNeighbors().iterator().next();
	Iterator it = getNeighbors().iterator();
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof LncRNA) {
		LncRNA l = (LncRNA)v;
		TreeSet<String> factors = l.getFactors();
		if (factors.size()>1) {
		    System.err.println("Error! found more than one factor!");
		    break;
		}
		ABIRead[] lncRNAReads = l.getFactorReads(factors.first());
		if ((double)ABIReads.getReads(chr, getStart(), getEnd(), lncRNAReads).length/lncRNAReads.length>GeneTranscriptsGraph.MIN_READS_PROPORTION)
		    return l;
	    }
	}
	return null;
    }

    public MacroRNA getUniqueMacroRNA() {
	if (numNeighbors()==1 && getNeighbors().iterator().next() instanceof MacroRNA) 
	    return (MacroRNA)getNeighbors().iterator().next();
	Iterator it = getNeighbors().iterator();
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof MacroRNA) {
		MacroRNA m = (MacroRNA)v;
		if (m.getStrand()==forward) {
		    TreeSet<String> factors = m.getFactors();
		    if (factors.size()>1) {
			System.err.println("Error! found more than one factor!");
			break;
		    }
		    ABIRead[] macroRNAreads = m.getFactorReads(factors.first());
		    if ((double)ABIReads.getReads(chr, getStart(), getEnd(), macroRNAreads).length/macroRNAreads.length>GeneTranscriptsGraph.MIN_READS_PROPORTION)
			return m;
		}
	    }
	}
	return null;
    }

    public MiRNA getUniqueMiRNA() {
	if (numNeighbors()==1 && getNeighbors().iterator().next() instanceof MiRNA) 
	    return (MiRNA)getNeighbors().iterator().next();
	Iterator it = getNeighbors().iterator();
	while (it.hasNext()) {
	    UndirectedSparseVertex v = (UndirectedSparseVertex)it.next();
	    if (v instanceof MiRNA) 
		return (MiRNA)v;
	}
	return null;
    }

    public boolean isNonCodingAnnotated() {
	return type.equals(NON_CODING_ANNOTATED_STR) || type.equals(LNC_RNA_STR) || type.equals(MACRO_RNA_STR);
    }

    public boolean isCodingAnnotated() {
	return type.equals(CODING_STR) || type.equals(CODING_FRAGMENT_STR) || type.equals(MULTI_GENE_STR);
    }


    //////////////////////////////////////////////////////////////////

    public int compareTo(Object s) {
	if (!(s instanceof TranscribedRegion2)) {
	    System.err.println("Error comparing with TranscribedRegion " + s.toString());
	    return 0;
	}
	TranscribedRegion2 t = (TranscribedRegion2)s;
	if (chr<t.chr)
	    return -1;
	else if (chr>t.chr)
	    return 1;
	else {
	    //Sort them in descending order on the negative strand
	    if (!t.forward && !forward) {
		if (-end<-t.end)
		    return -1;
		else if (-end>-t.end)
		    return 1;
		else {
		    if (-start<-t.start)
			return -1;
		    else if (-start>-t.start)
			return 1;
		    else
			return 0;
		}
	    }
	    else {
		//Both on positive strand or on different strands
		if (start<t.start)
		    return -1;
		else if (start>t.start)
		    return 1;
		else {
		    if (end<t.end)
			return -1;
		    else if (end>t.end)
			return 1;
		    else
			return 0;
		}
	    }
	}
    }

    public boolean equals(Object s) {
	if (!(s instanceof TranscribedRegion2)) {
	    System.err.println("Error comparing with TranscribedRegion " + s.toString());
	    return false;
	}
	TranscribedRegion2 t = (TranscribedRegion2)s;
	return start==t.start && end==t.end && level==t.level && chip==t.chip && chr==t.chr;
    }

    public String toString() {
	return "chr" + chr + ":" + start + "-" + end + "\t" + (end-start) + "\t" + format.format(level) + (type==null ? "" : "\t" + type); // + "\t" + format.format(errorReduction) + "\t" + format.format(logLike) + "\t" + format.format(logLike0);
    }

    public int hashCode() {
	return start*7 + end*13 + chr*31 + nReads*29;
    }

}