//Use this class to represent transcribed regions that are identified
//either by transcript calling or from the annotation. In particular,
//we want to use the class to merge results from the two approaches
//and keeping track of all the reads by assigning them to one unique
//transcribed region.

import java.util.*;
import java.text.*;

class TranscribedRegionMerge implements Comparable {

    //If we can assign the region to a gene, this is the one
    protected RefSeqUCSCEnsemblGene gene = null;
    protected MacroRNA macroRNA = null;
    protected LncRNA lncRNA = null;
    //The reads assigned to this region
    protected ABIRead[] reads = new ABIRead[0];
    protected int chr, start, end;
    protected boolean strand, conservedStart;
    protected String type;

    public static DecimalFormat format = new DecimalFormat("###.#######");
    public static Random random = new Random();

    public TranscribedRegionMerge(RefSeqUCSCEnsemblGene g) {
	gene = g;
	chr = g.getChromosome();
	strand = g.getStrand();
	start = g.getStrand() ? g.getStart() : g.getEnd();
	end = !g.getStrand() ? g.getStart() : g.getEnd();
    }

    public TranscribedRegionMerge(MacroRNA m) {
	macroRNA = m;
	chr = m.getChromosome();
	strand = m.getStrand();
	start = m.getStart(); //m.getStrand() ? m.getStart() : m.getEnd();
	end = m.getEnd(); //!m.getStrand() ? m.getStart() : m.getEnd();
    }

    public TranscribedRegionMerge(LncRNA m, boolean st) {
	lncRNA = m;
	chr = m.getChromosome();
	strand = st;
	start = m.getStart();
	end = m.getEnd();
    }

    public TranscribedRegionMerge(int c, boolean st, int s, int e) {
	chr = c;
	strand = st;
	start = s;
	end = e;
    }

    public TranscribedRegionMerge(TranscribedRegion2 t2) {
	chr = t2.getChromosome();
	strand = t2.getStrand();
	start = t2.getStart();
	end = t2.getEnd();
    }

    ///////////////////////////////////////////////////////

    public void assignReads(ABIRead[] rnaReads) {
	int startInd = Arrays.binarySearch(rnaReads, new ABIRead(chr, strand, start));
	startInd = startInd<0 ? -startInd-1 : startInd;
	int endInd = Arrays.binarySearch(rnaReads, new ABIRead(chr, strand, end));
	endInd = endInd<0 ? -endInd-1 : endInd;
	ABIRead[] rs = new ABIRead[endInd - startInd];
	for (int i=startInd; i<endInd; i++)
	    rs[i-startInd] = rnaReads[i];
	if (reads.length==0)
	    reads = rs;
	else { //Need to merge the two arrays
	    ABIRead[] tmp = new ABIRead[rs.length + reads.length];
	    for (int i=0; i<reads.length; i++) 
		tmp[i] = reads[i];
	    for (int i=0; i<rs.length; i++) 
		tmp[reads.length + i] = rs[i];
	    Arrays.sort(tmp);
	    reads = tmp;
	}
    }

    public double getReadDensity() { return (double)reads.length/getLength(); }

    public int getLength() { return end - start; }

    public int getStart() { return start; }
    
    public int getEnd() { return end; }

    public int getChromosome() { return chr; }

    public boolean getStrand() { return strand; }

    public void setConservedStart(boolean c) { conservedStart = c; }
    
    public boolean isConservedStart() { return conservedStart; }

    public ABIRead[] getReads() { return reads; }

    public RefSeqUCSCEnsemblGene getGene() { return gene; }

    public boolean merge(TreeSet<TranscribedRegion2> t2s) {
	boolean merged = false;
	for (TranscribedRegion2 t2 : t2s)
	    merged = merged || merge(t2);
	return merged;
    }

    public boolean merge(TranscribedRegion2 t2) {
	int oldStart = start, oldEnd = end;
	if (chr==t2.getChromosome() && strand==t2.getStrand()) {
	    start = Math.min(t2.getStart(), start);
	    end = Math.max(t2.getEnd(), end);
	}
	return oldStart!=start || oldEnd!=end;
    }

    public TreeSet<TranscribedRegion2> findOverlappingTranscript(TreeSet<TranscribedRegion2> segments) {
	TreeSet<TranscribedRegion2> overlapping = new TreeSet<TranscribedRegion2>();
	for (TranscribedRegion2 t2 : segments) {
	    if (overlaps(t2.getChromosome(), t2.getStrand(), t2.getStart(), t2.getEnd())) {
		//System.out.println(t2.toString() + " overlaps " + toString());
		overlapping.add(t2);
	    }
	}
	return overlapping;
    }

    public boolean overlaps(int c, boolean st, int s, int e) {
	if (c==chr && st==strand) {
	    if ((start<s && end>s) || (start<e && end>e) || (s<start && e>end))
		return true;
	}
	return false;
    }
    
    //////////////////////////////////////////////////////////////////

    public int compareTo(Object s) {
	if (!(s instanceof TranscribedRegionMerge)) {
	    System.err.println("Error comparing with TranscribedRegionMerge " + s.toString());
	    return 0;
	}
	TranscribedRegionMerge t = (TranscribedRegionMerge)s;
	if (chr<t.chr)
	    return -1;
	else if (chr>t.chr)
	    return 1;
	else {
	    if (start<t.start)
		return -1;
	    else if (start>t.start)
		return 1;
	    else {
		if (end<t.end)
		    return -1;
		else if (end>t.end)
		    return 1;
		else
		    return 0;
	    }
	}
    }

    public boolean equals(Object s) {
	if (!(s instanceof TranscribedRegionMerge)) {
	    System.err.println("Error comparing with TranscribedRegionMerge " + s.toString());
	    return false;
	}
	TranscribedRegionMerge t = (TranscribedRegionMerge)s;
	return start==t.start && end==t.end && strand==t.strand && chr==t.chr;
    }

    public String toString() {
	return "chr" + chr + ":" + start + "-" + end + "\t" + (end-start) + "\t" + format.format(getReadDensity());
    }

    public int hashCode() {
	return start*7 + end*13 + chr*31;
    }


}