//Use this class to represent a cluster of transcripts.

import java.io.*;
import java.util.*;
import java.text.DecimalFormat;
import edu.uci.ics.jung.graph.impl.*;


public class TranscriptCluster implements Comparable{

    protected TreeSet<TranscribedRegion> transcripts = new TreeSet<TranscribedRegion>();
    protected int start, end, chr, maxDist;



    ///////////////////////////////////////////////////////

    public TranscriptCluster(TranscribedRegion t) {
	chr = t.getChromosome();
	start = t.getStart();
	end = t.getEnd();
	transcripts.add(t);
    }

    public TranscriptCluster(TranscribedRegion t, int d) {
	chr = t.getChromosome();
	start = t.getStart();
	end = t.getEnd();
	transcripts.add(t);
	maxDist = d;
    }

    ///////////////////////////////////////////////////////

    public void setMaxDistance(int d) { maxDist = d; }

    public boolean add(TranscribedRegion t) {
	//System.out.println(t.toString() + " overlaps " + toString() + "\t:" + overlapping(t));
	//System.out.println(t.toString() + " distance " + toString() + "\t:" + nearestDist(t));
	if ((nearestDist(t)>maxDist && !overlapping(t)) || chr!=t.getChromosome())
	    return false;
	//System.out.println("Adding " + t.toString() + " to " + toString());
	transcripts.add(t);
	start = Math.min(start, t.getStart());
	end = Math.max(end, t.getEnd());
	return true;
    }

    public boolean overlapping(TranscribedRegion t) { return overlapping(t.getStart()) && overlapping(t.getEnd()); }

    public boolean overlapping(int p) { return start<p && end>p; }

    public int nearestDist(TranscribedRegion t) { return Math.min(nearestDist(t.getStart()), nearestDist(t.getEnd())); }

    public int nearestDist(int p) { return Math.min(Math.abs(start - p), Math.abs(end - p)); }

    public int[] head2headDistances() {
	//Calculate the head-to-head distances for all transcripts in
	//the cluster.
	Vector<Integer> dists = new Vector<Integer>();
	for (TranscribedRegion t : transcripts) {
	    //Only consider transcripts that are not overlapping
	    int minDist = Integer.MAX_VALUE;
	    for (TranscribedRegion t2 : transcripts) {
		if (!t.equals(t2) && t.getStrand()!=t2.getStrand() && !t.overlaps(t2) && !t2.overlaps(t)) {
		    if (t.getStrand() && t.getStart()>t2.getEnd())
			minDist = Math.min(minDist, Math.abs(t.getStart() - t2.getEnd()));
		    else if (!t.getStrand() && t.getEnd()<t2.getStart())
			minDist = Math.min(minDist, Math.abs(t2.getStart() - t.getEnd()));
		}
	    }
	    if (minDist<Integer.MAX_VALUE)
		dists.addElement(minDist);
	}
	int ds[] = new int[dists.size()], i = 0;
	for (Integer d : dists)
	    ds[i++] = d;
	return ds;
    }

    public int[] tail2tailDistances() {
	Vector<Integer> dists = new Vector<Integer>();
	for (TranscribedRegion t : transcripts) {
	    //Only consider transcripts that are not overlapping
	    int minDist = Integer.MAX_VALUE;
	    for (TranscribedRegion t2 : transcripts) {
		if (!t.equals(t2) && t.getStrand()!=t2.getStrand() && !t.overlaps(t2) && !t2.overlaps(t)) {
		    if (t.getStrand() && t.getEnd()<t2.getStart())
			minDist = Math.min(minDist, Math.abs(t.getEnd() - t2.getStart()));
		    else if (!t.getStrand() && t.getStart()>t2.getEnd())
			minDist = Math.min(minDist, Math.abs(t.getStart() - t2.getEnd()));
		}
	    }
	    if (minDist<Integer.MAX_VALUE)
		dists.addElement(minDist);
	}
	int ds[] = new int[dists.size()], i = 0;
	for (Integer d : dists)
	    ds[i++] = d;
	return ds;
    }

    public int[] tail2headDistances() {
	//Find the distance to the nearest transcript downstream
	Vector<Integer> dists = new Vector<Integer>();
	for (TranscribedRegion t : transcripts) {
	    //Only consider transcripts that are not fully overlapping
	    int minDist = Integer.MAX_VALUE;
	    for (TranscribedRegion t2 : transcripts) {
		if (!t.equals(t2) && t.getStrand()==t2.getStrand()) {
		    if (t.getStrand() && t.getEnd()<t2.getStart())
			minDist = Math.min(minDist, Math.abs(t.getEnd() - t2.getStart()));
		    else if (!t.getStrand() && t.getStart()>t2.getEnd())
			minDist = Math.min(minDist, Math.abs(t.getStart() - t2.getEnd()));
		}
	    }
	    if (minDist<Integer.MAX_VALUE)
		dists.addElement(minDist);
	}
	int ds[] = new int[dists.size()], i = 0;
	for (Integer d : dists)
	    ds[i++] = d;
	return ds;
    }
    
    ////////////////////////////////////////////

    public int number5pOverlap() {
	//Count how many transcripts that have another one overlapping
	//their 5'-end. Exclude those transcripts where there is full overlap.
	int nOverlap = 0;
	for (TranscribedRegion t : transcripts) {
	    for (TranscribedRegion t2 : transcripts) {
		if (!t.equals(t2) && t.getStrand()!=t2.getStrand() && t.overlaps(t2)) {
		    
		}
	    }
	}
	return nOverlap;
    }

    ////////////////////////////////////////////

    public TreeSet<TranscribedRegion> getTranscripts() { return transcripts; }

    public boolean contains(TranscribedRegion t) { return transcripts.contains(t); }

    //////////////////////////////////////////////////////////////////

    public String toString() {
	return "chr" + chr + ":" + start + "-" + end + "(" + transcripts.size() + ")";
    }

    public int compareTo(Object s) {
	if (!(s instanceof TranscriptCluster)) {
	    System.err.println("Error comparing with TranscriptCluster " + s.toString());
	    return 0;
	}
	TranscriptCluster t = (TranscriptCluster)s;
	if (chr<t.chr)
	    return -1;
	else if (chr>t.chr)
	    return 1;
	else {
	    if (start<t.start)
		return -1;
	    else if (start>t.start)
		return 1;
	    else {
		if (end<t.end)
		    return -1;
		else if (end>t.end)
		    return 1;
		else
		    return 0;
	    }
	}
    }

    public boolean equals(Object s) {
	if (!(s instanceof TranscriptCluster)) {
	    System.err.println("Error comparing with TranscriptCluster " + s.toString());
	    return false;
	}
	TranscriptCluster t = (TranscriptCluster)s;
	return start==t.start && end==t.end && chr==t.chr;
    }

}