//Use this class to keep track of all the reads for a given chromosome
//and experiment. There are also methods for grouping them into
//tentative transcripts and assessing the quality of the partition.

import java.io.*;
import java.util.*;
import java.text.DecimalFormat;
import umontreal.iro.lecuyer.probdist.*;
import umontreal.iro.lecuyer.stat.Tally;
import umontreal.iro.lecuyer.stat.TallyStore;
import org.apache.commons.math.util.MathUtils;

//import flanagan.integration.Integration;
//import flanagan.integration.IntegralFunction;



public class TranscriptFinder {

    public static final int MIN_READS_BLOCK = 50;
    public static final int MAX_ITERATIONS = 50;

    public static final String GAUSSIAN_ERROR = "gauss";
    public static final String POISSON_ERROR = "poisson";

    protected ABIRead reads[];
    protected int readsInd = 0, genomeInd = 0;
    protected int chr; 
    protected boolean log = true;
    private int maxReadLength = 35;
    private boolean forward = true;
    protected double sigmaHat[];
    protected Tally coeffsTally[];
    protected String species;
    //Length-scale -> location -> coeff-val
    protected TreeMap<Integer, TreeMap<Integer, Double>> localMax = new TreeMap<Integer, TreeMap<Integer, Double>>();
    //Keep track of all the local max, this will speed up calculations
    //since we can avoid recomputing the coefficients
    


    //Keep track of the location of break points and the average
    //expression levels. Make sure to use log-read counts for stability.
    protected ReducedModel reducedModel = null;

    protected int readGroupInd = 0;

    public static DecimalFormat format = new DecimalFormat("###.####");
    public static Random random = new Random();

    //////////////////////////////////////////////////////////////////

    public TranscriptFinder(String fName, int c, boolean strand, boolean flatten, int rLength, String sp, boolean l) {
	species = sp;
	//reads = ABIReads.loadMRNAReadsChr(fName, strand, c, maxReadLength, flatten);
	chr = c;
	maxReadLength = rLength;
	forward = strand;
	/*TreeMap<Integer, Integer> hist = ABIReads.calculateReadLengthDistribution(reads);
	if (hist.size()>1) {
	    System.out.println("Mean read length is " + format.format(ABIReads.meanReadLength(reads)) + " and the distribution is:");
	    for (Integer k : hist.keySet()) 
		System.out.println(k + "\t" + format.format((double)hist.get(k)/reads.length));
		}*/
	log = l; //!flatten;
    }

    public TranscriptFinder(ABIRead[] rs, int c, boolean strand, int rLength, String sp, boolean l) {
	species = sp;
	reads = rs;
	chr = c;
	maxReadLength = rLength;
	forward = strand;
	/*TreeMap<Integer, Integer> hist = ABIReads.calculateReadLengthDistribution(reads);
	if (hist.size()>1) {
	    System.out.println("Mean read length is " + format.format(ABIReads.meanReadLength(reads)) + " and the distribution is:");
	    for (Integer k : hist.keySet()) 
		System.out.println(k + "\t" + format.format((double)hist.get(k)/reads.length));
		}*/
	log = l; //!flatten;
    }

    protected TranscriptFinder() { }

    //////////////////////////////////////////////////////////////////


    public void removeHaarCoefficients() {
	readGroupInd = 0;
	localMax = new TreeMap<Integer, TreeMap<Integer, Double>>();
	coeffsTally = null;
	readsInd = 0;
	genomeInd = 0;
    }

    public void calculateHaarCoefficients(int lVec[], int binSize, boolean chip) { calculateHaarCoefficients(lVec, binSize, chip, null); }

    public void calculateHaarCoefficients(int lVec[], int binSize, boolean chip, String saveInfoFileName) { calculateHaarCoefficients(lVec, binSize, null); }

    public void calculateHaarCoefficients(int lVec[], int binSize, String saveInfoFileName) {
	/*if (species.equals("sc"))
	    calculateHaarCoefficientsAll(lVec, binSize, saveInfoFileName);
	    else*/
	    calculateHaarCoefficientsBlock(lVec, binSize, saveInfoFileName);
    }
	
    public void calculateHaarCoefficientsBuffer(int lVec[]) {
	//pre-compute the normalization coefficients for the Haar wavelets
	double normCoeffs[] = new double[lVec.length];
	for (int i=0; i<lVec.length; i++)
	    normCoeffs[i] = 1/Math.sqrt(Math.pow(2, lVec[i] + 1));
	TreeMap<Double, Integer> coeffsShortestHist = new TreeMap<Double, Integer>();
	for (int i=0; i<lVec.length; i++)
	    localMax.put(lVec[i], new TreeMap<Integer, Double>());
	int winLen = (int)Math.pow(2, lVec[lVec.length - 1]);
	double readsHist[] = calculateHistogram(0, winLen);
	double w[] = new double[lVec.length], wp[] = new double[lVec.length], wpp[] = new double[lVec.length];
	//Calculate the first two coefficients
	for (int i=0; i<lVec.length; i++) 
	    wpp[i] = calculateHaarCoefficient(lVec[i], 0, readsHist, i==0 ? 0 : wpp[i-1], log, normCoeffs[i]);
	readsHist = shiftHistogram(readsHist, ++genomeInd + winLen, 2*winLen);
	for (int i=0; i<lVec.length; i++) 
	    wp[i] = calculateHaarCoefficientInc(lVec[i], 1, readsHist, wpp[i], log, normCoeffs[i]);
	readsHist = shiftHistogram(readsHist, ++genomeInd + winLen, 2*winLen);
	while (genomeInd<numberOfBasePairs(chr, species)) {
	    for (int i=0; i<lVec.length; i++) { //Find coefficients for each subband
		w[i] = calculateHaarCoefficientInc(lVec[i], Math.min(genomeInd, readsHist.length/2), readsHist, wp[i], log, normCoeffs[i]);
		//Check if the previous coeff was a local maximum
		if (i<lVec.length-1 && isLocalExtrema(w, i)) 
		    localMax.get(lVec[i]).put(genomeInd, wp[i]);
		//shift the coefficients
		wpp[i] = wp[i];
		wp[i] = w[i];
	    }
	    coeffsShortestHist.put(Math.abs(w[0]), coeffsShortestHist.containsKey(Math.abs(w[0])) ? coeffsShortestHist.get(Math.abs(w[0])) + 1 : 1);
	    readsHist = shiftHistogram(readsHist, ++genomeInd + winLen, 2*winLen);
	    //System.out.println(Arrays.toString(readsHist));
	}
	//Calculate the FDR by considering all the localMax for each band
	//double sigmaHat = estimateStdDev(coeffsShortest);
	sigmaHat[0] = estimateStdDev(coeffsShortestHist);
	//System.out.println("sigmaHat = " + format.format(sigmaHat));
    }

    public void calculateHaarCoefficientsBlock(int lVec[], int binSize) { calculateHaarCoefficientsBlock(lVec, binSize, null); }

    public void calculateHaarCoefficientsBlock(int lVec[], int binSize, String saveInfoFileName) {
	//pre-compute the normalization coefficients for the Haar wavelets
	double normCoeffs[] = new double[lVec.length];
	for (int i=0; i<lVec.length; i++)
	    normCoeffs[i] = 1/Math.sqrt(Math.pow(2, lVec[i] + 1));
	//Since the variability in terms of number of reads for a
	//given locus (and hence the magnitude of the coefficients) is
	//much smaller for ChIP than for RNA-Seq, we no longer need to
	//create bins for the histogram.
	coeffsTally = new Tally[lVec.length];
	sigmaHat = new double[lVec.length];
	for (int i=0; i<lVec.length; i++) {
	    localMax.put(lVec[i], new TreeMap<Integer, Double>());
	    coeffsTally[i] = new Tally("coeffs " + lVec[i]);
	}
	try {
	    BufferedWriter bw = saveInfoFileName!=null ? new BufferedWriter(new FileWriter(saveInfoFileName + "_coeffs_" + chr + ".bed")) : null;
	    int winLen = (int)Math.pow(2, lVec[lVec.length - 1]);
	    while (moreReadGroups()/* && reads[readGroupInd].getPosition()<3500000*/) {
		ABIRead[] rs = nextReadGroup(2*winLen);
		double readsHist[] = calculateHistogram(rs, maxReadLength, winLen, binSize, saveInfoFileName);
		//System.out.println("Found block with length " + readsHist.length + ", containing " + rs.length + " reads starting at " + (rs[0].getPosition() - maxReadLength));
		double wp = 0;
		for (int i=0; i<lVec.length; i++) { //Find coefficients for each subband
		    double w[] = new double[readsHist.length + 2*winLen];
		    //Calculate the first coefficient
		    w[0] = i==0 ? calculateHaarCoefficient(lVec[i] - binSize, 0, readsHist, log, normCoeffs[i]) : calculateHaarCoefficient(lVec[i] - binSize, 0, readsHist, wp, log, normCoeffs[i]);
		    for (int j=1; j<readsHist.length; j++) //subsequent coefficients can be calculated incrementally and hence significantly faster
			w[j] = calculateHaarCoefficientInc(lVec[i] - binSize, j, readsHist, w[j-1], log, normCoeffs[i]);
		    localMax.get(lVec[i]).putAll(findLocalMax(w, rs[0].getPosition() - Math.min(winLen, rs[0].getPosition()), binSize));
		    wp = w[0]; //to speed up calculation of first coeff at the next length scale
		    int b = (int)Math.pow(2, binSize);
		    for (int j=0; j<w.length; j++) {
			if (j>0 && j<w.length-1 && isLocalExtrema(w, j))
			    coeffsTally[i].add(w[j]);
			if (bw!=null && i==0 && (rs[0].getPosition() - winLen/b + (j + 1)*b)<PeakFinder2.CHROMOSOME_LENGTH[chr-1]) 
			    bw.write("chr" + chr + "\t" + (rs[0].getPosition() - winLen + j*b) + "\t" + (rs[0].getPosition() - winLen + (j + 1)*b) + "\t" + format.format(w[j]) + "\n"); 
		    }
		}
	    }
	    for (int i=0; i<coeffsTally.length; i++) {
		sigmaHat[i] = coeffsTally[i].numberObs()>1 ? coeffsTally[i].standardDeviation() : 0;
		//System.out.println("sigmaHat = " + format.format(coeffsTally[i].standardDeviation()) + " for mean " + format.format(coeffsTally[i].average()) + " from " + coeffsTally[i].numberObs());
	    }
	    if (bw!=null) { 
		bw.close(); 
		String name = saveInfoFileName.substring(saveInfoFileName.lastIndexOf("/") + 1) + "_coeffs";
		String colorStr = CreateWigFiles.getColorString(name);
		int priority = CreateWigFiles.getPriority(name);
		String trackInfoStr = "track type=bigWig name=\"" + name + "_" + chr + "\" description=\"" + name + "_" + chr + "\" priority=" + priority + " visibility=dense autoScale=off maxHeightPixels=10:25:70 viewLimits=0:2 color=" + colorStr + " smoothingWindow=off bigDataUrl=" + FindTranscripts.URL_NAME + name + "_" + chr + ".bw";
		CreateWigFiles.addTrackInfo(trackInfoStr, FindTranscripts.TRACK_INFO_FILE_NAME);
		//CreateWigFiles.createWWWLink(saveInfoFileName + "_coeffs_" + chr + ".bed", "/www/" + FindTranscripts.WWW_DIR_NAME, species);
	    }
	}
	catch (IOException e) { 

	}
    }
    
    public void calculateHaarCoefficientsAll(int lVec[], int binSize) { calculateHaarCoefficientsAll(lVec, binSize, null); }

    public void calculateHaarCoefficientsAll(int lVec[], int binSize, String saveInfoFileName) {
	//pre-compute the normalization coefficients for the Haar wavelets
	double normCoeffs[] = new double[lVec.length];
	Tally coeffsTally[] = new Tally[lVec.length];
	sigmaHat = new double[lVec.length];	
	for (int i=0; i<lVec.length; i++) {
	    normCoeffs[i] = 1/Math.sqrt(Math.pow(2, lVec[i] + 1));
	    localMax.put(lVec[i], new TreeMap<Integer, Double>());
	    coeffsTally[i] = new Tally();
	}
	int winLen = (int)Math.pow(2, lVec[lVec.length - 1]);
	double readsHist[] = calculateHistogram(reads, maxReadLength, winLen, binSize);
	System.out.println("Found histogram with length " + readsHist.length + " starting at " + reads[0].getPosition());
	double wp = 0;
	for (int i=0; i<lVec.length; i++) { //Find coefficients for each subband
	    double w[] = new double[readsHist.length + 2*winLen];
	    //Calculate the first coefficient
	    w[0] = i==0 ? calculateHaarCoefficient(lVec[i], 0, readsHist, log, normCoeffs[i]) : calculateHaarCoefficient(lVec[i], 0, readsHist, wp, log, normCoeffs[i]);
	    for (int j=1; j<readsHist.length; j++)  //subsequent coefficients can be calculated incrementally and hence significantly faster
		w[j] = calculateHaarCoefficientInc(lVec[i], j, readsHist, w[j], log, normCoeffs[i]);
	    //w[j] = calculateHaarCoefficientInc(lVec[i], j, readsHist, w[j-1], normCoeffs[i]);
	    //We do not need to store all coefficients. We can
	    //save memory by only saving the ones for the
	    //shortest scale (needed for estimating sigmaHat)
	    //and the local maxima with corresponding
	    //locations (the only ones that may get selected
	    //later on).
	    localMax.get(lVec[i]).putAll(findLocalMax(w, reads[0].getPosition() - winLen, binSize));
	    wp = w[0]; //to speed up calculation of first coeff at the next length scale
	    for (int j=0; j<w.length; j++)
		coeffsTally[i].add(w[j]);
	}
	for (int i=0; i<coeffsTally.length; i++) 
	    sigmaHat[i] = coeffsTally[i].numberObs()>1 ? coeffsTally[i].standardDeviation() : 0; 
	//System.out.println("sigmaHat = " + format.format(sigmaHat));
    }

    ///////////////////////////////////////////////////////

    public int findRefinedReducedModel(int lMin, int lMax, int binSize, double fdr, boolean mergeSegments, boolean gmm, String saveInfoFileName, RefSeqGenes geneInfo, Vector<Peak> breakPoints, Vector<Peak> peaks, int maxDivergentTSSDistance, String costFunction) { return findRefinedReducedModel(lMin, lMax, binSize, fdr, mergeSegments, gmm, saveInfoFileName, geneInfo, null, null, null, null, breakPoints, peaks, maxDivergentTSSDistance, costFunction, -1); }

    public int findRefinedReducedModel(int lMin, int lMax, int binSize, double fdr, boolean mergeSegments, boolean gmm, String saveInfoFileName, RefSeqGenes geneInfo, Vector<Peak> breakPoints, Vector<Peak> peaks, int maxDivergentTSSDistance, String costFunction, int nIterations) { return findRefinedReducedModel(lMin, lMax, binSize, fdr, mergeSegments, gmm, saveInfoFileName, geneInfo, null, null, null, null, breakPoints, peaks, maxDivergentTSSDistance, costFunction, nIterations); }

    public int findRefinedReducedModel(int lMin, int lMax, int binSize, double fdr, boolean mergeSegments, boolean gmm, String saveInfoFileName, RefSeqGenes geneInfo, TreeSet<LncRNA> lncRNAs, TreeSet<MiRNA> miRNAs, TreeSet<MacroRNA> macroRNAs, TreeSet<Enhancer2> enhancers, Vector<Peak> breakPoints, Vector<Peak> peaks, int maxDivergentTSSDistance, String costFunction) { return findRefinedReducedModel(lMin, lMax, binSize, fdr, mergeSegments, gmm, saveInfoFileName, geneInfo, lncRNAs, miRNAs, macroRNAs, enhancers, breakPoints, peaks, maxDivergentTSSDistance, costFunction, -1); }

    public int findRefinedReducedModel(int lMin, int lMax, int binSize, double fdr, boolean mergeSegments, boolean gmm, String saveInfoFileName, RefSeqGenes geneInfo, TreeSet<LncRNA> lncRNAs, TreeSet<MiRNA> miRNAs, TreeSet<MacroRNA> macroRNAs, TreeSet<Enhancer2> enhancers, Vector<Peak> breakPointPeaks, Vector<Peak> peaks, int maxDivergentTSSDistance, String costFunction, int nIterationsMax) {
	//Use this method to find the best model, by applying the
	//wavelet method iteratively
	double cost = reducedModel==null ? 0 : 0/*OptimalTranscriptParameters.evaluateCostFunction(costFunction, reducedModel, geneInfo.getGeneIDsKeep(forward), peaks, maxDivergentTSSDistance, !forward)*/, prevCost = 0;
	if (cost>=prevCost) {
	    int[] lVec = FindTranscripts.getLengthVec(lMin, lMax, 1);
	    Vector<Integer> breakPoints = findBreakpoints(lVec, fdr, null, breakPointPeaks);
	    if (breakPoints.size()>0) {
		reducedModel = findReducedModel(reducedModel, 0, lVec, fdr, breakPoints, mergeSegments, gmm, binSize, saveInfoFileName);
		//reducedModel.refineTranscribedRegionBoundaries(getReads(), binSize, lMin, maxReadLength, log);
		//reducedModel.compare2Annotation(maxDivergentTSSDistance, geneInfo, lncRNAs, miRNAs, macroRNAs, enhancers);
		//reducedModel.categorizeTranscribedRegions(lncRNAs!=null);
		//Find out how many of the transcripts have a peak near their start
		TreeMap<TranscribedRegion, Peak> transcriptsWithPeaks = peaks==null ? new TreeMap<TranscribedRegion, Peak>() : null;//reducedModel.compare2Peaks(peaks, maxDivergentTSSDistance);
		//FindTranscripts.printStats(fdr, reducedModel, geneIDsKeepS, transcriptsWithPeaks, peaks);
		//Iterative refinement
		//cost = OptimalTranscriptParameters.evaluateCostFunction(costFunction, reducedModel, geneInfo.getGeneIDsKeep(forward), peaks, maxDivergentTSSDistance, !forward);
		int nUnique = reducedModel.numberOfTranscriptsWithOneGene(), prevUnique = 0, nIterations = 1, lmi = lMin, lma = lMax;
		//reducedModel.removeAll();
		while (nIterations<MAX_ITERATIONS && (nIterations<nIterationsMax || nIterationsMax<0)) { //Try again to see if we can find more interesting stuff the second time aroun
		    //fdr /= 2; //increase the FDR to be more stringent
		    nIterations++;
		    int lVecMore[] = FindTranscripts.getLengthVec(lmi, lma, 1);
		    breakPoints = findBreakpoints(lVecMore, fdr, reducedModel, peaks);
		    ReducedModel r = null; //findReducedModel((ReducedModel)reducedModel.clone(), nIterations, lVecMore, fdr, breakPoints, mergeSegments, gmm, binSize, saveInfoFileName);
		    //r.removeShortSegments();
		    //r.refineTranscribedRegionBoundaries(getReads(), binSize, lmi, maxReadLength, log);
		    //r.compare2Annotation(maxDivergentTSSDistance, geneInfo, lncRNAs, miRNAs, macroRNAs, enhancers);
		    //r.categorizeTranscribedRegions(lncRNAs!=null);
		    //FindTranscripts.printTypeStats(r, getReads().length);
		    //Find out how many of the transcripts have a peak near their start
		    //transcriptsWithPeaks = peaks==null ? new TreeMap<TranscribedRegion, Peak>() : r.compare2Peaks(peaks, maxDivergentTSSDistance);
		    //FindTranscripts.printStats(fdr, r, geneIDsKeepS, transcriptsWithPeaks, peaks);
		    prevCost = cost;
		    //cost = OptimalTranscriptParameters.evaluateCostFunction(costFunction, r, geneInfo.getGeneIDsKeep(forward), peaks, maxDivergentTSSDistance, !forward);
		    //r.removeAll();
		    //comparing to the annotation is a little bit cheating
		    //and hence we need a criterion that does not rely on
		    //the annotation. We can either have a fixed boundary
		    //(which we are allowed to optimize based on the
		    //annotation of a single chromosome) or we could 
		    System.out.println("Cost is " + format.format(cost) + " and previous round had " + format.format(prevCost));
		    if (prevCost<cost || nIterationsMax>0) 
			reducedModel = r;
		    else 
			break;
		}
		return nIterations;
	    }
	}
	return 0;
    }

    public ReducedModel getReducedModel() { return reducedModel; }

    ///////////////////////////////////////////////////////

    public Vector<Integer> findBreakpointsInput(int lVec[], double fdr) { return null; } //ugly solution
	

    public Vector<Integer> findBreakpoints(int lVec[], double fdr) { return findBreakpoints(lVec, fdr, null); }

    public Vector<Integer> findBreakpoints(int lVec[], double fdr, ReducedModel reducedModel) { return findBreakpoints(lVec, fdr, reducedModel, null); }
	
    public Vector<Integer> findBreakpoints(int lVec[], double fdr, ReducedModel reducedModel, Vector<Peak> peaks) {
	//Keep track of breakpoints stemming from local maxima and
	//local minima separately. In particular, we want to allow for
	//the situation where breakpoints are placed closed to one
	//another if they correspond to different types of
	//extrema. This should allow us to smooth over a longer length
	//scale and still be able to catch short transcribed regions.
	Vector<Integer> breakPoints = new Vector<Integer>(), breakPointsMax = new Vector<Integer>(), breakPointsMin = new Vector<Integer>();
	//for (int i=0; i<lVec.length; i++) { 
	for (int i=lVec.length-1; i>=0; i--) { 
	    //System.out.println("Found " + breakPoints.size() + " break points for scale " + lVec[i]);
	    if (reducedModel!=null) {
		//Remove the coefficients that are located in segments
		coeffsTally[i] = new Tally();
		if (localMax.containsKey(lVec[i])) {
		    for (Integer pos : localMax.get(lVec[i]).keySet()) {
			//if (!reducedModel.segmentOverlaps(pos))
			//  coeffsTally[i].add(localMax.get(lVec[i]).get(pos));
		    }
		}
	    }
	    //Invert the mapping so that we go from abs locMax -> position
	    TreeMap<Double, Integer> maxLoc = new TreeMap<Double, Integer>();
	    for (Integer pos : localMax.get(lVec[i]).keySet()) {
		//remove the break points that are sitting on top of regions that are already known to be segments
		/*if (reducedModel==null || !reducedModel.segmentOverlaps(pos)) {
		    double v = localMax.get(lVec[i]).get(pos);
		    if (!maxLoc.containsKey(v))
			maxLoc.put(v, pos);
		    else { //collision
			while (maxLoc.containsKey(v))
			    v += random.nextGaussian()*.00000001;
			maxLoc.put(v, pos);
		    }
		    }*/
	    }
	    if (coeffsTally[i].numberObs()>1) {
		double zThreshold = calculateThreshold(coeffsTally[i], fdr);
		double zThresholdUp = calculateThreshold(coeffsTally[i], fdr, true);
		double zThresholdLow = calculateThreshold(coeffsTally[i], fdr, false);
		//System.out.println("Threshold is " + format.format(zThresholdLow) + " and " + format.format(zThresholdUp) + " for " + lVec[i]);
		//determine which coefficients to keep, we want to avoid using overlapping ones.
		Vector<Double> maxs = new Vector<Double>(maxLoc.keySet());
		for (int j=maxs.size()-1; j>=0; j--) {
		    double m = maxs.elementAt(j);
		    int pos = maxLoc.get(m);
		    if (m>0 && m>zThresholdUp && !nearbyBreakPoint(breakPoints, reducedModel, pos, lVec[i])) {
			breakPoints.addElement(pos);
			//System.out.println("Found breakpoint " + pos + " with maxima " + format.format(m) + " at scale " + lVec[i]);
		    }
		    else if (m<0 && m<zThresholdLow && !nearbyBreakPoint(breakPoints, reducedModel, pos, lVec[i])) {
			breakPoints.addElement(pos);
			//System.out.println("Found breakpoint " + pos + " with minima " + format.format(m) + " at scale " + lVec[i]);
		    }
		}
	    }
	}
	if (reducedModel!=null) {
	    /*for (TranscribedRegion t : reducedModel.getSegments()) { //Add breakspoints corresponding to the start and end of each existing segment
		breakPoints.add(t.getStart());
		breakPoints.add(t.getEnd());
		}*/
	}
	if (peaks!=null) { //Finally, add breakpoints corresponding to Pol2 peaks
	    for (Peak p : peaks) {
		//if (!nearbyBreakPoint(breakPoints, reducedModel, p.getCenter(), lVec[0])) {
		//breakPoints.add(p.getStart());
		//breakPoints.add(p.getEnd());
		      breakPoints.add(p.getCenter() - 50);
		      breakPoints.add(p.getCenter() + 50);
		    //}
	    }
	    //System.out.println("Added break-points for " + peaks.size());
	}
	if (breakPoints.size()>0) {
	    breakPoints.setElementAt(0, 0);
	    Collections.sort(breakPoints);
	}
	//System.out.println("Found a total of " + breakPoints.size() + " break points.");
	return breakPoints;
    }

    public ReducedModel findReducedModel(boolean merge) { 
	//if (merge)
	//  reducedModel.merge(reads, maxReadLength, log);
	return reducedModel; 
    }

    public ReducedModel findReducedModel(int lVec[], double fdr, Vector<Integer> breakPoints, Vector<Integer> breakPointsInput, boolean mergeSegments, boolean gmm, int binSize) { return findReducedModel(lVec, fdr, breakPoints, mergeSegments, gmm, binSize, null); }

    public ReducedModel findReducedModel(int lVec[], double fdr, Vector<Integer> breakPoints, boolean mergeSegments, boolean gmm, int binSize) { return findReducedModel(lVec, fdr, breakPoints, mergeSegments, gmm, binSize, null); }

    public ReducedModel findReducedModel(int lVec[], double fdr, Vector<Integer> breakPoints, Vector<Integer> breakPointsInput, boolean mergeSegments, boolean gmm, int binSize, String saveInfoFileName) { return findReducedModel(lVec, fdr, breakPoints, mergeSegments, gmm, binSize, saveInfoFileName); }

    public ReducedModel findReducedModel(int lVec[], double fdr, Vector<Integer> breakPoints, boolean mergeSegments, boolean gmm, int binSize, String saveInfoFileName) {
	//Now we can reconstruct the signal by averaging between the
	//breakpoints. We need to add a L_1 penalty to make sure that
	//not all the segments get values.
	//We need to find out which density corresponds to a given FDR
	//for each length scale
	//if (gmm)
	    //reducedModel = new ReducedModel(reads, breakPoints, chr, maxReadLength, lVec, mergeSegments, log, binSize, species, false, fdr, saveInfoFileName);
	//else {	    
	//  double cutOffDens[] = calculateReadLevelCutOffs(lVec, fdr);
	//reducedModel = new ReducedModel(reads, breakPoints, chr, maxReadLength, lVec, //cutOffDens, mergeSegments, log, species, false);
	//}
	return null;//reducedModel;
    }

    public ReducedModel findReducedModel(ReducedModel reducedModel, int nIterations, int lVec[], double fdr, Vector<Integer> breakPoints, boolean mergeSegments, boolean gmm, int binSize, String saveInfoFileName) {
	//Use this method when we have a model which we would like to refine and find additional segments for.
	if (reducedModel==null)
	    return findReducedModel(lVec, fdr, breakPoints, mergeSegments, gmm, binSize, saveInfoFileName);
	//if (gmm)
	//  reducedModel.createMoreSegments(nIterations, reads, breakPoints, maxReadLength, lVec, mergeSegments, log, binSize, fdr, saveInfoFileName);
	//System.out.println(reducedModel.toString());
	return reducedModel;
    }

    public double[] calculateReadLevelCutOffs(int[] lVec, double fdr) {
	double cutOffs[] = new double[lVec.length];
	for (int i=0; i<lVec.length; i++) { 
	    TreeMap<Double, Integer> densHist = ABIReads.calculateReadDensity(chr, reads, (int)Math.pow(2, lVec[i]), ABIReads.meanReadLength(reads), maxReadLength, species, false, log);
	    TreeMap<Double, Integer> densHistVar = ABIReads.calculateReadDensityVariance(chr, reads, (int)Math.pow(2, lVec[i]), ABIReads.meanReadLength(reads), maxReadLength, species, false, log);
	    int totBins = numberOfBins((int)Math.pow(2, lVec[i])), nBins = 0;
	    Iterator it = densHist.keySet().iterator();
	    double keyDens = 0;
	    //System.out.println("There were " + totBins + " bins in total and threshold is " + format.format((1-fdr)*(double)totBins));
	    //for (Double k : densHistVar.keySet())
	    //System.out.println(format.format(k) + "\t" + format.format(densHistVar.get(k)));
	    while (nBins<(int)Math.round((1-fdr)*(double)totBins) && it.hasNext()) {
		keyDens = (Double)it.next();
		nBins += densHist.get(keyDens);
		//System.out.println(format.format(keyDens) + "\t" + densHist.get(keyDens));
	    }
	    //System.out.println("Cut-off density is " + format.format(keyDens) + " for length " + lVec[i] + " with " + nBins);
	    cutOffs[i] = keyDens;
	}
	return cutOffs;
    }

    public boolean nearbyBreakPoint(Vector<Integer> breakPoints, int pos, int l) { return nearbyBreakPoint(breakPoints, null, pos, l); }

    public boolean nearbyBreakPoint(Vector<Integer> breakPoints, ReducedModel reducedModel, int pos, int l) {
	for (Integer b : breakPoints) {
	    if (b + Math.pow(2, l)>pos && b - Math.pow(2, l)<pos)
		return true;
	}
	if (reducedModel!=null) {
	    /*for (TranscribedRegion t : reducedModel.getSegments()) {
		if ((t.getStart() + Math.pow(2, l)>pos && t.getStart() - Math.pow(2, l)<pos) || (t.getEnd() + Math.pow(2, l)>pos && t.getEnd() - Math.pow(2, l)<pos))
		    return true;
		    }*/
	}
	return false;
    }

    public int numberOfBins(int lengthScale) { return numberOfBins(lengthScale, chr, species); }

    public static int numberOfBins(int lengthScale, int chr, String species) {
	if (species.equals("mm"))
	    return RefSeqGenes.CHROMOSOME_LENGTH[chr - 1]/lengthScale;
	else if (species.equals("hs"))
	    return RefSeqGenes.CHROMOSOME_LENGTH_HS[chr - 1]/lengthScale;
	else if (species.equals("sc"))
	    return RefSeqGenes.CHROMOSOME_LENGTH_SC[chr - 1]/lengthScale;
	else if (species.equals("pf"))
	    return RefSeqGenes.CHROMOSOME_LENGTH_PF[chr - 1]/lengthScale;
	System.err.println("Warning, no genome length information for " + species);
	return 0;
    }

    public static int numberOfBasePairs(int chr, String species) { return numberOfBins(1, chr, species); }

    ////////////////////////////////////////////

    //This is a simpler version than the one in the original paper by
    //BenYaacov et al. Just invert the normal distribution specified
    //by sigmaHat and then find the cut-off required for the desired FDR.
    public static double calculateThreshold(Tally coeffsTally, double fdr, boolean upper) { return NormalDist.inverseF(coeffsTally.average(), coeffsTally.numberObs()>1 ? coeffsTally.standardDeviation() : 0, upper ? 1-fdr/2 : fdr/2); }

    //Assume that it is centered
    public static double calculateThreshold(Tally coeffsTally, double fdr) { return NormalDist.inverseF(0, coeffsTally.numberObs()>1 ? coeffsTally.standardDeviation() : 0, 1-fdr/2); }

    public static double calculateThreshold(double locMax[], double sigmaHat, double fdr) { return calculateThreshold(locMax, sigmaHat, fdr, null); }

    

    public static double calculateThreshold(double locMax[], double sigmaHat, double fdr, String saveFileName) {
	//Returns the threshold for which we need to go above for a
	//coefficient to be included.
	Arrays.sort(locMax); //ascending order
	double locMaxRev[] = new double[locMax.length];
	for (int i=0; i<locMax.length; i++) { locMaxRev[i] = locMax[locMax.length - i - 1]; }
	double thres = fdr/locMaxRev.length;
	if (saveFileName!=null) {
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(saveFileName));
		for (int j=0; j<locMax.length; j++) 
		    bw.write(format.format(locMax[j]) + "\n");
		bw.close();
	    }
	    catch (IOException e) { }	    
	}
	for (int i=0; i<locMaxRev.length; i++) {
	    double pVal = 2*NormalDist.barF01(locMaxRev[i]/sigmaHat);
	    if (pVal>(i+1)*thres) { //over the threshold, take the previous value
		int thresInd = i==0 ? i : i-1;
		//System.out.println("Found threshold " + format.format(locMaxRev[thresInd]) + " with thres " + (i*thres) + " at " + thresInd + " with p-val " + pVal + " for " + locMax.length);
		return locMaxRev[thresInd];
	    }
	}
	return locMaxRev.length>0 ? locMaxRev[0] + 1 : 1;
    }

    public static double estimateStdDev(Vector<double[]> dss) {
	//Concatenate the first line of all the vectors first
	int n = 0, i = 0;
	for (double[] ds : dss)
	    n += ds.length;
	double d[] = new double[n];
	for (double[] ds : dss) {
	    for (int j=0; j<ds.length; j++)
		d[i++] = Math.abs(ds[j]);
	}
	return estimateStdDev(d);
    }

    public static double estimateStdDev(double d[]) {
	//use the median absolute deviation
	double c[] = new double[d.length];
	Tally tl = new Tally("medians, log");
	for (int i=0; i<d.length; i++) { 
	    c[i] = Math.abs(d[i]); 
	    if (c[i]>0)
		tl.add(Math.log(c[i]));
	    else
		tl.add(Math.log(.0000001));
	}
	Arrays.sort(c);
	//System.out.println(tl.shortReportHeader());
	//System.out.println(tl.reportAndCIStudent(.95));
	//System.out.println("log-normal median = " + Math.exp(tl.average()));
	//System.out.println("exact median = " + format.format(c[c.length/2]));
	return c[c.length/2]/.6745; //constant taken from paper
    }

    public static double estimateStdDev(TreeMap<Integer, Integer> coeffHist, double binSize) {
	//use the median absolute deviation
	int n = 0, m = 0;
	for (Integer k : coeffHist.keySet())
	    n += coeffHist.get(k);
	for (Integer k : coeffHist.keySet()) {
	    m += coeffHist.get(k);
	    if (m>n/2)
		return binSize*((double)k)/.6745; //constant taken from paper
	}
	System.err.println("Error, could not find median!");
	return 0;
    }

    public static double estimateStdDev(TreeMap<Double, Integer> coeffHist) {
	//use the median absolute deviation
	int n = 0, m = 0;
	for (Double k : coeffHist.keySet())
	    n += coeffHist.get(k);
	for (Double k : coeffHist.keySet()) {
	    m += coeffHist.get(k);
	    if (m>n/2)
		return k/.6745; //constant taken from paper
	}
	System.err.println("Error, could not find median!");
	return 0;
    }

    public static double estimateStdDev(Tally t) {
	//System.out.println(t.shortReportHeader());
	//System.out.println(t.reportAndCIStudent(.95));
	double m = t.average();
	//System.out.println("approx median = " + format.format(m));
	return m/.6745; //constant taken from paper
    }


    ///////////////////////////////////////////////////////

    public static TreeMap<Integer, Double> findLocalMax(double w[], int startPos, int binSize) { return findLocalMax(w, startPos, binSize, 1); }

    public static TreeMap<Integer, Double> findLocalMax(double w[], int startPos, int binSize, double var) {
	//Return all the local maxima found in the vector w. Try
	//dividing the square of the coefficients with the estimated
	//variance to clean up the signal (not likely to help much
	//since we are looking for local maxima).
	TreeMap<Integer, Double> locValMap = new TreeMap<Integer, Double>();
	int n = 0, j = 0, bin = (int)Math.pow(2, binSize);
	for (int i=1; i<w.length-1; i++) {
	    if (isLocalExtrema(w, i))
		locValMap.put(startPos + i*bin, w[i]);
	}
	return locValMap;
    }

    public static boolean isLocalExtrema(double w[], int i) { 
	//This function makes sure that we are not at a plateau that
	//will go further up or down once it has ended
	if (Math.abs(w[i-1])<Math.abs(w[i]) && Math.abs(w[i+1])<Math.abs(w[i])) //True extrema
	    return true;
	else if (Math.abs(w[i-1])==Math.abs(w[i]) && Math.abs(w[i+1])<Math.abs(w[i])) {
	    int j = i-1;
	    while (j>0 && w[j]==w[j+1])
		j--;
	    return Math.abs(w[j])<Math.abs(w[i]);
	}
	else if (Math.abs(w[i-1])<Math.abs(w[i]) && Math.abs(w[i+1])==Math.abs(w[i])) { //start or end of plateau, could be considered an extrema
	    int j = i+1;
	    while (j<w.length && w[j]==w[j-1])
		j++;
	    return Math.abs(w[j])<Math.abs(w[i]);
	}
	return false;
    }

    ////////////////////////////////////////////

    public double[] calculateHistogram(int start, int end) { return calculateHistogram(start, end, this.reads, this.chr, this.maxReadLength); }
    
    public static double[] calculateHistogram(int start, int end, ABIRead[] reads, int chr, int maxReadLength) { return calculateHistogram(start, end, reads, chr, maxReadLength, false); }

    public static double[] calculateHistogram(int start, int end, ABIRead[] reads, int chr, int maxReadLength, boolean log) {
	int startInd = Arrays.binarySearch(reads, new ABIRead(chr, reads[0].getStrand(), start - maxReadLength));
	startInd = startInd<0 ? -startInd-1 : startInd;
	int endInd = Arrays.binarySearch(reads, new ABIRead(chr, reads[0].getStrand(), end));
	endInd = endInd<0 ? -endInd-1 : endInd;	
	//System.out.println("StartInd is " + startInd + " for " + start + " and endInd is " + endInd + " for " + end);
	double hist[] = new double[end - start];
	for (int i=startInd; i<endInd; i++) {
	    for (int j=0; j<reads[i].getLength(); j++) { 
		if (reads[i].getPosition() - start + j>=0 && reads[i].getPosition() - start + j<hist.length)
		    hist[reads[i].getPosition() - start + j]++; 
	    }
	}
	if (log) {
	    for (int i=0; i<hist.length; i++)
		hist[i] = Math.log(hist[i] + 1);
	}
	return hist;
    }

    public double[] calculateHistogram(ABIRead rs[]) { return calculateHistogram(rs, maxReadLength, forward); }

    public static double[] calculateHistogram(ABIRead rs[], int maxReadLen, boolean strand) {
	//Unfold all the reads and create a histogram. This may be a
	//very long array if the number of reads is large. 
	Arrays.sort(rs);
	double hist[] = new double[rs[rs.length-1].getPosition() - rs[0].getPosition() + 2*maxReadLen];
	int r0 = rs[0].getPosition();
	Arrays.fill(hist, 0);
	for (int i=0; i<rs.length; i++) {
	    for (int j=0; j<rs[i].getLength(); j++) { hist[rs[i].getPosition() - r0 + j + maxReadLen]++; }
	}
	return hist;
    }

    public static double[] calculateHistogram(ABIRead rs[], int maxReadLen, int pad, int binSize) { return calculateHistogram(rs, maxReadLen, pad, binSize, null); }

    public static double[] calculateHistogram(ABIRead rs[], int maxReadLen, int pad, int binSize, String saveInfoFileName) {
	//pad the histogram with pad zeros before and after
	Arrays.sort(rs);
	int bin = (int)Math.pow(2, binSize), padBefore = Math.min(pad, rs[0].getPosition()), len = rs[rs.length-1].getPosition() - rs[0].getPosition() + padBefore + pad + maxReadLen;
	double hist[] = new double[len/bin+1];
	int r0 = rs[0].getPosition();
	Arrays.fill(hist, 0);
	for (int i=0; i<rs.length; i++) {
	    for (int j=0; j<rs[i].getLength(); j++) { hist[(rs[i].getPosition() - r0 + j + padBefore)/bin]++; }
	}
	if (saveInfoFileName!=null) {
	    String name = saveInfoFileName + "_hist_" + rs[0].getChromosome() + ".txt";
	    System.out.println("Saving histograms to " + name);
	    try {
		BufferedWriter bw = new BufferedWriter(new FileWriter(name));
		for (int i=0; i<hist.length; i++)
		    bw.write((r0 + i*bin) + "\t" + hist[i] + "\n");
		bw.close();
	    }
	    catch (IOException e) {
		System.err.println("Error saving histogram to " + name);
	    }
	}
	return hist;
    }

    public static double calculateVariance(double hist[], boolean log) {
	double v = 0, m = TranscribedRegion.calculateAverageReadLevel(hist, log);
	for (int i=0; i<hist.length; i++)
	    v += log ? (Math.log(hist[i] + 1) - m)*(Math.log(hist[i] + 1) - m) : (hist[i] - m)*(hist[i] - m);
	return v/(hist.length - 1);
    }

    /////////////////////////////////

    public static double calculateHaarCoefficientNReads(int l, int n, int maxReadLen, boolean log, double normFactor) {
	//Calculate the coefficient that one would obtain if there were n reads in the window of length 2^l - 1.
	double hist[] = new double[(int)Math.pow(2, l) - 1], r = (double)(maxReadLen*n)/hist.length;
	if (r>1)
	    Arrays.fill(hist, r);
	else {
	    for (int i=0; i<maxReadLen*n; i++)
		hist[i] = 1;
	}
	return calculateHaarCoefficient(l, 0, hist, log, normFactor);
    }

    public static double calculateHaarCoefficient(int l, int ind, double hist[], boolean log, double normFactor) {
	//Use this when calculating the coefficient for the first
	//location and length scale, before we can use previously
	//computed results.
	double w = 0;
	for (int i=ind; i<Math.min(hist.length, ind + Math.pow(2, l) - 1); i++)
	    w += log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	for (int i=Math.max(0, ind - (int)Math.pow(2, l)); i<ind - 1; i++)
	    w -= log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	return w;
    }

    public static double calculateHaarCoefficient(int l, int ind, double hist[], double wpl, boolean log, double normFactor) {
	//Since we have already split the set into groups where we
	//know that the bases before and after the first and last read
	//are zero, we can speed up the calculation by ignoring those
	//bases when looping.
	double w = wpl/Math.sqrt(2);
	for (int i=ind + (int)Math.pow(2, l - 1); i<Math.min(hist.length, ind + Math.pow(2, l) - 1); i++)
	    w += log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	for (int i=Math.max(0, ind - (int)Math.pow(2, l)); i<ind - (int)Math.pow(2, l - 1) - 1; i++)
	    w -= log ? Math.log(hist[i] + 1)*normFactor : hist[i]*normFactor;
	return w;
    }
    
    public static double calculateHaarCoefficientInc(int l, int ind, double hist[], double wpp, boolean log, double normFactor) {
	//Save time by only doing the changes compared to the previous
	//coefficient:
	//w_n = w_(n-1) + y(n+2^l-1) + y(n-2^l-1) - 2*y(n-1)
	int l2 = (int)Math.pow(2, l);
	double yForwardAdd = ind + l2<hist.length ? (log ? Math.log(hist[ind + l2] + 1)*normFactor : (double)hist[ind + l2]*normFactor) : 0;
	double yBackwardAdd = ind - 1 - l2>0 ? (log ? Math.log(hist[ind - 1 - l2] + 1)*normFactor : (double)hist[ind - 1 - l2]*normFactor) : 0;
	double ySub = ind>0 ? (log ? Math.log(hist[ind - 1] + 1)*normFactor : (double)hist[ind - 1]*normFactor) : 0;
	return wpp - 2*ySub + yForwardAdd + yBackwardAdd;
    }

    //////////////////////////////////////////////////////////////////
    //For calculating statistics about the density of the coverage

    public ABIRead[] getReads() { return reads; }


    //////////////////////////////////////////////////////////////////
    //For using the Bayesian method
    //Use this stuff for implementing the method suggested by Hutter (2007)

    public void calculateBayesianPiecewiseConstantRegression(int maxDist, int binSize, int maxSegments, int nLogs, String errorModel, int gmm, RefSeqGenes geneInfo) {
	//reducedModel = new ReducedModel(chr, forward, false, species);
	int prevSegEnd = 3002017, bin = (int)Math.pow(2, binSize);
	//reducedModel.addSegment(prevSegEnd, reads[0].getPosition(), 0, 0);
	//Get the first reads in blocks
	//while (moreReadGroups() && reads[readGroupInd].getPosition()<3500000) {
	while (moreReadGroups()) {
	    ABIRead[] rs = nextReadGroup(maxDist);
	    //Get the histogram at the given coarse-graining
	    int blockLength = (rs[rs.length-1].getPosition() - rs[0].getPosition() + maxReadLength);
	    System.out.println("Found block with length " + blockLength + ", containing " + rs.length + " reads starting at " + rs[0].getPosition());
	    if (rs.length<MIN_READS_BLOCK) {
		//System.out.println("Too few reads or, skipping block...");
	    }
	    else if (blockLength/bin<=1) {
		//add one very short segment, this should be quite unusual
		//reducedModel.addSegment(prevSegEnd, rs[0].getPosition(), 0, 0);
		double y = calculateHistogram(rs, maxReadLength, 0, binSize)[0];
		for (int i=0; i<nLogs; i++) { y = Math.log(y + 1); }
		//reducedModel.addSegment(rs[0].getPosition(), rs[0].getPosition() + bin, Math.log(1 + Math.log(1 + y)), 0);
		prevSegEnd = rs[0].getPosition() + bin;
	    }
	    else {
		//Add a segment representing the connection between this one and the next block
		//reducedModel.addSegment(prevSegEnd, rs[0].getPosition(), 0, 0);
		double y[] = calculateHistogram(rs, maxReadLength, 0, binSize), sigma2s[] = null, ws[] = null;
		int n = y.length;
		int nSegmentsMax = Math.min(n, maxSegments);
		for (int i=0; i<n; i++) {
		    for (int j=0; j<nLogs; j++) 
			y[i] = Math.log(1 + Math.log(1 + y[i]));
		}
		double[][] a0, a1, a2;
		if (gmm>1) {
		    //estimate the parameters for the GMM locally
		    double nus[] = estimateNus(errorModel, geneInfo, rs, forward, bin, nLogs, gmm);
		    double rho2s[] = estimateRho2s(errorModel, geneInfo, rs, forward, nLogs, bin, nus, gmm);
		    sigma2s = estimateSigma2s(errorModel, geneInfo, rs, forward, bin, nLogs, gmm);
		    ws = estimateWeights(geneInfo, forward, rs[0].getPosition(), rs[0].getPosition() + n*bin, gmm);
		    System.out.println("Priors for the expressed portions are:\nnu\trho2\tsigma2\tw");
		    for (int i=0; i<gmm; i++)
			System.out.println(format.format(nus[i]) + "\t" + format.format(rho2s[i]) + "\t" + format.format(sigma2s[i]) + "\t" + format.format(ws[i]));
		    double tmp0[][][] = estimateA0(errorModel, y, nus, rho2s, sigma2s, ws);
		    double tmp1[][][] = estimateA1(errorModel, y, nus, rho2s, sigma2s, ws, tmp0);
		    double tmp2[][][] = estimateA2(errorModel, y, nus, rho2s, sigma2s, ws, tmp0, tmp1);
		    a0 = tmp0[gmm];
		    a1 = tmp1[gmm];
		    a2 = tmp2[gmm];
		}
		else {
		    //Estimate the hyper-priors
		    TallyStore observations = new TallyStore();
		    for (int i=0; i<n; i++) 
			observations.add(y[i]);
		    double nu = estimateNu(errorModel, observations), rho2 = estimateRho2(errorModel, observations), sigma2 = estimateSigma2(errorModel, observations);
		    a0 = estimateA0(errorModel, y, nu, rho2, sigma2);
		    a1 = estimateA1(errorModel, y, nu, rho2, sigma2, a0);
		    a2 = estimateA2(errorModel, y, nu, rho2, sigma2, a0, a1);
		}
		//Calculate the regression
		double[][] l = calculateLSum(a0, nSegmentsMax, n);
		double[][] r = calculateRSum(a0, nSegmentsMax, n);
		//Calculate the evidence and Find the most likely number of segments
		double evidence = -Double.MAX_VALUE; 
		double nSegmentProb[] = new double[nSegmentsMax+1], maxProb = -Double.MAX_VALUE;
		int mostLikelySegInd = 1;
		for (int k=1; k<=nSegmentsMax; k++) {
		    nSegmentProb[k] = l[k][n] - org.apache.commons.math.util.MathUtils.binomialCoefficientLog(n-1, k-1);
		    if (maxProb<nSegmentProb[k]) {
			maxProb = nSegmentProb[k];
			mostLikelySegInd = k;
		    }
		    evidence = logSum(evidence, l[k][n] - org.apache.commons.math.util.MathUtils.binomialCoefficientLog(n-1, k-1));
		}
		for (int k=1; k<nSegmentsMax; k++) 
		    nSegmentProb[k] -= evidence;
		evidence -= Math.log(nSegmentsMax);
		//System.out.println("E = " + evidence);
		//System.out.println("nSegmentProb " + Arrays.toString(nSegmentProb));
		System.out.println("Most likely number of segments is " + (mostLikelySegInd) + " with log probability " + format.format(maxProb));
		//calculate the probability for the segment locations as a mixture of 1 to nSegmentsMax
		double norm = 0, boundaryLocationProb[] = new double[n+1];
		if (true) { //mixtureBoundary) {
		    double d[][] = new double[nSegmentsMax+1][n+1];
		    for (int i=0; i<=nSegmentsMax; i++) {
			for (int j=0; j<=n; j++) {
			    d[i][j] = -Double.MAX_VALUE;
			    for (int h=i; h<=nSegmentsMax; h++)
				d[i][j] = logSum(d[i][j], r[h-i][j] - org.apache.commons.math.util.MathUtils.binomialCoefficientLog(n-1, h-1));
			}
		    }
		    for (int i=0; i<=n; i++) {
			//for (int j=0; j<=mostLikelySegInd; j++) {
			for (int j=1; j<=nSegmentsMax; j++) 
			    //boundaryLocationProb[i] += Math.exp(l[j][i] + r[mostLikelySegInd-j][i] - l[mostLikelySegInd][n-1]); 
			    boundaryLocationProb[i] += Math.exp(l[j][i] + d[j][i] - evidence)/nSegmentsMax; 
			norm += boundaryLocationProb[i];
		    }
		    for (int j=0; j<=nSegmentsMax; j++) 
			norm -= j*Math.exp(nSegmentProb[j]);
		    if (Math.abs(norm - 1)>1e-9)
			System.err.println("Normalization error, total boundary probability is " + norm);
		    boundaryLocationProb[0] = 1;
		    System.out.println(Arrays.toString(boundaryLocationProb));
		}
		int boundaryLocations[] = findBoundaryLocations(mostLikelySegInd, boundaryLocationProb);
		System.out.println("Boundary locations are " + Arrays.toString(boundaryLocations));
		//Calculate the estimated levels
		double estimatedLevels[] = new double[mostLikelySegInd], estimatedStdDevs[] = new double[mostLikelySegInd];
		for (int i=1; i<=mostLikelySegInd; i++) {
		    estimatedLevels[i-1] = Math.exp(a1[boundaryLocations[i-1]][boundaryLocations[i]] - a0[boundaryLocations[i-1]][boundaryLocations[i]]);
		    estimatedStdDevs[i-1] = Math.sqrt(Math.exp(a2[boundaryLocations[i-1]][boundaryLocations[i]] - a0[boundaryLocations[i-1]][boundaryLocations[i]]) - estimatedLevels[i-1]*estimatedLevels[i-1]);
		    /*if (estimatedStdDevs[i-1]>estimatedLevels[i-1])
			reducedModel.addSegment(rs[0].getPosition() + boundaryLocations[i-1]*bin, rs[0].getPosition() + boundaryLocations[i]*bin, 0, 0);
		    else
		    reducedModel.addSegment(rs[0].getPosition() + boundaryLocations[i-1]*bin, rs[0].getPosition() + boundaryLocations[i]*bin, estimatedLevels[i-1], estimatedStdDevs[i-1]);*/
		}
		prevSegEnd = rs[0].getPosition() + bin*boundaryLocations[boundaryLocations.length-1];
		System.out.println("Estimated levels are " + Arrays.toString(estimatedLevels));
		System.out.println("Estimated std devs are " + Arrays.toString(estimatedStdDevs));
		//////////////////////
		//calculate the relative likelihood
		double expectedRelativeLogLikelihood = 0, varianceRelativeLogLikelihood = n/2;
		if (gmm>1) {
		    for (int i=0; i<ws.length; i++)
			expectedRelativeLogLikelihood += ws[i]*n*Math.log(2*Math.PI*Math.exp(1)*sigma2s[i])/2;
		}
		//System.out.println("Relative log-likelihood is " + format.format((expectedRelativeLogLikelihood - evidence)/Math.sqrt(varianceRelativeLogLikelihood)));
		//Do the Bayesian regression
	    }
	}
    }

    public static double[] estimateWeights(RefSeqGenes geneInfo, boolean forward, int start, int end, int gmm) {
	//Find the proportion of exons/introns between start and end
	double ws[] = new double[gmm];
	if (gmm==2) { //one prior for coding regions, one for intergenic
	    ws[0] = 1 - (double)geneInfo.codingRegionLength(forward, start, end)/(end - start);
	    ws[1] = 1 - ws[0];
	}
	else { //One for intergenic, introns and exons
	    ws[1] = (double)geneInfo.intronLength(forward, start, end)/(end - start);
	    ws[2] = (double)geneInfo.exonLength(forward, start, end)/(end - start);
	    ws[0] = 1 - ws[1] - ws[2];
	}
	return ws;
    }

    public int[] findBoundaryLocations(int mostLikelySegInd, double boundaryLocationProb[]) {
	int boundaryLocations[] = new int[mostLikelySegInd+1], n = boundaryLocationProb.length;
	TreeMap<Double, Integer> tmp = new TreeMap<Double, Integer>();
	for (int i=0; i<boundaryLocationProb.length; i++)
	    tmp.put(boundaryLocationProb[i], i);
	//Vector<Double> keys = new Vector<Double>(tmp.keySet());
	//for (int i=0; i<=mostLikelySegInd; i++) 
	//boundaryLocations[i] = tmp.get(keys.elementAt(tmp.size()-1-i));
	for (Double p : tmp.keySet()) {
	    if (--n<boundaryLocations.length)
		boundaryLocations[n] = tmp.get(p);
	}
	Arrays.sort(boundaryLocations);
	return boundaryLocations;
    }

    public static double logSum(double x, double y) { return logSum(x, y, true); }

    public static double logSum(double x, double y, boolean logs) {
	//calculate log(x + y). If logs is true, then x and y have already been logged
	if (y>x) { //swap
	    double tmp = x;
	    x = y;
	    y = tmp;
	}
	if (!logs) {
	    x = Math.log(x);
	    y = Math.log(y);
	}
	if (x>y+40)
	    return x; //speed up
	return x + Math.log(1 + Math.exp(y - x));
    }

    public double[][] calculateLSum(double a0[][], int nSegmentsMax, int n) {
	//Calculate all the sums required to initialize the L-matrix
	double l[][] = new double[nSegmentsMax+1][n+1];
	for (int i=0; i<=n; i++) 
	    l[0][i] = -Double.MAX_VALUE; //log scale of delta function
	l[0][0] = 0;
	for (int k=0; k<nSegmentsMax; k++) {
	    for (int i=0; i<=n; i++) {
		l[k+1][i] = -Double.MAX_VALUE; //l[k][k] + a0[k][i];
		for (int h=k; h<=i-1; h++)
		    l[k+1][i] = logSum(l[k+1][i], l[k][h] + a0[h][i]);
		//if (i<5)
		//  System.out.println("i = " + i + ", k = " + k + ", l[k+1][i] = " + l[k+1][i] + " = " + format.format(a0[k+1][i]));
	    }
	}
	return l;
    }

    public double[][] calculateRSum(double a0[][], int nSegmentsMax, int n) {
	//Calculate all the sums required to initialize the R-matrix
	double r[][] = new double[nSegmentsMax+1][n+1];
	for (int i=0; i<n; i++) 
	    r[0][i] = -Double.MAX_VALUE;
	r[0][n] = 0;
	for (int k=0; k<nSegmentsMax; k++) {
	    for (int i=0; i<=n; i++) {
		r[k+1][i] = -Double.MAX_VALUE;
		for (int h=i+1; h<=n-k; h++)
		    r[k+1][i] = logSum(r[k+1][i], r[k][h] + a0[i][h]);
	    }
	}
	return r;
    }

    public double estimateSigma2(String errorModel, TallyStore y) {
	double sigma2 = 0;
	double d[] = y.getArray();
	if (errorModel.equals(GAUSSIAN_ERROR) || errorModel.equals(POISSON_ERROR)) {
	    for (int i=1; i<d.length; i++)
		sigma2 += (d[i] - d[i-1])*(d[i] - d[i-1]);
	    sigma2 /= 2*(d.length - 1);
	}
	else {
	    double delta[] = new double[d.length-1];
	    for (int i=1; i<d.length; i++)
		delta[i-1] = d[i] - d[i-1];
	    Arrays.sort(delta);
	    double sigma = (delta[3*delta.length/4] - delta[delta.length/4])/4;
	    sigma2 = sigma*sigma;
	}
	return sigma2;
    }

    public static double[] estimateSigma2s(String errorModel, RefSeqGenes geneInfo, ABIRead[] reads, boolean forward, int bin, int nLogs, int gmm) {
	double sigma2s[] = new double[gmm];
	if (gmm==2) {
	    sigma2s[0] = geneInfo.varianceWithinIntergenicDensity(reads, forward, true, Math.min(64, bin), nLogs, true); //0; 
	    sigma2s[1] = geneInfo.varianceWithinCodingRegionDensity(reads, forward, true, Math.min(64, bin), nLogs, true); //0; 
	}
	else if (gmm==3) {
	    sigma2s[0] = geneInfo.varianceWithinIntergenicDensity(reads, forward, true, Math.min(64, bin), nLogs, true); //0; 
	    sigma2s[1] = geneInfo.varianceWithinIntronDensity(reads, forward, true, Math.min(64, bin), nLogs, true); 
	    sigma2s[2] = geneInfo.varianceWithinExonDensity(reads, forward, true, Math.min(16, bin), nLogs, true); //0; 
	}
	if (sigma2s[0]==0)
	    sigma2s[0] = sigma2s[1]/10;
	return sigma2s;
    }

    public double estimateNu(String errorModel, TallyStore y) {
	if (errorModel.equals(GAUSSIAN_ERROR) || errorModel.equals(POISSON_ERROR)) 
	    return y.average();
	double tmp[] = (y.clone()).getArray();
	Arrays.sort(tmp);
	return tmp[tmp.length/2];
    }

    public static double[] estimateNus(String errorModel, RefSeqGenes geneInfo, ABIRead[] reads, boolean forward, int bin, int nLogs, int gmm) {
	double nus[] = new double[gmm];
	if (gmm==2) {
	    nus[0] = geneInfo.averageIntergenicDensity(reads, forward, true, Math.min(64, bin), nLogs); 
	    nus[1] = geneInfo.averageCodingRegionDensity(reads, forward, Math.min(64, bin), nLogs, true);
	}
	else if (gmm==3) {
	    nus[0] = geneInfo.averageIntergenicDensity(reads, forward, true, Math.min(64, bin), nLogs); 
	    nus[1] = geneInfo.averageIntronDensity(reads, forward, true, Math.min(bin, 64), nLogs, true);
	    nus[2] = geneInfo.averageExonDensity(reads, forward, true, Math.min(16, bin), nLogs, true);
	}
	return nus;
    }


    public double estimateRho2(String errorModel, TallyStore y) {
	if (errorModel.equals(GAUSSIAN_ERROR) || errorModel.equals(POISSON_ERROR)) 
	    return y.variance();
	double tmp[] = (y.clone()).getArray();
	Arrays.sort(tmp);
	double rho = (tmp[3*tmp.length/4] - tmp[tmp.length/4])/2;
	return rho*rho;
    }

    public static double[] estimateRho2s(String errorModel, RefSeqGenes geneInfo, ABIRead[] reads, boolean forward, int nLogs, int bin, double[] nus, int gmm) {
	double rho2s[] = new double[gmm];
	if (gmm==2) {
	    rho2s[0] = geneInfo.varianceIntergenicDensity(reads, forward, true, nus[0], Math.min(64, bin), nLogs); 
	    rho2s[1] = geneInfo.varianceCodingRegionDensity(reads, forward, true, nus[1], Math.min(64, bin), nLogs, true); 
	}
	else if (gmm==3) {
	    rho2s[0] = geneInfo.varianceIntergenicDensity(reads, forward, true, nus[0], Math.min(64, bin), nLogs); 
	    rho2s[1] = geneInfo.varianceIntronDensity(reads, forward, true, nus[1], Math.min(64, bin), nLogs, true); 
	    //Use a more fine-grained bin-size for the exons since we
	    //know that these are short and have higher fluctuations
	    rho2s[2] = geneInfo.varianceExonDensity(reads, forward, true, nus[1], Math.min(16, bin), nLogs, true); 
	}
	if (rho2s[0]==0)
	    rho2s[0] = rho2s[1]/10;
	return rho2s;
    }

    public double[][][] estimateA0(String errorModel, double y[], double nu[], double rho2[], double sigma2[], double w[]) {
	//Use this for the mixture model prior
	int n = y.length;
	double[][][] a0 = new double[nu.length+1][n+1][n+1];
	for (int v=0; v<nu.length; v++) {
	    //a0 = estimateA0(true, y, nu[v], rho2[v], sigma2[v]);
	    a0[v] = estimateA0(errorModel, y, nu[v], rho2[v], sigma2[v]);
	    for (int i=0; i<a0[v].length; i++) {
		for (int j=0; j<a0[v][i].length; j++) 
		    a0[nu.length][i][j] += w[v]*a0[v][i][j];
	    }
	}
	return a0;
    }

    public double[][] estimateA0(String errorModel, double y[], double nu, double rho2, double sigma2) {
	//returns the log of the actual quantity
	int n = y.length;
	double[][] a0 = new double[n+1][n+1];
	if (errorModel.equals(GAUSSIAN_ERROR)) {
	    for (int i=0; i<=n; i++) {
		double m = 0, s = 0;
		for (int j=i+1; j<=n; j++) {
		    int d = j - i;
		    if (d<=0)
			a0[i][j] = -Double.MAX_VALUE;
		    m += y[j-1] - nu;
		    s += (y[j-1] - nu)*(y[j-1] - nu);
		    a0[i][j] = (m*m/(d + sigma2/rho2) - s)/(2*sigma2) - d*Math.log(2*Math.PI*sigma2)/2 - Math.log(1 + d*rho2/sigma2)/2;
		}
	    }
	}
	else if (errorModel.equals(POISSON_ERROR)) {
	    //Assume Poisson distribution of tags with a Gamma-prior
	    //for the rate. The rates for the Gamma are determined
	    //from the mean and variance of the reads, nu and rho2.
	    double beta = rho2/nu, alpha = nu/beta, gamma = 1/(1+beta);
	    System.out.println("alpha = " + alpha + ", beta = " + beta + " and gamma = " + gamma);
	    for (int i=0; i<=n; i++) {
		double m = 0, normFactor = 0;
		for (int j=i+1; j<=n; j++) {
		    int d = j - i;
		    if (d<=0)
			a0[i][j] = -Double.MAX_VALUE;
		    m += y[j-1];
		    normFactor -= org.apache.commons.math.special.Gamma.logGamma(1 + y[j-1]);
		    //Calculate the log of the adjusted NB density
		    a0[i][j] = org.apache.commons.math.special.Gamma.logGamma(m + alpha) - org.apache.commons.math.special.Gamma.logGamma(alpha) - org.apache.commons.math.special.Gamma.logGamma(m + 1) + alpha*Math.log(gamma) + m*Math.log(1 - gamma) + org.apache.commons.math.special.Gamma.logGamma(1 + m) - normFactor;
		    if (i==0)
			System.out.println(m + "\t" + a0[i][j]);
		}
	    }
	}
	else  //Now we have to do numerical integration
	    a0 = getCauchyEstimate(y, nu, Math.sqrt(rho2), Math.sqrt(sigma2), 0);
	return a0;
    }

    public double[][] getCauchyEstimate(double y[], double nu, double rho, double sigma, int r) {
	//Assume Cauchy distributed noise and priors
	int n = y.length;
	double[][] a = new double[n+1][n+1];
	//find out the parameters of the grid that we wish to use for
	//the numerical integration
	double gridStart = nu - 25*rho, gridStep = sigma/10;
	int gridLength = (int)Math.round((nu + 50*rho)/gridStep) + 1;
	System.out.println("gridStart = " + gridStart + ". gridStep = " + gridStep + " and gridLength = " + gridLength);
	double logPymu[] = new double[gridLength];
	for (int i=0; i<=n; i++) {
	    for (int u=0; u<logPymu.length; u++) //initialize the grid
		logPymu[u] = logCauchyPDF(gridStart + u*gridStep, nu, rho);
	    for (int j=0; j<n; j++) {
		int d = j - i;
		double s = -Double.MAX_VALUE;
		if (d<=0) {
		    a[i][j] = r==0 ? -Double.MAX_VALUE : 0;
		    for (int u=0; u<logPymu.length; u++)
			s += logSum(s, logPymu[u]);
		}
		for (int u=0; u<gridLength; u++) {
		    logPymu[u] += logCauchyPDF(y[j], gridStart + u*gridStep, sigma);
		    s = logSum(s, logPymu[u]);
		}
		a[i][j] = r==0 ? s : 0;
		if (r==1) {
		    for (int u=0; u<gridLength; u++)
			a[i][j] += (gridStart + u*gridStep)*Math.exp(logPymu[u] - s);
		}
		else if (r==2) {
		    for (int u=0; u<gridLength; u++)
			a[i][j] += (gridStart + u*gridStep)*(gridStart + u*gridStep)*Math.exp(logPymu[u] - s);
		}
	    }
	}
	return a;
    }

    public static double logCauchyPDF(double x, double a, double b) { return b/(Math.PI*(b*b + (x - a)*(x - a))); }

    public double[][][] estimateA1(String errorModel, double y[], double nu[], double rho2[], double sigma2[], double w[], double a0[][][]) {
	//Use this for the mixture model prior
	int n = y.length;
	double[][][] a1 = new double[nu.length+1][n+1][n+1];
	for (int v=0; v<nu.length; v++) {
	    a1[v] = estimateA1(errorModel, y, nu[v], rho2[v], sigma2[v], a0[v]);
	    for (int i=0; i<a1[v].length; i++) {
		for (int j=0; j<a1[v][i].length; j++) 
		    a1[nu.length][i][j] += w[v]*a1[v][i][j];
	    }
	}
	return a1;
    }

    public double[][] estimateA1(String errorModel, double y[], double nu, double rho2, double sigma2, double a0[][]) {
	int n = y.length;
	double[][] a1 = new double[n+1][n+1];
	if (errorModel.equals(GAUSSIAN_ERROR)) {
	    for (int i=0; i<=n; i++) {
		double m = 0;
		for (int j=i+1; j<=n; j++) {
		    int d = j - i;
		    if (d<=0)
			a1[i][j] = 0;
		    m += y[j-1] - nu;
		    a1[i][j] = a0[i][j] + Math.log(nu + m/d);
		}
	    }
	}
	else if (errorModel.equals(POISSON_ERROR)) {
	    double beta = rho2/nu, alpha = nu/beta, gamma = 1/(1+beta);
	    for (int i=0; i<=n; i++) {
		double m = 0, normFactor = 0;
		for (int j=i+1; j<=n; j++) {
		    int d = j - i;
		    if (d<=0)
			a0[i][j] = -Double.MAX_VALUE;
		    m += y[j-1];
		    normFactor -= org.apache.commons.math.special.Gamma.logGamma(1 + y[j-1]);
		    //Calculate the log of the adjusted NB density
		    a1[i][j] = Math.log(m + 1) + org.apache.commons.math.special.Gamma.logGamma(m + 1 + alpha) - org.apache.commons.math.special.Gamma.logGamma(alpha) - org.apache.commons.math.special.Gamma.logGamma(m + 2) + alpha*Math.log(gamma) + (m + 1)*Math.log(1 - gamma) + org.apache.commons.math.special.Gamma.logGamma(2 + m) - normFactor;
		}
	    }
	}
	else  //Now we have to do numerical integration
	    a1 = getCauchyEstimate(y, nu, Math.sqrt(rho2), Math.sqrt(sigma2), 1);
	return a1;
    }

    public double[][][] estimateA2(String errorModel, double y[], double nu[], double rho2[], double sigma2[], double w[], double a0[][][], double a1[][][]) {
	//Use this for the mixture model prior
	int n = y.length;
	double[][][] a2 = new double[nu.length+1][n+1][n+1];
	for (int v=0; v<nu.length; v++) {
	    a2[v] = estimateA2(errorModel, y, nu[v], rho2[v], sigma2[v], a0[v], a1[v]);
	    for (int i=0; i<a2[v].length; i++) {
		for (int j=0; j<a2[v][i].length; j++) 
		    a2[nu.length][i][j] += w[v]*a2[v][i][j];
	    }
	}
	return a2;
    }

    public double[][] estimateA2(String errorModel, double y[], double nu, double rho2, double sigma2, double a0[][], double a1[][]) {
	int n = y.length;
	double[][] a2 = new double[n+1][n+1];
	if (errorModel.equals(GAUSSIAN_ERROR)) {
	    for (int i=0; i<=n; i++) {
		for (int j=i+1; j<=n; j++) 
		    a2[i][j] = a0[i][j] + logSum(2*(a1[i][j] - a0[i][j]), sigma2/(j - i));
	    }
	}
	else if (errorModel.equals(POISSON_ERROR)) {
	    double beta = rho2/nu, alpha = nu/beta, gamma = 1/(1+beta);
	    for (int i=0; i<=n; i++) {
		double m = 0, normFactor = 0;
		for (int j=i+1; j<=n; j++) {
		    int d = j - i;
		    if (d<=0)
			a0[i][j] = -Double.MAX_VALUE;
		    m += y[j-1];
		    normFactor -= org.apache.commons.math.special.Gamma.logGamma(1 + y[j-1]);
		    //Calculate the log of the adjusted NB density
		    a1[i][j] = Math.log(m + 2) + Math.log(m + 1) + org.apache.commons.math.special.Gamma.logGamma(m + 2 + alpha) - org.apache.commons.math.special.Gamma.logGamma(alpha) - org.apache.commons.math.special.Gamma.logGamma(m + 3) + alpha*Math.log(gamma) + (m + 2)*Math.log(1 - gamma) + org.apache.commons.math.special.Gamma.logGamma(3 + m) - normFactor;
		}
	    }
	}
	else  //Now we have to do numerical integration
	    a2 = getCauchyEstimate(y, nu, Math.sqrt(rho2), Math.sqrt(sigma2), 2);
	return a2;
    }


    ///////////////////////////////////////////////////////
    //Use these methods when we want to split up the reads into
    //isolated clusters.

    public boolean moreReadGroups() { return readGroupInd<reads.length; }

    public ABIRead[] nextReadGroup(int sepDist) {
	int startInd = readGroupInd, endInd;
	while (readGroupInd<reads.length-1 && reads[readGroupInd+1].getPosition() - reads[readGroupInd].getPosition()<sepDist) 
	    readGroupInd++;
	endInd = ++readGroupInd;
	ABIRead rs[] = new ABIRead[endInd - startInd];
	for (int i=startInd; i<endInd; i++)
	    rs[i-startInd] = reads[i];
	return rs;
    }

    public double[] shiftHistogram(double[] hist, int newLastWinInd, int maxLen) {
	double h[] = new double[Math.min(maxLen, hist.length + 1)];
	if (h.length==maxLen) { //move all the stuff one bp
	    for (int i=0; i<hist.length-1; i++)
		h[i] = hist[i+1];
	}
	else {
	    for (int i=0; i<hist.length; i++)
		h[i] = hist[i];
	}
	int i = 0; //check if there is anything to be added in the last slot
	while (readsInd+i<reads.length && reads[readsInd+i].getPosition()<newLastWinInd) {
	    if (reads[readsInd+i].overlaps(newLastWinInd)) //reads may have different length and hence we cannot be sure that there will be overlap
		h[h.length - 1]++;
	    i++;
	}
	//step readsInd to ignore reads which may no longer contribute
	while (readsInd<reads.length && reads[readsInd].getPosition()<newLastWinInd && !reads[readsInd].overlaps(newLastWinInd)) { 
	    readsInd++; 
	    //System.out.println(readsInd + "\t" + genomeInd + "\t" + newLastWinInd);
	}
	return h;
    }

}
/*
class CauchyDensityPriorIntegrand implements IntegralFunction{

    //The parameters for the noise
    private double mu = 0, sigma = 0;
    //The parameters for the prior
    private double nu = 0, rho = 0;
    //The moment we wish to compute
    private int r = 0;

    public double function(double x){ 
	return b/(Math.PI*(b*b + (x - a)*(x - a))); 
    }

    //public void setA(double a){ this.a = a; }

    //public void setB(double b){ this.b = b; }
    
    public void setR(double r){ this.r = r; }

}

class CauchyDensity implements IntegralFunction{

    private double a = 0, b = 0;

    public double function(double x){ return b/(Math.PI*(b*b + (x - a)*(x - a))); }

    public void setA(double a){ this.a = a; }

    public void setB(double b){ this.b = b; }

}
*/