//Use this class to look through the win_dist files with a given
//threshold to determine which windows are significant. It merges
//overlapping windows to count the effective number of peaks. This
//class is designed to be Matlab-friendly and can be called from with
//in Matlab.
//
//Example:
//
//javaaddpath '/home/mhemberg/Projects/ChIP-Seq/TKData';
//wins = [];
//winFinder = javaObject('WindowFinder', '../Npas4/KCl/Npas4_KCl', .001);
//for i=1:21
//  wins = [wins; double(javaMethod('findSignificantWindows', winFinder, i, 240))];
//end
//disp(['Found ',num2str(length(wins)),' significant windows.']);

import jargs.gnu.CmdLineParser;
import java.io.*;
import java.util.*;
import java.util.regex.*;
import java.text.DecimalFormat;

public class WindowFinder {

    public static final int MAX_BUFFER_SIZE = 25000;

    protected String fileName;
    protected FDRCutOffs cutOffs;
    protected Vector mergedWindows[];
    protected int windowSize = 240, windowStep = 10, totReads = 0, totInputReads = 0, chromosome;
    protected BufferedReader windowReader = null;
    //Keep track of the position in the file to speed up the jump to
    //that location if we have failed to rewind (this happens quite
    //frequently). In the vector
    protected Vector byteCounts = new Vector();
    //Normalization factor for writing bed-files
    protected double normalizationFactor = 0;

    public static DecimalFormat format = new DecimalFormat("###.########");

    public static Pattern windowPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");

    public static final String WIN_LIST_SUFFIX = "_win_dist_";
    public static final String PEAKS_SUFFIX = "_peaks_";
    public static final String PEAKS_SUFFIX_NON_SIGNIFICANT = "_peaks_ns_";
    public static final String READ_STATS_SUFFIX = "_read_stats_";
    

    //////////////////////////////////////////////////////////////////

    public WindowFinder(String f, double fdr, boolean locFDR) {
	fileName = f;
	cutOffs = new FDRCutOffs(fileName, fdr, locFDR);
    }

    public WindowFinder(String f, boolean locFDR, int w, int winStep, int nMappabilityBins, int nGCBins) {
	fileName = f;
	windowSize = w;
	windowStep = winStep;
	cutOffs = new FDRCutOffs(fileName, windowSize, nMappabilityBins, nGCBins);
	for (int i=0; i<nMappabilityBins; i++) {
	    for (int j=0; j<nGCBins; j++) {
		//System.out.println("m = " + i + ", g = " + j + ", cut-off = " + cutOffs.getFDRThreshold(i*windowSize/nMappabilityBins, j*windowSize/nGCBins, 9, 2));
	    }
	}
    }

    public WindowFinder(String f, int chr, int fragmentLength, boolean locFDR, int w, int winStep, int nMappabilityBins, int nGCBins, boolean hs) throws IOException {
	fileName = f;
	windowSize = w;
	windowStep = winStep;
	cutOffs = new FDRCutOffs(fileName, windowSize, nMappabilityBins, nGCBins);
	//Open the file with the list of windows
	windowReader = new BufferedReader(new FileReader(fileName + WIN_LIST_SUFFIX + chr + ".txt"), MAX_BUFFER_SIZE);
	chromosome = chr;
	normalizationFactor = getBedFileNormalizationFactor(fileName, chr, fragmentLength, hs);

    }

    protected WindowFinder(WindowFinder wf, long skip) throws IOException {
	//Copy constructor
	fileName = wf.fileName;
	windowSize = wf.windowSize;
	windowStep = wf.windowStep;
	cutOffs = wf.cutOffs;
	chromosome = wf.chromosome;
	//Open the file with the list of windows
	windowReader = new BufferedReader(new FileReader(fileName + WIN_LIST_SUFFIX + chromosome + ".txt"));
	windowReader.skip(skip); //Fast forward
    }

    //////////////////////////////////////////////////////////////////

    public Vector[] findSignificantWindows(int chr, int fragmentLength, double fdrs[], double inputFDR, boolean hs) { return findSignificantWindows(chr, fragmentLength, fdrs, inputFDR, hs, false); }

    public Vector[] findSignificantWindows(int chr, int fragmentLength, double fdrs[], double inputFDR, boolean hs, boolean savePeakFile) {
	mergedWindows = new Vector[fdrs.length];
	for (int i=0; i<fdrs.length; i++) 
	    mergedWindows[i] = new Vector();
	try {
	    String c = "chr" + chr;
	    double normalizationFactor = 0;
	    int totReads = 0;
	    BufferedWriter bedFileWriter = null;
	    if (savePeakFile && fdrs.length==1) { //We can save a peak file
		bedFileWriter = new BufferedWriter(new FileWriter(fileName + PEAKS_SUFFIX + chr + ".bed"));
		if (chr==1) { writeBedFileHeader(bedFileWriter, fileName, true);}
		else if (chr==20) { c = "chrX"; }
		else if (chr==21) { c = "chrY"; }
		normalizationFactor = getBedFileNormalizationFactor(fileName, chr, fragmentLength, hs);
	    }
	    //Open the file with the list of windows
	    windowReader = new BufferedReader(new FileReader(fileName + WIN_LIST_SUFFIX + chr + ".txt"));
	    String line = windowReader.readLine();
	    //>start mappability gc #ChIP deltaChIP #Input deltaInput
	    int deltaRead[] = new int[fdrs.length], deltaInput[] = new int[fdrs.length];
	    //Keep track on how far away from the actual window-start
	    //where the changes have occured.
	    int windowShift = windowSize/2;// - FDRCutOffs.FRAGMENT_LENGTH;
	    Arrays.fill(deltaRead, 0);
	    Arrays.fill(deltaInput, 0);
	    Window[] win = new Window[fdrs.length];
	    for (int i=0; i<win.length; i++) { win[i] = null; }
	    double readsPoisson = 0;
	    while (line!=null) { //Find the ones which meet the threshold
		Matcher m = windowPattern.matcher(line);
		if (m.find()) {
		    int start = Integer.parseInt(m.group(1));
		    int map = Integer.parseInt(m.group(2));
		    int gc = Integer.parseInt(m.group(3));
		    int nReads = Integer.parseInt(m.group(4));
		    int newReads = Integer.parseInt(m.group(5));
		    int nInput = Integer.parseInt(m.group(6));
		    int newInput = Integer.parseInt(m.group(7));
		    double threshold = cutOffs.getFDRThreshold(map, gc, nReads, nInput);
		    double inputThreshold = cutOffs.getInputFDRThreshold(map, gc, nInput);
		    //if (nReads>50) 
		    //	System.out.println(start + "\t" + nReads + "\t" + nInput + "\t" + format.format(threshold));
		    //Do not calculate an average when merging
		    //windows, should hopefully not make much of a
		    //difference.
		    readsPoisson = cutOffs.getReadsPoisson(map, gc);
		    for (int i=0; i<fdrs.length; i++) { //Consider all the cut-offs simultaneously
			if (threshold<fdrs[i] && inputThreshold>inputFDR) {
			    if (win[i]!=null && start<=win[i].getEnd() + windowSize - windowShift) { //We can merge the windows
				//Add the reads in the windows between
				win[i].addReads(newReads + deltaRead[i]);
				win[i].addInput(newInput + deltaInput[i]);
				win[i].setEnd(start + windowShift + windowStep);
				win[i].addReadsPoisson(cutOffs.getReadsPoisson(map, gc));
				win[i].setDetectionDistance(/*nReads - nInput - cutOffs.getDiffThreshold(fdrs[i], map, gc) + 1*/1, nReads);
				//Not correcting for input
				win[i].setCenter(nReads, start + fragmentLength); //Ugly as ....
				deltaRead[i] = 0;
				deltaInput[i] = 0;
				if (bedFileWriter!=null)
				    bedFileWriter.write(c + "\t" + (start + windowShift) + "\t" + (start + windowShift + windowStep) + "\t" + format.format(normalizationFactor*(double)nReads) + "\n");
			    }
			    else { //New window, save the old one
				if (win[i]!=null) { 
				    //System.out.println("Found window " + win[i].toString());
				    mergedWindows[i].addElement(win[i]); 
				}
				win[i] = new Window(chr, start + fragmentLength/* + windowShift*/, start + fragmentLength + /*windowShift + */windowStep, nReads, nInput, cutOffs.getReadsPoisson(map, gc), (double)gc/map, 1/*, nReads - nInput - cutOffs.getDiffThreshold(fdrs[i], map, gc) + 1*/);
				//System.out.println("Found window " + win[i].toString() + " with nReads " + nReads + " and " + nInput);
				deltaRead[i] = 0;
				deltaInput[i] = 0;
			    }
			}
			else if (win[i]!=null && start<=win[i].getEnd()) { //Save the differences in case we run into a significant window later
			    deltaRead[i] += newReads;
			    deltaInput[i] += newInput;
			    //We also need to extend the last region
			    //of the .bed file to make sure that we
			    //plot the intensity for the entire window
			    //bedFileWriter.write(c + "\t" + start + "\t" + (start + windowStep) + "\t" + format.format(normalizationFactor*(double)peakHeight) + "\n");
			}
		    }
		}
		line = windowReader.readLine();
	    }
	    for (int i=0; i<fdrs.length; i++) { //Add the last window
		if (win[i]!=null)
		    mergedWindows[i].addElement(win[i]);
		System.out.println(chr + "\t" + (fdrs[i]) + "\t" + fileName + "\t" + mergedWindows[i].size() + "\t" + format.format(cutOffs.calculateExpectedPower()) + "\t" + format.format(cutOffs.calculateStdDevPower()));
	    }
	    windowReader.close();
	    if (bedFileWriter!=null) { bedFileWriter.close(); }
	}
	catch (IOException e) {
	    System.err.println("Error reading " + fileName);
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
	//int[][][] mergedWins = new int[fdrs.length][mergedWindows[i].size()][5];
	//for (int i=0; i<fdrs.length; i++) {
	//    for (int j=0; j<mergedWindows[i].size(); j++) 
	//    mergedWins[i] = (int[])mergedWindows.elementAt(i);
	//return mergedWins;
	return mergedWindows;
    }

    protected static void writeBedFileHeader(BufferedWriter bedFileWriter, String fileName, boolean significantWindows) throws IOException {
	String tfName = fileName.substring(fileName.lastIndexOf("/") + 1);
	bedFileWriter.write("browser position chr1:21315319-21327381\n");
	if (significantWindows)
	    bedFileWriter.write("track type=wiggle_0 name=\"" + tfName + PEAKS_SUFFIX + "data\" description=\"" + tfName + PEAKS_SUFFIX + "data\" visibility=dense autoScale=off viewLimits=0:2 color=\"255,0,0\" maxHeightPixels=70:40:10 priority=1\n");
	else
	    bedFileWriter.write("track type=wiggle_0 name=\"" + tfName + PEAKS_SUFFIX_NON_SIGNIFICANT + "data\" description=\"" + tfName + PEAKS_SUFFIX_NON_SIGNIFICANT + "data\" visibility=hide autoScale=off viewLimits=0:2 color=\"0,255,0\" maxHeightPixels=70:40:10 priority=1\n");
    }

    protected static void writeCodingRegionsBedFileHeader(BufferedWriter bedFileWriter, String fileName) throws IOException {
	//use this for coding regions
	String tfName = fileName.substring(fileName.lastIndexOf("/") + 1);
	bedFileWriter.write("browser position chr1:21315319-21327381\n");
	bedFileWriter.write("track type=wiggle_0 name=\"" + tfName + FindOverlappingPeaks.CODING_REGION_PEAKS_SUFFIX + "data\" description=\"" + tfName + FindOverlappingPeaks.CODING_REGION_PEAKS_SUFFIX + "data\" visibility=dense autoScale=off viewLimits=-30:30 color=\"0,0,255\" maxHeightPixels=70:40:10 priority=1\n");
    }

    public static double getBedFileNormalizationFactor(String fileName, int chr, int fragmentLength, boolean hs) throws IOException {
	//Find the total number of reads by parsing the read_stats-file
	BufferedReader statsReader = new BufferedReader(new FileReader(fileName.endsWith(".bed") ? fileName.replace(".bed", READ_STATS_SUFFIX + chr + "_E" + fragmentLength + ".txt") : fileName + READ_STATS_SUFFIX + chr + "_E" + fragmentLength + ".txt"));
	String line = statsReader.readLine();
	Pattern statsPattern = Pattern.compile("(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)\\s+(\\d+)");
	int n = 0;
	while (line!=null) {
	    Matcher m = statsPattern.matcher(line);
	    if (m.find() && Integer.parseInt(m.group(1))==fragmentLength) { n = Integer.parseInt(m.group(4)); }
	    line = statsReader.readLine();
	}
	if (hs)
	    return (double)CreateBedFiles.READS_NORMALIZATION_FACTOR*(double)PeakFinder2.CHROMOSOME_LENGTH_HS[chr-1]/((double)n*(double)PeakFinder2.TOTAL_CHROMOSOME_LENGTH_HS);
	else
	    return (double)CreateBedFiles.READS_NORMALIZATION_FACTOR*(double)PeakFinder2.CHROMOSOME_LENGTH[chr-1]/((double)n*(double)PeakFinder2.TOTAL_CHROMOSOME_LENGTH);
    }

    //////////////////////////////////////////////////////////////////

    
    public Window findReadsInCodingRegion(RefSeqGene gene) { return findReadsInCodingRegion(gene, null); }

    public Window findReadsInCodingRegion(RefSeqGene gene, BufferedWriter bedFileWriter) { 
	if (gene.getChromosome()==chromosome)
	    return gene.getStrand() ? findReadsInRegion(gene.getStart(), gene.getEnd(), bedFileWriter, true) : findReadsInRegion(gene.getEnd(), gene.getStart(), bedFileWriter, true); 
	return null;
    }

    public Window findReadsInRegion(int start, int end) { return findReadsInRegion(start, end, null, true); }

    public Window findReadsInRegion(int start, int end, BufferedWriter bedFileWriter, boolean tryAgain) {
	//Find the total number of reads as well as the bias due to
	//mappability and GC-content in the region between start and
	//end
	Window win = null;
	try {
	    String c = "chr" + chromosome;
	    long byteCount = byteCounts.size()>0 ? ((Long)byteCounts.lastElement()).longValue() : 0;
	    if (chromosome==20) { c = "chrX"; }
	    else if (chromosome==21) { c = "chrY"; }
	    String line = windowReader.readLine();
	    if (line==null) { return win; }
	    byteCount += (long)line.length();
	    //>start mappability gc #ChIP deltaChIP #Input deltaInput
	    if (Integer.parseInt(line.substring(0, line.indexOf("\t")))>start) { //Check if we have read beyond the start of the gene already. This can happen since we sometimes have overlapping genes. If that is the case, we reset the reader to the previous mark and start over
		try {
		    windowReader.reset();
		    line = windowReader.readLine();
		    //System.out.println("Found overlapping genes, rewinding...");
		}
		catch (IOException e) {
		    //System.err.println("Failed to rewind...");
		    //Create a new object and find the information for
		    //this gene
		    if (tryAgain) {
			//closeWindowReader(); //Close the stream
			//We want to skip to where the previous gene started
			WindowFinder wf = new WindowFinder(this, 0); //byteCounts.size()>0 ? ((Long)byteCounts.lastElement()).longValue() : 0);
			win = wf.findReadsInRegion(start, end, bedFileWriter, false);
			wf.closeWindowReader();
			//windowReader = new BufferedReader(new FileReader(fileName + WIN_LIST_SUFFIX + chromosome + ".txt"), MAX_BUFFER_SIZE); //Open the stream again
			return win;
		    }
		}
	    }
	    windowReader.mark(MAX_BUFFER_SIZE);
	    int s = 0, windowShift = windowSize - FDRCutOffs.FRAGMENT_LENGTH;
	    boolean firstWindowFound = false;
	    while (line!=null && s+windowSize<=end) { //Find the ones which are within the region of interest
		Matcher m = windowPattern.matcher(line);
		if (m.find()) {
		    s = Integer.parseInt(m.group(1));
		    int map = Integer.parseInt(m.group(2));
		    int gc = Integer.parseInt(m.group(3));
		    int nReads = Integer.parseInt(m.group(4));
		    int newReads = Integer.parseInt(m.group(5));
		    int nInput = Integer.parseInt(m.group(6));
		    int newInput = Integer.parseInt(m.group(7));
		    if (s>=start && s+windowSize<=end) {
			if (!firstWindowFound) { //Store the start of the gene
			    byteCounts.addElement(new Long(byteCount)); 
			    firstWindowFound = true;
			}
			//We can merge the windows
			if (win==null) 
			    win = new Window(chromosome, start/* + windowShift*/, start/* + windowShift*/ + windowStep, nReads, nInput, cutOffs.getReadsPoisson(map, gc), cutOffs.getInputPoisson(map, gc), (double)gc/windowSize);
			else { ///Add the reads in the windows between
			    win.addReads(newReads);
			    win.addInput(newInput);
			    win.setEnd(start + windowShift + windowStep);
			    win.addReadsPoisson(cutOffs.getReadsPoisson(map, gc));
			    win.addInputPoisson(cutOffs.getInputPoisson(map, gc));
			}
			if (bedFileWriter!=null)
			    bedFileWriter.write(c + "\t" + (start + windowShift) + "\t" + (start + windowShift + windowStep) + "\t" + format.format(normalizationFactor*(double)nReads) + "\n");
		    }
		}
		line = windowReader.readLine();
		if (line!=null) { byteCount += line.length(); }
	    }
	}
	catch (IOException e) {
	    System.err.println("Error reading " + fileName);
	    System.err.println("Caught exception " + e.toString());
	    StackTraceElement s[] = e.getStackTrace();
	    for(int i=0; i<s.length; i++)
		System.err.println(s[i].toString());
	}
	//if (win==null)
	//  win = new Window(chromosome, start, start + windowSize, 0, 0, cutOffs.getReadsPoisson(map, gc), cutOffs.getInputPoisson(map, gc));
	return win;
    }

    public void closeWindowReader() throws IOException { windowReader.close(); }

    public int numberOfReads() { return totReads; }
    
    public int numberOfInputReads() { return totInputReads; }

    public int numberOfWindowsVectors() { return mergedWindows.length; }
	
    public int[][] getMergedWindows() { return getMergedWindows(0); }

    public int[][] getMergedWindows(int i) {
	if (mergedWindows!=null && i<mergedWindows.length) {
	    int[][] mergedWins = new int[mergedWindows[i].size()][5];
	    for (int j=0; j<mergedWindows[i].size(); j++) 
		mergedWins[j] = (int[])mergedWindows[i].elementAt(j);
	    return mergedWins;
	}
	System.err.println("Cannot find windows for index " + i);
	return null;
    }

    public int[] getNextWindow() throws IOException {
	//Use this method to read one line from the file and return it as a window
	String line = windowReader.readLine();
	if (line!=null) {
	    Matcher m = windowPattern.matcher(line);
	    if (m.find()) 
		//start mappability gc reads newReads input newInput
		return new int[]{Integer.parseInt(m.group(1)), Integer.parseInt(m.group(2)), Integer.parseInt(m.group(3)), Integer.parseInt(m.group(4)), Integer.parseInt(m.group(5)), Integer.parseInt(m.group(6)), Integer.parseInt(m.group(7))};
	}
	return null;
    }

    //////////////////////////////////////////////////////////////////
    //Use these methods for finding the number of reads surrounding a
    //specific feature, eg a TSS.



}
	
