//Use this class to represent one microRNA

import java.util.*;
import edu.uci.ics.jung.graph.impl.*;

public class MiRNA extends UndirectedSparseVertex implements Comparable {

    //The start is always the 5' end and will thus be larger than end
    //for genes on the - strand.
    protected int chr, id, start, end;
    protected boolean strand, precursor;
    //Information about the precursor transcript available in some
    //cases from the paper of Landgraf et al (2007)
    protected int precursorStart, precursorEnd;
    protected String name;
    //Keep track of the number of reads, can either be across coding
    //region or just at TSS
    protected TreeMap<String, ABIRead[]> factorReads = new TreeMap<String, ABIRead[]>();

    //////////////////////////////////////////////////////////////////

    public MiRNA(int c, int s, int e, boolean st, String n) {
	chr = c;
	strand = st;
	start = s;
	end = e;
	name = n;
    }

    public MiRNA(int c, int s, int e, boolean st, String n, int pStart, int pEnd) {
	chr = c;
	strand = st;
	start = s;
	end = e;
	name = n;
	precursorStart = pStart;
	precursorEnd = pEnd;
	precursor = true;
    }


    ///////////////////////////////////////////////////////

    public int getStart() { return start; }

    public int getEnd() { return end; }

    public int getPrecursorStart() { return precursorStart; }

    public int getPrecursorEnd() { return precursorEnd; }

    public boolean hasPrecursor() { return precursor; }

    public int getLength() { return Math.abs(start - end); }

    //////////////////////////////////////////////////////////////////
    //Stuff for adding read info

    public void setFactorReads(String tfName, ABIRead[] reads) { factorReads.put(tfName, reads); }

    public ABIRead[] getFactorReads(String tfName) { return factorReads.containsKey(tfName) ? factorReads.get(tfName) : null; }

    public double getReadDensity(String tfName) { return (double)getFactorReads(tfName).length/getLength(); }

    //////////////////////////////////////////////////////////////////

    public boolean getStrand() { return strand; }

    public int getChromosome() { return chr; }


    /////////////////////////////////

    public boolean overlaps(int st, int en) {
	//if (strand)
	//  return en>start && start<end;
	//return st<start && en>end;
	if (strand) {
	    if ((st>start && st<end) || (en>start && en<end)) //Start or en of region inside gene
		return true;
	    else if ((start>st && start<en) || (end>st && end<en)) //Start or en of exon inside region
		return true;
	}
	else {
	    if ((st<start && st>end) || (en<start && en>end)) //Start or en of region inside gene
		return true;
	    else if ((start>st && start<en) || (end>st && end<en)) //Start or en of exon inside region
		return true;
	}
	return false;
    }


    //////////////////////////////////////////////////////////////////
    //Stuff for the JUNG package
    
    //////////////////////////////////////////////////////////////////

    public int compareTo(Object o) {
	//Sort according to chromosomal position
	if (!(o instanceof MiRNA))
	    System.err.println("ERROR comparing miRNA!");
	MiRNA li = (MiRNA)o;
	if (chr<li.chr)
	    return -1;
	else if (chr>li.chr)
	    return 1;
	else {
	    if (getStart()<li.getStart())
		return -1;
	    else if (getStart()>li.getStart())
		return 1;
	    else
		return 0;
	}
    }

    public String toString() {
	return chr + "(" + (strand ? "+" : "-") + ")" + start + "-" + end;
    }

    public boolean equals(Object o) {
	if (!(o instanceof MiRNA))
	    System.err.println("ERROR comparing miRNA!");
	MiRNA mi = (MiRNA)o;
	return mi.chr==chr && mi.start==start && mi.end==end;
    }

    public int hashCode() { return chr*29 + 5*end + 13*start; }

}
